#![no_std]

#[derive(Debug)]
#[cfg_attr(feature = "defmt", derive(defmt::Format))]
pub struct UnknownLocale;

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod POSIX {
    pub mod LC_ADDRESS {
        /// `None`
        pub const COUNTRY_AB3: Option<&str> = None;
        /// `None`
        pub const COUNTRY_ABTWO: Option<&str> = None;
        /// `None`
        pub const COUNTRY_CAR: Option<&str> = None;
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `None`
        pub const COUNTRY_NAME: Option<&str> = None;
        /// `None`
        pub const COUNTRY_NUM: Option<i64> = None;
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `None`
        pub const LANG_LIB: Option<&str> = None;
        /// `None`
        pub const LANG_NAME: Option<&str> = None;
        /// `None`
        pub const LANG_TERM: Option<&str> = None;
        /// `None`
        pub const POSTAL_FMT: Option<&str> = None;
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `None`
        pub const ADDRESS: Option<&str> = None;
        /// `None`
        pub const CATEGORY: Option<&[&[&str]]> = None;
        /// `None`
        pub const CONTACT: Option<&str> = None;
        /// `None`
        pub const DATE: Option<&str> = None;
        /// `None`
        pub const EMAIL: Option<&str> = None;
        /// `None`
        pub const FAX: Option<&str> = None;
        /// `None`
        pub const LANGUAGE: Option<&str> = None;
        /// `None`
        pub const REVISION: Option<&str> = None;
        /// `None`
        pub const SOURCE: Option<&str> = None;
        /// `None`
        pub const TEL: Option<&str> = None;
        /// `None`
        pub const TERRITORY: Option<&str> = None;
        /// `None`
        pub const TITLE: Option<&str> = None;
    }
    pub mod LC_MESSAGES {
        /// `"^[nN]"`
        pub const NOEXPR: &str = "^[nN]";
        /// `Some("No")`
        pub const NOSTR: Option<&str> = Some("No");
        /// `"^[yY]"`
        pub const YESEXPR: &str = "^[yY]";
        /// `Some("Yes")`
        pub const YESSTR: Option<&str> = Some("Yes");
    }
    pub mod LC_MONETARY {
        /// `""`
        pub const CURRENCY_SYMBOL: &str = "";
        /// `-1`
        pub const FRAC_DIGITS: i64 = -1;
        /// `""`
        pub const INT_CURR_SYMBOL: &str = "";
        /// `-1`
        pub const INT_FRAC_DIGITS: i64 = -1;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[-1]`
        pub const MON_GROUPING: &[i64] = &[-1];
        /// `""`
        pub const MON_THOUSANDS_SEP: &str = "";
        /// `""`
        pub const NEGATIVE_SIGN: &str = "";
        /// `-1`
        pub const N_CS_PRECEDES: i64 = -1;
        /// `-1`
        pub const N_SEP_BY_SPACE: i64 = -1;
        /// `-1`
        pub const N_SIGN_POSN: i64 = -1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `-1`
        pub const P_CS_PRECEDES: i64 = -1;
        /// `-1`
        pub const P_SEP_BY_SPACE: i64 = -1;
        /// `-1`
        pub const P_SIGN_POSN: i64 = -1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[-1]`
        pub const GROUPING: &[i64] = &[-1];
        /// `""`
        pub const THOUSANDS_SEP: &str = "";
    }
    pub mod LC_TELEPHONE {
        /// `None`
        pub const INT_PREFIX: Option<&str> = None;
        /// `None`
        pub const INT_SELECT: Option<&str> = None;
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `None`
        pub const TEL_INT_FMT: Option<&str> = None;
    }
    pub mod LC_TIME {
        /// `&["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"]`
        pub const ABDAY: &[&str] = &["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"];
        /// `&["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["AM", "PM"]`
        pub const AM_PM: &[&str] = &["AM", "PM"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a %b %e %H:%M:%S %Z %Y")`
        pub const DATE_FMT: Option<&str> = Some("%a %b %e %H:%M:%S %Z %Y");
        /// `&["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"]`
        pub const DAY: &[&str] = &["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"];
        /// `"%m/%d/%y"`
        pub const D_FMT: &str = "%m/%d/%y";
        /// `"%a %b %e %H:%M:%S %Y"`
        pub const D_T_FMT: &str = "%a %b %e %H:%M:%S %Y";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"]`
        pub const MON: &[&str] = &["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"];
        /// `"%H:%M:%S"`
        pub const T_FMT: &str = "%H:%M:%S";
        /// `"%I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p";
        /// `None`
        pub const WEEK: Option<&[i64]> = None;
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod aa_DJ {
    pub mod LC_ADDRESS {
        /// `Some("DJI")`
        pub const COUNTRY_AB3: Option<&str> = Some("DJI");
        /// `Some("DJ")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("DJ");
        /// `Some("DJI")`
        pub const COUNTRY_CAR: Option<&str> = Some("DJI");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Yabuuti")`
        pub const COUNTRY_NAME: Option<&str> = Some("Yabuuti");
        /// `Some(262)`
        pub const COUNTRY_NUM: Option<i64> = Some(262);
        /// `Some("DJI")`
        pub const COUNTRY_POST: Option<&str> = Some("DJI");
        /// `Some("aa")`
        pub const LANG_AB: Option<&str> = Some("aa");
        /// `Some("aar")`
        pub const LANG_LIB: Option<&str> = Some("aar");
        /// `Some("Qafar")`
        pub const LANG_NAME: Option<&str> = Some("Qafar");
        /// `Some("aar")`
        pub const LANG_TERM: Option<&str> = Some("aar");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("7802 Solomon Seal Dr., Springfield, VA 22152, USA")`
        pub const ADDRESS: Option<&str> = Some("7802 Solomon Seal Dr., Springfield, VA 22152, USA");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_TIME"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_TIME"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2003-07-05")`
        pub const DATE: Option<&str> = Some("2003-07-05");
        /// `Some("locales@geez.org")`
        pub const EMAIL: Option<&str> = Some("locales@geez.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Afar")`
        pub const LANGUAGE: Option<&str> = Some("Afar");
        /// `Some("0.20")`
        pub const REVISION: Option<&str> = Some("0.20");
        /// `Some("Ge'ez Frontier Foundation")`
        pub const SOURCE: Option<&str> = Some("Ge'ez Frontier Foundation");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Djibouti")`
        pub const TERRITORY: Option<&str> = Some("Djibouti");
        /// `Some("Afar language locale for Djibouti (Cadu/Laaqo Dialects).")`
        pub const TITLE: Option<&str> = Some("Afar language locale for Djibouti (Cadu/Laaqo Dialects).");
    }
    pub use super::aa_ET::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"Fdj"`
        pub const CURRENCY_SYMBOL: &str = "Fdj";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"DJF "`
        pub const INT_CURR_SYMBOL: &str = "DJF ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const MON_THOUSANDS_SEP: &str = "\u{202f}";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[0, 0]`
        pub const GROUPING: &[i64] = &[0, 0];
        /// `""`
        pub const THOUSANDS_SEP: &str = "";
    }
    pub mod LC_TELEPHONE {
        /// `Some("253")`
        pub const INT_PREFIX: Option<&str> = Some("253");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `Some("%a-%l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("%a-%l");
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["Aca", "Etl", "Tal", "Arb", "Kam", "Gum", "Sab"]`
        pub const ABDAY: &[&str] = &["Aca", "Etl", "Tal", "Arb", "Kam", "Gum", "Sab"];
        /// `&["Qun", "Nah", "Cig", "Agd", "Cax", "Qas", "Qad", "Leq", "Way", "Dit", "Xim", "Kax"]`
        pub const ABMON: &[&str] = &["Qun", "Nah", "Cig", "Agd", "Cax", "Qas", "Qad", "Leq", "Way", "Dit", "Xim", "Kax"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["saaku", "carra"]`
        pub const AM_PM: &[&str] = &["saaku", "carra"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a %b %e %r %Z %Y")`
        pub const DATE_FMT: Option<&str> = Some("%a %b %e %r %Z %Y");
        /// `&["Acaada", "Etleeni", "Talaata", "Arbaqa", "Kamiisi", "Gumqata", "Sabti"]`
        pub const DAY: &[&str] = &["Acaada", "Etleeni", "Talaata", "Arbaqa", "Kamiisi", "Gumqata", "Sabti"];
        /// `"%d.%m.%Y"`
        pub const D_FMT: &str = "%d.%m.%Y";
        /// `"%a %d %b %Y %l:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %l:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(7)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(7);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Qunxa Garablu", "Naharsi Kudo", "Ciggilta Kudo", "Agda Baxis", "Caxah Alsa", "Qasa Dirri", "Qado Dirri", "Leqeeni", "Waysu", "Diteli", "Ximoli", "Kaxxa Garablu"]`
        pub const MON: &[&str] = &["Qunxa Garablu", "Naharsi Kudo", "Ciggilta Kudo", "Agda Baxis", "Caxah Alsa", "Qasa Dirri", "Qado Dirri", "Leqeeni", "Waysu", "Diteli", "Ximoli", "Kaxxa Garablu"];
        /// `"%l:%M:%S %p"`
        pub const T_FMT: &str = "%l:%M:%S %p";
        /// `"%l:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%l:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod aa_ER {
    pub mod LC_ADDRESS {
        /// `Some("ERI")`
        pub const COUNTRY_AB3: Option<&str> = Some("ERI");
        /// `Some("ER")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("ER");
        /// `Some("ER")`
        pub const COUNTRY_CAR: Option<&str> = Some("ER");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Eretria")`
        pub const COUNTRY_NAME: Option<&str> = Some("Eretria");
        /// `Some(232)`
        pub const COUNTRY_NUM: Option<i64> = Some(232);
        /// `Some("ERI")`
        pub const COUNTRY_POST: Option<&str> = Some("ERI");
        /// `Some("aa")`
        pub const LANG_AB: Option<&str> = Some("aa");
        /// `Some("aar")`
        pub const LANG_LIB: Option<&str> = Some("aar");
        /// `Some("Qafar")`
        pub const LANG_NAME: Option<&str> = Some("Qafar");
        /// `Some("aar")`
        pub const LANG_TERM: Option<&str> = Some("aar");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("7802 Solomon Seal Dr., Springfield, VA 22152, USA")`
        pub const ADDRESS: Option<&str> = Some("7802 Solomon Seal Dr., Springfield, VA 22152, USA");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_TIME"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_TIME"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2003-07-05")`
        pub const DATE: Option<&str> = Some("2003-07-05");
        /// `Some("locales@geez.org")`
        pub const EMAIL: Option<&str> = Some("locales@geez.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Afar")`
        pub const LANGUAGE: Option<&str> = Some("Afar");
        /// `Some("0.20")`
        pub const REVISION: Option<&str> = Some("0.20");
        /// `Some("Ge'ez Frontier Foundation")`
        pub const SOURCE: Option<&str> = Some("Ge'ez Frontier Foundation");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Eritrea")`
        pub const TERRITORY: Option<&str> = Some("Eritrea");
        /// `Some("Afar language locale for Eritrea (Cadu/Laaqo Dialects).")`
        pub const TITLE: Option<&str> = Some("Afar language locale for Eritrea (Cadu/Laaqo Dialects).");
    }
    pub use super::aa_ET::LC_MESSAGES;
    pub use super::ti_ER::LC_MONETARY;
    pub use super::ti_ER::LC_NUMERIC;
    pub use super::ti_ER::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Aca", "Etl", "Tal", "Arb", "Kam", "Gum", "Sab"]`
        pub const ABDAY: &[&str] = &["Aca", "Etl", "Tal", "Arb", "Kam", "Gum", "Sab"];
        /// `&["Qun", "Nah", "Cig", "Agd", "Cax", "Qas", "Qad", "Leq", "Way", "Dit", "Xim", "Kax"]`
        pub const ABMON: &[&str] = &["Qun", "Nah", "Cig", "Agd", "Cax", "Qas", "Qad", "Leq", "Way", "Dit", "Xim", "Kax"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["saaku", "carra"]`
        pub const AM_PM: &[&str] = &["saaku", "carra"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%A, %B %e, %r %Z %Y")`
        pub const DATE_FMT: Option<&str> = Some("%A, %B %e, %r %Z %Y");
        /// `&["Acaada", "Etleeni", "Talaata", "Arbaqa", "Kamiisi", "Gumqata", "Sabti"]`
        pub const DAY: &[&str] = &["Acaada", "Etleeni", "Talaata", "Arbaqa", "Kamiisi", "Gumqata", "Sabti"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%A, %B %e, %Y %l:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A, %B %e, %Y %l:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Qunxa Garablu", "Naharsi Kudo", "Ciggilta Kudo", "Agda Baxis", "Caxah Alsa", "Qasa Dirri", "Qado Dirri", "Liiqen", "Waysu", "Diteli", "Ximoli", "Kaxxa Garablu"]`
        pub const MON: &[&str] = &["Qunxa Garablu", "Naharsi Kudo", "Ciggilta Kudo", "Agda Baxis", "Caxah Alsa", "Qasa Dirri", "Qado Dirri", "Liiqen", "Waysu", "Diteli", "Ximoli", "Kaxxa Garablu"];
        /// `"%l:%M:%S %p"`
        pub const T_FMT: &str = "%l:%M:%S %p";
        /// `"%l:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%l:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod aa_ER_saaho {
    pub use super::aa_ER::LC_ADDRESS;
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("7802 Solomon Seal Dr., Springfield, VA 22152, USA")`
        pub const ADDRESS: Option<&str> = Some("7802 Solomon Seal Dr., Springfield, VA 22152, USA");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_TIME"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_TIME"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2003-07-05")`
        pub const DATE: Option<&str> = Some("2003-07-05");
        /// `Some("locales@geez.org")`
        pub const EMAIL: Option<&str> = Some("locales@geez.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Afar")`
        pub const LANGUAGE: Option<&str> = Some("Afar");
        /// `Some("0.20")`
        pub const REVISION: Option<&str> = Some("0.20");
        /// `Some("Ge'ez Frontier Foundation")`
        pub const SOURCE: Option<&str> = Some("Ge'ez Frontier Foundation");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Eritrea")`
        pub const TERRITORY: Option<&str> = Some("Eritrea");
        /// `Some("Afar language locale for Eritrea (Saaho Dialect).")`
        pub const TITLE: Option<&str> = Some("Afar language locale for Eritrea (Saaho Dialect).");
    }
    pub use super::aa_ET::LC_MESSAGES;
    pub use super::aa_ER::LC_MONETARY;
    pub use super::aa_ER::LC_NUMERIC;
    pub use super::aa_ER::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Nab", "San", "Sal", "Rab", "Cam", "Jum", "Qun"]`
        pub const ABDAY: &[&str] = &["Nab", "San", "Sal", "Rab", "Cam", "Jum", "Qun"];
        /// `&["Qun", "Nah", "Cig", "Agd", "Cax", "Qas", "Qad", "Leq", "Way", "Dit", "Xim", "Kax"]`
        pub const ABMON: &[&str] = &["Qun", "Nah", "Cig", "Agd", "Cax", "Qas", "Qad", "Leq", "Way", "Dit", "Xim", "Kax"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["saaku", "carra"]`
        pub const AM_PM: &[&str] = &["saaku", "carra"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%A, %B %e, %r %Z %Y")`
        pub const DATE_FMT: Option<&str> = Some("%A, %B %e, %r %Z %Y");
        /// `&["Naba Sambat", "Sani", "Salus", "Rabuq", "Camus", "Jumqata", "Qunxa Sambat"]`
        pub const DAY: &[&str] = &["Naba Sambat", "Sani", "Salus", "Rabuq", "Camus", "Jumqata", "Qunxa Sambat"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%A, %B %e, %Y %l:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A, %B %e, %Y %l:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Qunxa Garablu", "Naharsi Kudo", "Ciggilta Kudo", "Agda Baxis", "Caxah Alsa", "Qasa Dirri", "Qado Dirri", "Liiqen", "Waysu", "Diteli", "Ximoli", "Kaxxa Garablu"]`
        pub const MON: &[&str] = &["Qunxa Garablu", "Naharsi Kudo", "Ciggilta Kudo", "Agda Baxis", "Caxah Alsa", "Qasa Dirri", "Qado Dirri", "Liiqen", "Waysu", "Diteli", "Ximoli", "Kaxxa Garablu"];
        /// `"%l:%M:%S %p"`
        pub const T_FMT: &str = "%l:%M:%S %p";
        /// `"%l:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%l:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod aa_ET {
    pub mod LC_ADDRESS {
        /// `Some("ETH")`
        pub const COUNTRY_AB3: Option<&str> = Some("ETH");
        /// `Some("ET")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("ET");
        /// `Some("ETH")`
        pub const COUNTRY_CAR: Option<&str> = Some("ETH");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Otobbia")`
        pub const COUNTRY_NAME: Option<&str> = Some("Otobbia");
        /// `Some(231)`
        pub const COUNTRY_NUM: Option<i64> = Some(231);
        /// `Some("ETH")`
        pub const COUNTRY_POST: Option<&str> = Some("ETH");
        /// `Some("aa")`
        pub const LANG_AB: Option<&str> = Some("aa");
        /// `Some("aar")`
        pub const LANG_LIB: Option<&str> = Some("aar");
        /// `Some("Qafar")`
        pub const LANG_NAME: Option<&str> = Some("Qafar");
        /// `Some("aar")`
        pub const LANG_TERM: Option<&str> = Some("aar");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("7802 Solomon Seal Dr., Springfield, VA 22152, USA")`
        pub const ADDRESS: Option<&str> = Some("7802 Solomon Seal Dr., Springfield, VA 22152, USA");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_TIME"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_TIME"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2003-07-05")`
        pub const DATE: Option<&str> = Some("2003-07-05");
        /// `Some("locales@geez.org")`
        pub const EMAIL: Option<&str> = Some("locales@geez.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Afar")`
        pub const LANGUAGE: Option<&str> = Some("Afar");
        /// `Some("0.20")`
        pub const REVISION: Option<&str> = Some("0.20");
        /// `Some("Ge'ez Frontier Foundation")`
        pub const SOURCE: Option<&str> = Some("Ge'ez Frontier Foundation");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Ethiopia")`
        pub const TERRITORY: Option<&str> = Some("Ethiopia");
        /// `Some("Afar language locale for Ethiopia (Cadu/Carra Dialects).")`
        pub const TITLE: Option<&str> = Some("Afar language locale for Ethiopia (Cadu/Carra Dialects).");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0mnMN]"`
        pub const NOEXPR: &str = "^[-0mnMN]";
        /// `Some("Maleey")`
        pub const NOSTR: Option<&str> = Some("Maleey");
        /// `"^[+1yY]"`
        pub const YESEXPR: &str = "^[+1yY]";
        /// `Some("Yeey")`
        pub const YESSTR: Option<&str> = Some("Yeey");
    }
    pub use super::ti_ET::LC_MONETARY;
    pub use super::ti_ET::LC_NUMERIC;
    pub use super::ti_ET::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Aca", "Etl", "Tal", "Arb", "Kam", "Gum", "Sab"]`
        pub const ABDAY: &[&str] = &["Aca", "Etl", "Tal", "Arb", "Kam", "Gum", "Sab"];
        /// `&["Qun", "Nah", "Cig", "Agd", "Cax", "Qas", "Qad", "Leq", "Way", "Dit", "Xim", "Kax"]`
        pub const ABMON: &[&str] = &["Qun", "Nah", "Cig", "Agd", "Cax", "Qas", "Qad", "Leq", "Way", "Dit", "Xim", "Kax"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["saaku", "carra"]`
        pub const AM_PM: &[&str] = &["saaku", "carra"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%A, %B %e, %r %Z %Y")`
        pub const DATE_FMT: Option<&str> = Some("%A, %B %e, %r %Z %Y");
        /// `&["Acaada", "Etleeni", "Talaata", "Arbaqa", "Kamiisi", "Gumqata", "Sabti"]`
        pub const DAY: &[&str] = &["Acaada", "Etleeni", "Talaata", "Arbaqa", "Kamiisi", "Gumqata", "Sabti"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%A, %B %e, %Y %l:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A, %B %e, %Y %l:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Qunxa Garablu", "Naharsi Kudo", "Ciggilta Kudo", "Agda Baxis", "Caxah Alsa", "Qasa Dirri", "Qado Dirri", "Liiqen", "Waysu", "Diteli", "Ximoli", "Kaxxa Garablu"]`
        pub const MON: &[&str] = &["Qunxa Garablu", "Naharsi Kudo", "Ciggilta Kudo", "Agda Baxis", "Caxah Alsa", "Qasa Dirri", "Qado Dirri", "Liiqen", "Waysu", "Diteli", "Ximoli", "Kaxxa Garablu"];
        /// `"%l:%M:%S %p"`
        pub const T_FMT: &str = "%l:%M:%S %p";
        /// `"%l:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%l:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod af_ZA {
    pub mod LC_ADDRESS {
        /// `Some("ZAF")`
        pub const COUNTRY_AB3: Option<&str> = Some("ZAF");
        /// `Some("ZA")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("ZA");
        /// `Some("ZA")`
        pub const COUNTRY_CAR: Option<&str> = Some("ZA");
        /// `Some("0")`
        pub const COUNTRY_ISBN: Option<&str> = Some("0");
        /// `Some("Suid-Afrika")`
        pub const COUNTRY_NAME: Option<&str> = Some("Suid-Afrika");
        /// `Some(710)`
        pub const COUNTRY_NUM: Option<i64> = Some(710);
        /// `Some("ZA")`
        pub const COUNTRY_POST: Option<&str> = Some("ZA");
        /// `Some("af")`
        pub const LANG_AB: Option<&str> = Some("af");
        /// `Some("afr")`
        pub const LANG_LIB: Option<&str> = Some("afr");
        /// `Some("Afrikaans")`
        pub const LANG_NAME: Option<&str> = Some("Afrikaans");
        /// `Some("afr")`
        pub const LANG_TERM: Option<&str> = Some("afr");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Box 28364, Sunnyside, 0132, South Africa")`
        pub const ADDRESS: Option<&str> = Some("Box 28364, Sunnyside, 0132, South Africa");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Dwayne Bailey")`
        pub const CONTACT: Option<&str> = Some("Dwayne Bailey");
        /// `Some("2005-10-13")`
        pub const DATE: Option<&str> = Some("2005-10-13");
        /// `Some("dwayne@translate.org.za")`
        pub const EMAIL: Option<&str> = Some("dwayne@translate.org.za");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Afrikaans")`
        pub const LANGUAGE: Option<&str> = Some("Afrikaans");
        /// `Some("1.2.1")`
        pub const REVISION: Option<&str> = Some("1.2.1");
        /// `Some("Zuza Software Foundation (Translate.org.za)")`
        pub const SOURCE: Option<&str> = Some("Zuza Software Foundation (Translate.org.za)");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("South Africa")`
        pub const TERRITORY: Option<&str> = Some("South Africa");
        /// `Some("Afrikaans locale for South Africa")`
        pub const TITLE: Option<&str> = Some("Afrikaans locale for South Africa");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("nee")`
        pub const NOSTR: Option<&str> = Some("nee");
        /// `"^[+1jJyY]"`
        pub const YESEXPR: &str = "^[+1jJyY]";
        /// `Some("ja")`
        pub const YESSTR: Option<&str> = Some("ja");
    }
    pub use super::en_ZA::LC_MONETARY;
    pub use super::en_ZA::LC_NUMERIC;
    pub use super::en_ZA::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["So", "Ma", "Di", "Wo", "Do", "Vr", "Sa"]`
        pub const ABDAY: &[&str] = &["So", "Ma", "Di", "Wo", "Do", "Vr", "Sa"];
        /// `&["Jan", "Feb", "Mrt", "Apr", "Mei", "Jun", "Jul", "Aug", "Sep", "Okt", "Nov", "Des"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Mrt", "Apr", "Mei", "Jun", "Jul", "Aug", "Sep", "Okt", "Nov", "Des"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["VM", "NM"]`
        pub const AM_PM: &[&str] = &["VM", "NM"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Sondag", "Maandag", "Dinsdag", "Woensdag", "Donderdag", "Vrydag", "Saterdag"]`
        pub const DAY: &[&str] = &["Sondag", "Maandag", "Dinsdag", "Woensdag", "Donderdag", "Vrydag", "Saterdag"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Januarie", "Februarie", "Maart", "April", "Mei", "Junie", "Julie", "Augustus", "September", "Oktober", "November", "Desember"]`
        pub const MON: &[&str] = &["Januarie", "Februarie", "Maart", "April", "Mei", "Junie", "Julie", "Augustus", "September", "Oktober", "November", "Desember"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `"%l:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%l:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod agr_PE {
    pub mod LC_ADDRESS {
        /// `Some("PER")`
        pub const COUNTRY_AB3: Option<&str> = Some("PER");
        /// `Some("PE")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("PE");
        /// `Some("PE")`
        pub const COUNTRY_CAR: Option<&str> = Some("PE");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Perú")`
        pub const COUNTRY_NAME: Option<&str> = Some("Perú");
        /// `Some(604)`
        pub const COUNTRY_NUM: Option<i64> = Some(604);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("")`
        pub const LANG_AB: Option<&str> = Some("");
        /// `Some("agr")`
        pub const LANG_LIB: Option<&str> = Some("agr");
        /// `Some("Awajún")`
        pub const LANG_NAME: Option<&str> = Some("Awajún");
        /// `Some("")`
        pub const LANG_TERM: Option<&str> = Some("");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2016-08-20")`
        pub const DATE: Option<&str> = Some("2016-08-20");
        /// `Some("libc-alpha@sourceware.org")`
        pub const EMAIL: Option<&str> = Some("libc-alpha@sourceware.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Aguaruna")`
        pub const LANGUAGE: Option<&str> = Some("Aguaruna");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("somosazucar.org")`
        pub const SOURCE: Option<&str> = Some("somosazucar.org");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Peru")`
        pub const TERRITORY: Option<&str> = Some("Peru");
        /// `Some("Awajún (agr) locale for Peru")`
        pub const TITLE: Option<&str> = Some("Awajún (agr) locale for Peru");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0aAnN]"`
        pub const NOEXPR: &str = "^[-0aAnN]";
        /// `Some("atsa")`
        pub const NOSTR: Option<&str> = Some("atsa");
        /// `"^[+1eEsSyY]"`
        pub const YESEXPR: &str = "^[+1eEsSyY]";
        /// `Some("ehe")`
        pub const YESSTR: Option<&str> = Some("ehe");
    }
    pub use super::es_PE::LC_MONETARY;
    pub use super::es_PE::LC_NUMERIC;
    pub use super::es_PE::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Tun", "Ach", "Kug", "Sak", "Shim", "Im", "Bat"]`
        pub const ABDAY: &[&str] = &["Tun", "Ach", "Kug", "Sak", "Shim", "Im", "Bat"];
        /// `&["Pet", "Kup", "Uya", "Tay", "Keg", "Teg", "Kun", "Yag", "Dait", "Ipam", "Shin", "Sak"]`
        pub const ABMON: &[&str] = &["Pet", "Kup", "Uya", "Tay", "Keg", "Teg", "Kun", "Yag", "Dait", "Ipam", "Shin", "Sak"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["VM", "NM"]`
        pub const AM_PM: &[&str] = &["VM", "NM"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a %b %e %H:%M:%S %Z %Y")`
        pub const DATE_FMT: Option<&str> = Some("%a %b %e %H:%M:%S %Z %Y");
        /// `&["Tuntuamtin", "Achutin", "Kugkuktin", "Saketin", "Shimpitin", "Imaptin", "Bataetin"]`
        pub const DAY: &[&str] = &["Tuntuamtin", "Achutin", "Kugkuktin", "Saketin", "Shimpitin", "Imaptin", "Bataetin"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(1)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(1);
        /// `Some(2)`
        pub const FIRST_WORKDAY: Option<i64> = Some(2);
        /// `&["Petsatin", "Kupitin", "Uyaitin", "Tayutin", "Kegketin", "Tegmatin", "Kuntutin", "Yagkujutin", "Daiktatin", "Ipamtatin", "Shinutin", "Sakamtin"]`
        pub const MON: &[&str] = &["Petsatin", "Kupitin", "Uyaitin", "Tayutin", "Kegketin", "Tegmatin", "Kuntutin", "Yagkujutin", "Daiktatin", "Ipamtatin", "Shinutin", "Sakamtin"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `"%I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p";
        /// `Some(&[7, 19971130, 7])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 7]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ak_GH {
    pub mod LC_ADDRESS {
        /// `Some("GHA")`
        pub const COUNTRY_AB3: Option<&str> = Some("GHA");
        /// `Some("GH")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("GH");
        /// `Some("GH")`
        pub const COUNTRY_CAR: Option<&str> = Some("GH");
        /// `Some("9964")`
        pub const COUNTRY_ISBN: Option<&str> = Some("9964");
        /// `Some("Gaana")`
        pub const COUNTRY_NAME: Option<&str> = Some("Gaana");
        /// `Some(288)`
        pub const COUNTRY_NUM: Option<i64> = Some(288);
        /// `Some("GH")`
        pub const COUNTRY_POST: Option<&str> = Some("GH");
        /// `Some("ak")`
        pub const LANG_AB: Option<&str> = Some("ak");
        /// `None`
        pub const LANG_LIB: Option<&str> = None;
        /// `Some("Akan")`
        pub const LANG_NAME: Option<&str> = Some("Akan");
        /// `Some("aka")`
        pub const LANG_TERM: Option<&str> = Some("aka");
        /// `Some("%a%N%f%N%d%N%b%N%h %s %e %r%N%T, %S %z%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%a%N%f%N%d%N%b%N%h %s %e %r%N%T, %S %z%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
        ]);
        /// `Some("sugarlabs.org")`
        pub const CONTACT: Option<&str> = Some("sugarlabs.org");
        /// `Some("2013-08-24")`
        pub const DATE: Option<&str> = Some("2013-08-24");
        /// `Some("libc-alpha@sourceware.org")`
        pub const EMAIL: Option<&str> = Some("libc-alpha@sourceware.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Akan")`
        pub const LANGUAGE: Option<&str> = Some("Akan");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Sugar Labs / OLPC")`
        pub const SOURCE: Option<&str> = Some("Sugar Labs / OLPC");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Ghana")`
        pub const TERRITORY: Option<&str> = Some("Ghana");
        /// `Some("Akan locale for Ghana")`
        pub const TITLE: Option<&str> = Some("Akan locale for Ghana");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0dDnN]"`
        pub const NOEXPR: &str = "^[-0dDnN]";
        /// `Some("Daabi")`
        pub const NOSTR: Option<&str> = Some("Daabi");
        /// `"^[+1yY]"`
        pub const YESEXPR: &str = "^[+1yY]";
        /// `Some("Yiw")`
        pub const YESSTR: Option<&str> = Some("Yiw");
    }
    pub mod LC_MONETARY {
        /// `"GH₵"`
        pub const CURRENCY_SYMBOL: &str = "GH₵";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"GHS "`
        pub const INT_CURR_SYMBOL: &str = "GHS ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("233")`
        pub const INT_PREFIX: Option<&str> = Some("233");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `Some("%a %l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("%a %l");
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["Kwe", "Dwo", "Ben", "Wuk", "Yaw", "Fia", "Mem"]`
        pub const ABDAY: &[&str] = &["Kwe", "Dwo", "Ben", "Wuk", "Yaw", "Fia", "Mem"];
        /// `&["S-Ɔ", "K-Ɔ", "E-Ɔ", "E-O", "E-K", "O-A", "A-K", "D-Ɔ", "F-Ɛ", "Ɔ-A", "Ɔ-O", "M-Ɔ"]`
        pub const ABMON: &[&str] = &["S-Ɔ", "K-Ɔ", "E-Ɔ", "E-O", "E-K", "O-A", "A-K", "D-Ɔ", "F-Ɛ", "Ɔ-A", "Ɔ-O", "M-Ɔ"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["AN", "EW"]`
        pub const AM_PM: &[&str] = &["AN", "EW"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Kwesida", "Dwowda", "Benada", "Wukuda", "Yawda", "Fida", "Memeneda"]`
        pub const DAY: &[&str] = &["Kwesida", "Dwowda", "Benada", "Wukuda", "Yawda", "Fida", "Memeneda"];
        /// `"%Y/%m/%d"`
        pub const D_FMT: &str = "%Y/%m/%d";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Sanda-Ɔpɛpɔn", "Kwakwar-Ɔgyefuo", "Ebɔw-Ɔbenem", "Ebɔbira-Oforisuo", "Esusow Aketseaba-Kɔtɔnimba", "Obirade-Ayɛwohomumu", "Ayɛwoho-Kitawonsa", "Difuu-Ɔsandaa", "Fankwa-Ɛbɔ", "Ɔbɛsɛ-Ahinime", "Ɔberɛfɛw-Obubuo", "Mumu-Ɔpɛnimba"]`
        pub const MON: &[&str] = &["Sanda-Ɔpɛpɔn", "Kwakwar-Ɔgyefuo", "Ebɔw-Ɔbenem", "Ebɔbira-Oforisuo", "Esusow Aketseaba-Kɔtɔnimba", "Obirade-Ayɛwohomumu", "Ayɛwoho-Kitawonsa", "Difuu-Ɔsandaa", "Fankwa-Ɛbɔ", "Ɔbɛsɛ-Ahinime", "Ɔberɛfɛw-Obubuo", "Mumu-Ɔpɛnimba"];
        /// `"%I:%M:%S %p"`
        pub const T_FMT: &str = "%I:%M:%S %p";
        /// `"%I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod am_ET {
    pub mod LC_ADDRESS {
        /// `Some("ETH")`
        pub const COUNTRY_AB3: Option<&str> = Some("ETH");
        /// `Some("ET")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("ET");
        /// `Some("ETH")`
        pub const COUNTRY_CAR: Option<&str> = Some("ETH");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("ኢትዮጵያ")`
        pub const COUNTRY_NAME: Option<&str> = Some("ኢትዮጵያ");
        /// `Some(231)`
        pub const COUNTRY_NUM: Option<i64> = Some(231);
        /// `Some("ETH")`
        pub const COUNTRY_POST: Option<&str> = Some("ETH");
        /// `Some("am")`
        pub const LANG_AB: Option<&str> = Some("am");
        /// `Some("amh")`
        pub const LANG_LIB: Option<&str> = Some("amh");
        /// `Some("አማርኛ")`
        pub const LANG_NAME: Option<&str> = Some("አማርኛ");
        /// `Some("amh")`
        pub const LANG_TERM: Option<&str> = Some("amh");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("7802 Solomon Seal Dr., Springfield, VA 22152, USA")`
        pub const ADDRESS: Option<&str> = Some("7802 Solomon Seal Dr., Springfield, VA 22152, USA");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_TIME"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_TIME"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2003-07-05")`
        pub const DATE: Option<&str> = Some("2003-07-05");
        /// `Some("locales@geez.org")`
        pub const EMAIL: Option<&str> = Some("locales@geez.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Amharic")`
        pub const LANGUAGE: Option<&str> = Some("Amharic");
        /// `Some("0.20")`
        pub const REVISION: Option<&str> = Some("0.20");
        /// `Some("Ge'ez Frontier Foundation")`
        pub const SOURCE: Option<&str> = Some("Ge'ez Frontier Foundation");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Ethiopia")`
        pub const TERRITORY: Option<&str> = Some("Ethiopia");
        /// `Some("Amharic language locale for Ethiopia.")`
        pub const TITLE: Option<&str> = Some("Amharic language locale for Ethiopia.");
    }
    pub mod LC_MESSAGES {
        /// `"^([-0nNይ]|አይ)"`
        pub const NOEXPR: &str = "^([-0nNይ]|አይ)";
        /// `Some("አይ")`
        pub const NOSTR: Option<&str> = Some("አይ");
        /// `"^([+1yYዎ]|አዎን)"`
        pub const YESEXPR: &str = "^([+1yYዎ]|አዎን)";
        /// `Some("አዎን")`
        pub const YESSTR: Option<&str> = Some("አዎን");
    }
    pub use super::ti_ET::LC_MONETARY;
    pub use super::ti_ET::LC_NUMERIC;
    pub use super::ti_ET::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["እሑድ", "ሰኞ ", "ማክሰ", "ረቡዕ", "ሐሙስ", "ዓርብ", "ቅዳሜ"]`
        pub const ABDAY: &[&str] = &["እሑድ", "ሰኞ ", "ማክሰ", "ረቡዕ", "ሐሙስ", "ዓርብ", "ቅዳሜ"];
        /// `&["ጃንዩ", "ፌብሩ", "ማርች", "ኤፕረ", "ሜይ ", "ጁን ", "ጁላይ", "ኦገስ", "ሴፕቴ", "ኦክተ", "ኖቬም", "ዲሴም"]`
        pub const ABMON: &[&str] = &["ጃንዩ", "ፌብሩ", "ማርች", "ኤፕረ", "ሜይ ", "ጁን ", "ጁላይ", "ኦገስ", "ሴፕቴ", "ኦክተ", "ኖቬም", "ዲሴም"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["ጡዋት", "ከሰዓት"]`
        pub const AM_PM: &[&str] = &["ጡዋት", "ከሰዓት"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%A፣ %B %e ቀን %r %Z %Y ዓ/ም")`
        pub const DATE_FMT: Option<&str> = Some("%A፣ %B %e ቀን %r %Z %Y ዓ/ም");
        /// `&["እሑድ", "ሰኞ", "ማክሰኞ", "ረቡዕ", "ሐሙስ", "ዓርብ", "ቅዳሜ"]`
        pub const DAY: &[&str] = &["እሑድ", "ሰኞ", "ማክሰኞ", "ረቡዕ", "ሐሙስ", "ዓርብ", "ቅዳሜ"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%A፣ %B %e ቀን %Y %l:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A፣ %B %e ቀን %Y %l:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["ጃንዩወሪ", "ፌብሩወሪ", "ማርች", "ኤፕሪል", "ሜይ", "ጁን", "ጁላይ", "ኦገስት", "ሴፕቴምበር", "ኦክቶበር", "ኖቬምበር", "ዲሴምበር"]`
        pub const MON: &[&str] = &["ጃንዩወሪ", "ፌብሩወሪ", "ማርች", "ኤፕሪል", "ሜይ", "ጁን", "ጁላይ", "ኦገስት", "ሴፕቴምበር", "ኦክቶበር", "ኖቬምበር", "ዲሴምበር"];
        /// `"%l:%M:%S %p"`
        pub const T_FMT: &str = "%l:%M:%S %p";
        /// `"%l:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%l:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod an_ES {
    pub mod LC_ADDRESS {
        /// `Some("ESP")`
        pub const COUNTRY_AB3: Option<&str> = Some("ESP");
        /// `Some("ES")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("ES");
        /// `Some("E")`
        pub const COUNTRY_CAR: Option<&str> = Some("E");
        /// `Some("978-84")`
        pub const COUNTRY_ISBN: Option<&str> = Some("978-84");
        /// `Some("Espanya")`
        pub const COUNTRY_NAME: Option<&str> = Some("Espanya");
        /// `Some(724)`
        pub const COUNTRY_NUM: Option<i64> = Some(724);
        /// `Some("E")`
        pub const COUNTRY_POST: Option<&str> = Some("E");
        /// `Some("an")`
        pub const LANG_AB: Option<&str> = Some("an");
        /// `Some("arg")`
        pub const LANG_LIB: Option<&str> = Some("arg");
        /// `Some("aragonés")`
        pub const LANG_NAME: Option<&str> = Some("aragonés");
        /// `Some("arg")`
        pub const LANG_TERM: Option<&str> = Some("arg");
        /// `Some("%a%N%f%N%d%N%b%N%s %h %e %r%N%z %T, %S%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%a%N%f%N%d%N%b%N%s %h %e %r%N%z %T, %S%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Jordi Mallach Pérez, Juan Pablo Martínez")`
        pub const CONTACT: Option<&str> = Some("Jordi Mallach Pérez, Juan Pablo Martínez");
        /// `Some("2018-02-05")`
        pub const DATE: Option<&str> = Some("2018-02-05");
        /// `Some("bug-glibc-locales@gnu.org, softaragones@softaragones.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org, softaragones@softaragones.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Aragonese")`
        pub const LANGUAGE: Option<&str> = Some("Aragonese");
        /// `Some("1.2")`
        pub const REVISION: Option<&str> = Some("1.2");
        /// `Some("Softaragones")`
        pub const SOURCE: Option<&str> = Some("Softaragones");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Spain")`
        pub const TERRITORY: Option<&str> = Some("Spain");
        /// `Some("Aragonese locale for Spain")`
        pub const TITLE: Option<&str> = Some("Aragonese locale for Spain");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("no")`
        pub const NOSTR: Option<&str> = Some("no");
        /// `"^[+1sSyY]"`
        pub const YESEXPR: &str = "^[+1sSyY]";
        /// `Some("sí")`
        pub const YESSTR: Option<&str> = Some("sí");
    }
    pub use super::es_ES::LC_MONETARY;
    pub use super::es_ES::LC_NUMERIC;
    pub use super::es_ES::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["dom", "lun", "mar", "mie", "chu", "vie", "sab"]`
        pub const ABDAY: &[&str] = &["dom", "lun", "mar", "mie", "chu", "vie", "sab"];
        /// `&["chi", "feb", "mar", "abr", "may", "chn", "chl", "ago", "set", "oct", "nov", "avi"]`
        pub const ABMON: &[&str] = &["chi", "feb", "mar", "abr", "may", "chn", "chl", "ago", "set", "oct", "nov", "avi"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["domingo", "luns", "martes", "mierques", "chueves", "viernes", "sabado"]`
        pub const DAY: &[&str] = &["domingo", "luns", "martes", "mierques", "chueves", "viernes", "sabado"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["chinero", "febrero", "marzo", "abril", "mayo", "chunyo", "chuliol", "agosto", "setiembre", "octubre", "noviembre", "aviento"]`
        pub const MON: &[&str] = &["chinero", "febrero", "marzo", "abril", "mayo", "chunyo", "chuliol", "agosto", "setiembre", "octubre", "noviembre", "aviento"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod anp_IN {
    pub mod LC_ADDRESS {
        /// `Some("IND")`
        pub const COUNTRY_AB3: Option<&str> = Some("IND");
        /// `Some("IN")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IN");
        /// `Some("IND")`
        pub const COUNTRY_CAR: Option<&str> = Some("IND");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("भारत")`
        pub const COUNTRY_NAME: Option<&str> = Some("भारत");
        /// `Some(356)`
        pub const COUNTRY_NUM: Option<i64> = Some(356);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("anp")`
        pub const LANG_LIB: Option<&str> = Some("anp");
        /// `Some("अ\u{902}गिका")`
        pub const LANG_NAME: Option<&str> = Some("अ\u{902}गिका");
        /// `Some("anp")`
        pub const LANG_TERM: Option<&str> = Some("anp");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2013-10-24")`
        pub const DATE: Option<&str> = Some("2013-10-24");
        /// `Some("bhashaghar@googlegroups.com")`
        pub const EMAIL: Option<&str> = Some("bhashaghar@googlegroups.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Angika")`
        pub const LANGUAGE: Option<&str> = Some("Angika");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("India")`
        pub const TERRITORY: Option<&str> = Some("India");
        /// `Some("Angika language locale for India")`
        pub const TITLE: Option<&str> = Some("Angika language locale for India");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0नइnN]"`
        pub const NOEXPR: &str = "^[-0नइnN]";
        /// `Some("नही\u{902}")`
        pub const NOSTR: Option<&str> = Some("नही\u{902}");
        /// `"^[+1हवyY]"`
        pub const YESEXPR: &str = "^[+1हवyY]";
        /// `Some("हा\u{901}")`
        pub const YESSTR: Option<&str> = Some("हा\u{901}");
    }
    pub use super::hi_IN::LC_MONETARY;
    pub use super::hi_IN::LC_NUMERIC;
    pub use super::hi_IN::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["रवि", "सोम", "म\u{902}गल", "ब\u{941}ध", "ब\u{943}हस\u{94d}पति", "श\u{941}क\u{94d}र", "शनि"]`
        pub const ABDAY: &[&str] = &["रवि", "सोम", "म\u{902}गल", "ब\u{941}ध", "ब\u{943}हस\u{94d}पति", "श\u{941}क\u{94d}र", "शनि"];
        /// `&["जनवरी", "फरवरी", "मार\u{94d}च", "अप\u{94d}र\u{948}ल", "मई", "ज\u{942}न", "ज\u{941}लाई", "अगस\u{94d}त", "सित\u{902}बर", "अक\u{94d}ट\u{942}बर", "नव\u{902}बर", "दिस\u{902}बर"]`
        pub const ABMON: &[&str] = &["जनवरी", "फरवरी", "मार\u{94d}च", "अप\u{94d}र\u{948}ल", "मई", "ज\u{942}न", "ज\u{941}लाई", "अगस\u{94d}त", "सित\u{902}बर", "अक\u{94d}ट\u{942}बर", "नव\u{902}बर", "दिस\u{902}बर"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["प\u{942}र\u{94d}वाह\u{94d}न", "अपराह\u{94d}न"]`
        pub const AM_PM: &[&str] = &["प\u{942}र\u{94d}वाह\u{94d}न", "अपराह\u{94d}न"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["रविवार", "सोमवार", "म\u{902}गलवार", "ब\u{941}धवार", "ब\u{943}हस\u{94d}पतिवार", "श\u{941}क\u{94d}रवार", "शनिवार"]`
        pub const DAY: &[&str] = &["रविवार", "सोमवार", "म\u{902}गलवार", "ब\u{941}धवार", "ब\u{943}हस\u{94d}पतिवार", "श\u{941}क\u{94d}रवार", "शनिवार"];
        /// `"%-d/%-m/%y"`
        pub const D_FMT: &str = "%-d/%-m/%y";
        /// `"%A %d %b %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A %d %b %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["जनवरी", "फरवरी", "मार\u{94d}च", "अप\u{94d}र\u{948}ल", "मई", "ज\u{942}न", "ज\u{941}लाई", "अगस\u{94d}त", "सित\u{902}बर", "अक\u{94d}ट\u{942}बर", "नव\u{902}बर", "दिस\u{902}बर"]`
        pub const MON: &[&str] = &["जनवरी", "फरवरी", "मार\u{94d}च", "अप\u{94d}र\u{948}ल", "मई", "ज\u{942}न", "ज\u{941}लाई", "अगस\u{94d}त", "सित\u{902}बर", "अक\u{94d}ट\u{942}बर", "नव\u{902}बर", "दिस\u{902}बर"];
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT: &str = "%I:%M:%S %p %Z";
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p %Z";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ar_AE {
    pub mod LC_ADDRESS {
        /// `Some("ARE")`
        pub const COUNTRY_AB3: Option<&str> = Some("ARE");
        /// `Some("AE")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("AE");
        /// `Some("UAE")`
        pub const COUNTRY_CAR: Option<&str> = Some("UAE");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("الإمارات العربية المتحدة")`
        pub const COUNTRY_NAME: Option<&str> = Some("الإمارات العربية المتحدة");
        /// `Some(784)`
        pub const COUNTRY_NUM: Option<i64> = Some(784);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ar")`
        pub const LANG_AB: Option<&str> = Some("ar");
        /// `Some("ara")`
        pub const LANG_LIB: Option<&str> = Some("ara");
        /// `Some("العربية")`
        pub const LANG_NAME: Option<&str> = Some("العربية");
        /// `Some("ara")`
        pub const LANG_TERM: Option<&str> = Some("ara");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan")`
        pub const ADDRESS: Option<&str> = Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-07-20")`
        pub const DATE: Option<&str> = Some("2000-07-20");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Arabic")`
        pub const LANGUAGE: Option<&str> = Some("Arabic");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("IBM Globalization Center of Competency, Yamato Software Laboratory")`
        pub const SOURCE: Option<&str> = Some("IBM Globalization Center of Competency, Yamato Software Laboratory");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("United Arab Emirates")`
        pub const TERRITORY: Option<&str> = Some("United Arab Emirates");
        /// `Some("Arabic language locale for United Arab Emirates")`
        pub const TITLE: Option<&str> = Some("Arabic language locale for United Arab Emirates");
    }
    pub use super::ar_EG::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"د.إ."`
        pub const CURRENCY_SYMBOL: &str = "د.إ.";
        /// `3`
        pub const FRAC_DIGITS: i64 = 3;
        /// `"AED "`
        pub const INT_CURR_SYMBOL: &str = "AED ";
        /// `3`
        pub const INT_FRAC_DIGITS: i64 = 3;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `2`
        pub const N_SIGN_POSN: i64 = 2;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("971")`
        pub const INT_PREFIX: Option<&str> = Some("971");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c ;%a ;%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c ;%a ;%l");
    }
    pub mod LC_TIME {
        /// `&["ح", "ن", "ث", "ر", "خ", "ج", "س"]`
        pub const ABDAY: &[&str] = &["ح", "ن", "ث", "ر", "خ", "ج", "س"];
        /// `&["ينا", "فبر", "مار", "أبر", "ماي", "يون", "يول", "أغس", "سبت", "أكت", "نوف", "ديس"]`
        pub const ABMON: &[&str] = &["ينا", "فبر", "مار", "أبر", "ماي", "يون", "يول", "أغس", "سبت", "أكت", "نوف", "ديس"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["ص", "م"]`
        pub const AM_PM: &[&str] = &["ص", "م"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت "]`
        pub const DAY: &[&str] = &["الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت "];
        /// `"%d %b, %Y"`
        pub const D_FMT: &str = "%d %b, %Y";
        /// `"%d %b, %Y %Z %I:%M:%S %p"`
        pub const D_T_FMT: &str = "%d %b, %Y %Z %I:%M:%S %p";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(7)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(7);
        /// `Some(1)`
        pub const FIRST_WORKDAY: Option<i64> = Some(1);
        /// `&["يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"]`
        pub const MON: &[&str] = &["يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"];
        /// `"%Z %I:%M:%S %p"`
        pub const T_FMT: &str = "%Z %I:%M:%S %p";
        /// `"%Z %I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%Z %I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ar_BH {
    pub mod LC_ADDRESS {
        /// `Some("BHR")`
        pub const COUNTRY_AB3: Option<&str> = Some("BHR");
        /// `Some("BH")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("BH");
        /// `Some("BRN")`
        pub const COUNTRY_CAR: Option<&str> = Some("BRN");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("البحرين")`
        pub const COUNTRY_NAME: Option<&str> = Some("البحرين");
        /// `Some(48)`
        pub const COUNTRY_NUM: Option<i64> = Some(48);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ar")`
        pub const LANG_AB: Option<&str> = Some("ar");
        /// `Some("ara")`
        pub const LANG_LIB: Option<&str> = Some("ara");
        /// `Some("العربية")`
        pub const LANG_NAME: Option<&str> = Some("العربية");
        /// `Some("ara")`
        pub const LANG_TERM: Option<&str> = Some("ara");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan")`
        pub const ADDRESS: Option<&str> = Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-07-20")`
        pub const DATE: Option<&str> = Some("2000-07-20");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Arabic")`
        pub const LANGUAGE: Option<&str> = Some("Arabic");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("IBM Globalization Center of Competency, Yamato Software Laboratory")`
        pub const SOURCE: Option<&str> = Some("IBM Globalization Center of Competency, Yamato Software Laboratory");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Bahrain")`
        pub const TERRITORY: Option<&str> = Some("Bahrain");
        /// `Some("Arabic language locale for Bahrain")`
        pub const TITLE: Option<&str> = Some("Arabic language locale for Bahrain");
    }
    pub use super::ar_EG::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"د.ب."`
        pub const CURRENCY_SYMBOL: &str = "د.ب.";
        /// `3`
        pub const FRAC_DIGITS: i64 = 3;
        /// `"BHD "`
        pub const INT_CURR_SYMBOL: &str = "BHD ";
        /// `3`
        pub const INT_FRAC_DIGITS: i64 = 3;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `2`
        pub const N_SIGN_POSN: i64 = 2;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("973")`
        pub const INT_PREFIX: Option<&str> = Some("973");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c ;%a ;%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c ;%a ;%l");
    }
    pub mod LC_TIME {
        /// `&["ح", "ن", "ث", "ر", "خ", "ج", "س"]`
        pub const ABDAY: &[&str] = &["ح", "ن", "ث", "ر", "خ", "ج", "س"];
        /// `&["ينا", "فبر", "مار", "أبر", "ماي", "يون", "يول", "أغس", "سبت", "أكت", "نوف", "ديس"]`
        pub const ABMON: &[&str] = &["ينا", "فبر", "مار", "أبر", "ماي", "يون", "يول", "أغس", "سبت", "أكت", "نوف", "ديس"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["ص", "م"]`
        pub const AM_PM: &[&str] = &["ص", "م"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"]`
        pub const DAY: &[&str] = &["الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"];
        /// `"%d %b, %Y"`
        pub const D_FMT: &str = "%d %b, %Y";
        /// `"%d %b, %Y %Z %I:%M:%S %p"`
        pub const D_T_FMT: &str = "%d %b, %Y %Z %I:%M:%S %p";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(7)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(7);
        /// `Some(1)`
        pub const FIRST_WORKDAY: Option<i64> = Some(1);
        /// `&["يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"]`
        pub const MON: &[&str] = &["يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"];
        /// `"%Z %I:%M:%S %p"`
        pub const T_FMT: &str = "%Z %I:%M:%S %p";
        /// `"%Z %I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%Z %I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ar_DZ {
    pub mod LC_ADDRESS {
        /// `Some("DZA")`
        pub const COUNTRY_AB3: Option<&str> = Some("DZA");
        /// `Some("DZ")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("DZ");
        /// `Some("DZ")`
        pub const COUNTRY_CAR: Option<&str> = Some("DZ");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("الجزائر")`
        pub const COUNTRY_NAME: Option<&str> = Some("الجزائر");
        /// `Some(12)`
        pub const COUNTRY_NUM: Option<i64> = Some(12);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ar")`
        pub const LANG_AB: Option<&str> = Some("ar");
        /// `Some("ara")`
        pub const LANG_LIB: Option<&str> = Some("ara");
        /// `Some("العربية")`
        pub const LANG_NAME: Option<&str> = Some("العربية");
        /// `Some("ara")`
        pub const LANG_TERM: Option<&str> = Some("ara");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan")`
        pub const ADDRESS: Option<&str> = Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-07-20")`
        pub const DATE: Option<&str> = Some("2000-07-20");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Arabic")`
        pub const LANGUAGE: Option<&str> = Some("Arabic");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("IBM Globalization Center of Competency, Yamato Software Laboratory")`
        pub const SOURCE: Option<&str> = Some("IBM Globalization Center of Competency, Yamato Software Laboratory");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Algeria")`
        pub const TERRITORY: Option<&str> = Some("Algeria");
        /// `Some("Arabic language locale for Algeria")`
        pub const TITLE: Option<&str> = Some("Arabic language locale for Algeria");
    }
    pub use super::ar_EG::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"د.ج."`
        pub const CURRENCY_SYMBOL: &str = "د.ج.";
        /// `3`
        pub const FRAC_DIGITS: i64 = 3;
        /// `"DZD "`
        pub const INT_CURR_SYMBOL: &str = "DZD ";
        /// `3`
        pub const INT_FRAC_DIGITS: i64 = 3;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `2`
        pub const N_SIGN_POSN: i64 = 2;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("213")`
        pub const INT_PREFIX: Option<&str> = Some("213");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c ;%a ;%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c ;%a ;%l");
    }
    pub mod LC_TIME {
        /// `&["ح", "ن", "ث", "ر", "خ", "ج", "س"]`
        pub const ABDAY: &[&str] = &["ح", "ن", "ث", "ر", "خ", "ج", "س"];
        /// `&["جانفي", "فيفري", "مارس", "أفريل", "ماي", "جوان", "جويلية", "أوت", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"]`
        pub const ABMON: &[&str] = &["جانفي", "فيفري", "مارس", "أفريل", "ماي", "جوان", "جويلية", "أوت", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["ص", "م"]`
        pub const AM_PM: &[&str] = &["ص", "م"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"]`
        pub const DAY: &[&str] = &["الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"];
        /// `"%d %b, %Y"`
        pub const D_FMT: &str = "%d %b, %Y";
        /// `"%d %b, %Y %Z %I:%M:%S %p"`
        pub const D_T_FMT: &str = "%d %b, %Y %Z %I:%M:%S %p";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(7)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(7);
        /// `Some(1)`
        pub const FIRST_WORKDAY: Option<i64> = Some(1);
        /// `&["جانفي", "فيفري", "مارس", "أفريل", "ماي", "جوان", "جويلية", "أوت", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"]`
        pub const MON: &[&str] = &["جانفي", "فيفري", "مارس", "أفريل", "ماي", "جوان", "جويلية", "أوت", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"];
        /// `"%Z %I:%M:%S %p"`
        pub const T_FMT: &str = "%Z %I:%M:%S %p";
        /// `"%Z %I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%Z %I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ar_EG {
    pub mod LC_ADDRESS {
        /// `Some("EGY")`
        pub const COUNTRY_AB3: Option<&str> = Some("EGY");
        /// `Some("EG")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("EG");
        /// `Some("ET")`
        pub const COUNTRY_CAR: Option<&str> = Some("ET");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("مصر")`
        pub const COUNTRY_NAME: Option<&str> = Some("مصر");
        /// `Some(818)`
        pub const COUNTRY_NUM: Option<i64> = Some(818);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ar")`
        pub const LANG_AB: Option<&str> = Some("ar");
        /// `Some("ara")`
        pub const LANG_LIB: Option<&str> = Some("ara");
        /// `Some("العربية")`
        pub const LANG_NAME: Option<&str> = Some("العربية");
        /// `Some("ara")`
        pub const LANG_TERM: Option<&str> = Some("ara");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan")`
        pub const ADDRESS: Option<&str> = Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-07-20")`
        pub const DATE: Option<&str> = Some("2000-07-20");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Arabic")`
        pub const LANGUAGE: Option<&str> = Some("Arabic");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("IBM Globalization Center of Competency, Yamato Software Laboratory")`
        pub const SOURCE: Option<&str> = Some("IBM Globalization Center of Competency, Yamato Software Laboratory");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Egypt")`
        pub const TERRITORY: Option<&str> = Some("Egypt");
        /// `Some("Arabic language locale for Egypt")`
        pub const TITLE: Option<&str> = Some("Arabic language locale for Egypt");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0لnN]"`
        pub const NOEXPR: &str = "^[-0لnN]";
        /// `Some("لا")`
        pub const NOSTR: Option<&str> = Some("لا");
        /// `"^[+1نyY]"`
        pub const YESEXPR: &str = "^[+1نyY]";
        /// `Some("نعم")`
        pub const YESSTR: Option<&str> = Some("نعم");
    }
    pub mod LC_MONETARY {
        /// `"ج.م."`
        pub const CURRENCY_SYMBOL: &str = "ج.م.";
        /// `3`
        pub const FRAC_DIGITS: i64 = 3;
        /// `"EGP "`
        pub const INT_CURR_SYMBOL: &str = "EGP ";
        /// `3`
        pub const INT_FRAC_DIGITS: i64 = 3;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `2`
        pub const N_SIGN_POSN: i64 = 2;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("20")`
        pub const INT_PREFIX: Option<&str> = Some("20");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c ;%a ;%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c ;%a ;%l");
    }
    pub mod LC_TIME {
        /// `&["ح", "ن", "ث", "ر", "خ", "ج", "س"]`
        pub const ABDAY: &[&str] = &["ح", "ن", "ث", "ر", "خ", "ج", "س"];
        /// `&["ينا", "فبر", "مار", "أبر", "ماي", "يون", "يول", "أغس", "سبت", "أكت", "نوف", "ديس"]`
        pub const ABMON: &[&str] = &["ينا", "فبر", "مار", "أبر", "ماي", "يون", "يول", "أغس", "سبت", "أكت", "نوف", "ديس"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["ص", "م"]`
        pub const AM_PM: &[&str] = &["ص", "م"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"]`
        pub const DAY: &[&str] = &["الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"];
        /// `"%d %b, %Y"`
        pub const D_FMT: &str = "%d %b, %Y";
        /// `"%d %b, %Y %Z %I:%M:%S %p"`
        pub const D_T_FMT: &str = "%d %b, %Y %Z %I:%M:%S %p";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(7)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(7);
        /// `Some(1)`
        pub const FIRST_WORKDAY: Option<i64> = Some(1);
        /// `&["يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"]`
        pub const MON: &[&str] = &["يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"];
        /// `"%Z %I:%M:%S %p"`
        pub const T_FMT: &str = "%Z %I:%M:%S %p";
        /// `"%Z %I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%Z %I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ar_IN {
    pub mod LC_ADDRESS {
        /// `Some("IND")`
        pub const COUNTRY_AB3: Option<&str> = Some("IND");
        /// `Some("IN")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IN");
        /// `Some("IND")`
        pub const COUNTRY_CAR: Option<&str> = Some("IND");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("الهند")`
        pub const COUNTRY_NAME: Option<&str> = Some("الهند");
        /// `Some(356)`
        pub const COUNTRY_NUM: Option<i64> = Some(356);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ar")`
        pub const LANG_AB: Option<&str> = Some("ar");
        /// `Some("ara")`
        pub const LANG_LIB: Option<&str> = Some("ara");
        /// `Some("العربية")`
        pub const LANG_NAME: Option<&str> = Some("العربية");
        /// `Some("ara")`
        pub const LANG_TERM: Option<&str> = Some("ara");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan")`
        pub const ADDRESS: Option<&str> = Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000,October,27 (XML source:2000,July,20)")`
        pub const DATE: Option<&str> = Some("2000,October,27 (XML source:2000,July,20)");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Arabic")`
        pub const LANGUAGE: Option<&str> = Some("Arabic");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("IBM Globalization Center of Competency, Yamato Software Laboratory")`
        pub const SOURCE: Option<&str> = Some("IBM Globalization Center of Competency, Yamato Software Laboratory");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("India")`
        pub const TERRITORY: Option<&str> = Some("India");
        /// `Some("Arabic language locale for India")`
        pub const TITLE: Option<&str> = Some("Arabic language locale for India");
    }
    pub use super::ar_EG::LC_MESSAGES;
    pub use super::hi_IN::LC_MONETARY;
    pub use super::hi_IN::LC_NUMERIC;
    pub use super::hi_IN::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["ح", "ن", "ث", "ر", "خ", "ج", "س"]`
        pub const ABDAY: &[&str] = &["ح", "ن", "ث", "ر", "خ", "ج", "س"];
        /// `&["ينا", "فبر", "مار", "أبر", "ماي", "يون", "يول", "أغس", "سبت", "أكت", "نوف", "ديس"]`
        pub const ABMON: &[&str] = &["ينا", "فبر", "مار", "أبر", "ماي", "يون", "يول", "أغس", "سبت", "أكت", "نوف", "ديس"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["ص", "م"]`
        pub const AM_PM: &[&str] = &["ص", "م"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"]`
        pub const DAY: &[&str] = &["الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"];
        /// `"%-d/%-m/%y"`
        pub const D_FMT: &str = "%-d/%-m/%y";
        /// `"%A %d %B %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A %d %B %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"]`
        pub const MON: &[&str] = &["يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"];
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT: &str = "%I:%M:%S %p %Z";
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p %Z";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ar_IQ {
    pub mod LC_ADDRESS {
        /// `Some("IRQ")`
        pub const COUNTRY_AB3: Option<&str> = Some("IRQ");
        /// `Some("IQ")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IQ");
        /// `Some("IRQ")`
        pub const COUNTRY_CAR: Option<&str> = Some("IRQ");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("العراق")`
        pub const COUNTRY_NAME: Option<&str> = Some("العراق");
        /// `Some(368)`
        pub const COUNTRY_NUM: Option<i64> = Some(368);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ar")`
        pub const LANG_AB: Option<&str> = Some("ar");
        /// `Some("ara")`
        pub const LANG_LIB: Option<&str> = Some("ara");
        /// `Some("العربية")`
        pub const LANG_NAME: Option<&str> = Some("العربية");
        /// `Some("ara")`
        pub const LANG_TERM: Option<&str> = Some("ara");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan")`
        pub const ADDRESS: Option<&str> = Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-07-20")`
        pub const DATE: Option<&str> = Some("2000-07-20");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Arabic")`
        pub const LANGUAGE: Option<&str> = Some("Arabic");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("IBM Globalization Center of Competency, Yamato Software Laboratory")`
        pub const SOURCE: Option<&str> = Some("IBM Globalization Center of Competency, Yamato Software Laboratory");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Iraq")`
        pub const TERRITORY: Option<&str> = Some("Iraq");
        /// `Some("Arabic language locale for Iraq")`
        pub const TITLE: Option<&str> = Some("Arabic language locale for Iraq");
    }
    pub use super::ar_EG::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"د.ع."`
        pub const CURRENCY_SYMBOL: &str = "د.ع.";
        /// `3`
        pub const FRAC_DIGITS: i64 = 3;
        /// `"IQD "`
        pub const INT_CURR_SYMBOL: &str = "IQD ";
        /// `3`
        pub const INT_FRAC_DIGITS: i64 = 3;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `2`
        pub const N_SIGN_POSN: i64 = 2;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("964")`
        pub const INT_PREFIX: Option<&str> = Some("964");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c ;%a ;%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c ;%a ;%l");
    }
    pub mod LC_TIME {
        /// `&["ح", "ن", "ث", "ر", "خ", "ج", "س"]`
        pub const ABDAY: &[&str] = &["ح", "ن", "ث", "ر", "خ", "ج", "س"];
        /// `&["كانون الثاني", "شباط", "آذار", "نيسان", "أيار", "حزيران", "تموز", "آب", "أيلول", "تشرين الأول", "تشرين الثاني", "كانون الأول"]`
        pub const ABMON: &[&str] = &["كانون الثاني", "شباط", "آذار", "نيسان", "أيار", "حزيران", "تموز", "آب", "أيلول", "تشرين الأول", "تشرين الثاني", "كانون الأول"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["ص", "م"]`
        pub const AM_PM: &[&str] = &["ص", "م"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"]`
        pub const DAY: &[&str] = &["الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"];
        /// `"%d %b, %Y"`
        pub const D_FMT: &str = "%d %b, %Y";
        /// `"%d %b, %Y %Z %I:%M:%S %p"`
        pub const D_T_FMT: &str = "%d %b, %Y %Z %I:%M:%S %p";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(7)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(7);
        /// `Some(1)`
        pub const FIRST_WORKDAY: Option<i64> = Some(1);
        /// `&["كانون الثاني", "شباط", "آذار", "نيسان", "أيار", "حزيران", "تموز", "آب", "أيلول", "تشرين الأول", "تشرين الثاني", "كانون الأول"]`
        pub const MON: &[&str] = &["كانون الثاني", "شباط", "آذار", "نيسان", "أيار", "حزيران", "تموز", "آب", "أيلول", "تشرين الأول", "تشرين الثاني", "كانون الأول"];
        /// `"%Z %I:%M:%S %p"`
        pub const T_FMT: &str = "%Z %I:%M:%S %p";
        /// `"%Z %I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%Z %I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ar_JO {
    pub mod LC_ADDRESS {
        /// `Some("JOR")`
        pub const COUNTRY_AB3: Option<&str> = Some("JOR");
        /// `Some("JO")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("JO");
        /// `Some("HKJ")`
        pub const COUNTRY_CAR: Option<&str> = Some("HKJ");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("الأردن")`
        pub const COUNTRY_NAME: Option<&str> = Some("الأردن");
        /// `Some(400)`
        pub const COUNTRY_NUM: Option<i64> = Some(400);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ar")`
        pub const LANG_AB: Option<&str> = Some("ar");
        /// `Some("ara")`
        pub const LANG_LIB: Option<&str> = Some("ara");
        /// `Some("العربية")`
        pub const LANG_NAME: Option<&str> = Some("العربية");
        /// `Some("ara")`
        pub const LANG_TERM: Option<&str> = Some("ara");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan")`
        pub const ADDRESS: Option<&str> = Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-07-20")`
        pub const DATE: Option<&str> = Some("2000-07-20");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Arabic")`
        pub const LANGUAGE: Option<&str> = Some("Arabic");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("IBM Globalization Center of Competency, Yamato Software Laboratory")`
        pub const SOURCE: Option<&str> = Some("IBM Globalization Center of Competency, Yamato Software Laboratory");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Jordan")`
        pub const TERRITORY: Option<&str> = Some("Jordan");
        /// `Some("Arabic language locale for Jordan")`
        pub const TITLE: Option<&str> = Some("Arabic language locale for Jordan");
    }
    pub use super::ar_EG::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"د.أ."`
        pub const CURRENCY_SYMBOL: &str = "د.أ.";
        /// `3`
        pub const FRAC_DIGITS: i64 = 3;
        /// `"JOD "`
        pub const INT_CURR_SYMBOL: &str = "JOD ";
        /// `3`
        pub const INT_FRAC_DIGITS: i64 = 3;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `2`
        pub const N_SIGN_POSN: i64 = 2;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("962")`
        pub const INT_PREFIX: Option<&str> = Some("962");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c ;%a ;%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c ;%a ;%l");
    }
    pub mod LC_TIME {
        /// `&["ح", "ن", "ث", "ر", "خ", "ج", "س"]`
        pub const ABDAY: &[&str] = &["ح", "ن", "ث", "ر", "خ", "ج", "س"];
        /// `&["كانون الثاني", "شباط", "آذار", "نيسان", "أيار", "حزيران", "تموز", "آب", "أيلول", "تشرين الأول", "تشرين الثاني", "كانون الأول"]`
        pub const ABMON: &[&str] = &["كانون الثاني", "شباط", "آذار", "نيسان", "أيار", "حزيران", "تموز", "آب", "أيلول", "تشرين الأول", "تشرين الثاني", "كانون الأول"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["ص", "م"]`
        pub const AM_PM: &[&str] = &["ص", "م"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"]`
        pub const DAY: &[&str] = &["الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"];
        /// `"%d %b, %Y"`
        pub const D_FMT: &str = "%d %b, %Y";
        /// `"%d %b, %Y %Z %I:%M:%S %p"`
        pub const D_T_FMT: &str = "%d %b, %Y %Z %I:%M:%S %p";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(7)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(7);
        /// `Some(1)`
        pub const FIRST_WORKDAY: Option<i64> = Some(1);
        /// `&["كانون الثاني", "شباط", "آذار", "نيسان", "أيار", "حزيران", "تموز", "آب", "أيلول", "تشرين الأول", "تشرين الثاني", "كانون الأول"]`
        pub const MON: &[&str] = &["كانون الثاني", "شباط", "آذار", "نيسان", "أيار", "حزيران", "تموز", "آب", "أيلول", "تشرين الأول", "تشرين الثاني", "كانون الأول"];
        /// `"%Z %I:%M:%S %p"`
        pub const T_FMT: &str = "%Z %I:%M:%S %p";
        /// `"%Z %I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%Z %I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ar_KW {
    pub mod LC_ADDRESS {
        /// `Some("KWT")`
        pub const COUNTRY_AB3: Option<&str> = Some("KWT");
        /// `Some("KW")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("KW");
        /// `Some("KWT")`
        pub const COUNTRY_CAR: Option<&str> = Some("KWT");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("الكويت")`
        pub const COUNTRY_NAME: Option<&str> = Some("الكويت");
        /// `Some(414)`
        pub const COUNTRY_NUM: Option<i64> = Some(414);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ar")`
        pub const LANG_AB: Option<&str> = Some("ar");
        /// `Some("ara")`
        pub const LANG_LIB: Option<&str> = Some("ara");
        /// `Some("العربية")`
        pub const LANG_NAME: Option<&str> = Some("العربية");
        /// `Some("ara")`
        pub const LANG_TERM: Option<&str> = Some("ara");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan")`
        pub const ADDRESS: Option<&str> = Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-07-20")`
        pub const DATE: Option<&str> = Some("2000-07-20");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Arabic")`
        pub const LANGUAGE: Option<&str> = Some("Arabic");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("IBM Globalization Center of Competency, Yamato Software Laboratory")`
        pub const SOURCE: Option<&str> = Some("IBM Globalization Center of Competency, Yamato Software Laboratory");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Kuwait")`
        pub const TERRITORY: Option<&str> = Some("Kuwait");
        /// `Some("Arabic language locale for Kuwait")`
        pub const TITLE: Option<&str> = Some("Arabic language locale for Kuwait");
    }
    pub use super::ar_EG::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"د.ك."`
        pub const CURRENCY_SYMBOL: &str = "د.ك.";
        /// `3`
        pub const FRAC_DIGITS: i64 = 3;
        /// `"KWD "`
        pub const INT_CURR_SYMBOL: &str = "KWD ";
        /// `3`
        pub const INT_FRAC_DIGITS: i64 = 3;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `2`
        pub const N_SIGN_POSN: i64 = 2;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("965")`
        pub const INT_PREFIX: Option<&str> = Some("965");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c ;%a ;%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c ;%a ;%l");
    }
    pub mod LC_TIME {
        /// `&["ح", "ن", "ث", "ر", "خ", "ج", "س"]`
        pub const ABDAY: &[&str] = &["ح", "ن", "ث", "ر", "خ", "ج", "س"];
        /// `&["ينا", "فبر", "مار", "أبر", "ماي", "يون", "يول", "أغس", "سبت", "أكت", "نوف", "ديس"]`
        pub const ABMON: &[&str] = &["ينا", "فبر", "مار", "أبر", "ماي", "يون", "يول", "أغس", "سبت", "أكت", "نوف", "ديس"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["ص", "م"]`
        pub const AM_PM: &[&str] = &["ص", "م"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"]`
        pub const DAY: &[&str] = &["الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"];
        /// `"%d %b, %Y"`
        pub const D_FMT: &str = "%d %b, %Y";
        /// `"%d %b, %Y %Z %I:%M:%S %p"`
        pub const D_T_FMT: &str = "%d %b, %Y %Z %I:%M:%S %p";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(7)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(7);
        /// `Some(1)`
        pub const FIRST_WORKDAY: Option<i64> = Some(1);
        /// `&["يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"]`
        pub const MON: &[&str] = &["يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"];
        /// `"%Z %I:%M:%S %p"`
        pub const T_FMT: &str = "%Z %I:%M:%S %p";
        /// `"%Z %I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%Z %I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ar_LB {
    pub mod LC_ADDRESS {
        /// `Some("LBN")`
        pub const COUNTRY_AB3: Option<&str> = Some("LBN");
        /// `Some("LB")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("LB");
        /// `Some("RL")`
        pub const COUNTRY_CAR: Option<&str> = Some("RL");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("لبنان")`
        pub const COUNTRY_NAME: Option<&str> = Some("لبنان");
        /// `Some(422)`
        pub const COUNTRY_NUM: Option<i64> = Some(422);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ar")`
        pub const LANG_AB: Option<&str> = Some("ar");
        /// `Some("ara")`
        pub const LANG_LIB: Option<&str> = Some("ara");
        /// `Some("العربية")`
        pub const LANG_NAME: Option<&str> = Some("العربية");
        /// `Some("ara")`
        pub const LANG_TERM: Option<&str> = Some("ara");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan")`
        pub const ADDRESS: Option<&str> = Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-07-20")`
        pub const DATE: Option<&str> = Some("2000-07-20");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Arabic")`
        pub const LANGUAGE: Option<&str> = Some("Arabic");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("IBM Globalization Center of Competency, Yamato Software Laboratory")`
        pub const SOURCE: Option<&str> = Some("IBM Globalization Center of Competency, Yamato Software Laboratory");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Lebanon")`
        pub const TERRITORY: Option<&str> = Some("Lebanon");
        /// `Some("Arabic language locale for Lebanon")`
        pub const TITLE: Option<&str> = Some("Arabic language locale for Lebanon");
    }
    pub use super::ar_EG::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"ل.ل."`
        pub const CURRENCY_SYMBOL: &str = "ل.ل.";
        /// `3`
        pub const FRAC_DIGITS: i64 = 3;
        /// `"LBP "`
        pub const INT_CURR_SYMBOL: &str = "LBP ";
        /// `3`
        pub const INT_FRAC_DIGITS: i64 = 3;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `2`
        pub const N_SIGN_POSN: i64 = 2;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("961")`
        pub const INT_PREFIX: Option<&str> = Some("961");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c ;%a ;%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c ;%a ;%l");
    }
    pub mod LC_TIME {
        /// `&["ح", "ن", "ث", "ر", "خ", "ج", "س"]`
        pub const ABDAY: &[&str] = &["ح", "ن", "ث", "ر", "خ", "ج", "س"];
        /// `&["كانون الثاني", "شباط", "آذار", "نيسان", "أيار", "حزيران", "تموز", "آب", "أيلول", "تشرين الأول", "تشرين الثاني", "كانون الأول"]`
        pub const ABMON: &[&str] = &["كانون الثاني", "شباط", "آذار", "نيسان", "أيار", "حزيران", "تموز", "آب", "أيلول", "تشرين الأول", "تشرين الثاني", "كانون الأول"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["ص", "م"]`
        pub const AM_PM: &[&str] = &["ص", "م"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"]`
        pub const DAY: &[&str] = &["الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"];
        /// `"%d %b, %Y"`
        pub const D_FMT: &str = "%d %b, %Y";
        /// `"%d %b, %Y %Z %I:%M:%S %p"`
        pub const D_T_FMT: &str = "%d %b, %Y %Z %I:%M:%S %p";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["كانون الثاني", "شباط", "آذار", "نيسان", "أيار", "حزيران", "تموز", "آب", "أيلول", "تشرين الأول", "تشرين الثاني", "كانون الأول"]`
        pub const MON: &[&str] = &["كانون الثاني", "شباط", "آذار", "نيسان", "أيار", "حزيران", "تموز", "آب", "أيلول", "تشرين الأول", "تشرين الثاني", "كانون الأول"];
        /// `"%Z %I:%M:%S %p"`
        pub const T_FMT: &str = "%Z %I:%M:%S %p";
        /// `"%Z %I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%Z %I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ar_LY {
    pub mod LC_ADDRESS {
        /// `Some("LBY")`
        pub const COUNTRY_AB3: Option<&str> = Some("LBY");
        /// `Some("LY")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("LY");
        /// `Some("LAR")`
        pub const COUNTRY_CAR: Option<&str> = Some("LAR");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("ليبيا")`
        pub const COUNTRY_NAME: Option<&str> = Some("ليبيا");
        /// `Some(434)`
        pub const COUNTRY_NUM: Option<i64> = Some(434);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ar")`
        pub const LANG_AB: Option<&str> = Some("ar");
        /// `Some("ara")`
        pub const LANG_LIB: Option<&str> = Some("ara");
        /// `Some("العربية")`
        pub const LANG_NAME: Option<&str> = Some("العربية");
        /// `Some("ara")`
        pub const LANG_TERM: Option<&str> = Some("ara");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan")`
        pub const ADDRESS: Option<&str> = Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-07-20")`
        pub const DATE: Option<&str> = Some("2000-07-20");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Arabic")`
        pub const LANGUAGE: Option<&str> = Some("Arabic");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("IBM Globalization Center of Competency, Yamato Software Laboratory")`
        pub const SOURCE: Option<&str> = Some("IBM Globalization Center of Competency, Yamato Software Laboratory");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Libya")`
        pub const TERRITORY: Option<&str> = Some("Libya");
        /// `Some("Arabic language locale for Libyan Arab Jamahiriya")`
        pub const TITLE: Option<&str> = Some("Arabic language locale for Libyan Arab Jamahiriya");
    }
    pub use super::ar_EG::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"د.ل."`
        pub const CURRENCY_SYMBOL: &str = "د.ل.";
        /// `3`
        pub const FRAC_DIGITS: i64 = 3;
        /// `"LYD "`
        pub const INT_CURR_SYMBOL: &str = "LYD ";
        /// `3`
        pub const INT_FRAC_DIGITS: i64 = 3;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `2`
        pub const N_SIGN_POSN: i64 = 2;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("218")`
        pub const INT_PREFIX: Option<&str> = Some("218");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c ;%a ;%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c ;%a ;%l");
    }
    pub mod LC_TIME {
        /// `&["ح", "ن", "ث", "ر", "خ", "ج", "س"]`
        pub const ABDAY: &[&str] = &["ح", "ن", "ث", "ر", "خ", "ج", "س"];
        /// `&["ينا", "فبر", "مار", "أبر", "ماي", "يون", "يول", "أغس", "سبت", "أكت", "نوف", "ديس"]`
        pub const ABMON: &[&str] = &["ينا", "فبر", "مار", "أبر", "ماي", "يون", "يول", "أغس", "سبت", "أكت", "نوف", "ديس"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["ص", "م"]`
        pub const AM_PM: &[&str] = &["ص", "م"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"]`
        pub const DAY: &[&str] = &["الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"];
        /// `"%d %b, %Y"`
        pub const D_FMT: &str = "%d %b, %Y";
        /// `"%d %b, %Y %Z %I:%M:%S %p"`
        pub const D_T_FMT: &str = "%d %b, %Y %Z %I:%M:%S %p";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(7)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(7);
        /// `Some(1)`
        pub const FIRST_WORKDAY: Option<i64> = Some(1);
        /// `&["يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"]`
        pub const MON: &[&str] = &["يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"];
        /// `"%Z %I:%M:%S %p"`
        pub const T_FMT: &str = "%Z %I:%M:%S %p";
        /// `"%Z %I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%Z %I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ar_MA {
    pub mod LC_ADDRESS {
        /// `Some("MAR")`
        pub const COUNTRY_AB3: Option<&str> = Some("MAR");
        /// `Some("MA")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("MA");
        /// `Some("MA")`
        pub const COUNTRY_CAR: Option<&str> = Some("MA");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("المغرب")`
        pub const COUNTRY_NAME: Option<&str> = Some("المغرب");
        /// `Some(504)`
        pub const COUNTRY_NUM: Option<i64> = Some(504);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ar")`
        pub const LANG_AB: Option<&str> = Some("ar");
        /// `Some("ara")`
        pub const LANG_LIB: Option<&str> = Some("ara");
        /// `Some("العربية")`
        pub const LANG_NAME: Option<&str> = Some("العربية");
        /// `Some("ara")`
        pub const LANG_TERM: Option<&str> = Some("ara");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan")`
        pub const ADDRESS: Option<&str> = Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-07-20")`
        pub const DATE: Option<&str> = Some("2000-07-20");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Arabic")`
        pub const LANGUAGE: Option<&str> = Some("Arabic");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("IBM Globalization Center of Competency, Yamato Software Laboratory")`
        pub const SOURCE: Option<&str> = Some("IBM Globalization Center of Competency, Yamato Software Laboratory");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Morocco")`
        pub const TERRITORY: Option<&str> = Some("Morocco");
        /// `Some("Arabic language locale for Morocco")`
        pub const TITLE: Option<&str> = Some("Arabic language locale for Morocco");
    }
    pub use super::ar_EG::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"د.م."`
        pub const CURRENCY_SYMBOL: &str = "د.م.";
        /// `3`
        pub const FRAC_DIGITS: i64 = 3;
        /// `"MAD "`
        pub const INT_CURR_SYMBOL: &str = "MAD ";
        /// `3`
        pub const INT_FRAC_DIGITS: i64 = 3;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `2`
        pub const N_SIGN_POSN: i64 = 2;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("212")`
        pub const INT_PREFIX: Option<&str> = Some("212");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c ;%a ;%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c ;%a ;%l");
    }
    pub mod LC_TIME {
        /// `&["ح", "ن", "ث", "ر", "خ", "ج", "س"]`
        pub const ABDAY: &[&str] = &["ح", "ن", "ث", "ر", "خ", "ج", "س"];
        /// `&["يناير", "فبراير", "مارس", "أبريل", "ماي", "يونيو", "يوليوز", "غشت", "شتنبر", "أكتوبر", "نونبر", "دجنبر"]`
        pub const ABMON: &[&str] = &["يناير", "فبراير", "مارس", "أبريل", "ماي", "يونيو", "يوليوز", "غشت", "شتنبر", "أكتوبر", "نونبر", "دجنبر"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["ص", "م"]`
        pub const AM_PM: &[&str] = &["ص", "م"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"]`
        pub const DAY: &[&str] = &["الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"];
        /// `"%d %b, %Y"`
        pub const D_FMT: &str = "%d %b, %Y";
        /// `"%d %b, %Y %Z %H:%M:%S"`
        pub const D_T_FMT: &str = "%d %b, %Y %Z %H:%M:%S";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(7)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(7);
        /// `Some(1)`
        pub const FIRST_WORKDAY: Option<i64> = Some(1);
        /// `&["يناير", "فبراير", "مارس", "أبريل", "ماي", "يونيو", "يوليوز", "غشت", "شتنبر", "أكتوبر", "نونبر", "دجنبر"]`
        pub const MON: &[&str] = &["يناير", "فبراير", "مارس", "أبريل", "ماي", "يونيو", "يوليوز", "غشت", "شتنبر", "أكتوبر", "نونبر", "دجنبر"];
        /// `"%Z %H:%M:%S"`
        pub const T_FMT: &str = "%Z %H:%M:%S";
        /// `"%l:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%l:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ar_OM {
    pub mod LC_ADDRESS {
        /// `Some("OMN")`
        pub const COUNTRY_AB3: Option<&str> = Some("OMN");
        /// `Some("OM")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("OM");
        /// `Some("OM")`
        pub const COUNTRY_CAR: Option<&str> = Some("OM");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("ع\u{64f}مان")`
        pub const COUNTRY_NAME: Option<&str> = Some("ع\u{64f}مان");
        /// `Some(512)`
        pub const COUNTRY_NUM: Option<i64> = Some(512);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ar")`
        pub const LANG_AB: Option<&str> = Some("ar");
        /// `Some("ara")`
        pub const LANG_LIB: Option<&str> = Some("ara");
        /// `Some("العربية")`
        pub const LANG_NAME: Option<&str> = Some("العربية");
        /// `Some("ara")`
        pub const LANG_TERM: Option<&str> = Some("ara");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan")`
        pub const ADDRESS: Option<&str> = Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-07-20")`
        pub const DATE: Option<&str> = Some("2000-07-20");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Arabic")`
        pub const LANGUAGE: Option<&str> = Some("Arabic");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("IBM Globalization Center of Competency, Yamato Software Laboratory")`
        pub const SOURCE: Option<&str> = Some("IBM Globalization Center of Competency, Yamato Software Laboratory");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Oman")`
        pub const TERRITORY: Option<&str> = Some("Oman");
        /// `Some("Arabic language locale for Oman")`
        pub const TITLE: Option<&str> = Some("Arabic language locale for Oman");
    }
    pub use super::ar_EG::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"ر.ع."`
        pub const CURRENCY_SYMBOL: &str = "ر.ع.";
        /// `3`
        pub const FRAC_DIGITS: i64 = 3;
        /// `"OMR "`
        pub const INT_CURR_SYMBOL: &str = "OMR ";
        /// `3`
        pub const INT_FRAC_DIGITS: i64 = 3;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `2`
        pub const N_SIGN_POSN: i64 = 2;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("968")`
        pub const INT_PREFIX: Option<&str> = Some("968");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c ;%a ;%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c ;%a ;%l");
    }
    pub mod LC_TIME {
        /// `&["ح", "ن", "ث", "ر", "خ", "ج", "س"]`
        pub const ABDAY: &[&str] = &["ح", "ن", "ث", "ر", "خ", "ج", "س"];
        /// `&["ينا", "فبر", "مار", "أبر", "ماي", "يون", "يول", "أغس", "سبت", "أكت", "نوف", "ديس"]`
        pub const ABMON: &[&str] = &["ينا", "فبر", "مار", "أبر", "ماي", "يون", "يول", "أغس", "سبت", "أكت", "نوف", "ديس"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["ص", "م"]`
        pub const AM_PM: &[&str] = &["ص", "م"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"]`
        pub const DAY: &[&str] = &["الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"];
        /// `"%d %b, %Y"`
        pub const D_FMT: &str = "%d %b, %Y";
        /// `"%d %b, %Y %Z %I:%M:%S %p"`
        pub const D_T_FMT: &str = "%d %b, %Y %Z %I:%M:%S %p";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(7)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(7);
        /// `Some(1)`
        pub const FIRST_WORKDAY: Option<i64> = Some(1);
        /// `&["يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"]`
        pub const MON: &[&str] = &["يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"];
        /// `"%Z %I:%M:%S %p"`
        pub const T_FMT: &str = "%Z %I:%M:%S %p";
        /// `"%Z %I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%Z %I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ar_QA {
    pub mod LC_ADDRESS {
        /// `Some("QAT")`
        pub const COUNTRY_AB3: Option<&str> = Some("QAT");
        /// `Some("QA")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("QA");
        /// `Some("Q")`
        pub const COUNTRY_CAR: Option<&str> = Some("Q");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("قطر")`
        pub const COUNTRY_NAME: Option<&str> = Some("قطر");
        /// `Some(634)`
        pub const COUNTRY_NUM: Option<i64> = Some(634);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ar")`
        pub const LANG_AB: Option<&str> = Some("ar");
        /// `Some("ara")`
        pub const LANG_LIB: Option<&str> = Some("ara");
        /// `Some("العربية")`
        pub const LANG_NAME: Option<&str> = Some("العربية");
        /// `Some("ara")`
        pub const LANG_TERM: Option<&str> = Some("ara");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan")`
        pub const ADDRESS: Option<&str> = Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-07-20")`
        pub const DATE: Option<&str> = Some("2000-07-20");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Arabic")`
        pub const LANGUAGE: Option<&str> = Some("Arabic");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("IBM Globalization Center of Competency, Yamato Software Laboratory")`
        pub const SOURCE: Option<&str> = Some("IBM Globalization Center of Competency, Yamato Software Laboratory");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Qatar")`
        pub const TERRITORY: Option<&str> = Some("Qatar");
        /// `Some("Arabic language locale for Qatar")`
        pub const TITLE: Option<&str> = Some("Arabic language locale for Qatar");
    }
    pub use super::ar_EG::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"ر.ق."`
        pub const CURRENCY_SYMBOL: &str = "ر.ق.";
        /// `3`
        pub const FRAC_DIGITS: i64 = 3;
        /// `"QAR "`
        pub const INT_CURR_SYMBOL: &str = "QAR ";
        /// `3`
        pub const INT_FRAC_DIGITS: i64 = 3;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `2`
        pub const N_SIGN_POSN: i64 = 2;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("974")`
        pub const INT_PREFIX: Option<&str> = Some("974");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c ;%a ;%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c ;%a ;%l");
    }
    pub mod LC_TIME {
        /// `&["ح", "ن", "ث", "ر", "خ", "ج", "س"]`
        pub const ABDAY: &[&str] = &["ح", "ن", "ث", "ر", "خ", "ج", "س"];
        /// `&["ينا", "فبر", "مار", "أبر", "ماي", "يون", "يول", "أغس", "سبت", "أكت", "نوف", "ديس"]`
        pub const ABMON: &[&str] = &["ينا", "فبر", "مار", "أبر", "ماي", "يون", "يول", "أغس", "سبت", "أكت", "نوف", "ديس"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["ص", "م"]`
        pub const AM_PM: &[&str] = &["ص", "م"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"]`
        pub const DAY: &[&str] = &["الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"];
        /// `"%d %b, %Y"`
        pub const D_FMT: &str = "%d %b, %Y";
        /// `"%d %b, %Y %Z %I:%M:%S %p"`
        pub const D_T_FMT: &str = "%d %b, %Y %Z %I:%M:%S %p";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(7)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(7);
        /// `Some(1)`
        pub const FIRST_WORKDAY: Option<i64> = Some(1);
        /// `&["يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"]`
        pub const MON: &[&str] = &["يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"];
        /// `"%Z %I:%M:%S %p"`
        pub const T_FMT: &str = "%Z %I:%M:%S %p";
        /// `"%Z %I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%Z %I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ar_SA {
    pub mod LC_ADDRESS {
        /// `Some("SAU")`
        pub const COUNTRY_AB3: Option<&str> = Some("SAU");
        /// `Some("SA")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("SA");
        /// `Some("KSA")`
        pub const COUNTRY_CAR: Option<&str> = Some("KSA");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("المملكة العربية السعودية")`
        pub const COUNTRY_NAME: Option<&str> = Some("المملكة العربية السعودية");
        /// `Some(682)`
        pub const COUNTRY_NUM: Option<i64> = Some(682);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ar")`
        pub const LANG_AB: Option<&str> = Some("ar");
        /// `Some("ara")`
        pub const LANG_LIB: Option<&str> = Some("ara");
        /// `Some("العربية")`
        pub const LANG_NAME: Option<&str> = Some("العربية");
        /// `Some("ara")`
        pub const LANG_TERM: Option<&str> = Some("ara");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Arabic")`
        pub const LANGUAGE: Option<&str> = Some("Arabic");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Saudi Arabia")`
        pub const TERRITORY: Option<&str> = Some("Saudi Arabia");
        /// `Some("Arabic locale for Saudi Arabia")`
        pub const TITLE: Option<&str> = Some("Arabic locale for Saudi Arabia");
    }
    pub use super::ar_EG::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"ر.س"`
        pub const CURRENCY_SYMBOL: &str = "ر.س";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"SAR "`
        pub const INT_CURR_SYMBOL: &str = "SAR ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[-1]`
        pub const MON_GROUPING: &[i64] = &[-1];
        /// `""`
        pub const MON_THOUSANDS_SEP: &str = "";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[-1]`
        pub const GROUPING: &[i64] = &[-1];
        /// `""`
        pub const THOUSANDS_SEP: &str = "";
    }
    pub mod LC_TELEPHONE {
        /// `Some("966")`
        pub const INT_PREFIX: Option<&str> = Some("966");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["ح", "ن", "ث", "ر", "خ", "ج", "س"]`
        pub const ABDAY: &[&str] = &["ح", "ن", "ث", "ر", "خ", "ج", "س"];
        /// `&["يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"]`
        pub const ABMON: &[&str] = &["يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["الأحد", "الإثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعـة", "السبت"]`
        pub const DAY: &[&str] = &["الأحد", "الإثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعـة", "السبت"];
        /// `"%A %e %B %Y"`
        pub const D_FMT: &str = "%A %e %B %Y";
        /// `"%A %e %B %Y  %k:%M:%S"`
        pub const D_T_FMT: &str = "%A %e %B %Y  %k:%M:%S";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `Some("")`
        pub const ERA_D_FMT: Option<&str> = Some("");
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `Some(1)`
        pub const FIRST_WORKDAY: Option<i64> = Some(1);
        /// `&["يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"]`
        pub const MON: &[&str] = &["يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"];
        /// `"%k:%M:%S"`
        pub const T_FMT: &str = "%k:%M:%S";
        /// `"%k:%M:%S"`
        pub const T_FMT_AMPM: &str = "%k:%M:%S";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ar_SD {
    pub mod LC_ADDRESS {
        /// `Some("SDN")`
        pub const COUNTRY_AB3: Option<&str> = Some("SDN");
        /// `Some("SD")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("SD");
        /// `Some("SUD")`
        pub const COUNTRY_CAR: Option<&str> = Some("SUD");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("السودان")`
        pub const COUNTRY_NAME: Option<&str> = Some("السودان");
        /// `Some(729)`
        pub const COUNTRY_NUM: Option<i64> = Some(729);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ar")`
        pub const LANG_AB: Option<&str> = Some("ar");
        /// `Some("ara")`
        pub const LANG_LIB: Option<&str> = Some("ara");
        /// `Some("العربية")`
        pub const LANG_NAME: Option<&str> = Some("العربية");
        /// `Some("ara")`
        pub const LANG_TERM: Option<&str> = Some("ara");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan")`
        pub const ADDRESS: Option<&str> = Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-07-20")`
        pub const DATE: Option<&str> = Some("2000-07-20");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Arabic")`
        pub const LANGUAGE: Option<&str> = Some("Arabic");
        /// `Some("1.1")`
        pub const REVISION: Option<&str> = Some("1.1");
        /// `Some("IBM Globalization Center of Competency, Yamato Software Laboratory")`
        pub const SOURCE: Option<&str> = Some("IBM Globalization Center of Competency, Yamato Software Laboratory");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Sudan")`
        pub const TERRITORY: Option<&str> = Some("Sudan");
        /// `Some("Arabic language locale for Sudan")`
        pub const TITLE: Option<&str> = Some("Arabic language locale for Sudan");
    }
    pub use super::ar_EG::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"ج.س."`
        pub const CURRENCY_SYMBOL: &str = "ج.س.";
        /// `3`
        pub const FRAC_DIGITS: i64 = 3;
        /// `"SDG "`
        pub const INT_CURR_SYMBOL: &str = "SDG ";
        /// `3`
        pub const INT_FRAC_DIGITS: i64 = 3;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `2`
        pub const N_SIGN_POSN: i64 = 2;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("249")`
        pub const INT_PREFIX: Option<&str> = Some("249");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c ;%a ;%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c ;%a ;%l");
    }
    pub mod LC_TIME {
        /// `&["ح", "ن", "ث", "ر", "خ", "ج", "س"]`
        pub const ABDAY: &[&str] = &["ح", "ن", "ث", "ر", "خ", "ج", "س"];
        /// `&["ينا", "فبر", "مار", "أبر", "ماي", "يون", "يول", "أغس", "سبت", "أكت", "نوف", "ديس"]`
        pub const ABMON: &[&str] = &["ينا", "فبر", "مار", "أبر", "ماي", "يون", "يول", "أغس", "سبت", "أكت", "نوف", "ديس"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["ص", "م"]`
        pub const AM_PM: &[&str] = &["ص", "م"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"]`
        pub const DAY: &[&str] = &["الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"];
        /// `"%d %b, %Y"`
        pub const D_FMT: &str = "%d %b, %Y";
        /// `"%d %b, %Y %Z %I:%M:%S %p"`
        pub const D_T_FMT: &str = "%d %b, %Y %Z %I:%M:%S %p";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(7)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(7);
        /// `Some(1)`
        pub const FIRST_WORKDAY: Option<i64> = Some(1);
        /// `&["يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"]`
        pub const MON: &[&str] = &["يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"];
        /// `"%Z %I:%M:%S %p"`
        pub const T_FMT: &str = "%Z %I:%M:%S %p";
        /// `"%Z %I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%Z %I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ar_SS {
    pub mod LC_ADDRESS {
        /// `Some("SSD")`
        pub const COUNTRY_AB3: Option<&str> = Some("SSD");
        /// `Some("SS")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("SS");
        /// `Some("SUD")`
        pub const COUNTRY_CAR: Option<&str> = Some("SUD");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("جنوب السودان")`
        pub const COUNTRY_NAME: Option<&str> = Some("جنوب السودان");
        /// `Some(728)`
        pub const COUNTRY_NUM: Option<i64> = Some(728);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ar")`
        pub const LANG_AB: Option<&str> = Some("ar");
        /// `Some("ara")`
        pub const LANG_LIB: Option<&str> = Some("ara");
        /// `Some("العربية")`
        pub const LANG_NAME: Option<&str> = Some("العربية");
        /// `Some("ara")`
        pub const LANG_TERM: Option<&str> = Some("ara");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan")`
        pub const ADDRESS: Option<&str> = Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-07-20")`
        pub const DATE: Option<&str> = Some("2000-07-20");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Arabic")`
        pub const LANGUAGE: Option<&str> = Some("Arabic");
        /// `Some("1.1")`
        pub const REVISION: Option<&str> = Some("1.1");
        /// `Some("IBM Globalization Center of Competency, Yamato Software Laboratory")`
        pub const SOURCE: Option<&str> = Some("IBM Globalization Center of Competency, Yamato Software Laboratory");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("South Sudan")`
        pub const TERRITORY: Option<&str> = Some("South Sudan");
        /// `Some("Arabic language locale for South Sudan")`
        pub const TITLE: Option<&str> = Some("Arabic language locale for South Sudan");
    }
    pub use super::ar_EG::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"£"`
        pub const CURRENCY_SYMBOL: &str = "£";
        /// `3`
        pub const FRAC_DIGITS: i64 = 3;
        /// `"SSP "`
        pub const INT_CURR_SYMBOL: &str = "SSP ";
        /// `3`
        pub const INT_FRAC_DIGITS: i64 = 3;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `2`
        pub const N_SIGN_POSN: i64 = 2;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("211")`
        pub const INT_PREFIX: Option<&str> = Some("211");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c ;%a ;%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c ;%a ;%l");
    }
    pub mod LC_TIME {
        /// `&["ح", "ن", "ث", "ر", "خ", "ج", "س"]`
        pub const ABDAY: &[&str] = &["ح", "ن", "ث", "ر", "خ", "ج", "س"];
        /// `&["ينا", "فبر", "مار", "أبر", "ماي", "يون", "يول", "أغس", "سبت", "أكت", "نوف", "ديس"]`
        pub const ABMON: &[&str] = &["ينا", "فبر", "مار", "أبر", "ماي", "يون", "يول", "أغس", "سبت", "أكت", "نوف", "ديس"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["ص", "م"]`
        pub const AM_PM: &[&str] = &["ص", "م"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"]`
        pub const DAY: &[&str] = &["الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"];
        /// `"%d %b, %Y"`
        pub const D_FMT: &str = "%d %b, %Y";
        /// `"%d %b, %Y %Z %I:%M:%S %p"`
        pub const D_T_FMT: &str = "%d %b, %Y %Z %I:%M:%S %p";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"]`
        pub const MON: &[&str] = &["يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"];
        /// `"%Z %I:%M:%S %p"`
        pub const T_FMT: &str = "%Z %I:%M:%S %p";
        /// `"%Z %I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%Z %I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ar_SY {
    pub mod LC_ADDRESS {
        /// `Some("SYR")`
        pub const COUNTRY_AB3: Option<&str> = Some("SYR");
        /// `Some("SY")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("SY");
        /// `Some("SYR")`
        pub const COUNTRY_CAR: Option<&str> = Some("SYR");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("سوريا")`
        pub const COUNTRY_NAME: Option<&str> = Some("سوريا");
        /// `Some(760)`
        pub const COUNTRY_NUM: Option<i64> = Some(760);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ar")`
        pub const LANG_AB: Option<&str> = Some("ar");
        /// `Some("ara")`
        pub const LANG_LIB: Option<&str> = Some("ara");
        /// `Some("العربية")`
        pub const LANG_NAME: Option<&str> = Some("العربية");
        /// `Some("ara")`
        pub const LANG_TERM: Option<&str> = Some("ara");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan")`
        pub const ADDRESS: Option<&str> = Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-07-20")`
        pub const DATE: Option<&str> = Some("2000-07-20");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Arabic")`
        pub const LANGUAGE: Option<&str> = Some("Arabic");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("IBM Globalization Center of Competency, Yamato Software Laboratory")`
        pub const SOURCE: Option<&str> = Some("IBM Globalization Center of Competency, Yamato Software Laboratory");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Syria")`
        pub const TERRITORY: Option<&str> = Some("Syria");
        /// `Some("Arabic language locale for Syrian Arab Republic")`
        pub const TITLE: Option<&str> = Some("Arabic language locale for Syrian Arab Republic");
    }
    pub use super::ar_EG::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"ل.س."`
        pub const CURRENCY_SYMBOL: &str = "ل.س.";
        /// `3`
        pub const FRAC_DIGITS: i64 = 3;
        /// `"SYP "`
        pub const INT_CURR_SYMBOL: &str = "SYP ";
        /// `3`
        pub const INT_FRAC_DIGITS: i64 = 3;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `2`
        pub const N_SIGN_POSN: i64 = 2;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("963")`
        pub const INT_PREFIX: Option<&str> = Some("963");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c ;%a ;%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c ;%a ;%l");
    }
    pub mod LC_TIME {
        /// `&["ح", "ن", "ث", "ر", "خ", "ج", "س"]`
        pub const ABDAY: &[&str] = &["ح", "ن", "ث", "ر", "خ", "ج", "س"];
        /// `&["كانون الثاني", "شباط", "آذار", "نيسان", "أيار", "حزيران", "تموز", "آب", "أيلول", "تشرين الأول", "تشرين الثاني", "كانون الأول"]`
        pub const ABMON: &[&str] = &["كانون الثاني", "شباط", "آذار", "نيسان", "أيار", "حزيران", "تموز", "آب", "أيلول", "تشرين الأول", "تشرين الثاني", "كانون الأول"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["ص", "م"]`
        pub const AM_PM: &[&str] = &["ص", "م"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"]`
        pub const DAY: &[&str] = &["الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"];
        /// `"%d %b, %Y"`
        pub const D_FMT: &str = "%d %b, %Y";
        /// `"%d %b, %Y %Z %I:%M:%S %p"`
        pub const D_T_FMT: &str = "%d %b, %Y %Z %I:%M:%S %p";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(7)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(7);
        /// `Some(1)`
        pub const FIRST_WORKDAY: Option<i64> = Some(1);
        /// `&["كانون الثاني", "شباط", "آذار", "نيسان", "أيار", "حزيران", "تموز", "آب", "أيلول", "تشرين الأول", "تشرين الثاني", "كانون الأول"]`
        pub const MON: &[&str] = &["كانون الثاني", "شباط", "آذار", "نيسان", "أيار", "حزيران", "تموز", "آب", "أيلول", "تشرين الأول", "تشرين الثاني", "كانون الأول"];
        /// `"%Z %I:%M:%S %p"`
        pub const T_FMT: &str = "%Z %I:%M:%S %p";
        /// `"%Z %I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%Z %I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ar_TN {
    pub mod LC_ADDRESS {
        /// `Some("TUN")`
        pub const COUNTRY_AB3: Option<&str> = Some("TUN");
        /// `Some("TN")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("TN");
        /// `Some("TN")`
        pub const COUNTRY_CAR: Option<&str> = Some("TN");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("تونس")`
        pub const COUNTRY_NAME: Option<&str> = Some("تونس");
        /// `Some(788)`
        pub const COUNTRY_NUM: Option<i64> = Some(788);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ar")`
        pub const LANG_AB: Option<&str> = Some("ar");
        /// `Some("ara")`
        pub const LANG_LIB: Option<&str> = Some("ara");
        /// `Some("العربية")`
        pub const LANG_NAME: Option<&str> = Some("العربية");
        /// `Some("ara")`
        pub const LANG_TERM: Option<&str> = Some("ara");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan")`
        pub const ADDRESS: Option<&str> = Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-07-20")`
        pub const DATE: Option<&str> = Some("2000-07-20");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Arabic")`
        pub const LANGUAGE: Option<&str> = Some("Arabic");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("IBM Globalization Center of Competency, Yamato Software Laboratory")`
        pub const SOURCE: Option<&str> = Some("IBM Globalization Center of Competency, Yamato Software Laboratory");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Tunisia")`
        pub const TERRITORY: Option<&str> = Some("Tunisia");
        /// `Some("Arabic language locale for Tunisia")`
        pub const TITLE: Option<&str> = Some("Arabic language locale for Tunisia");
    }
    pub use super::ar_EG::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"د.ت."`
        pub const CURRENCY_SYMBOL: &str = "د.ت.";
        /// `3`
        pub const FRAC_DIGITS: i64 = 3;
        /// `"TND "`
        pub const INT_CURR_SYMBOL: &str = "TND ";
        /// `3`
        pub const INT_FRAC_DIGITS: i64 = 3;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `2`
        pub const N_SIGN_POSN: i64 = 2;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("216")`
        pub const INT_PREFIX: Option<&str> = Some("216");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c ;%a ;%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c ;%a ;%l");
    }
    pub mod LC_TIME {
        /// `&["ح", "ن", "ث", "ر", "خ", "ج", "س"]`
        pub const ABDAY: &[&str] = &["ح", "ن", "ث", "ر", "خ", "ج", "س"];
        /// `&["جانفي", "فيفري", "مارس", "أفريل", "ماي", "جوان", "جويلية", "أوت", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"]`
        pub const ABMON: &[&str] = &["جانفي", "فيفري", "مارس", "أفريل", "ماي", "جوان", "جويلية", "أوت", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["ص", "م"]`
        pub const AM_PM: &[&str] = &["ص", "م"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"]`
        pub const DAY: &[&str] = &["الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"];
        /// `"%d %b, %Y"`
        pub const D_FMT: &str = "%d %b, %Y";
        /// `"%d %b, %Y %Z %I:%M:%S %p"`
        pub const D_T_FMT: &str = "%d %b, %Y %Z %I:%M:%S %p";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `Some(1)`
        pub const FIRST_WORKDAY: Option<i64> = Some(1);
        /// `&["جانفي", "فيفري", "مارس", "أفريل", "ماي", "جوان", "جويلية", "أوت", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"]`
        pub const MON: &[&str] = &["جانفي", "فيفري", "مارس", "أفريل", "ماي", "جوان", "جويلية", "أوت", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"];
        /// `"%Z %I:%M:%S %p"`
        pub const T_FMT: &str = "%Z %I:%M:%S %p";
        /// `"%Z %I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%Z %I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ar_YE {
    pub mod LC_ADDRESS {
        /// `Some("YEM")`
        pub const COUNTRY_AB3: Option<&str> = Some("YEM");
        /// `Some("YE")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("YE");
        /// `Some("YAR")`
        pub const COUNTRY_CAR: Option<&str> = Some("YAR");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("اليمن")`
        pub const COUNTRY_NAME: Option<&str> = Some("اليمن");
        /// `Some(887)`
        pub const COUNTRY_NUM: Option<i64> = Some(887);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ar")`
        pub const LANG_AB: Option<&str> = Some("ar");
        /// `Some("ara")`
        pub const LANG_LIB: Option<&str> = Some("ara");
        /// `Some("العربية")`
        pub const LANG_NAME: Option<&str> = Some("العربية");
        /// `Some("ara")`
        pub const LANG_TERM: Option<&str> = Some("ara");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan")`
        pub const ADDRESS: Option<&str> = Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-07-20")`
        pub const DATE: Option<&str> = Some("2000-07-20");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Arabic")`
        pub const LANGUAGE: Option<&str> = Some("Arabic");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("IBM Globalization Center of Competency, Yamato Software Laboratory")`
        pub const SOURCE: Option<&str> = Some("IBM Globalization Center of Competency, Yamato Software Laboratory");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Yemen")`
        pub const TERRITORY: Option<&str> = Some("Yemen");
        /// `Some("Arabic language locale for Yemen")`
        pub const TITLE: Option<&str> = Some("Arabic language locale for Yemen");
    }
    pub use super::ar_EG::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"ر.ي."`
        pub const CURRENCY_SYMBOL: &str = "ر.ي.";
        /// `3`
        pub const FRAC_DIGITS: i64 = 3;
        /// `"YER "`
        pub const INT_CURR_SYMBOL: &str = "YER ";
        /// `3`
        pub const INT_FRAC_DIGITS: i64 = 3;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `2`
        pub const N_SIGN_POSN: i64 = 2;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("967")`
        pub const INT_PREFIX: Option<&str> = Some("967");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c ;%a ;%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c ;%a ;%l");
    }
    pub mod LC_TIME {
        /// `&["ح", "ن", "ث", "ر", "خ", "ج", "س"]`
        pub const ABDAY: &[&str] = &["ح", "ن", "ث", "ر", "خ", "ج", "س"];
        /// `&["ينا", "فبر", "مار", "أبر", "ماي", "يون", "يول", "أغس", "سبت", "أكت", "نوف", "ديس"]`
        pub const ABMON: &[&str] = &["ينا", "فبر", "مار", "أبر", "ماي", "يون", "يول", "أغس", "سبت", "أكت", "نوف", "ديس"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["ص", "م"]`
        pub const AM_PM: &[&str] = &["ص", "م"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"]`
        pub const DAY: &[&str] = &["الأحد", "الاثنين", "الثلاثاء", "الأربعاء", "الخميس", "الجمعة", "السبت"];
        /// `"%d %b, %Y"`
        pub const D_FMT: &str = "%d %b, %Y";
        /// `"%d %b, %Y %Z %I:%M:%S %p"`
        pub const D_T_FMT: &str = "%d %b, %Y %Z %I:%M:%S %p";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `Some(1)`
        pub const FIRST_WORKDAY: Option<i64> = Some(1);
        /// `&["يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"]`
        pub const MON: &[&str] = &["يناير", "فبراير", "مارس", "أبريل", "مايو", "يونيو", "يوليو", "أغسطس", "سبتمبر", "أكتوبر", "نوفمبر", "ديسمبر"];
        /// `"%Z %I:%M:%S %p"`
        pub const T_FMT: &str = "%Z %I:%M:%S %p";
        /// `"%Z %I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%Z %I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod as_IN {
    pub mod LC_ADDRESS {
        /// `None`
        pub const COUNTRY_AB3: Option<&str> = None;
        /// `None`
        pub const COUNTRY_ABTWO: Option<&str> = None;
        /// `Some("IND")`
        pub const COUNTRY_CAR: Option<&str> = Some("IND");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("ভ\u{9be}ৰত")`
        pub const COUNTRY_NAME: Option<&str> = Some("ভ\u{9be}ৰত");
        /// `Some(356)`
        pub const COUNTRY_NUM: Option<i64> = Some(356);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("as")`
        pub const LANG_AB: Option<&str> = Some("as");
        /// `Some("asm")`
        pub const LANG_LIB: Option<&str> = Some("asm");
        /// `Some("অসমীয\u{9bc}\u{9be}")`
        pub const LANG_NAME: Option<&str> = Some("অসমীয\u{9bc}\u{9be}");
        /// `Some("asm")`
        pub const LANG_TERM: Option<&str> = Some("asm");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2006-05-25")`
        pub const DATE: Option<&str> = Some("2006-05-25");
        /// `Some("bug-glibc@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Assamese")`
        pub const LANGUAGE: Option<&str> = Some("Assamese");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Amitakhya Phukan, Red Hat")`
        pub const SOURCE: Option<&str> = Some("Amitakhya Phukan, Red Hat");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("India")`
        pub const TERRITORY: Option<&str> = Some("India");
        /// `Some("Assamese language locale for India")`
        pub const TITLE: Option<&str> = Some("Assamese language locale for India");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNন]"`
        pub const NOEXPR: &str = "^[-0nNন]";
        /// `Some("নহয়")`
        pub const NOSTR: Option<&str> = Some("নহয়");
        /// `"^[+1yYহ]"`
        pub const YESEXPR: &str = "^[+1yYহ]";
        /// `Some("হয়")`
        pub const YESSTR: Option<&str> = Some("হয়");
    }
    pub use super::hi_IN::LC_MONETARY;
    pub use super::en_IN::LC_NUMERIC;
    pub mod LC_TELEPHONE {
        /// `Some("91")`
        pub const INT_PREFIX: Option<&str> = Some("91");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c ;%a ;%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c ;%a ;%l");
    }
    pub mod LC_TIME {
        /// `&["দেও", "সোম", "মঙ\u{9cd}গল", "ব\u{9c1}ধ", "ব\u{9c3}হষ\u{9cd}পতি", "শ\u{9c1}ক\u{9cd}ৰ", "শনি"]`
        pub const ABDAY: &[&str] = &["দেও", "সোম", "মঙ\u{9cd}গল", "ব\u{9c1}ধ", "ব\u{9c3}হষ\u{9cd}পতি", "শ\u{9c1}ক\u{9cd}ৰ", "শনি"];
        /// `&["জ\u{9be}ন\u{9c1}", "ফেব\u{9cd}ৰ\u{9c1}", "ম\u{9be}ৰ\u{9cd}চ", "এপ\u{9cd}ৰিল", "মে", "জ\u{9c1}ন", "জ\u{9c1}ল\u{9be}ই", "আগ", "সেপ\u{9cd}ট", "অক\u{9cd}টো", "নভে", "ডিসে"]`
        pub const ABMON: &[&str] = &["জ\u{9be}ন\u{9c1}", "ফেব\u{9cd}ৰ\u{9c1}", "ম\u{9be}ৰ\u{9cd}চ", "এপ\u{9cd}ৰিল", "মে", "জ\u{9c1}ন", "জ\u{9c1}ল\u{9be}ই", "আগ", "সেপ\u{9cd}ট", "অক\u{9cd}টো", "নভে", "ডিসে"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["প\u{9c2}ৰ\u{9cd}ব\u{9cd}ব\u{9be}হ\u{9cd}ন", "অপৰ\u{9be}হ\u{9cd}ন"]`
        pub const AM_PM: &[&str] = &["প\u{9c2}ৰ\u{9cd}ব\u{9cd}ব\u{9be}হ\u{9cd}ন", "অপৰ\u{9be}হ\u{9cd}ন"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["দেওব\u{9be}ৰ", "সোমব\u{9be}ৰ", "মঙ\u{9cd}গলব\u{9be}ৰ", "ব\u{9c1}ধব\u{9be}ৰ", "ব\u{9c3}হষ\u{9cd}পতিব\u{9be}ৰ", "শ\u{9c1}ক\u{9cd}ৰব\u{9be}ৰ", "শনিব\u{9be}ৰ"]`
        pub const DAY: &[&str] = &["দেওব\u{9be}ৰ", "সোমব\u{9be}ৰ", "মঙ\u{9cd}গলব\u{9be}ৰ", "ব\u{9c1}ধব\u{9be}ৰ", "ব\u{9c3}হষ\u{9cd}পতিব\u{9be}ৰ", "শ\u{9c1}ক\u{9cd}ৰব\u{9be}ৰ", "শনিব\u{9be}ৰ"];
        /// `"%e-%m-%Y"`
        pub const D_FMT: &str = "%e-%m-%Y";
        /// `"%e %B, %Y %I.%M.%S %p %Z"`
        pub const D_T_FMT: &str = "%e %B, %Y %I.%M.%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["জ\u{9be}ন\u{9c1}ৱ\u{9be}ৰী", "ফেব\u{9cd}ৰ\u{9c1}ৱ\u{9be}ৰী", "ম\u{9be}ৰ\u{9cd}চ", "এপ\u{9cd}ৰিল", "মে", "জ\u{9c1}ন", "জ\u{9c1}ল\u{9be}ই", "আগষ\u{9cd}ট", "ছেপ\u{9cd}তেম\u{9cd}বৰ", "অক\u{9cd}টোবৰ", "নৱেম\u{9cd}বৰ", "ডিচেম\u{9cd}বৰ"]`
        pub const MON: &[&str] = &["জ\u{9be}ন\u{9c1}ৱ\u{9be}ৰী", "ফেব\u{9cd}ৰ\u{9c1}ৱ\u{9be}ৰী", "ম\u{9be}ৰ\u{9cd}চ", "এপ\u{9cd}ৰিল", "মে", "জ\u{9c1}ন", "জ\u{9c1}ল\u{9be}ই", "আগষ\u{9cd}ট", "ছেপ\u{9cd}তেম\u{9cd}বৰ", "অক\u{9cd}টোবৰ", "নৱেম\u{9cd}বৰ", "ডিচেম\u{9cd}বৰ"];
        /// `"%I.%M.%S %p"`
        pub const T_FMT: &str = "%I.%M.%S %p";
        /// `"%I.%M.%S %p"`
        pub const T_FMT_AMPM: &str = "%I.%M.%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ast_ES {
    pub mod LC_ADDRESS {
        /// `Some("ESP")`
        pub const COUNTRY_AB3: Option<&str> = Some("ESP");
        /// `Some("ES")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("ES");
        /// `Some("E")`
        pub const COUNTRY_CAR: Option<&str> = Some("E");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("España")`
        pub const COUNTRY_NAME: Option<&str> = Some("España");
        /// `Some(724)`
        pub const COUNTRY_NUM: Option<i64> = Some(724);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("ast")`
        pub const LANG_LIB: Option<&str> = Some("ast");
        /// `Some("asturianu")`
        pub const LANG_NAME: Option<&str> = Some("asturianu");
        /// `Some("ast")`
        pub const LANG_TERM: Option<&str> = Some("ast");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Jordi Mallach")`
        pub const CONTACT: Option<&str> = Some("Jordi Mallach");
        /// `Some("2005-08-26")`
        pub const DATE: Option<&str> = Some("2005-08-26");
        /// `Some("jordi@gnu.org")`
        pub const EMAIL: Option<&str> = Some("jordi@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Asturian")`
        pub const LANGUAGE: Option<&str> = Some("Asturian");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Spain")`
        pub const TERRITORY: Option<&str> = Some("Spain");
        /// `Some("Asturian locale for Spain")`
        pub const TITLE: Option<&str> = Some("Asturian locale for Spain");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("non")`
        pub const NOSTR: Option<&str> = Some("non");
        /// `"^[+1sSyY]"`
        pub const YESEXPR: &str = "^[+1sSyY]";
        /// `Some("sí")`
        pub const YESSTR: Option<&str> = Some("sí");
    }
    pub use super::es_ES::LC_MONETARY;
    pub use super::es_ES::LC_NUMERIC;
    pub use super::es_ES::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["dom", "llu", "mar", "mié", "xue", "vie", "sáb"]`
        pub const ABDAY: &[&str] = &["dom", "llu", "mar", "mié", "xue", "vie", "sáb"];
        /// `&["xin", "feb", "mar", "abr", "may", "xun", "xnt", "ago", "set", "och", "pay", "avi"]`
        pub const ABMON: &[&str] = &["xin", "feb", "mar", "abr", "may", "xun", "xnt", "ago", "set", "och", "pay", "avi"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `Some(&["xineru", "febreru", "marzu", "abril", "mayu", "xunu", "xunetu", "agostu", "setiembre", "ochobre", "payares", "avientu"])`
        pub const ALT_MON: Option<&[&str]> = Some(&["xineru", "febreru", "marzu", "abril", "mayu", "xunu", "xunetu", "agostu", "setiembre", "ochobre", "payares", "avientu"]);
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["domingu", "llunes", "martes", "miércoles", "xueves", "vienres", "sábadu"]`
        pub const DAY: &[&str] = &["domingu", "llunes", "martes", "miércoles", "xueves", "vienres", "sábadu"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["de xineru", "de febreru", "de marzu", "d’abril", "de mayu", "de xunu", "de xunetu", "d’agostu", "de setiembre", "d’ochobre", "de payares", "d’avientu"]`
        pub const MON: &[&str] = &["de xineru", "de febreru", "de marzu", "d’abril", "de mayu", "de xunu", "de xunetu", "d’agostu", "de setiembre", "d’ochobre", "de payares", "d’avientu"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ayc_PE {
    pub mod LC_ADDRESS {
        /// `Some("PER")`
        pub const COUNTRY_AB3: Option<&str> = Some("PER");
        /// `Some("PE")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("PE");
        /// `Some("PE")`
        pub const COUNTRY_CAR: Option<&str> = Some("PE");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Piruw")`
        pub const COUNTRY_NAME: Option<&str> = Some("Piruw");
        /// `Some(604)`
        pub const COUNTRY_NUM: Option<i64> = Some(604);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ay")`
        pub const LANG_AB: Option<&str> = Some("ay");
        /// `Some("ayc")`
        pub const LANG_LIB: Option<&str> = Some("ayc");
        /// `Some("Aymar aru")`
        pub const LANG_NAME: Option<&str> = Some("Aymar aru");
        /// `Some("ayc")`
        pub const LANG_TERM: Option<&str> = Some("ayc");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2011-11-13")`
        pub const DATE: Option<&str> = Some("2011-11-13");
        /// `Some("libc-alpha@sourceware.org")`
        pub const EMAIL: Option<&str> = Some("libc-alpha@sourceware.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Aymara")`
        pub const LANGUAGE: Option<&str> = Some("Aymara");
        /// `Some("1.1")`
        pub const REVISION: Option<&str> = Some("1.1");
        /// `Some("runasimipi.org")`
        pub const SOURCE: Option<&str> = Some("runasimipi.org");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Peru")`
        pub const TERRITORY: Option<&str> = Some("Peru");
        /// `Some("Aymara (ayc) locale for Peru")`
        pub const TITLE: Option<&str> = Some("Aymara (ayc) locale for Peru");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0jJnN]"`
        pub const NOEXPR: &str = "^[-0jJnN]";
        /// `Some("Janiwa")`
        pub const NOSTR: Option<&str> = Some("Janiwa");
        /// `"^[+1uUsSyY]"`
        pub const YESEXPR: &str = "^[+1uUsSyY]";
        /// `Some("Ukhamawa")`
        pub const YESSTR: Option<&str> = Some("Ukhamawa");
    }
    pub use super::es_PE::LC_MONETARY;
    pub use super::es_PE::LC_NUMERIC;
    pub use super::es_PE::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["tum", "lun", "mar", "mir", "juy", "wir", "saw"]`
        pub const ABDAY: &[&str] = &["tum", "lun", "mar", "mir", "juy", "wir", "saw"];
        /// `&["ini", "phi", "mar", "awr", "may", "jun", "jul", "awu", "sit", "ukt", "nuw", "ris"]`
        pub const ABMON: &[&str] = &["ini", "phi", "mar", "awr", "may", "jun", "jul", "awu", "sit", "ukt", "nuw", "ris"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["VM", "NM"]`
        pub const AM_PM: &[&str] = &["VM", "NM"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["tuminku", "lunisa", "martisa", "mirkulisa", "juywisa", "wirnisa", "sawäru"]`
        pub const DAY: &[&str] = &["tuminku", "lunisa", "martisa", "mirkulisa", "juywisa", "wirnisa", "sawäru"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["inïru", "phiwriru", "marsu", "awrila", "mayu", "junyu", "julyu", "awustu", "sitimri", "uktuwri", "nuwimri", "risimri"]`
        pub const MON: &[&str] = &["inïru", "phiwriru", "marsu", "awrila", "mayu", "junyu", "julyu", "awustu", "sitimri", "uktuwri", "nuwimri", "risimri"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `"%I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod az_AZ {
    pub mod LC_ADDRESS {
        /// `Some("AZE")`
        pub const COUNTRY_AB3: Option<&str> = Some("AZE");
        /// `Some("AZ")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("AZ");
        /// `Some("AZ")`
        pub const COUNTRY_CAR: Option<&str> = Some("AZ");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Azərbaycan")`
        pub const COUNTRY_NAME: Option<&str> = Some("Azərbaycan");
        /// `Some(31)`
        pub const COUNTRY_NUM: Option<i64> = Some(31);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("az")`
        pub const LANG_AB: Option<&str> = Some("az");
        /// `Some("aze")`
        pub const LANG_LIB: Option<&str> = Some("aze");
        /// `Some("azərbaycan")`
        pub const LANG_NAME: Option<&str> = Some("azərbaycan");
        /// `Some("aze")`
        pub const LANG_TERM: Option<&str> = Some("aze");
        /// `Some("???")`
        pub const POSTAL_FMT: Option<&str> = Some("???");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
        ]);
        /// `Some("Pablo Saratxaga")`
        pub const CONTACT: Option<&str> = Some("Pablo Saratxaga");
        /// `Some("2001-01-26")`
        pub const DATE: Option<&str> = Some("2001-01-26");
        /// `Some("pablo@mandrakesoft.com")`
        pub const EMAIL: Option<&str> = Some("pablo@mandrakesoft.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Azerbaijani")`
        pub const LANGUAGE: Option<&str> = Some("Azerbaijani");
        /// `Some("0.4")`
        pub const REVISION: Option<&str> = Some("0.4");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Azerbaijan")`
        pub const TERRITORY: Option<&str> = Some("Azerbaijan");
        /// `Some("Azeri language locale for Azerbaijan (latin)")`
        pub const TITLE: Option<&str> = Some("Azeri language locale for Azerbaijan (latin)");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0YyNn]"`
        pub const NOEXPR: &str = "^[-0YyNn]";
        /// `Some("yox")`
        pub const NOSTR: Option<&str> = Some("yox");
        /// `"^[+1bBhH]"`
        pub const YESEXPR: &str = "^[+1bBhH]";
        /// `Some("hə")`
        pub const YESSTR: Option<&str> = Some("hə");
    }
    pub mod LC_MONETARY {
        /// `"₼"`
        pub const CURRENCY_SYMBOL: &str = "₼";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"AZN "`
        pub const INT_CURR_SYMBOL: &str = "AZN ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const MON_THOUSANDS_SEP: &str = "\u{202f}";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const THOUSANDS_SEP: &str = ".";
    }
    pub mod LC_TELEPHONE {
        /// `Some("994")`
        pub const INT_PREFIX: Option<&str> = Some("994");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["baz", "ber", "çax", "çər", "cax", "cüm", "şnb"]`
        pub const ABDAY: &[&str] = &["baz", "ber", "çax", "çər", "cax", "cüm", "şnb"];
        /// `&["Yan", "Fev", "Mar", "Apr", "May", "İyn", "İyl", "Avq", "Sen", "Okt", "Noy", "Dek"]`
        pub const ABMON: &[&str] = &["Yan", "Fev", "Mar", "Apr", "May", "İyn", "İyl", "Avq", "Sen", "Okt", "Noy", "Dek"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["bazar günü", "bazar ertəsi", "çərşənbə axşamı", "çərşənbə", "cümə axşamı", "cümə", "şənbə"]`
        pub const DAY: &[&str] = &["bazar günü", "bazar ertəsi", "çərşənbə axşamı", "çərşənbə", "cümə axşamı", "cümə", "şənbə"];
        /// `"%d.%m.%Y"`
        pub const D_FMT: &str = "%d.%m.%Y";
        /// `"%A, %d %B %Y %T"`
        pub const D_T_FMT: &str = "%A, %d %B %Y %T";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["yanvar", "fevral", "mart", "aprel", "may", "iyun", "iyul", "avqust", "sentyabr", "oktyabr", "noyabr", "dekabr"]`
        pub const MON: &[&str] = &["yanvar", "fevral", "mart", "aprel", "may", "iyun", "iyul", "avqust", "sentyabr", "oktyabr", "noyabr", "dekabr"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod az_IR {
    pub mod LC_ADDRESS {
        /// `Some("IRN")`
        pub const COUNTRY_AB3: Option<&str> = Some("IRN");
        /// `Some("IR")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IR");
        /// `Some("IR")`
        pub const COUNTRY_CAR: Option<&str> = Some("IR");
        /// `Some("964")`
        pub const COUNTRY_ISBN: Option<&str> = Some("964");
        /// `Some("ایران")`
        pub const COUNTRY_NAME: Option<&str> = Some("ایران");
        /// `Some(364)`
        pub const COUNTRY_NUM: Option<i64> = Some(364);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("az")`
        pub const LANG_AB: Option<&str> = Some("az");
        /// `Some("aze")`
        pub const LANG_LIB: Option<&str> = Some("aze");
        /// `Some("تۆرکجه")`
        pub const LANG_NAME: Option<&str> = Some("تۆرکجه");
        /// `Some("azb")`
        pub const LANG_TERM: Option<&str> = Some("azb");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
        ]);
        /// `Some("Mousa Moradi")`
        pub const CONTACT: Option<&str> = Some("Mousa Moradi");
        /// `Some("2014-05-01")`
        pub const DATE: Option<&str> = Some("2014-05-01");
        /// `Some("mousamk@gmail.com")`
        pub const EMAIL: Option<&str> = Some("mousamk@gmail.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("South Azerbaijani")`
        pub const LANGUAGE: Option<&str> = Some("South Azerbaijani");
        /// `Some("0.2")`
        pub const REVISION: Option<&str> = Some("0.2");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("(+98) 914 492 1058")`
        pub const TEL: Option<&str> = Some("(+98) 914 492 1058");
        /// `Some("Iran")`
        pub const TERRITORY: Option<&str> = Some("Iran");
        /// `Some("South Azerbaijani language locale for Iran")`
        pub const TITLE: Option<&str> = Some("South Azerbaijani language locale for Iran");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNیخ]"`
        pub const NOEXPR: &str = "^[-0nNیخ]";
        /// `Some("یو\u{652}خ")`
        pub const NOSTR: Option<&str> = Some("یو\u{652}خ");
        /// `"^[+1yYهب]"`
        pub const YESEXPR: &str = "^[+1yYهب]";
        /// `Some("هن")`
        pub const YESSTR: Option<&str> = Some("هن");
    }
    pub use super::fa_IR::LC_MONETARY;
    pub use super::fa_IR::LC_NUMERIC;
    pub use super::fa_IR::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["یکشنبه", "دوشنبه", "سه\u{200c}شنبه", "چارشنبه", "جۆمعه آخشامی", "جۆمعه", "شنبه"]`
        pub const ABDAY: &[&str] = &["یکشنبه", "دوشنبه", "سه\u{200c}شنبه", "چارشنبه", "جۆمعه آخشامی", "جۆمعه", "شنبه"];
        /// `&["ژانویه", "فوریه", "مارس", "آوریل", "مئی", "ژوئن", "جولای", "آقۇست", "سپتامبر", "او\u{652}کتو\u{652}بر", "نو\u{652}وامبر", "دسامبر"]`
        pub const ABMON: &[&str] = &["ژانویه", "فوریه", "مارس", "آوریل", "مئی", "ژوئن", "جولای", "آقۇست", "سپتامبر", "او\u{652}کتو\u{652}بر", "نو\u{652}وامبر", "دسامبر"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `Some(&["۰۰", "۰۱", "۰۲", "۰۳", "۰۴", "۰۵", "۰۶", "۰۷", "۰۸", "۰۹", "۱۰", "۱۱", "۱۲", "۱۳", "۱۴", "۱۵", "۱۶", "۱۷", "۱۸", "۱۹", "۲۰", "۲۱", "۲۲", "۲۳", "۲۴", "۲۵", "۲۶", "۲۷", "۲۸", "۲۹", "۳۰", "۳۱", "۳۲", "۳۳", "۳۴", "۳۵", "۳۶", "۳۷", "۳۸", "۳۹", "۴۰", "۴۱", "۴۲", "۴۳", "۴۴", "۴۵", "۴۶", "۴۷", "۴۸", "۴۹", "۵۰", "۵۱", "۵۲", "۵۳", "۵۴", "۵۵", "۵۶", "۵۷", "۵۸", "۵۹", "۶۰", "۶۱", "۶۲", "۶۳", "۶۴", "۶۵", "۶۶", "۶۷", "۶۸", "۶۹", "۷۰", "۷۱", "۷۲", "۷۳", "۷۴", "۷۵", "۷۶", "۷۷", "۷۸", "۷۹", "۸۰", "۸۱", "۸۲", "۸۳", "۸۴", "۸۵", "۸۶", "۸۷", "۸۸", "۸۹", "۹۰", "۹۱", "۹۲", "۹۳", "۹۴", "۹۵", "۹۶", "۹۷", "۹۸", "۹۹"])`
        pub const ALT_DIGITS: Option<&[&str]> = Some(&["۰۰", "۰۱", "۰۲", "۰۳", "۰۴", "۰۵", "۰۶", "۰۷", "۰۸", "۰۹", "۱۰", "۱۱", "۱۲", "۱۳", "۱۴", "۱۵", "۱۶", "۱۷", "۱۸", "۱۹", "۲۰", "۲۱", "۲۲", "۲۳", "۲۴", "۲۵", "۲۶", "۲۷", "۲۸", "۲۹", "۳۰", "۳۱", "۳۲", "۳۳", "۳۴", "۳۵", "۳۶", "۳۷", "۳۸", "۳۹", "۴۰", "۴۱", "۴۲", "۴۳", "۴۴", "۴۵", "۴۶", "۴۷", "۴۸", "۴۹", "۵۰", "۵۱", "۵۲", "۵۳", "۵۴", "۵۵", "۵۶", "۵۷", "۵۸", "۵۹", "۶۰", "۶۱", "۶۲", "۶۳", "۶۴", "۶۵", "۶۶", "۶۷", "۶۸", "۶۹", "۷۰", "۷۱", "۷۲", "۷۳", "۷۴", "۷۵", "۷۶", "۷۷", "۷۸", "۷۹", "۸۰", "۸۱", "۸۲", "۸۳", "۸۴", "۸۵", "۸۶", "۸۷", "۸۸", "۸۹", "۹۰", "۹۱", "۹۲", "۹۳", "۹۴", "۹۵", "۹۶", "۹۷", "۹۸", "۹۹"]);
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `Some(3)`
        pub const CAL_DIRECTION: Option<i64> = Some(3);
        /// `Some("\u{202b}%A %Oe %B %Oy، ساعات %OH:%OM:%OS (%Z)\u{202c}")`
        pub const DATE_FMT: Option<&str> = Some("\u{202b}%A %Oe %B %Oy، ساعات %OH:%OM:%OS (%Z)\u{202c}");
        /// `&["یکشنبه", "دوشنبه", "سه\u{200c}شنبه", "چارشنبه", "جۆمعه آخشامی", "جۆمعه", "شنبه"]`
        pub const DAY: &[&str] = &["یکشنبه", "دوشنبه", "سه\u{200c}شنبه", "چارشنبه", "جۆمعه آخشامی", "جۆمعه", "شنبه"];
        /// `"%Oy/%Om/%Od"`
        pub const D_FMT: &str = "%Oy/%Om/%Od";
        /// `"\u{202b}%A %Oe %B %Oy، %OH:%OM:%OS\u{202c}"`
        pub const D_T_FMT: &str = "\u{202b}%A %Oe %B %Oy، %OH:%OM:%OS\u{202c}";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(7)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(7);
        /// `Some(7)`
        pub const FIRST_WORKDAY: Option<i64> = Some(7);
        /// `&["ژانویه", "فوریه", "مارس", "آوریل", "مئی", "ژوئن", "جولای", "آقۇست", "سپتامبر", "او\u{652}کتو\u{652}بر", "نو\u{652}وامبر", "دسامبر"]`
        pub const MON: &[&str] = &["ژانویه", "فوریه", "مارس", "آوریل", "مئی", "ژوئن", "جولای", "آقۇست", "سپتامبر", "او\u{652}کتو\u{652}بر", "نو\u{652}وامبر", "دسامبر"];
        /// `"%OH:%OM:%OS"`
        pub const T_FMT: &str = "%OH:%OM:%OS";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `None`
        pub const WEEK: Option<&[i64]> = None;
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod be_BY {
    pub mod LC_ADDRESS {
        /// `Some("BLR")`
        pub const COUNTRY_AB3: Option<&str> = Some("BLR");
        /// `Some("BY")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("BY");
        /// `Some("BY")`
        pub const COUNTRY_CAR: Option<&str> = Some("BY");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Беларусь")`
        pub const COUNTRY_NAME: Option<&str> = Some("Беларусь");
        /// `Some(112)`
        pub const COUNTRY_NUM: Option<i64> = Some(112);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("be")`
        pub const LANG_AB: Option<&str> = Some("be");
        /// `Some("bel")`
        pub const LANG_LIB: Option<&str> = Some("bel");
        /// `Some("беларуская")`
        pub const LANG_NAME: Option<&str> = Some("беларуская");
        /// `Some("bel")`
        pub const LANG_TERM: Option<&str> = Some("bel");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Belarusian")`
        pub const LANGUAGE: Option<&str> = Some("Belarusian");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Belarus")`
        pub const TERRITORY: Option<&str> = Some("Belarus");
        /// `Some("Belarusian locale for Belarus")`
        pub const TITLE: Option<&str> = Some("Belarusian locale for Belarus");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNНн]"`
        pub const NOEXPR: &str = "^[-0nNНн]";
        /// `Some("не")`
        pub const NOSTR: Option<&str> = Some("не");
        /// `"^[+1yYТт]"`
        pub const YESEXPR: &str = "^[+1yYТт]";
        /// `Some("так")`
        pub const YESSTR: Option<&str> = Some("так");
    }
    pub mod LC_MONETARY {
        /// `"руб"`
        pub const CURRENCY_SYMBOL: &str = "руб";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"BYR "`
        pub const INT_CURR_SYMBOL: &str = "BYR ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const MON_THOUSANDS_SEP: &str = "\u{202f}";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const THOUSANDS_SEP: &str = ".";
    }
    pub mod LC_TELEPHONE {
        /// `Some("375")`
        pub const INT_PREFIX: Option<&str> = Some("375");
        /// `Some("8~10")`
        pub const INT_SELECT: Option<&str> = Some("8~10");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["Няд", "Пан", "Аўт", "Срд", "Чцв", "Пят", "Суб"]`
        pub const ABDAY: &[&str] = &["Няд", "Пан", "Аўт", "Срд", "Чцв", "Пят", "Суб"];
        /// `&["сту", "лют", "сак", "кра", "мая", "чэр", "ліп", "жні", "вер", "кас", "ліс", "сне"]`
        pub const ABMON: &[&str] = &["сту", "лют", "сак", "кра", "мая", "чэр", "ліп", "жні", "вер", "кас", "ліс", "сне"];
        /// `Some(&["сту", "лют", "сак", "кра", "май", "чэр", "ліп", "жні", "вер", "кас", "ліс", "сне"])`
        pub const AB_ALT_MON: Option<&[&str]> = Some(&["сту", "лют", "сак", "кра", "май", "чэр", "ліп", "жні", "вер", "кас", "ліс", "сне"]);
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `Some(&["студзень", "люты", "сакавік", "красавік", "май", "чэрвень", "ліпень", "жнівень", "верасень", "кастрычнік", "лістапад", "снежань"])`
        pub const ALT_MON: Option<&[&str]> = Some(&["студзень", "люты", "сакавік", "красавік", "май", "чэрвень", "ліпень", "жнівень", "верасень", "кастрычнік", "лістапад", "снежань"]);
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Нядзеля", "Панядзелак", "Аўторак", "Серада", "Чацвер", "Пятніца", "Субота"]`
        pub const DAY: &[&str] = &["Нядзеля", "Панядзелак", "Аўторак", "Серада", "Чацвер", "Пятніца", "Субота"];
        /// `"%d.%m.%Y"`
        pub const D_FMT: &str = "%d.%m.%Y";
        /// `"%a %d %b %Y %T"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["студзеня", "лютага", "сакавіка", "красавіка", "мая", "чэрвеня", "ліпеня", "жніўня", "верасня", "кастрычніка", "лістапада", "снежня"]`
        pub const MON: &[&str] = &["студзеня", "лютага", "сакавіка", "красавіка", "мая", "чэрвеня", "ліпеня", "жніўня", "верасня", "кастрычніка", "лістапада", "снежня"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod be_BY_latin {
    pub mod LC_ADDRESS {
        /// `Some("BLR")`
        pub const COUNTRY_AB3: Option<&str> = Some("BLR");
        /// `Some("BY")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("BY");
        /// `Some("BY")`
        pub const COUNTRY_CAR: Option<&str> = Some("BY");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Biełaruś")`
        pub const COUNTRY_NAME: Option<&str> = Some("Biełaruś");
        /// `Some(112)`
        pub const COUNTRY_NUM: Option<i64> = Some(112);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("be")`
        pub const LANG_AB: Option<&str> = Some("be");
        /// `Some("bel")`
        pub const LANG_LIB: Option<&str> = Some("bel");
        /// `Some("biełaruskaja")`
        pub const LANG_NAME: Option<&str> = Some("biełaruskaja");
        /// `Some("bel")`
        pub const LANG_TERM: Option<&str> = Some("bel");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2005-09-15")`
        pub const DATE: Option<&str> = Some("2005-09-15");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Belarusian")`
        pub const LANGUAGE: Option<&str> = Some("Belarusian");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Belarus")`
        pub const TERRITORY: Option<&str> = Some("Belarus");
        /// `Some("Belarusian Latin-Script locale for Belarus")`
        pub const TITLE: Option<&str> = Some("Belarusian Latin-Script locale for Belarus");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0Nn]"`
        pub const NOEXPR: &str = "^[-0Nn]";
        /// `None`
        pub const NOSTR: Option<&str> = None;
        /// `"^[+1TtYy]"`
        pub const YESEXPR: &str = "^[+1TtYy]";
        /// `None`
        pub const YESSTR: Option<&str> = None;
    }
    pub mod LC_MONETARY {
        /// `"Rub"`
        pub const CURRENCY_SYMBOL: &str = "Rub";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"BYR "`
        pub const INT_CURR_SYMBOL: &str = "BYR ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const MON_THOUSANDS_SEP: &str = "\u{202f}";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub use super::be_BY::LC_NUMERIC;
    pub use super::be_BY::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Nia", "Pan", "Aŭt", "Sie", "Čać", "Pia", "Sub"]`
        pub const ABDAY: &[&str] = &["Nia", "Pan", "Aŭt", "Sie", "Čać", "Pia", "Sub"];
        /// `&["Stu", "Lut", "Sak", "Kra", "Maj", "Čer", "Lip", "Žni", "Vie", "Kas", "Lis", "Śni"]`
        pub const ABMON: &[&str] = &["Stu", "Lut", "Sak", "Kra", "Maj", "Čer", "Lip", "Žni", "Vie", "Kas", "Lis", "Śni"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `Some(&["Studzień", "Luty", "Sakavik", "Krasavik", "Maj", "Červień", "Lipień", "Žnivień", "Vierasień", "Kastryčnik", "Listapad", "Śniežań"])`
        pub const ALT_MON: Option<&[&str]> = Some(&["Studzień", "Luty", "Sakavik", "Krasavik", "Maj", "Červień", "Lipień", "Žnivień", "Vierasień", "Kastryčnik", "Listapad", "Śniežań"]);
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Niadziela", "Paniadziełak", "Aŭtorak", "Sierada", "Čaćvier", "Piatnica", "Subota"]`
        pub const DAY: &[&str] = &["Niadziela", "Paniadziełak", "Aŭtorak", "Sierada", "Čaćvier", "Piatnica", "Subota"];
        /// `"%d.%m.%Y"`
        pub const D_FMT: &str = "%d.%m.%Y";
        /// `"%a %d %b %Y %T"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["studzienia", "lutaha", "sakavika", "krasavika", "maja", "červienia", "lipienia", "žniŭnia", "vieraśnia", "kastryčnika", "listapada", "śniežnia"]`
        pub const MON: &[&str] = &["studzienia", "lutaha", "sakavika", "krasavika", "maja", "červienia", "lipienia", "žniŭnia", "vieraśnia", "kastryčnika", "listapada", "śniežnia"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod bem_ZM {
    pub mod LC_ADDRESS {
        /// `Some("ZMB")`
        pub const COUNTRY_AB3: Option<&str> = Some("ZMB");
        /// `Some("ZM")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("ZM");
        /// `Some("Z")`
        pub const COUNTRY_CAR: Option<&str> = Some("Z");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Zambia")`
        pub const COUNTRY_NAME: Option<&str> = Some("Zambia");
        /// `Some(894)`
        pub const COUNTRY_NUM: Option<i64> = Some(894);
        /// `Some("ZM")`
        pub const COUNTRY_POST: Option<&str> = Some("ZM");
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("bem")`
        pub const LANG_LIB: Option<&str> = Some("bem");
        /// `Some("Ichibemba")`
        pub const LANG_NAME: Option<&str> = Some("Ichibemba");
        /// `Some("bem")`
        pub const LANG_TERM: Option<&str> = Some("bem");
        /// `Some("%a%N%f%N%e %h%N%b %s%N%z%N%T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%a%N%f%N%e %h%N%b %s%N%z%N%T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `None`
        pub const ADDRESS: Option<&str> = None;
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
        ]);
        /// `Some("Martin Benjamin")`
        pub const CONTACT: Option<&str> = Some("Martin Benjamin");
        /// `Some("2011-04-18")`
        pub const DATE: Option<&str> = Some("2011-04-18");
        /// `Some("locales@africanlocalization.net")`
        pub const EMAIL: Option<&str> = Some("locales@africanlocalization.net");
        /// `None`
        pub const FAX: Option<&str> = None;
        /// `Some("Bemba")`
        pub const LANGUAGE: Option<&str> = Some("Bemba");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("ANLoc")`
        pub const SOURCE: Option<&str> = Some("ANLoc");
        /// `None`
        pub const TEL: Option<&str> = None;
        /// `Some("Zambia")`
        pub const TERRITORY: Option<&str> = Some("Zambia");
        /// `Some("Bemba locale for Zambia")`
        pub const TITLE: Option<&str> = Some("Bemba locale for Zambia");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNaA]"`
        pub const NOEXPR: &str = "^[-0nNaA]";
        /// `Some("Awe")`
        pub const NOSTR: Option<&str> = Some("Awe");
        /// `"^[+1yYeE]"`
        pub const YESEXPR: &str = "^[+1yYeE]";
        /// `Some("Ee")`
        pub const YESSTR: Option<&str> = Some("Ee");
    }
    pub mod LC_MONETARY {
        /// `"K"`
        pub const CURRENCY_SYMBOL: &str = "K";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"ZMW "`
        pub const INT_CURR_SYMBOL: &str = "ZMW ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("260")`
        pub const INT_PREFIX: Option<&str> = Some("260");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `Some("%A %l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("%A %l");
        /// `Some("+%c %A %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %A %l");
    }
    pub mod LC_TIME {
        /// `&["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"]`
        pub const ABDAY: &[&str] = &["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"];
        /// `&["Jan", "Feb", "Mac", "Epr", "Mei", "Jun", "Jul", "Oga", "Sep", "Okt", "Nov", "Dis"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Mac", "Epr", "Mei", "Jun", "Jul", "Oga", "Sep", "Okt", "Nov", "Dis"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["uluchelo", "akasuba"]`
        pub const AM_PM: &[&str] = &["uluchelo", "akasuba"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Pa Mulungu", "Palichimo", "Palichibuli", "Palichitatu", "Palichine", "Palichisano", "Pachibelushi"]`
        pub const DAY: &[&str] = &["Pa Mulungu", "Palichimo", "Palichibuli", "Palichitatu", "Palichine", "Palichisano", "Pachibelushi"];
        /// `"%m/%d/%Y"`
        pub const D_FMT: &str = "%m/%d/%Y";
        /// `"%a %d %b %Y %R %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %R %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Januari", "Februari", "Machi", "Epreo", "Mei", "Juni", "Julai", "Ogasti", "Septemba", "Oktoba", "Novemba", "Disemba"]`
        pub const MON: &[&str] = &["Januari", "Februari", "Machi", "Epreo", "Mei", "Juni", "Julai", "Ogasti", "Septemba", "Oktoba", "Novemba", "Disemba"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `"%I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ber_DZ {
    pub mod LC_ADDRESS {
        /// `Some("DZA")`
        pub const COUNTRY_AB3: Option<&str> = Some("DZA");
        /// `Some("DZ")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("DZ");
        /// `Some("DZ")`
        pub const COUNTRY_CAR: Option<&str> = Some("DZ");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Lezzayer")`
        pub const COUNTRY_NAME: Option<&str> = Some("Lezzayer");
        /// `Some(12)`
        pub const COUNTRY_NUM: Option<i64> = Some(12);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("ber")`
        pub const LANG_LIB: Option<&str> = Some("ber");
        /// `Some("Tamaziɣt")`
        pub const LANG_NAME: Option<&str> = Some("Tamaziɣt");
        /// `Some("ber")`
        pub const LANG_TERM: Option<&str> = Some("ber");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
        ]);
        /// `Some("Pablo Saratxaga")`
        pub const CONTACT: Option<&str> = Some("Pablo Saratxaga");
        /// `Some("2002-04-16")`
        pub const DATE: Option<&str> = Some("2002-04-16");
        /// `Some("pablo@mandrakesoft.com")`
        pub const EMAIL: Option<&str> = Some("pablo@mandrakesoft.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Berber")`
        pub const LANGUAGE: Option<&str> = Some("Berber");
        /// `Some("0.1")`
        pub const REVISION: Option<&str> = Some("0.1");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Algeria")`
        pub const TERRITORY: Option<&str> = Some("Algeria");
        /// `Some("Berber language locale for Algeria (latin)")`
        pub const TITLE: Option<&str> = Some("Berber language locale for Algeria (latin)");
    }
    pub use super::kab_DZ::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"د.ج."`
        pub const CURRENCY_SYMBOL: &str = "د.ج.";
        /// `3`
        pub const FRAC_DIGITS: i64 = 3;
        /// `"DZD "`
        pub const INT_CURR_SYMBOL: &str = "DZD ";
        /// `3`
        pub const INT_FRAC_DIGITS: i64 = 3;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `2`
        pub const N_SIGN_POSN: i64 = 2;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("213")`
        pub const INT_PREFIX: Option<&str> = Some("213");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c ;%a ;%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c ;%a ;%l");
    }
    pub mod LC_TIME {
        /// `&["baz", "bir", "iki", "üçü", "dör", "beş", "alt"]`
        pub const ABDAY: &[&str] = &["baz", "bir", "iki", "üçü", "dör", "beş", "alt"];
        /// `&["Yan", "Fev", "Mar", "Apr", "May", "İyn", "İyl", "Avq", "Sen", "Okt", "Noy", "Dek"]`
        pub const ABMON: &[&str] = &["Yan", "Fev", "Mar", "Apr", "May", "İyn", "İyl", "Avq", "Sen", "Okt", "Noy", "Dek"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["bazar günü", "birinci gün", "ikinci gün", "üçüncü gün", "dördüncü gün", "beşinci gün", "altıncı gün"]`
        pub const DAY: &[&str] = &["bazar günü", "birinci gün", "ikinci gün", "üçüncü gün", "dördüncü gün", "beşinci gün", "altıncı gün"];
        /// `"%d.%m.%Y"`
        pub const D_FMT: &str = "%d.%m.%Y";
        /// `"%A, %d %B %Y %T"`
        pub const D_T_FMT: &str = "%A, %d %B %Y %T";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(7)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(7);
        /// `Some(1)`
        pub const FIRST_WORKDAY: Option<i64> = Some(1);
        /// `&["yanvar", "fevral", "mart", "aprel", "may", "iyun", "iyul", "avqust", "sentyabr", "oktyabr", "noyabr", "dekabr"]`
        pub const MON: &[&str] = &["yanvar", "fevral", "mart", "aprel", "may", "iyun", "iyul", "avqust", "sentyabr", "oktyabr", "noyabr", "dekabr"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ber_MA {
    pub mod LC_ADDRESS {
        /// `Some("MAR")`
        pub const COUNTRY_AB3: Option<&str> = Some("MAR");
        /// `Some("MA")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("MA");
        /// `Some("MA")`
        pub const COUNTRY_CAR: Option<&str> = Some("MA");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("ⵜⴰⴳⵍⴷⵉⵜ ⵏ ⵍⵎⵖⵔⵉⴱ")`
        pub const COUNTRY_NAME: Option<&str> = Some("ⵜⴰⴳⵍⴷⵉⵜ ⵏ ⵍⵎⵖⵔⵉⴱ");
        /// `Some(504)`
        pub const COUNTRY_NUM: Option<i64> = Some(504);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("ber")`
        pub const LANG_LIB: Option<&str> = Some("ber");
        /// `Some("ⵜⴰⵎⴰⵣⵉⵖⵜ")`
        pub const LANG_NAME: Option<&str> = Some("ⵜⴰⵎⴰⵣⵉⵖⵜ");
        /// `Some("ber")`
        pub const LANG_TERM: Option<&str> = Some("ber");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
        ]);
        /// `Some("Pablo Saratxaga")`
        pub const CONTACT: Option<&str> = Some("Pablo Saratxaga");
        /// `Some("2002-06-26")`
        pub const DATE: Option<&str> = Some("2002-06-26");
        /// `Some("pablo@mandrakesoft.com")`
        pub const EMAIL: Option<&str> = Some("pablo@mandrakesoft.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Berber")`
        pub const LANGUAGE: Option<&str> = Some("Berber");
        /// `Some("0.1")`
        pub const REVISION: Option<&str> = Some("0.1");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Morocco")`
        pub const TERRITORY: Option<&str> = Some("Morocco");
        /// `Some("Berber language locale for Morocco (tifinagh)")`
        pub const TITLE: Option<&str> = Some("Berber language locale for Morocco (tifinagh)");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0YyNn]"`
        pub const NOEXPR: &str = "^[-0YyNn]";
        /// `Some("yox")`
        pub const NOSTR: Option<&str> = Some("yox");
        /// `"^[+1Bb]"`
        pub const YESEXPR: &str = "^[+1Bb]";
        /// `Some("bəli")`
        pub const YESSTR: Option<&str> = Some("bəli");
    }
    pub mod LC_MONETARY {
        /// `"ⴷ.ⵎ."`
        pub const CURRENCY_SYMBOL: &str = "ⴷ.ⵎ.";
        /// `3`
        pub const FRAC_DIGITS: i64 = 3;
        /// `"MAD "`
        pub const INT_CURR_SYMBOL: &str = "MAD ";
        /// `3`
        pub const INT_FRAC_DIGITS: i64 = 3;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `2`
        pub const N_SIGN_POSN: i64 = 2;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("212")`
        pub const INT_PREFIX: Option<&str> = Some("212");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c ;%a ;%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c ;%a ;%l");
    }
    pub mod LC_TIME {
        /// `&["baz", "bir", "iki", "üçü", "dör", "beş", "alt"]`
        pub const ABDAY: &[&str] = &["baz", "bir", "iki", "üçü", "dör", "beş", "alt"];
        /// `&["Yan", "Fev", "Mar", "Apr", "May", "İyn", "İyl", "Avq", "Sen", "Okt", "Noy", "Dek"]`
        pub const ABMON: &[&str] = &["Yan", "Fev", "Mar", "Apr", "May", "İyn", "İyl", "Avq", "Sen", "Okt", "Noy", "Dek"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["bazar günü", "birinci gün", "ikinci gün", "üçüncü gün", "dördüncü gün", "beşinci gün", "altıncı gün"]`
        pub const DAY: &[&str] = &["bazar günü", "birinci gün", "ikinci gün", "üçüncü gün", "dördüncü gün", "beşinci gün", "altıncı gün"];
        /// `"%d.%m.%Y"`
        pub const D_FMT: &str = "%d.%m.%Y";
        /// `"%A, %d %B %Y %T"`
        pub const D_T_FMT: &str = "%A, %d %B %Y %T";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(7)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(7);
        /// `Some(1)`
        pub const FIRST_WORKDAY: Option<i64> = Some(1);
        /// `&["yanvar", "fevral", "mart", "aprel", "may", "iyun", "iyul", "avqust", "sentyabr", "oktyabr", "noyabr", "dekabr"]`
        pub const MON: &[&str] = &["yanvar", "fevral", "mart", "aprel", "may", "iyun", "iyul", "avqust", "sentyabr", "oktyabr", "noyabr", "dekabr"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod bg_BG {
    pub mod LC_ADDRESS {
        /// `Some("BGR")`
        pub const COUNTRY_AB3: Option<&str> = Some("BGR");
        /// `Some("BG")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("BG");
        /// `Some("BG")`
        pub const COUNTRY_CAR: Option<&str> = Some("BG");
        /// `Some("954")`
        pub const COUNTRY_ISBN: Option<&str> = Some("954");
        /// `Some("България")`
        pub const COUNTRY_NAME: Option<&str> = Some("България");
        /// `Some(100)`
        pub const COUNTRY_NUM: Option<i64> = Some(100);
        /// `Some("BG")`
        pub const COUNTRY_POST: Option<&str> = Some("BG");
        /// `Some("bg")`
        pub const LANG_AB: Option<&str> = Some("bg");
        /// `Some("bul")`
        pub const LANG_LIB: Option<&str> = Some("bul");
        /// `Some("български")`
        pub const LANG_NAME: Option<&str> = Some("български");
        /// `Some("bul")`
        pub const LANG_TERM: Option<&str> = Some("bul");
        /// `Some("%f%N%a%N%d%N%b%N%sN%h, %e, %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%sN%h, %e, %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("develop@linux.zonebg.com")`
        pub const ADDRESS: Option<&str> = Some("develop@linux.zonebg.com");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Delyan Toshev")`
        pub const CONTACT: Option<&str> = Some("Delyan Toshev");
        /// `Some("2002-09-10")`
        pub const DATE: Option<&str> = Some("2002-09-10");
        /// `Some("delyant@yahoo.com")`
        pub const EMAIL: Option<&str> = Some("delyant@yahoo.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Bulgarian")`
        pub const LANGUAGE: Option<&str> = Some("Bulgarian");
        /// `Some("2.0.1")`
        pub const REVISION: Option<&str> = Some("2.0.1");
        /// `Some("Linux Society Bulgaria")`
        pub const SOURCE: Option<&str> = Some("Linux Society Bulgaria");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Bulgaria")`
        pub const TERRITORY: Option<&str> = Some("Bulgaria");
        /// `Some("Bulgarian locale for Bulgaria")`
        pub const TITLE: Option<&str> = Some("Bulgarian locale for Bulgaria");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNkKНн]"`
        pub const NOEXPR: &str = "^[-0nNkKНн]";
        /// `Some("не")`
        pub const NOSTR: Option<&str> = Some("не");
        /// `"^[+1yYdDoOДд]"`
        pub const YESEXPR: &str = "^[+1yYdDoOДд]";
        /// `Some("да")`
        pub const YESSTR: Option<&str> = Some("да");
    }
    pub mod LC_MONETARY {
        /// `"лв."`
        pub const CURRENCY_SYMBOL: &str = "лв.";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"BGN "`
        pub const INT_CURR_SYMBOL: &str = "BGN ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const MON_THOUSANDS_SEP: &str = "\u{202f}";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `""`
        pub const THOUSANDS_SEP: &str = "";
    }
    pub mod LC_TELEPHONE {
        /// `Some("359")`
        pub const INT_PREFIX: Option<&str> = Some("359");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `Some("(0%a) %l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("(0%a) %l");
        /// `Some("(+%c %a) %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("(+%c %a) %l");
    }
    pub mod LC_TIME {
        /// `&["нд", "пн", "вт", "ср", "чт", "пт", "сб"]`
        pub const ABDAY: &[&str] = &["нд", "пн", "вт", "ср", "чт", "пт", "сб"];
        /// `&["яну", "фев", "мар", "апр", "май", "юни", "юли", "авг", "сеп", "окт", "ное", "дек"]`
        pub const ABMON: &[&str] = &["яну", "фев", "мар", "апр", "май", "юни", "юли", "авг", "сеп", "окт", "ное", "дек"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["неделя", "понеделник", "вторник", "сряда", "четвъртък", "петък", "събота"]`
        pub const DAY: &[&str] = &["неделя", "понеделник", "вторник", "сряда", "четвъртък", "петък", "събота"];
        /// `"%e.%m.%Y"`
        pub const D_FMT: &str = "%e.%m.%Y";
        /// `"%e.%m.%Y (%a) %k:%M:%S %Z"`
        pub const D_T_FMT: &str = "%e.%m.%Y (%a) %k:%M:%S %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["януари", "февруари", "март", "април", "май", "юни", "юли", "август", "септември", "октомври", "ноември", "декември"]`
        pub const MON: &[&str] = &["януари", "февруари", "март", "април", "май", "юни", "юли", "август", "септември", "октомври", "ноември", "декември"];
        /// `"%k:%M:%S"`
        pub const T_FMT: &str = "%k:%M:%S";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod bhb_IN {
    pub mod LC_ADDRESS {
        /// `Some("IND")`
        pub const COUNTRY_AB3: Option<&str> = Some("IND");
        /// `Some("IN")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IN");
        /// `Some("IND")`
        pub const COUNTRY_CAR: Option<&str> = Some("IND");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("भारत")`
        pub const COUNTRY_NAME: Option<&str> = Some("भारत");
        /// `Some(356)`
        pub const COUNTRY_NUM: Option<i64> = Some(356);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("")`
        pub const LANG_AB: Option<&str> = Some("");
        /// `Some("bhb")`
        pub const LANG_LIB: Option<&str> = Some("bhb");
        /// `Some("भीली")`
        pub const LANG_NAME: Option<&str> = Some("भीली");
        /// `Some("bhb")`
        pub const LANG_TERM: Option<&str> = Some("bhb");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2014-09-30")`
        pub const DATE: Option<&str> = Some("2014-09-30");
        /// `Some("alexey.merzlyakov@samsung.com")`
        pub const EMAIL: Option<&str> = Some("alexey.merzlyakov@samsung.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Bhili")`
        pub const LANGUAGE: Option<&str> = Some("Bhili");
        /// `Some("0.1")`
        pub const REVISION: Option<&str> = Some("0.1");
        /// `Some("Samsung Electronics Co., Ltd.")`
        pub const SOURCE: Option<&str> = Some("Samsung Electronics Co., Ltd.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("India")`
        pub const TERRITORY: Option<&str> = Some("India");
        /// `Some("Bhili(devanagari) language locale for India")`
        pub const TITLE: Option<&str> = Some("Bhili(devanagari) language locale for India");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `None`
        pub const NOSTR: Option<&str> = None;
        /// `"^[+1yY]"`
        pub const YESEXPR: &str = "^[+1yY]";
        /// `None`
        pub const YESSTR: Option<&str> = None;
    }
    pub use super::hi_IN::LC_MONETARY;
    pub use super::hi_IN::LC_NUMERIC;
    pub use super::hi_IN::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"]`
        pub const ABDAY: &[&str] = &["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"];
        /// `&["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["AM", "PM"]`
        pub const AM_PM: &[&str] = &["AM", "PM"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"]`
        pub const DAY: &[&str] = &["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"];
        /// `"%-d/%-m/%y"`
        pub const D_FMT: &str = "%-d/%-m/%y";
        /// `"%A %d %b %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A %d %b %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"]`
        pub const MON: &[&str] = &["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"];
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT: &str = "%I:%M:%S %p %Z";
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p %Z";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod bho_IN {
    pub mod LC_ADDRESS {
        /// `Some("IND")`
        pub const COUNTRY_AB3: Option<&str> = Some("IND");
        /// `Some("IN")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IN");
        /// `Some("IND")`
        pub const COUNTRY_CAR: Option<&str> = Some("IND");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("भारत")`
        pub const COUNTRY_NAME: Option<&str> = Some("भारत");
        /// `Some(356)`
        pub const COUNTRY_NUM: Option<i64> = Some(356);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("bho")`
        pub const LANG_LIB: Option<&str> = Some("bho");
        /// `Some("भोजप\u{941}री")`
        pub const LANG_NAME: Option<&str> = Some("भोजप\u{941}री");
        /// `Some("bho")`
        pub const LANG_TERM: Option<&str> = Some("bho");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2011-12-20")`
        pub const DATE: Option<&str> = Some("2011-12-20");
        /// `Some("bhashaghar@googlegroups.com")`
        pub const EMAIL: Option<&str> = Some("bhashaghar@googlegroups.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Bhojpuri")`
        pub const LANGUAGE: Option<&str> = Some("Bhojpuri");
        /// `Some("0.1")`
        pub const REVISION: Option<&str> = Some("0.1");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("India")`
        pub const TERRITORY: Option<&str> = Some("India");
        /// `Some("Bhojpuri language locale for India")`
        pub const TITLE: Option<&str> = Some("Bhojpuri language locale for India");
    }
    pub use super::hi_IN::LC_MESSAGES;
    pub use super::hi_IN::LC_MONETARY;
    pub use super::hi_IN::LC_NUMERIC;
    pub use super::hi_IN::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["रवि", "सोम", "म\u{902}गल", "ब\u{941}ध", "ग\u{941}र\u{941}", "श\u{941}क\u{94d}र", "शनि"]`
        pub const ABDAY: &[&str] = &["रवि", "सोम", "म\u{902}गल", "ब\u{941}ध", "ग\u{941}र\u{941}", "श\u{941}क\u{94d}र", "शनि"];
        /// `&["जनवरी", "फरवरी", "मार\u{94d}च", "अप\u{94d}र\u{948}ल", "मई", "ज\u{942}न", "ज\u{941}लाई", "अगस\u{94d}त", "सितम\u{94d}बर", "अक\u{94d}ट\u{942}बर", "नवम\u{94d}बर", "दिसम\u{94d}बर"]`
        pub const ABMON: &[&str] = &["जनवरी", "फरवरी", "मार\u{94d}च", "अप\u{94d}र\u{948}ल", "मई", "ज\u{942}न", "ज\u{941}लाई", "अगस\u{94d}त", "सितम\u{94d}बर", "अक\u{94d}ट\u{942}बर", "नवम\u{94d}बर", "दिसम\u{94d}बर"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["प\u{942}र\u{94d}वाह\u{94d}न", "अपराह\u{94d}न"]`
        pub const AM_PM: &[&str] = &["प\u{942}र\u{94d}वाह\u{94d}न", "अपराह\u{94d}न"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["रविवार", "सोमवार", "म\u{902}गलवार", "ब\u{941}धवार", "ग\u{941}र\u{941}वार", "श\u{941}क\u{94d}रवार", "शनिवार"]`
        pub const DAY: &[&str] = &["रविवार", "सोमवार", "म\u{902}गलवार", "ब\u{941}धवार", "ग\u{941}र\u{941}वार", "श\u{941}क\u{94d}रवार", "शनिवार"];
        /// `"%-d/%-m/%y"`
        pub const D_FMT: &str = "%-d/%-m/%y";
        /// `"%A %d %b %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A %d %b %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["जनवरी", "फरवरी", "मार\u{94d}च", "अप\u{94d}र\u{948}ल", "मई", "ज\u{942}न", "ज\u{941}लाई", "अगस\u{94d}त", "सितम\u{94d}बर", "अक\u{94d}ट\u{942}बर", "नवम\u{94d}बर", "दिसम\u{94d}बर"]`
        pub const MON: &[&str] = &["जनवरी", "फरवरी", "मार\u{94d}च", "अप\u{94d}र\u{948}ल", "मई", "ज\u{942}न", "ज\u{941}लाई", "अगस\u{94d}त", "सितम\u{94d}बर", "अक\u{94d}ट\u{942}बर", "नवम\u{94d}बर", "दिसम\u{94d}बर"];
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT: &str = "%I:%M:%S %p %Z";
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p %Z";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod bho_NP {
    pub mod LC_ADDRESS {
        /// `Some("NPL")`
        pub const COUNTRY_AB3: Option<&str> = Some("NPL");
        /// `Some("NP")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("NP");
        /// `Some("NEP")`
        pub const COUNTRY_CAR: Option<&str> = Some("NEP");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("न\u{947}पाल")`
        pub const COUNTRY_NAME: Option<&str> = Some("न\u{947}पाल");
        /// `Some(524)`
        pub const COUNTRY_NUM: Option<i64> = Some(524);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("bho")`
        pub const LANG_LIB: Option<&str> = Some("bho");
        /// `Some("भोजप\u{941}री")`
        pub const LANG_NAME: Option<&str> = Some("भोजप\u{941}री");
        /// `Some("bho")`
        pub const LANG_TERM: Option<&str> = Some("bho");
        /// `Some("%f%N%h%s%N%T")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%h%s%N%T");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2017-07-24")`
        pub const DATE: Option<&str> = Some("2017-07-24");
        /// `Some("akhilesh.k@samusng.com")`
        pub const EMAIL: Option<&str> = Some("akhilesh.k@samusng.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Bhojpuri")`
        pub const LANGUAGE: Option<&str> = Some("Bhojpuri");
        /// `Some("0.1")`
        pub const REVISION: Option<&str> = Some("0.1");
        /// `Some("Samsung Electronics Co., Ltd.")`
        pub const SOURCE: Option<&str> = Some("Samsung Electronics Co., Ltd.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Nepal")`
        pub const TERRITORY: Option<&str> = Some("Nepal");
        /// `Some("Bhojpuri language locale for Nepal")`
        pub const TITLE: Option<&str> = Some("Bhojpuri language locale for Nepal");
    }
    pub use super::bho_IN::LC_MESSAGES;
    pub use super::ne_NP::LC_MONETARY;
    pub use super::ne_NP::LC_NUMERIC;
    pub use super::ne_NP::LC_TELEPHONE;
    pub use super::bho_IN::LC_TIME;
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod bi_VU {
    pub mod LC_ADDRESS {
        /// `Some("VUT")`
        pub const COUNTRY_AB3: Option<&str> = Some("VUT");
        /// `Some("VU")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("VU");
        /// `Some("VU")`
        pub const COUNTRY_CAR: Option<&str> = Some("VU");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Vanuatu")`
        pub const COUNTRY_NAME: Option<&str> = Some("Vanuatu");
        /// `Some(548)`
        pub const COUNTRY_NUM: Option<i64> = Some(548);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("bi")`
        pub const LANG_AB: Option<&str> = Some("bi");
        /// `Some("bis")`
        pub const LANG_LIB: Option<&str> = Some("bis");
        /// `Some("Bislama")`
        pub const LANG_NAME: Option<&str> = Some("Bislama");
        /// `Some("bis")`
        pub const LANG_TERM: Option<&str> = Some("bis");
        /// `Some("%a%b%s%c")`
        pub const POSTAL_FMT: Option<&str> = Some("%a%b%s%c");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2017-07-13")`
        pub const DATE: Option<&str> = Some("2017-07-13");
        /// `Some("akhilesh.k@samsung.com & maninder1.s@samsung.com")`
        pub const EMAIL: Option<&str> = Some("akhilesh.k@samsung.com & maninder1.s@samsung.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Bislama")`
        pub const LANGUAGE: Option<&str> = Some("Bislama");
        /// `Some("0.1")`
        pub const REVISION: Option<&str> = Some("0.1");
        /// `Some("Samsung Electronics Co., Ltd.")`
        pub const SOURCE: Option<&str> = Some("Samsung Electronics Co., Ltd.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Vanuatu")`
        pub const TERRITORY: Option<&str> = Some("Vanuatu");
        /// `Some("Bislama language locale for Vanuatu")`
        pub const TITLE: Option<&str> = Some("Bislama language locale for Vanuatu");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("No")`
        pub const NOSTR: Option<&str> = Some("No");
        /// `"^[+1yY]"`
        pub const YESEXPR: &str = "^[+1yY]";
        /// `Some("Yes")`
        pub const YESSTR: Option<&str> = Some("Yes");
    }
    pub mod LC_MONETARY {
        /// `"VT"`
        pub const CURRENCY_SYMBOL: &str = "VT";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"VUV "`
        pub const INT_CURR_SYMBOL: &str = "VUV ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `Some(1)`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = Some(1);
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `Some(1)`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = Some(1);
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("678")`
        pub const INT_PREFIX: Option<&str> = Some("678");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %l");
    }
    pub mod LC_TIME {
        /// `&["San", "Man", "Tus", "Wen", "Tas", "Fra", "Sar"]`
        pub const ABDAY: &[&str] = &["San", "Man", "Tus", "Wen", "Tas", "Fra", "Sar"];
        /// `&["Jan", "Feb", "Maj", "Epr", "Mei", "Jun", "Jul", "Ogs", "Sep", "Okt", "Nov", "Dis"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Maj", "Epr", "Mei", "Jun", "Jul", "Ogs", "Sep", "Okt", "Nov", "Dis"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["AM", "PM"]`
        pub const AM_PM: &[&str] = &["AM", "PM"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Sande", "Mande", "Tusde", "Wenesde", "Tosde", "Fraede", "Sarere"]`
        pub const DAY: &[&str] = &["Sande", "Mande", "Tusde", "Wenesde", "Tosde", "Fraede", "Sarere"];
        /// `"%A %d %b %Y"`
        pub const D_FMT: &str = "%A %d %b %Y";
        /// `"%A %d %b %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A %d %b %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Jenuware", "Febwari", "Maj", "Epril", "Mei", "Jun", "Julae", "Ogis", "Septemba", "Oktoba", "Novemba", "Disemba"]`
        pub const MON: &[&str] = &["Jenuware", "Febwari", "Maj", "Epril", "Mei", "Jun", "Julae", "Ogis", "Septemba", "Oktoba", "Novemba", "Disemba"];
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT: &str = "%I:%M:%S %p %Z";
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p %Z";
        /// `None`
        pub const WEEK: Option<&[i64]> = None;
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod bn_BD {
    pub mod LC_ADDRESS {
        /// `Some("BGD")`
        pub const COUNTRY_AB3: Option<&str> = Some("BGD");
        /// `Some("BD")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("BD");
        /// `Some("BD")`
        pub const COUNTRY_CAR: Option<&str> = Some("BD");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("ব\u{9be}ংল\u{9be}দেশ")`
        pub const COUNTRY_NAME: Option<&str> = Some("ব\u{9be}ংল\u{9be}দেশ");
        /// `Some(50)`
        pub const COUNTRY_NUM: Option<i64> = Some(50);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("bn")`
        pub const LANG_AB: Option<&str> = Some("bn");
        /// `Some("ben")`
        pub const LANG_LIB: Option<&str> = Some("ben");
        /// `Some("ব\u{9be}ংল\u{9be}")`
        pub const LANG_NAME: Option<&str> = Some("ব\u{9be}ংল\u{9be}");
        /// `Some("ben")`
        pub const LANG_TERM: Option<&str> = Some("ben");
        /// `Some("%a%N%f%N%d%N%b%N%h %s %e %r%N%T %z%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%a%N%f%N%d%N%b%N%h %s %e %r%N%T %z%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Dhaka, Bangladesh")`
        pub const ADDRESS: Option<&str> = Some("Dhaka, Bangladesh");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_NAME"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_NAME"],
        ]);
        /// `Some("Taneem Ahmed, Jamil Ahmed")`
        pub const CONTACT: Option<&str> = Some("Taneem Ahmed, Jamil Ahmed");
        /// `Some("2007-01-10")`
        pub const DATE: Option<&str> = Some("2007-01-10");
        /// `Some("taneem@bengalinux.org, jamil@bengalinux.org")`
        pub const EMAIL: Option<&str> = Some("taneem@bengalinux.org, jamil@bengalinux.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Bangla")`
        pub const LANGUAGE: Option<&str> = Some("Bangla");
        /// `Some("0.5")`
        pub const REVISION: Option<&str> = Some("0.5");
        /// `Some("Ankur Group, http://www.ankurbangla.org, http://www.bengalinux.org")`
        pub const SOURCE: Option<&str> = Some("Ankur Group, http://www.ankurbangla.org, http://www.bengalinux.org");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Bangladesh")`
        pub const TERRITORY: Option<&str> = Some("Bangladesh");
        /// `Some("Bangla language locale for Bangladesh")`
        pub const TITLE: Option<&str> = Some("Bangla language locale for Bangladesh");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNন]"`
        pub const NOEXPR: &str = "^[-0nNন]";
        /// `Some("ন\u{9be}")`
        pub const NOSTR: Option<&str> = Some("ন\u{9be}");
        /// `"^[+1yYহ]"`
        pub const YESEXPR: &str = "^[+1yYহ]";
        /// `Some("হ\u{9cd}য\u{9be}\u{981}")`
        pub const YESSTR: Option<&str> = Some("হ\u{9cd}য\u{9be}\u{981}");
    }
    pub mod LC_MONETARY {
        /// `"৳"`
        pub const CURRENCY_SYMBOL: &str = "৳";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"BDT "`
        pub const INT_CURR_SYMBOL: &str = "BDT ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 2]`
        pub const MON_GROUPING: &[i64] = &[3, 2];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3, 2]`
        pub const GROUPING: &[i64] = &[3, 2];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("880")`
        pub const INT_PREFIX: Option<&str> = Some("880");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `Some("%a %l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("%a %l");
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["রবি", "সোম", "মঙ\u{9cd}গল", "ব\u{9c1}ধ", "ব\u{9c3}হঃ", "শ\u{9c1}ক\u{9cd}র", "শনি"]`
        pub const ABDAY: &[&str] = &["রবি", "সোম", "মঙ\u{9cd}গল", "ব\u{9c1}ধ", "ব\u{9c3}হঃ", "শ\u{9c1}ক\u{9cd}র", "শনি"];
        /// `&["জ\u{9be}ন\u{9c1}", "ফেব", "ম\u{9be}র\u{9cd}চ", "এপ\u{9cd}রিল", "মে", "জ\u{9c1}ন", "জ\u{9c1}ল\u{9be}ই", "আগস\u{9cd}ট", "সেপ\u{9cd}টেম\u{9cd}বর", "অক\u{9cd}টোবর", "নভেম\u{9cd}বর", "ডিসেম\u{9cd}বর"]`
        pub const ABMON: &[&str] = &["জ\u{9be}ন\u{9c1}", "ফেব", "ম\u{9be}র\u{9cd}চ", "এপ\u{9cd}রিল", "মে", "জ\u{9c1}ন", "জ\u{9c1}ল\u{9be}ই", "আগস\u{9cd}ট", "সেপ\u{9cd}টেম\u{9cd}বর", "অক\u{9cd}টোবর", "নভেম\u{9cd}বর", "ডিসেম\u{9cd}বর"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["প\u{9c2}র\u{9cd}ব\u{9be}হ\u{9cd}ণ", "অপর\u{9be}হ\u{9cd}ণ"]`
        pub const AM_PM: &[&str] = &["প\u{9c2}র\u{9cd}ব\u{9be}হ\u{9cd}ণ", "অপর\u{9be}হ\u{9cd}ণ"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["রবিব\u{9be}র", "সোমব\u{9be}র", "মঙ\u{9cd}গলব\u{9be}র", "ব\u{9c1}ধব\u{9be}র", "ব\u{9c3}হস\u{9cd}পতিব\u{9be}র", "শ\u{9c1}ক\u{9cd}রব\u{9be}র", "শনিব\u{9be}র"]`
        pub const DAY: &[&str] = &["রবিব\u{9be}র", "সোমব\u{9be}র", "মঙ\u{9cd}গলব\u{9be}র", "ব\u{9c1}ধব\u{9be}র", "ব\u{9c3}হস\u{9cd}পতিব\u{9be}র", "শ\u{9c1}ক\u{9cd}রব\u{9be}র", "শনিব\u{9be}র"];
        /// `"%-d/%-m/%y"`
        pub const D_FMT: &str = "%-d/%-m/%y";
        /// `"%A %d %b %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A %d %b %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(6)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(6);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["জ\u{9be}ন\u{9c1}য\u{9bc}\u{9be}রী", "ফেব\u{9cd}র\u{9c1}য\u{9bc}\u{9be}রী", "ম\u{9be}র\u{9cd}চ", "এপ\u{9cd}রিল", "মে", "জ\u{9c1}ন", "জ\u{9c1}ল\u{9be}ই", "আগস\u{9cd}ট", "সেপ\u{9cd}টেম\u{9cd}বর", "অক\u{9cd}টোবর", "নভেম\u{9cd}বর", "ডিসেম\u{9cd}বর"]`
        pub const MON: &[&str] = &["জ\u{9be}ন\u{9c1}য\u{9bc}\u{9be}রী", "ফেব\u{9cd}র\u{9c1}য\u{9bc}\u{9be}রী", "ম\u{9be}র\u{9cd}চ", "এপ\u{9cd}রিল", "মে", "জ\u{9c1}ন", "জ\u{9c1}ল\u{9be}ই", "আগস\u{9cd}ট", "সেপ\u{9cd}টেম\u{9cd}বর", "অক\u{9cd}টোবর", "নভেম\u{9cd}বর", "ডিসেম\u{9cd}বর"];
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT: &str = "%I:%M:%S %p %Z";
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p %Z";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod bn_IN {
    pub mod LC_ADDRESS {
        /// `Some("IND")`
        pub const COUNTRY_AB3: Option<&str> = Some("IND");
        /// `Some("IN")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IN");
        /// `Some("IND")`
        pub const COUNTRY_CAR: Option<&str> = Some("IND");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("ভ\u{9be}রত")`
        pub const COUNTRY_NAME: Option<&str> = Some("ভ\u{9be}রত");
        /// `Some(356)`
        pub const COUNTRY_NUM: Option<i64> = Some(356);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("bn")`
        pub const LANG_AB: Option<&str> = Some("bn");
        /// `Some("ben")`
        pub const LANG_LIB: Option<&str> = Some("ben");
        /// `Some("ব\u{9be}ংল\u{9be}")`
        pub const LANG_NAME: Option<&str> = Some("ব\u{9be}ংল\u{9be}");
        /// `Some("ben")`
        pub const LANG_TERM: Option<&str> = Some("ben");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan")`
        pub const ADDRESS: Option<&str> = Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `None`
        pub const CONTACT: Option<&str> = None;
        /// `Some("2006-05-29")`
        pub const DATE: Option<&str> = Some("2006-05-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Bangla")`
        pub const LANGUAGE: Option<&str> = Some("Bangla");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("IBM Globalization Center of Competency, Yamato Software Laboratory")`
        pub const SOURCE: Option<&str> = Some("IBM Globalization Center of Competency, Yamato Software Laboratory");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("India")`
        pub const TERRITORY: Option<&str> = Some("India");
        /// `Some("Bangla language locale for India")`
        pub const TITLE: Option<&str> = Some("Bangla language locale for India");
    }
    pub use super::bn_BD::LC_MESSAGES;
    pub use super::hi_IN::LC_MONETARY;
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3, 2]`
        pub const GROUPING: &[i64] = &[3, 2];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub use super::hi_IN::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["রবি", "সোম", "মঙ\u{9cd}গল", "ব\u{9c1}ধ", "ব\u{9c3}হস\u{9cd}পতি", "শ\u{9c1}ক\u{9cd}র", "শনি"]`
        pub const ABDAY: &[&str] = &["রবি", "সোম", "মঙ\u{9cd}গল", "ব\u{9c1}ধ", "ব\u{9c3}হস\u{9cd}পতি", "শ\u{9c1}ক\u{9cd}র", "শনি"];
        /// `&["জ\u{9be}ন\u{9c1}", "ফেব", "ম\u{9be}র\u{9cd}চ", "এপ\u{9cd}রিল", "মে", "জ\u{9c1}ন", "জ\u{9c1}ল\u{9be}ই", "আগস\u{9cd}ট", "সেপ\u{9cd}টেম\u{9cd}বর", "অক\u{9cd}টোবর", "নভেম\u{9cd}বর", "ডিসেম\u{9cd}বর"]`
        pub const ABMON: &[&str] = &["জ\u{9be}ন\u{9c1}", "ফেব", "ম\u{9be}র\u{9cd}চ", "এপ\u{9cd}রিল", "মে", "জ\u{9c1}ন", "জ\u{9c1}ল\u{9be}ই", "আগস\u{9cd}ট", "সেপ\u{9cd}টেম\u{9cd}বর", "অক\u{9cd}টোবর", "নভেম\u{9cd}বর", "ডিসেম\u{9cd}বর"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["প\u{9c2}র\u{9cd}ব\u{9be}হ\u{9cd}ণ", "অপর\u{9be}হ\u{9cd}ণ"]`
        pub const AM_PM: &[&str] = &["প\u{9c2}র\u{9cd}ব\u{9be}হ\u{9cd}ণ", "অপর\u{9be}হ\u{9cd}ণ"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["রবিব\u{9be}র", "সোমব\u{9be}র", "মঙ\u{9cd}গলব\u{9be}র", "ব\u{9c1}ধব\u{9be}র", "ব\u{9c3}হস\u{9cd}পতিব\u{9be}র", "শ\u{9c1}ক\u{9cd}রব\u{9be}র", "শনিব\u{9be}র"]`
        pub const DAY: &[&str] = &["রবিব\u{9be}র", "সোমব\u{9be}র", "মঙ\u{9cd}গলব\u{9be}র", "ব\u{9c1}ধব\u{9be}র", "ব\u{9c3}হস\u{9cd}পতিব\u{9be}র", "শ\u{9c1}ক\u{9cd}রব\u{9be}র", "শনিব\u{9be}র"];
        /// `"%-d/%-m/%y"`
        pub const D_FMT: &str = "%-d/%-m/%y";
        /// `"%A %d %b %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A %d %b %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["জ\u{9be}ন\u{9c1}য\u{9bc}\u{9be}রী", "ফেব\u{9cd}র\u{9c1}য\u{9bc}\u{9be}রী", "ম\u{9be}র\u{9cd}চ", "এপ\u{9cd}রিল", "মে", "জ\u{9c1}ন", "জ\u{9c1}ল\u{9be}ই", "আগস\u{9cd}ট", "সেপ\u{9cd}টেম\u{9cd}বর", "অক\u{9cd}টোবর", "নভেম\u{9cd}বর", "ডিসেম\u{9cd}বর"]`
        pub const MON: &[&str] = &["জ\u{9be}ন\u{9c1}য\u{9bc}\u{9be}রী", "ফেব\u{9cd}র\u{9c1}য\u{9bc}\u{9be}রী", "ম\u{9be}র\u{9cd}চ", "এপ\u{9cd}রিল", "মে", "জ\u{9c1}ন", "জ\u{9c1}ল\u{9be}ই", "আগস\u{9cd}ট", "সেপ\u{9cd}টেম\u{9cd}বর", "অক\u{9cd}টোবর", "নভেম\u{9cd}বর", "ডিসেম\u{9cd}বর"];
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT: &str = "%I:%M:%S %p %Z";
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p %Z";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod bo_CN {
    pub mod LC_ADDRESS {
        /// `Some("CHN")`
        pub const COUNTRY_AB3: Option<&str> = Some("CHN");
        /// `Some("CN")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("CN");
        /// `Some("CHN")`
        pub const COUNTRY_CAR: Option<&str> = Some("CHN");
        /// `Some("7")`
        pub const COUNTRY_ISBN: Option<&str> = Some("7");
        /// `Some("ར\u{f92}\u{fb1}་ནག")`
        pub const COUNTRY_NAME: Option<&str> = Some("ར\u{f92}\u{fb1}་ནག");
        /// `Some(156)`
        pub const COUNTRY_NUM: Option<i64> = Some(156);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("bo")`
        pub const LANG_AB: Option<&str> = Some("bo");
        /// `Some("tib")`
        pub const LANG_LIB: Option<&str> = Some("tib");
        /// `Some("བ\u{f7c}ད་ས\u{f90}ད་")`
        pub const LANG_NAME: Option<&str> = Some("བ\u{f7c}ད་ས\u{f90}ད་");
        /// `Some("bod")`
        pub const LANG_TERM: Option<&str> = Some("bod");
        /// `Some("%c%N%T%N%s %h %e %r%N%b%N%d%N%f%N%a%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%c%N%T%N%s %h %e %r%N%b%N%d%N%f%N%a%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2007-11-06")`
        pub const DATE: Option<&str> = Some("2007-11-06");
        /// `Some("bug-glibc@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Tibetan")`
        pub const LANGUAGE: Option<&str> = Some("Tibetan");
        /// `Some("0.1")`
        pub const REVISION: Option<&str> = Some("0.1");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("China")`
        pub const TERRITORY: Option<&str> = Some("China");
        /// `Some("Tibetan language locale for P.R. of China")`
        pub const TITLE: Option<&str> = Some("Tibetan language locale for P.R. of China");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNམ]"`
        pub const NOEXPR: &str = "^[-0nNམ]";
        /// `Some("མ\u{f72}ན།")`
        pub const NOSTR: Option<&str> = Some("མ\u{f72}ན།");
        /// `"^[+1yYཨ]"`
        pub const YESEXPR: &str = "^[+1yYཨ]";
        /// `Some("ཡ\u{f72}ན།")`
        pub const YESSTR: Option<&str> = Some("ཡ\u{f72}ན།");
    }
    pub use super::zh_CN::LC_MONETARY;
    pub use super::zh_CN::LC_NUMERIC;
    pub use super::zh_CN::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["ཉ\u{f72}་", "ཟ\u{fb3}་", "མ\u{f72}ར་", "ལ\u{fb7}ག་", "པ\u{f74}ར་", "སངས་", "ས\u{fa4}\u{f7a}ན་"]`
        pub const ABDAY: &[&str] = &["ཉ\u{f72}་", "ཟ\u{fb3}་", "མ\u{f72}ར་", "ལ\u{fb7}ག་", "པ\u{f74}ར་", "སངས་", "ས\u{fa4}\u{f7a}ན་"];
        /// `&["ཟ\u{fb3}་༡", "ཟ\u{fb3}་༢", "ཟ\u{fb3}་༣", "ཟ\u{fb3}་༤", "ཟ\u{fb3}་༥", "ཟ\u{fb3}་༦", "ཟ\u{fb3}་༧", "ཟ\u{fb3}་༨", "ཟ\u{fb3}་༩", "ཟ\u{fb3}་༡༠", "ཟ\u{fb3}་༡༡", "ཟ\u{fb3}་༡༢"]`
        pub const ABMON: &[&str] = &["ཟ\u{fb3}་༡", "ཟ\u{fb3}་༢", "ཟ\u{fb3}་༣", "ཟ\u{fb3}་༤", "ཟ\u{fb3}་༥", "ཟ\u{fb3}་༦", "ཟ\u{fb3}་༧", "ཟ\u{fb3}་༨", "ཟ\u{fb3}་༩", "ཟ\u{fb3}་༡༠", "ཟ\u{fb3}་༡༡", "ཟ\u{fb3}་༡༢"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["ངས་ཆ", "ཕ\u{fb1}\u{f72}་ཆ"]`
        pub const AM_PM: &[&str] = &["ངས་ཆ", "ཕ\u{fb1}\u{f72}་ཆ"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["གཟའ་ཉ\u{f72}་མ་", "གཟའ་ཟ\u{fb3}་བ་", "གཟའ་མ\u{f72}ག་དམར་", "གཟའ་ལ\u{fb7}ག་ཕ་", "གཟའ་པ\u{f74}ར་བ\u{f74}་", "གཟའ་པ་སངས་", "གཟའ་ས\u{fa4}\u{f7a}ན་ཕ་"]`
        pub const DAY: &[&str] = &["གཟའ་ཉ\u{f72}་མ་", "གཟའ་ཟ\u{fb3}་བ་", "གཟའ་མ\u{f72}ག་དམར་", "གཟའ་ལ\u{fb7}ག་ཕ་", "གཟའ་པ\u{f74}ར་བ\u{f74}་", "གཟའ་པ་སངས་", "གཟའ་ས\u{fa4}\u{f7a}ན་ཕ་"];
        /// `"པས\u{fb1}\u{f72}་ལ\u{f7c}%yཟལ%mཚ\u{f7a}ས%d"`
        pub const D_FMT: &str = "པས\u{fb1}\u{f72}་ལ\u{f7c}%yཟལ%mཚ\u{f7a}ས%d";
        /// `"པས\u{fb1}\u{f72}་ལ\u{f7c}%yཟལ%mཚ\u{f7a}ས%dཆ\u{f74}་ཚ\u{f7c}ད%Hཀསར་མ%Mཀསར་ཆ%S"`
        pub const D_T_FMT: &str = "པས\u{fb1}\u{f72}་ལ\u{f7c}%yཟལ%mཚ\u{f7a}ས%dཆ\u{f74}་ཚ\u{f7c}ད%Hཀསར་མ%Mཀསར་ཆ%S";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["ཟ\u{fb3}་བ་དང་པ་", "ཟ\u{fb3}་བ་གཉ\u{f72}ས་པ་", "ཟ\u{fb3}་བ་གས\u{f74}མ་པ་", "ཟ\u{fb3}་བ་བཞ\u{f72}་པ་", "ཟ\u{fb3}་བ་ལ\u{f94}་ཕ་", "ཟ\u{fb3}་བ་ད\u{fb2}\u{f74}ག་པ་", "ཟ\u{fb3}་བ་བད\u{f74}ནཔ་", "ཟ\u{fb3}་བ་བར\u{f92}\u{fb1}ད་པ་", "ཟ\u{fb3}་བ་དག\u{f74}་པ་", "ཟ\u{fb3}་བ་བཅ\u{f74}་པ་", "ཟ\u{fb3}་བ་བཅ\u{f74}་གཅ\u{f72}ག་པ་", "ཟ\u{fb3}་བ་བཅ\u{f74}་གཉ\u{f72}ས་པ་"]`
        pub const MON: &[&str] = &["ཟ\u{fb3}་བ་དང་པ་", "ཟ\u{fb3}་བ་གཉ\u{f72}ས་པ་", "ཟ\u{fb3}་བ་གས\u{f74}མ་པ་", "ཟ\u{fb3}་བ་བཞ\u{f72}་པ་", "ཟ\u{fb3}་བ་ལ\u{f94}་ཕ་", "ཟ\u{fb3}་བ་ད\u{fb2}\u{f74}ག་པ་", "ཟ\u{fb3}་བ་བད\u{f74}ནཔ་", "ཟ\u{fb3}་བ་བར\u{f92}\u{fb1}ད་པ་", "ཟ\u{fb3}་བ་དག\u{f74}་པ་", "ཟ\u{fb3}་བ་བཅ\u{f74}་པ་", "ཟ\u{fb3}་བ་བཅ\u{f74}་གཅ\u{f72}ག་པ་", "ཟ\u{fb3}་བ་བཅ\u{f74}་གཉ\u{f72}ས་པ་"];
        /// `"ཆ\u{f74}་ཚ\u{f7c}ད%Hཀསར་མ%Mཀསར་ཆ%S"`
        pub const T_FMT: &str = "ཆ\u{f74}་ཚ\u{f7c}ད%Hཀསར་མ%Mཀསར་ཆ%S";
        /// `"ཆ\u{f74}་ཚ\u{f7c}ད%Iཀསར་མ%Mཀསར་ཆ%S %p"`
        pub const T_FMT_AMPM: &str = "ཆ\u{f74}་ཚ\u{f7c}ད%Iཀསར་མ%Mཀསར་ཆ%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod bo_IN {
    pub mod LC_ADDRESS {
        /// `Some("IND")`
        pub const COUNTRY_AB3: Option<&str> = Some("IND");
        /// `Some("IN")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IN");
        /// `Some("IND")`
        pub const COUNTRY_CAR: Option<&str> = Some("IND");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("ར\u{f92}\u{fb1}་གར་")`
        pub const COUNTRY_NAME: Option<&str> = Some("ར\u{f92}\u{fb1}་གར་");
        /// `Some(356)`
        pub const COUNTRY_NUM: Option<i64> = Some(356);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("bo")`
        pub const LANG_AB: Option<&str> = Some("bo");
        /// `Some("tib")`
        pub const LANG_LIB: Option<&str> = Some("tib");
        /// `Some("བ\u{f7c}ད་ས\u{f90}ད་")`
        pub const LANG_NAME: Option<&str> = Some("བ\u{f7c}ད་ས\u{f90}ད་");
        /// `Some("bod")`
        pub const LANG_TERM: Option<&str> = Some("bod");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2007-11-06")`
        pub const DATE: Option<&str> = Some("2007-11-06");
        /// `Some("bug-glibc@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Tibetan")`
        pub const LANGUAGE: Option<&str> = Some("Tibetan");
        /// `Some("0.1")`
        pub const REVISION: Option<&str> = Some("0.1");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("India")`
        pub const TERRITORY: Option<&str> = Some("India");
        /// `Some("Tibetan language locale for India")`
        pub const TITLE: Option<&str> = Some("Tibetan language locale for India");
    }
    pub use super::bo_CN::LC_MESSAGES;
    pub use super::hi_IN::LC_MONETARY;
    pub use super::bo_CN::LC_NUMERIC;
    pub use super::hi_IN::LC_TELEPHONE;
    pub use super::bo_CN::LC_TIME;
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod br_FR {
    pub mod LC_ADDRESS {
        /// `Some("FRA")`
        pub const COUNTRY_AB3: Option<&str> = Some("FRA");
        /// `Some("FR")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("FR");
        /// `Some("F")`
        pub const COUNTRY_CAR: Option<&str> = Some("F");
        /// `Some("979-10")`
        pub const COUNTRY_ISBN: Option<&str> = Some("979-10");
        /// `Some("Frañs")`
        pub const COUNTRY_NAME: Option<&str> = Some("Frañs");
        /// `Some(250)`
        pub const COUNTRY_NUM: Option<i64> = Some(250);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("br")`
        pub const LANG_AB: Option<&str> = Some("br");
        /// `Some("bre")`
        pub const LANG_LIB: Option<&str> = Some("bre");
        /// `Some("brezhoneg")`
        pub const LANG_NAME: Option<&str> = Some("brezhoneg");
        /// `Some("bre")`
        pub const LANG_TERM: Option<&str> = Some("bre");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
        ]);
        /// `Some("Thierry Vignaud")`
        pub const CONTACT: Option<&str> = Some("Thierry Vignaud");
        /// `Some("2001-01-28")`
        pub const DATE: Option<&str> = Some("2001-01-28");
        /// `Some("thierry.vignaud@gmail.com")`
        pub const EMAIL: Option<&str> = Some("thierry.vignaud@gmail.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Breton")`
        pub const LANGUAGE: Option<&str> = Some("Breton");
        /// `Some("0.54")`
        pub const REVISION: Option<&str> = Some("0.54");
        /// `Some("Denise.Peden@enst-bretagne.fr (Denise Derrien-Peden)")`
        pub const SOURCE: Option<&str> = Some("Denise.Peden@enst-bretagne.fr (Denise Derrien-Peden)");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("France")`
        pub const TERRITORY: Option<&str> = Some("France");
        /// `Some("Breton language locale for France")`
        pub const TITLE: Option<&str> = Some("Breton language locale for France");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0kKnN]"`
        pub const NOEXPR: &str = "^[-0kKnN]";
        /// `Some("ket")`
        pub const NOSTR: Option<&str> = Some("ket");
        /// `"^[+1oOyY]"`
        pub const YESEXPR: &str = "^[+1oOyY]";
        /// `Some("ya")`
        pub const YESSTR: Option<&str> = Some("ya");
    }
    pub use super::fr_FR::LC_MONETARY;
    pub use super::fr_FR::LC_NUMERIC;
    pub use super::fr_FR::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["sul", "lun", "meu", "mer", "yao", "gwe", "sad"]`
        pub const ABDAY: &[&str] = &["sul", "lun", "meu", "mer", "yao", "gwe", "sad"];
        /// `&["Gen ", "Cʼhw", "Meu ", "Ebr ", "Mae ", "Mezh", "Gou ", "Eos ", "Gwe ", "Her ", "Du  ", "Ker "]`
        pub const ABMON: &[&str] = &["Gen ", "Cʼhw", "Meu ", "Ebr ", "Mae ", "Mezh", "Gou ", "Eos ", "Gwe ", "Her ", "Du  ", "Ker "];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&[" ", " "]`
        pub const AM_PM: &[&str] = &[" ", " "];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["sul", "lun", "meurzh", "mercʼher", "yaou", "gwener", "sadorn"]`
        pub const DAY: &[&str] = &["sul", "lun", "meurzh", "mercʼher", "yaou", "gwener", "sadorn"];
        /// `"%d.%m.%Y"`
        pub const D_FMT: &str = "%d.%m.%Y";
        /// `"Dʼar %A %d a viz %B %Y"`
        pub const D_T_FMT: &str = "Dʼar %A %d a viz %B %Y";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Genver", "Cʼhwevrer", "Meurzh", "Ebrel", "Mae", "Mezheven", "Gouere", "Eost", "Gwengolo", "Here", "Du", "Kerzu"]`
        pub const MON: &[&str] = &["Genver", "Cʼhwevrer", "Meurzh", "Ebrel", "Mae", "Mezheven", "Gouere", "Eost", "Gwengolo", "Here", "Du", "Kerzu"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `"%Ie%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%Ie%M:%S %p";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod br_FR_euro {
    pub use super::br_FR::LC_ADDRESS;
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("https://www.gnu.org/software/libc/")`
        pub const ADDRESS: Option<&str> = Some("https://www.gnu.org/software/libc/");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `None`
        pub const CONTACT: Option<&str> = None;
        /// `Some("2002-02-28")`
        pub const DATE: Option<&str> = Some("2002-02-28");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Breton")`
        pub const LANGUAGE: Option<&str> = Some("Breton");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Free Software Foundation, Inc.")`
        pub const SOURCE: Option<&str> = Some("Free Software Foundation, Inc.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("France")`
        pub const TERRITORY: Option<&str> = Some("France");
        /// `Some("Breton locale for France with Euro")`
        pub const TITLE: Option<&str> = Some("Breton locale for France with Euro");
    }
    pub use super::br_FR::LC_MESSAGES;
    pub use super::br_FR::LC_MONETARY;
    pub use super::br_FR::LC_NUMERIC;
    pub use super::br_FR::LC_TELEPHONE;
    pub use super::br_FR::LC_TIME;
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod brx_IN {
    pub mod LC_ADDRESS {
        /// `Some("IND")`
        pub const COUNTRY_AB3: Option<&str> = Some("IND");
        /// `Some("IN")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IN");
        /// `Some("IND")`
        pub const COUNTRY_CAR: Option<&str> = Some("IND");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("भारत")`
        pub const COUNTRY_NAME: Option<&str> = Some("भारत");
        /// `Some(356)`
        pub const COUNTRY_NUM: Option<i64> = Some(356);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("brx")`
        pub const LANG_LIB: Option<&str> = Some("brx");
        /// `Some("बड\u{93c}ो")`
        pub const LANG_NAME: Option<&str> = Some("बड\u{93c}ो");
        /// `Some("brx")`
        pub const LANG_TERM: Option<&str> = Some("brx");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Level 1, Tower X, Cybercity, Magarpatta City, Hadapsar, Pune-411013 ")`
        pub const ADDRESS: Option<&str> = Some("Level 1, Tower X, Cybercity, Magarpatta City, Hadapsar, Pune-411013 ");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2011-10-17")`
        pub const DATE: Option<&str> = Some("2011-10-17");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Bodo")`
        pub const LANGUAGE: Option<&str> = Some("Bodo");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Red Hat Pune")`
        pub const SOURCE: Option<&str> = Some("Red Hat Pune");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("India")`
        pub const TERRITORY: Option<&str> = Some("India");
        /// `Some("Bodo language locale for India")`
        pub const TITLE: Option<&str> = Some("Bodo language locale for India");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNन]"`
        pub const NOEXPR: &str = "^[-0nNन]";
        /// `Some("नही\u{902}")`
        pub const NOSTR: Option<&str> = Some("नही\u{902}");
        /// `"^[+1yYह]"`
        pub const YESEXPR: &str = "^[+1yYह]";
        /// `Some("हा\u{901}")`
        pub const YESSTR: Option<&str> = Some("हा\u{901}");
    }
    pub use super::hi_IN::LC_MONETARY;
    pub use super::hi_IN::LC_NUMERIC;
    pub use super::hi_IN::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["रबि", "सम", "म\u{902}गल", "ब\u{941}द", "बिसथि", "स\u{941}ख\u{941}र", "स\u{941}नि"]`
        pub const ABDAY: &[&str] = &["रबि", "सम", "म\u{902}गल", "ब\u{941}द", "बिसथि", "स\u{941}ख\u{941}र", "स\u{941}नि"];
        /// `&["जान\u{941}वारी", "फ\u{947}ब\u{94d}र\u{941}वारी", "मार\u{94d}स", "एप\u{94d}रिल", "म\u{947}", "ज\u{941}न", "ज\u{941}लाइ", "आगस\u{94d}थ", "स\u{947}बथ\u{947}ज\u{94d}ब\u{93c}र", "अखथबर", "नब\u{947}ज\u{94d}ब\u{93c}र", "दिस\u{947}ज\u{94d}ब\u{93c}र"]`
        pub const ABMON: &[&str] = &["जान\u{941}वारी", "फ\u{947}ब\u{94d}र\u{941}वारी", "मार\u{94d}स", "एप\u{94d}रिल", "म\u{947}", "ज\u{941}न", "ज\u{941}लाइ", "आगस\u{94d}थ", "स\u{947}बथ\u{947}ज\u{94d}ब\u{93c}र", "अखथबर", "नब\u{947}ज\u{94d}ब\u{93c}र", "दिस\u{947}ज\u{94d}ब\u{93c}र"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["फ\u{941}\u{902}.", "ब\u{947}लास\u{947}."]`
        pub const AM_PM: &[&str] = &["फ\u{941}\u{902}.", "ब\u{947}लास\u{947}."];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["रबिबार", "सोबार", "म\u{902}गलबार", "ब\u{941}दबार", "बिसथिबार", "स\u{941}ख\u{941}रबार", "स\u{941}निबार"]`
        pub const DAY: &[&str] = &["रबिबार", "सोबार", "म\u{902}गलबार", "ब\u{941}दबार", "बिसथिबार", "स\u{941}ख\u{941}रबार", "स\u{941}निबार"];
        /// `"%-m/%-d/%y"`
        pub const D_FMT: &str = "%-m/%-d/%y";
        /// `"%A %d %b %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A %d %b %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["जान\u{941}वारी", "फ\u{947}ब\u{94d}र\u{941}वारी", "मार\u{94d}स", "एफ\u{94d}रिल", "म\u{947}", "ज\u{941}न", "ज\u{941}लाइ", "आगस\u{94d}थ", "स\u{947}बथ\u{947}ज\u{94d}ब\u{93c}र", "अखथबर", "नब\u{947}ज\u{94d}ब\u{93c}र", "दिस\u{947}ज\u{94d}ब\u{93c}र"]`
        pub const MON: &[&str] = &["जान\u{941}वारी", "फ\u{947}ब\u{94d}र\u{941}वारी", "मार\u{94d}स", "एफ\u{94d}रिल", "म\u{947}", "ज\u{941}न", "ज\u{941}लाइ", "आगस\u{94d}थ", "स\u{947}बथ\u{947}ज\u{94d}ब\u{93c}र", "अखथबर", "नब\u{947}ज\u{94d}ब\u{93c}र", "दिस\u{947}ज\u{94d}ब\u{93c}र"];
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT: &str = "%I:%M:%S %p %Z";
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p %Z";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod bs_BA {
    pub mod LC_ADDRESS {
        /// `Some("BIH")`
        pub const COUNTRY_AB3: Option<&str> = Some("BIH");
        /// `Some("BA")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("BA");
        /// `Some("BIH")`
        pub const COUNTRY_CAR: Option<&str> = Some("BIH");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Bosna i Hercegovina")`
        pub const COUNTRY_NAME: Option<&str> = Some("Bosna i Hercegovina");
        /// `Some(70)`
        pub const COUNTRY_NUM: Option<i64> = Some(70);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("bs")`
        pub const LANG_AB: Option<&str> = Some("bs");
        /// `Some("bos")`
        pub const LANG_LIB: Option<&str> = Some("bos");
        /// `Some("bosanski")`
        pub const LANG_NAME: Option<&str> = Some("bosanski");
        /// `Some("bos")`
        pub const LANG_TERM: Option<&str> = Some("bos");
        /// `Some("%a%N%f%N%d%N%b%N%h %s %e %r%N%T %z%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%a%N%f%N%d%N%b%N%h %s %e %r%N%T %z%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
        ]);
        /// `Some("Tomislav Vujec")`
        pub const CONTACT: Option<&str> = Some("Tomislav Vujec");
        /// `Some("2004-01-09")`
        pub const DATE: Option<&str> = Some("2004-01-09");
        /// `Some("tvujec@carnet.hr")`
        pub const EMAIL: Option<&str> = Some("tvujec@carnet.hr");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Bosnian")`
        pub const LANGUAGE: Option<&str> = Some("Bosnian");
        /// `Some("0.4")`
        pub const REVISION: Option<&str> = Some("0.4");
        /// `Some("adapted from Croatian locale")`
        pub const SOURCE: Option<&str> = Some("adapted from Croatian locale");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Bosnia & Herzegovina")`
        pub const TERRITORY: Option<&str> = Some("Bosnia & Herzegovina");
        /// `Some("Bosnian language locale for Bosnia and Herzegowina")`
        pub const TITLE: Option<&str> = Some("Bosnian language locale for Bosnia and Herzegowina");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("ne")`
        pub const NOSTR: Option<&str> = Some("ne");
        /// `"^[+1dDyY]"`
        pub const YESEXPR: &str = "^[+1dDyY]";
        /// `Some("da")`
        pub const YESSTR: Option<&str> = Some("da");
    }
    pub mod LC_MONETARY {
        /// `"KM"`
        pub const CURRENCY_SYMBOL: &str = "KM";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"BAM "`
        pub const INT_CURR_SYMBOL: &str = "BAM ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const MON_THOUSANDS_SEP: &str = "\u{202f}";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[0, 0]`
        pub const GROUPING: &[i64] = &[0, 0];
        /// `""`
        pub const THOUSANDS_SEP: &str = "";
    }
    pub mod LC_TELEPHONE {
        /// `Some("387")`
        pub const INT_PREFIX: Option<&str> = Some("387");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["Ned", "Pon", "Uto", "Sri", "Čet", "Pet", "Sub"]`
        pub const ABDAY: &[&str] = &["Ned", "Pon", "Uto", "Sri", "Čet", "Pet", "Sub"];
        /// `&["Jan", "Feb", "Mar", "Apr", "Maj", "Jun", "Jul", "Aug", "Sep", "Okt", "Nov", "Dec"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Mar", "Apr", "Maj", "Jun", "Jul", "Aug", "Sep", "Okt", "Nov", "Dec"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Nedjelja", "Ponedjeljak", "Utorak", "Srijeda", "Četvrtak", "Petak", "Subota"]`
        pub const DAY: &[&str] = &["Nedjelja", "Ponedjeljak", "Utorak", "Srijeda", "Četvrtak", "Petak", "Subota"];
        /// `"%d.%m.%Y"`
        pub const D_FMT: &str = "%d.%m.%Y";
        /// `"%a %d %b %Y %T"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Januar", "Februar", "Mart", "April", "Maj", "Juni", "Juli", "August", "Septembar", "Oktobar", "Novembar", "Decembar"]`
        pub const MON: &[&str] = &["Januar", "Februar", "Mart", "April", "Maj", "Juni", "Juli", "August", "Septembar", "Oktobar", "Novembar", "Decembar"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod byn_ER {
    pub mod LC_ADDRESS {
        /// `Some("ERI")`
        pub const COUNTRY_AB3: Option<&str> = Some("ERI");
        /// `Some("ER")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("ER");
        /// `Some("ER")`
        pub const COUNTRY_CAR: Option<&str> = Some("ER");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("ኤርትራ")`
        pub const COUNTRY_NAME: Option<&str> = Some("ኤርትራ");
        /// `Some(232)`
        pub const COUNTRY_NUM: Option<i64> = Some(232);
        /// `Some("ERI")`
        pub const COUNTRY_POST: Option<&str> = Some("ERI");
        /// `Some("")`
        pub const LANG_AB: Option<&str> = Some("");
        /// `Some("byn")`
        pub const LANG_LIB: Option<&str> = Some("byn");
        /// `Some("ብሊን")`
        pub const LANG_NAME: Option<&str> = Some("ብሊን");
        /// `Some("byn")`
        pub const LANG_TERM: Option<&str> = Some("byn");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("7802 Solomon Seal Dr., Springfield, VA 22152, USA")`
        pub const ADDRESS: Option<&str> = Some("7802 Solomon Seal Dr., Springfield, VA 22152, USA");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_TIME"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_TIME"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2003-11-01")`
        pub const DATE: Option<&str> = Some("2003-11-01");
        /// `Some("locales@geez.org")`
        pub const EMAIL: Option<&str> = Some("locales@geez.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Blin")`
        pub const LANGUAGE: Option<&str> = Some("Blin");
        /// `Some("0.21")`
        pub const REVISION: Option<&str> = Some("0.21");
        /// `Some("Ge'ez Frontier Foundation")`
        pub const SOURCE: Option<&str> = Some("Ge'ez Frontier Foundation");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Eritrea")`
        pub const TERRITORY: Option<&str> = Some("Eritrea");
        /// `Some("Blin language locale for Eritrea")`
        pub const TITLE: Option<&str> = Some("Blin language locale for Eritrea");
    }
    pub use super::ti_ER::LC_MESSAGES;
    pub use super::ti_ER::LC_MONETARY;
    pub use super::ti_ER::LC_NUMERIC;
    pub use super::ti_ER::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["ሰ/ቅ", "ሰኑ", "ሰሊጝ", "ለጓ", "ኣምድ", "ኣርብ", "ሰ/ሽ"]`
        pub const ABDAY: &[&str] = &["ሰ/ቅ", "ሰኑ", "ሰሊጝ", "ለጓ", "ኣምድ", "ኣርብ", "ሰ/ሽ"];
        /// `&["ልደት", "ካብኽ", "ክብላ", "ፋጅኺ", "ክቢቅ", "ም/ት", "ኰር", "ማርያ", "ያኸኒ", "መተሉ", "ም/ም", "ተሕሳ"]`
        pub const ABMON: &[&str] = &["ልደት", "ካብኽ", "ክብላ", "ፋጅኺ", "ክቢቅ", "ም/ት", "ኰር", "ማርያ", "ያኸኒ", "መተሉ", "ም/ም", "ተሕሳ"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["ፋዱስ ጃብ", "ፋዱስ ደምቢ"]`
        pub const AM_PM: &[&str] = &["ፋዱስ ጃብ", "ፋዱስ ደምቢ"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%A፡ %B %e ግርጋ %r %Z %Y ኣድ")`
        pub const DATE_FMT: Option<&str> = Some("%A፡ %B %e ግርጋ %r %Z %Y ኣድ");
        /// `&["ሰንበር ቅዳዅ", "ሰኑ", "ሰሊጝ", "ለጓ ወሪ ለብዋ", "ኣምድ", "ኣርብ", "ሰንበር ሽጓዅ"]`
        pub const DAY: &[&str] = &["ሰንበር ቅዳዅ", "ሰኑ", "ሰሊጝ", "ለጓ ወሪ ለብዋ", "ኣምድ", "ኣርብ", "ሰንበር ሽጓዅ"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%A፡ %B %e ግርጋ %Y %l:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A፡ %B %e ግርጋ %Y %l:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["ልደትሪ", "ካብኽብቲ", "ክብላ", "ፋጅኺሪ", "ክቢቅሪ", "ምኪኤል ትጓ\u{305}ኒሪ", "ኰርኩ", "ማርያም ትሪ", "ያኸኒ መሳቅለሪ", "መተሉ", "ምኪኤል መሽወሪ", "ተሕሳስሪ"]`
        pub const MON: &[&str] = &["ልደትሪ", "ካብኽብቲ", "ክብላ", "ፋጅኺሪ", "ክቢቅሪ", "ምኪኤል ትጓ\u{305}ኒሪ", "ኰርኩ", "ማርያም ትሪ", "ያኸኒ መሳቅለሪ", "መተሉ", "ምኪኤል መሽወሪ", "ተሕሳስሪ"];
        /// `"%l:%M:%S %p"`
        pub const T_FMT: &str = "%l:%M:%S %p";
        /// `"%l:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%l:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ca_AD {
    pub mod LC_ADDRESS {
        /// `Some("AND")`
        pub const COUNTRY_AB3: Option<&str> = Some("AND");
        /// `Some("AD")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("AD");
        /// `Some("AND")`
        pub const COUNTRY_CAR: Option<&str> = Some("AND");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Andorra")`
        pub const COUNTRY_NAME: Option<&str> = Some("Andorra");
        /// `Some(20)`
        pub const COUNTRY_NUM: Option<i64> = Some(20);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ca")`
        pub const LANG_AB: Option<&str> = Some("ca");
        /// `Some("cat")`
        pub const LANG_LIB: Option<&str> = Some("cat");
        /// `Some("català")`
        pub const LANG_NAME: Option<&str> = Some("català");
        /// `Some("cat")`
        pub const LANG_TERM: Option<&str> = Some("cat");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2006-01-16")`
        pub const DATE: Option<&str> = Some("2006-01-16");
        /// `Some("")`
        pub const EMAIL: Option<&str> = Some("");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Catalan")`
        pub const LANGUAGE: Option<&str> = Some("Catalan");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Robert Millan")`
        pub const SOURCE: Option<&str> = Some("Robert Millan");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Andorra")`
        pub const TERRITORY: Option<&str> = Some("Andorra");
        /// `Some("Catalan locale for Andorra ")`
        pub const TITLE: Option<&str> = Some("Catalan locale for Andorra ");
    }
    pub use super::ca_ES::LC_MESSAGES;
    pub use super::ca_ES::LC_MONETARY;
    pub use super::ca_ES::LC_NUMERIC;
    pub mod LC_TELEPHONE {
        /// `Some("376")`
        pub const INT_PREFIX: Option<&str> = Some("376");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub use super::ca_ES::LC_TIME;
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ca_ES {
    pub mod LC_ADDRESS {
        /// `Some("ESP")`
        pub const COUNTRY_AB3: Option<&str> = Some("ESP");
        /// `Some("ES")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("ES");
        /// `Some("E")`
        pub const COUNTRY_CAR: Option<&str> = Some("E");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Espanya")`
        pub const COUNTRY_NAME: Option<&str> = Some("Espanya");
        /// `Some(724)`
        pub const COUNTRY_NUM: Option<i64> = Some(724);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ca")`
        pub const LANG_AB: Option<&str> = Some("ca");
        /// `Some("cat")`
        pub const LANG_LIB: Option<&str> = Some("cat");
        /// `Some("català")`
        pub const LANG_NAME: Option<&str> = Some("català");
        /// `Some("cat")`
        pub const LANG_TERM: Option<&str> = Some("cat");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Catalan")`
        pub const LANGUAGE: Option<&str> = Some("Catalan");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Spain")`
        pub const TERRITORY: Option<&str> = Some("Spain");
        /// `Some("Catalan locale for Spain")`
        pub const TITLE: Option<&str> = Some("Catalan locale for Spain");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("no")`
        pub const NOSTR: Option<&str> = Some("no");
        /// `"^[+1sSyY]"`
        pub const YESEXPR: &str = "^[+1sSyY]";
        /// `Some("sí")`
        pub const YESSTR: Option<&str> = Some("sí");
    }
    pub mod LC_MONETARY {
        /// `"€"`
        pub const CURRENCY_SYMBOL: &str = "€";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"EUR "`
        pub const INT_CURR_SYMBOL: &str = "EUR ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const MON_THOUSANDS_SEP: &str = ".";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const THOUSANDS_SEP: &str = ".";
    }
    pub mod LC_TELEPHONE {
        /// `Some("34")`
        pub const INT_PREFIX: Option<&str> = Some("34");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["dg.", "dl.", "dt.", "dc.", "dj.", "dv.", "ds."]`
        pub const ABDAY: &[&str] = &["dg.", "dl.", "dt.", "dc.", "dj.", "dv.", "ds."];
        /// `&["de gen.", "de febr.", "de març", "d’abr.", "de maig", "de juny", "de jul.", "d’ag.", "de set.", "d’oct.", "de nov.", "de des."]`
        pub const ABMON: &[&str] = &["de gen.", "de febr.", "de març", "d’abr.", "de maig", "de juny", "de jul.", "d’ag.", "de set.", "d’oct.", "de nov.", "de des."];
        /// `Some(&["gen.", "febr.", "març", "abr.", "maig", "juny", "jul.", "ag.", "set.", "oct.", "nov.", "des."])`
        pub const AB_ALT_MON: Option<&[&str]> = Some(&["gen.", "febr.", "març", "abr.", "maig", "juny", "jul.", "ag.", "set.", "oct.", "nov.", "des."]);
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `Some(&["gener", "febrer", "març", "abril", "maig", "juny", "juliol", "agost", "setembre", "octubre", "novembre", "desembre"])`
        pub const ALT_MON: Option<&[&str]> = Some(&["gener", "febrer", "març", "abril", "maig", "juny", "juliol", "agost", "setembre", "octubre", "novembre", "desembre"]);
        /// `&["a. m.", "p. m."]`
        pub const AM_PM: &[&str] = &["a. m.", "p. m."];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["diumenge", "dilluns", "dimarts", "dimecres", "dijous", "divendres", "dissabte"]`
        pub const DAY: &[&str] = &["diumenge", "dilluns", "dimarts", "dimecres", "dijous", "divendres", "dissabte"];
        /// `"%-d/%-m/%y"`
        pub const D_FMT: &str = "%-d/%-m/%y";
        /// `"%A, %-d %B de %Y, %T %Z"`
        pub const D_T_FMT: &str = "%A, %-d %B de %Y, %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["de gener", "de febrer", "de març", "d’abril", "de maig", "de juny", "de juliol", "d’agost", "de setembre", "d’octubre", "de novembre", "de desembre"]`
        pub const MON: &[&str] = &["de gener", "de febrer", "de març", "d’abril", "de maig", "de juny", "de juliol", "d’agost", "de setembre", "d’octubre", "de novembre", "de desembre"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `"%l:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%l:%M:%S %p";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ca_ES_euro {
    pub use super::ca_ES::LC_ADDRESS;
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("https://www.gnu.org/software/libc/")`
        pub const ADDRESS: Option<&str> = Some("https://www.gnu.org/software/libc/");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-08-20")`
        pub const DATE: Option<&str> = Some("2000-08-20");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Catalan")`
        pub const LANGUAGE: Option<&str> = Some("Catalan");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Free Software Foundation, Inc.")`
        pub const SOURCE: Option<&str> = Some("Free Software Foundation, Inc.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Spain")`
        pub const TERRITORY: Option<&str> = Some("Spain");
        /// `Some("Catalan locale for Catalonia with Euro")`
        pub const TITLE: Option<&str> = Some("Catalan locale for Catalonia with Euro");
    }
    pub use super::ca_ES::LC_MESSAGES;
    pub use super::ca_ES::LC_MONETARY;
    pub use super::ca_ES::LC_NUMERIC;
    pub use super::ca_ES::LC_TELEPHONE;
    pub use super::ca_ES::LC_TIME;
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ca_ES_valencia {
    pub use super::ca_ES::LC_ADDRESS;
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Jordi Mallach")`
        pub const CONTACT: Option<&str> = Some("Jordi Mallach");
        /// `Some("2006-04-06")`
        pub const DATE: Option<&str> = Some("2006-04-06");
        /// `Some("jordi@gnu.org")`
        pub const EMAIL: Option<&str> = Some("jordi@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Catalan")`
        pub const LANGUAGE: Option<&str> = Some("Catalan");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Spain")`
        pub const TERRITORY: Option<&str> = Some("Spain");
        /// `Some("Valencian (southern Catalan) locale for Spain with Euro")`
        pub const TITLE: Option<&str> = Some("Valencian (southern Catalan) locale for Spain with Euro");
    }
    pub use super::ca_ES::LC_MESSAGES;
    pub use super::ca_ES::LC_MONETARY;
    pub use super::ca_ES::LC_NUMERIC;
    pub use super::ca_ES::LC_TELEPHONE;
    pub use super::ca_ES::LC_TIME;
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ca_FR {
    pub mod LC_ADDRESS {
        /// `Some("FRA")`
        pub const COUNTRY_AB3: Option<&str> = Some("FRA");
        /// `Some("FR")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("FR");
        /// `Some("F")`
        pub const COUNTRY_CAR: Option<&str> = Some("F");
        /// `Some("979-10")`
        pub const COUNTRY_ISBN: Option<&str> = Some("979-10");
        /// `Some("França")`
        pub const COUNTRY_NAME: Option<&str> = Some("França");
        /// `Some(250)`
        pub const COUNTRY_NUM: Option<i64> = Some(250);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ca")`
        pub const LANG_AB: Option<&str> = Some("ca");
        /// `Some("cat")`
        pub const LANG_LIB: Option<&str> = Some("cat");
        /// `Some("català")`
        pub const LANG_NAME: Option<&str> = Some("català");
        /// `Some("cat")`
        pub const LANG_TERM: Option<&str> = Some("cat");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2006-01-16")`
        pub const DATE: Option<&str> = Some("2006-01-16");
        /// `Some("")`
        pub const EMAIL: Option<&str> = Some("");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Catalan")`
        pub const LANGUAGE: Option<&str> = Some("Catalan");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Robert Millan")`
        pub const SOURCE: Option<&str> = Some("Robert Millan");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("France")`
        pub const TERRITORY: Option<&str> = Some("France");
        /// `Some("Catalan locale for France ")`
        pub const TITLE: Option<&str> = Some("Catalan locale for France ");
    }
    pub use super::ca_ES::LC_MESSAGES;
    pub use super::ca_ES::LC_MONETARY;
    pub use super::ca_ES::LC_NUMERIC;
    pub use super::fr_FR::LC_TELEPHONE;
    pub use super::ca_ES::LC_TIME;
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ca_IT {
    pub mod LC_ADDRESS {
        /// `Some("ITA")`
        pub const COUNTRY_AB3: Option<&str> = Some("ITA");
        /// `Some("IT")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IT");
        /// `Some("I")`
        pub const COUNTRY_CAR: Option<&str> = Some("I");
        /// `Some("978-88,979-12")`
        pub const COUNTRY_ISBN: Option<&str> = Some("978-88,979-12");
        /// `Some("Itàlia")`
        pub const COUNTRY_NAME: Option<&str> = Some("Itàlia");
        /// `Some(380)`
        pub const COUNTRY_NUM: Option<i64> = Some(380);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ca")`
        pub const LANG_AB: Option<&str> = Some("ca");
        /// `Some("cat")`
        pub const LANG_LIB: Option<&str> = Some("cat");
        /// `Some("català")`
        pub const LANG_NAME: Option<&str> = Some("català");
        /// `Some("cat")`
        pub const LANG_TERM: Option<&str> = Some("cat");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2006-01-16")`
        pub const DATE: Option<&str> = Some("2006-01-16");
        /// `Some("")`
        pub const EMAIL: Option<&str> = Some("");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Catalan")`
        pub const LANGUAGE: Option<&str> = Some("Catalan");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Robert Millan")`
        pub const SOURCE: Option<&str> = Some("Robert Millan");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Italy")`
        pub const TERRITORY: Option<&str> = Some("Italy");
        /// `Some("Catalan locale for Italy (L'Alguer) ")`
        pub const TITLE: Option<&str> = Some("Catalan locale for Italy (L'Alguer) ");
    }
    pub use super::ca_ES::LC_MESSAGES;
    pub use super::ca_ES::LC_MONETARY;
    pub use super::ca_ES::LC_NUMERIC;
    pub use super::it_IT::LC_TELEPHONE;
    pub use super::ca_ES::LC_TIME;
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ce_RU {
    pub mod LC_ADDRESS {
        /// `Some("RUS")`
        pub const COUNTRY_AB3: Option<&str> = Some("RUS");
        /// `Some("RU")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("RU");
        /// `Some("RUS")`
        pub const COUNTRY_CAR: Option<&str> = Some("RUS");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Росси")`
        pub const COUNTRY_NAME: Option<&str> = Some("Росси");
        /// `Some(643)`
        pub const COUNTRY_NUM: Option<i64> = Some(643);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ce")`
        pub const LANG_AB: Option<&str> = Some("ce");
        /// `None`
        pub const LANG_LIB: Option<&str> = None;
        /// `Some("нохчийн")`
        pub const LANG_NAME: Option<&str> = Some("нохчийн");
        /// `Some("che")`
        pub const LANG_TERM: Option<&str> = Some("che");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("364024, ChR, Grozny, pr-kt of M. Esambayev, 13")`
        pub const ADDRESS: Option<&str> = Some("364024, ChR, Grozny, pr-kt of M. Esambayev, 13");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
        ]);
        /// `None`
        pub const CONTACT: Option<&str> = None;
        /// `Some("2014-08-25")`
        pub const DATE: Option<&str> = Some("2014-08-25");
        /// `None`
        pub const EMAIL: Option<&str> = None;
        /// `None`
        pub const FAX: Option<&str> = None;
        /// `Some("Chechen")`
        pub const LANGUAGE: Option<&str> = Some("Chechen");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("ANCHR")`
        pub const SOURCE: Option<&str> = Some("ANCHR");
        /// `None`
        pub const TEL: Option<&str> = None;
        /// `Some("Russia")`
        pub const TERRITORY: Option<&str> = Some("Russia");
        /// `Some("Chechen locale for RUSSIAN FEDERATION")`
        pub const TITLE: Option<&str> = Some("Chechen locale for RUSSIAN FEDERATION");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNМм]"`
        pub const NOEXPR: &str = "^[-0nNМм]";
        /// `None`
        pub const NOSTR: Option<&str> = None;
        /// `"^[+1yYДд]"`
        pub const YESEXPR: &str = "^[+1yYДд]";
        /// `None`
        pub const YESSTR: Option<&str> = None;
    }
    pub mod LC_MONETARY {
        /// `"₽"`
        pub const CURRENCY_SYMBOL: &str = "₽";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"RUB "`
        pub const INT_CURR_SYMBOL: &str = "RUB ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const MON_THOUSANDS_SEP: &str = "\u{202f}";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub use super::ru_RU::LC_NUMERIC;
    pub mod LC_TELEPHONE {
        /// `Some("7")`
        pub const INT_PREFIX: Option<&str> = Some("7");
        /// `Some("0~10")`
        pub const INT_SELECT: Option<&str> = Some("0~10");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["КӀ", "Ор", "Ши", "Кх", "Еа", "ПӀ", "Шо"]`
        pub const ABDAY: &[&str] = &["КӀ", "Ор", "Ши", "Кх", "Еа", "ПӀ", "Шо"];
        /// `&["янв", "фев", "мар", "апр", "май", "июн", "июл", "авг", "сен", "окт", "ноя", "дек"]`
        pub const ABMON: &[&str] = &["янв", "фев", "мар", "апр", "май", "июн", "июл", "авг", "сен", "окт", "ноя", "дек"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["КӀиранан де", "Оршотан де", "Шинарин де", "Кхаарин де", "Еарин де", "ПӀераскан де", "Шот де"]`
        pub const DAY: &[&str] = &["КӀиранан де", "Оршотан де", "Шинарин де", "Кхаарин де", "Еарин де", "ПӀераскан де", "Шот де"];
        /// `"%Y.%d.%m"`
        pub const D_FMT: &str = "%Y.%d.%m";
        /// `"%Y %d %b %a %T"`
        pub const D_T_FMT: &str = "%Y %d %b %a %T";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Январь", "Февраль", "Март", "Апрель", "Май", "Июнь", "Июль", "Август", "Сентябрь", "Октябрь", "Ноябрь", "Декабрь"]`
        pub const MON: &[&str] = &["Январь", "Февраль", "Март", "Апрель", "Май", "Июнь", "Июль", "Август", "Сентябрь", "Октябрь", "Ноябрь", "Декабрь"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod chr_US {
    pub mod LC_ADDRESS {
        /// `Some("USA")`
        pub const COUNTRY_AB3: Option<&str> = Some("USA");
        /// `Some("US")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("US");
        /// `Some("USA")`
        pub const COUNTRY_CAR: Option<&str> = Some("USA");
        /// `Some("0")`
        pub const COUNTRY_ISBN: Option<&str> = Some("0");
        /// `Some("ᏌᏊ ᎢᏳᎾᎵᏍᏔᏅ ᏍᎦᏚᎩ")`
        pub const COUNTRY_NAME: Option<&str> = Some("ᏌᏊ ᎢᏳᎾᎵᏍᏔᏅ ᏍᎦᏚᎩ");
        /// `Some(840)`
        pub const COUNTRY_NUM: Option<i64> = Some(840);
        /// `Some("USA")`
        pub const COUNTRY_POST: Option<&str> = Some("USA");
        /// `Some("")`
        pub const LANG_AB: Option<&str> = Some("");
        /// `Some("chr")`
        pub const LANG_LIB: Option<&str> = Some("chr");
        /// `Some("ᏣᎳᎩ")`
        pub const LANG_NAME: Option<&str> = Some("ᏣᎳᎩ");
        /// `Some("chr")`
        pub const LANG_TERM: Option<&str> = Some("chr");
        /// `Some("%a%N%f%N%d%N%b%N%h %s %e %r%N%T, %S %z%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%a%N%f%N%d%N%b%N%h %s %e %r%N%T, %S %z%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("http://cldr.unicode.org/index/process")`
        pub const ADDRESS: Option<&str> = Some("http://cldr.unicode.org/index/process");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Joseph Erb")`
        pub const CONTACT: Option<&str> = Some("Joseph Erb");
        /// `Some("2010-10-20")`
        pub const DATE: Option<&str> = Some("2010-10-20");
        /// `Some("josepherb7@gmail.com")`
        pub const EMAIL: Option<&str> = Some("josepherb7@gmail.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Cherokee")`
        pub const LANGUAGE: Option<&str> = Some("Cherokee");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Cherokee Nation")`
        pub const SOURCE: Option<&str> = Some("Cherokee Nation");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("United States")`
        pub const TERRITORY: Option<&str> = Some("United States");
        /// `Some("Cherokee language locale for United States")`
        pub const TITLE: Option<&str> = Some("Cherokee language locale for United States");
    }
    pub mod LC_MESSAGES {
        /// `"^([-0nN]|ᎥᏝ)"`
        pub const NOEXPR: &str = "^([-0nN]|ᎥᏝ)";
        /// `Some("ᎥᏝ")`
        pub const NOSTR: Option<&str> = Some("ᎥᏝ");
        /// `"^([+1yY]|ᎥᎥ)"`
        pub const YESEXPR: &str = "^([+1yY]|ᎥᎥ)";
        /// `Some("ᎥᎥ")`
        pub const YESSTR: Option<&str> = Some("ᎥᎥ");
    }
    pub use super::en_US::LC_MONETARY;
    pub use super::en_US::LC_NUMERIC;
    pub use super::en_US::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["ᏆᏍᎬ", "ᏉᏅᎯ", "ᏔᎵᏁ", "ᏦᎢᏁ", "ᏅᎩᏁ", "ᏧᎾᎩ", "ᏈᏕᎾ"]`
        pub const ABDAY: &[&str] = &["ᏆᏍᎬ", "ᏉᏅᎯ", "ᏔᎵᏁ", "ᏦᎢᏁ", "ᏅᎩᏁ", "ᏧᎾᎩ", "ᏈᏕᎾ"];
        /// `&["ᎤᏃ", "ᎧᎦ", "ᎠᏅ", "ᎧᏬ", "ᎠᏂ", "ᏕᎭ", "ᎫᏰ", "ᎦᎶ", "ᏚᎵ", "ᏚᏂ", "ᏅᏓ", "ᎥᏍ"]`
        pub const ABMON: &[&str] = &["ᎤᏃ", "ᎧᎦ", "ᎠᏅ", "ᎧᏬ", "ᎠᏂ", "ᏕᎭ", "ᎫᏰ", "ᎦᎶ", "ᏚᎵ", "ᏚᏂ", "ᏅᏓ", "ᎥᏍ"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["ᏌᎾᎴ", "ᏒᎯᏱᎢᏗᏢ"]`
        pub const AM_PM: &[&str] = &["ᏌᎾᎴ", "ᏒᎯᏱᎢᏗᏢ"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["ᎤᎾᏙᏓᏆᏍᎬ", "ᎤᎾᏙᏓᏉᏅᎯ", "ᏔᎵᏁᎢᎦ", "ᏦᎢᏁᎢᎦ", "ᏅᎩᏁᎢᎦ", "ᏧᎾᎩᎶᏍᏗ", "ᎤᎾᏙᏓᏈᏕᎾ"]`
        pub const DAY: &[&str] = &["ᎤᎾᏙᏓᏆᏍᎬ", "ᎤᎾᏙᏓᏉᏅᎯ", "ᏔᎵᏁᎢᎦ", "ᏦᎢᏁᎢᎦ", "ᏅᎩᏁᎢᎦ", "ᏧᎾᎩᎶᏍᏗ", "ᎤᎾᏙᏓᏈᏕᎾ"];
        /// `"%m/%d/%Y"`
        pub const D_FMT: &str = "%m/%d/%Y";
        /// `"%a %d %b %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["ᎤᏃᎸᏔᏅ", "ᎧᎦᎵ", "ᎠᏅᏱ", "ᎧᏬᏂ", "ᎠᏂᏍᎬᏘ", "ᏕᎭᎷᏱ", "ᎫᏰᏉᏂ", "ᎦᎶᏂ", "ᏚᎵᏍᏗ", "ᏚᏂᏅᏗ", "ᏅᏓᏕᏆ", "ᎥᏍᎩᏱ"]`
        pub const MON: &[&str] = &["ᎤᏃᎸᏔᏅ", "ᎧᎦᎵ", "ᎠᏅᏱ", "ᎧᏬᏂ", "ᎠᏂᏍᎬᏘ", "ᏕᎭᎷᏱ", "ᎫᏰᏉᏂ", "ᎦᎶᏂ", "ᏚᎵᏍᏗ", "ᏚᏂᏅᏗ", "ᏅᏓᏕᏆ", "ᎥᏍᎩᏱ"];
        /// `"%I:%M:%S %p"`
        pub const T_FMT: &str = "%I:%M:%S %p";
        /// `"%I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod cmn_TW {
    pub mod LC_ADDRESS {
        /// `Some("TWN")`
        pub const COUNTRY_AB3: Option<&str> = Some("TWN");
        /// `Some("TW")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("TW");
        /// `Some("RC")`
        pub const COUNTRY_CAR: Option<&str> = Some("RC");
        /// `Some("957")`
        pub const COUNTRY_ISBN: Option<&str> = Some("957");
        /// `Some("中華民國")`
        pub const COUNTRY_NAME: Option<&str> = Some("中華民國");
        /// `Some(158)`
        pub const COUNTRY_NUM: Option<i64> = Some(158);
        /// `Some("TW")`
        pub const COUNTRY_POST: Option<&str> = Some("TW");
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("cmn")`
        pub const LANG_LIB: Option<&str> = Some("cmn");
        /// `Some("漢語官話")`
        pub const LANG_NAME: Option<&str> = Some("漢語官話");
        /// `Some("cmn")`
        pub const LANG_TERM: Option<&str> = Some("cmn");
        /// `Some("%c%N%T%N%s %h %e %r%N%b%N%d%N%f%N%a%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%c%N%T%N%s %h %e %r%N%b%N%d%N%f%N%a%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2017-07-20")`
        pub const DATE: Option<&str> = Some("2017-07-20");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Mandarin Chinese")`
        pub const LANGUAGE: Option<&str> = Some("Mandarin Chinese");
        /// `Some("0.2")`
        pub const REVISION: Option<&str> = Some("0.2");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Taiwan")`
        pub const TERRITORY: Option<&str> = Some("Taiwan");
        /// `Some("Mandarin Chinese locale for the Republic of China")`
        pub const TITLE: Option<&str> = Some("Mandarin Chinese locale for the Republic of China");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNｎＮ不否]"`
        pub const NOEXPR: &str = "^[-0nNｎＮ不否]";
        /// `Some("不是")`
        pub const NOSTR: Option<&str> = Some("不是");
        /// `"^[+1yYｙＹ是]"`
        pub const YESEXPR: &str = "^[+1yYｙＹ是]";
        /// `Some("是")`
        pub const YESSTR: Option<&str> = Some("是");
    }
    pub mod LC_MONETARY {
        /// `"NT$"`
        pub const CURRENCY_SYMBOL: &str = "NT$";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"TWD "`
        pub const INT_CURR_SYMBOL: &str = "TWD ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `Some(1)`
        pub const INT_N_CS_PRECEDES: Option<i64> = Some(1);
        /// `Some(0)`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = Some(0);
        /// `Some(1)`
        pub const INT_N_SIGN_POSN: Option<i64> = Some(1);
        /// `Some(1)`
        pub const INT_P_CS_PRECEDES: Option<i64> = Some(1);
        /// `Some(0)`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = Some(0);
        /// `Some(1)`
        pub const INT_P_SIGN_POSN: Option<i64> = Some(1);
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[4]`
        pub const MON_GROUPING: &[i64] = &[4];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[4]`
        pub const GROUPING: &[i64] = &[4];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("886")`
        pub const INT_PREFIX: Option<&str> = Some("886");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `Some("%A-%l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("%A-%l");
        /// `Some("+%c-%a-%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c-%a-%l");
    }
    pub mod LC_TIME {
        /// `&["日", "一", "二", "三", "四", "五", "六"]`
        pub const ABDAY: &[&str] = &["日", "一", "二", "三", "四", "五", "六"];
        /// `&[" 1月", " 2月", " 3月", " 4月", " 5月", " 6月", " 7月", " 8月", " 9月", "10月", "11月", "12月"]`
        pub const ABMON: &[&str] = &[" 1月", " 2月", " 3月", " 4月", " 5月", " 6月", " 7月", " 8月", " 9月", "10月", "11月", "12月"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["上午", "下午"]`
        pub const AM_PM: &[&str] = &["上午", "下午"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%Y年 %b %-d號 %A %H:%M:%S %Z")`
        pub const DATE_FMT: Option<&str> = Some("%Y年 %b %-d號 %A %H:%M:%S %Z");
        /// `&["星期日", "星期一", "星期二", "星期三", "星期四", "星期五", "星期六"]`
        pub const DAY: &[&str] = &["星期日", "星期一", "星期二", "星期三", "星期四", "星期五", "星期六"];
        /// `"%Y年%m月%d號"`
        pub const D_FMT: &str = "%Y年%m月%d號";
        /// `"%Y年%m月%d號 (%A) %H點%M分%S秒"`
        pub const D_T_FMT: &str = "%Y年%m月%d號 (%A) %H點%M分%S秒";
        /// `Some(&["+:2:1913/01/01:+*:民國:%EC%Ey年", "+:1:1912/01/01:1912/12/31:民國:%EC元年", "+:1:1911/12/31:-*:民前:%EC%Ey年"])`
        pub const ERA: Option<&[&str]> = Some(&["+:2:1913/01/01:+*:民國:%EC%Ey年", "+:1:1912/01/01:1912/12/31:民國:%EC元年", "+:1:1911/12/31:-*:民前:%EC%Ey年"]);
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["一月", "二月", "三月", "四月", "五月", "六月", "七月", "八月", "九月", "十月", "十一月", "十二月"]`
        pub const MON: &[&str] = &["一月", "二月", "三月", "四月", "五月", "六月", "七月", "八月", "九月", "十月", "十一月", "十二月"];
        /// `"%H點%M分%S秒"`
        pub const T_FMT: &str = "%H點%M分%S秒";
        /// `"%p %I點%M分%S秒"`
        pub const T_FMT_AMPM: &str = "%p %I點%M分%S秒";
        /// `None`
        pub const WEEK: Option<&[i64]> = None;
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod crh_UA {
    pub mod LC_ADDRESS {
        /// `Some("UKR")`
        pub const COUNTRY_AB3: Option<&str> = Some("UKR");
        /// `Some("UA")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("UA");
        /// `Some("UA")`
        pub const COUNTRY_CAR: Option<&str> = Some("UA");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `None`
        pub const COUNTRY_NAME: Option<&str> = None;
        /// `Some(804)`
        pub const COUNTRY_NUM: Option<i64> = Some(804);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("crh")`
        pub const LANG_LIB: Option<&str> = Some("crh");
        /// `Some("qırımtatar tili")`
        pub const LANG_NAME: Option<&str> = Some("qırımtatar tili");
        /// `Some("crh")`
        pub const LANG_TERM: Option<&str> = Some("crh");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_NAME"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_NAME"],
        ]);
        /// `Some("Reşat SABIQ")`
        pub const CONTACT: Option<&str> = Some("Reşat SABIQ");
        /// `Some("2009-08-16")`
        pub const DATE: Option<&str> = Some("2009-08-16");
        /// `Some("tilde.birlik@gmail.com")`
        pub const EMAIL: Option<&str> = Some("tilde.birlik@gmail.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Crimean Tatar")`
        pub const LANGUAGE: Option<&str> = Some("Crimean Tatar");
        /// `Some("0.4")`
        pub const REVISION: Option<&str> = Some("0.4");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Ukraine")`
        pub const TERRITORY: Option<&str> = Some("Ukraine");
        /// `Some("Crimean Tatar (Crimean Turkish) language locale for Ukraine")`
        pub const TITLE: Option<&str> = Some("Crimean Tatar (Crimean Turkish) language locale for Ukraine");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNhH]"`
        pub const NOEXPR: &str = "^[-0nNhH]";
        /// `Some("hayır")`
        pub const NOSTR: Option<&str> = Some("hayır");
        /// `"^[+1yYeE]"`
        pub const YESEXPR: &str = "^[+1yYeE]";
        /// `Some("ebet")`
        pub const YESSTR: Option<&str> = Some("ebet");
    }
    pub mod LC_MONETARY {
        /// `"₴"`
        pub const CURRENCY_SYMBOL: &str = "₴";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"UAH "`
        pub const INT_CURR_SYMBOL: &str = "UAH ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const MON_THOUSANDS_SEP: &str = "\u{202f}";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const THOUSANDS_SEP: &str = ".";
    }
    pub mod LC_TELEPHONE {
        /// `Some("380")`
        pub const INT_PREFIX: Option<&str> = Some("380");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["Baz", "Ber", "Sal", "Çar", "Caq", "Cum", "Cer"]`
        pub const ABDAY: &[&str] = &["Baz", "Ber", "Sal", "Çar", "Caq", "Cum", "Cer"];
        /// `&["Yan", "Fev", "Mar", "Apr", "May", "İyn", "İyl", "Avg", "Sen", "Okt", "Noy", "Dek"]`
        pub const ABMON: &[&str] = &["Yan", "Fev", "Mar", "Apr", "May", "İyn", "İyl", "Avg", "Sen", "Okt", "Noy", "Dek"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["ÜE", "ÜS"]`
        pub const AM_PM: &[&str] = &["ÜE", "ÜS"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Bazar", "Bazarertesi", "Salı", "Çarşembe", "Cumaaqşamı", "Cuma", "Cumaertesi"]`
        pub const DAY: &[&str] = &["Bazar", "Bazarertesi", "Salı", "Çarşembe", "Cumaaqşamı", "Cuma", "Cumaertesi"];
        /// `"%d.%m.%Y"`
        pub const D_FMT: &str = "%d.%m.%Y";
        /// `"%a %d %b %Y %T"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Yanvar", "Fevral", "Mart", "Aprel", "Mayıs", "İyun", "İyul", "Avgust", "Sentâbr", "Oktâbr", "Noyabr", "Dekabr"]`
        pub const MON: &[&str] = &["Yanvar", "Fevral", "Mart", "Aprel", "Mayıs", "İyun", "İyul", "Avgust", "Sentâbr", "Oktâbr", "Noyabr", "Dekabr"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `"%I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod cs_CZ {
    pub mod LC_ADDRESS {
        /// `Some("CZE")`
        pub const COUNTRY_AB3: Option<&str> = Some("CZE");
        /// `Some("CZ")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("CZ");
        /// `Some("CZ")`
        pub const COUNTRY_CAR: Option<&str> = Some("CZ");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Česká republika")`
        pub const COUNTRY_NAME: Option<&str> = Some("Česká republika");
        /// `Some(203)`
        pub const COUNTRY_NUM: Option<i64> = Some(203);
        /// `Some("CZ")`
        pub const COUNTRY_POST: Option<&str> = Some("CZ");
        /// `Some("cs")`
        pub const LANG_AB: Option<&str> = Some("cs");
        /// `Some("cze")`
        pub const LANG_LIB: Option<&str> = Some("cze");
        /// `Some("čeština")`
        pub const LANG_NAME: Option<&str> = Some("čeština");
        /// `Some("ces")`
        pub const LANG_TERM: Option<&str> = Some("ces");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("https://www.gnu.org/software/libc/")`
        pub const ADDRESS: Option<&str> = Some("https://www.gnu.org/software/libc/");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-28")`
        pub const DATE: Option<&str> = Some("2000-06-28");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Czech")`
        pub const LANGUAGE: Option<&str> = Some("Czech");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Free Software Foundation, Inc.")`
        pub const SOURCE: Option<&str> = Some("Free Software Foundation, Inc.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Czech Republic")`
        pub const TERRITORY: Option<&str> = Some("Czech Republic");
        /// `Some("Czech locale for the Czech Republic")`
        pub const TITLE: Option<&str> = Some("Czech locale for the Czech Republic");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("ne")`
        pub const NOSTR: Option<&str> = Some("ne");
        /// `"^[+1aAyY]"`
        pub const YESEXPR: &str = "^[+1aAyY]";
        /// `Some("ano")`
        pub const YESSTR: Option<&str> = Some("ano");
    }
    pub mod LC_MONETARY {
        /// `"Kč"`
        pub const CURRENCY_SYMBOL: &str = "Kč";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"CZK "`
        pub const INT_CURR_SYMBOL: &str = "CZK ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const MON_THOUSANDS_SEP: &str = "\u{202f}";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const THOUSANDS_SEP: &str = "\u{202f}";
    }
    pub mod LC_TELEPHONE {
        /// `Some("420")`
        pub const INT_PREFIX: Option<&str> = Some("420");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `Some("(0%a) %l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("(0%a) %l");
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["Ne", "Po", "Út", "St", "Čt", "Pá", "So"]`
        pub const ABDAY: &[&str] = &["Ne", "Po", "Út", "St", "Čt", "Pá", "So"];
        /// `&["led", "úno", "bře", "dub", "kvě", "čen", "čec", "srp", "zář", "říj", "lis", "pro"]`
        pub const ABMON: &[&str] = &["led", "úno", "bře", "dub", "kvě", "čen", "čec", "srp", "zář", "říj", "lis", "pro"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `Some(&["leden", "únor", "březen", "duben", "květen", "červen", "červenec", "srpen", "září", "říjen", "listopad", "prosinec"])`
        pub const ALT_MON: Option<&[&str]> = Some(&["leden", "únor", "březen", "duben", "květen", "červen", "červenec", "srpen", "září", "říjen", "listopad", "prosinec"]);
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Neděle", "Pondělí", "Úterý", "Středa", "Čtvrtek", "Pátek", "Sobota"]`
        pub const DAY: &[&str] = &["Neděle", "Pondělí", "Úterý", "Středa", "Čtvrtek", "Pátek", "Sobota"];
        /// `"%-d.%-m.%Y"`
        pub const D_FMT: &str = "%-d.%-m.%Y";
        /// `"%a\u{a0}%-d.\u{a0}%B\u{a0}%Y,\u{a0}%H:%M:%S\u{a0}%Z"`
        pub const D_T_FMT: &str = "%a\u{a0}%-d.\u{a0}%B\u{a0}%Y,\u{a0}%H:%M:%S\u{a0}%Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["ledna", "února", "března", "dubna", "května", "června", "července", "srpna", "září", "října", "listopadu", "prosince"]`
        pub const MON: &[&str] = &["ledna", "února", "března", "dubna", "května", "června", "července", "srpna", "září", "října", "listopadu", "prosince"];
        /// `"%H:%M:%S"`
        pub const T_FMT: &str = "%H:%M:%S";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod csb_PL {
    pub mod LC_ADDRESS {
        /// `Some("POL")`
        pub const COUNTRY_AB3: Option<&str> = Some("POL");
        /// `Some("PL")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("PL");
        /// `Some("PL")`
        pub const COUNTRY_CAR: Option<&str> = Some("PL");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Pòlskô")`
        pub const COUNTRY_NAME: Option<&str> = Some("Pòlskô");
        /// `Some(616)`
        pub const COUNTRY_NUM: Option<i64> = Some(616);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("csb")`
        pub const LANG_LIB: Option<&str> = Some("csb");
        /// `Some("kaszëbsczi jãzëk")`
        pub const LANG_NAME: Option<&str> = Some("kaszëbsczi jãzëk");
        /// `Some("csb")`
        pub const LANG_TERM: Option<&str> = Some("csb");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Michal Ostrowski")`
        pub const CONTACT: Option<&str> = Some("Michal Ostrowski");
        /// `Some("2006-07-25")`
        pub const DATE: Option<&str> = Some("2006-07-25");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Kashubian")`
        pub const LANGUAGE: Option<&str> = Some("Kashubian");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("csb_PL locale")`
        pub const SOURCE: Option<&str> = Some("csb_PL locale");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Poland")`
        pub const TERRITORY: Option<&str> = Some("Poland");
        /// `Some("Kashubian locale for Poland")`
        pub const TITLE: Option<&str> = Some("Kashubian locale for Poland");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("nié")`
        pub const NOSTR: Option<&str> = Some("nié");
        /// `"^[+1JjTtYy]"`
        pub const YESEXPR: &str = "^[+1JjTtYy]";
        /// `Some("jo")`
        pub const YESSTR: Option<&str> = Some("jo");
    }
    pub use super::pl_PL::LC_MONETARY;
    pub use super::pl_PL::LC_NUMERIC;
    pub use super::pl_PL::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["nie", "pòn", "wtó", "str", "czw", "pią", "sob"]`
        pub const ABDAY: &[&str] = &["nie", "pòn", "wtó", "str", "czw", "pią", "sob"];
        /// `&["stë", "gro", "str", "łżë", "maj", "cze", "lëp", "zél", "séw", "ruj", "lës", "gòd"]`
        pub const ABMON: &[&str] = &["stë", "gro", "str", "łżë", "maj", "cze", "lëp", "zél", "séw", "ruj", "lës", "gòd"];
        /// `Some(&["stë", "gro", "str", "łżë", "môj", "cze", "lëp", "zél", "séw", "ruj", "lës", "gòd"])`
        pub const AB_ALT_MON: Option<&[&str]> = Some(&["stë", "gro", "str", "łżë", "môj", "cze", "lëp", "zél", "séw", "ruj", "lës", "gòd"]);
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `Some(&["stëcznik", "gromicznik", "strëmiannik", "łżëkwiat", "môj", "czerwińc", "lëpińc", "zélnik", "séwnik", "rujan", "lëstopadnik", "gòdnik"])`
        pub const ALT_MON: Option<&[&str]> = Some(&["stëcznik", "gromicznik", "strëmiannik", "łżëkwiat", "môj", "czerwińc", "lëpińc", "zélnik", "séwnik", "rujan", "lëstopadnik", "gòdnik"]);
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["niedzela", "pòniedzôłk", "wtórk", "strzoda", "czwiôrtk", "piątk", "sobòta"]`
        pub const DAY: &[&str] = &["niedzela", "pòniedzôłk", "wtórk", "strzoda", "czwiôrtk", "piątk", "sobòta"];
        /// `"%Y-%m-%d"`
        pub const D_FMT: &str = "%Y-%m-%d";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["stëcznika", "gromicznika", "strëmiannika", "łżëkwiata", "maja", "czerwińca", "lëpińca", "zélnika", "séwnika", "rujana", "lëstopadnika", "gòdnika"]`
        pub const MON: &[&str] = &["stëcznika", "gromicznika", "strëmiannika", "łżëkwiata", "maja", "czerwińca", "lëpińca", "zélnika", "séwnika", "rujana", "lëstopadnika", "gòdnika"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod cv_RU {
    pub mod LC_ADDRESS {
        /// `Some("RUS")`
        pub const COUNTRY_AB3: Option<&str> = Some("RUS");
        /// `Some("RU")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("RU");
        /// `Some("RUS")`
        pub const COUNTRY_CAR: Option<&str> = Some("RUS");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `None`
        pub const COUNTRY_NAME: Option<&str> = None;
        /// `Some(643)`
        pub const COUNTRY_NUM: Option<i64> = Some(643);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("cv")`
        pub const LANG_AB: Option<&str> = Some("cv");
        /// `Some("chv")`
        pub const LANG_LIB: Option<&str> = Some("chv");
        /// `Some("Чӑвашла")`
        pub const LANG_NAME: Option<&str> = Some("Чӑвашла");
        /// `Some("chv")`
        pub const LANG_TERM: Option<&str> = Some("chv");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jorgens Alle 8, DK-1615 Kobenhavn V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jorgens Alle 8, DK-1615 Kobenhavn V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Chuvash")`
        pub const LANGUAGE: Option<&str> = Some("Chuvash");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Russia")`
        pub const TERRITORY: Option<&str> = Some("Russia");
        /// `Some("Chuvash locale for Russia")`
        pub const TITLE: Option<&str> = Some("Chuvash locale for Russia");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `None`
        pub const NOSTR: Option<&str> = None;
        /// `"^[+1yY]"`
        pub const YESEXPR: &str = "^[+1yY]";
        /// `None`
        pub const YESSTR: Option<&str> = None;
    }
    pub mod LC_MONETARY {
        /// `"₽"`
        pub const CURRENCY_SYMBOL: &str = "₽";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"RUB "`
        pub const INT_CURR_SYMBOL: &str = "RUB ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const MON_THOUSANDS_SEP: &str = "\u{202f}";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub use super::ru_RU::LC_NUMERIC;
    pub use super::ru_RU::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["vr", "tn", "yt", "jn", "kş", "er", "šm"]`
        pub const ABDAY: &[&str] = &["vr", "tn", "yt", "jn", "kş", "er", "šm"];
        /// `&["KĂR", "NAR", "PUŠ", "AKA", "ŞU", "ŞĔR", "UTĂ", "ŞUR", "AVĂ", "JUP", "CÜK", "RAŠ"]`
        pub const ABMON: &[&str] = &["KĂR", "NAR", "PUŠ", "AKA", "ŞU", "ŞĔR", "UTĂ", "ŞUR", "AVĂ", "JUP", "CÜK", "RAŠ"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["vyrsarnikun", "tuntikun", "ytlarikun", "junkun", "kĕşnernikun", "ernekun", "šămatkun"]`
        pub const DAY: &[&str] = &["vyrsarnikun", "tuntikun", "ytlarikun", "junkun", "kĕşnernikun", "ernekun", "šămatkun"];
        /// `"%d.%m.%Y"`
        pub const D_FMT: &str = "%d.%m.%Y";
        /// `"%a %d %b %Y %T"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["kărlac", "narăs", "puš", "aka", "şu", "şĕrtme", "ută", "şurla", "avăn", "jupa", "cük", "raštav"]`
        pub const MON: &[&str] = &["kărlac", "narăs", "puš", "aka", "şu", "şĕrtme", "ută", "şurla", "avăn", "jupa", "cük", "raštav"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod cy_GB {
    pub mod LC_ADDRESS {
        /// `Some("GBR")`
        pub const COUNTRY_AB3: Option<&str> = Some("GBR");
        /// `Some("GB")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("GB");
        /// `Some("GB")`
        pub const COUNTRY_CAR: Option<&str> = Some("GB");
        /// `Some("0")`
        pub const COUNTRY_ISBN: Option<&str> = Some("0");
        /// `Some("Y Deyrnas Unedig")`
        pub const COUNTRY_NAME: Option<&str> = Some("Y Deyrnas Unedig");
        /// `Some(826)`
        pub const COUNTRY_NUM: Option<i64> = Some(826);
        /// `Some("GB")`
        pub const COUNTRY_POST: Option<&str> = Some("GB");
        /// `Some("cy")`
        pub const LANG_AB: Option<&str> = Some("cy");
        /// `Some("wel")`
        pub const LANG_LIB: Option<&str> = Some("wel");
        /// `Some("Cymraeg")`
        pub const LANG_NAME: Option<&str> = Some("Cymraeg");
        /// `Some("cym")`
        pub const LANG_TERM: Option<&str> = Some("cym");
        /// `Some("%d%N%f%N%d%N%b%N%s %h 5e %r%N%C%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%d%N%f%N%d%N%b%N%s %h 5e %r%N%C%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_MONETARY"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_MONETARY"],
        ]);
        /// `Some("Pablo Saratxaga")`
        pub const CONTACT: Option<&str> = Some("Pablo Saratxaga");
        /// `Some("2004-09-27")`
        pub const DATE: Option<&str> = Some("2004-09-27");
        /// `Some("pablo@mandrakesoft.com")`
        pub const EMAIL: Option<&str> = Some("pablo@mandrakesoft.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Welsh")`
        pub const LANGUAGE: Option<&str> = Some("Welsh");
        /// `Some("0.9")`
        pub const REVISION: Option<&str> = Some("0.9");
        /// `Some("thanks to Dafydd Tomos (dafydd@imaginet.co.uk)")`
        pub const SOURCE: Option<&str> = Some("thanks to Dafydd Tomos (dafydd@imaginet.co.uk)");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("United Kingdom")`
        pub const TERRITORY: Option<&str> = Some("United Kingdom");
        /// `Some("Welsh language locale for Great Britain")`
        pub const TITLE: Option<&str> = Some("Welsh language locale for Great Britain");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("na")`
        pub const NOSTR: Option<&str> = Some("na");
        /// `"^[+1iItTyY]"`
        pub const YESEXPR: &str = "^[+1iItTyY]";
        /// `Some("ie")`
        pub const YESSTR: Option<&str> = Some("ie");
    }
    pub use super::en_GB::LC_MONETARY;
    pub use super::en_GB::LC_NUMERIC;
    pub use super::en_GB::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Sul", "Llu", "Maw", "Mer", "Iau", "Gwe", "Sad"]`
        pub const ABDAY: &[&str] = &["Sul", "Llu", "Maw", "Mer", "Iau", "Gwe", "Sad"];
        /// `&["Ion", "Chw", "Maw", "Ebr", "Mai", "Meh", "Gor", "Aws", "Med", "Hyd", "Tach", "Rha"]`
        pub const ABMON: &[&str] = &["Ion", "Chw", "Maw", "Ebr", "Mai", "Meh", "Gor", "Aws", "Med", "Hyd", "Tach", "Rha"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["am", "pm"]`
        pub const AM_PM: &[&str] = &["am", "pm"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a %e %b %H:%M:%S %Z %Y")`
        pub const DATE_FMT: Option<&str> = Some("%a %e %b %H:%M:%S %Z %Y");
        /// `&["Sul", "Llun", "Mawrth", "Mercher", "Iau", "Gwener", "Sadwrn"]`
        pub const DAY: &[&str] = &["Sul", "Llun", "Mawrth", "Mercher", "Iau", "Gwener", "Sadwrn"];
        /// `"%d.%m.%y"`
        pub const D_FMT: &str = "%d.%m.%y";
        /// `"Dydd %A %d mis %B %Y %T %Z"`
        pub const D_T_FMT: &str = "Dydd %A %d mis %B %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Ionawr", "Chwefror", "Mawrth", "Ebrill", "Mai", "Mehefin", "Gorffennaf", "Awst", "Medi", "Hydref", "Tachwedd", "Rhagfyr"]`
        pub const MON: &[&str] = &["Ionawr", "Chwefror", "Mawrth", "Ebrill", "Mai", "Mehefin", "Gorffennaf", "Awst", "Medi", "Hydref", "Tachwedd", "Rhagfyr"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `"%l:%M:%S %P %Z"`
        pub const T_FMT_AMPM: &str = "%l:%M:%S %P %Z";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod da_DK {
    pub mod LC_ADDRESS {
        /// `Some("DNK")`
        pub const COUNTRY_AB3: Option<&str> = Some("DNK");
        /// `Some("DK")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("DK");
        /// `Some("DK")`
        pub const COUNTRY_CAR: Option<&str> = Some("DK");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Danmark")`
        pub const COUNTRY_NAME: Option<&str> = Some("Danmark");
        /// `Some(208)`
        pub const COUNTRY_NUM: Option<i64> = Some(208);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("da")`
        pub const LANG_AB: Option<&str> = Some("da");
        /// `Some("dan")`
        pub const LANG_LIB: Option<&str> = Some("dan");
        /// `Some("dansk")`
        pub const LANG_NAME: Option<&str> = Some("dansk");
        /// `Some("dan")`
        pub const LANG_TERM: Option<&str> = Some("dan");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Kollegievej 6, DK-2920 Charlottenlund, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Kollegievej 6, DK-2920 Charlottenlund, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Danish")`
        pub const LANGUAGE: Option<&str> = Some("Danish");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Danish Standards Association")`
        pub const SOURCE: Option<&str> = Some("Danish Standards Association");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Denmark")`
        pub const TERRITORY: Option<&str> = Some("Denmark");
        /// `Some("Danish locale for Denmark")`
        pub const TITLE: Option<&str> = Some("Danish locale for Denmark");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0Nn]"`
        pub const NOEXPR: &str = "^[-0Nn]";
        /// `Some("nej")`
        pub const NOSTR: Option<&str> = Some("nej");
        /// `"^[+1JjYy]"`
        pub const YESEXPR: &str = "^[+1JjYy]";
        /// `Some("ja")`
        pub const YESSTR: Option<&str> = Some("ja");
    }
    pub mod LC_MONETARY {
        /// `"kr."`
        pub const CURRENCY_SYMBOL: &str = "kr.";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"DKK "`
        pub const INT_CURR_SYMBOL: &str = "DKK ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const MON_THOUSANDS_SEP: &str = ".";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `2`
        pub const N_SEP_BY_SPACE: i64 = 2;
        /// `4`
        pub const N_SIGN_POSN: i64 = 4;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `2`
        pub const P_SEP_BY_SPACE: i64 = 2;
        /// `4`
        pub const P_SIGN_POSN: i64 = 4;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const THOUSANDS_SEP: &str = ".";
    }
    pub mod LC_TELEPHONE {
        /// `Some("45")`
        pub const INT_PREFIX: Option<&str> = Some("45");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["søn", "man", "tir", "ons", "tor", "fre", "lør"]`
        pub const ABDAY: &[&str] = &["søn", "man", "tir", "ons", "tor", "fre", "lør"];
        /// `&["jan", "feb", "mar", "apr", "maj", "jun", "jul", "aug", "sep", "okt", "nov", "dec"]`
        pub const ABMON: &[&str] = &["jan", "feb", "mar", "apr", "maj", "jun", "jul", "aug", "sep", "okt", "nov", "dec"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a %e %b %H:%M:%S %Z %Y")`
        pub const DATE_FMT: Option<&str> = Some("%a %e %b %H:%M:%S %Z %Y");
        /// `&["søndag", "mandag", "tirsdag", "onsdag", "torsdag", "fredag", "lørdag"]`
        pub const DAY: &[&str] = &["søndag", "mandag", "tirsdag", "onsdag", "torsdag", "fredag", "lørdag"];
        /// `"%d-%m-%Y"`
        pub const D_FMT: &str = "%d-%m-%Y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["januar", "februar", "marts", "april", "maj", "juni", "juli", "august", "september", "oktober", "november", "december"]`
        pub const MON: &[&str] = &["januar", "februar", "marts", "april", "maj", "juni", "juli", "august", "september", "oktober", "november", "december"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod de_AT {
    pub mod LC_ADDRESS {
        /// `Some("AUT")`
        pub const COUNTRY_AB3: Option<&str> = Some("AUT");
        /// `Some("AT")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("AT");
        /// `Some("A")`
        pub const COUNTRY_CAR: Option<&str> = Some("A");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Österreich")`
        pub const COUNTRY_NAME: Option<&str> = Some("Österreich");
        /// `Some(40)`
        pub const COUNTRY_NUM: Option<i64> = Some(40);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("de")`
        pub const LANG_AB: Option<&str> = Some("de");
        /// `Some("ger")`
        pub const LANG_LIB: Option<&str> = Some("ger");
        /// `Some("Deutsch")`
        pub const LANG_NAME: Option<&str> = Some("Deutsch");
        /// `Some("deu")`
        pub const LANG_TERM: Option<&str> = Some("deu");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Postfach 130, A-1021 Wien")`
        pub const ADDRESS: Option<&str> = Some("Postfach 130, A-1021 Wien");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Gerhard Budin")`
        pub const CONTACT: Option<&str> = Some("Gerhard Budin");
        /// `Some("2000-06-28")`
        pub const DATE: Option<&str> = Some("2000-06-28");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Austrian German")`
        pub const LANGUAGE: Option<&str> = Some("Austrian German");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("O:sterreiches Normungsinstitut")`
        pub const SOURCE: Option<&str> = Some("O:sterreiches Normungsinstitut");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Austria")`
        pub const TERRITORY: Option<&str> = Some("Austria");
        /// `Some("German locale for Austria")`
        pub const TITLE: Option<&str> = Some("German locale for Austria");
    }
    pub use super::de_DE::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"€"`
        pub const CURRENCY_SYMBOL: &str = "€";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"EUR "`
        pub const INT_CURR_SYMBOL: &str = "EUR ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const MON_THOUSANDS_SEP: &str = "\u{202f}";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub use super::de_DE::LC_NUMERIC;
    pub mod LC_TELEPHONE {
        /// `Some("43")`
        pub const INT_PREFIX: Option<&str> = Some("43");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `Some("%a %l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("%a %l");
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["So", "Mo", "Di", "Mi", "Do", "Fr", "Sa"]`
        pub const ABDAY: &[&str] = &["So", "Mo", "Di", "Mi", "Do", "Fr", "Sa"];
        /// `&["Jän", "Feb", "Mär", "Apr", "Mai", "Jun", "Jul", "Aug", "Sep", "Okt", "Nov", "Dez"]`
        pub const ABMON: &[&str] = &["Jän", "Feb", "Mär", "Apr", "Mai", "Jun", "Jul", "Aug", "Sep", "Okt", "Nov", "Dez"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Sonntag", "Montag", "Dienstag", "Mittwoch", "Donnerstag", "Freitag", "Samstag"]`
        pub const DAY: &[&str] = &["Sonntag", "Montag", "Dienstag", "Mittwoch", "Donnerstag", "Freitag", "Samstag"];
        /// `"%Y-%m-%d"`
        pub const D_FMT: &str = "%Y-%m-%d";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Jänner", "Februar", "März", "April", "Mai", "Juni", "Juli", "August", "September", "Oktober", "November", "Dezember"]`
        pub const MON: &[&str] = &["Jänner", "Februar", "März", "April", "Mai", "Juni", "Juli", "August", "September", "Oktober", "November", "Dezember"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod de_AT_euro {
    pub use super::de_AT::LC_ADDRESS;
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Postfach 130, A-1021 Wien")`
        pub const ADDRESS: Option<&str> = Some("Postfach 130, A-1021 Wien");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Gerhard Budin")`
        pub const CONTACT: Option<&str> = Some("Gerhard Budin");
        /// `Some("2000-08-20")`
        pub const DATE: Option<&str> = Some("2000-08-20");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("German")`
        pub const LANGUAGE: Option<&str> = Some("German");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("O:sterreiches Normungsinstitut")`
        pub const SOURCE: Option<&str> = Some("O:sterreiches Normungsinstitut");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Austria")`
        pub const TERRITORY: Option<&str> = Some("Austria");
        /// `Some("German locale for Austria with Euro")`
        pub const TITLE: Option<&str> = Some("German locale for Austria with Euro");
    }
    pub use super::de_AT::LC_MESSAGES;
    pub use super::de_AT::LC_MONETARY;
    pub use super::de_AT::LC_NUMERIC;
    pub use super::de_AT::LC_TELEPHONE;
    pub use super::de_AT::LC_TIME;
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod de_BE {
    pub mod LC_ADDRESS {
        /// `Some("BEL")`
        pub const COUNTRY_AB3: Option<&str> = Some("BEL");
        /// `Some("BE")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("BE");
        /// `Some("B")`
        pub const COUNTRY_CAR: Option<&str> = Some("B");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Belgien")`
        pub const COUNTRY_NAME: Option<&str> = Some("Belgien");
        /// `Some(56)`
        pub const COUNTRY_NUM: Option<i64> = Some(56);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("de")`
        pub const LANG_AB: Option<&str> = Some("de");
        /// `Some("ger")`
        pub const LANG_LIB: Option<&str> = Some("ger");
        /// `Some("Deutsch")`
        pub const LANG_NAME: Option<&str> = Some("Deutsch");
        /// `Some("deu")`
        pub const LANG_TERM: Option<&str> = Some("deu");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("German")`
        pub const LANGUAGE: Option<&str> = Some("German");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Belgium")`
        pub const TERRITORY: Option<&str> = Some("Belgium");
        /// `Some("German locale for Belgium")`
        pub const TITLE: Option<&str> = Some("German locale for Belgium");
    }
    pub use super::de_DE::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"€"`
        pub const CURRENCY_SYMBOL: &str = "€";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"EUR "`
        pub const INT_CURR_SYMBOL: &str = "EUR ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const MON_THOUSANDS_SEP: &str = ".";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `4`
        pub const N_SIGN_POSN: i64 = 4;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `4`
        pub const P_SIGN_POSN: i64 = 4;
    }
    pub use super::de_DE::LC_NUMERIC;
    pub use super::fr_BE::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["So", "Mo", "Di", "Mi", "Do", "Fr", "Sa"]`
        pub const ABDAY: &[&str] = &["So", "Mo", "Di", "Mi", "Do", "Fr", "Sa"];
        /// `&["Jan", "Feb", "Mär", "Apr", "Mai", "Jun", "Jul", "Aug", "Sep", "Okt", "Nov", "Dez"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Mär", "Apr", "Mai", "Jun", "Jul", "Aug", "Sep", "Okt", "Nov", "Dez"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Sonntag", "Montag", "Dienstag", "Mittwoch", "Donnerstag", "Freitag", "Samstag"]`
        pub const DAY: &[&str] = &["Sonntag", "Montag", "Dienstag", "Mittwoch", "Donnerstag", "Freitag", "Samstag"];
        /// `"%Y-%m-%d"`
        pub const D_FMT: &str = "%Y-%m-%d";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Januar", "Februar", "März", "April", "Mai", "Juni", "Juli", "August", "September", "Oktober", "November", "Dezember"]`
        pub const MON: &[&str] = &["Januar", "Februar", "März", "April", "Mai", "Juni", "Juli", "August", "September", "Oktober", "November", "Dezember"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod de_BE_euro {
    pub use super::de_BE::LC_ADDRESS;
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("https://www.gnu.org/software/libc/")`
        pub const ADDRESS: Option<&str> = Some("https://www.gnu.org/software/libc/");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-08-21")`
        pub const DATE: Option<&str> = Some("2000-08-21");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("German")`
        pub const LANGUAGE: Option<&str> = Some("German");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Free Software Foundation, Inc.")`
        pub const SOURCE: Option<&str> = Some("Free Software Foundation, Inc.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Belgium")`
        pub const TERRITORY: Option<&str> = Some("Belgium");
        /// `Some("German locale for Belgium with Euro")`
        pub const TITLE: Option<&str> = Some("German locale for Belgium with Euro");
    }
    pub use super::de_BE::LC_MESSAGES;
    pub use super::de_BE::LC_MONETARY;
    pub use super::de_BE::LC_NUMERIC;
    pub use super::de_BE::LC_TELEPHONE;
    pub use super::de_BE::LC_TIME;
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod de_CH {
    pub mod LC_ADDRESS {
        /// `Some("CHE")`
        pub const COUNTRY_AB3: Option<&str> = Some("CHE");
        /// `Some("CH")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("CH");
        /// `Some("CH")`
        pub const COUNTRY_CAR: Option<&str> = Some("CH");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Schweiz")`
        pub const COUNTRY_NAME: Option<&str> = Some("Schweiz");
        /// `Some(756)`
        pub const COUNTRY_NUM: Option<i64> = Some(756);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("de")`
        pub const LANG_AB: Option<&str> = Some("de");
        /// `Some("ger")`
        pub const LANG_LIB: Option<&str> = Some("ger");
        /// `Some("Deutsch")`
        pub const LANG_NAME: Option<&str> = Some("Deutsch");
        /// `Some("deu")`
        pub const LANG_TERM: Option<&str> = Some("deu");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2007-09-23")`
        pub const DATE: Option<&str> = Some("2007-09-23");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Swiss High German")`
        pub const LANGUAGE: Option<&str> = Some("Swiss High German");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Switzerland")`
        pub const TERRITORY: Option<&str> = Some("Switzerland");
        /// `Some("German locale for Switzerland")`
        pub const TITLE: Option<&str> = Some("German locale for Switzerland");
    }
    pub use super::de_DE::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"CHF"`
        pub const CURRENCY_SYMBOL: &str = "CHF";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"CHF "`
        pub const INT_CURR_SYMBOL: &str = "CHF ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"’"`
        pub const MON_THOUSANDS_SEP: &str = "’";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `4`
        pub const N_SIGN_POSN: i64 = 4;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `4`
        pub const P_SIGN_POSN: i64 = 4;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `"’"`
        pub const THOUSANDS_SEP: &str = "’";
    }
    pub mod LC_TELEPHONE {
        /// `Some("41")`
        pub const INT_PREFIX: Option<&str> = Some("41");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["So", "Mo", "Di", "Mi", "Do", "Fr", "Sa"]`
        pub const ABDAY: &[&str] = &["So", "Mo", "Di", "Mi", "Do", "Fr", "Sa"];
        /// `&["Jan", "Feb", "Mär", "Apr", "Mai", "Jun", "Jul", "Aug", "Sep", "Okt", "Nov", "Dez"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Mär", "Apr", "Mai", "Jun", "Jul", "Aug", "Sep", "Okt", "Nov", "Dez"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Sonntag", "Montag", "Dienstag", "Mittwoch", "Donnerstag", "Freitag", "Samstag"]`
        pub const DAY: &[&str] = &["Sonntag", "Montag", "Dienstag", "Mittwoch", "Donnerstag", "Freitag", "Samstag"];
        /// `"%d.%m.%Y"`
        pub const D_FMT: &str = "%d.%m.%Y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Januar", "Februar", "März", "April", "Mai", "Juni", "Juli", "August", "September", "Oktober", "November", "Dezember"]`
        pub const MON: &[&str] = &["Januar", "Februar", "März", "April", "Mai", "Juni", "Juli", "August", "September", "Oktober", "November", "Dezember"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod de_DE {
    pub mod LC_ADDRESS {
        /// `Some("DEU")`
        pub const COUNTRY_AB3: Option<&str> = Some("DEU");
        /// `Some("DE")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("DE");
        /// `Some("D")`
        pub const COUNTRY_CAR: Option<&str> = Some("D");
        /// `Some("3")`
        pub const COUNTRY_ISBN: Option<&str> = Some("3");
        /// `Some("Deutschland")`
        pub const COUNTRY_NAME: Option<&str> = Some("Deutschland");
        /// `Some(276)`
        pub const COUNTRY_NUM: Option<i64> = Some(276);
        /// `Some("D")`
        pub const COUNTRY_POST: Option<&str> = Some("D");
        /// `Some("de")`
        pub const LANG_AB: Option<&str> = Some("de");
        /// `Some("ger")`
        pub const LANG_LIB: Option<&str> = Some("ger");
        /// `Some("Deutsch")`
        pub const LANG_NAME: Option<&str> = Some("Deutsch");
        /// `Some("deu")`
        pub const LANG_TERM: Option<&str> = Some("deu");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("https://www.gnu.org/software/libc/")`
        pub const ADDRESS: Option<&str> = Some("https://www.gnu.org/software/libc/");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-24")`
        pub const DATE: Option<&str> = Some("2000-06-24");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("German")`
        pub const LANGUAGE: Option<&str> = Some("German");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Free Software Foundation, Inc.")`
        pub const SOURCE: Option<&str> = Some("Free Software Foundation, Inc.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Germany")`
        pub const TERRITORY: Option<&str> = Some("Germany");
        /// `Some("German locale for Germany")`
        pub const TITLE: Option<&str> = Some("German locale for Germany");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("nein")`
        pub const NOSTR: Option<&str> = Some("nein");
        /// `"^[+1jJyY]"`
        pub const YESEXPR: &str = "^[+1jJyY]";
        /// `Some("ja")`
        pub const YESSTR: Option<&str> = Some("ja");
    }
    pub mod LC_MONETARY {
        /// `"€"`
        pub const CURRENCY_SYMBOL: &str = "€";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"EUR "`
        pub const INT_CURR_SYMBOL: &str = "EUR ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const MON_THOUSANDS_SEP: &str = ".";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const THOUSANDS_SEP: &str = ".";
    }
    pub mod LC_TELEPHONE {
        /// `Some("49")`
        pub const INT_PREFIX: Option<&str> = Some("49");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `Some("%A %l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("%A %l");
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["So", "Mo", "Di", "Mi", "Do", "Fr", "Sa"]`
        pub const ABDAY: &[&str] = &["So", "Mo", "Di", "Mi", "Do", "Fr", "Sa"];
        /// `&["Jan", "Feb", "Mär", "Apr", "Mai", "Jun", "Jul", "Aug", "Sep", "Okt", "Nov", "Dez"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Mär", "Apr", "Mai", "Jun", "Jul", "Aug", "Sep", "Okt", "Nov", "Dez"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a %-d. %b %H:%M:%S %Z %Y")`
        pub const DATE_FMT: Option<&str> = Some("%a %-d. %b %H:%M:%S %Z %Y");
        /// `&["Sonntag", "Montag", "Dienstag", "Mittwoch", "Donnerstag", "Freitag", "Samstag"]`
        pub const DAY: &[&str] = &["Sonntag", "Montag", "Dienstag", "Mittwoch", "Donnerstag", "Freitag", "Samstag"];
        /// `"%d.%m.%Y"`
        pub const D_FMT: &str = "%d.%m.%Y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Januar", "Februar", "März", "April", "Mai", "Juni", "Juli", "August", "September", "Oktober", "November", "Dezember"]`
        pub const MON: &[&str] = &["Januar", "Februar", "März", "April", "Mai", "Juni", "Juli", "August", "September", "Oktober", "November", "Dezember"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod de_DE_euro {
    pub use super::de_DE::LC_ADDRESS;
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("https://www.gnu.org/software/libc/")`
        pub const ADDRESS: Option<&str> = Some("https://www.gnu.org/software/libc/");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-24")`
        pub const DATE: Option<&str> = Some("2000-06-24");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("German")`
        pub const LANGUAGE: Option<&str> = Some("German");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Free Software Foundation, Inc.")`
        pub const SOURCE: Option<&str> = Some("Free Software Foundation, Inc.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Germany")`
        pub const TERRITORY: Option<&str> = Some("Germany");
        /// `Some("German locale for Germany with Euro")`
        pub const TITLE: Option<&str> = Some("German locale for Germany with Euro");
    }
    pub use super::de_DE::LC_MESSAGES;
    pub use super::de_DE::LC_MONETARY;
    pub use super::de_DE::LC_NUMERIC;
    pub use super::de_DE::LC_TELEPHONE;
    pub use super::de_DE::LC_TIME;
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod de_IT {
    pub mod LC_ADDRESS {
        /// `Some("ITA")`
        pub const COUNTRY_AB3: Option<&str> = Some("ITA");
        /// `Some("IT")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IT");
        /// `Some("I")`
        pub const COUNTRY_CAR: Option<&str> = Some("I");
        /// `Some("978-88,979-12")`
        pub const COUNTRY_ISBN: Option<&str> = Some("978-88,979-12");
        /// `Some("Italien")`
        pub const COUNTRY_NAME: Option<&str> = Some("Italien");
        /// `Some(380)`
        pub const COUNTRY_NUM: Option<i64> = Some(380);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("de")`
        pub const LANG_AB: Option<&str> = Some("de");
        /// `Some("ger")`
        pub const LANG_LIB: Option<&str> = Some("ger");
        /// `Some("Deutsch")`
        pub const LANG_NAME: Option<&str> = Some("Deutsch");
        /// `Some("deu")`
        pub const LANG_TERM: Option<&str> = Some("deu");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("I-39100 Bolzano/Bozen")`
        pub const ADDRESS: Option<&str> = Some("I-39100 Bolzano/Bozen");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Matthias Dieter Wallno:fer")`
        pub const CONTACT: Option<&str> = Some("Matthias Dieter Wallno:fer");
        /// `Some("2000-06-28")`
        pub const DATE: Option<&str> = Some("2000-06-28");
        /// `Some("libc-locales@sourceware.org")`
        pub const EMAIL: Option<&str> = Some("libc-locales@sourceware.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("German")`
        pub const LANGUAGE: Option<&str> = Some("German");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Italy")`
        pub const TERRITORY: Option<&str> = Some("Italy");
        /// `Some("German language locale for Italy")`
        pub const TITLE: Option<&str> = Some("German language locale for Italy");
    }
    pub use super::de_DE::LC_MESSAGES;
    pub use super::it_IT::LC_MONETARY;
    pub use super::de_DE::LC_NUMERIC;
    pub use super::it_IT::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["So", "Mo", "Di", "Mi", "Do", "Fr", "Sa"]`
        pub const ABDAY: &[&str] = &["So", "Mo", "Di", "Mi", "Do", "Fr", "Sa"];
        /// `&["Jän", "Feb", "Mär", "Apr", "Mai", "Jun", "Jul", "Aug", "Sep", "Okt", "Nov", "Dez"]`
        pub const ABMON: &[&str] = &["Jän", "Feb", "Mär", "Apr", "Mai", "Jun", "Jul", "Aug", "Sep", "Okt", "Nov", "Dez"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a %-d. %b %H:%M:%S %Z %Y")`
        pub const DATE_FMT: Option<&str> = Some("%a %-d. %b %H:%M:%S %Z %Y");
        /// `&["Sonntag", "Montag", "Dienstag", "Mittwoch", "Donnerstag", "Freitag", "Samstag"]`
        pub const DAY: &[&str] = &["Sonntag", "Montag", "Dienstag", "Mittwoch", "Donnerstag", "Freitag", "Samstag"];
        /// `"%d.%m.%Y"`
        pub const D_FMT: &str = "%d.%m.%Y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Jänner", "Februar", "März", "April", "Mai", "Juni", "Juli", "August", "September", "Oktober", "November", "Dezember"]`
        pub const MON: &[&str] = &["Jänner", "Februar", "März", "April", "Mai", "Juni", "Juli", "August", "September", "Oktober", "November", "Dezember"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod de_LI {
    pub mod LC_ADDRESS {
        /// `Some("LIE")`
        pub const COUNTRY_AB3: Option<&str> = Some("LIE");
        /// `Some("LI")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("LI");
        /// `Some("FL")`
        pub const COUNTRY_CAR: Option<&str> = Some("FL");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Liechtenstein")`
        pub const COUNTRY_NAME: Option<&str> = Some("Liechtenstein");
        /// `Some(438)`
        pub const COUNTRY_NUM: Option<i64> = Some(438);
        /// `Some("LI")`
        pub const COUNTRY_POST: Option<&str> = Some("LI");
        /// `Some("de")`
        pub const LANG_AB: Option<&str> = Some("de");
        /// `Some("ger")`
        pub const LANG_LIB: Option<&str> = Some("ger");
        /// `Some("Deutsch")`
        pub const LANG_NAME: Option<&str> = Some("Deutsch");
        /// `Some("deu")`
        pub const LANG_TERM: Option<&str> = Some("deu");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2007-11-27")`
        pub const DATE: Option<&str> = Some("2007-11-27");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("German")`
        pub const LANGUAGE: Option<&str> = Some("German");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Liechtenstein")`
        pub const TERRITORY: Option<&str> = Some("Liechtenstein");
        /// `Some("German locale for Liechtenstein")`
        pub const TITLE: Option<&str> = Some("German locale for Liechtenstein");
    }
    pub use super::de_CH::LC_MESSAGES;
    pub use super::de_CH::LC_MONETARY;
    pub use super::de_CH::LC_NUMERIC;
    pub mod LC_TELEPHONE {
        /// `Some("423")`
        pub const INT_PREFIX: Option<&str> = Some("423");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub use super::de_CH::LC_TIME;
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod de_LU {
    pub mod LC_ADDRESS {
        /// `Some("LUX")`
        pub const COUNTRY_AB3: Option<&str> = Some("LUX");
        /// `Some("LU")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("LU");
        /// `Some("L")`
        pub const COUNTRY_CAR: Option<&str> = Some("L");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Luxemburg")`
        pub const COUNTRY_NAME: Option<&str> = Some("Luxemburg");
        /// `Some(442)`
        pub const COUNTRY_NUM: Option<i64> = Some(442);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("de")`
        pub const LANG_AB: Option<&str> = Some("de");
        /// `Some("ger")`
        pub const LANG_LIB: Option<&str> = Some("ger");
        /// `Some("Deutsch")`
        pub const LANG_NAME: Option<&str> = Some("Deutsch");
        /// `Some("deu")`
        pub const LANG_TERM: Option<&str> = Some("deu");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("German")`
        pub const LANGUAGE: Option<&str> = Some("German");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Luxembourg")`
        pub const TERRITORY: Option<&str> = Some("Luxembourg");
        /// `Some("German locale for Luxemburg")`
        pub const TITLE: Option<&str> = Some("German locale for Luxemburg");
    }
    pub use super::de_DE::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"€"`
        pub const CURRENCY_SYMBOL: &str = "€";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"EUR "`
        pub const INT_CURR_SYMBOL: &str = "EUR ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const MON_THOUSANDS_SEP: &str = ".";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `4`
        pub const N_SIGN_POSN: i64 = 4;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `4`
        pub const P_SIGN_POSN: i64 = 4;
    }
    pub use super::de_DE::LC_NUMERIC;
    pub mod LC_TELEPHONE {
        /// `Some("352")`
        pub const INT_PREFIX: Option<&str> = Some("352");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["So", "Mo", "Di", "Mi", "Do", "Fr", "Sa"]`
        pub const ABDAY: &[&str] = &["So", "Mo", "Di", "Mi", "Do", "Fr", "Sa"];
        /// `&["Jan", "Feb", "Mär", "Apr", "Mai", "Jun", "Jul", "Aug", "Sep", "Okt", "Nov", "Dez"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Mär", "Apr", "Mai", "Jun", "Jul", "Aug", "Sep", "Okt", "Nov", "Dez"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Sonntag", "Montag", "Dienstag", "Mittwoch", "Donnerstag", "Freitag", "Samstag"]`
        pub const DAY: &[&str] = &["Sonntag", "Montag", "Dienstag", "Mittwoch", "Donnerstag", "Freitag", "Samstag"];
        /// `"%Y-%m-%d"`
        pub const D_FMT: &str = "%Y-%m-%d";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Januar", "Februar", "März", "April", "Mai", "Juni", "Juli", "August", "September", "Oktober", "November", "Dezember"]`
        pub const MON: &[&str] = &["Januar", "Februar", "März", "April", "Mai", "Juni", "Juli", "August", "September", "Oktober", "November", "Dezember"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod de_LU_euro {
    pub use super::de_LU::LC_ADDRESS;
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("https://www.gnu.org/software/libc/")`
        pub const ADDRESS: Option<&str> = Some("https://www.gnu.org/software/libc/");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-08-20")`
        pub const DATE: Option<&str> = Some("2000-08-20");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("German")`
        pub const LANGUAGE: Option<&str> = Some("German");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Free Software Foundation, Inc.")`
        pub const SOURCE: Option<&str> = Some("Free Software Foundation, Inc.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Luxembourg")`
        pub const TERRITORY: Option<&str> = Some("Luxembourg");
        /// `Some("German locale for Luxemburg with Euro")`
        pub const TITLE: Option<&str> = Some("German locale for Luxemburg with Euro");
    }
    pub use super::de_DE::LC_MESSAGES;
    pub use super::de_LU::LC_MONETARY;
    pub use super::de_DE::LC_NUMERIC;
    pub use super::de_LU::LC_TELEPHONE;
    pub use super::de_LU::LC_TIME;
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod doi_IN {
    pub mod LC_ADDRESS {
        /// `Some("IND")`
        pub const COUNTRY_AB3: Option<&str> = Some("IND");
        /// `Some("IN")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IN");
        /// `Some("IND")`
        pub const COUNTRY_CAR: Option<&str> = Some("IND");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("भारत")`
        pub const COUNTRY_NAME: Option<&str> = Some("भारत");
        /// `Some(356)`
        pub const COUNTRY_NUM: Option<i64> = Some(356);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("doi")`
        pub const LANG_LIB: Option<&str> = Some("doi");
        /// `Some("डोगरी")`
        pub const LANG_NAME: Option<&str> = Some("डोगरी");
        /// `Some("doi")`
        pub const LANG_TERM: Option<&str> = Some("doi");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Level 1, Tower X, Cybercity, Magarpatta City, Hadapsar, Pune-411013 ")`
        pub const ADDRESS: Option<&str> = Some("Level 1, Tower X, Cybercity, Magarpatta City, Hadapsar, Pune-411013 ");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2012-02-16")`
        pub const DATE: Option<&str> = Some("2012-02-16");
        /// `Some("libc-alpha@sourceware.org")`
        pub const EMAIL: Option<&str> = Some("libc-alpha@sourceware.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Dogri")`
        pub const LANGUAGE: Option<&str> = Some("Dogri");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Red Hat Pune")`
        pub const SOURCE: Option<&str> = Some("Red Hat Pune");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("India")`
        pub const TERRITORY: Option<&str> = Some("India");
        /// `Some("Dogri language locale for India")`
        pub const TITLE: Option<&str> = Some("Dogri language locale for India");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNन]"`
        pub const NOEXPR: &str = "^[-0nNन]";
        /// `Some("ना")`
        pub const NOSTR: Option<&str> = Some("ना");
        /// `"^[+1yYऑ]"`
        pub const YESEXPR: &str = "^[+1yYऑ]";
        /// `Some("ऑह")`
        pub const YESSTR: Option<&str> = Some("ऑह");
    }
    pub use super::hi_IN::LC_MONETARY;
    pub use super::hi_IN::LC_NUMERIC;
    pub use super::hi_IN::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["ऐत", "सोम", "म\u{902}गल", "ब\u{941}ध", "बीर", "श\u{941}क\u{94d}कर", "श\u{94d}नीचर"]`
        pub const ABDAY: &[&str] = &["ऐत", "सोम", "म\u{902}गल", "ब\u{941}ध", "बीर", "श\u{941}क\u{94d}कर", "श\u{94d}नीचर"];
        /// `&["जनवरी", "फरवरी", "मार\u{94d}च", "एप\u{94d}र\u{948}ल", "म\u{947}ई", "ज\u{942}न", "ज\u{942}ल\u{948}", "अगस\u{94d}त", "सित\u{902}बर", "अक\u{94d}त\u{942}बर", "नव\u{902}बर", "दिस\u{902}बर"]`
        pub const ABMON: &[&str] = &["जनवरी", "फरवरी", "मार\u{94d}च", "एप\u{94d}र\u{948}ल", "म\u{947}ई", "ज\u{942}न", "ज\u{942}ल\u{948}", "अगस\u{94d}त", "सित\u{902}बर", "अक\u{94d}त\u{942}बर", "नव\u{902}बर", "दिस\u{902}बर"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["सञ\u{902}", "सब\u{947}र"]`
        pub const AM_PM: &[&str] = &["सञ\u{902}", "सब\u{947}र"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["ऐतबार", "सोमबार", "म\u{902}गलबर", "ब\u{941}धबार", "बीरबार", "श\u{941}क\u{94d}करबार", "श\u{94d}नीचरबार"]`
        pub const DAY: &[&str] = &["ऐतबार", "सोमबार", "म\u{902}गलबर", "ब\u{941}धबार", "बीरबार", "श\u{941}क\u{94d}करबार", "श\u{94d}नीचरबार"];
        /// `"%-d/%-m/%y"`
        pub const D_FMT: &str = "%-d/%-m/%y";
        /// `"%A %d %b %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A %d %b %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["जनवरी", "फरवरी", "मार\u{94d}च", "एप\u{94d}र\u{948}ल", "म\u{947}ई", "ज\u{942}न", "ज\u{942}ल\u{948}", "अगस\u{94d}त", "सित\u{902}बर", "अक\u{94d}त\u{942}बर", "नव\u{902}बर", "दिस\u{902}बर"]`
        pub const MON: &[&str] = &["जनवरी", "फरवरी", "मार\u{94d}च", "एप\u{94d}र\u{948}ल", "म\u{947}ई", "ज\u{942}न", "ज\u{942}ल\u{948}", "अगस\u{94d}त", "सित\u{902}बर", "अक\u{94d}त\u{942}बर", "नव\u{902}बर", "दिस\u{902}बर"];
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT: &str = "%I:%M:%S %p %Z";
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p %Z";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod dsb_DE {
    pub mod LC_ADDRESS {
        /// `Some("DEU")`
        pub const COUNTRY_AB3: Option<&str> = Some("DEU");
        /// `Some("DE")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("DE");
        /// `Some("D")`
        pub const COUNTRY_CAR: Option<&str> = Some("D");
        /// `Some("3")`
        pub const COUNTRY_ISBN: Option<&str> = Some("3");
        /// `Some("Nimska")`
        pub const COUNTRY_NAME: Option<&str> = Some("Nimska");
        /// `Some(276)`
        pub const COUNTRY_NUM: Option<i64> = Some(276);
        /// `Some("D")`
        pub const COUNTRY_POST: Option<&str> = Some("D");
        /// `Some("")`
        pub const LANG_AB: Option<&str> = Some("");
        /// `Some("dsb")`
        pub const LANG_LIB: Option<&str> = Some("dsb");
        /// `Some("dolnoserbšćina")`
        pub const LANG_NAME: Option<&str> = Some("dolnoserbšćina");
        /// `Some("dsb")`
        pub const LANG_TERM: Option<&str> = Some("dsb");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("")`
        pub const DATE: Option<&str> = Some("");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Lower Sorbian")`
        pub const LANGUAGE: Option<&str> = Some("Lower Sorbian");
        /// `Some("0.1")`
        pub const REVISION: Option<&str> = Some("0.1");
        /// `Some("Information from Michael Wolf")`
        pub const SOURCE: Option<&str> = Some("Information from Michael Wolf");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Germany")`
        pub const TERRITORY: Option<&str> = Some("Germany");
        /// `Some("Lower Sorbian locale for Germany")`
        pub const TITLE: Option<&str> = Some("Lower Sorbian locale for Germany");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("ně")`
        pub const NOSTR: Option<&str> = Some("ně");
        /// `"^[+1jJhHyY]"`
        pub const YESEXPR: &str = "^[+1jJhHyY]";
        /// `Some("jo")`
        pub const YESSTR: Option<&str> = Some("jo");
    }
    pub use super::de_DE::LC_MONETARY;
    pub use super::de_DE::LC_NUMERIC;
    pub use super::de_DE::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Nj", "Pó", "Wa", "Sr", "St", "Pě", "So"]`
        pub const ABDAY: &[&str] = &["Nj", "Pó", "Wa", "Sr", "St", "Pě", "So"];
        /// `&["Jan", "Feb", "Měr", "Apr", "Maj", "Jun", "Jul", "Awg", "Sep", "Okt", "Now", "Dec"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Měr", "Apr", "Maj", "Jun", "Jul", "Awg", "Sep", "Okt", "Now", "Dec"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `Some(&["Januar", "Februar", "Měrc", "Apryl", "Maj", "Junij", "Julij", "Awgust", "September", "Oktober", "Nowember", "December"])`
        pub const ALT_MON: Option<&[&str]> = Some(&["Januar", "Februar", "Měrc", "Apryl", "Maj", "Junij", "Julij", "Awgust", "September", "Oktober", "Nowember", "December"]);
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Njeźela", "Pónjeźele", "Wałtora", "Srjoda", "Stwórtk", "Pětk", "Sobota"]`
        pub const DAY: &[&str] = &["Njeźela", "Pónjeźele", "Wałtora", "Srjoda", "Stwórtk", "Pětk", "Sobota"];
        /// `"%d.%m.%Y"`
        pub const D_FMT: &str = "%d.%m.%Y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["januara", "februara", "měrca", "apryla", "maja", "junija", "julija", "awgusta", "septembra", "oktobra", "nowembra", "decembra"]`
        pub const MON: &[&str] = &["januara", "februara", "měrca", "apryla", "maja", "junija", "julija", "awgusta", "septembra", "oktobra", "nowembra", "decembra"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod dv_MV {
    pub mod LC_ADDRESS {
        /// `Some("MDV")`
        pub const COUNTRY_AB3: Option<&str> = Some("MDV");
        /// `Some("MV")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("MV");
        /// `Some("MV")`
        pub const COUNTRY_CAR: Option<&str> = Some("MV");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("ދ\u{7a8}ވ\u{7ac}ހ\u{7a8}ރ\u{7a7}އ\u{7b0}ޖ\u{7ac}")`
        pub const COUNTRY_NAME: Option<&str> = Some("ދ\u{7a8}ވ\u{7ac}ހ\u{7a8}ރ\u{7a7}އ\u{7b0}ޖ\u{7ac}");
        /// `Some(462)`
        pub const COUNTRY_NUM: Option<i64> = Some(462);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("dv")`
        pub const LANG_AB: Option<&str> = Some("dv");
        /// `Some("div")`
        pub const LANG_LIB: Option<&str> = Some("div");
        /// `Some("ދ\u{7a8}ވ\u{7ac}ހ\u{7a8}")`
        pub const LANG_NAME: Option<&str> = Some("ދ\u{7a8}ވ\u{7ac}ހ\u{7a8}");
        /// `Some("div")`
        pub const LANG_TERM: Option<&str> = Some("div");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2006-05-13")`
        pub const DATE: Option<&str> = Some("2006-05-13");
        /// `Some("sofwath@hotmail.com")`
        pub const EMAIL: Option<&str> = Some("sofwath@hotmail.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Divehi")`
        pub const LANGUAGE: Option<&str> = Some("Divehi");
        /// `Some("0.1")`
        pub const REVISION: Option<&str> = Some("0.1");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Maldives")`
        pub const TERRITORY: Option<&str> = Some("Maldives");
        /// `Some("Dhivehi Language Locale for Maldives")`
        pub const TITLE: Option<&str> = Some("Dhivehi Language Locale for Maldives");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `None`
        pub const NOSTR: Option<&str> = None;
        /// `"^[+1yY]"`
        pub const YESEXPR: &str = "^[+1yY]";
        /// `None`
        pub const YESSTR: Option<&str> = None;
    }
    pub mod LC_MONETARY {
        /// `"ރ."`
        pub const CURRENCY_SYMBOL: &str = "ރ.";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"MVR "`
        pub const INT_CURR_SYMBOL: &str = "MVR ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `2`
        pub const P_SIGN_POSN: i64 = 2;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("960")`
        pub const INT_PREFIX: Option<&str> = Some("960");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `Some("%a %l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("%a %l");
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["އ\u{7a7}ދ\u{7a9}އ\u{7b0}ތ\u{7a6}", "ހ\u{7af}މ\u{7a6}", "އ\u{7a6}ނ\u{7b0}ގ\u{7a7}ރ\u{7a6}", "ބ\u{7aa}ދ\u{7a6}", "ބ\u{7aa}ރ\u{7a7}ސ\u{7b0}ފ\u{7a6}ތ\u{7a8}", "ހ\u{7aa}ކ\u{7aa}ރ\u{7aa}", "ހ\u{7ae}ނ\u{7a8}ހ\u{7a8}ރ\u{7aa}"]`
        pub const ABDAY: &[&str] = &["އ\u{7a7}ދ\u{7a9}އ\u{7b0}ތ\u{7a6}", "ހ\u{7af}މ\u{7a6}", "އ\u{7a6}ނ\u{7b0}ގ\u{7a7}ރ\u{7a6}", "ބ\u{7aa}ދ\u{7a6}", "ބ\u{7aa}ރ\u{7a7}ސ\u{7b0}ފ\u{7a6}ތ\u{7a8}", "ހ\u{7aa}ކ\u{7aa}ރ\u{7aa}", "ހ\u{7ae}ނ\u{7a8}ހ\u{7a8}ރ\u{7aa}"];
        /// `&["ޖ\u{7ac}ނ\u{7aa}އ\u{7a6}ރ\u{7a9}", "ފ\u{7ac}ބ\u{7b0}ރ\u{7aa}އ\u{7a6}ރ\u{7a9}", "މ\u{7a7}ރޗ\u{7b0}", "އ\u{7ac}ޕ\u{7b0}ރ\u{7a9}ލ\u{7b0}", "މ\u{7ac}އ\u{7a8}", "ޖ\u{7ab}ނ\u{7b0}", "ޖ\u{7aa}ލ\u{7a6}އ\u{7a8}", "އ\u{7ae}ގ\u{7a6}ސ\u{7b0}ޓ\u{7b0}", "ސ\u{7ac}ޕ\u{7b0}ޓ\u{7ac}ނ\u{7b0}ބ\u{7a6}ރ", "އ\u{7ae}ކ\u{7b0}ޓ\u{7ab}ބ\u{7a6}ރ", "ނ\u{7ae}ވ\u{7ac}ނ\u{7b0}ބ\u{7a6}ރ", "ޑ\u{7a8}ސ\u{7ac}ނ\u{7b0}ބ\u{7a6}ރ"]`
        pub const ABMON: &[&str] = &["ޖ\u{7ac}ނ\u{7aa}އ\u{7a6}ރ\u{7a9}", "ފ\u{7ac}ބ\u{7b0}ރ\u{7aa}އ\u{7a6}ރ\u{7a9}", "މ\u{7a7}ރޗ\u{7b0}", "އ\u{7ac}ޕ\u{7b0}ރ\u{7a9}ލ\u{7b0}", "މ\u{7ac}އ\u{7a8}", "ޖ\u{7ab}ނ\u{7b0}", "ޖ\u{7aa}ލ\u{7a6}އ\u{7a8}", "އ\u{7ae}ގ\u{7a6}ސ\u{7b0}ޓ\u{7b0}", "ސ\u{7ac}ޕ\u{7b0}ޓ\u{7ac}ނ\u{7b0}ބ\u{7a6}ރ", "އ\u{7ae}ކ\u{7b0}ޓ\u{7ab}ބ\u{7a6}ރ", "ނ\u{7ae}ވ\u{7ac}ނ\u{7b0}ބ\u{7a6}ރ", "ޑ\u{7a8}ސ\u{7ac}ނ\u{7b0}ބ\u{7a6}ރ"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["މކ", "މފ"]`
        pub const AM_PM: &[&str] = &["މކ", "މފ"];
        /// `Some(3)`
        pub const CAL_DIRECTION: Option<i64> = Some(3);
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["އ\u{7a7}ދ\u{7a9}އ\u{7b0}ތ\u{7a6}", "ހ\u{7af}މ\u{7a6}", "އ\u{7a6}ނ\u{7b0}ގ\u{7a7}ރ\u{7a6}", "ބ\u{7aa}ދ\u{7a6}", "ބ\u{7aa}ރ\u{7a7}ސ\u{7b0}ފ\u{7a6}ތ\u{7a8}", "ހ\u{7aa}ކ\u{7aa}ރ\u{7aa}", "ހ\u{7ae}ނ\u{7a8}ހ\u{7a8}ރ\u{7aa}"]`
        pub const DAY: &[&str] = &["އ\u{7a7}ދ\u{7a9}އ\u{7b0}ތ\u{7a6}", "ހ\u{7af}މ\u{7a6}", "އ\u{7a6}ނ\u{7b0}ގ\u{7a7}ރ\u{7a6}", "ބ\u{7aa}ދ\u{7a6}", "ބ\u{7aa}ރ\u{7a7}ސ\u{7b0}ފ\u{7a6}ތ\u{7a8}", "ހ\u{7aa}ކ\u{7aa}ރ\u{7aa}", "ހ\u{7ae}ނ\u{7a8}ހ\u{7a8}ރ\u{7aa}"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%Z %H:%M:%S %Y %b %d %a"`
        pub const D_T_FMT: &str = "%Z %H:%M:%S %Y %b %d %a";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(6)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(6);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["ޖ\u{7ac}ނ\u{7aa}އ\u{7a6}ރ\u{7a9}", "ފ\u{7ac}ބ\u{7b0}ރ\u{7aa}އ\u{7a6}ރ\u{7a9}", "މ\u{7a7}ރޗ\u{7b0}", "އ\u{7ac}ޕ\u{7b0}ރ\u{7a9}ލ\u{7b0}", "މ\u{7ac}އ\u{7a8}", "ޖ\u{7ab}ނ\u{7b0}", "ޖ\u{7aa}ލ\u{7a6}އ\u{7a8}", "އ\u{7ae}ގ\u{7a6}ސ\u{7b0}ޓ\u{7b0}", "ސ\u{7ac}ޕ\u{7b0}ޓ\u{7ac}ނ\u{7b0}ބ\u{7a6}ރ", "އ\u{7ae}ކ\u{7b0}ޓ\u{7ab}ބ\u{7a6}ރ", "ނ\u{7ae}ވ\u{7ac}ނ\u{7b0}ބ\u{7a6}ރ", "ޑ\u{7a8}ސ\u{7ac}ނ\u{7b0}ބ\u{7a6}ރ"]`
        pub const MON: &[&str] = &["ޖ\u{7ac}ނ\u{7aa}އ\u{7a6}ރ\u{7a9}", "ފ\u{7ac}ބ\u{7b0}ރ\u{7aa}އ\u{7a6}ރ\u{7a9}", "މ\u{7a7}ރޗ\u{7b0}", "އ\u{7ac}ޕ\u{7b0}ރ\u{7a9}ލ\u{7b0}", "މ\u{7ac}އ\u{7a8}", "ޖ\u{7ab}ނ\u{7b0}", "ޖ\u{7aa}ލ\u{7a6}އ\u{7a8}", "އ\u{7ae}ގ\u{7a6}ސ\u{7b0}ޓ\u{7b0}", "ސ\u{7ac}ޕ\u{7b0}ޓ\u{7ac}ނ\u{7b0}ބ\u{7a6}ރ", "އ\u{7ae}ކ\u{7b0}ޓ\u{7ab}ބ\u{7a6}ރ", "ނ\u{7ae}ވ\u{7ac}ނ\u{7b0}ބ\u{7a6}ރ", "ޑ\u{7a8}ސ\u{7ac}ނ\u{7b0}ބ\u{7a6}ރ"];
        /// `"%H:%M:%S"`
        pub const T_FMT: &str = "%H:%M:%S";
        /// `"%P %I:%M:%S"`
        pub const T_FMT_AMPM: &str = "%P %I:%M:%S";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod dz_BT {
    pub mod LC_ADDRESS {
        /// `Some("BTN")`
        pub const COUNTRY_AB3: Option<&str> = Some("BTN");
        /// `Some("BT")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("BT");
        /// `Some("BHT")`
        pub const COUNTRY_CAR: Option<&str> = Some("BHT");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("འབ\u{fb2}\u{f74}ག")`
        pub const COUNTRY_NAME: Option<&str> = Some("འབ\u{fb2}\u{f74}ག");
        /// `Some(64)`
        pub const COUNTRY_NUM: Option<i64> = Some(64);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("dz")`
        pub const LANG_AB: Option<&str> = Some("dz");
        /// `Some("dzo")`
        pub const LANG_LIB: Option<&str> = Some("dzo");
        /// `Some("ར\u{fab}\u{f7c}ང་ཁ")`
        pub const LANG_NAME: Option<&str> = Some("ར\u{fab}\u{f7c}ང་ཁ");
        /// `Some("dzo")`
        pub const LANG_TERM: Option<&str> = Some("dzo");
        /// `Some("%f%N%a%N%d%N%r%t%e%t%b%N%h%t%s%N%T%N%S%N%z%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%r%t%e%t%b%N%h%t%s%N%T%N%S%N%z%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Kanglung, Bhutan")`
        pub const ADDRESS: Option<&str> = Some("Kanglung, Bhutan");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2004-09-03")`
        pub const DATE: Option<&str> = Some("2004-09-03");
        /// `Some("bug-glibc@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Dzongkha")`
        pub const LANGUAGE: Option<&str> = Some("Dzongkha");
        /// `Some("0.3")`
        pub const REVISION: Option<&str> = Some("0.3");
        /// `Some("Sherubtse College")`
        pub const SOURCE: Option<&str> = Some("Sherubtse College");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Bhutan")`
        pub const TERRITORY: Option<&str> = Some("Bhutan");
        /// `Some("Dzongkha language locale for Bhutan")`
        pub const TITLE: Option<&str> = Some("Dzongkha language locale for Bhutan");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNམ]"`
        pub const NOEXPR: &str = "^[-0nNམ]";
        /// `Some("མ\u{f7a}ན་")`
        pub const NOSTR: Option<&str> = Some("མ\u{f7a}ན་");
        /// `"^[+1yYཨ]"`
        pub const YESEXPR: &str = "^[+1yYཨ]";
        /// `Some("ཨ\u{f72}ན་")`
        pub const YESSTR: Option<&str> = Some("ཨ\u{f72}ན་");
    }
    pub mod LC_MONETARY {
        /// `"Nu."`
        pub const CURRENCY_SYMBOL: &str = "Nu.";
        /// `3`
        pub const FRAC_DIGITS: i64 = 3;
        /// `"BTN "`
        pub const INT_CURR_SYMBOL: &str = "BTN ";
        /// `3`
        pub const INT_FRAC_DIGITS: i64 = 3;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 2]`
        pub const MON_GROUPING: &[i64] = &[3, 2];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `4`
        pub const N_SIGN_POSN: i64 = 4;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `4`
        pub const P_SIGN_POSN: i64 = 4;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3, 2]`
        pub const GROUPING: &[i64] = &[3, 2];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("975")`
        pub const INT_PREFIX: Option<&str> = Some("975");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["ཟ\u{fb3}་", "མ\u{f72}ར་", "ལ\u{fb7}ག་", "པ\u{f74}ར་", "སངས་", "ས\u{fa4}\u{f7a}ན་", "ཉ\u{f72}་"]`
        pub const ABDAY: &[&str] = &["ཟ\u{fb3}་", "མ\u{f72}ར་", "ལ\u{fb7}ག་", "པ\u{f74}ར་", "སངས་", "ས\u{fa4}\u{f7a}ན་", "ཉ\u{f72}་"];
        /// `&["ཟ\u{fb3}་༡", "ཟ\u{fb3}་༢", "ཟ\u{fb3}་༣", "ཟ\u{fb3}་༤", "ཟ\u{fb3}་༥", "ཟ\u{fb3}་༦", "ཟ\u{fb3}་༧", "ཟ\u{fb3}་༨", "ཟ\u{fb3}་༩", "ཟ\u{fb3}་༡༠", "ཟ\u{fb3}་༡༡", "ཟ\u{fb3}་༡༢"]`
        pub const ABMON: &[&str] = &["ཟ\u{fb3}་༡", "ཟ\u{fb3}་༢", "ཟ\u{fb3}་༣", "ཟ\u{fb3}་༤", "ཟ\u{fb3}་༥", "ཟ\u{fb3}་༦", "ཟ\u{fb3}་༧", "ཟ\u{fb3}་༨", "ཟ\u{fb3}་༩", "ཟ\u{fb3}་༡༠", "ཟ\u{fb3}་༡༡", "ཟ\u{fb3}་༡༢"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["ངས་ཆ", "ཕ\u{fb1}\u{f72}་ཆ"]`
        pub const AM_PM: &[&str] = &["ངས་ཆ", "ཕ\u{fb1}\u{f72}་ཆ"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["གཟའ་ཟ\u{fb3}་བ་", "གཟའ་མ\u{f72}ག་དམར་", "གཟའ་ལ\u{fb7}ག་ཕ་", "གཟའ་པ\u{f74}ར་བ\u{f74}་", "གཟའ་པ་སངས་", "གཟའ་ས\u{fa4}\u{f7a}ན་ཕ་", "གཟའ་ཉ\u{f72}་མ་"]`
        pub const DAY: &[&str] = &["གཟའ་ཟ\u{fb3}་བ་", "གཟའ་མ\u{f72}ག་དམར་", "གཟའ་ལ\u{fb7}ག་ཕ་", "གཟའ་པ\u{f74}ར་བ\u{f74}་", "གཟའ་པ་སངས་", "གཟའ་ས\u{fa4}\u{f7a}ན་ཕ་", "གཟའ་ཉ\u{f72}་མ་"];
        /// `"པས\u{fb1}\u{f72}་ལ\u{f7c}%yཟལ%mཚ\u{f7a}ས%d"`
        pub const D_FMT: &str = "པས\u{fb1}\u{f72}་ལ\u{f7c}%yཟལ%mཚ\u{f7a}ས%d";
        /// `"པས\u{fb1}\u{f72}་ལ\u{f7c}%yཟལ%mཚ\u{f7a}ས%dཆ\u{f74}་ཚ\u{f7c}ད%Hཀསར་མ%Mཀསར་ཆ%S"`
        pub const D_T_FMT: &str = "པས\u{fb1}\u{f72}་ལ\u{f7c}%yཟལ%mཚ\u{f7a}ས%dཆ\u{f74}་ཚ\u{f7c}ད%Hཀསར་མ%Mཀསར་ཆ%S";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["ཟ\u{fb3}་བ་དང་པ་", "ཟ\u{fb3}་བ་གཉ\u{f72}ས་པ་", "ཟ\u{fb3}་བ་གས\u{f74}མ་པ་", "ཟ\u{fb3}་བ་བཞ\u{f72}་པ་", "ཟ\u{fb3}་བ་ལ\u{f94}་ཕ་", "ཟ\u{fb3}་བ་ད\u{fb2}\u{f74}ག་པ་", "ཟ\u{fb3}་བ་བད\u{f74}ནཔ་", "ཟ\u{fb3}་བ་བར\u{f92}\u{fb1}ད་པ་", "ཟ\u{fb3}་བ་དག\u{f74}་པ་", "ཟ\u{fb3}་བ་བཅ\u{f74}་པ་", "ཟ\u{fb3}་བ་བཅ\u{f74}་གཅ\u{f72}ག་པ་", "ཟ\u{fb3}་བ་བཅ\u{f74}་གཉ\u{f72}ས་པ་"]`
        pub const MON: &[&str] = &["ཟ\u{fb3}་བ་དང་པ་", "ཟ\u{fb3}་བ་གཉ\u{f72}ས་པ་", "ཟ\u{fb3}་བ་གས\u{f74}མ་པ་", "ཟ\u{fb3}་བ་བཞ\u{f72}་པ་", "ཟ\u{fb3}་བ་ལ\u{f94}་ཕ་", "ཟ\u{fb3}་བ་ད\u{fb2}\u{f74}ག་པ་", "ཟ\u{fb3}་བ་བད\u{f74}ནཔ་", "ཟ\u{fb3}་བ་བར\u{f92}\u{fb1}ད་པ་", "ཟ\u{fb3}་བ་དག\u{f74}་པ་", "ཟ\u{fb3}་བ་བཅ\u{f74}་པ་", "ཟ\u{fb3}་བ་བཅ\u{f74}་གཅ\u{f72}ག་པ་", "ཟ\u{fb3}་བ་བཅ\u{f74}་གཉ\u{f72}ས་པ་"];
        /// `"ཆ\u{f74}་ཚ\u{f7c}ད%Hཀསར་མ%Mཀསར་ཆ%S"`
        pub const T_FMT: &str = "ཆ\u{f74}་ཚ\u{f7c}ད%Hཀསར་མ%Mཀསར་ཆ%S";
        /// `"ཆ\u{f74}་ཚ\u{f7c}ད%Iཀསར་མ%Mཀསར་ཆ%S %p"`
        pub const T_FMT_AMPM: &str = "ཆ\u{f74}་ཚ\u{f7c}ད%Iཀསར་མ%Mཀསར་ཆ%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod el_CY {
    pub mod LC_ADDRESS {
        /// `Some("CYP")`
        pub const COUNTRY_AB3: Option<&str> = Some("CYP");
        /// `Some("CY")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("CY");
        /// `Some("CY")`
        pub const COUNTRY_CAR: Option<&str> = Some("CY");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Κύπρος")`
        pub const COUNTRY_NAME: Option<&str> = Some("Κύπρος");
        /// `Some(196)`
        pub const COUNTRY_NUM: Option<i64> = Some(196);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("el")`
        pub const LANG_AB: Option<&str> = Some("el");
        /// `Some("gre")`
        pub const LANG_LIB: Option<&str> = Some("gre");
        /// `Some("Ελληνικά")`
        pub const LANG_NAME: Option<&str> = Some("Ελληνικά");
        /// `Some("ell")`
        pub const LANG_TERM: Option<&str> = Some("ell");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Konstantinos Margaritis, M. Asias 50, Nafplion 21100, Greece")`
        pub const ADDRESS: Option<&str> = Some("Konstantinos Margaritis, M. Asias 50, Nafplion 21100, Greece");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2004-10-20")`
        pub const DATE: Option<&str> = Some("2004-10-20");
        /// `Some("bug-glibc@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Greek")`
        pub const LANGUAGE: Option<&str> = Some("Greek");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Greek Debian Translation Team")`
        pub const SOURCE: Option<&str> = Some("Greek Debian Translation Team");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Cyprus")`
        pub const TERRITORY: Option<&str> = Some("Cyprus");
        /// `Some("Greek locale for Cyprus")`
        pub const TITLE: Option<&str> = Some("Greek locale for Cyprus");
    }
    pub use super::el_GR::LC_MESSAGES;
    pub use super::el_GR::LC_MONETARY;
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[0, 0]`
        pub const GROUPING: &[i64] = &[0, 0];
        /// `"."`
        pub const THOUSANDS_SEP: &str = ".";
    }
    pub mod LC_TELEPHONE {
        /// `Some("357")`
        pub const INT_PREFIX: Option<&str> = Some("357");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["Κυρ", "Δευ", "Τρι", "Τετ", "Πεμ", "Παρ", "Σαβ"]`
        pub const ABDAY: &[&str] = &["Κυρ", "Δευ", "Τρι", "Τετ", "Πεμ", "Παρ", "Σαβ"];
        /// `&["Ιαν", "Φεβ", "Μαρ", "Απρ", "Μαΐ", "Ιουν", "Ιουλ", "Αυγ", "Σεπ", "Οκτ", "Νοε", "Δεκ"]`
        pub const ABMON: &[&str] = &["Ιαν", "Φεβ", "Μαρ", "Απρ", "Μαΐ", "Ιουν", "Ιουλ", "Αυγ", "Σεπ", "Οκτ", "Νοε", "Δεκ"];
        /// `Some(&["Ιαν", "Φεβ", "Μάρ", "Απρ", "Μάι", "Ιούν", "Ιούλ", "Αύγ", "Σεπ", "Οκτ", "Νοέ", "Δεκ"])`
        pub const AB_ALT_MON: Option<&[&str]> = Some(&["Ιαν", "Φεβ", "Μάρ", "Απρ", "Μάι", "Ιούν", "Ιούλ", "Αύγ", "Σεπ", "Οκτ", "Νοέ", "Δεκ"]);
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `Some(&["Ιανουάριος", "Φεβρουάριος", "Μάρτιος", "Απρίλιος", "Μάιος", "Ιούνιος", "Ιούλιος", "Αύγουστος", "Σεπτέμβριος", "Οκτώβριος", "Νοέμβριος", "Δεκέμβριος"])`
        pub const ALT_MON: Option<&[&str]> = Some(&["Ιανουάριος", "Φεβρουάριος", "Μάρτιος", "Απρίλιος", "Μάιος", "Ιούνιος", "Ιούλιος", "Αύγουστος", "Σεπτέμβριος", "Οκτώβριος", "Νοέμβριος", "Δεκέμβριος"]);
        /// `&["πμ", "μμ"]`
        pub const AM_PM: &[&str] = &["πμ", "μμ"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a %d %b %Y %r %Z")`
        pub const DATE_FMT: Option<&str> = Some("%a %d %b %Y %r %Z");
        /// `&["Κυριακή", "Δευτέρα", "Τρίτη", "Τετάρτη", "Πέμπτη", "Παρασκευή", "Σάββατο"]`
        pub const DAY: &[&str] = &["Κυριακή", "Δευτέρα", "Τρίτη", "Τετάρτη", "Πέμπτη", "Παρασκευή", "Σάββατο"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%a %d %b %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Ιανουαρίου", "Φεβρουαρίου", "Μαρτίου", "Απριλίου", "Μαΐου", "Ιουνίου", "Ιουλίου", "Αυγούστου", "Σεπτεμβρίου", "Οκτωβρίου", "Νοεμβρίου", "Δεκεμβρίου"]`
        pub const MON: &[&str] = &["Ιανουαρίου", "Φεβρουαρίου", "Μαρτίου", "Απριλίου", "Μαΐου", "Ιουνίου", "Ιουλίου", "Αυγούστου", "Σεπτεμβρίου", "Οκτωβρίου", "Νοεμβρίου", "Δεκεμβρίου"];
        /// `"%I:%M:%S %p"`
        pub const T_FMT: &str = "%I:%M:%S %p";
        /// `"%I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod el_GR {
    pub mod LC_ADDRESS {
        /// `Some("GRC")`
        pub const COUNTRY_AB3: Option<&str> = Some("GRC");
        /// `Some("GR")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("GR");
        /// `Some("GR")`
        pub const COUNTRY_CAR: Option<&str> = Some("GR");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Ελλάδα")`
        pub const COUNTRY_NAME: Option<&str> = Some("Ελλάδα");
        /// `Some(300)`
        pub const COUNTRY_NUM: Option<i64> = Some(300);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("el")`
        pub const LANG_AB: Option<&str> = Some("el");
        /// `Some("gre")`
        pub const LANG_LIB: Option<&str> = Some("gre");
        /// `Some("Ελληνικά")`
        pub const LANG_NAME: Option<&str> = Some("Ελληνικά");
        /// `Some("ell")`
        pub const LANG_TERM: Option<&str> = Some("ell");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jorgens Alle 8, DK-1615 Kobenhavn V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jorgens Alle 8, DK-1615 Kobenhavn V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Greek")`
        pub const LANGUAGE: Option<&str> = Some("Greek");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Greece")`
        pub const TERRITORY: Option<&str> = Some("Greece");
        /// `Some("Greek locale for Greece")`
        pub const TITLE: Option<&str> = Some("Greek locale for Greece");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNοΟόΌ]"`
        pub const NOEXPR: &str = "^[-0nNοΟόΌ]";
        /// `Some("όχι")`
        pub const NOSTR: Option<&str> = Some("όχι");
        /// `"^[+1yYνΝ]"`
        pub const YESEXPR: &str = "^[+1yYνΝ]";
        /// `Some("ναι")`
        pub const YESSTR: Option<&str> = Some("ναι");
    }
    pub mod LC_MONETARY {
        /// `"€"`
        pub const CURRENCY_SYMBOL: &str = "€";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"EUR "`
        pub const INT_CURR_SYMBOL: &str = "EUR ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `"."`
        pub const MON_THOUSANDS_SEP: &str = ".";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[0, 0]`
        pub const GROUPING: &[i64] = &[0, 0];
        /// `"."`
        pub const THOUSANDS_SEP: &str = ".";
    }
    pub mod LC_TELEPHONE {
        /// `Some("30")`
        pub const INT_PREFIX: Option<&str> = Some("30");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["Κυρ", "Δευ", "Τρι", "Τετ", "Πεμ", "Παρ", "Σαβ"]`
        pub const ABDAY: &[&str] = &["Κυρ", "Δευ", "Τρι", "Τετ", "Πεμ", "Παρ", "Σαβ"];
        /// `&["Ιαν", "Φεβ", "Μαρ", "Απρ", "Μαΐ", "Ιουν", "Ιουλ", "Αυγ", "Σεπ", "Οκτ", "Νοε", "Δεκ"]`
        pub const ABMON: &[&str] = &["Ιαν", "Φεβ", "Μαρ", "Απρ", "Μαΐ", "Ιουν", "Ιουλ", "Αυγ", "Σεπ", "Οκτ", "Νοε", "Δεκ"];
        /// `Some(&["Ιαν", "Φεβ", "Μάρ", "Απρ", "Μάι", "Ιούν", "Ιούλ", "Αύγ", "Σεπ", "Οκτ", "Νοέ", "Δεκ"])`
        pub const AB_ALT_MON: Option<&[&str]> = Some(&["Ιαν", "Φεβ", "Μάρ", "Απρ", "Μάι", "Ιούν", "Ιούλ", "Αύγ", "Σεπ", "Οκτ", "Νοέ", "Δεκ"]);
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `Some(&["Ιανουάριος", "Φεβρουάριος", "Μάρτιος", "Απρίλιος", "Μάιος", "Ιούνιος", "Ιούλιος", "Αύγουστος", "Σεπτέμβριος", "Οκτώβριος", "Νοέμβριος", "Δεκέμβριος"])`
        pub const ALT_MON: Option<&[&str]> = Some(&["Ιανουάριος", "Φεβρουάριος", "Μάρτιος", "Απρίλιος", "Μάιος", "Ιούνιος", "Ιούλιος", "Αύγουστος", "Σεπτέμβριος", "Οκτώβριος", "Νοέμβριος", "Δεκέμβριος"]);
        /// `&["πμ", "μμ"]`
        pub const AM_PM: &[&str] = &["πμ", "μμ"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a %d %b %Y %r %Z")`
        pub const DATE_FMT: Option<&str> = Some("%a %d %b %Y %r %Z");
        /// `&["Κυριακή", "Δευτέρα", "Τρίτη", "Τετάρτη", "Πέμπτη", "Παρασκευή", "Σάββατο"]`
        pub const DAY: &[&str] = &["Κυριακή", "Δευτέρα", "Τρίτη", "Τετάρτη", "Πέμπτη", "Παρασκευή", "Σάββατο"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%a %d %b %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Ιανουαρίου", "Φεβρουαρίου", "Μαρτίου", "Απριλίου", "Μαΐου", "Ιουνίου", "Ιουλίου", "Αυγούστου", "Σεπτεμβρίου", "Οκτωβρίου", "Νοεμβρίου", "Δεκεμβρίου"]`
        pub const MON: &[&str] = &["Ιανουαρίου", "Φεβρουαρίου", "Μαρτίου", "Απριλίου", "Μαΐου", "Ιουνίου", "Ιουλίου", "Αυγούστου", "Σεπτεμβρίου", "Οκτωβρίου", "Νοεμβρίου", "Δεκεμβρίου"];
        /// `"%I:%M:%S %p"`
        pub const T_FMT: &str = "%I:%M:%S %p";
        /// `"%I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod el_GR_euro {
    pub use super::el_GR::LC_ADDRESS;
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("https://www.gnu.org/software/libc/")`
        pub const ADDRESS: Option<&str> = Some("https://www.gnu.org/software/libc/");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-24")`
        pub const DATE: Option<&str> = Some("2000-06-24");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Greek")`
        pub const LANGUAGE: Option<&str> = Some("Greek");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Free Software Foundation, Inc.")`
        pub const SOURCE: Option<&str> = Some("Free Software Foundation, Inc.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Greece")`
        pub const TERRITORY: Option<&str> = Some("Greece");
        /// `Some("Greek locale for Greece with Euro")`
        pub const TITLE: Option<&str> = Some("Greek locale for Greece with Euro");
    }
    pub use super::el_GR::LC_MESSAGES;
    pub use super::el_GR::LC_MONETARY;
    pub use super::el_GR::LC_NUMERIC;
    pub use super::el_GR::LC_TELEPHONE;
    pub use super::el_GR::LC_TIME;
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod en_AG {
    pub mod LC_ADDRESS {
        /// `Some("ATG")`
        pub const COUNTRY_AB3: Option<&str> = Some("ATG");
        /// `Some("AG")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("AG");
        /// `Some("AG")`
        pub const COUNTRY_CAR: Option<&str> = Some("AG");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Antigua & Barbuda")`
        pub const COUNTRY_NAME: Option<&str> = Some("Antigua & Barbuda");
        /// `Some(28)`
        pub const COUNTRY_NUM: Option<i64> = Some(28);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("en")`
        pub const LANG_AB: Option<&str> = Some("en");
        /// `Some("eng")`
        pub const LANG_LIB: Option<&str> = Some("eng");
        /// `Some("English")`
        pub const LANG_NAME: Option<&str> = Some("English");
        /// `Some("eng")`
        pub const LANG_TERM: Option<&str> = Some("eng");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("https://www.gnu.org/software/libc/")`
        pub const ADDRESS: Option<&str> = Some("https://www.gnu.org/software/libc/");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2008-09-16")`
        pub const DATE: Option<&str> = Some("2008-09-16");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("English")`
        pub const LANGUAGE: Option<&str> = Some("English");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Free Software Foundation, Inc.")`
        pub const SOURCE: Option<&str> = Some("Free Software Foundation, Inc.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Antigua & Barbuda")`
        pub const TERRITORY: Option<&str> = Some("Antigua & Barbuda");
        /// `Some("English language locale for Antigua and Barbuda")`
        pub const TITLE: Option<&str> = Some("English language locale for Antigua and Barbuda");
    }
    pub use super::en_GB::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"$"`
        pub const CURRENCY_SYMBOL: &str = "$";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"XCD "`
        pub const INT_CURR_SYMBOL: &str = "XCD ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub use super::en_GB::LC_NUMERIC;
    pub mod LC_TELEPHONE {
        /// `Some("1")`
        pub const INT_PREFIX: Option<&str> = Some("1");
        /// `Some("011")`
        pub const INT_SELECT: Option<&str> = Some("011");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"]`
        pub const ABDAY: &[&str] = &["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"];
        /// `&["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["am", "pm"]`
        pub const AM_PM: &[&str] = &["am", "pm"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a %e %b %H:%M:%S %Z %Y")`
        pub const DATE_FMT: Option<&str> = Some("%a %e %b %H:%M:%S %Z %Y");
        /// `&["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"]`
        pub const DAY: &[&str] = &["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"]`
        pub const MON: &[&str] = &["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `"%l:%M:%S %P %Z"`
        pub const T_FMT_AMPM: &str = "%l:%M:%S %P %Z";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod en_AU {
    pub mod LC_ADDRESS {
        /// `Some("AUS")`
        pub const COUNTRY_AB3: Option<&str> = Some("AUS");
        /// `Some("AU")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("AU");
        /// `Some("AUS")`
        pub const COUNTRY_CAR: Option<&str> = Some("AUS");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Australia")`
        pub const COUNTRY_NAME: Option<&str> = Some("Australia");
        /// `Some(36)`
        pub const COUNTRY_NUM: Option<i64> = Some(36);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("en")`
        pub const LANG_AB: Option<&str> = Some("en");
        /// `Some("eng")`
        pub const LANG_LIB: Option<&str> = Some("eng");
        /// `Some("English")`
        pub const LANG_NAME: Option<&str> = Some("English");
        /// `Some("eng")`
        pub const LANG_TERM: Option<&str> = Some("eng");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Australian English")`
        pub const LANGUAGE: Option<&str> = Some("Australian English");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Australia")`
        pub const TERRITORY: Option<&str> = Some("Australia");
        /// `Some("English locale for Australia")`
        pub const TITLE: Option<&str> = Some("English locale for Australia");
    }
    pub use super::en_US::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"$"`
        pub const CURRENCY_SYMBOL: &str = "$";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"AUD "`
        pub const INT_CURR_SYMBOL: &str = "AUD ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("61")`
        pub const INT_PREFIX: Option<&str> = Some("61");
        /// `Some("0011")`
        pub const INT_SELECT: Option<&str> = Some("0011");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"]`
        pub const ABDAY: &[&str] = &["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"];
        /// `&["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["AM", "PM"]`
        pub const AM_PM: &[&str] = &["AM", "PM"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"]`
        pub const DAY: &[&str] = &["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"]`
        pub const MON: &[&str] = &["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `"%I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod en_BW {
    pub mod LC_ADDRESS {
        /// `Some("BWA")`
        pub const COUNTRY_AB3: Option<&str> = Some("BWA");
        /// `Some("BW")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("BW");
        /// `Some("BW")`
        pub const COUNTRY_CAR: Option<&str> = Some("BW");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Botswana")`
        pub const COUNTRY_NAME: Option<&str> = Some("Botswana");
        /// `Some(72)`
        pub const COUNTRY_NUM: Option<i64> = Some(72);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("en")`
        pub const LANG_AB: Option<&str> = Some("en");
        /// `Some("eng")`
        pub const LANG_LIB: Option<&str> = Some("eng");
        /// `Some("English")`
        pub const LANG_NAME: Option<&str> = Some("English");
        /// `Some("eng")`
        pub const LANG_TERM: Option<&str> = Some("eng");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("English")`
        pub const LANGUAGE: Option<&str> = Some("English");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Botswana")`
        pub const TERRITORY: Option<&str> = Some("Botswana");
        /// `Some("English locale for Botswana")`
        pub const TITLE: Option<&str> = Some("English locale for Botswana");
    }
    pub use super::en_ZA::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"P"`
        pub const CURRENCY_SYMBOL: &str = "P";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"BWP "`
        pub const INT_CURR_SYMBOL: &str = "BWP ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub use super::en_ZA::LC_NUMERIC;
    pub mod LC_TELEPHONE {
        /// `Some("267")`
        pub const INT_PREFIX: Option<&str> = Some("267");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub use super::en_ZA::LC_TIME;
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod en_CA {
    pub mod LC_ADDRESS {
        /// `Some("CAN")`
        pub const COUNTRY_AB3: Option<&str> = Some("CAN");
        /// `Some("CA")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("CA");
        /// `Some("CDN")`
        pub const COUNTRY_CAR: Option<&str> = Some("CDN");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Canada")`
        pub const COUNTRY_NAME: Option<&str> = Some("Canada");
        /// `Some(124)`
        pub const COUNTRY_NUM: Option<i64> = Some(124);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("en")`
        pub const LANG_AB: Option<&str> = Some("en");
        /// `Some("eng")`
        pub const LANG_LIB: Option<&str> = Some("eng");
        /// `Some("English")`
        pub const LANG_NAME: Option<&str> = Some("English");
        /// `Some("eng")`
        pub const LANG_TERM: Option<&str> = Some("eng");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Canadian English")`
        pub const LANGUAGE: Option<&str> = Some("Canadian English");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Canada")`
        pub const TERRITORY: Option<&str> = Some("Canada");
        /// `Some("English locale for Canada")`
        pub const TITLE: Option<&str> = Some("English locale for Canada");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("no")`
        pub const NOSTR: Option<&str> = Some("no");
        /// `"^[+1yYoO]"`
        pub const YESEXPR: &str = "^[+1yYoO]";
        /// `Some("yes")`
        pub const YESSTR: Option<&str> = Some("yes");
    }
    pub mod LC_MONETARY {
        /// `"$"`
        pub const CURRENCY_SYMBOL: &str = "$";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"CAD "`
        pub const INT_CURR_SYMBOL: &str = "CAD ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("1")`
        pub const INT_PREFIX: Option<&str> = Some("1");
        /// `Some("011")`
        pub const INT_SELECT: Option<&str> = Some("011");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"]`
        pub const ABDAY: &[&str] = &["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"];
        /// `&["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["AM", "PM"]`
        pub const AM_PM: &[&str] = &["AM", "PM"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"]`
        pub const DAY: &[&str] = &["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"];
        /// `"%Y-%m-%d"`
        pub const D_FMT: &str = "%Y-%m-%d";
        /// `"%a %d %b %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"]`
        pub const MON: &[&str] = &["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"];
        /// `"%I:%M:%S %p"`
        pub const T_FMT: &str = "%I:%M:%S %p";
        /// `"%I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod en_DK {
    pub mod LC_ADDRESS {
        /// `Some("DNK")`
        pub const COUNTRY_AB3: Option<&str> = Some("DNK");
        /// `Some("DK")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("DK");
        /// `Some("DK")`
        pub const COUNTRY_CAR: Option<&str> = Some("DK");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Denmark")`
        pub const COUNTRY_NAME: Option<&str> = Some("Denmark");
        /// `Some(208)`
        pub const COUNTRY_NUM: Option<i64> = Some(208);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("en")`
        pub const LANG_AB: Option<&str> = Some("en");
        /// `Some("eng")`
        pub const LANG_LIB: Option<&str> = Some("eng");
        /// `Some("English")`
        pub const LANG_NAME: Option<&str> = Some("English");
        /// `Some("eng")`
        pub const LANG_TERM: Option<&str> = Some("eng");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Kollegievej 6, DK-2920 Charlottenlund, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Kollegievej 6, DK-2920 Charlottenlund, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("English")`
        pub const LANGUAGE: Option<&str> = Some("English");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Danish Standards Association")`
        pub const SOURCE: Option<&str> = Some("Danish Standards Association");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Denmark")`
        pub const TERRITORY: Option<&str> = Some("Denmark");
        /// `Some("English locale for Denmark")`
        pub const TITLE: Option<&str> = Some("English locale for Denmark");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("no")`
        pub const NOSTR: Option<&str> = Some("no");
        /// `"^[+1yYjJsSoO]"`
        pub const YESEXPR: &str = "^[+1yYjJsSoO]";
        /// `Some("yes")`
        pub const YESSTR: Option<&str> = Some("yes");
    }
    pub mod LC_MONETARY {
        /// `"kr."`
        pub const CURRENCY_SYMBOL: &str = "kr.";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"DKK "`
        pub const INT_CURR_SYMBOL: &str = "DKK ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const MON_THOUSANDS_SEP: &str = ".";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const THOUSANDS_SEP: &str = ".";
    }
    pub use super::da_DK::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"]`
        pub const ABDAY: &[&str] = &["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"];
        /// `&["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"]`
        pub const DAY: &[&str] = &["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"];
        /// `"%Y-%m-%d"`
        pub const D_FMT: &str = "%Y-%m-%d";
        /// `"%Y-%m-%dT%T %Z"`
        pub const D_T_FMT: &str = "%Y-%m-%dT%T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"]`
        pub const MON: &[&str] = &["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod en_GB {
    pub mod LC_ADDRESS {
        /// `Some("GBR")`
        pub const COUNTRY_AB3: Option<&str> = Some("GBR");
        /// `Some("GB")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("GB");
        /// `Some("GB")`
        pub const COUNTRY_CAR: Option<&str> = Some("GB");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("United Kingdom")`
        pub const COUNTRY_NAME: Option<&str> = Some("United Kingdom");
        /// `Some(826)`
        pub const COUNTRY_NUM: Option<i64> = Some(826);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("en")`
        pub const LANG_AB: Option<&str> = Some("en");
        /// `Some("eng")`
        pub const LANG_LIB: Option<&str> = Some("eng");
        /// `Some("English")`
        pub const LANG_NAME: Option<&str> = Some("English");
        /// `Some("eng")`
        pub const LANG_TERM: Option<&str> = Some("eng");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Keld Simonsen")`
        pub const CONTACT: Option<&str> = Some("Keld Simonsen");
        /// `Some("2000-06-28")`
        pub const DATE: Option<&str> = Some("2000-06-28");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("British English")`
        pub const LANGUAGE: Option<&str> = Some("British English");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("United Kingdom")`
        pub const TERRITORY: Option<&str> = Some("United Kingdom");
        /// `Some("English locale for Britain")`
        pub const TITLE: Option<&str> = Some("English locale for Britain");
    }
    pub use super::en_US::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"£"`
        pub const CURRENCY_SYMBOL: &str = "£";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"GBP "`
        pub const INT_CURR_SYMBOL: &str = "GBP ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("44")`
        pub const INT_PREFIX: Option<&str> = Some("44");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `Some("%A %l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("%A %l");
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"]`
        pub const ABDAY: &[&str] = &["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"];
        /// `&["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["am", "pm"]`
        pub const AM_PM: &[&str] = &["am", "pm"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a %e %b %H:%M:%S %Z %Y")`
        pub const DATE_FMT: Option<&str> = Some("%a %e %b %H:%M:%S %Z %Y");
        /// `&["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"]`
        pub const DAY: &[&str] = &["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"]`
        pub const MON: &[&str] = &["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `"%l:%M:%S %P %Z"`
        pub const T_FMT_AMPM: &str = "%l:%M:%S %P %Z";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod en_HK {
    pub mod LC_ADDRESS {
        /// `Some("HKG")`
        pub const COUNTRY_AB3: Option<&str> = Some("HKG");
        /// `Some("HK")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("HK");
        /// `Some("HK")`
        pub const COUNTRY_CAR: Option<&str> = Some("HK");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Hong Kong SAR China")`
        pub const COUNTRY_NAME: Option<&str> = Some("Hong Kong SAR China");
        /// `Some(344)`
        pub const COUNTRY_NUM: Option<i64> = Some(344);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("en")`
        pub const LANG_AB: Option<&str> = Some("en");
        /// `Some("eng")`
        pub const LANG_LIB: Option<&str> = Some("eng");
        /// `Some("English")`
        pub const LANG_NAME: Option<&str> = Some("English");
        /// `Some("eng")`
        pub const LANG_TERM: Option<&str> = Some("eng");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan")`
        pub const ADDRESS: Option<&str> = Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000,October,27 (XML source:2000,July,20)")`
        pub const DATE: Option<&str> = Some("2000,October,27 (XML source:2000,July,20)");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("English")`
        pub const LANGUAGE: Option<&str> = Some("English");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("IBM Globalization Center of Competency, Yamato Software Laboratory")`
        pub const SOURCE: Option<&str> = Some("IBM Globalization Center of Competency, Yamato Software Laboratory");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Hong Kong SAR China")`
        pub const TERRITORY: Option<&str> = Some("Hong Kong SAR China");
        /// `Some("English locale for Hong Kong")`
        pub const TITLE: Option<&str> = Some("English locale for Hong Kong");
    }
    pub use super::en_US::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"HK$"`
        pub const CURRENCY_SYMBOL: &str = "HK$";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"HKD "`
        pub const INT_CURR_SYMBOL: &str = "HKD ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `0`
        pub const N_SIGN_POSN: i64 = 0;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("852")`
        pub const INT_PREFIX: Option<&str> = Some("852");
        /// `Some("001")`
        pub const INT_SELECT: Option<&str> = Some("001");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c ;%a ;%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c ;%a ;%l");
    }
    pub mod LC_TIME {
        /// `&["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"]`
        pub const ABDAY: &[&str] = &["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"];
        /// `&["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["AM", "PM"]`
        pub const AM_PM: &[&str] = &["AM", "PM"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"]`
        pub const DAY: &[&str] = &["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"];
        /// `"%A, %B %d, %Y"`
        pub const D_FMT: &str = "%A, %B %d, %Y";
        /// `"%A, %B %d, %Y %p%I:%M:%S %Z"`
        pub const D_T_FMT: &str = "%A, %B %d, %Y %p%I:%M:%S %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"]`
        pub const MON: &[&str] = &["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"];
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT: &str = "%I:%M:%S %p %Z";
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p %Z";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod en_IE {
    pub mod LC_ADDRESS {
        /// `Some("IRL")`
        pub const COUNTRY_AB3: Option<&str> = Some("IRL");
        /// `Some("IE")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IE");
        /// `Some("IRL")`
        pub const COUNTRY_CAR: Option<&str> = Some("IRL");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Ireland")`
        pub const COUNTRY_NAME: Option<&str> = Some("Ireland");
        /// `Some(372)`
        pub const COUNTRY_NUM: Option<i64> = Some(372);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("en")`
        pub const LANG_AB: Option<&str> = Some("en");
        /// `Some("eng")`
        pub const LANG_LIB: Option<&str> = Some("eng");
        /// `Some("English")`
        pub const LANG_NAME: Option<&str> = Some("English");
        /// `Some("eng")`
        pub const LANG_TERM: Option<&str> = Some("eng");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("English")`
        pub const LANGUAGE: Option<&str> = Some("English");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Ireland")`
        pub const TERRITORY: Option<&str> = Some("Ireland");
        /// `Some("English locale for Ireland")`
        pub const TITLE: Option<&str> = Some("English locale for Ireland");
    }
    pub use super::en_GB::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"€"`
        pub const CURRENCY_SYMBOL: &str = "€";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"EUR "`
        pub const INT_CURR_SYMBOL: &str = "EUR ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub use super::en_GB::LC_NUMERIC;
    pub mod LC_TELEPHONE {
        /// `Some("353")`
        pub const INT_PREFIX: Option<&str> = Some("353");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"]`
        pub const ABDAY: &[&str] = &["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"];
        /// `&["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"]`
        pub const DAY: &[&str] = &["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"]`
        pub const MON: &[&str] = &["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod en_IE_euro {
    pub use super::en_IE::LC_ADDRESS;
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("https://www.gnu.org/software/libc/")`
        pub const ADDRESS: Option<&str> = Some("https://www.gnu.org/software/libc/");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-08-20")`
        pub const DATE: Option<&str> = Some("2000-08-20");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("English")`
        pub const LANGUAGE: Option<&str> = Some("English");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Free Software Foundation, Inc.")`
        pub const SOURCE: Option<&str> = Some("Free Software Foundation, Inc.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Ireland")`
        pub const TERRITORY: Option<&str> = Some("Ireland");
        /// `Some("English locale for Ireland with Euro")`
        pub const TITLE: Option<&str> = Some("English locale for Ireland with Euro");
    }
    pub use super::en_IE::LC_MESSAGES;
    pub use super::en_IE::LC_MONETARY;
    pub use super::en_GB::LC_NUMERIC;
    pub use super::en_IE::LC_TELEPHONE;
    pub use super::en_IE::LC_TIME;
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod en_IL {
    pub mod LC_ADDRESS {
        /// `Some("ISR")`
        pub const COUNTRY_AB3: Option<&str> = Some("ISR");
        /// `Some("IL")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IL");
        /// `Some("IL")`
        pub const COUNTRY_CAR: Option<&str> = Some("IL");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Israel")`
        pub const COUNTRY_NAME: Option<&str> = Some("Israel");
        /// `Some(376)`
        pub const COUNTRY_NUM: Option<i64> = Some(376);
        /// `Some("ISR")`
        pub const COUNTRY_POST: Option<&str> = Some("ISR");
        /// `Some("en")`
        pub const LANG_AB: Option<&str> = Some("en");
        /// `Some("eng")`
        pub const LANG_LIB: Option<&str> = Some("eng");
        /// `Some("English")`
        pub const LANG_NAME: Option<&str> = Some("English");
        /// `Some("eng")`
        pub const LANG_TERM: Option<&str> = Some("eng");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("https://www.gnu.org/software/libc/")`
        pub const ADDRESS: Option<&str> = Some("https://www.gnu.org/software/libc/");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2016-04-19")`
        pub const DATE: Option<&str> = Some("2016-04-19");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("English")`
        pub const LANGUAGE: Option<&str> = Some("English");
        /// `Some("1.1")`
        pub const REVISION: Option<&str> = Some("1.1");
        /// `Some("Free Software Foundation, Inc.")`
        pub const SOURCE: Option<&str> = Some("Free Software Foundation, Inc.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Israel")`
        pub const TERRITORY: Option<&str> = Some("Israel");
        /// `Some("English locale for Israel")`
        pub const TITLE: Option<&str> = Some("English locale for Israel");
    }
    pub use super::en_US::LC_MESSAGES;
    pub use super::he_IL::LC_MONETARY;
    pub use super::he_IL::LC_NUMERIC;
    pub use super::he_IL::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"]`
        pub const ABDAY: &[&str] = &["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"];
        /// `&["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"]`
        pub const DAY: &[&str] = &["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `Some(1)`
        pub const FIRST_WORKDAY: Option<i64> = Some(1);
        /// `&["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"]`
        pub const MON: &[&str] = &["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod en_IN {
    pub mod LC_ADDRESS {
        /// `Some("IND")`
        pub const COUNTRY_AB3: Option<&str> = Some("IND");
        /// `Some("IN")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IN");
        /// `Some("IND")`
        pub const COUNTRY_CAR: Option<&str> = Some("IND");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("India")`
        pub const COUNTRY_NAME: Option<&str> = Some("India");
        /// `Some(356)`
        pub const COUNTRY_NUM: Option<i64> = Some(356);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("en")`
        pub const LANG_AB: Option<&str> = Some("en");
        /// `Some("eng")`
        pub const LANG_LIB: Option<&str> = Some("eng");
        /// `Some("English")`
        pub const LANG_NAME: Option<&str> = Some("English");
        /// `Some("eng")`
        pub const LANG_TERM: Option<&str> = Some("eng");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan")`
        pub const ADDRESS: Option<&str> = Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000,October,27 (XML source:2000,July,20)")`
        pub const DATE: Option<&str> = Some("2000,October,27 (XML source:2000,July,20)");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("English")`
        pub const LANGUAGE: Option<&str> = Some("English");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("IBM Globalization Center of Competency, Yamato Software Laboratory")`
        pub const SOURCE: Option<&str> = Some("IBM Globalization Center of Competency, Yamato Software Laboratory");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("India")`
        pub const TERRITORY: Option<&str> = Some("India");
        /// `Some("English language locale for India")`
        pub const TITLE: Option<&str> = Some("English language locale for India");
    }
    pub use super::en_US::LC_MESSAGES;
    pub use super::hi_IN::LC_MONETARY;
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3, 2]`
        pub const GROUPING: &[i64] = &[3, 2];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub use super::hi_IN::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"]`
        pub const ABDAY: &[&str] = &["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"];
        /// `&["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["AM", "PM"]`
        pub const AM_PM: &[&str] = &["AM", "PM"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"]`
        pub const DAY: &[&str] = &["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%A %d %B %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A %d %B %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"]`
        pub const MON: &[&str] = &["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"];
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT: &str = "%I:%M:%S %p %Z";
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p %Z";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod en_NG {
    pub mod LC_ADDRESS {
        /// `Some("NGA")`
        pub const COUNTRY_AB3: Option<&str> = Some("NGA");
        /// `Some("NG")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("NG");
        /// `Some("WAN")`
        pub const COUNTRY_CAR: Option<&str> = Some("WAN");
        /// `Some("978")`
        pub const COUNTRY_ISBN: Option<&str> = Some("978");
        /// `Some("Nigeria")`
        pub const COUNTRY_NAME: Option<&str> = Some("Nigeria");
        /// `Some(566)`
        pub const COUNTRY_NUM: Option<i64> = Some(566);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("en")`
        pub const LANG_AB: Option<&str> = Some("en");
        /// `Some("eng")`
        pub const LANG_LIB: Option<&str> = Some("eng");
        /// `Some("English")`
        pub const LANG_NAME: Option<&str> = Some("English");
        /// `Some("eng")`
        pub const LANG_TERM: Option<&str> = Some("eng");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2006-02-01")`
        pub const DATE: Option<&str> = Some("2006-02-01");
        /// `Some("")`
        pub const EMAIL: Option<&str> = Some("");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("English")`
        pub const LANGUAGE: Option<&str> = Some("English");
        /// `Some("0.2")`
        pub const REVISION: Option<&str> = Some("0.2");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Nigeria")`
        pub const TERRITORY: Option<&str> = Some("Nigeria");
        /// `Some("English locale for Nigeria")`
        pub const TITLE: Option<&str> = Some("English locale for Nigeria");
    }
    pub use super::en_US::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"₦"`
        pub const CURRENCY_SYMBOL: &str = "₦";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"NGN "`
        pub const INT_CURR_SYMBOL: &str = "NGN ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("234")`
        pub const INT_PREFIX: Option<&str> = Some("234");
        /// `Some("009")`
        pub const INT_SELECT: Option<&str> = Some("009");
        /// `Some("(%A) %l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("(%A) %l");
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"]`
        pub const ABDAY: &[&str] = &["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"];
        /// `&["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"]`
        pub const DAY: &[&str] = &["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"]`
        pub const MON: &[&str] = &["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod en_NZ {
    pub mod LC_ADDRESS {
        /// `Some("NZL")`
        pub const COUNTRY_AB3: Option<&str> = Some("NZL");
        /// `Some("NZ")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("NZ");
        /// `Some("NZ")`
        pub const COUNTRY_CAR: Option<&str> = Some("NZ");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("New Zealand")`
        pub const COUNTRY_NAME: Option<&str> = Some("New Zealand");
        /// `Some(554)`
        pub const COUNTRY_NUM: Option<i64> = Some(554);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("en")`
        pub const LANG_AB: Option<&str> = Some("en");
        /// `Some("eng")`
        pub const LANG_LIB: Option<&str> = Some("eng");
        /// `Some("English")`
        pub const LANG_NAME: Option<&str> = Some("English");
        /// `Some("eng")`
        pub const LANG_TERM: Option<&str> = Some("eng");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("English")`
        pub const LANGUAGE: Option<&str> = Some("English");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("New Zealand")`
        pub const TERRITORY: Option<&str> = Some("New Zealand");
        /// `Some("English locale for New Zealand")`
        pub const TITLE: Option<&str> = Some("English locale for New Zealand");
    }
    pub use super::en_US::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"$"`
        pub const CURRENCY_SYMBOL: &str = "$";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"NZD "`
        pub const INT_CURR_SYMBOL: &str = "NZD ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("64")`
        pub const INT_PREFIX: Option<&str> = Some("64");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"]`
        pub const ABDAY: &[&str] = &["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"];
        /// `&["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["AM", "PM"]`
        pub const AM_PM: &[&str] = &["AM", "PM"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"]`
        pub const DAY: &[&str] = &["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"]`
        pub const MON: &[&str] = &["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `"%I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod en_PH {
    pub mod LC_ADDRESS {
        /// `Some("PHL")`
        pub const COUNTRY_AB3: Option<&str> = Some("PHL");
        /// `Some("PH")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("PH");
        /// `Some("RP")`
        pub const COUNTRY_CAR: Option<&str> = Some("RP");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Philippines")`
        pub const COUNTRY_NAME: Option<&str> = Some("Philippines");
        /// `Some(608)`
        pub const COUNTRY_NUM: Option<i64> = Some(608);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("en")`
        pub const LANG_AB: Option<&str> = Some("en");
        /// `Some("eng")`
        pub const LANG_LIB: Option<&str> = Some("eng");
        /// `Some("English")`
        pub const LANG_NAME: Option<&str> = Some("English");
        /// `Some("eng")`
        pub const LANG_TERM: Option<&str> = Some("eng");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan")`
        pub const ADDRESS: Option<&str> = Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000,October,27 (XML source:2000,July,20)")`
        pub const DATE: Option<&str> = Some("2000,October,27 (XML source:2000,July,20)");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("English")`
        pub const LANGUAGE: Option<&str> = Some("English");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("IBM Globalization Center of Competency, Yamato Software Laboratory")`
        pub const SOURCE: Option<&str> = Some("IBM Globalization Center of Competency, Yamato Software Laboratory");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Philippines")`
        pub const TERRITORY: Option<&str> = Some("Philippines");
        /// `Some("English language locale for Philippines")`
        pub const TITLE: Option<&str> = Some("English language locale for Philippines");
    }
    pub use super::en_US::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"₱"`
        pub const CURRENCY_SYMBOL: &str = "₱";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"PHP "`
        pub const INT_CURR_SYMBOL: &str = "PHP ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `0`
        pub const N_SIGN_POSN: i64 = 0;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("63")`
        pub const INT_PREFIX: Option<&str> = Some("63");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c ;%a ;%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c ;%a ;%l");
    }
    pub mod LC_TIME {
        /// `&["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"]`
        pub const ABDAY: &[&str] = &["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"];
        /// `&["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["AM", "PM"]`
        pub const AM_PM: &[&str] = &["AM", "PM"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"]`
        pub const DAY: &[&str] = &["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"];
        /// `"%A, %d %B, %Y"`
        pub const D_FMT: &str = "%A, %d %B, %Y";
        /// `"%A, %d %B, %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A, %d %B, %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"]`
        pub const MON: &[&str] = &["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"];
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT: &str = "%I:%M:%S %p %Z";
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p %Z";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod en_SC {
    pub mod LC_ADDRESS {
        /// `Some("SYC")`
        pub const COUNTRY_AB3: Option<&str> = Some("SYC");
        /// `Some("SC")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("SC");
        /// `Some("SY")`
        pub const COUNTRY_CAR: Option<&str> = Some("SY");
        /// `Some("978-99931")`
        pub const COUNTRY_ISBN: Option<&str> = Some("978-99931");
        /// `Some("Seychelles")`
        pub const COUNTRY_NAME: Option<&str> = Some("Seychelles");
        /// `Some(690)`
        pub const COUNTRY_NUM: Option<i64> = Some(690);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("en")`
        pub const LANG_AB: Option<&str> = Some("en");
        /// `Some("eng")`
        pub const LANG_LIB: Option<&str> = Some("eng");
        /// `Some("English")`
        pub const LANG_NAME: Option<&str> = Some("English");
        /// `Some("eng")`
        pub const LANG_TERM: Option<&str> = Some("eng");
        /// `Some("%f%N%h%s%N%T")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%h%s%N%T");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2017-07-24")`
        pub const DATE: Option<&str> = Some("2017-07-24");
        /// `Some("akhilesh.k@samsung.com")`
        pub const EMAIL: Option<&str> = Some("akhilesh.k@samsung.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("English")`
        pub const LANGUAGE: Option<&str> = Some("English");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Samsung Electronics Co., Ltd.")`
        pub const SOURCE: Option<&str> = Some("Samsung Electronics Co., Ltd.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Seychelles")`
        pub const TERRITORY: Option<&str> = Some("Seychelles");
        /// `Some("English locale for the Seychelles")`
        pub const TITLE: Option<&str> = Some("English locale for the Seychelles");
    }
    pub use super::en_GB::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"SR"`
        pub const CURRENCY_SYMBOL: &str = "SR";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"SCR "`
        pub const INT_CURR_SYMBOL: &str = "SCR ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `Some(1)`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = Some(1);
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `Some(1)`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = Some(1);
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub use super::en_GB::LC_NUMERIC;
    pub mod LC_TELEPHONE {
        /// `Some("248")`
        pub const INT_PREFIX: Option<&str> = Some("248");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %l");
    }
    pub use super::en_GB::LC_TIME;
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod en_SG {
    pub mod LC_ADDRESS {
        /// `Some("SGP")`
        pub const COUNTRY_AB3: Option<&str> = Some("SGP");
        /// `Some("SG")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("SG");
        /// `Some("SGP")`
        pub const COUNTRY_CAR: Option<&str> = Some("SGP");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Singapore")`
        pub const COUNTRY_NAME: Option<&str> = Some("Singapore");
        /// `Some(702)`
        pub const COUNTRY_NUM: Option<i64> = Some(702);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("en")`
        pub const LANG_AB: Option<&str> = Some("en");
        /// `Some("eng")`
        pub const LANG_LIB: Option<&str> = Some("eng");
        /// `Some("English")`
        pub const LANG_NAME: Option<&str> = Some("English");
        /// `Some("eng")`
        pub const LANG_TERM: Option<&str> = Some("eng");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan")`
        pub const ADDRESS: Option<&str> = Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000,October,27 (XML source:2000,July,20)")`
        pub const DATE: Option<&str> = Some("2000,October,27 (XML source:2000,July,20)");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("English")`
        pub const LANGUAGE: Option<&str> = Some("English");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("IBM Globalization Center of Competency, Yamato Software Laboratory")`
        pub const SOURCE: Option<&str> = Some("IBM Globalization Center of Competency, Yamato Software Laboratory");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Singapore")`
        pub const TERRITORY: Option<&str> = Some("Singapore");
        /// `Some("English language locale for Singapore")`
        pub const TITLE: Option<&str> = Some("English language locale for Singapore");
    }
    pub use super::en_US::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"$"`
        pub const CURRENCY_SYMBOL: &str = "$";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"SGD "`
        pub const INT_CURR_SYMBOL: &str = "SGD ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `0`
        pub const N_SIGN_POSN: i64 = 0;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("65")`
        pub const INT_PREFIX: Option<&str> = Some("65");
        /// `Some("001")`
        pub const INT_SELECT: Option<&str> = Some("001");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c ;%a ;%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c ;%a ;%l");
    }
    pub mod LC_TIME {
        /// `&["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"]`
        pub const ABDAY: &[&str] = &["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"];
        /// `&["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["AM", "PM"]`
        pub const AM_PM: &[&str] = &["AM", "PM"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"]`
        pub const DAY: &[&str] = &["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%a %d %b %Y %I:%M:%S %p"`
        pub const D_T_FMT: &str = "%a %d %b %Y %I:%M:%S %p";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"]`
        pub const MON: &[&str] = &["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `"%I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod en_US {
    pub mod LC_ADDRESS {
        /// `Some("USA")`
        pub const COUNTRY_AB3: Option<&str> = Some("USA");
        /// `Some("US")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("US");
        /// `Some("USA")`
        pub const COUNTRY_CAR: Option<&str> = Some("USA");
        /// `Some("0")`
        pub const COUNTRY_ISBN: Option<&str> = Some("0");
        /// `Some("United States")`
        pub const COUNTRY_NAME: Option<&str> = Some("United States");
        /// `Some(840)`
        pub const COUNTRY_NUM: Option<i64> = Some(840);
        /// `Some("USA")`
        pub const COUNTRY_POST: Option<&str> = Some("USA");
        /// `Some("en")`
        pub const LANG_AB: Option<&str> = Some("en");
        /// `Some("eng")`
        pub const LANG_LIB: Option<&str> = Some("eng");
        /// `Some("English")`
        pub const LANG_NAME: Option<&str> = Some("English");
        /// `Some("eng")`
        pub const LANG_TERM: Option<&str> = Some("eng");
        /// `Some("%a%N%f%N%d%N%b%N%h %s %e %r%N%T, %S %z%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%a%N%f%N%d%N%b%N%h %s %e %r%N%T, %S %z%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("https://www.gnu.org/software/libc/")`
        pub const ADDRESS: Option<&str> = Some("https://www.gnu.org/software/libc/");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-24")`
        pub const DATE: Option<&str> = Some("2000-06-24");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("American English")`
        pub const LANGUAGE: Option<&str> = Some("American English");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Free Software Foundation, Inc.")`
        pub const SOURCE: Option<&str> = Some("Free Software Foundation, Inc.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("United States")`
        pub const TERRITORY: Option<&str> = Some("United States");
        /// `Some("English locale for the USA")`
        pub const TITLE: Option<&str> = Some("English locale for the USA");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("no")`
        pub const NOSTR: Option<&str> = Some("no");
        /// `"^[+1yY]"`
        pub const YESEXPR: &str = "^[+1yY]";
        /// `Some("yes")`
        pub const YESSTR: Option<&str> = Some("yes");
    }
    pub mod LC_MONETARY {
        /// `"$"`
        pub const CURRENCY_SYMBOL: &str = "$";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"USD "`
        pub const INT_CURR_SYMBOL: &str = "USD ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `Some(1)`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = Some(1);
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `Some(1)`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = Some(1);
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("1")`
        pub const INT_PREFIX: Option<&str> = Some("1");
        /// `Some("11")`
        pub const INT_SELECT: Option<&str> = Some("11");
        /// `Some("(%a) %l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("(%a) %l");
        /// `Some("+%c (%a) %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c (%a) %l");
    }
    pub mod LC_TIME {
        /// `&["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"]`
        pub const ABDAY: &[&str] = &["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"];
        /// `&["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["AM", "PM"]`
        pub const AM_PM: &[&str] = &["AM", "PM"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a %d %b %Y %r %Z")`
        pub const DATE_FMT: Option<&str> = Some("%a %d %b %Y %r %Z");
        /// `&["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"]`
        pub const DAY: &[&str] = &["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"];
        /// `"%m/%d/%Y"`
        pub const D_FMT: &str = "%m/%d/%Y";
        /// `"%a %d %b %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"]`
        pub const MON: &[&str] = &["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"];
        /// `"%I:%M:%S %p"`
        pub const T_FMT: &str = "%I:%M:%S %p";
        /// `"%I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod en_ZA {
    pub mod LC_ADDRESS {
        /// `Some("ZAF")`
        pub const COUNTRY_AB3: Option<&str> = Some("ZAF");
        /// `Some("ZA")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("ZA");
        /// `Some("ZA")`
        pub const COUNTRY_CAR: Option<&str> = Some("ZA");
        /// `Some("0")`
        pub const COUNTRY_ISBN: Option<&str> = Some("0");
        /// `Some("South Africa")`
        pub const COUNTRY_NAME: Option<&str> = Some("South Africa");
        /// `Some(710)`
        pub const COUNTRY_NUM: Option<i64> = Some(710);
        /// `Some("ZA")`
        pub const COUNTRY_POST: Option<&str> = Some("ZA");
        /// `Some("en")`
        pub const LANG_AB: Option<&str> = Some("en");
        /// `Some("eng")`
        pub const LANG_LIB: Option<&str> = Some("eng");
        /// `Some("English")`
        pub const LANG_NAME: Option<&str> = Some("English");
        /// `Some("eng")`
        pub const LANG_TERM: Option<&str> = Some("eng");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `Some("Translate.org.za")`
        pub const ABBREVIATION: Option<&str> = Some("Translate.org.za");
        /// `Some("Box 28364, Sunnyside, 0132, South Africa")`
        pub const ADDRESS: Option<&str> = Some("Box 28364, Sunnyside, 0132, South Africa");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
        ]);
        /// `Some("Dwayne Bailey")`
        pub const CONTACT: Option<&str> = Some("Dwayne Bailey");
        /// `Some("2007-04-19")`
        pub const DATE: Option<&str> = Some("2007-04-19");
        /// `Some("dwayne@translate.org.za")`
        pub const EMAIL: Option<&str> = Some("dwayne@translate.org.za");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("English")`
        pub const LANGUAGE: Option<&str> = Some("English");
        /// `Some("1.3")`
        pub const REVISION: Option<&str> = Some("1.3");
        /// `Some("Zuza Software Foundation (Translate.org.za)")`
        pub const SOURCE: Option<&str> = Some("Zuza Software Foundation (Translate.org.za)");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("South Africa")`
        pub const TERRITORY: Option<&str> = Some("South Africa");
        /// `Some("English locale for South Africa")`
        pub const TITLE: Option<&str> = Some("English locale for South Africa");
    }
    pub use super::en_US::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"R"`
        pub const CURRENCY_SYMBOL: &str = "R";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"ZAR "`
        pub const INT_CURR_SYMBOL: &str = "ZAR ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("27")`
        pub const INT_PREFIX: Option<&str> = Some("27");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `Some("(%A) %l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("(%A) %l");
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"]`
        pub const ABDAY: &[&str] = &["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"];
        /// `&["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"]`
        pub const DAY: &[&str] = &["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"]`
        pub const MON: &[&str] = &["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod en_ZM {
    pub mod LC_ADDRESS {
        /// `Some("ZMB")`
        pub const COUNTRY_AB3: Option<&str> = Some("ZMB");
        /// `Some("ZM")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("ZM");
        /// `Some("Z")`
        pub const COUNTRY_CAR: Option<&str> = Some("Z");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Zambia")`
        pub const COUNTRY_NAME: Option<&str> = Some("Zambia");
        /// `Some(894)`
        pub const COUNTRY_NUM: Option<i64> = Some(894);
        /// `Some("ZM")`
        pub const COUNTRY_POST: Option<&str> = Some("ZM");
        /// `Some("en")`
        pub const LANG_AB: Option<&str> = Some("en");
        /// `Some("eng")`
        pub const LANG_LIB: Option<&str> = Some("eng");
        /// `Some("English")`
        pub const LANG_NAME: Option<&str> = Some("English");
        /// `Some("eng")`
        pub const LANG_TERM: Option<&str> = Some("eng");
        /// `Some("%a%N%f%N%e %h%N%b %s%N%z%N%T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%a%N%f%N%e %h%N%b %s%N%z%N%T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `None`
        pub const ADDRESS: Option<&str> = None;
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
        ]);
        /// `Some("Martin Benjamin")`
        pub const CONTACT: Option<&str> = Some("Martin Benjamin");
        /// `Some("2011-04-18")`
        pub const DATE: Option<&str> = Some("2011-04-18");
        /// `Some("locales@africanlocalization.net")`
        pub const EMAIL: Option<&str> = Some("locales@africanlocalization.net");
        /// `None`
        pub const FAX: Option<&str> = None;
        /// `Some("English")`
        pub const LANGUAGE: Option<&str> = Some("English");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("ANLoc")`
        pub const SOURCE: Option<&str> = Some("ANLoc");
        /// `None`
        pub const TEL: Option<&str> = None;
        /// `Some("Zambia")`
        pub const TERRITORY: Option<&str> = Some("Zambia");
        /// `Some("English locale for Zambia")`
        pub const TITLE: Option<&str> = Some("English locale for Zambia");
    }
    pub use super::bem_ZM::LC_MESSAGES;
    pub use super::bem_ZM::LC_MONETARY;
    pub use super::en_GB::LC_NUMERIC;
    pub use super::bem_ZM::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"]`
        pub const ABDAY: &[&str] = &["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"];
        /// `&["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["am", "pm"]`
        pub const AM_PM: &[&str] = &["am", "pm"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a %e %b %H:%M:%S %Z %Y")`
        pub const DATE_FMT: Option<&str> = Some("%a %e %b %H:%M:%S %Z %Y");
        /// `&["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"]`
        pub const DAY: &[&str] = &["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"]`
        pub const MON: &[&str] = &["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `"%l:%M:%S %P %Z"`
        pub const T_FMT_AMPM: &str = "%l:%M:%S %P %Z";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod en_ZW {
    pub mod LC_ADDRESS {
        /// `Some("ZWE")`
        pub const COUNTRY_AB3: Option<&str> = Some("ZWE");
        /// `Some("ZW")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("ZW");
        /// `Some("ZW")`
        pub const COUNTRY_CAR: Option<&str> = Some("ZW");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Zimbabwe")`
        pub const COUNTRY_NAME: Option<&str> = Some("Zimbabwe");
        /// `Some(716)`
        pub const COUNTRY_NUM: Option<i64> = Some(716);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("en")`
        pub const LANG_AB: Option<&str> = Some("en");
        /// `Some("eng")`
        pub const LANG_LIB: Option<&str> = Some("eng");
        /// `Some("English")`
        pub const LANG_NAME: Option<&str> = Some("English");
        /// `Some("eng")`
        pub const LANG_TERM: Option<&str> = Some("eng");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("English")`
        pub const LANGUAGE: Option<&str> = Some("English");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Zimbabwe")`
        pub const TERRITORY: Option<&str> = Some("Zimbabwe");
        /// `Some("English locale for Zimbabwe")`
        pub const TITLE: Option<&str> = Some("English locale for Zimbabwe");
    }
    pub use super::en_ZA::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"$"`
        pub const CURRENCY_SYMBOL: &str = "$";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"USD "`
        pub const INT_CURR_SYMBOL: &str = "USD ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub use super::en_ZA::LC_NUMERIC;
    pub mod LC_TELEPHONE {
        /// `Some("263")`
        pub const INT_PREFIX: Option<&str> = Some("263");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub use super::en_ZA::LC_TIME;
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod eo {
    pub mod LC_ADDRESS {
        /// `None`
        pub const COUNTRY_AB3: Option<&str> = None;
        /// `None`
        pub const COUNTRY_ABTWO: Option<&str> = None;
        /// `None`
        pub const COUNTRY_CAR: Option<&str> = None;
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `None`
        pub const COUNTRY_NAME: Option<&str> = None;
        /// `None`
        pub const COUNTRY_NUM: Option<i64> = None;
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("eo")`
        pub const LANG_AB: Option<&str> = Some("eo");
        /// `None`
        pub const LANG_LIB: Option<&str> = None;
        /// `Some("esperanto")`
        pub const LANG_NAME: Option<&str> = Some("esperanto");
        /// `Some("epo")`
        pub const LANG_TERM: Option<&str> = Some("epo");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2016-06-10")`
        pub const DATE: Option<&str> = Some("2016-06-10");
        /// `Some("debian-esperanto@lists.debian.org")`
        pub const EMAIL: Option<&str> = Some("debian-esperanto@lists.debian.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Esperanto")`
        pub const LANGUAGE: Option<&str> = Some("Esperanto");
        /// `Some("draft")`
        pub const REVISION: Option<&str> = Some("draft");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("")`
        pub const TERRITORY: Option<&str> = Some("");
        /// `Some("Esperanto language locale")`
        pub const TITLE: Option<&str> = Some("Esperanto language locale");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("Ne")`
        pub const NOSTR: Option<&str> = Some("Ne");
        /// `"^[+1JjYy]"`
        pub const YESEXPR: &str = "^[+1JjYy]";
        /// `Some("Jes")`
        pub const YESSTR: Option<&str> = Some("Jes");
    }
    pub mod LC_MONETARY {
        /// `"¤"`
        pub const CURRENCY_SYMBOL: &str = "¤";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"XDR "`
        pub const INT_CURR_SYMBOL: &str = "XDR ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const MON_THOUSANDS_SEP: &str = "\u{202f}";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[0, 0]`
        pub const GROUPING: &[i64] = &[0, 0];
        /// `""`
        pub const THOUSANDS_SEP: &str = "";
    }
    pub mod LC_TELEPHONE {
        /// `None`
        pub const INT_PREFIX: Option<&str> = None;
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a%t%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a%t%l");
    }
    pub mod LC_TIME {
        /// `&["dim", "lun", "mar", "mer", "ĵaŭ", "ven", "sab"]`
        pub const ABDAY: &[&str] = &["dim", "lun", "mar", "mer", "ĵaŭ", "ven", "sab"];
        /// `&["Jan", "Feb", "Mar", "Apr", "Maj", "Jun", "Jul", "Aŭg", "Sep", "Okt", "Nov", "Dec"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Mar", "Apr", "Maj", "Jun", "Jul", "Aŭg", "Sep", "Okt", "Nov", "Dec"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["dimanĉo", "lundo", "mardo", "merkredo", "ĵaŭdo", "vendredo", "sabato"]`
        pub const DAY: &[&str] = &["dimanĉo", "lundo", "mardo", "merkredo", "ĵaŭdo", "vendredo", "sabato"];
        /// `"%Y-%m-%d"`
        pub const D_FMT: &str = "%Y-%m-%d";
        /// `"%a %d %b %Y %T %z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Januaro", "Februaro", "Marto", "Aprilo", "Majo", "Junio", "Julio", "Aŭgusto", "Septembro", "Oktobro", "Novembro", "Decembro"]`
        pub const MON: &[&str] = &["Januaro", "Februaro", "Marto", "Aprilo", "Majo", "Junio", "Julio", "Aŭgusto", "Septembro", "Oktobro", "Novembro", "Decembro"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod es_AR {
    pub mod LC_ADDRESS {
        /// `Some("ARG")`
        pub const COUNTRY_AB3: Option<&str> = Some("ARG");
        /// `Some("AR")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("AR");
        /// `Some("RA")`
        pub const COUNTRY_CAR: Option<&str> = Some("RA");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Argentina")`
        pub const COUNTRY_NAME: Option<&str> = Some("Argentina");
        /// `Some(32)`
        pub const COUNTRY_NUM: Option<i64> = Some(32);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("es")`
        pub const LANG_AB: Option<&str> = Some("es");
        /// `Some("spa")`
        pub const LANG_LIB: Option<&str> = Some("spa");
        /// `Some("español")`
        pub const LANG_NAME: Option<&str> = Some("español");
        /// `Some("spa")`
        pub const LANG_TERM: Option<&str> = Some("spa");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Spanish")`
        pub const LANGUAGE: Option<&str> = Some("Spanish");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Argentina")`
        pub const TERRITORY: Option<&str> = Some("Argentina");
        /// `Some("Spanish locale for Argentina")`
        pub const TITLE: Option<&str> = Some("Spanish locale for Argentina");
    }
    pub use super::es_ES::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"$"`
        pub const CURRENCY_SYMBOL: &str = "$";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"ARS "`
        pub const INT_CURR_SYMBOL: &str = "ARS ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const MON_THOUSANDS_SEP: &str = ".";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const THOUSANDS_SEP: &str = ".";
    }
    pub mod LC_TELEPHONE {
        /// `Some("54")`
        pub const INT_PREFIX: Option<&str> = Some("54");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["dom", "lun", "mar", "mié", "jue", "vie", "sáb"]`
        pub const ABDAY: &[&str] = &["dom", "lun", "mar", "mié", "jue", "vie", "sáb"];
        /// `&["ene", "feb", "mar", "abr", "may", "jun", "jul", "ago", "sep", "oct", "nov", "dic"]`
        pub const ABMON: &[&str] = &["ene", "feb", "mar", "abr", "may", "jun", "jul", "ago", "sep", "oct", "nov", "dic"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"]`
        pub const DAY: &[&str] = &["domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"]`
        pub const MON: &[&str] = &["enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod es_BO {
    pub mod LC_ADDRESS {
        /// `Some("BOL")`
        pub const COUNTRY_AB3: Option<&str> = Some("BOL");
        /// `Some("BO")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("BO");
        /// `Some("BOL")`
        pub const COUNTRY_CAR: Option<&str> = Some("BOL");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Bolivia")`
        pub const COUNTRY_NAME: Option<&str> = Some("Bolivia");
        /// `Some(68)`
        pub const COUNTRY_NUM: Option<i64> = Some(68);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("es")`
        pub const LANG_AB: Option<&str> = Some("es");
        /// `Some("spa")`
        pub const LANG_LIB: Option<&str> = Some("spa");
        /// `Some("español")`
        pub const LANG_NAME: Option<&str> = Some("español");
        /// `Some("spa")`
        pub const LANG_TERM: Option<&str> = Some("spa");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Spanish")`
        pub const LANGUAGE: Option<&str> = Some("Spanish");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Bolivia")`
        pub const TERRITORY: Option<&str> = Some("Bolivia");
        /// `Some("Spanish locale for Bolivia")`
        pub const TITLE: Option<&str> = Some("Spanish locale for Bolivia");
    }
    pub use super::es_ES::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"Bs"`
        pub const CURRENCY_SYMBOL: &str = "Bs";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"BOB "`
        pub const INT_CURR_SYMBOL: &str = "BOB ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const MON_THOUSANDS_SEP: &str = ".";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub use super::es_ES::LC_NUMERIC;
    pub mod LC_TELEPHONE {
        /// `Some("591")`
        pub const INT_PREFIX: Option<&str> = Some("591");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["dom", "lun", "mar", "mié", "jue", "vie", "sáb"]`
        pub const ABDAY: &[&str] = &["dom", "lun", "mar", "mié", "jue", "vie", "sáb"];
        /// `&["ene", "feb", "mar", "abr", "may", "jun", "jul", "ago", "sep", "oct", "nov", "dic"]`
        pub const ABMON: &[&str] = &["ene", "feb", "mar", "abr", "may", "jun", "jul", "ago", "sep", "oct", "nov", "dic"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"]`
        pub const DAY: &[&str] = &["domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"]`
        pub const MON: &[&str] = &["enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod es_CL {
    pub mod LC_ADDRESS {
        /// `Some("CHL")`
        pub const COUNTRY_AB3: Option<&str> = Some("CHL");
        /// `Some("CL")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("CL");
        /// `Some("RCH")`
        pub const COUNTRY_CAR: Option<&str> = Some("RCH");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Chile")`
        pub const COUNTRY_NAME: Option<&str> = Some("Chile");
        /// `Some(152)`
        pub const COUNTRY_NUM: Option<i64> = Some(152);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("es")`
        pub const LANG_AB: Option<&str> = Some("es");
        /// `Some("spa")`
        pub const LANG_LIB: Option<&str> = Some("spa");
        /// `Some("español")`
        pub const LANG_NAME: Option<&str> = Some("español");
        /// `Some("spa")`
        pub const LANG_TERM: Option<&str> = Some("spa");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Spanish")`
        pub const LANGUAGE: Option<&str> = Some("Spanish");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Chile")`
        pub const TERRITORY: Option<&str> = Some("Chile");
        /// `Some("Spanish locale for Chile")`
        pub const TITLE: Option<&str> = Some("Spanish locale for Chile");
    }
    pub use super::es_ES::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"$"`
        pub const CURRENCY_SYMBOL: &str = "$";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"CLP "`
        pub const INT_CURR_SYMBOL: &str = "CLP ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const MON_THOUSANDS_SEP: &str = ".";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub use super::es_ES::LC_NUMERIC;
    pub mod LC_TELEPHONE {
        /// `Some("56")`
        pub const INT_PREFIX: Option<&str> = Some("56");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub use super::es_BO::LC_TIME;
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod es_CO {
    pub mod LC_ADDRESS {
        /// `Some("COL")`
        pub const COUNTRY_AB3: Option<&str> = Some("COL");
        /// `Some("CO")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("CO");
        /// `Some("CO")`
        pub const COUNTRY_CAR: Option<&str> = Some("CO");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Colombia")`
        pub const COUNTRY_NAME: Option<&str> = Some("Colombia");
        /// `Some(170)`
        pub const COUNTRY_NUM: Option<i64> = Some(170);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("es")`
        pub const LANG_AB: Option<&str> = Some("es");
        /// `Some("spa")`
        pub const LANG_LIB: Option<&str> = Some("spa");
        /// `Some("español")`
        pub const LANG_NAME: Option<&str> = Some("español");
        /// `Some("spa")`
        pub const LANG_TERM: Option<&str> = Some("spa");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Spanish")`
        pub const LANGUAGE: Option<&str> = Some("Spanish");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Colombia")`
        pub const TERRITORY: Option<&str> = Some("Colombia");
        /// `Some("Spanish locale for Colombia")`
        pub const TITLE: Option<&str> = Some("Spanish locale for Colombia");
    }
    pub use super::es_ES::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"$"`
        pub const CURRENCY_SYMBOL: &str = "$";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"COP "`
        pub const INT_CURR_SYMBOL: &str = "COP ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const MON_THOUSANDS_SEP: &str = ".";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub use super::es_ES::LC_NUMERIC;
    pub mod LC_TELEPHONE {
        /// `Some("57")`
        pub const INT_PREFIX: Option<&str> = Some("57");
        /// `Some("009")`
        pub const INT_SELECT: Option<&str> = Some("009");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["dom", "lun", "mar", "mié", "jue", "vie", "sáb"]`
        pub const ABDAY: &[&str] = &["dom", "lun", "mar", "mié", "jue", "vie", "sáb"];
        /// `&["ene", "feb", "mar", "abr", "may", "jun", "jul", "ago", "sep", "oct", "nov", "dic"]`
        pub const ABMON: &[&str] = &["ene", "feb", "mar", "abr", "may", "jun", "jul", "ago", "sep", "oct", "nov", "dic"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["AM", "PM"]`
        pub const AM_PM: &[&str] = &["AM", "PM"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"]`
        pub const DAY: &[&str] = &["domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"]`
        pub const MON: &[&str] = &["enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `"%I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod es_CR {
    pub mod LC_ADDRESS {
        /// `Some("CRI")`
        pub const COUNTRY_AB3: Option<&str> = Some("CRI");
        /// `Some("CR")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("CR");
        /// `Some("CR")`
        pub const COUNTRY_CAR: Option<&str> = Some("CR");
        /// `Some("9930,9977,9968")`
        pub const COUNTRY_ISBN: Option<&str> = Some("9930,9977,9968");
        /// `Some("Costa Rica")`
        pub const COUNTRY_NAME: Option<&str> = Some("Costa Rica");
        /// `Some(188)`
        pub const COUNTRY_NUM: Option<i64> = Some(188);
        /// `Some("CR")`
        pub const COUNTRY_POST: Option<&str> = Some("CR");
        /// `Some("es")`
        pub const LANG_AB: Option<&str> = Some("es");
        /// `Some("spa")`
        pub const LANG_LIB: Option<&str> = Some("spa");
        /// `Some("español")`
        pub const LANG_NAME: Option<&str> = Some("español");
        /// `Some("spa")`
        pub const LANG_TERM: Option<&str> = Some("spa");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("https://www.gnu.org/software/libc/")`
        pub const ADDRESS: Option<&str> = Some("https://www.gnu.org/software/libc/");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2009-12-23")`
        pub const DATE: Option<&str> = Some("2009-12-23");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Spanish")`
        pub const LANGUAGE: Option<&str> = Some("Spanish");
        /// `Some("1.1")`
        pub const REVISION: Option<&str> = Some("1.1");
        /// `Some("Free Software Foundation, Inc.")`
        pub const SOURCE: Option<&str> = Some("Free Software Foundation, Inc.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Costa Rica")`
        pub const TERRITORY: Option<&str> = Some("Costa Rica");
        /// `Some("Spanish locale for Costa Rica")`
        pub const TITLE: Option<&str> = Some("Spanish locale for Costa Rica");
    }
    pub use super::es_ES::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"₡"`
        pub const CURRENCY_SYMBOL: &str = "₡";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"CRC "`
        pub const INT_CURR_SYMBOL: &str = "CRC ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const MON_THOUSANDS_SEP: &str = "\u{202f}";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const THOUSANDS_SEP: &str = "\u{202f}";
    }
    pub mod LC_TELEPHONE {
        /// `Some("506")`
        pub const INT_PREFIX: Option<&str> = Some("506");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `Some("%l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("%l");
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["dom", "lun", "mar", "mié", "jue", "vie", "sáb"]`
        pub const ABDAY: &[&str] = &["dom", "lun", "mar", "mié", "jue", "vie", "sáb"];
        /// `&["ene", "feb", "mar", "abr", "may", "jun", "jul", "ago", "sep", "oct", "nov", "dic"]`
        pub const ABMON: &[&str] = &["ene", "feb", "mar", "abr", "may", "jun", "jul", "ago", "sep", "oct", "nov", "dic"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["a.m.", "p.m."]`
        pub const AM_PM: &[&str] = &["a.m.", "p.m."];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a %e %b %H:%M:%S %Z %Y")`
        pub const DATE_FMT: Option<&str> = Some("%a %e %b %H:%M:%S %Z %Y");
        /// `&["domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"]`
        pub const DAY: &[&str] = &["domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"]`
        pub const MON: &[&str] = &["enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `"%I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod es_CU {
    pub mod LC_ADDRESS {
        /// `Some("CUB")`
        pub const COUNTRY_AB3: Option<&str> = Some("CUB");
        /// `Some("CU")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("CU");
        /// `Some("C")`
        pub const COUNTRY_CAR: Option<&str> = Some("C");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Cuba")`
        pub const COUNTRY_NAME: Option<&str> = Some("Cuba");
        /// `Some(192)`
        pub const COUNTRY_NUM: Option<i64> = Some(192);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("es")`
        pub const LANG_AB: Option<&str> = Some("es");
        /// `Some("spa")`
        pub const LANG_LIB: Option<&str> = Some("spa");
        /// `Some("español")`
        pub const LANG_NAME: Option<&str> = Some("español");
        /// `Some("spa")`
        pub const LANG_TERM: Option<&str> = Some("spa");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("San Antonio de los Baøos Highway, Kilometer 1 1/2, University of Informatic Sciences ")`
        pub const ADDRESS: Option<&str> = Some("San Antonio de los Baøos Highway, Kilometer 1 1/2, University of Informatic Sciences ");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2011-10-19")`
        pub const DATE: Option<&str> = Some("2011-10-19");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Spanish")`
        pub const LANGUAGE: Option<&str> = Some("Spanish");
        /// `Some("0.1")`
        pub const REVISION: Option<&str> = Some("0.1");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Cuba")`
        pub const TERRITORY: Option<&str> = Some("Cuba");
        /// `Some("Spanish locale for Cuba")`
        pub const TITLE: Option<&str> = Some("Spanish locale for Cuba");
    }
    pub use super::es_ES::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"$"`
        pub const CURRENCY_SYMBOL: &str = "$";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"CUP "`
        pub const INT_CURR_SYMBOL: &str = "CUP ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const MON_THOUSANDS_SEP: &str = ".";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[0, 0]`
        pub const GROUPING: &[i64] = &[0, 0];
        /// `"."`
        pub const THOUSANDS_SEP: &str = ".";
    }
    pub mod LC_TELEPHONE {
        /// `Some("53")`
        pub const INT_PREFIX: Option<&str> = Some("53");
        /// `Some("119")`
        pub const INT_SELECT: Option<&str> = Some("119");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub use super::es_BO::LC_TIME;
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod es_DO {
    pub mod LC_ADDRESS {
        /// `Some("DOM")`
        pub const COUNTRY_AB3: Option<&str> = Some("DOM");
        /// `Some("DO")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("DO");
        /// `Some("DOM")`
        pub const COUNTRY_CAR: Option<&str> = Some("DOM");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("República Dominicana")`
        pub const COUNTRY_NAME: Option<&str> = Some("República Dominicana");
        /// `Some(214)`
        pub const COUNTRY_NUM: Option<i64> = Some(214);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("es")`
        pub const LANG_AB: Option<&str> = Some("es");
        /// `Some("spa")`
        pub const LANG_LIB: Option<&str> = Some("spa");
        /// `Some("español")`
        pub const LANG_NAME: Option<&str> = Some("español");
        /// `Some("spa")`
        pub const LANG_TERM: Option<&str> = Some("spa");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Spanish")`
        pub const LANGUAGE: Option<&str> = Some("Spanish");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Dominican Republic")`
        pub const TERRITORY: Option<&str> = Some("Dominican Republic");
        /// `Some("Spanish locale for Dominican Republic")`
        pub const TITLE: Option<&str> = Some("Spanish locale for Dominican Republic");
    }
    pub use super::es_ES::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"RD$"`
        pub const CURRENCY_SYMBOL: &str = "RD$";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"DOP "`
        pub const INT_CURR_SYMBOL: &str = "DOP ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("1")`
        pub const INT_PREFIX: Option<&str> = Some("1");
        /// `Some("011")`
        pub const INT_SELECT: Option<&str> = Some("011");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["dom", "lun", "mar", "mié", "jue", "vie", "sáb"]`
        pub const ABDAY: &[&str] = &["dom", "lun", "mar", "mié", "jue", "vie", "sáb"];
        /// `&["ene", "feb", "mar", "abr", "may", "jun", "jul", "ago", "sep", "oct", "nov", "dic"]`
        pub const ABMON: &[&str] = &["ene", "feb", "mar", "abr", "may", "jun", "jul", "ago", "sep", "oct", "nov", "dic"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"]`
        pub const DAY: &[&str] = &["domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"]`
        pub const MON: &[&str] = &["enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod es_EC {
    pub mod LC_ADDRESS {
        /// `Some("ECU")`
        pub const COUNTRY_AB3: Option<&str> = Some("ECU");
        /// `Some("EC")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("EC");
        /// `Some("EC")`
        pub const COUNTRY_CAR: Option<&str> = Some("EC");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Ecuador")`
        pub const COUNTRY_NAME: Option<&str> = Some("Ecuador");
        /// `Some(218)`
        pub const COUNTRY_NUM: Option<i64> = Some(218);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("es")`
        pub const LANG_AB: Option<&str> = Some("es");
        /// `Some("spa")`
        pub const LANG_LIB: Option<&str> = Some("spa");
        /// `Some("español")`
        pub const LANG_NAME: Option<&str> = Some("español");
        /// `Some("spa")`
        pub const LANG_TERM: Option<&str> = Some("spa");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Spanish")`
        pub const LANGUAGE: Option<&str> = Some("Spanish");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Ecuador")`
        pub const TERRITORY: Option<&str> = Some("Ecuador");
        /// `Some("Spanish locale for Ecuador")`
        pub const TITLE: Option<&str> = Some("Spanish locale for Ecuador");
    }
    pub use super::es_ES::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"$"`
        pub const CURRENCY_SYMBOL: &str = "$";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"USD "`
        pub const INT_CURR_SYMBOL: &str = "USD ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const MON_THOUSANDS_SEP: &str = ".";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub use super::es_ES::LC_NUMERIC;
    pub mod LC_TELEPHONE {
        /// `Some("593")`
        pub const INT_PREFIX: Option<&str> = Some("593");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub use super::es_BO::LC_TIME;
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod es_ES {
    pub mod LC_ADDRESS {
        /// `Some("ESP")`
        pub const COUNTRY_AB3: Option<&str> = Some("ESP");
        /// `Some("ES")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("ES");
        /// `Some("E")`
        pub const COUNTRY_CAR: Option<&str> = Some("E");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("España")`
        pub const COUNTRY_NAME: Option<&str> = Some("España");
        /// `Some(724)`
        pub const COUNTRY_NUM: Option<i64> = Some(724);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("es")`
        pub const LANG_AB: Option<&str> = Some("es");
        /// `Some("spa")`
        pub const LANG_LIB: Option<&str> = Some("spa");
        /// `Some("español")`
        pub const LANG_NAME: Option<&str> = Some("español");
        /// `Some("spa")`
        pub const LANG_TERM: Option<&str> = Some("spa");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("European Spanish")`
        pub const LANGUAGE: Option<&str> = Some("European Spanish");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Spain")`
        pub const TERRITORY: Option<&str> = Some("Spain");
        /// `Some("Spanish locale for Spain")`
        pub const TITLE: Option<&str> = Some("Spanish locale for Spain");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("no")`
        pub const NOSTR: Option<&str> = Some("no");
        /// `"^[+1sSyY]"`
        pub const YESEXPR: &str = "^[+1sSyY]";
        /// `Some("sí")`
        pub const YESSTR: Option<&str> = Some("sí");
    }
    pub mod LC_MONETARY {
        /// `"€"`
        pub const CURRENCY_SYMBOL: &str = "€";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"EUR "`
        pub const INT_CURR_SYMBOL: &str = "EUR ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const MON_THOUSANDS_SEP: &str = ".";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const THOUSANDS_SEP: &str = ".";
    }
    pub mod LC_TELEPHONE {
        /// `Some("34")`
        pub const INT_PREFIX: Option<&str> = Some("34");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["dom", "lun", "mar", "mié", "jue", "vie", "sáb"]`
        pub const ABDAY: &[&str] = &["dom", "lun", "mar", "mié", "jue", "vie", "sáb"];
        /// `&["ene", "feb", "mar", "abr", "may", "jun", "jul", "ago", "sep", "oct", "nov", "dic"]`
        pub const ABMON: &[&str] = &["ene", "feb", "mar", "abr", "may", "jun", "jul", "ago", "sep", "oct", "nov", "dic"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"]`
        pub const DAY: &[&str] = &["domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"]`
        pub const MON: &[&str] = &["enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod es_ES_euro {
    pub use super::es_ES::LC_ADDRESS;
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("https://www.gnu.org/software/libc/")`
        pub const ADDRESS: Option<&str> = Some("https://www.gnu.org/software/libc/");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Spanish")`
        pub const LANGUAGE: Option<&str> = Some("Spanish");
        /// `None`
        pub const REVISION: Option<&str> = None;
        /// `Some("Free Software Foundation, Inc.")`
        pub const SOURCE: Option<&str> = Some("Free Software Foundation, Inc.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Spain")`
        pub const TERRITORY: Option<&str> = Some("Spain");
        /// `Some("Spanish locale for Spain with Euro")`
        pub const TITLE: Option<&str> = Some("Spanish locale for Spain with Euro");
    }
    pub use super::es_ES::LC_MESSAGES;
    pub use super::es_ES::LC_MONETARY;
    pub use super::es_ES::LC_NUMERIC;
    pub use super::es_ES::LC_TELEPHONE;
    pub use super::es_ES::LC_TIME;
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod es_GT {
    pub mod LC_ADDRESS {
        /// `Some("GTM")`
        pub const COUNTRY_AB3: Option<&str> = Some("GTM");
        /// `Some("GT")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("GT");
        /// `Some("GCA")`
        pub const COUNTRY_CAR: Option<&str> = Some("GCA");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Guatemala")`
        pub const COUNTRY_NAME: Option<&str> = Some("Guatemala");
        /// `Some(320)`
        pub const COUNTRY_NUM: Option<i64> = Some(320);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("es")`
        pub const LANG_AB: Option<&str> = Some("es");
        /// `Some("spa")`
        pub const LANG_LIB: Option<&str> = Some("spa");
        /// `Some("español")`
        pub const LANG_NAME: Option<&str> = Some("español");
        /// `Some("spa")`
        pub const LANG_TERM: Option<&str> = Some("spa");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Spanish")`
        pub const LANGUAGE: Option<&str> = Some("Spanish");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Guatemala")`
        pub const TERRITORY: Option<&str> = Some("Guatemala");
        /// `Some("Spanish locale for Guatemala")`
        pub const TITLE: Option<&str> = Some("Spanish locale for Guatemala");
    }
    pub use super::es_ES::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"Q"`
        pub const CURRENCY_SYMBOL: &str = "Q";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"GTQ "`
        pub const INT_CURR_SYMBOL: &str = "GTQ ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("502")`
        pub const INT_PREFIX: Option<&str> = Some("502");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["dom", "lun", "mar", "mié", "jue", "vie", "sáb"]`
        pub const ABDAY: &[&str] = &["dom", "lun", "mar", "mié", "jue", "vie", "sáb"];
        /// `&["ene", "feb", "mar", "abr", "may", "jun", "jul", "ago", "sep", "oct", "nov", "dic"]`
        pub const ABMON: &[&str] = &["ene", "feb", "mar", "abr", "may", "jun", "jul", "ago", "sep", "oct", "nov", "dic"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"]`
        pub const DAY: &[&str] = &["domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"]`
        pub const MON: &[&str] = &["enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod es_HN {
    pub mod LC_ADDRESS {
        /// `Some("HND")`
        pub const COUNTRY_AB3: Option<&str> = Some("HND");
        /// `Some("HN")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("HN");
        /// `Some("HN")`
        pub const COUNTRY_CAR: Option<&str> = Some("HN");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Honduras")`
        pub const COUNTRY_NAME: Option<&str> = Some("Honduras");
        /// `Some(340)`
        pub const COUNTRY_NUM: Option<i64> = Some(340);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("es")`
        pub const LANG_AB: Option<&str> = Some("es");
        /// `Some("spa")`
        pub const LANG_LIB: Option<&str> = Some("spa");
        /// `Some("español")`
        pub const LANG_NAME: Option<&str> = Some("español");
        /// `Some("spa")`
        pub const LANG_TERM: Option<&str> = Some("spa");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Spanish")`
        pub const LANGUAGE: Option<&str> = Some("Spanish");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Honduras")`
        pub const TERRITORY: Option<&str> = Some("Honduras");
        /// `Some("Spanish locale for Honduras")`
        pub const TITLE: Option<&str> = Some("Spanish locale for Honduras");
    }
    pub use super::es_ES::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"L"`
        pub const CURRENCY_SYMBOL: &str = "L";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"HNL "`
        pub const INT_CURR_SYMBOL: &str = "HNL ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("504")`
        pub const INT_PREFIX: Option<&str> = Some("504");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["dom", "lun", "mar", "mié", "jue", "vie", "sáb"]`
        pub const ABDAY: &[&str] = &["dom", "lun", "mar", "mié", "jue", "vie", "sáb"];
        /// `&["ene", "feb", "mar", "abr", "may", "jun", "jul", "ago", "sep", "oct", "nov", "dic"]`
        pub const ABMON: &[&str] = &["ene", "feb", "mar", "abr", "may", "jun", "jul", "ago", "sep", "oct", "nov", "dic"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"]`
        pub const DAY: &[&str] = &["domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"]`
        pub const MON: &[&str] = &["enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod es_MX {
    pub mod LC_ADDRESS {
        /// `Some("MEX")`
        pub const COUNTRY_AB3: Option<&str> = Some("MEX");
        /// `Some("MX")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("MX");
        /// `Some("MEX")`
        pub const COUNTRY_CAR: Option<&str> = Some("MEX");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("México")`
        pub const COUNTRY_NAME: Option<&str> = Some("México");
        /// `Some(484)`
        pub const COUNTRY_NUM: Option<i64> = Some(484);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("es")`
        pub const LANG_AB: Option<&str> = Some("es");
        /// `Some("spa")`
        pub const LANG_LIB: Option<&str> = Some("spa");
        /// `Some("español")`
        pub const LANG_NAME: Option<&str> = Some("español");
        /// `Some("spa")`
        pub const LANG_TERM: Option<&str> = Some("spa");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Mexican Spanish")`
        pub const LANGUAGE: Option<&str> = Some("Mexican Spanish");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Mexico")`
        pub const TERRITORY: Option<&str> = Some("Mexico");
        /// `Some("Spanish locale for Mexico")`
        pub const TITLE: Option<&str> = Some("Spanish locale for Mexico");
    }
    pub use super::es_ES::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"$"`
        pub const CURRENCY_SYMBOL: &str = "$";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"MXN "`
        pub const INT_CURR_SYMBOL: &str = "MXN ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const THOUSANDS_SEP: &str = "\u{202f}";
    }
    pub mod LC_TELEPHONE {
        /// `Some("52")`
        pub const INT_PREFIX: Option<&str> = Some("52");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["dom", "lun", "mar", "mié", "jue", "vie", "sáb"]`
        pub const ABDAY: &[&str] = &["dom", "lun", "mar", "mié", "jue", "vie", "sáb"];
        /// `&["ene", "feb", "mar", "abr", "may", "jun", "jul", "ago", "sep", "oct", "nov", "dic"]`
        pub const ABMON: &[&str] = &["ene", "feb", "mar", "abr", "may", "jun", "jul", "ago", "sep", "oct", "nov", "dic"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"]`
        pub const DAY: &[&str] = &["domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"]`
        pub const MON: &[&str] = &["enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod es_NI {
    pub mod LC_ADDRESS {
        /// `Some("NIC")`
        pub const COUNTRY_AB3: Option<&str> = Some("NIC");
        /// `Some("NI")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("NI");
        /// `Some("NIC")`
        pub const COUNTRY_CAR: Option<&str> = Some("NIC");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Nicaragua")`
        pub const COUNTRY_NAME: Option<&str> = Some("Nicaragua");
        /// `Some(558)`
        pub const COUNTRY_NUM: Option<i64> = Some(558);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("es")`
        pub const LANG_AB: Option<&str> = Some("es");
        /// `Some("spa")`
        pub const LANG_LIB: Option<&str> = Some("spa");
        /// `Some("español")`
        pub const LANG_NAME: Option<&str> = Some("español");
        /// `Some("spa")`
        pub const LANG_TERM: Option<&str> = Some("spa");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("https://www.gnu.org/software/libc/")`
        pub const ADDRESS: Option<&str> = Some("https://www.gnu.org/software/libc/");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-08-21")`
        pub const DATE: Option<&str> = Some("2000-08-21");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Spanish")`
        pub const LANGUAGE: Option<&str> = Some("Spanish");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Free Software Foundation, Inc.")`
        pub const SOURCE: Option<&str> = Some("Free Software Foundation, Inc.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Nicaragua")`
        pub const TERRITORY: Option<&str> = Some("Nicaragua");
        /// `Some("Spanish locale for Nicaragua")`
        pub const TITLE: Option<&str> = Some("Spanish locale for Nicaragua");
    }
    pub use super::es_ES::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"C$"`
        pub const CURRENCY_SYMBOL: &str = "C$";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"NIO "`
        pub const INT_CURR_SYMBOL: &str = "NIO ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("505")`
        pub const INT_PREFIX: Option<&str> = Some("505");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["dom", "lun", "mar", "mié", "jue", "vie", "sáb"]`
        pub const ABDAY: &[&str] = &["dom", "lun", "mar", "mié", "jue", "vie", "sáb"];
        /// `&["ene", "feb", "mar", "abr", "may", "jun", "jul", "ago", "sep", "oct", "nov", "dic"]`
        pub const ABMON: &[&str] = &["ene", "feb", "mar", "abr", "may", "jun", "jul", "ago", "sep", "oct", "nov", "dic"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["AM", "PM"]`
        pub const AM_PM: &[&str] = &["AM", "PM"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"]`
        pub const DAY: &[&str] = &["domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"]`
        pub const MON: &[&str] = &["enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `"%I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod es_PA {
    pub mod LC_ADDRESS {
        /// `Some("PAN")`
        pub const COUNTRY_AB3: Option<&str> = Some("PAN");
        /// `Some("PA")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("PA");
        /// `Some("PA")`
        pub const COUNTRY_CAR: Option<&str> = Some("PA");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Panamá")`
        pub const COUNTRY_NAME: Option<&str> = Some("Panamá");
        /// `Some(591)`
        pub const COUNTRY_NUM: Option<i64> = Some(591);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("es")`
        pub const LANG_AB: Option<&str> = Some("es");
        /// `Some("spa")`
        pub const LANG_LIB: Option<&str> = Some("spa");
        /// `Some("español")`
        pub const LANG_NAME: Option<&str> = Some("español");
        /// `Some("spa")`
        pub const LANG_TERM: Option<&str> = Some("spa");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Spanish")`
        pub const LANGUAGE: Option<&str> = Some("Spanish");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Panama")`
        pub const TERRITORY: Option<&str> = Some("Panama");
        /// `Some("Spanish locale for Panama")`
        pub const TITLE: Option<&str> = Some("Spanish locale for Panama");
    }
    pub use super::es_ES::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"B/."`
        pub const CURRENCY_SYMBOL: &str = "B/.";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"PAB "`
        pub const INT_CURR_SYMBOL: &str = "PAB ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("507")`
        pub const INT_PREFIX: Option<&str> = Some("507");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["dom", "lun", "mar", "mié", "jue", "vie", "sáb"]`
        pub const ABDAY: &[&str] = &["dom", "lun", "mar", "mié", "jue", "vie", "sáb"];
        /// `&["ene", "feb", "mar", "abr", "may", "jun", "jul", "ago", "sep", "oct", "nov", "dic"]`
        pub const ABMON: &[&str] = &["ene", "feb", "mar", "abr", "may", "jun", "jul", "ago", "sep", "oct", "nov", "dic"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"]`
        pub const DAY: &[&str] = &["domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"]`
        pub const MON: &[&str] = &["enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod es_PE {
    pub mod LC_ADDRESS {
        /// `Some("PER")`
        pub const COUNTRY_AB3: Option<&str> = Some("PER");
        /// `Some("PE")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("PE");
        /// `Some("PE")`
        pub const COUNTRY_CAR: Option<&str> = Some("PE");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Perú")`
        pub const COUNTRY_NAME: Option<&str> = Some("Perú");
        /// `Some(604)`
        pub const COUNTRY_NUM: Option<i64> = Some(604);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("es")`
        pub const LANG_AB: Option<&str> = Some("es");
        /// `Some("spa")`
        pub const LANG_LIB: Option<&str> = Some("spa");
        /// `Some("español")`
        pub const LANG_NAME: Option<&str> = Some("español");
        /// `Some("spa")`
        pub const LANG_TERM: Option<&str> = Some("spa");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Spanish")`
        pub const LANGUAGE: Option<&str> = Some("Spanish");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Peru")`
        pub const TERRITORY: Option<&str> = Some("Peru");
        /// `Some("Spanish locale for Peru")`
        pub const TITLE: Option<&str> = Some("Spanish locale for Peru");
    }
    pub use super::es_ES::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"S/"`
        pub const CURRENCY_SYMBOL: &str = "S/";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"PEN "`
        pub const INT_CURR_SYMBOL: &str = "PEN ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub use super::es_ES::LC_NUMERIC;
    pub mod LC_TELEPHONE {
        /// `Some("51")`
        pub const INT_PREFIX: Option<&str> = Some("51");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["dom", "lun", "mar", "mié", "jue", "vie", "sáb"]`
        pub const ABDAY: &[&str] = &["dom", "lun", "mar", "mié", "jue", "vie", "sáb"];
        /// `&["ene", "feb", "mar", "abr", "may", "jun", "jul", "ago", "set", "oct", "nov", "dic"]`
        pub const ABMON: &[&str] = &["ene", "feb", "mar", "abr", "may", "jun", "jul", "ago", "set", "oct", "nov", "dic"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["AM", "PM"]`
        pub const AM_PM: &[&str] = &["AM", "PM"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"]`
        pub const DAY: &[&str] = &["domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "setiembre", "octubre", "noviembre", "diciembre"]`
        pub const MON: &[&str] = &["enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "setiembre", "octubre", "noviembre", "diciembre"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `"%I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod es_PR {
    pub mod LC_ADDRESS {
        /// `Some("PRI")`
        pub const COUNTRY_AB3: Option<&str> = Some("PRI");
        /// `Some("PR")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("PR");
        /// `Some("PR")`
        pub const COUNTRY_CAR: Option<&str> = Some("PR");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Puerto Rico")`
        pub const COUNTRY_NAME: Option<&str> = Some("Puerto Rico");
        /// `Some(630)`
        pub const COUNTRY_NUM: Option<i64> = Some(630);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("es")`
        pub const LANG_AB: Option<&str> = Some("es");
        /// `Some("spa")`
        pub const LANG_LIB: Option<&str> = Some("spa");
        /// `Some("español")`
        pub const LANG_NAME: Option<&str> = Some("español");
        /// `Some("spa")`
        pub const LANG_TERM: Option<&str> = Some("spa");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("https://www.gnu.org/software/libc/")`
        pub const ADDRESS: Option<&str> = Some("https://www.gnu.org/software/libc/");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-08-21")`
        pub const DATE: Option<&str> = Some("2000-08-21");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Spanish")`
        pub const LANGUAGE: Option<&str> = Some("Spanish");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Free Software Foundation, Inc.")`
        pub const SOURCE: Option<&str> = Some("Free Software Foundation, Inc.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Puerto Rico")`
        pub const TERRITORY: Option<&str> = Some("Puerto Rico");
        /// `Some("Spanish locale for Puerto Rico")`
        pub const TITLE: Option<&str> = Some("Spanish locale for Puerto Rico");
    }
    pub use super::es_ES::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"$"`
        pub const CURRENCY_SYMBOL: &str = "$";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"USD "`
        pub const INT_CURR_SYMBOL: &str = "USD ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("1")`
        pub const INT_PREFIX: Option<&str> = Some("1");
        /// `Some("011")`
        pub const INT_SELECT: Option<&str> = Some("011");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["dom", "lun", "mar", "mié", "jue", "vie", "sáb"]`
        pub const ABDAY: &[&str] = &["dom", "lun", "mar", "mié", "jue", "vie", "sáb"];
        /// `&["ene", "feb", "mar", "abr", "may", "jun", "jul", "ago", "sep", "oct", "nov", "dic"]`
        pub const ABMON: &[&str] = &["ene", "feb", "mar", "abr", "may", "jun", "jul", "ago", "sep", "oct", "nov", "dic"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"]`
        pub const DAY: &[&str] = &["domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"]`
        pub const MON: &[&str] = &["enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod es_PY {
    pub mod LC_ADDRESS {
        /// `Some("PRY")`
        pub const COUNTRY_AB3: Option<&str> = Some("PRY");
        /// `Some("PY")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("PY");
        /// `Some("PY")`
        pub const COUNTRY_CAR: Option<&str> = Some("PY");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Paraguay")`
        pub const COUNTRY_NAME: Option<&str> = Some("Paraguay");
        /// `Some(600)`
        pub const COUNTRY_NUM: Option<i64> = Some(600);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("es")`
        pub const LANG_AB: Option<&str> = Some("es");
        /// `Some("spa")`
        pub const LANG_LIB: Option<&str> = Some("spa");
        /// `Some("español")`
        pub const LANG_NAME: Option<&str> = Some("español");
        /// `Some("spa")`
        pub const LANG_TERM: Option<&str> = Some("spa");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Spanish")`
        pub const LANGUAGE: Option<&str> = Some("Spanish");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Paraguay")`
        pub const TERRITORY: Option<&str> = Some("Paraguay");
        /// `Some("Spanish locale for Paraguay")`
        pub const TITLE: Option<&str> = Some("Spanish locale for Paraguay");
    }
    pub use super::es_ES::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"Gs."`
        pub const CURRENCY_SYMBOL: &str = "Gs.";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"PYG "`
        pub const INT_CURR_SYMBOL: &str = "PYG ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const MON_THOUSANDS_SEP: &str = ".";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub use super::es_ES::LC_NUMERIC;
    pub mod LC_TELEPHONE {
        /// `Some("595")`
        pub const INT_PREFIX: Option<&str> = Some("595");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["dom", "lun", "mar", "mié", "jue", "vie", "sáb"]`
        pub const ABDAY: &[&str] = &["dom", "lun", "mar", "mié", "jue", "vie", "sáb"];
        /// `&["ene", "feb", "mar", "abr", "may", "jun", "jul", "ago", "sep", "oct", "nov", "dic"]`
        pub const ABMON: &[&str] = &["ene", "feb", "mar", "abr", "may", "jun", "jul", "ago", "sep", "oct", "nov", "dic"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"]`
        pub const DAY: &[&str] = &["domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"]`
        pub const MON: &[&str] = &["enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod es_SV {
    pub mod LC_ADDRESS {
        /// `Some("SLV")`
        pub const COUNTRY_AB3: Option<&str> = Some("SLV");
        /// `Some("SV")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("SV");
        /// `Some("ES")`
        pub const COUNTRY_CAR: Option<&str> = Some("ES");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("El Salvador")`
        pub const COUNTRY_NAME: Option<&str> = Some("El Salvador");
        /// `Some(222)`
        pub const COUNTRY_NUM: Option<i64> = Some(222);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("es")`
        pub const LANG_AB: Option<&str> = Some("es");
        /// `Some("spa")`
        pub const LANG_LIB: Option<&str> = Some("spa");
        /// `Some("español")`
        pub const LANG_NAME: Option<&str> = Some("español");
        /// `Some("spa")`
        pub const LANG_TERM: Option<&str> = Some("spa");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Spanish")`
        pub const LANGUAGE: Option<&str> = Some("Spanish");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("El Salvador")`
        pub const TERRITORY: Option<&str> = Some("El Salvador");
        /// `Some("Spanish locale for El Salvador")`
        pub const TITLE: Option<&str> = Some("Spanish locale for El Salvador");
    }
    pub use super::es_ES::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"$"`
        pub const CURRENCY_SYMBOL: &str = "$";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"USD "`
        pub const INT_CURR_SYMBOL: &str = "USD ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("503")`
        pub const INT_PREFIX: Option<&str> = Some("503");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["dom", "lun", "mar", "mié", "jue", "vie", "sáb"]`
        pub const ABDAY: &[&str] = &["dom", "lun", "mar", "mié", "jue", "vie", "sáb"];
        /// `&["ene", "feb", "mar", "abr", "may", "jun", "jul", "ago", "sep", "oct", "nov", "dic"]`
        pub const ABMON: &[&str] = &["ene", "feb", "mar", "abr", "may", "jun", "jul", "ago", "sep", "oct", "nov", "dic"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"]`
        pub const DAY: &[&str] = &["domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"]`
        pub const MON: &[&str] = &["enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod es_US {
    pub mod LC_ADDRESS {
        /// `Some("USA")`
        pub const COUNTRY_AB3: Option<&str> = Some("USA");
        /// `Some("US")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("US");
        /// `Some("USA")`
        pub const COUNTRY_CAR: Option<&str> = Some("USA");
        /// `Some("0")`
        pub const COUNTRY_ISBN: Option<&str> = Some("0");
        /// `Some("Estados Unidos")`
        pub const COUNTRY_NAME: Option<&str> = Some("Estados Unidos");
        /// `Some(840)`
        pub const COUNTRY_NUM: Option<i64> = Some(840);
        /// `Some("USA")`
        pub const COUNTRY_POST: Option<&str> = Some("USA");
        /// `Some("es")`
        pub const LANG_AB: Option<&str> = Some("es");
        /// `Some("spa")`
        pub const LANG_LIB: Option<&str> = Some("spa");
        /// `Some("español")`
        pub const LANG_NAME: Option<&str> = Some("español");
        /// `Some("spa")`
        pub const LANG_TERM: Option<&str> = Some("spa");
        /// `Some("%a%N%f%N%d%N%b%N%h %s %e %r%N%T, %S %z%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%a%N%f%N%d%N%b%N%h %s %e %r%N%T, %S %z%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Spanish")`
        pub const LANGUAGE: Option<&str> = Some("Spanish");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("United States")`
        pub const TERRITORY: Option<&str> = Some("United States");
        /// `Some("Spanish locale for the USA")`
        pub const TITLE: Option<&str> = Some("Spanish locale for the USA");
    }
    pub use super::es_ES::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"$"`
        pub const CURRENCY_SYMBOL: &str = "$";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"USD "`
        pub const INT_CURR_SYMBOL: &str = "USD ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub use super::en_US::LC_NUMERIC;
    pub mod LC_TELEPHONE {
        /// `Some("1")`
        pub const INT_PREFIX: Option<&str> = Some("1");
        /// `Some("11")`
        pub const INT_SELECT: Option<&str> = Some("11");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["dom", "lun", "mar", "mié", "jue", "vie", "sáb"]`
        pub const ABDAY: &[&str] = &["dom", "lun", "mar", "mié", "jue", "vie", "sáb"];
        /// `&["ene", "feb", "mar", "abr", "may", "jun", "jul", "ago", "sep", "oct", "nov", "dic"]`
        pub const ABMON: &[&str] = &["ene", "feb", "mar", "abr", "may", "jun", "jul", "ago", "sep", "oct", "nov", "dic"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"]`
        pub const DAY: &[&str] = &["domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"]`
        pub const MON: &[&str] = &["enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod es_UY {
    pub mod LC_ADDRESS {
        /// `Some("URY")`
        pub const COUNTRY_AB3: Option<&str> = Some("URY");
        /// `Some("UY")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("UY");
        /// `Some("UY")`
        pub const COUNTRY_CAR: Option<&str> = Some("UY");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Uruguay")`
        pub const COUNTRY_NAME: Option<&str> = Some("Uruguay");
        /// `Some(858)`
        pub const COUNTRY_NUM: Option<i64> = Some(858);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("es")`
        pub const LANG_AB: Option<&str> = Some("es");
        /// `Some("spa")`
        pub const LANG_LIB: Option<&str> = Some("spa");
        /// `Some("español")`
        pub const LANG_NAME: Option<&str> = Some("español");
        /// `Some("spa")`
        pub const LANG_TERM: Option<&str> = Some("spa");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Spanish")`
        pub const LANGUAGE: Option<&str> = Some("Spanish");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Uruguay")`
        pub const TERRITORY: Option<&str> = Some("Uruguay");
        /// `Some("Spanish locale for Uruguay")`
        pub const TITLE: Option<&str> = Some("Spanish locale for Uruguay");
    }
    pub use super::es_ES::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"$"`
        pub const CURRENCY_SYMBOL: &str = "$";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"UYU "`
        pub const INT_CURR_SYMBOL: &str = "UYU ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const MON_THOUSANDS_SEP: &str = ".";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub use super::es_ES::LC_NUMERIC;
    pub mod LC_TELEPHONE {
        /// `Some("598")`
        pub const INT_PREFIX: Option<&str> = Some("598");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["dom", "lun", "mar", "mié", "jue", "vie", "sáb"]`
        pub const ABDAY: &[&str] = &["dom", "lun", "mar", "mié", "jue", "vie", "sáb"];
        /// `&["ene", "feb", "mar", "abr", "may", "jun", "jul", "ago", "set", "oct", "nov", "dic"]`
        pub const ABMON: &[&str] = &["ene", "feb", "mar", "abr", "may", "jun", "jul", "ago", "set", "oct", "nov", "dic"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"]`
        pub const DAY: &[&str] = &["domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "setiembre", "octubre", "noviembre", "diciembre"]`
        pub const MON: &[&str] = &["enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "setiembre", "octubre", "noviembre", "diciembre"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod es_VE {
    pub mod LC_ADDRESS {
        /// `Some("VEN")`
        pub const COUNTRY_AB3: Option<&str> = Some("VEN");
        /// `Some("VE")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("VE");
        /// `Some("YV")`
        pub const COUNTRY_CAR: Option<&str> = Some("YV");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Venezuela")`
        pub const COUNTRY_NAME: Option<&str> = Some("Venezuela");
        /// `Some(862)`
        pub const COUNTRY_NUM: Option<i64> = Some(862);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("es")`
        pub const LANG_AB: Option<&str> = Some("es");
        /// `Some("spa")`
        pub const LANG_LIB: Option<&str> = Some("spa");
        /// `Some("español")`
        pub const LANG_NAME: Option<&str> = Some("español");
        /// `Some("spa")`
        pub const LANG_TERM: Option<&str> = Some("spa");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Spanish")`
        pub const LANGUAGE: Option<&str> = Some("Spanish");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Venezuela")`
        pub const TERRITORY: Option<&str> = Some("Venezuela");
        /// `Some("Spanish locale for Venezuela")`
        pub const TITLE: Option<&str> = Some("Spanish locale for Venezuela");
    }
    pub use super::es_ES::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"Bs."`
        pub const CURRENCY_SYMBOL: &str = "Bs.";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"VEF "`
        pub const INT_CURR_SYMBOL: &str = "VEF ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const MON_THOUSANDS_SEP: &str = ".";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub use super::es_ES::LC_NUMERIC;
    pub mod LC_TELEPHONE {
        /// `Some("58")`
        pub const INT_PREFIX: Option<&str> = Some("58");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["dom", "lun", "mar", "mié", "jue", "vie", "sáb"]`
        pub const ABDAY: &[&str] = &["dom", "lun", "mar", "mié", "jue", "vie", "sáb"];
        /// `&["ene", "feb", "mar", "abr", "may", "jun", "jul", "ago", "sep", "oct", "nov", "dic"]`
        pub const ABMON: &[&str] = &["ene", "feb", "mar", "abr", "may", "jun", "jul", "ago", "sep", "oct", "nov", "dic"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["am", "pm"]`
        pub const AM_PM: &[&str] = &["am", "pm"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"]`
        pub const DAY: &[&str] = &["domingo", "lunes", "martes", "miércoles", "jueves", "viernes", "sábado"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"]`
        pub const MON: &[&str] = &["enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `"%I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod et_EE {
    pub mod LC_ADDRESS {
        /// `Some("EST")`
        pub const COUNTRY_AB3: Option<&str> = Some("EST");
        /// `Some("EE")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("EE");
        /// `Some("EST")`
        pub const COUNTRY_CAR: Option<&str> = Some("EST");
        /// `Some("978-9949,978-9985")`
        pub const COUNTRY_ISBN: Option<&str> = Some("978-9949,978-9985");
        /// `Some("Eesti")`
        pub const COUNTRY_NAME: Option<&str> = Some("Eesti");
        /// `Some(233)`
        pub const COUNTRY_NUM: Option<i64> = Some(233);
        /// `Some("EE")`
        pub const COUNTRY_POST: Option<&str> = Some("EE");
        /// `Some("et")`
        pub const LANG_AB: Option<&str> = Some("et");
        /// `Some("est")`
        pub const LANG_LIB: Option<&str> = Some("est");
        /// `Some("eesti")`
        pub const LANG_NAME: Option<&str> = Some("eesti");
        /// `Some("est")`
        pub const LANG_TERM: Option<&str> = Some("est");
        /// `Some("%a%N%f%N%d%N%b%N%s%t%h%t%e%t%r%N%C-%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%a%N%f%N%d%N%b%N%s%t%h%t%e%t%r%N%C-%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("To-nisma:gi 8, Tallinn, EE0100 Estonia")`
        pub const ADDRESS: Option<&str> = Some("To-nisma:gi 8, Tallinn, EE0100 Estonia");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Estonian")`
        pub const LANGUAGE: Option<&str> = Some("Estonian");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Estonian Informatics Fund")`
        pub const SOURCE: Option<&str> = Some("Estonian Informatics Fund");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Estonia")`
        pub const TERRITORY: Option<&str> = Some("Estonia");
        /// `Some("Estonian locale for Estonia")`
        pub const TITLE: Option<&str> = Some("Estonian locale for Estonia");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0EeNn]"`
        pub const NOEXPR: &str = "^[-0EeNn]";
        /// `Some("ei")`
        pub const NOSTR: Option<&str> = Some("ei");
        /// `"^[+1JjYy]"`
        pub const YESEXPR: &str = "^[+1JjYy]";
        /// `Some("jah")`
        pub const YESSTR: Option<&str> = Some("jah");
    }
    pub mod LC_MONETARY {
        /// `"€"`
        pub const CURRENCY_SYMBOL: &str = "€";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"EUR "`
        pub const INT_CURR_SYMBOL: &str = "EUR ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const MON_THOUSANDS_SEP: &str = "\u{202f}";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const THOUSANDS_SEP: &str = "\u{202f}";
    }
    pub mod LC_TELEPHONE {
        /// `Some("372")`
        pub const INT_PREFIX: Option<&str> = Some("372");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["P", "E", "T", "K", "N", "R", "L"]`
        pub const ABDAY: &[&str] = &["P", "E", "T", "K", "N", "R", "L"];
        /// `&["jaan ", "veebr", "märts", "apr  ", "mai  ", "juuni", "juuli", "aug  ", "sept ", "okt  ", "nov  ", "dets "]`
        pub const ABMON: &[&str] = &["jaan ", "veebr", "märts", "apr  ", "mai  ", "juuni", "juuli", "aug  ", "sept ", "okt  ", "nov  ", "dets "];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["pühapäev", "esmaspäev", "teisipäev", "kolmapäev", "neljapäev", "reede", "laupäev"]`
        pub const DAY: &[&str] = &["pühapäev", "esmaspäev", "teisipäev", "kolmapäev", "neljapäev", "reede", "laupäev"];
        /// `"%d.%m.%Y"`
        pub const D_FMT: &str = "%d.%m.%Y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["jaanuar", "veebruar", "märts", "aprill", "mai", "juuni", "juuli", "august", "september", "oktoober", "november", "detsember"]`
        pub const MON: &[&str] = &["jaanuar", "veebruar", "märts", "aprill", "mai", "juuni", "juuli", "august", "september", "oktoober", "november", "detsember"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod eu_ES {
    pub mod LC_ADDRESS {
        /// `Some("ESP")`
        pub const COUNTRY_AB3: Option<&str> = Some("ESP");
        /// `Some("ES")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("ES");
        /// `Some("E")`
        pub const COUNTRY_CAR: Option<&str> = Some("E");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Espainia")`
        pub const COUNTRY_NAME: Option<&str> = Some("Espainia");
        /// `Some(724)`
        pub const COUNTRY_NUM: Option<i64> = Some(724);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("eu")`
        pub const LANG_AB: Option<&str> = Some("eu");
        /// `Some("baq")`
        pub const LANG_LIB: Option<&str> = Some("baq");
        /// `Some("euskara")`
        pub const LANG_NAME: Option<&str> = Some("euskara");
        /// `Some("eus")`
        pub const LANG_TERM: Option<&str> = Some("eus");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Basque")`
        pub const LANGUAGE: Option<&str> = Some("Basque");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Spain")`
        pub const TERRITORY: Option<&str> = Some("Spain");
        /// `Some("Basque locale for Spain")`
        pub const TITLE: Option<&str> = Some("Basque locale for Spain");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0eEnN]"`
        pub const NOEXPR: &str = "^[-0eEnN]";
        /// `Some("ez")`
        pub const NOSTR: Option<&str> = Some("ez");
        /// `"^[+1bBsSyY]"`
        pub const YESEXPR: &str = "^[+1bBsSyY]";
        /// `Some("bai")`
        pub const YESSTR: Option<&str> = Some("bai");
    }
    pub mod LC_MONETARY {
        /// `"€"`
        pub const CURRENCY_SYMBOL: &str = "€";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"EUR "`
        pub const INT_CURR_SYMBOL: &str = "EUR ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const MON_THOUSANDS_SEP: &str = ".";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const THOUSANDS_SEP: &str = ".";
    }
    pub mod LC_TELEPHONE {
        /// `Some("34")`
        pub const INT_PREFIX: Option<&str> = Some("34");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["ig.", "al.", "ar.", "az.", "og.", "or.", "lr."]`
        pub const ABDAY: &[&str] = &["ig.", "al.", "ar.", "az.", "og.", "or.", "lr."];
        /// `&["urt", "ots", "mar", "api", "mai", "eka", "uzt", "abu", "ira", "urr", "aza", "abe"]`
        pub const ABMON: &[&str] = &["urt", "ots", "mar", "api", "mai", "eka", "uzt", "abu", "ira", "urr", "aza", "abe"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["igandea", "astelehena", "asteartea", "asteazkena", "osteguna", "ostirala", "larunbata"]`
        pub const DAY: &[&str] = &["igandea", "astelehena", "asteartea", "asteazkena", "osteguna", "ostirala", "larunbata"];
        /// `"%a, %Y.eko %bren %da"`
        pub const D_FMT: &str = "%a, %Y.eko %bren %da";
        /// `"%y-%m-%d %T %Z"`
        pub const D_T_FMT: &str = "%y-%m-%d %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["urtarrila", "otsaila", "martxoa", "apirila", "maiatza", "ekaina", "uztaila", "abuztua", "iraila", "urria", "azaroa", "abendua"]`
        pub const MON: &[&str] = &["urtarrila", "otsaila", "martxoa", "apirila", "maiatza", "ekaina", "uztaila", "abuztua", "iraila", "urria", "azaroa", "abendua"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod eu_ES_euro {
    pub use super::eu_ES::LC_ADDRESS;
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("https://www.gnu.org/software/libc/")`
        pub const ADDRESS: Option<&str> = Some("https://www.gnu.org/software/libc/");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-08-21")`
        pub const DATE: Option<&str> = Some("2000-08-21");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Basque")`
        pub const LANGUAGE: Option<&str> = Some("Basque");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Free Software Foundation, Inc.")`
        pub const SOURCE: Option<&str> = Some("Free Software Foundation, Inc.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Spain")`
        pub const TERRITORY: Option<&str> = Some("Spain");
        /// `Some("Basque language locale for Spain with Euro")`
        pub const TITLE: Option<&str> = Some("Basque language locale for Spain with Euro");
    }
    pub use super::eu_ES::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"€"`
        pub const CURRENCY_SYMBOL: &str = "€";
        /// `0`
        pub const FRAC_DIGITS: i64 = 0;
        /// `"EUR "`
        pub const INT_CURR_SYMBOL: &str = "EUR ";
        /// `0`
        pub const INT_FRAC_DIGITS: i64 = 0;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const MON_THOUSANDS_SEP: &str = ".";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub use super::eu_ES::LC_NUMERIC;
    pub use super::eu_ES::LC_TELEPHONE;
    pub use super::eu_ES::LC_TIME;
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod fa_IR {
    pub mod LC_ADDRESS {
        /// `Some("IRN")`
        pub const COUNTRY_AB3: Option<&str> = Some("IRN");
        /// `Some("IR")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IR");
        /// `Some("IR")`
        pub const COUNTRY_CAR: Option<&str> = Some("IR");
        /// `Some("964")`
        pub const COUNTRY_ISBN: Option<&str> = Some("964");
        /// `Some("ایران")`
        pub const COUNTRY_NAME: Option<&str> = Some("ایران");
        /// `Some(364)`
        pub const COUNTRY_NUM: Option<i64> = Some(364);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("fa")`
        pub const LANG_AB: Option<&str> = Some("fa");
        /// `Some("per")`
        pub const LANG_LIB: Option<&str> = Some("per");
        /// `Some("فارسی")`
        pub const LANG_NAME: Option<&str> = Some("فارسی");
        /// `Some("fas")`
        pub const LANG_TERM: Option<&str> = Some("fas");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("5, Shahid Ghasemi Habibollah, Azadi Ave, Tehran, Iran")`
        pub const ADDRESS: Option<&str> = Some("5, Shahid Ghasemi Habibollah, Azadi Ave, Tehran, Iran");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Roozbeh Pournader")`
        pub const CONTACT: Option<&str> = Some("Roozbeh Pournader");
        /// `Some("2005-04-06")`
        pub const DATE: Option<&str> = Some("2005-04-06");
        /// `Some("roozbeh@farsiweb.info")`
        pub const EMAIL: Option<&str> = Some("roozbeh@farsiweb.info");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Persian")`
        pub const LANGUAGE: Option<&str> = Some("Persian");
        /// `Some("3.0")`
        pub const REVISION: Option<&str> = Some("3.0");
        /// `Some("Sharif FarsiWeb, Inc.")`
        pub const SOURCE: Option<&str> = Some("Sharif FarsiWeb, Inc.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Iran")`
        pub const TERRITORY: Option<&str> = Some("Iran");
        /// `Some("Persian locale for Iran")`
        pub const TITLE: Option<&str> = Some("Persian locale for Iran");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNخنok]"`
        pub const NOEXPR: &str = "^[-0nNخنok]";
        /// `Some("نه")`
        pub const NOSTR: Option<&str> = Some("نه");
        /// `"^[+1yYآبHf]"`
        pub const YESEXPR: &str = "^[+1yYآبHf]";
        /// `Some("بله")`
        pub const YESSTR: Option<&str> = Some("بله");
    }
    pub mod LC_MONETARY {
        /// `"ریال"`
        pub const CURRENCY_SYMBOL: &str = "ریال";
        /// `0`
        pub const FRAC_DIGITS: i64 = 0;
        /// `"IRR "`
        pub const INT_CURR_SYMBOL: &str = "IRR ";
        /// `0`
        pub const INT_FRAC_DIGITS: i64 = 0;
        /// `Some(0)`
        pub const INT_N_CS_PRECEDES: Option<i64> = Some(0);
        /// `Some(1)`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = Some(1);
        /// `Some(1)`
        pub const INT_N_SIGN_POSN: Option<i64> = Some(1);
        /// `Some(0)`
        pub const INT_P_CS_PRECEDES: Option<i64> = Some(0);
        /// `Some(1)`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = Some(1);
        /// `Some(1)`
        pub const INT_P_SIGN_POSN: Option<i64> = Some(1);
        /// `"٫"`
        pub const MON_DECIMAL_POINT: &str = "٫";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `"٬"`
        pub const MON_THOUSANDS_SEP: &str = "٬";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("98")`
        pub const INT_PREFIX: Option<&str> = Some("98");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `Some("\u{202a}%A‒%l\u{202c}")`
        pub const TEL_DOM_FMT: Option<&str> = Some("\u{202a}%A‒%l\u{202c}");
        /// `Some("\u{202a}+%c %a %l\u{202c}")`
        pub const TEL_INT_FMT: Option<&str> = Some("\u{202a}+%c %a %l\u{202c}");
    }
    pub mod LC_TIME {
        /// `&["یکشنبه", "دوشنبه", "سه\u{200c}شنبه", "چهارشنبه", "پنجشنبه", "جمعه", "شنبه"]`
        pub const ABDAY: &[&str] = &["یکشنبه", "دوشنبه", "سه\u{200c}شنبه", "چهارشنبه", "پنجشنبه", "جمعه", "شنبه"];
        /// `&["ژانویه", "فوریه", "مارس", "آوریل", "مه", "ژوئن", "ژوئیه", "اوت", "سپتامبر", "اكتبر", "نوامبر", "دسامبر"]`
        pub const ABMON: &[&str] = &["ژانویه", "فوریه", "مارس", "آوریل", "مه", "ژوئن", "ژوئیه", "اوت", "سپتامبر", "اكتبر", "نوامبر", "دسامبر"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `Some(&["۰۰", "۰۱", "۰۲", "۰۳", "۰۴", "۰۵", "۰۶", "۰۷", "۰۸", "۰۹", "۱۰", "۱۱", "۱۲", "۱۳", "۱۴", "۱۵", "۱۶", "۱۷", "۱۸", "۱۹", "۲۰", "۲۱", "۲۲", "۲۳", "۲۴", "۲۵", "۲۶", "۲۷", "۲۸", "۲۹", "۳۰", "۳۱", "۳۲", "۳۳", "۳۴", "۳۵", "۳۶", "۳۷", "۳۸", "۳۹", "۴۰", "۴۱", "۴۲", "۴۳", "۴۴", "۴۵", "۴۶", "۴۷", "۴۸", "۴۹", "۵۰", "۵۱", "۵۲", "۵۳", "۵۴", "۵۵", "۵۶", "۵۷", "۵۸", "۵۹", "۶۰", "۶۱", "۶۲", "۶۳", "۶۴", "۶۵", "۶۶", "۶۷", "۶۸", "۶۹", "۷۰", "۷۱", "۷۲", "۷۳", "۷۴", "۷۵", "۷۶", "۷۷", "۷۸", "۷۹", "۸۰", "۸۱", "۸۲", "۸۳", "۸۴", "۸۵", "۸۶", "۸۷", "۸۸", "۸۹", "۹۰", "۹۱", "۹۲", "۹۳", "۹۴", "۹۵", "۹۶", "۹۷", "۹۸", "۹۹"])`
        pub const ALT_DIGITS: Option<&[&str]> = Some(&["۰۰", "۰۱", "۰۲", "۰۳", "۰۴", "۰۵", "۰۶", "۰۷", "۰۸", "۰۹", "۱۰", "۱۱", "۱۲", "۱۳", "۱۴", "۱۵", "۱۶", "۱۷", "۱۸", "۱۹", "۲۰", "۲۱", "۲۲", "۲۳", "۲۴", "۲۵", "۲۶", "۲۷", "۲۸", "۲۹", "۳۰", "۳۱", "۳۲", "۳۳", "۳۴", "۳۵", "۳۶", "۳۷", "۳۸", "۳۹", "۴۰", "۴۱", "۴۲", "۴۳", "۴۴", "۴۵", "۴۶", "۴۷", "۴۸", "۴۹", "۵۰", "۵۱", "۵۲", "۵۳", "۵۴", "۵۵", "۵۶", "۵۷", "۵۸", "۵۹", "۶۰", "۶۱", "۶۲", "۶۳", "۶۴", "۶۵", "۶۶", "۶۷", "۶۸", "۶۹", "۷۰", "۷۱", "۷۲", "۷۳", "۷۴", "۷۵", "۷۶", "۷۷", "۷۸", "۷۹", "۸۰", "۸۱", "۸۲", "۸۳", "۸۴", "۸۵", "۸۶", "۸۷", "۸۸", "۸۹", "۹۰", "۹۱", "۹۲", "۹۳", "۹۴", "۹۵", "۹۶", "۹۷", "۹۸", "۹۹"]);
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `Some(3)`
        pub const CAL_DIRECTION: Option<i64> = Some(3);
        /// `Some("\u{202b}%A %Oe %B %Oy، ساعت %OH:%OM:%OS (%Z)\u{202c}")`
        pub const DATE_FMT: Option<&str> = Some("\u{202b}%A %Oe %B %Oy، ساعت %OH:%OM:%OS (%Z)\u{202c}");
        /// `&["یکشنبه", "دوشنبه", "سه\u{200c}شنبه", "چهارشنبه", "پنجشنبه", "جمعه", "شنبه"]`
        pub const DAY: &[&str] = &["یکشنبه", "دوشنبه", "سه\u{200c}شنبه", "چهارشنبه", "پنجشنبه", "جمعه", "شنبه"];
        /// `"%Oy/%Om/%Od"`
        pub const D_FMT: &str = "%Oy/%Om/%Od";
        /// `"\u{202b}%A %Oe %B %Oy، %OH:%OM:%OS\u{202c}"`
        pub const D_T_FMT: &str = "\u{202b}%A %Oe %B %Oy، %OH:%OM:%OS\u{202c}";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(7)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(7);
        /// `Some(7)`
        pub const FIRST_WORKDAY: Option<i64> = Some(7);
        /// `&["ژانویه", "فوریه", "مارس", "آوریل", "مه", "ژوئن", "ژوئیه", "اوت", "سپتامبر", "اكتبر", "نوامبر", "دسامبر"]`
        pub const MON: &[&str] = &["ژانویه", "فوریه", "مارس", "آوریل", "مه", "ژوئن", "ژوئیه", "اوت", "سپتامبر", "اكتبر", "نوامبر", "دسامبر"];
        /// `"%OH:%OM:%OS"`
        pub const T_FMT: &str = "%OH:%OM:%OS";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ff_SN {
    pub mod LC_ADDRESS {
        /// `Some("SEN")`
        pub const COUNTRY_AB3: Option<&str> = Some("SEN");
        /// `Some("SN")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("SN");
        /// `Some("SN")`
        pub const COUNTRY_CAR: Option<&str> = Some("SN");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Senegaal")`
        pub const COUNTRY_NAME: Option<&str> = Some("Senegaal");
        /// `Some(686)`
        pub const COUNTRY_NUM: Option<i64> = Some(686);
        /// `Some("SN")`
        pub const COUNTRY_POST: Option<&str> = Some("SN");
        /// `Some("ff")`
        pub const LANG_AB: Option<&str> = Some("ff");
        /// `Some("ful")`
        pub const LANG_LIB: Option<&str> = Some("ful");
        /// `Some("Pulaar")`
        pub const LANG_NAME: Option<&str> = Some("Pulaar");
        /// `Some("ful")`
        pub const LANG_TERM: Option<&str> = Some("ful");
        /// `Some("%a%N%f%N%d%N%b%N%h %s %e %r%N%T, %S %z%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%a%N%f%N%d%N%b%N%h %s %e %r%N%T, %S %z%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `None`
        pub const ADDRESS: Option<&str> = None;
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
        ]);
        /// `Some("Ibrahima Sarr")`
        pub const CONTACT: Option<&str> = Some("Ibrahima Sarr");
        /// `Some("2011-03-28")`
        pub const DATE: Option<&str> = Some("2011-03-28");
        /// `Some("admin@pulaar-fulfulde.org")`
        pub const EMAIL: Option<&str> = Some("admin@pulaar-fulfulde.org");
        /// `None`
        pub const FAX: Option<&str> = None;
        /// `Some("Fulah")`
        pub const LANGUAGE: Option<&str> = Some("Fulah");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Pular-Fulfulde.org")`
        pub const SOURCE: Option<&str> = Some("Pular-Fulfulde.org");
        /// `None`
        pub const TEL: Option<&str> = None;
        /// `Some("Senegal")`
        pub const TERRITORY: Option<&str> = Some("Senegal");
        /// `Some("Fulah locale for Senegal")`
        pub const TITLE: Option<&str> = Some("Fulah locale for Senegal");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNaA]"`
        pub const NOEXPR: &str = "^[-0nNaA]";
        /// `Some("Alaa")`
        pub const NOSTR: Option<&str> = Some("Alaa");
        /// `"^[+1yYeE]"`
        pub const YESEXPR: &str = "^[+1yYeE]";
        /// `Some("Eey")`
        pub const YESSTR: Option<&str> = Some("Eey");
    }
    pub mod LC_MONETARY {
        /// `"CFA"`
        pub const CURRENCY_SYMBOL: &str = "CFA";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"XOF "`
        pub const INT_CURR_SYMBOL: &str = "XOF ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `"."`
        pub const THOUSANDS_SEP: &str = ".";
    }
    pub mod LC_TELEPHONE {
        /// `Some("221")`
        pub const INT_PREFIX: Option<&str> = Some("221");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `Some("%a %l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("%a %l");
        /// `Some("%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["dew", "aaɓ", "maw", "nje", "naa", "mwd", "hbi"]`
        pub const ABDAY: &[&str] = &["dew", "aaɓ", "maw", "nje", "naa", "mwd", "hbi"];
        /// `&["sii", "col", "mbo", "see", "duu", "kor", "mor", "juk", "slt", "yar", "jol", "bow"]`
        pub const ABMON: &[&str] = &["sii", "col", "mbo", "see", "duu", "kor", "mor", "juk", "slt", "yar", "jol", "bow"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["subaka", "kikiiɗe"]`
        pub const AM_PM: &[&str] = &["subaka", "kikiiɗe"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["dewo", "aaɓnde", "mawbaare", "njeslaare", "naasaande", "mawnde", "hoore-biir"]`
        pub const DAY: &[&str] = &["dewo", "aaɓnde", "mawbaare", "njeslaare", "naasaande", "mawnde", "hoore-biir"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%a %d %b %Y %R %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %R %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["siilo", "colte", "mbooy", "seeɗto", "duujal", "korse", "morso", "juko", "siilto", "yarkomaa", "jolal", "bowte"]`
        pub const MON: &[&str] = &["siilo", "colte", "mbooy", "seeɗto", "duujal", "korse", "morso", "juko", "siilto", "yarkomaa", "jolal", "bowte"];
        /// `"%R"`
        pub const T_FMT: &str = "%R";
        /// `"%l:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%l:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod fi_FI {
    pub mod LC_ADDRESS {
        /// `Some("FIN")`
        pub const COUNTRY_AB3: Option<&str> = Some("FIN");
        /// `Some("FI")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("FI");
        /// `Some("FIN")`
        pub const COUNTRY_CAR: Option<&str> = Some("FIN");
        /// `Some("952")`
        pub const COUNTRY_ISBN: Option<&str> = Some("952");
        /// `Some("Suomi")`
        pub const COUNTRY_NAME: Option<&str> = Some("Suomi");
        /// `Some(246)`
        pub const COUNTRY_NUM: Option<i64> = Some(246);
        /// `Some("FI")`
        pub const COUNTRY_POST: Option<&str> = Some("FI");
        /// `Some("fi")`
        pub const LANG_AB: Option<&str> = Some("fi");
        /// `Some("fin")`
        pub const LANG_LIB: Option<&str> = Some("fin");
        /// `Some("suomi")`
        pub const LANG_NAME: Option<&str> = Some("suomi");
        /// `Some("fin")`
        pub const LANG_TERM: Option<&str> = Some("fin");
        /// `Some("%f%N%d%N%b%N%a%N%s %h%t%e%t%r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%d%N%b%N%a%N%s %h%t%e%t%r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Finnish")`
        pub const LANGUAGE: Option<&str> = Some("Finnish");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Finland")`
        pub const TERRITORY: Option<&str> = Some("Finland");
        /// `Some("Finnish locale for Finland")`
        pub const TITLE: Option<&str> = Some("Finnish locale for Finland");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0EeNn]"`
        pub const NOEXPR: &str = "^[-0EeNn]";
        /// `Some("ei")`
        pub const NOSTR: Option<&str> = Some("ei");
        /// `"^[+1KkYy]"`
        pub const YESEXPR: &str = "^[+1KkYy]";
        /// `Some("kyllä")`
        pub const YESSTR: Option<&str> = Some("kyllä");
    }
    pub mod LC_MONETARY {
        /// `"€"`
        pub const CURRENCY_SYMBOL: &str = "€";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"EUR "`
        pub const INT_CURR_SYMBOL: &str = "EUR ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const MON_THOUSANDS_SEP: &str = "\u{202f}";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const THOUSANDS_SEP: &str = "\u{202f}";
    }
    pub mod LC_TELEPHONE {
        /// `Some("358")`
        pub const INT_PREFIX: Option<&str> = Some("358");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `Some("(%A) %l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("(%A) %l");
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["su", "ma", "ti", "ke", "to", "pe", "la"]`
        pub const ABDAY: &[&str] = &["su", "ma", "ti", "ke", "to", "pe", "la"];
        /// `&["tammi", "helmi", "maalis", "huhti", "touko", "kesä", "heinä", "elo", "syys", "loka", "marras", "joulu"]`
        pub const ABMON: &[&str] = &["tammi", "helmi", "maalis", "huhti", "touko", "kesä", "heinä", "elo", "syys", "loka", "marras", "joulu"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a %-d.%-m.%Y %H.%M.%S %z")`
        pub const DATE_FMT: Option<&str> = Some("%a %-d.%-m.%Y %H.%M.%S %z");
        /// `&["sunnuntai", "maanantai", "tiistai", "keskiviikko", "torstai", "perjantai", "lauantai"]`
        pub const DAY: &[&str] = &["sunnuntai", "maanantai", "tiistai", "keskiviikko", "torstai", "perjantai", "lauantai"];
        /// `"%d.%m.%Y"`
        pub const D_FMT: &str = "%d.%m.%Y";
        /// `"%a %e. %Bta %Y %H.%M.%S"`
        pub const D_T_FMT: &str = "%a %e. %Bta %Y %H.%M.%S";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["tammikuu", "helmikuu", "maaliskuu", "huhtikuu", "toukokuu", "kesäkuu", "heinäkuu", "elokuu", "syyskuu", "lokakuu", "marraskuu", "joulukuu"]`
        pub const MON: &[&str] = &["tammikuu", "helmikuu", "maaliskuu", "huhtikuu", "toukokuu", "kesäkuu", "heinäkuu", "elokuu", "syyskuu", "lokakuu", "marraskuu", "joulukuu"];
        /// `"%H.%M.%S"`
        pub const T_FMT: &str = "%H.%M.%S";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod fi_FI_euro {
    pub use super::fi_FI::LC_ADDRESS;
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("https://www.gnu.org/software/libc/")`
        pub const ADDRESS: Option<&str> = Some("https://www.gnu.org/software/libc/");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-08-20")`
        pub const DATE: Option<&str> = Some("2000-08-20");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Finnish")`
        pub const LANGUAGE: Option<&str> = Some("Finnish");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Free Software Foundation, Inc.")`
        pub const SOURCE: Option<&str> = Some("Free Software Foundation, Inc.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Finland")`
        pub const TERRITORY: Option<&str> = Some("Finland");
        /// `Some("Finnish locale for Finland with Euro")`
        pub const TITLE: Option<&str> = Some("Finnish locale for Finland with Euro");
    }
    pub use super::fi_FI::LC_MESSAGES;
    pub use super::fi_FI::LC_MONETARY;
    pub use super::fi_FI::LC_NUMERIC;
    pub use super::fi_FI::LC_TELEPHONE;
    pub use super::fi_FI::LC_TIME;
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod fil_PH {
    pub mod LC_ADDRESS {
        /// `Some("PHL")`
        pub const COUNTRY_AB3: Option<&str> = Some("PHL");
        /// `Some("PH")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("PH");
        /// `Some("RP")`
        pub const COUNTRY_CAR: Option<&str> = Some("RP");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Pilipinas")`
        pub const COUNTRY_NAME: Option<&str> = Some("Pilipinas");
        /// `Some(608)`
        pub const COUNTRY_NUM: Option<i64> = Some(608);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `None`
        pub const LANG_LIB: Option<&str> = None;
        /// `Some("Filipino")`
        pub const LANG_NAME: Option<&str> = Some("Filipino");
        /// `Some("fil")`
        pub const LANG_TERM: Option<&str> = Some("fil");
        /// `Some("%a%N%f%N%d%N%b%N%h %s %e %r%N%T %z%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%a%N%f%N%d%N%b%N%h %s %e %r%N%T %z%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_MONETARY"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_MONETARY"],
        ]);
        /// `Some("Rene Torres, Pablo Saratxaga")`
        pub const CONTACT: Option<&str> = Some("Rene Torres, Pablo Saratxaga");
        /// `Some("2005-02-02")`
        pub const DATE: Option<&str> = Some("2005-02-02");
        /// `Some("rgtorre@rocketmail.com, pablo@mandrakesoft.com")`
        pub const EMAIL: Option<&str> = Some("rgtorre@rocketmail.com, pablo@mandrakesoft.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Filipino")`
        pub const LANGUAGE: Option<&str> = Some("Filipino");
        /// `Some("0.5")`
        pub const REVISION: Option<&str> = Some("0.5");
        /// `Some("Rene Torres")`
        pub const SOURCE: Option<&str> = Some("Rene Torres");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Philippines")`
        pub const TERRITORY: Option<&str> = Some("Philippines");
        /// `Some("Filipino language locale for Philippines")`
        pub const TITLE: Option<&str> = Some("Filipino language locale for Philippines");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0hHnN]"`
        pub const NOEXPR: &str = "^[-0hHnN]";
        /// `Some("hindi")`
        pub const NOSTR: Option<&str> = Some("hindi");
        /// `"^[+1oOyY]"`
        pub const YESEXPR: &str = "^[+1oOyY]";
        /// `Some("oo")`
        pub const YESSTR: Option<&str> = Some("oo");
    }
    pub mod LC_MONETARY {
        /// `"₱"`
        pub const CURRENCY_SYMBOL: &str = "₱";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"PHP "`
        pub const INT_CURR_SYMBOL: &str = "PHP ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub use super::tl_PH::LC_NUMERIC;
    pub use super::en_PH::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Lin", "Lun", "Mar", "Miy", "Huw", "Biy", "Sab"]`
        pub const ABDAY: &[&str] = &["Lin", "Lun", "Mar", "Miy", "Huw", "Biy", "Sab"];
        /// `&["Ene", "Peb", "Mar", "Abr", "May", "Hun", "Hul", "Ago", "Set", "Okt", "Nob", "Dis"]`
        pub const ABMON: &[&str] = &["Ene", "Peb", "Mar", "Abr", "May", "Hun", "Hul", "Ago", "Set", "Okt", "Nob", "Dis"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["N.U.", "N.H."]`
        pub const AM_PM: &[&str] = &["N.U.", "N.H."];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Linggo", "Lunes", "Martes", "Miyerkoles", "Huwebes", "Biyernes", "Sabado"]`
        pub const DAY: &[&str] = &["Linggo", "Lunes", "Martes", "Miyerkoles", "Huwebes", "Biyernes", "Sabado"];
        /// `"%m/%d/%y"`
        pub const D_FMT: &str = "%m/%d/%y";
        /// `"%a %d %b %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Enero", "Pebrero", "Marso", "Abril", "Mayo", "Hunyo", "Hulyo", "Agosto", "Setyembre", "Oktubre", "Nobyembre", "Disyembre"]`
        pub const MON: &[&str] = &["Enero", "Pebrero", "Marso", "Abril", "Mayo", "Hunyo", "Hulyo", "Agosto", "Setyembre", "Oktubre", "Nobyembre", "Disyembre"];
        /// `"%I:%M:%S %p"`
        pub const T_FMT: &str = "%I:%M:%S %p";
        /// `"%I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod fo_FO {
    pub mod LC_ADDRESS {
        /// `Some("FRO")`
        pub const COUNTRY_AB3: Option<&str> = Some("FRO");
        /// `Some("FO")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("FO");
        /// `Some("FO")`
        pub const COUNTRY_CAR: Option<&str> = Some("FO");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Føroyar")`
        pub const COUNTRY_NAME: Option<&str> = Some("Føroyar");
        /// `Some(234)`
        pub const COUNTRY_NUM: Option<i64> = Some(234);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("fo")`
        pub const LANG_AB: Option<&str> = Some("fo");
        /// `None`
        pub const LANG_LIB: Option<&str> = None;
        /// `Some("føroyskt")`
        pub const LANG_NAME: Option<&str> = Some("føroyskt");
        /// `Some("fao")`
        pub const LANG_TERM: Option<&str> = Some("fao");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Kollegievej 6, DK-2920 Charlottenlund, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Kollegievej 6, DK-2920 Charlottenlund, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Faroese")`
        pub const LANGUAGE: Option<&str> = Some("Faroese");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Danish Standards Association")`
        pub const SOURCE: Option<&str> = Some("Danish Standards Association");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Faroe Islands")`
        pub const TERRITORY: Option<&str> = Some("Faroe Islands");
        /// `Some("Faroese locale for Faroe Islands")`
        pub const TITLE: Option<&str> = Some("Faroese locale for Faroe Islands");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0Nn]"`
        pub const NOEXPR: &str = "^[-0Nn]";
        /// `Some("nei")`
        pub const NOSTR: Option<&str> = Some("nei");
        /// `"^[+1JjYy]"`
        pub const YESEXPR: &str = "^[+1JjYy]";
        /// `Some("já")`
        pub const YESSTR: Option<&str> = Some("já");
    }
    pub use super::da_DK::LC_MONETARY;
    pub use super::da_DK::LC_NUMERIC;
    pub mod LC_TELEPHONE {
        /// `Some("298")`
        pub const INT_PREFIX: Option<&str> = Some("298");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["sun", "mán", "týs", "mik", "hós", "frí", "ley"]`
        pub const ABDAY: &[&str] = &["sun", "mán", "týs", "mik", "hós", "frí", "ley"];
        /// `&["jan", "feb", "mar", "apr", "mai", "jun", "jul", "aug", "sep", "okt", "nov", "des"]`
        pub const ABMON: &[&str] = &["jan", "feb", "mar", "apr", "mai", "jun", "jul", "aug", "sep", "okt", "nov", "des"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["sunnudagur", "mánadagur", "týsdagur", "mikudagur", "hósdagur", "fríggjadagur", "leygardagur"]`
        pub const DAY: &[&str] = &["sunnudagur", "mánadagur", "týsdagur", "mikudagur", "hósdagur", "fríggjadagur", "leygardagur"];
        /// `"%d/%m-%Y"`
        pub const D_FMT: &str = "%d/%m-%Y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["januar", "februar", "mars", "apríl", "mai", "juni", "juli", "august", "september", "oktober", "november", "desember"]`
        pub const MON: &[&str] = &["januar", "februar", "mars", "apríl", "mai", "juni", "juli", "august", "september", "oktober", "november", "desember"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod fr_BE {
    pub mod LC_ADDRESS {
        /// `Some("BEL")`
        pub const COUNTRY_AB3: Option<&str> = Some("BEL");
        /// `Some("BE")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("BE");
        /// `Some("B")`
        pub const COUNTRY_CAR: Option<&str> = Some("B");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Belgique")`
        pub const COUNTRY_NAME: Option<&str> = Some("Belgique");
        /// `Some(56)`
        pub const COUNTRY_NUM: Option<i64> = Some(56);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("fr")`
        pub const LANG_AB: Option<&str> = Some("fr");
        /// `Some("fre")`
        pub const LANG_LIB: Option<&str> = Some("fre");
        /// `Some("français")`
        pub const LANG_NAME: Option<&str> = Some("français");
        /// `Some("fra")`
        pub const LANG_TERM: Option<&str> = Some("fra");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("French")`
        pub const LANGUAGE: Option<&str> = Some("French");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Belgium")`
        pub const TERRITORY: Option<&str> = Some("Belgium");
        /// `Some("French locale for Belgium")`
        pub const TITLE: Option<&str> = Some("French locale for Belgium");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("non")`
        pub const NOSTR: Option<&str> = Some("non");
        /// `"^[+1oOjJyY]"`
        pub const YESEXPR: &str = "^[+1oOjJyY]";
        /// `Some("oui")`
        pub const YESSTR: Option<&str> = Some("oui");
    }
    pub mod LC_MONETARY {
        /// `"€"`
        pub const CURRENCY_SYMBOL: &str = "€";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"EUR "`
        pub const INT_CURR_SYMBOL: &str = "EUR ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const MON_THOUSANDS_SEP: &str = ".";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const THOUSANDS_SEP: &str = ".";
    }
    pub mod LC_TELEPHONE {
        /// `Some("32")`
        pub const INT_PREFIX: Option<&str> = Some("32");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["dim", "lun", "mar", "mer", "jeu", "ven", "sam"]`
        pub const ABDAY: &[&str] = &["dim", "lun", "mar", "mer", "jeu", "ven", "sam"];
        /// `&["jan", "fév", "mar", "avr", "mai", "jun", "jui", "aoû", "sep", "oct", "nov", "déc"]`
        pub const ABMON: &[&str] = &["jan", "fév", "mar", "avr", "mai", "jun", "jui", "aoû", "sep", "oct", "nov", "déc"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"]`
        pub const DAY: &[&str] = &["dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"]`
        pub const MON: &[&str] = &["janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod fr_BE_euro {
    pub use super::fr_BE::LC_ADDRESS;
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("https://www.gnu.org/software/libc/")`
        pub const ADDRESS: Option<&str> = Some("https://www.gnu.org/software/libc/");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-08-20")`
        pub const DATE: Option<&str> = Some("2000-08-20");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("French")`
        pub const LANGUAGE: Option<&str> = Some("French");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Free Software Foundation, Inc.")`
        pub const SOURCE: Option<&str> = Some("Free Software Foundation, Inc.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Belgium")`
        pub const TERRITORY: Option<&str> = Some("Belgium");
        /// `Some("French locale for Belgium with Euro")`
        pub const TITLE: Option<&str> = Some("French locale for Belgium with Euro");
    }
    pub use super::fr_BE::LC_MESSAGES;
    pub use super::fr_BE::LC_MONETARY;
    pub use super::fr_BE::LC_NUMERIC;
    pub use super::fr_BE::LC_TELEPHONE;
    pub use super::fr_BE::LC_TIME;
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod fr_CA {
    pub mod LC_ADDRESS {
        /// `Some("CAN")`
        pub const COUNTRY_AB3: Option<&str> = Some("CAN");
        /// `Some("CA")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("CA");
        /// `Some("CDN")`
        pub const COUNTRY_CAR: Option<&str> = Some("CDN");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Canada")`
        pub const COUNTRY_NAME: Option<&str> = Some("Canada");
        /// `Some(124)`
        pub const COUNTRY_NUM: Option<i64> = Some(124);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("fr")`
        pub const LANG_AB: Option<&str> = Some("fr");
        /// `Some("fre")`
        pub const LANG_LIB: Option<&str> = Some("fre");
        /// `Some("français")`
        pub const LANG_NAME: Option<&str> = Some("français");
        /// `Some("fra")`
        pub const LANG_TERM: Option<&str> = Some("fra");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Canadian French")`
        pub const LANGUAGE: Option<&str> = Some("Canadian French");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Canada")`
        pub const TERRITORY: Option<&str> = Some("Canada");
        /// `Some("French locale for Canada")`
        pub const TITLE: Option<&str> = Some("French locale for Canada");
    }
    pub use super::fr_FR::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"$"`
        pub const CURRENCY_SYMBOL: &str = "$";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"CAD "`
        pub const INT_CURR_SYMBOL: &str = "CAD ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const MON_THOUSANDS_SEP: &str = "\u{202f}";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `0`
        pub const N_SIGN_POSN: i64 = 0;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub use super::fr_FR::LC_NUMERIC;
    pub use super::en_CA::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["dim", "lun", "mar", "mer", "jeu", "ven", "sam"]`
        pub const ABDAY: &[&str] = &["dim", "lun", "mar", "mer", "jeu", "ven", "sam"];
        /// `&["jan", "fév", "mar", "avr", "mai", "jun", "jui", "aoû", "sep", "oct", "nov", "déc"]`
        pub const ABMON: &[&str] = &["jan", "fév", "mar", "avr", "mai", "jun", "jui", "aoû", "sep", "oct", "nov", "déc"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"]`
        pub const DAY: &[&str] = &["dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"];
        /// `"%Y-%m-%d"`
        pub const D_FMT: &str = "%Y-%m-%d";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"]`
        pub const MON: &[&str] = &["janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod fr_CH {
    pub mod LC_ADDRESS {
        /// `Some("CHE")`
        pub const COUNTRY_AB3: Option<&str> = Some("CHE");
        /// `Some("CH")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("CH");
        /// `Some("CH")`
        pub const COUNTRY_CAR: Option<&str> = Some("CH");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Suisse")`
        pub const COUNTRY_NAME: Option<&str> = Some("Suisse");
        /// `Some(756)`
        pub const COUNTRY_NUM: Option<i64> = Some(756);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("fr")`
        pub const LANG_AB: Option<&str> = Some("fr");
        /// `Some("fre")`
        pub const LANG_LIB: Option<&str> = Some("fre");
        /// `Some("français")`
        pub const LANG_NAME: Option<&str> = Some("français");
        /// `Some("fra")`
        pub const LANG_TERM: Option<&str> = Some("fra");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Swiss French")`
        pub const LANGUAGE: Option<&str> = Some("Swiss French");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Switzerland")`
        pub const TERRITORY: Option<&str> = Some("Switzerland");
        /// `Some("French locale for Switzerland")`
        pub const TITLE: Option<&str> = Some("French locale for Switzerland");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("non")`
        pub const NOSTR: Option<&str> = Some("non");
        /// `"^[+1OojJsSyY]"`
        pub const YESEXPR: &str = "^[+1OojJsSyY]";
        /// `Some("oui")`
        pub const YESSTR: Option<&str> = Some("oui");
    }
    pub use super::de_CH::LC_MONETARY;
    pub use super::de_CH::LC_NUMERIC;
    pub mod LC_TELEPHONE {
        /// `Some("41")`
        pub const INT_PREFIX: Option<&str> = Some("41");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["dim", "lun", "mar", "mer", "jeu", "ven", "sam"]`
        pub const ABDAY: &[&str] = &["dim", "lun", "mar", "mer", "jeu", "ven", "sam"];
        /// `&["jan", "fév", "mar", "avr", "mai", "jun", "jui", "aoû", "sep", "oct", "nov", "déc"]`
        pub const ABMON: &[&str] = &["jan", "fév", "mar", "avr", "mai", "jun", "jui", "aoû", "sep", "oct", "nov", "déc"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"]`
        pub const DAY: &[&str] = &["dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"];
        /// `"%d. %m. %y"`
        pub const D_FMT: &str = "%d. %m. %y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"]`
        pub const MON: &[&str] = &["janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod fr_FR {
    pub mod LC_ADDRESS {
        /// `Some("FRA")`
        pub const COUNTRY_AB3: Option<&str> = Some("FRA");
        /// `Some("FR")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("FR");
        /// `Some("F")`
        pub const COUNTRY_CAR: Option<&str> = Some("F");
        /// `Some("979-10")`
        pub const COUNTRY_ISBN: Option<&str> = Some("979-10");
        /// `Some("France")`
        pub const COUNTRY_NAME: Option<&str> = Some("France");
        /// `Some(250)`
        pub const COUNTRY_NUM: Option<i64> = Some(250);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("fr")`
        pub const LANG_AB: Option<&str> = Some("fr");
        /// `Some("fre")`
        pub const LANG_LIB: Option<&str> = Some("fre");
        /// `Some("français")`
        pub const LANG_NAME: Option<&str> = Some("français");
        /// `Some("fra")`
        pub const LANG_TERM: Option<&str> = Some("fra");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Traduc.org")`
        pub const CONTACT: Option<&str> = Some("Traduc.org");
        /// `Some("2008-03-15")`
        pub const DATE: Option<&str> = Some("2008-03-15");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("French")`
        pub const LANGUAGE: Option<&str> = Some("French");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("France")`
        pub const TERRITORY: Option<&str> = Some("France");
        /// `Some("French locale for France")`
        pub const TITLE: Option<&str> = Some("French locale for France");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("non")`
        pub const NOSTR: Option<&str> = Some("non");
        /// `"^[+1oOyY]"`
        pub const YESEXPR: &str = "^[+1oOyY]";
        /// `Some("oui")`
        pub const YESSTR: Option<&str> = Some("oui");
    }
    pub mod LC_MONETARY {
        /// `"€"`
        pub const CURRENCY_SYMBOL: &str = "€";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"EUR "`
        pub const INT_CURR_SYMBOL: &str = "EUR ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `"\u{202f}"`
        pub const MON_THOUSANDS_SEP: &str = "\u{202f}";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `"\u{202f}"`
        pub const THOUSANDS_SEP: &str = "\u{202f}";
    }
    pub mod LC_TELEPHONE {
        /// `Some("33")`
        pub const INT_PREFIX: Option<&str> = Some("33");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `Some("%a %l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("%a %l");
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."]`
        pub const ABDAY: &[&str] = &["dim.", "lun.", "mar.", "mer.", "jeu.", "ven.", "sam."];
        /// `&["janv.", "févr.", "mars", "avril", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."]`
        pub const ABMON: &[&str] = &["janv.", "févr.", "mars", "avril", "mai", "juin", "juil.", "août", "sept.", "oct.", "nov.", "déc."];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"]`
        pub const DAY: &[&str] = &["dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"]`
        pub const MON: &[&str] = &["janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod fr_FR_euro {
    pub use super::fr_FR::LC_ADDRESS;
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("https://www.gnu.org/software/libc/")`
        pub const ADDRESS: Option<&str> = Some("https://www.gnu.org/software/libc/");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `None`
        pub const CONTACT: Option<&str> = None;
        /// `Some("2000-08-20")`
        pub const DATE: Option<&str> = Some("2000-08-20");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("French")`
        pub const LANGUAGE: Option<&str> = Some("French");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Free Software Foundation, Inc.")`
        pub const SOURCE: Option<&str> = Some("Free Software Foundation, Inc.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("France")`
        pub const TERRITORY: Option<&str> = Some("France");
        /// `Some("French locale for France with Euro")`
        pub const TITLE: Option<&str> = Some("French locale for France with Euro");
    }
    pub use super::fr_FR::LC_MESSAGES;
    pub use super::fr_FR::LC_MONETARY;
    pub use super::fr_FR::LC_NUMERIC;
    pub use super::fr_FR::LC_TELEPHONE;
    pub use super::fr_FR::LC_TIME;
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod fr_LU {
    pub mod LC_ADDRESS {
        /// `Some("LUX")`
        pub const COUNTRY_AB3: Option<&str> = Some("LUX");
        /// `Some("LU")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("LU");
        /// `Some("L")`
        pub const COUNTRY_CAR: Option<&str> = Some("L");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Luxembourg")`
        pub const COUNTRY_NAME: Option<&str> = Some("Luxembourg");
        /// `Some(442)`
        pub const COUNTRY_NUM: Option<i64> = Some(442);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("fr")`
        pub const LANG_AB: Option<&str> = Some("fr");
        /// `Some("fre")`
        pub const LANG_LIB: Option<&str> = Some("fre");
        /// `Some("français")`
        pub const LANG_NAME: Option<&str> = Some("français");
        /// `Some("fra")`
        pub const LANG_TERM: Option<&str> = Some("fra");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("French")`
        pub const LANGUAGE: Option<&str> = Some("French");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Luxembourg")`
        pub const TERRITORY: Option<&str> = Some("Luxembourg");
        /// `Some("French locale for Luxemburg")`
        pub const TITLE: Option<&str> = Some("French locale for Luxemburg");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("non")`
        pub const NOSTR: Option<&str> = Some("non");
        /// `"^[+1oOyY]"`
        pub const YESEXPR: &str = "^[+1oOyY]";
        /// `Some("oui")`
        pub const YESSTR: Option<&str> = Some("oui");
    }
    pub mod LC_MONETARY {
        /// `"€"`
        pub const CURRENCY_SYMBOL: &str = "€";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"EUR "`
        pub const INT_CURR_SYMBOL: &str = "EUR ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const MON_THOUSANDS_SEP: &str = "\u{202f}";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const THOUSANDS_SEP: &str = "\u{202f}";
    }
    pub mod LC_TELEPHONE {
        /// `Some("352")`
        pub const INT_PREFIX: Option<&str> = Some("352");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["dim", "lun", "mar", "mer", "jeu", "ven", "sam"]`
        pub const ABDAY: &[&str] = &["dim", "lun", "mar", "mer", "jeu", "ven", "sam"];
        /// `&["jan", "fév", "mar", "avr", "mai", "jun", "jui", "aoû", "sep", "oct", "nov", "déc"]`
        pub const ABMON: &[&str] = &["jan", "fév", "mar", "avr", "mai", "jun", "jui", "aoû", "sep", "oct", "nov", "déc"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"]`
        pub const DAY: &[&str] = &["dimanche", "lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi"];
        /// `"%d.%m.%Y"`
        pub const D_FMT: &str = "%d.%m.%Y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"]`
        pub const MON: &[&str] = &["janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre", "décembre"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod fr_LU_euro {
    pub use super::fr_LU::LC_ADDRESS;
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("https://www.gnu.org/software/libc/")`
        pub const ADDRESS: Option<&str> = Some("https://www.gnu.org/software/libc/");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-08-20")`
        pub const DATE: Option<&str> = Some("2000-08-20");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("French")`
        pub const LANGUAGE: Option<&str> = Some("French");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Free Software Foundation, Inc.")`
        pub const SOURCE: Option<&str> = Some("Free Software Foundation, Inc.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Luxembourg")`
        pub const TERRITORY: Option<&str> = Some("Luxembourg");
        /// `Some("French locale for Luxemburg with Euro")`
        pub const TITLE: Option<&str> = Some("French locale for Luxemburg with Euro");
    }
    pub use super::fr_LU::LC_MESSAGES;
    pub use super::fr_LU::LC_MONETARY;
    pub use super::fr_LU::LC_NUMERIC;
    pub use super::fr_LU::LC_TELEPHONE;
    pub use super::fr_LU::LC_TIME;
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod fur_IT {
    pub mod LC_ADDRESS {
        /// `Some("ITA")`
        pub const COUNTRY_AB3: Option<&str> = Some("ITA");
        /// `Some("IT")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IT");
        /// `Some("I")`
        pub const COUNTRY_CAR: Option<&str> = Some("I");
        /// `Some("978-88,979-12")`
        pub const COUNTRY_ISBN: Option<&str> = Some("978-88,979-12");
        /// `Some("Italie")`
        pub const COUNTRY_NAME: Option<&str> = Some("Italie");
        /// `Some(380)`
        pub const COUNTRY_NUM: Option<i64> = Some(380);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `None`
        pub const LANG_LIB: Option<&str> = None;
        /// `Some("furlan")`
        pub const LANG_NAME: Option<&str> = Some("furlan");
        /// `Some("fur")`
        pub const LANG_TERM: Option<&str> = Some("fur");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Pablo Saratxaga")`
        pub const CONTACT: Option<&str> = Some("Pablo Saratxaga");
        /// `Some("2004-04-26")`
        pub const DATE: Option<&str> = Some("2004-04-26");
        /// `Some("pablo@mandrakesoft.com")`
        pub const EMAIL: Option<&str> = Some("pablo@mandrakesoft.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Friulian")`
        pub const LANGUAGE: Option<&str> = Some("Friulian");
        /// `Some("0.3")`
        pub const REVISION: Option<&str> = Some("0.3");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Italy")`
        pub const TERRITORY: Option<&str> = Some("Italy");
        /// `Some("Furlan locale for Italy")`
        pub const TITLE: Option<&str> = Some("Furlan locale for Italy");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("no")`
        pub const NOSTR: Option<&str> = Some("no");
        /// `"^[+1sSjJoOyY]"`
        pub const YESEXPR: &str = "^[+1sSjJoOyY]";
        /// `Some("sì")`
        pub const YESSTR: Option<&str> = Some("sì");
    }
    pub use super::it_IT::LC_MONETARY;
    pub use super::it_IT::LC_NUMERIC;
    pub use super::it_IT::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Dom", "Lun", "Mar", "Mie", "Joi", "Vin", "Sab"]`
        pub const ABDAY: &[&str] = &["Dom", "Lun", "Mar", "Mie", "Joi", "Vin", "Sab"];
        /// `&["Zen", "Fev", "Mar", "Avr", "Mai", "Jug", "Lui", "Avo", "Set", "Otu", "Nov", "Dic"]`
        pub const ABMON: &[&str] = &["Zen", "Fev", "Mar", "Avr", "Mai", "Jug", "Lui", "Avo", "Set", "Otu", "Nov", "Dic"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Domenie", "Lunis", "Martars", "Miercus", "Joibe", "Vinars", "Sabide"]`
        pub const DAY: &[&str] = &["Domenie", "Lunis", "Martars", "Miercus", "Joibe", "Vinars", "Sabide"];
        /// `"%d. %m. %y"`
        pub const D_FMT: &str = "%d. %m. %y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Zenâr", "Fevrâr", "Març", "Avrîl", "Mai", "Jugn", "Lui", "Avost", "Setembar", "Otubar", "Novembar", "Dicembar"]`
        pub const MON: &[&str] = &["Zenâr", "Fevrâr", "Març", "Avrîl", "Mai", "Jugn", "Lui", "Avost", "Setembar", "Otubar", "Novembar", "Dicembar"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod fy_DE {
    pub mod LC_ADDRESS {
        /// `Some("DEU")`
        pub const COUNTRY_AB3: Option<&str> = Some("DEU");
        /// `Some("DE")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("DE");
        /// `Some("D")`
        pub const COUNTRY_CAR: Option<&str> = Some("D");
        /// `Some("3")`
        pub const COUNTRY_ISBN: Option<&str> = Some("3");
        /// `Some("Dútslân")`
        pub const COUNTRY_NAME: Option<&str> = Some("Dútslân");
        /// `Some(276)`
        pub const COUNTRY_NUM: Option<i64> = Some(276);
        /// `Some("D")`
        pub const COUNTRY_POST: Option<&str> = Some("D");
        /// `Some("fy")`
        pub const LANG_AB: Option<&str> = Some("fy");
        /// `Some("fry")`
        pub const LANG_LIB: Option<&str> = Some("fry");
        /// `Some("Frysk")`
        pub const LANG_NAME: Option<&str> = Some("Frysk");
        /// `Some("fry")`
        pub const LANG_TERM: Option<&str> = Some("fry");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_ADDRESS"],
        ]);
        /// `Some("Kenneth Christiansen, Pablo Saratxaga")`
        pub const CONTACT: Option<&str> = Some("Kenneth Christiansen, Pablo Saratxaga");
        /// `Some("2003-11-30")`
        pub const DATE: Option<&str> = Some("2003-11-30");
        /// `Some("kenneth@gnu.org, pablo@mandriva.com")`
        pub const EMAIL: Option<&str> = Some("kenneth@gnu.org, pablo@mandriva.com");
        /// `None`
        pub const FAX: Option<&str> = None;
        /// `Some("Western Frisian")`
        pub const LANGUAGE: Option<&str> = Some("Western Frisian");
        /// `Some("0.1")`
        pub const REVISION: Option<&str> = Some("0.1");
        /// `Some("information from Kenneth Christiansen")`
        pub const SOURCE: Option<&str> = Some("information from Kenneth Christiansen");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Germany")`
        pub const TERRITORY: Option<&str> = Some("Germany");
        /// `Some("Sater Frisian and North Frisian Locale for Germany")`
        pub const TITLE: Option<&str> = Some("Sater Frisian and North Frisian Locale for Germany");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("nee")`
        pub const NOSTR: Option<&str> = Some("nee");
        /// `"^[+1jJyY]"`
        pub const YESEXPR: &str = "^[+1jJyY]";
        /// `Some("ja")`
        pub const YESSTR: Option<&str> = Some("ja");
    }
    pub use super::de_DE::LC_MONETARY;
    pub use super::de_DE::LC_NUMERIC;
    pub use super::de_DE::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Sdg", "Mdg", "Dsg", "Mwk", "Ddg", "Fdg", "Swd"]`
        pub const ABDAY: &[&str] = &["Sdg", "Mdg", "Dsg", "Mwk", "Ddg", "Fdg", "Swd"];
        /// `&["Jan", "Feb", "Moz", "Apr", "Mai", "Jun", "Jul", "Aug", "Sep", "Okt", "Now", "Dez"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Moz", "Apr", "Mai", "Jun", "Jul", "Aug", "Sep", "Okt", "Now", "Dez"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Sinndag", "Mondag", "Dingsdag", "Meddwäakj", "Donnadag", "Friedag", "Sinnowend"]`
        pub const DAY: &[&str] = &["Sinndag", "Mondag", "Dingsdag", "Meddwäakj", "Donnadag", "Friedag", "Sinnowend"];
        /// `"%d.%m.%Y"`
        pub const D_FMT: &str = "%d.%m.%Y";
        /// `"%a %d. %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d. %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Jaunuwoa", "Februwoa", "Moaz", "Aprell", "Mai", "Juni", "Juli", "August", "Septamba", "Oktoba", "Nowamba", "Dezamba"]`
        pub const MON: &[&str] = &["Jaunuwoa", "Februwoa", "Moaz", "Aprell", "Mai", "Juni", "Juli", "August", "Septamba", "Oktoba", "Nowamba", "Dezamba"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod fy_NL {
    pub mod LC_ADDRESS {
        /// `Some("NLD")`
        pub const COUNTRY_AB3: Option<&str> = Some("NLD");
        /// `Some("NL")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("NL");
        /// `Some("NL")`
        pub const COUNTRY_CAR: Option<&str> = Some("NL");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Nederlân")`
        pub const COUNTRY_NAME: Option<&str> = Some("Nederlân");
        /// `Some(528)`
        pub const COUNTRY_NUM: Option<i64> = Some(528);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("fy")`
        pub const LANG_AB: Option<&str> = Some("fy");
        /// `Some("fry")`
        pub const LANG_LIB: Option<&str> = Some("fry");
        /// `Some("West-Frysk")`
        pub const LANG_NAME: Option<&str> = Some("West-Frysk");
        /// `Some("fry")`
        pub const LANG_TERM: Option<&str> = Some("fry");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("https://www.gnu.org/software/libc/")`
        pub const ADDRESS: Option<&str> = Some("https://www.gnu.org/software/libc/");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2006-08-13")`
        pub const DATE: Option<&str> = Some("2006-08-13");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Western Frisian")`
        pub const LANGUAGE: Option<&str> = Some("Western Frisian");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Free Software Foundation, Inc.")`
        pub const SOURCE: Option<&str> = Some("Free Software Foundation, Inc.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Netherlands")`
        pub const TERRITORY: Option<&str> = Some("Netherlands");
        /// `Some("Frisian locale for the Netherlands")`
        pub const TITLE: Option<&str> = Some("Frisian locale for the Netherlands");
    }
    pub use super::nl_NL::LC_MESSAGES;
    pub use super::nl_NL::LC_MONETARY;
    pub use super::nl_NL::LC_NUMERIC;
    pub use super::nl_NL::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Sn", "Mo", "Ti", "Wo", "To", "Fr", "Sn"]`
        pub const ABDAY: &[&str] = &["Sn", "Mo", "Ti", "Wo", "To", "Fr", "Sn"];
        /// `&["Jan", "Feb", "Mrt", "Apr", "Mai", "Jun", "Jul", "Aug", "Sep", "Okt", "Nov", "Des"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Mrt", "Apr", "Mai", "Jun", "Jul", "Aug", "Sep", "Okt", "Nov", "Des"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Snein", "Moandei", "Tiisdei", "Woansdei", "Tongersdei", "Freed", "Sneon"]`
        pub const DAY: &[&str] = &["Snein", "Moandei", "Tiisdei", "Woansdei", "Tongersdei", "Freed", "Sneon"];
        /// `"%d-%m-%y"`
        pub const D_FMT: &str = "%d-%m-%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Jannewaris", "Febrewaris", "Maart", "April", "Maaie", "Juny", "July", "Augustus", "Septimber", "Oktober", "Novimber", "Desimber"]`
        pub const MON: &[&str] = &["Jannewaris", "Febrewaris", "Maart", "April", "Maaie", "Juny", "July", "Augustus", "Septimber", "Oktober", "Novimber", "Desimber"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ga_IE {
    pub mod LC_ADDRESS {
        /// `Some("IRL")`
        pub const COUNTRY_AB3: Option<&str> = Some("IRL");
        /// `Some("IE")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IE");
        /// `Some("IRL")`
        pub const COUNTRY_CAR: Option<&str> = Some("IRL");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Éire")`
        pub const COUNTRY_NAME: Option<&str> = Some("Éire");
        /// `Some(372)`
        pub const COUNTRY_NUM: Option<i64> = Some(372);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ga")`
        pub const LANG_AB: Option<&str> = Some("ga");
        /// `Some("gle")`
        pub const LANG_LIB: Option<&str> = Some("gle");
        /// `Some("Gaeilge")`
        pub const LANG_NAME: Option<&str> = Some("Gaeilge");
        /// `Some("gle")`
        pub const LANG_TERM: Option<&str> = Some("gle");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Glasnevin, Dublin 9, Ireland")`
        pub const ADDRESS: Option<&str> = Some("Glasnevin, Dublin 9, Ireland");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Irish")`
        pub const LANGUAGE: Option<&str> = Some("Irish");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("NSAI")`
        pub const SOURCE: Option<&str> = Some("NSAI");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Ireland")`
        pub const TERRITORY: Option<&str> = Some("Ireland");
        /// `Some("Irish locale for Ireland")`
        pub const TITLE: Option<&str> = Some("Irish locale for Ireland");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("níl")`
        pub const NOSTR: Option<&str> = Some("níl");
        /// `"^[+1tTyY]"`
        pub const YESEXPR: &str = "^[+1tTyY]";
        /// `Some("tá")`
        pub const YESSTR: Option<&str> = Some("tá");
    }
    pub mod LC_MONETARY {
        /// `"€"`
        pub const CURRENCY_SYMBOL: &str = "€";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"EUR "`
        pub const INT_CURR_SYMBOL: &str = "EUR ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("353")`
        pub const INT_PREFIX: Option<&str> = Some("353");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["Domh", "Luan", "Máirt", "Céad", "Déar", "Aoine", "Sath"]`
        pub const ABDAY: &[&str] = &["Domh", "Luan", "Máirt", "Céad", "Déar", "Aoine", "Sath"];
        /// `&["Ean", "Feabh", "Márta", "Aib", "Beal", "Meith", "Iúil", "Lún", "MFómh", "DFómh", "Samh", "Noll"]`
        pub const ABMON: &[&str] = &["Ean", "Feabh", "Márta", "Aib", "Beal", "Meith", "Iúil", "Lún", "MFómh", "DFómh", "Samh", "Noll"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Dé Domhnaigh", "Dé Luain", "Dé Máirt", "Dé Céadaoin", "Déardaoin", "Dé hAoine", "Dé Sathairn"]`
        pub const DAY: &[&str] = &["Dé Domhnaigh", "Dé Luain", "Dé Máirt", "Dé Céadaoin", "Déardaoin", "Dé hAoine", "Dé Sathairn"];
        /// `"%d.%m.%y"`
        pub const D_FMT: &str = "%d.%m.%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Eanáir", "Feabhra", "Márta", "Aibreán", "Bealtaine", "Meitheamh", "Iúil", "Lúnasa", "Meán Fómhair", "Deireadh Fómhair", "Samhain", "Nollaig"]`
        pub const MON: &[&str] = &["Eanáir", "Feabhra", "Márta", "Aibreán", "Bealtaine", "Meitheamh", "Iúil", "Lúnasa", "Meán Fómhair", "Deireadh Fómhair", "Samhain", "Nollaig"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ga_IE_euro {
    pub use super::ga_IE::LC_ADDRESS;
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("https://www.gnu.org/software/libc/")`
        pub const ADDRESS: Option<&str> = Some("https://www.gnu.org/software/libc/");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-08-21")`
        pub const DATE: Option<&str> = Some("2000-08-21");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Irish")`
        pub const LANGUAGE: Option<&str> = Some("Irish");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Free Software Foundation, Inc.")`
        pub const SOURCE: Option<&str> = Some("Free Software Foundation, Inc.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Ireland")`
        pub const TERRITORY: Option<&str> = Some("Ireland");
        /// `Some("Irish locale for Ireland with Euro")`
        pub const TITLE: Option<&str> = Some("Irish locale for Ireland with Euro");
    }
    pub use super::ga_IE::LC_MESSAGES;
    pub use super::ga_IE::LC_MONETARY;
    pub use super::ga_IE::LC_NUMERIC;
    pub use super::ga_IE::LC_TELEPHONE;
    pub use super::ga_IE::LC_TIME;
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod gd_GB {
    pub mod LC_ADDRESS {
        /// `Some("GBR")`
        pub const COUNTRY_AB3: Option<&str> = Some("GBR");
        /// `Some("GB")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("GB");
        /// `Some("GB")`
        pub const COUNTRY_CAR: Option<&str> = Some("GB");
        /// `Some("0")`
        pub const COUNTRY_ISBN: Option<&str> = Some("0");
        /// `Some("An Rìoghachd Aonaichte")`
        pub const COUNTRY_NAME: Option<&str> = Some("An Rìoghachd Aonaichte");
        /// `Some(826)`
        pub const COUNTRY_NUM: Option<i64> = Some(826);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("gd")`
        pub const LANG_AB: Option<&str> = Some("gd");
        /// `Some("gla")`
        pub const LANG_LIB: Option<&str> = Some("gla");
        /// `Some("Gàidhlig")`
        pub const LANG_NAME: Option<&str> = Some("Gàidhlig");
        /// `Some("gla")`
        pub const LANG_TERM: Option<&str> = Some("gla");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%S%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%S%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("1/2, 47 Wilton Street, Glasgow G20 6RT, UK")`
        pub const ADDRESS: Option<&str> = Some("1/2, 47 Wilton Street, Glasgow G20 6RT, UK");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
        ]);
        /// `Some("Michael Bauer")`
        pub const CONTACT: Option<&str> = Some("Michael Bauer");
        /// `Some("2013-04-25")`
        pub const DATE: Option<&str> = Some("2013-04-25");
        /// `Some("fios@akerbeltz.org")`
        pub const EMAIL: Option<&str> = Some("fios@akerbeltz.org");
        /// `None`
        pub const FAX: Option<&str> = None;
        /// `Some("Scottish Gaelic")`
        pub const LANGUAGE: Option<&str> = Some("Scottish Gaelic");
        /// `Some("1.2")`
        pub const REVISION: Option<&str> = Some("1.2");
        /// `Some("Akerbeltz")`
        pub const SOURCE: Option<&str> = Some("Akerbeltz");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("United Kingdom")`
        pub const TERRITORY: Option<&str> = Some("United Kingdom");
        /// `Some("Scots Gaelic language locale for Great Britain")`
        pub const TITLE: Option<&str> = Some("Scots Gaelic language locale for Great Britain");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0cCnN]"`
        pub const NOEXPR: &str = "^[-0cCnN]";
        /// `Some("chan eil")`
        pub const NOSTR: Option<&str> = Some("chan eil");
        /// `"^[+1tTyY]"`
        pub const YESEXPR: &str = "^[+1tTyY]";
        /// `Some("tha")`
        pub const YESSTR: Option<&str> = Some("tha");
    }
    pub use super::en_GB::LC_MONETARY;
    pub use super::en_GB::LC_NUMERIC;
    pub use super::en_GB::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["DiD", "DiL", "DiM", "DiC", "Dia", "Dih", "DiS"]`
        pub const ABDAY: &[&str] = &["DiD", "DiL", "DiM", "DiC", "Dia", "Dih", "DiS"];
        /// `&["Faoi", "Gearr", "Màrt", "Gibl", "Cèit", "Ògmh", "Iuch", "Lùna", "Sult", "Dàmh", "Samh", "Dùbh"]`
        pub const ABMON: &[&str] = &["Faoi", "Gearr", "Màrt", "Gibl", "Cèit", "Ògmh", "Iuch", "Lùna", "Sult", "Dàmh", "Samh", "Dùbh"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `Some(&["Am Faoilleach", "An Gearran", "Am Màrt", "An Giblean", "An Cèitean", "An t-Ògmhios", "An t-Iuchar", "An Lùnastal", "An t-Sultain", "An Dàmhair", "An t-Samhain", "An Dùbhlachd"])`
        pub const ALT_MON: Option<&[&str]> = Some(&["Am Faoilleach", "An Gearran", "Am Màrt", "An Giblean", "An Cèitean", "An t-Ògmhios", "An t-Iuchar", "An Lùnastal", "An t-Sultain", "An Dàmhair", "An t-Samhain", "An Dùbhlachd"]);
        /// `&["m", "f"]`
        pub const AM_PM: &[&str] = &["m", "f"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["DiDòmhnaich", "DiLuain", "DiMàirt", "DiCiadain", "DiarDaoin", "DihAoine", "DiSathairne"]`
        pub const DAY: &[&str] = &["DiDòmhnaich", "DiLuain", "DiMàirt", "DiCiadain", "DiarDaoin", "DihAoine", "DiSathairne"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["dhen Fhaoilleach", "dhen Ghearran", "dhen Mhàrt", "dhen Ghiblean", "dhen Chèitean", "dhen Ògmhios", "dhen Iuchar", "dhen Lùnastal", "dhen t-Sultain", "dhen Dàmhair", "dhen t-Samhain", "dhen Dùbhlachd"]`
        pub const MON: &[&str] = &["dhen Fhaoilleach", "dhen Ghearran", "dhen Mhàrt", "dhen Ghiblean", "dhen Chèitean", "dhen Ògmhios", "dhen Iuchar", "dhen Lùnastal", "dhen t-Sultain", "dhen Dàmhair", "dhen t-Samhain", "dhen Dùbhlachd"];
        /// `"%l:%M:%S %p"`
        pub const T_FMT: &str = "%l:%M:%S %p";
        /// `"%l:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%l:%M:%S %p";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod gez_ER {
    pub mod LC_ADDRESS {
        /// `Some("ERI")`
        pub const COUNTRY_AB3: Option<&str> = Some("ERI");
        /// `Some("ER")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("ER");
        /// `Some("ER")`
        pub const COUNTRY_CAR: Option<&str> = Some("ER");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("ኤርትራ")`
        pub const COUNTRY_NAME: Option<&str> = Some("ኤርትራ");
        /// `Some(232)`
        pub const COUNTRY_NUM: Option<i64> = Some(232);
        /// `Some("ERI")`
        pub const COUNTRY_POST: Option<&str> = Some("ERI");
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `None`
        pub const LANG_LIB: Option<&str> = None;
        /// `Some("ግዕዝ")`
        pub const LANG_NAME: Option<&str> = Some("ግዕዝ");
        /// `Some("gez")`
        pub const LANG_TERM: Option<&str> = Some("gez");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("7802 Solomon Seal Dr., Springfield, VA 22152, USA")`
        pub const ADDRESS: Option<&str> = Some("7802 Solomon Seal Dr., Springfield, VA 22152, USA");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_TIME"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_TIME"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2003-07-05")`
        pub const DATE: Option<&str> = Some("2003-07-05");
        /// `Some("locales@geez.org")`
        pub const EMAIL: Option<&str> = Some("locales@geez.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Geez")`
        pub const LANGUAGE: Option<&str> = Some("Geez");
        /// `Some("0.20")`
        pub const REVISION: Option<&str> = Some("0.20");
        /// `Some("Ge'ez Frontier Foundation")`
        pub const SOURCE: Option<&str> = Some("Ge'ez Frontier Foundation");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Eritrea")`
        pub const TERRITORY: Option<&str> = Some("Eritrea");
        /// `Some("Ge'ez language locale for Eritrea.")`
        pub const TITLE: Option<&str> = Some("Ge'ez language locale for Eritrea.");
    }
    pub use super::ti_ER::LC_MESSAGES;
    pub use super::ti_ER::LC_MONETARY;
    pub use super::ti_ER::LC_NUMERIC;
    pub use super::ti_ER::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["እኁድ", "ሰኑይ", "ሠሉስ", "ራብዕ", "ሐሙስ", "ዓርበ", "ቀዳሚ"]`
        pub const ABDAY: &[&str] = &["እኁድ", "ሰኑይ", "ሠሉስ", "ራብዕ", "ሐሙስ", "ዓርበ", "ቀዳሚ"];
        /// `&["ጠሐረ", "ከተተ", "መገበ", "አኀዘ", "ግንባ", "ሠንየ", "ሐመለ", "ነሐሰ", "ከረመ", "ጠቀመ", "ኀደረ", "ኀሠሠ"]`
        pub const ABMON: &[&str] = &["ጠሐረ", "ከተተ", "መገበ", "አኀዘ", "ግንባ", "ሠንየ", "ሐመለ", "ነሐሰ", "ከረመ", "ጠቀመ", "ኀደረ", "ኀሠሠ"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["ጽባሕ", "ምሴት"]`
        pub const AM_PM: &[&str] = &["ጽባሕ", "ምሴት"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%A፥%B፡%e፡መዓልት፡%r፡%Z፡%Y፡ዓ/ም")`
        pub const DATE_FMT: Option<&str> = Some("%A፥%B፡%e፡መዓልት፡%r፡%Z፡%Y፡ዓ/ም");
        /// `&["እኁድ", "ሰኑይ", "ሠሉስ", "ራብዕ", "ሐሙስ", "ዓርበ", "ቀዳሚት"]`
        pub const DAY: &[&str] = &["እኁድ", "ሰኑይ", "ሠሉስ", "ራብዕ", "ሐሙስ", "ዓርበ", "ቀዳሚት"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%A፥%B፡%e፡መዓልት፡%Y፡%l:%M:%S፡%p፡%Z"`
        pub const D_T_FMT: &str = "%A፥%B፡%e፡መዓልት፡%Y፡%l:%M:%S፡%p፡%Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["ጠሐረ", "ከተተ", "መገበ", "አኀዘ", "ግንባት", "ሠንየ", "ሐመለ", "ነሐሰ", "ከረመ", "ጠቀመ", "ኀደረ", "ኀሠሠ"]`
        pub const MON: &[&str] = &["ጠሐረ", "ከተተ", "መገበ", "አኀዘ", "ግንባት", "ሠንየ", "ሐመለ", "ነሐሰ", "ከረመ", "ጠቀመ", "ኀደረ", "ኀሠሠ"];
        /// `"%l:%M:%S፡%p"`
        pub const T_FMT: &str = "%l:%M:%S፡%p";
        /// `"%l:%M:%S፡%p"`
        pub const T_FMT_AMPM: &str = "%l:%M:%S፡%p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod gez_ER_abegede {
    pub use super::gez_ER::LC_ADDRESS;
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("7802 Solomon Seal Dr., Springfield, VA 22152, USA")`
        pub const ADDRESS: Option<&str> = Some("7802 Solomon Seal Dr., Springfield, VA 22152, USA");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_TIME"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_TIME"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2003-07-05")`
        pub const DATE: Option<&str> = Some("2003-07-05");
        /// `Some("locales@geez.org")`
        pub const EMAIL: Option<&str> = Some("locales@geez.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Geez")`
        pub const LANGUAGE: Option<&str> = Some("Geez");
        /// `Some("0.20")`
        pub const REVISION: Option<&str> = Some("0.20");
        /// `Some("Ge'ez Frontier Foundation")`
        pub const SOURCE: Option<&str> = Some("Ge'ez Frontier Foundation");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Eritrea")`
        pub const TERRITORY: Option<&str> = Some("Eritrea");
        /// `Some("Ge'ez language locale for Eritrea With Abegede Collation.")`
        pub const TITLE: Option<&str> = Some("Ge'ez language locale for Eritrea With Abegede Collation.");
    }
    pub use super::gez_ER::LC_MESSAGES;
    pub use super::gez_ER::LC_MONETARY;
    pub use super::gez_ER::LC_NUMERIC;
    pub use super::gez_ER::LC_TELEPHONE;
    pub use super::gez_ER::LC_TIME;
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod gez_ET {
    pub mod LC_ADDRESS {
        /// `Some("ETH")`
        pub const COUNTRY_AB3: Option<&str> = Some("ETH");
        /// `Some("ET")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("ET");
        /// `Some("ETH")`
        pub const COUNTRY_CAR: Option<&str> = Some("ETH");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("ኢትዮጵያ")`
        pub const COUNTRY_NAME: Option<&str> = Some("ኢትዮጵያ");
        /// `Some(231)`
        pub const COUNTRY_NUM: Option<i64> = Some(231);
        /// `Some("ETH")`
        pub const COUNTRY_POST: Option<&str> = Some("ETH");
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `None`
        pub const LANG_LIB: Option<&str> = None;
        /// `Some("ግዕዝ")`
        pub const LANG_NAME: Option<&str> = Some("ግዕዝ");
        /// `Some("gez")`
        pub const LANG_TERM: Option<&str> = Some("gez");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("7802 Solomon Seal Dr., Springfield, VA 22152, USA")`
        pub const ADDRESS: Option<&str> = Some("7802 Solomon Seal Dr., Springfield, VA 22152, USA");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_TIME"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_TIME"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2003-07-05")`
        pub const DATE: Option<&str> = Some("2003-07-05");
        /// `Some("locales@geez.org")`
        pub const EMAIL: Option<&str> = Some("locales@geez.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Geez")`
        pub const LANGUAGE: Option<&str> = Some("Geez");
        /// `Some("0.20")`
        pub const REVISION: Option<&str> = Some("0.20");
        /// `Some("Ge'ez Frontier Foundation")`
        pub const SOURCE: Option<&str> = Some("Ge'ez Frontier Foundation");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Ethiopia")`
        pub const TERRITORY: Option<&str> = Some("Ethiopia");
        /// `Some("Ge'ez language locale for Ethiopia")`
        pub const TITLE: Option<&str> = Some("Ge'ez language locale for Ethiopia");
    }
    pub use super::ti_ET::LC_MESSAGES;
    pub use super::ti_ET::LC_MONETARY;
    pub use super::ti_ET::LC_NUMERIC;
    pub use super::ti_ET::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["እኁድ", "ሰኑይ", "ሠሉስ", "ራብዕ", "ሐሙስ", "ዓርበ", "ቀዳሚ"]`
        pub const ABDAY: &[&str] = &["እኁድ", "ሰኑይ", "ሠሉስ", "ራብዕ", "ሐሙስ", "ዓርበ", "ቀዳሚ"];
        /// `&["ጃንዩ", "ፌብሩ", "ማርች", "ኤፕረ", "ሜይ ", "ጁን ", "ጁላይ", "ኦገስ", "ሴፕቴ", "ኦክተ", "ኖቬም", "ዲሴም"]`
        pub const ABMON: &[&str] = &["ጃንዩ", "ፌብሩ", "ማርች", "ኤፕረ", "ሜይ ", "ጁን ", "ጁላይ", "ኦገስ", "ሴፕቴ", "ኦክተ", "ኖቬም", "ዲሴም"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["ጽባሕ", "ምሴት"]`
        pub const AM_PM: &[&str] = &["ጽባሕ", "ምሴት"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%A፥%B፡%e፡መዓልት፡%r፡%Z፡%Y፡ዓ/ም")`
        pub const DATE_FMT: Option<&str> = Some("%A፥%B፡%e፡መዓልት፡%r፡%Z፡%Y፡ዓ/ም");
        /// `&["እኁድ", "ሰኑይ", "ሠሉስ", "ራብዕ", "ሐሙስ", "ዓርበ", "ቀዳሚት"]`
        pub const DAY: &[&str] = &["እኁድ", "ሰኑይ", "ሠሉስ", "ራብዕ", "ሐሙስ", "ዓርበ", "ቀዳሚት"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%A፥%B፡%e፡መዓልት፡%Y፡%l:%M:%S፡%p፡%Z"`
        pub const D_T_FMT: &str = "%A፥%B፡%e፡መዓልት፡%Y፡%l:%M:%S፡%p፡%Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["ጃንዩወሪ", "ፌብሩወሪ", "ማርች", "ኤፕረል", "ሜይ", "ጁን", "ጁላይ", "ኦገስት", "ሴፕቴምበር", "ኦክተውበር", "ኖቬምበር", "ዲሴምበር"]`
        pub const MON: &[&str] = &["ጃንዩወሪ", "ፌብሩወሪ", "ማርች", "ኤፕረል", "ሜይ", "ጁን", "ጁላይ", "ኦገስት", "ሴፕቴምበር", "ኦክተውበር", "ኖቬምበር", "ዲሴምበር"];
        /// `"%l:%M:%S፡%p"`
        pub const T_FMT: &str = "%l:%M:%S፡%p";
        /// `"%l:%M:%S፡%p"`
        pub const T_FMT_AMPM: &str = "%l:%M:%S፡%p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod gez_ET_abegede {
    pub use super::gez_ET::LC_ADDRESS;
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("7802 Solomon Seal Dr., Springfield, VA 22152, USA")`
        pub const ADDRESS: Option<&str> = Some("7802 Solomon Seal Dr., Springfield, VA 22152, USA");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_TIME"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_TIME"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2003-07-05")`
        pub const DATE: Option<&str> = Some("2003-07-05");
        /// `Some("locales@geez.org")`
        pub const EMAIL: Option<&str> = Some("locales@geez.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Geez")`
        pub const LANGUAGE: Option<&str> = Some("Geez");
        /// `Some("0.20")`
        pub const REVISION: Option<&str> = Some("0.20");
        /// `Some("Ge'ez Frontier Foundation")`
        pub const SOURCE: Option<&str> = Some("Ge'ez Frontier Foundation");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Ethiopia")`
        pub const TERRITORY: Option<&str> = Some("Ethiopia");
        /// `Some("Ge'ez language locale for Ethiopia With Abegede Collation")`
        pub const TITLE: Option<&str> = Some("Ge'ez language locale for Ethiopia With Abegede Collation");
    }
    pub use super::gez_ET::LC_MESSAGES;
    pub use super::gez_ET::LC_MONETARY;
    pub use super::gez_ET::LC_NUMERIC;
    pub use super::gez_ET::LC_TELEPHONE;
    pub use super::gez_ET::LC_TIME;
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod gl_ES {
    pub mod LC_ADDRESS {
        /// `Some("ESP")`
        pub const COUNTRY_AB3: Option<&str> = Some("ESP");
        /// `Some("ES")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("ES");
        /// `Some("E")`
        pub const COUNTRY_CAR: Option<&str> = Some("E");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("España")`
        pub const COUNTRY_NAME: Option<&str> = Some("España");
        /// `Some(724)`
        pub const COUNTRY_NUM: Option<i64> = Some(724);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("gl")`
        pub const LANG_AB: Option<&str> = Some("gl");
        /// `Some("glg")`
        pub const LANG_LIB: Option<&str> = Some("glg");
        /// `Some("galego")`
        pub const LANG_NAME: Option<&str> = Some("galego");
        /// `Some("glg")`
        pub const LANG_TERM: Option<&str> = Some("glg");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Facultade de Informática, Campus de Elviña, s/n, 15071 A Coruña, Spain")`
        pub const ADDRESS: Option<&str> = Some("Facultade de Informática, Campus de Elviña, s/n, 15071 A Coruña, Spain");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Galician")`
        pub const LANGUAGE: Option<&str> = Some("Galician");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("GPUL")`
        pub const SOURCE: Option<&str> = Some("GPUL");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Spain")`
        pub const TERRITORY: Option<&str> = Some("Spain");
        /// `Some("Galician locale for Spain")`
        pub const TITLE: Option<&str> = Some("Galician locale for Spain");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("non")`
        pub const NOSTR: Option<&str> = Some("non");
        /// `"^[+1sSyY]"`
        pub const YESEXPR: &str = "^[+1sSyY]";
        /// `Some("si")`
        pub const YESSTR: Option<&str> = Some("si");
    }
    pub mod LC_MONETARY {
        /// `"€"`
        pub const CURRENCY_SYMBOL: &str = "€";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"EUR "`
        pub const INT_CURR_SYMBOL: &str = "EUR ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const MON_THOUSANDS_SEP: &str = ".";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[0, 0]`
        pub const GROUPING: &[i64] = &[0, 0];
        /// `""`
        pub const THOUSANDS_SEP: &str = "";
    }
    pub mod LC_TELEPHONE {
        /// `Some("34")`
        pub const INT_PREFIX: Option<&str> = Some("34");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["Dom", "Lun", "Mar", "Mér", "Xov", "Ven", "Sáb"]`
        pub const ABDAY: &[&str] = &["Dom", "Lun", "Mar", "Mér", "Xov", "Ven", "Sáb"];
        /// `&["Xan", "Feb", "Mar", "Abr", "Mai", "Xuñ", "Xul", "Ago", "Set", "Out", "Nov", "Dec"]`
        pub const ABMON: &[&str] = &["Xan", "Feb", "Mar", "Abr", "Mai", "Xuñ", "Xul", "Ago", "Set", "Out", "Nov", "Dec"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Domingo", "Luns", "Martes", "Mércores", "Xoves", "Venres", "Sábado"]`
        pub const DAY: &[&str] = &["Domingo", "Luns", "Martes", "Mércores", "Xoves", "Venres", "Sábado"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Xaneiro", "Febreiro", "Marzo", "Abril", "Maio", "Xuño", "Xullo", "Agosto", "Setembro", "Outubro", "Novembro", "Decembro"]`
        pub const MON: &[&str] = &["Xaneiro", "Febreiro", "Marzo", "Abril", "Maio", "Xuño", "Xullo", "Agosto", "Setembro", "Outubro", "Novembro", "Decembro"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod gl_ES_euro {
    pub use super::gl_ES::LC_ADDRESS;
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("https://www.gnu.org/software/libc/")`
        pub const ADDRESS: Option<&str> = Some("https://www.gnu.org/software/libc/");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-08-21")`
        pub const DATE: Option<&str> = Some("2000-08-21");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Galician")`
        pub const LANGUAGE: Option<&str> = Some("Galician");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Free Software Foundation, Inc.")`
        pub const SOURCE: Option<&str> = Some("Free Software Foundation, Inc.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Spain")`
        pub const TERRITORY: Option<&str> = Some("Spain");
        /// `Some("Galician locale for Spain with Euro")`
        pub const TITLE: Option<&str> = Some("Galician locale for Spain with Euro");
    }
    pub use super::gl_ES::LC_MESSAGES;
    pub use super::gl_ES::LC_MONETARY;
    pub use super::gl_ES::LC_NUMERIC;
    pub use super::gl_ES::LC_TELEPHONE;
    pub use super::gl_ES::LC_TIME;
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod gu_IN {
    pub mod LC_ADDRESS {
        /// `Some("IND")`
        pub const COUNTRY_AB3: Option<&str> = Some("IND");
        /// `Some("IN")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IN");
        /// `Some("IND")`
        pub const COUNTRY_CAR: Option<&str> = Some("IND");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("ભારત")`
        pub const COUNTRY_NAME: Option<&str> = Some("ભારત");
        /// `Some(356)`
        pub const COUNTRY_NUM: Option<i64> = Some(356);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("gu")`
        pub const LANG_AB: Option<&str> = Some("gu");
        /// `Some("guj")`
        pub const LANG_LIB: Option<&str> = Some("guj");
        /// `Some("ગ\u{ac1}જરાતી")`
        pub const LANG_NAME: Option<&str> = Some("ગ\u{ac1}જરાતી");
        /// `Some("guj")`
        pub const LANG_TERM: Option<&str> = Some("guj");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2004-09-14")`
        pub const DATE: Option<&str> = Some("2004-09-14");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Gujarati")`
        pub const LANGUAGE: Option<&str> = Some("Gujarati");
        /// `Some("0.2")`
        pub const REVISION: Option<&str> = Some("0.2");
        /// `Some("IndLinux.org")`
        pub const SOURCE: Option<&str> = Some("IndLinux.org");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("India")`
        pub const TERRITORY: Option<&str> = Some("India");
        /// `Some("Gujarati Language Locale For India")`
        pub const TITLE: Option<&str> = Some("Gujarati Language Locale For India");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNન]"`
        pub const NOEXPR: &str = "^[-0nNન]";
        /// `Some("નહી\u{a82}")`
        pub const NOSTR: Option<&str> = Some("નહી\u{a82}");
        /// `"^[+1yYહ]"`
        pub const YESEXPR: &str = "^[+1yYહ]";
        /// `Some("હા")`
        pub const YESSTR: Option<&str> = Some("હા");
    }
    pub use super::hi_IN::LC_MONETARY;
    pub use super::hi_IN::LC_NUMERIC;
    pub use super::hi_IN::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["રવિ", "સોમ", "મ\u{a82}ગળ", "બ\u{ac1}ધ", "ગ\u{ac1}ર\u{ac1}", "શ\u{ac1}ક\u{acd}ર", "શનિ"]`
        pub const ABDAY: &[&str] = &["રવિ", "સોમ", "મ\u{a82}ગળ", "બ\u{ac1}ધ", "ગ\u{ac1}ર\u{ac1}", "શ\u{ac1}ક\u{acd}ર", "શનિ"];
        /// `&["જાન\u{acd}ય\u{ac1}", "ફ\u{ac7}બ\u{acd}ર\u{ac1}", "માર\u{acd}ચ", "એપ\u{acd}રિલ", "મ\u{ac7}", "જ\u{ac2}ન", "જ\u{ac1}લાઈ", "ઑગસ\u{acd}ટ", "સપ\u{acd}ટ\u{ac7}", "ઑક\u{acd}ટો", "નવ\u{ac7}", "ડિસ\u{ac7}"]`
        pub const ABMON: &[&str] = &["જાન\u{acd}ય\u{ac1}", "ફ\u{ac7}બ\u{acd}ર\u{ac1}", "માર\u{acd}ચ", "એપ\u{acd}રિલ", "મ\u{ac7}", "જ\u{ac2}ન", "જ\u{ac1}લાઈ", "ઑગસ\u{acd}ટ", "સપ\u{acd}ટ\u{ac7}", "ઑક\u{acd}ટો", "નવ\u{ac7}", "ડિસ\u{ac7}"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["સવાર\u{ac7}", "સા\u{a82}જ\u{ac7}"]`
        pub const AM_PM: &[&str] = &["સવાર\u{ac7}", "સા\u{a82}જ\u{ac7}"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["રવિવાર", "સોમવાર", "મ\u{a82}ગળવાર", "બ\u{ac1}ધવાર", "ગ\u{ac1}ર\u{ac1}વાર", "શ\u{ac1}ક\u{acd}રવાર", "શનિવાર"]`
        pub const DAY: &[&str] = &["રવિવાર", "સોમવાર", "મ\u{a82}ગળવાર", "બ\u{ac1}ધવાર", "ગ\u{ac1}ર\u{ac1}વાર", "શ\u{ac1}ક\u{acd}રવાર", "શનિવાર"];
        /// `"%-d/%-m/%y"`
        pub const D_FMT: &str = "%-d/%-m/%y";
        /// `"%A %d %b %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A %d %b %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["જાન\u{acd}ય\u{ac1}આરી", "ફ\u{ac7}બ\u{acd}ર\u{ac1}આરી", "માર\u{acd}ચ", "એપ\u{acd}રિલ", "મ\u{ac7}", "જ\u{ac2}ન", "જ\u{ac1}લાઈ", "ઑગસ\u{acd}ટ", "સપ\u{acd}ટ\u{ac7}મ\u{acd}બર", "ઑક\u{acd}ટોબર", "નવ\u{ac7}મ\u{acd}બર", "ડિસ\u{ac7}મ\u{acd}બર"]`
        pub const MON: &[&str] = &["જાન\u{acd}ય\u{ac1}આરી", "ફ\u{ac7}બ\u{acd}ર\u{ac1}આરી", "માર\u{acd}ચ", "એપ\u{acd}રિલ", "મ\u{ac7}", "જ\u{ac2}ન", "જ\u{ac1}લાઈ", "ઑગસ\u{acd}ટ", "સપ\u{acd}ટ\u{ac7}મ\u{acd}બર", "ઑક\u{acd}ટોબર", "નવ\u{ac7}મ\u{acd}બર", "ડિસ\u{ac7}મ\u{acd}બર"];
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT: &str = "%I:%M:%S %p %Z";
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p %Z";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod gv_GB {
    pub mod LC_ADDRESS {
        /// `Some("GBR")`
        pub const COUNTRY_AB3: Option<&str> = Some("GBR");
        /// `Some("GB")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("GB");
        /// `Some("GB")`
        pub const COUNTRY_CAR: Option<&str> = Some("GB");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Rywvaneth Unys")`
        pub const COUNTRY_NAME: Option<&str> = Some("Rywvaneth Unys");
        /// `Some(826)`
        pub const COUNTRY_NUM: Option<i64> = Some(826);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("gv")`
        pub const LANG_AB: Option<&str> = Some("gv");
        /// `Some("glv")`
        pub const LANG_LIB: Option<&str> = Some("glv");
        /// `Some("Gaelg")`
        pub const LANG_NAME: Option<&str> = Some("Gaelg");
        /// `Some("glv")`
        pub const LANG_TERM: Option<&str> = Some("glv");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Croí Lár, Ballinahalla, Maigh Cuilinn,, Co. Gaillimh, Ireland")`
        pub const ADDRESS: Option<&str> = Some("Croí Lár, Ballinahalla, Maigh Cuilinn,, Co. Gaillimh, Ireland");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Manx")`
        pub const LANGUAGE: Option<&str> = Some("Manx");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Alastair McKinstry")`
        pub const SOURCE: Option<&str> = Some("Alastair McKinstry");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("United Kingdom")`
        pub const TERRITORY: Option<&str> = Some("United Kingdom");
        /// `Some("Manx Gaelic locale for Britain")`
        pub const TITLE: Option<&str> = Some("Manx Gaelic locale for Britain");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `None`
        pub const NOSTR: Option<&str> = None;
        /// `"^[+1yY]"`
        pub const YESEXPR: &str = "^[+1yY]";
        /// `None`
        pub const YESSTR: Option<&str> = None;
    }
    pub mod LC_MONETARY {
        /// `"£"`
        pub const CURRENCY_SYMBOL: &str = "£";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"GBP "`
        pub const INT_CURR_SYMBOL: &str = "GBP ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("44")`
        pub const INT_PREFIX: Option<&str> = Some("44");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["Jed", "Jel", "Jem", "Jerc", "Jerd", "Jeh", "Jes"]`
        pub const ABDAY: &[&str] = &["Jed", "Jel", "Jem", "Jerc", "Jerd", "Jeh", "Jes"];
        /// `&["J-guer", "T-arree", "Mayrnt", "Avrril", "Boaldyn", "M-souree", "J-souree", "Luanistyn", "M-fouyir", "J-fouyir", "M.Houney", "M.Nollick"]`
        pub const ABMON: &[&str] = &["J-guer", "T-arree", "Mayrnt", "Avrril", "Boaldyn", "M-souree", "J-souree", "Luanistyn", "M-fouyir", "J-fouyir", "M.Houney", "M.Nollick"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Jedoonee", "Jelhein", "Jemayrt", "Jercean", "Jerdein", "Jeheiney", "Jesarn"]`
        pub const DAY: &[&str] = &["Jedoonee", "Jelhein", "Jemayrt", "Jercean", "Jerdein", "Jeheiney", "Jesarn"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Jerrey-geuree", "Toshiaght-arree", "Mayrnt", "Averil", "Boaldyn", "Mean-souree", "Jerrey-souree", "Luanistyn", "Mean-fouyir", "Jerrey-fouyir", "Mee Houney", "Mee ny Nollick"]`
        pub const MON: &[&str] = &["Jerrey-geuree", "Toshiaght-arree", "Mayrnt", "Averil", "Boaldyn", "Mean-souree", "Jerrey-souree", "Luanistyn", "Mean-fouyir", "Jerrey-fouyir", "Mee Houney", "Mee ny Nollick"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ha_NG {
    pub mod LC_ADDRESS {
        /// `Some("NGA")`
        pub const COUNTRY_AB3: Option<&str> = Some("NGA");
        /// `Some("NG")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("NG");
        /// `Some("WAN")`
        pub const COUNTRY_CAR: Option<&str> = Some("WAN");
        /// `Some("978")`
        pub const COUNTRY_ISBN: Option<&str> = Some("978");
        /// `Some("Najeriya")`
        pub const COUNTRY_NAME: Option<&str> = Some("Najeriya");
        /// `Some(566)`
        pub const COUNTRY_NUM: Option<i64> = Some(566);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ha")`
        pub const LANG_AB: Option<&str> = Some("ha");
        /// `Some("hau")`
        pub const LANG_LIB: Option<&str> = Some("hau");
        /// `Some("Hausa")`
        pub const LANG_NAME: Option<&str> = Some("Hausa");
        /// `Some("hau")`
        pub const LANG_TERM: Option<&str> = Some("hau");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2006-02-01")`
        pub const DATE: Option<&str> = Some("2006-02-01");
        /// `Some("pablo@mandriva.com")`
        pub const EMAIL: Option<&str> = Some("pablo@mandriva.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Hausa")`
        pub const LANGUAGE: Option<&str> = Some("Hausa");
        /// `Some("0.2")`
        pub const REVISION: Option<&str> = Some("0.2");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Nigeria")`
        pub const TERRITORY: Option<&str> = Some("Nigeria");
        /// `Some("Hausa locale for Nigeria")`
        pub const TITLE: Option<&str> = Some("Hausa locale for Nigeria");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0bBaAnN]"`
        pub const NOEXPR: &str = "^[-0bBaAnN]";
        /// `Some("aʼa")`
        pub const NOSTR: Option<&str> = Some("aʼa");
        /// `"^[+1TtiIYy]"`
        pub const YESEXPR: &str = "^[+1TtiIYy]";
        /// `Some("i")`
        pub const YESSTR: Option<&str> = Some("i");
    }
    pub use super::en_NG::LC_MONETARY;
    pub use super::en_NG::LC_NUMERIC;
    pub use super::en_NG::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Lah", "Lit", "Tal", "Lar", "Alh", "Jum", "Asa"]`
        pub const ABDAY: &[&str] = &["Lah", "Lit", "Tal", "Lar", "Alh", "Jum", "Asa"];
        /// `&["Jan", "Fab", "Mar", "Afi", "May", "Yun", "Yul", "Agu", "Sat", "Okt", "Nuw", "Dis"]`
        pub const ABMON: &[&str] = &["Jan", "Fab", "Mar", "Afi", "May", "Yun", "Yul", "Agu", "Sat", "Okt", "Nuw", "Dis"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["AM", "PM"]`
        pub const AM_PM: &[&str] = &["AM", "PM"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Lahadi", "Litini", "Talata", "Laraba", "Alhamis", "Juma'a", "Asabar"]`
        pub const DAY: &[&str] = &["Lahadi", "Litini", "Talata", "Laraba", "Alhamis", "Juma'a", "Asabar"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"ranar %A, %d ga %B cikin %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "ranar %A, %d ga %B cikin %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Janairu", "Faburairu", "Maris", "Afirilu", "Mayu", "Yuni", "Yuli", "Agusta", "Satumba", "Oktoba", "Nuwamba", "Disamba"]`
        pub const MON: &[&str] = &["Janairu", "Faburairu", "Maris", "Afirilu", "Mayu", "Yuni", "Yuli", "Agusta", "Satumba", "Oktoba", "Nuwamba", "Disamba"];
        /// `"%I:%M:%S %p"`
        pub const T_FMT: &str = "%I:%M:%S %p";
        /// `"%I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod hak_TW {
    pub mod LC_ADDRESS {
        /// `Some("TWN")`
        pub const COUNTRY_AB3: Option<&str> = Some("TWN");
        /// `Some("TW")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("TW");
        /// `Some("RC")`
        pub const COUNTRY_CAR: Option<&str> = Some("RC");
        /// `Some("957")`
        pub const COUNTRY_ISBN: Option<&str> = Some("957");
        /// `Some("中華民國")`
        pub const COUNTRY_NAME: Option<&str> = Some("中華民國");
        /// `Some(158)`
        pub const COUNTRY_NUM: Option<i64> = Some(158);
        /// `Some("TW")`
        pub const COUNTRY_POST: Option<&str> = Some("TW");
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("hak")`
        pub const LANG_LIB: Option<&str> = Some("hak");
        /// `Some("客家話")`
        pub const LANG_NAME: Option<&str> = Some("客家話");
        /// `Some("hak")`
        pub const LANG_TERM: Option<&str> = Some("hak");
        /// `Some("%c%N%T%N%s %h %e %r%N%b%N%d%N%f%N%a%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%c%N%T%N%s %h %e %r%N%b%N%d%N%f%N%a%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2013-06-02")`
        pub const DATE: Option<&str> = Some("2013-06-02");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Hakka Chinese")`
        pub const LANGUAGE: Option<&str> = Some("Hakka Chinese");
        /// `Some("0.1")`
        pub const REVISION: Option<&str> = Some("0.1");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Taiwan")`
        pub const TERRITORY: Option<&str> = Some("Taiwan");
        /// `Some("Hakka Chinese locale for the Republic of China")`
        pub const TITLE: Option<&str> = Some("Hakka Chinese locale for the Republic of China");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNｎＮ毋]"`
        pub const NOEXPR: &str = "^[-0nNｎＮ毋]";
        /// `None`
        pub const NOSTR: Option<&str> = None;
        /// `"^[+1yYｙＹ係]"`
        pub const YESEXPR: &str = "^[+1yYｙＹ係]";
        /// `None`
        pub const YESSTR: Option<&str> = None;
    }
    pub mod LC_MONETARY {
        /// `"NT$"`
        pub const CURRENCY_SYMBOL: &str = "NT$";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"TWD "`
        pub const INT_CURR_SYMBOL: &str = "TWD ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `Some(1)`
        pub const INT_N_CS_PRECEDES: Option<i64> = Some(1);
        /// `Some(0)`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = Some(0);
        /// `Some(1)`
        pub const INT_N_SIGN_POSN: Option<i64> = Some(1);
        /// `Some(1)`
        pub const INT_P_CS_PRECEDES: Option<i64> = Some(1);
        /// `Some(0)`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = Some(0);
        /// `Some(1)`
        pub const INT_P_SIGN_POSN: Option<i64> = Some(1);
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[4]`
        pub const MON_GROUPING: &[i64] = &[4];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[4]`
        pub const GROUPING: &[i64] = &[4];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("886")`
        pub const INT_PREFIX: Option<&str> = Some("886");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `Some("%A-%l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("%A-%l");
        /// `Some("+%c-%a-%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c-%a-%l");
    }
    pub mod LC_TIME {
        /// `&["日", "一", "二", "三", "四", "五", "六"]`
        pub const ABDAY: &[&str] = &["日", "一", "二", "三", "四", "五", "六"];
        /// `&[" 1月", " 2月", " 3月", " 4月", " 5月", " 6月", " 7月", " 8月", " 9月", "10月", "11月", "12月"]`
        pub const ABMON: &[&str] = &[" 1月", " 2月", " 3月", " 4月", " 5月", " 6月", " 7月", " 8月", " 9月", "10月", "11月", "12月"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["上晝", "下晝"]`
        pub const AM_PM: &[&str] = &["上晝", "下晝"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%Y年 %b %e日 %A %H:%M:%S %Z")`
        pub const DATE_FMT: Option<&str> = Some("%Y年 %b %e日 %A %H:%M:%S %Z");
        /// `&["禮拜日", "禮拜一", "禮拜二", "禮拜三", "禮拜四", "禮拜五", "禮拜六"]`
        pub const DAY: &[&str] = &["禮拜日", "禮拜一", "禮拜二", "禮拜三", "禮拜四", "禮拜五", "禮拜六"];
        /// `"%Y年%m月%d日"`
        pub const D_FMT: &str = "%Y年%m月%d日";
        /// `"%Y年%m月%d日 (%A) %H點%M分%S秒"`
        pub const D_T_FMT: &str = "%Y年%m月%d日 (%A) %H點%M分%S秒";
        /// `Some(&["+:2:1913/01/01:+*:民國:%EC%Ey年", "+:1:1912/01/01:1912/12/31:民國:%EC元年", "+:1:1911/12/31:-*:民前:%EC%Ey年"])`
        pub const ERA: Option<&[&str]> = Some(&["+:2:1913/01/01:+*:民國:%EC%Ey年", "+:1:1912/01/01:1912/12/31:民國:%EC元年", "+:1:1911/12/31:-*:民前:%EC%Ey年"]);
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["一月", "二月", "三月", "四月", "五月", "六月", "七月", "八月", "九月", "十月", "十一月", "十二月"]`
        pub const MON: &[&str] = &["一月", "二月", "三月", "四月", "五月", "六月", "七月", "八月", "九月", "十月", "十一月", "十二月"];
        /// `"%H點%M分%S秒"`
        pub const T_FMT: &str = "%H點%M分%S秒";
        /// `"%p %I點%M分%S秒"`
        pub const T_FMT_AMPM: &str = "%p %I點%M分%S秒";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod he_IL {
    pub mod LC_ADDRESS {
        /// `Some("ISR")`
        pub const COUNTRY_AB3: Option<&str> = Some("ISR");
        /// `Some("IL")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IL");
        /// `Some("IL")`
        pub const COUNTRY_CAR: Option<&str> = Some("IL");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("ישראל")`
        pub const COUNTRY_NAME: Option<&str> = Some("ישראל");
        /// `Some(376)`
        pub const COUNTRY_NUM: Option<i64> = Some(376);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("he")`
        pub const LANG_AB: Option<&str> = Some("he");
        /// `Some("heb")`
        pub const LANG_LIB: Option<&str> = Some("heb");
        /// `Some("עברית")`
        pub const LANG_NAME: Option<&str> = Some("עברית");
        /// `Some("heb")`
        pub const LANG_TERM: Option<&str> = Some("heb");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jo/rgens Alle 8, DK-1615 Ko/benhavn V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jo/rgens Alle 8, DK-1615 Ko/benhavn V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Hebrew")`
        pub const LANGUAGE: Option<&str> = Some("Hebrew");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Israel")`
        pub const TERRITORY: Option<&str> = Some("Israel");
        /// `Some("Hebrew locale for Israel")`
        pub const TITLE: Option<&str> = Some("Hebrew locale for Israel");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNל]"`
        pub const NOEXPR: &str = "^[-0nNל]";
        /// `Some("לא")`
        pub const NOSTR: Option<&str> = Some("לא");
        /// `"^[+1yYכ]"`
        pub const YESEXPR: &str = "^[+1yYכ]";
        /// `Some("כן")`
        pub const YESSTR: Option<&str> = Some("כן");
    }
    pub mod LC_MONETARY {
        /// `"₪"`
        pub const CURRENCY_SYMBOL: &str = "₪";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"ILS "`
        pub const INT_CURR_SYMBOL: &str = "ILS ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `2`
        pub const N_SIGN_POSN: i64 = 2;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `2`
        pub const P_SIGN_POSN: i64 = 2;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("972")`
        pub const INT_PREFIX: Option<&str> = Some("972");
        /// `Some("00,012,013,014,018")`
        pub const INT_SELECT: Option<&str> = Some("00,012,013,014,018");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["א'", "ב'", "ג'", "ד'", "ה'", "ו'", "ש'"]`
        pub const ABDAY: &[&str] = &["א'", "ב'", "ג'", "ד'", "ה'", "ו'", "ש'"];
        /// `&["ינו", "פבר", "מרץ", "אפר", "מאי", "יונ", "יול", "אוג", "ספט", "אוק", "נוב", "דצמ"]`
        pub const ABMON: &[&str] = &["ינו", "פבר", "מרץ", "אפר", "מאי", "יונ", "יול", "אוג", "ספט", "אוק", "נוב", "דצמ"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["AM", "PM"]`
        pub const AM_PM: &[&str] = &["AM", "PM"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["ראשון", "שני", "שלישי", "רביעי", "חמישי", "שישי", "שבת"]`
        pub const DAY: &[&str] = &["ראשון", "שני", "שלישי", "רביעי", "חמישי", "שישי", "שבת"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%Z %H:%M:%S %Y %b %d %a"`
        pub const D_T_FMT: &str = "%Z %H:%M:%S %Y %b %d %a";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `Some(1)`
        pub const FIRST_WORKDAY: Option<i64> = Some(1);
        /// `&["ינואר", "פברואר", "מרץ", "אפריל", "מאי", "יוני", "יולי", "אוגוסט", "ספטמבר", "אוקטובר", "נובמבר", "דצמבר"]`
        pub const MON: &[&str] = &["ינואר", "פברואר", "מרץ", "אפריל", "מאי", "יוני", "יולי", "אוגוסט", "ספטמבר", "אוקטובר", "נובמבר", "דצמבר"];
        /// `"%H:%M:%S"`
        pub const T_FMT: &str = "%H:%M:%S";
        /// `"%I:%M:%S %P"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %P";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod hi_IN {
    pub mod LC_ADDRESS {
        /// `Some("IND")`
        pub const COUNTRY_AB3: Option<&str> = Some("IND");
        /// `Some("IN")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IN");
        /// `Some("IND")`
        pub const COUNTRY_CAR: Option<&str> = Some("IND");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("भारत")`
        pub const COUNTRY_NAME: Option<&str> = Some("भारत");
        /// `Some(356)`
        pub const COUNTRY_NUM: Option<i64> = Some(356);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("hi")`
        pub const LANG_AB: Option<&str> = Some("hi");
        /// `Some("hin")`
        pub const LANG_LIB: Option<&str> = Some("hin");
        /// `Some("हिन\u{94d}दी")`
        pub const LANG_NAME: Option<&str> = Some("हिन\u{94d}दी");
        /// `Some("hin")`
        pub const LANG_TERM: Option<&str> = Some("hin");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan")`
        pub const ADDRESS: Option<&str> = Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-07-21")`
        pub const DATE: Option<&str> = Some("2000-07-21");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Hindi")`
        pub const LANGUAGE: Option<&str> = Some("Hindi");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("IBM Globalization Center of Competency, Yamato Software Laboratory")`
        pub const SOURCE: Option<&str> = Some("IBM Globalization Center of Competency, Yamato Software Laboratory");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("India")`
        pub const TERRITORY: Option<&str> = Some("India");
        /// `Some("Hindi language locale for India")`
        pub const TITLE: Option<&str> = Some("Hindi language locale for India");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNन]"`
        pub const NOEXPR: &str = "^[-0nNन]";
        /// `Some("नही\u{902}")`
        pub const NOSTR: Option<&str> = Some("नही\u{902}");
        /// `"^[+1yYह]"`
        pub const YESEXPR: &str = "^[+1yYह]";
        /// `Some("हा\u{901}")`
        pub const YESSTR: Option<&str> = Some("हा\u{901}");
    }
    pub mod LC_MONETARY {
        /// `"₹"`
        pub const CURRENCY_SYMBOL: &str = "₹";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"INR "`
        pub const INT_CURR_SYMBOL: &str = "INR ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 2]`
        pub const MON_GROUPING: &[i64] = &[3, 2];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("91")`
        pub const INT_PREFIX: Option<&str> = Some("91");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c ;%a ;%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c ;%a ;%l");
    }
    pub mod LC_TIME {
        /// `&["रवि", "सोम", "म\u{902}गल", "ब\u{941}ध", "ग\u{941}र\u{941}", "श\u{941}क\u{94d}र", "शनि"]`
        pub const ABDAY: &[&str] = &["रवि", "सोम", "म\u{902}गल", "ब\u{941}ध", "ग\u{941}र\u{941}", "श\u{941}क\u{94d}र", "शनि"];
        /// `&["जन॰", "फ\u{93c}र॰", "मार\u{94d}च", "अप\u{94d}र\u{948}ल", "मई", "ज\u{942}न", "ज\u{941}ल॰", "अग॰", "सित॰", "अक\u{94d}त\u{942}॰", "नव॰", "दिस॰"]`
        pub const ABMON: &[&str] = &["जन॰", "फ\u{93c}र॰", "मार\u{94d}च", "अप\u{94d}र\u{948}ल", "मई", "ज\u{942}न", "ज\u{941}ल॰", "अग॰", "सित॰", "अक\u{94d}त\u{942}॰", "नव॰", "दिस॰"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["प\u{942}र\u{94d}वाह\u{94d}न", "अपराह\u{94d}न"]`
        pub const AM_PM: &[&str] = &["प\u{942}र\u{94d}वाह\u{94d}न", "अपराह\u{94d}न"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["रविवार", "सोमवार", "म\u{902}गलवार", "ब\u{941}धवार", "ग\u{941}र\u{941}वार", "श\u{941}क\u{94d}रवार", "शनिवार"]`
        pub const DAY: &[&str] = &["रविवार", "सोमवार", "म\u{902}गलवार", "ब\u{941}धवार", "ग\u{941}र\u{941}वार", "श\u{941}क\u{94d}रवार", "शनिवार"];
        /// `"%-d/%-m/%y"`
        pub const D_FMT: &str = "%-d/%-m/%y";
        /// `"%A %d %b %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A %d %b %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["जनवरी", "फ\u{93c}रवरी", "मार\u{94d}च", "अप\u{94d}र\u{948}ल", "मई", "ज\u{942}न", "ज\u{941}लाई", "अगस\u{94d}त", "सित\u{902}बर", "अक\u{94d}त\u{942}बर", "नव\u{902}बर", "दिस\u{902}बर"]`
        pub const MON: &[&str] = &["जनवरी", "फ\u{93c}रवरी", "मार\u{94d}च", "अप\u{94d}र\u{948}ल", "मई", "ज\u{942}न", "ज\u{941}लाई", "अगस\u{94d}त", "सित\u{902}बर", "अक\u{94d}त\u{942}बर", "नव\u{902}बर", "दिस\u{902}बर"];
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT: &str = "%I:%M:%S %p %Z";
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p %Z";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod hif_FJ {
    pub mod LC_ADDRESS {
        /// `Some("FJI")`
        pub const COUNTRY_AB3: Option<&str> = Some("FJI");
        /// `Some("FJ")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("FJ");
        /// `Some("FJI")`
        pub const COUNTRY_CAR: Option<&str> = Some("FJI");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Fiji")`
        pub const COUNTRY_NAME: Option<&str> = Some("Fiji");
        /// `Some(242)`
        pub const COUNTRY_NUM: Option<i64> = Some(242);
        /// `Some("")`
        pub const COUNTRY_POST: Option<&str> = Some("");
        /// `Some("")`
        pub const LANG_AB: Option<&str> = Some("");
        /// `Some("")`
        pub const LANG_LIB: Option<&str> = Some("");
        /// `Some("हिन\u{94d}दी")`
        pub const LANG_NAME: Option<&str> = Some("हिन\u{94d}दी");
        /// `Some("hif")`
        pub const LANG_TERM: Option<&str> = Some("hif");
        /// `Some("%a%N%h%t%s%N%N%T%c")`
        pub const POSTAL_FMT: Option<&str> = Some("%a%N%h%t%s%N%N%T%c");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2017-03-21")`
        pub const DATE: Option<&str> = Some("2017-03-21");
        /// `Some("akhilesh.k@samsung.com")`
        pub const EMAIL: Option<&str> = Some("akhilesh.k@samsung.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Fiji Hindi")`
        pub const LANGUAGE: Option<&str> = Some("Fiji Hindi");
        /// `Some("0.1")`
        pub const REVISION: Option<&str> = Some("0.1");
        /// `Some("Samsung Electronics Co., Ltd.")`
        pub const SOURCE: Option<&str> = Some("Samsung Electronics Co., Ltd.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Fiji")`
        pub const TERRITORY: Option<&str> = Some("Fiji");
        /// `Some("Fiji Hindi (Latin) language locale for Fiji")`
        pub const TITLE: Option<&str> = Some("Fiji Hindi (Latin) language locale for Fiji");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("Nahi")`
        pub const NOSTR: Option<&str> = Some("Nahi");
        /// `"^[+1hHyY]"`
        pub const YESEXPR: &str = "^[+1hHyY]";
        /// `Some("Haan")`
        pub const YESSTR: Option<&str> = Some("Haan");
    }
    pub mod LC_MONETARY {
        /// `"FJ$"`
        pub const CURRENCY_SYMBOL: &str = "FJ$";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"FJD "`
        pub const INT_CURR_SYMBOL: &str = "FJD ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `Some(1)`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = Some(1);
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `Some(1)`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = Some(1);
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub use super::hi_IN::LC_NUMERIC;
    pub mod LC_TELEPHONE {
        /// `Some("679")`
        pub const INT_PREFIX: Option<&str> = Some("679");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %l");
    }
    pub mod LC_TIME {
        /// `&["Ravi", "Som", "Mangal", "Budh", "Guru", "Shukra", "Shani"]`
        pub const ABDAY: &[&str] = &["Ravi", "Som", "Mangal", "Budh", "Guru", "Shukra", "Shani"];
        /// `&["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["Purvahan", "Aparaahna"]`
        pub const AM_PM: &[&str] = &["Purvahan", "Aparaahna"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%A %d %b %Y %I:%M:%S %p %Z")`
        pub const DATE_FMT: Option<&str> = Some("%A %d %b %Y %I:%M:%S %p %Z");
        /// `&["Ravivar", "Somvar", "Mangalvar", "Budhvar", "Guruvar", "Shukravar", "Shanivar"]`
        pub const DAY: &[&str] = &["Ravivar", "Somvar", "Mangalvar", "Budhvar", "Guruvar", "Shukravar", "Shanivar"];
        /// `"%A %d %b %Y"`
        pub const D_FMT: &str = "%A %d %b %Y";
        /// `"%A %d %b %Y %I:%M:%S %p"`
        pub const D_T_FMT: &str = "%A %d %b %Y %I:%M:%S %p";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"]`
        pub const MON: &[&str] = &["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"];
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT: &str = "%I:%M:%S %p %Z";
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p %Z";
        /// `None`
        pub const WEEK: Option<&[i64]> = None;
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod hne_IN {
    pub mod LC_ADDRESS {
        /// `Some("IND")`
        pub const COUNTRY_AB3: Option<&str> = Some("IND");
        /// `Some("IN")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IN");
        /// `Some("IND")`
        pub const COUNTRY_CAR: Option<&str> = Some("IND");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("भारत")`
        pub const COUNTRY_NAME: Option<&str> = Some("भारत");
        /// `Some(356)`
        pub const COUNTRY_NUM: Option<i64> = Some(356);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `None`
        pub const LANG_LIB: Option<&str> = None;
        /// `Some("छत\u{94d}तीसगढ\u{93c}ी")`
        pub const LANG_NAME: Option<&str> = Some("छत\u{94d}तीसगढ\u{93c}ी");
        /// `Some("hne")`
        pub const LANG_TERM: Option<&str> = Some("hne");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Marisfot III, Marigold Premises, East-Wing, Kalyaninagar, Pune, India-411014")`
        pub const ADDRESS: Option<&str> = Some("Marisfot III, Marigold Premises, East-Wing, Kalyaninagar, Pune, India-411014");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2008-12-03")`
        pub const DATE: Option<&str> = Some("2008-12-03");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Chhattisgarhi")`
        pub const LANGUAGE: Option<&str> = Some("Chhattisgarhi");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Red Hat, Pune")`
        pub const SOURCE: Option<&str> = Some("Red Hat, Pune");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("India")`
        pub const TERRITORY: Option<&str> = Some("India");
        /// `Some("Chhattisgarhi language locale for India")`
        pub const TITLE: Option<&str> = Some("Chhattisgarhi language locale for India");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNन]"`
        pub const NOEXPR: &str = "^[-0nNन]";
        /// `Some("नइ")`
        pub const NOSTR: Option<&str> = Some("नइ");
        /// `"^[+1yYह]"`
        pub const YESEXPR: &str = "^[+1yYह]";
        /// `Some("हव")`
        pub const YESSTR: Option<&str> = Some("हव");
    }
    pub use super::hi_IN::LC_MONETARY;
    pub use super::hi_IN::LC_NUMERIC;
    pub use super::hi_IN::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["इत", "सोम", "म\u{902}ग", "ब\u{941}ध", "बिर", "स\u{941}क", "सनि"]`
        pub const ABDAY: &[&str] = &["इत", "सोम", "म\u{902}ग", "ब\u{941}ध", "बिर", "स\u{941}क", "सनि"];
        /// `&["जन", "फर", "मार\u{94d}च", "अप", "मई", "ज\u{942}न", "ज\u{941}ला", "अग", "सित", "अकट\u{942}", "नव", "दिस"]`
        pub const ABMON: &[&str] = &["जन", "फर", "मार\u{94d}च", "अप", "मई", "ज\u{942}न", "ज\u{941}ला", "अग", "सित", "अकट\u{942}", "नव", "दिस"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["बिहिनिया\u{901}", "म\u{902}झनिया\u{901}"]`
        pub const AM_PM: &[&str] = &["बिहिनिया\u{901}", "म\u{902}झनिया\u{901}"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["इतवार", "सोमवार", "म\u{902}गलवार", "ब\u{941}धवार", "बिरसपत", "स\u{941}करवार", "सनिवार"]`
        pub const DAY: &[&str] = &["इतवार", "सोमवार", "म\u{902}गलवार", "ब\u{941}धवार", "बिरसपत", "स\u{941}करवार", "सनिवार"];
        /// `"%-d/%-m/%y"`
        pub const D_FMT: &str = "%-d/%-m/%y";
        /// `"%A %d %b %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A %d %b %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["जनवरी", "फरवरी", "मार\u{94d}च", "अपर\u{947}ल", "मई", "ज\u{942}न", "ज\u{941}लाई", "अगस\u{94d}त", "सितमबर", "अकट\u{942}बर", "नवमबर", "दिसमबर"]`
        pub const MON: &[&str] = &["जनवरी", "फरवरी", "मार\u{94d}च", "अपर\u{947}ल", "मई", "ज\u{942}न", "ज\u{941}लाई", "अगस\u{94d}त", "सितमबर", "अकट\u{942}बर", "नवमबर", "दिसमबर"];
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT: &str = "%I:%M:%S %p %Z";
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p %Z";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod hr_HR {
    pub mod LC_ADDRESS {
        /// `Some("HRV")`
        pub const COUNTRY_AB3: Option<&str> = Some("HRV");
        /// `Some("HR")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("HR");
        /// `Some("HR")`
        pub const COUNTRY_CAR: Option<&str> = Some("HR");
        /// `Some("978-953")`
        pub const COUNTRY_ISBN: Option<&str> = Some("978-953");
        /// `Some("Hrvatska")`
        pub const COUNTRY_NAME: Option<&str> = Some("Hrvatska");
        /// `Some(191)`
        pub const COUNTRY_NUM: Option<i64> = Some(191);
        /// `Some("HR")`
        pub const COUNTRY_POST: Option<&str> = Some("HR");
        /// `Some("hr")`
        pub const LANG_AB: Option<&str> = Some("hr");
        /// `Some("hrv")`
        pub const LANG_LIB: Option<&str> = Some("hrv");
        /// `Some("hrvatski")`
        pub const LANG_NAME: Option<&str> = Some("hrvatski");
        /// `Some("hrv")`
        pub const LANG_TERM: Option<&str> = Some("hrv");
        /// `Some("%f%N%d%N%a%N%s %h%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%d%N%a%N%s %h%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2016-04-16")`
        pub const DATE: Option<&str> = Some("2016-04-16");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Croatian")`
        pub const LANGUAGE: Option<&str> = Some("Croatian");
        /// `Some("2.3")`
        pub const REVISION: Option<&str> = Some("2.3");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Croatia")`
        pub const TERRITORY: Option<&str> = Some("Croatia");
        /// `Some("Croatian locale for Croatia")`
        pub const TITLE: Option<&str> = Some("Croatian locale for Croatia");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("ne")`
        pub const NOSTR: Option<&str> = Some("ne");
        /// `"^[+1dDyY]"`
        pub const YESEXPR: &str = "^[+1dDyY]";
        /// `Some("da")`
        pub const YESSTR: Option<&str> = Some("da");
    }
    pub mod LC_MONETARY {
        /// `"kn"`
        pub const CURRENCY_SYMBOL: &str = "kn";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"HRK "`
        pub const INT_CURR_SYMBOL: &str = "HRK ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `Some(1)`
        pub const INT_N_CS_PRECEDES: Option<i64> = Some(1);
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `Some(1)`
        pub const INT_P_CS_PRECEDES: Option<i64> = Some(1);
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const MON_THOUSANDS_SEP: &str = ".";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const THOUSANDS_SEP: &str = ".";
    }
    pub mod LC_TELEPHONE {
        /// `Some("385")`
        pub const INT_PREFIX: Option<&str> = Some("385");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `Some("%A %l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("%A %l");
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["ned", "pon", "uto", "sri", "čet", "pet", "sub"]`
        pub const ABDAY: &[&str] = &["ned", "pon", "uto", "sri", "čet", "pet", "sub"];
        /// `&["sij", "velj", "ožu", "tra", "svi", "lip", "srp", "kol", "ruj", "lis", "stu", "pro"]`
        pub const ABMON: &[&str] = &["sij", "velj", "ožu", "tra", "svi", "lip", "srp", "kol", "ruj", "lis", "stu", "pro"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `Some(&["siječanj", "veljača", "ožujak", "travanj", "svibanj", "lipanj", "srpanj", "kolovoz", "rujan", "listopad", "studeni", "prosinac"])`
        pub const ALT_MON: Option<&[&str]> = Some(&["siječanj", "veljača", "ožujak", "travanj", "svibanj", "lipanj", "srpanj", "kolovoz", "rujan", "listopad", "studeni", "prosinac"]);
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a, %e.%m.%Y.  %H:%M:%S %Z")`
        pub const DATE_FMT: Option<&str> = Some("%a, %e.%m.%Y.  %H:%M:%S %Z");
        /// `&["nedjelja", "ponedjeljak", "utorak", "srijeda", "četvrtak", "petak", "subota"]`
        pub const DAY: &[&str] = &["nedjelja", "ponedjeljak", "utorak", "srijeda", "četvrtak", "petak", "subota"];
        /// `"%d.%m.%Y"`
        pub const D_FMT: &str = "%d.%m.%Y";
        /// `"%A, %d. %B %Y. %T %Z"`
        pub const D_T_FMT: &str = "%A, %d. %B %Y. %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `Some(2)`
        pub const FIRST_WORKDAY: Option<i64> = Some(2);
        /// `&["siječnja", "veljače", "ožujka", "travnja", "svibnja", "lipnja", "srpnja", "kolovoza", "rujna", "listopada", "studenoga", "prosinca"]`
        pub const MON: &[&str] = &["siječnja", "veljače", "ožujka", "travnja", "svibnja", "lipnja", "srpnja", "kolovoza", "rujna", "listopada", "studenoga", "prosinca"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod hsb_DE {
    pub mod LC_ADDRESS {
        /// `Some("DEU")`
        pub const COUNTRY_AB3: Option<&str> = Some("DEU");
        /// `Some("DE")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("DE");
        /// `Some("D")`
        pub const COUNTRY_CAR: Option<&str> = Some("D");
        /// `Some("3")`
        pub const COUNTRY_ISBN: Option<&str> = Some("3");
        /// `Some("Němska")`
        pub const COUNTRY_NAME: Option<&str> = Some("Němska");
        /// `Some(276)`
        pub const COUNTRY_NUM: Option<i64> = Some(276);
        /// `Some("D")`
        pub const COUNTRY_POST: Option<&str> = Some("D");
        /// `Some("")`
        pub const LANG_AB: Option<&str> = Some("");
        /// `Some("hsb")`
        pub const LANG_LIB: Option<&str> = Some("hsb");
        /// `Some("hornjoserbšćina")`
        pub const LANG_NAME: Option<&str> = Some("hornjoserbšćina");
        /// `Some("hsb")`
        pub const LANG_TERM: Option<&str> = Some("hsb");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Andrzej Krzysztofowicz")`
        pub const CONTACT: Option<&str> = Some("Andrzej Krzysztofowicz");
        /// `Some("2004-09-09")`
        pub const DATE: Option<&str> = Some("2004-09-09");
        /// `Some("ankry@mif.pg.gda.pl")`
        pub const EMAIL: Option<&str> = Some("ankry@mif.pg.gda.pl");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Upper Sorbian")`
        pub const LANGUAGE: Option<&str> = Some("Upper Sorbian");
        /// `Some("0.1")`
        pub const REVISION: Option<&str> = Some("0.1");
        /// `Some("Information from Michael Wolf")`
        pub const SOURCE: Option<&str> = Some("Information from Michael Wolf");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Germany")`
        pub const TERRITORY: Option<&str> = Some("Germany");
        /// `Some("Upper Sorbian locale for Germany")`
        pub const TITLE: Option<&str> = Some("Upper Sorbian locale for Germany");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("ně")`
        pub const NOSTR: Option<&str> = Some("ně");
        /// `"^[+1hHyY]"`
        pub const YESEXPR: &str = "^[+1hHyY]";
        /// `Some("haj")`
        pub const YESSTR: Option<&str> = Some("haj");
    }
    pub use super::de_DE::LC_MONETARY;
    pub use super::de_DE::LC_NUMERIC;
    pub use super::de_DE::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Nj", "Pó", "Wu", "Sr", "Št", "Pj", "So"]`
        pub const ABDAY: &[&str] = &["Nj", "Pó", "Wu", "Sr", "Št", "Pj", "So"];
        /// `&["Jan", "Feb", "Měr", "Apr", "Mej", "Jun", "Jul", "Awg", "Sep", "Okt", "Now", "Dec"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Měr", "Apr", "Mej", "Jun", "Jul", "Awg", "Sep", "Okt", "Now", "Dec"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `Some(&["Januar", "Februar", "Měrc", "Apryl", "Meja", "Junij", "Julij", "Awgust", "September", "Oktober", "Nowember", "December"])`
        pub const ALT_MON: Option<&[&str]> = Some(&["Januar", "Februar", "Měrc", "Apryl", "Meja", "Junij", "Julij", "Awgust", "September", "Oktober", "Nowember", "December"]);
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Njedźela", "Póndźela", "Wutora", "Srjeda", "Štvórtk", "Pjatk", "Sobota"]`
        pub const DAY: &[&str] = &["Njedźela", "Póndźela", "Wutora", "Srjeda", "Štvórtk", "Pjatk", "Sobota"];
        /// `"%d.%m.%Y"`
        pub const D_FMT: &str = "%d.%m.%Y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["januara", "februara", "měrca", "apryla", "meje", "junija", "julija", "awgusta", "septembra", "oktobra", "nowembra", "decembra"]`
        pub const MON: &[&str] = &["januara", "februara", "měrca", "apryla", "meje", "junija", "julija", "awgusta", "septembra", "oktobra", "nowembra", "decembra"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ht_HT {
    pub mod LC_ADDRESS {
        /// `Some("HTI")`
        pub const COUNTRY_AB3: Option<&str> = Some("HTI");
        /// `Some("HT")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("HT");
        /// `Some("RH")`
        pub const COUNTRY_CAR: Option<&str> = Some("RH");
        /// `Some("99935")`
        pub const COUNTRY_ISBN: Option<&str> = Some("99935");
        /// `Some("Ayiti")`
        pub const COUNTRY_NAME: Option<&str> = Some("Ayiti");
        /// `Some(332)`
        pub const COUNTRY_NUM: Option<i64> = Some(332);
        /// `Some("HT")`
        pub const COUNTRY_POST: Option<&str> = Some("HT");
        /// `Some("ht")`
        pub const LANG_AB: Option<&str> = Some("ht");
        /// `Some("hat")`
        pub const LANG_LIB: Option<&str> = Some("hat");
        /// `Some("Kreyòl ayisyen")`
        pub const LANG_NAME: Option<&str> = Some("Kreyòl ayisyen");
        /// `Some("hat")`
        pub const LANG_TERM: Option<&str> = Some("hat");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %R%N%%Z %t%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %R%N%%Z %t%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
        ]);
        /// `Some("sugarlabs.org")`
        pub const CONTACT: Option<&str> = Some("sugarlabs.org");
        /// `Some("2013-08-09")`
        pub const DATE: Option<&str> = Some("2013-08-09");
        /// `Some("libc-alpha@sourceware.org")`
        pub const EMAIL: Option<&str> = Some("libc-alpha@sourceware.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Haitian Creole")`
        pub const LANGUAGE: Option<&str> = Some("Haitian Creole");
        /// `Some("1.1")`
        pub const REVISION: Option<&str> = Some("1.1");
        /// `Some("Sugar Labs / OLPC")`
        pub const SOURCE: Option<&str> = Some("Sugar Labs / OLPC");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Haiti")`
        pub const TERRITORY: Option<&str> = Some("Haiti");
        /// `Some("Kreyol locale for Haiti")`
        pub const TITLE: Option<&str> = Some("Kreyol locale for Haiti");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("Non")`
        pub const NOSTR: Option<&str> = Some("Non");
        /// `"^[+1wWoOyY]"`
        pub const YESEXPR: &str = "^[+1wWoOyY]";
        /// `Some("Wi")`
        pub const YESSTR: Option<&str> = Some("Wi");
    }
    pub mod LC_MONETARY {
        /// `"g"`
        pub const CURRENCY_SYMBOL: &str = "g";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"HTG "`
        pub const INT_CURR_SYMBOL: &str = "HTG ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `"\u{202f}"`
        pub const MON_THOUSANDS_SEP: &str = "\u{202f}";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `"\u{202f}"`
        pub const THOUSANDS_SEP: &str = "\u{202f}";
    }
    pub mod LC_TELEPHONE {
        /// `Some("509")`
        pub const INT_PREFIX: Option<&str> = Some("509");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `Some("%a %l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("%a %l");
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["dim", "len", "mad", "mèk", "jed", "van", "sam"]`
        pub const ABDAY: &[&str] = &["dim", "len", "mad", "mèk", "jed", "van", "sam"];
        /// `&["jan", "fev", "mas", "avr", "me", "jen", "jiy", "out", "sep", "okt", "nov", "des"]`
        pub const ABMON: &[&str] = &["jan", "fev", "mas", "avr", "me", "jen", "jiy", "out", "sep", "okt", "nov", "des"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["dimanch", "lendi", "madi", "mèkredi", "jedi", "vandredi", "samdi"]`
        pub const DAY: &[&str] = &["dimanch", "lendi", "madi", "mèkredi", "jedi", "vandredi", "samdi"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["janvye", "fevriye", "mas", "avril", "me", "jen", "jiyè", "out", "septanm", "oktòb", "novanm", "desanm"]`
        pub const MON: &[&str] = &["janvye", "fevriye", "mas", "avril", "me", "jen", "jiyè", "out", "septanm", "oktòb", "novanm", "desanm"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod hu_HU {
    pub mod LC_ADDRESS {
        /// `Some("HUN")`
        pub const COUNTRY_AB3: Option<&str> = Some("HUN");
        /// `Some("HU")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("HU");
        /// `Some("H")`
        pub const COUNTRY_CAR: Option<&str> = Some("H");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Magyarország")`
        pub const COUNTRY_NAME: Option<&str> = Some("Magyarország");
        /// `Some(348)`
        pub const COUNTRY_NUM: Option<i64> = Some(348);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("hu")`
        pub const LANG_AB: Option<&str> = Some("hu");
        /// `Some("hun")`
        pub const LANG_LIB: Option<&str> = Some("hun");
        /// `Some("magyar")`
        pub const LANG_NAME: Option<&str> = Some("magyar");
        /// `Some("hun")`
        pub const LANG_TERM: Option<&str> = Some("hun");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jorgens Alle 8, DK-1615 Kobenhavn V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jorgens Alle 8, DK-1615 Kobenhavn V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2001-01-29")`
        pub const DATE: Option<&str> = Some("2001-01-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Hungarian")`
        pub const LANGUAGE: Option<&str> = Some("Hungarian");
        /// `Some("4.7")`
        pub const REVISION: Option<&str> = Some("4.7");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Hungary")`
        pub const TERRITORY: Option<&str> = Some("Hungary");
        /// `Some("Hungarian locale for Hungary")`
        pub const TITLE: Option<&str> = Some("Hungarian locale for Hungary");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("nem")`
        pub const NOSTR: Option<&str> = Some("nem");
        /// `"^[+1IiYy]"`
        pub const YESEXPR: &str = "^[+1IiYy]";
        /// `Some("igen")`
        pub const YESSTR: Option<&str> = Some("igen");
    }
    pub mod LC_MONETARY {
        /// `"Ft"`
        pub const CURRENCY_SYMBOL: &str = "Ft";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"HUF "`
        pub const INT_CURR_SYMBOL: &str = "HUF ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `Some(1)`
        pub const INT_N_CS_PRECEDES: Option<i64> = Some(1);
        /// `Some(1)`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = Some(1);
        /// `Some(1)`
        pub const INT_N_SIGN_POSN: Option<i64> = Some(1);
        /// `Some(1)`
        pub const INT_P_CS_PRECEDES: Option<i64> = Some(1);
        /// `Some(1)`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = Some(1);
        /// `Some(1)`
        pub const INT_P_SIGN_POSN: Option<i64> = Some(1);
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const MON_THOUSANDS_SEP: &str = ".";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const THOUSANDS_SEP: &str = ".";
    }
    pub mod LC_TELEPHONE {
        /// `Some("36")`
        pub const INT_PREFIX: Option<&str> = Some("36");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["v", "h", "k", "sze", "cs", "p", "szo"]`
        pub const ABDAY: &[&str] = &["v", "h", "k", "sze", "cs", "p", "szo"];
        /// `&["jan", "febr", "márc", "ápr", "máj", "jún", "júl", "aug", "szept", "okt", "nov", "dec"]`
        pub const ABMON: &[&str] = &["jan", "febr", "márc", "ápr", "máj", "jún", "júl", "aug", "szept", "okt", "nov", "dec"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%Y. %b. %-e., %A, %H:%M:%S %Z")`
        pub const DATE_FMT: Option<&str> = Some("%Y. %b. %-e., %A, %H:%M:%S %Z");
        /// `&["vasárnap", "hétfő", "kedd", "szerda", "csütörtök", "péntek", "szombat"]`
        pub const DAY: &[&str] = &["vasárnap", "hétfő", "kedd", "szerda", "csütörtök", "péntek", "szombat"];
        /// `"%Y-%m-%d"`
        pub const D_FMT: &str = "%Y-%m-%d";
        /// `"%Y. %b. %-e., %A, %H:%M:%S %Z"`
        pub const D_T_FMT: &str = "%Y. %b. %-e., %A, %H:%M:%S %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["január", "február", "március", "április", "május", "június", "július", "augusztus", "szeptember", "október", "november", "december"]`
        pub const MON: &[&str] = &["január", "február", "március", "április", "május", "június", "július", "augusztus", "szeptember", "október", "november", "december"];
        /// `"%H:%M:%S"`
        pub const T_FMT: &str = "%H:%M:%S";
        /// `"%H:%M:%S"`
        pub const T_FMT_AMPM: &str = "%H:%M:%S";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod hy_AM {
    pub mod LC_ADDRESS {
        /// `Some("ARM")`
        pub const COUNTRY_AB3: Option<&str> = Some("ARM");
        /// `Some("AM")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("AM");
        /// `Some("AM")`
        pub const COUNTRY_CAR: Option<&str> = Some("AM");
        /// `Some("978-9939,978-99930,978-99941")`
        pub const COUNTRY_ISBN: Option<&str> = Some("978-9939,978-99930,978-99941");
        /// `Some("Հայաստան")`
        pub const COUNTRY_NAME: Option<&str> = Some("Հայաստան");
        /// `Some(51)`
        pub const COUNTRY_NUM: Option<i64> = Some(51);
        /// `Some("AM")`
        pub const COUNTRY_POST: Option<&str> = Some("AM");
        /// `Some("hy")`
        pub const LANG_AB: Option<&str> = Some("hy");
        /// `Some("arm")`
        pub const LANG_LIB: Option<&str> = Some("arm");
        /// `Some("հայերեն")`
        pub const LANG_NAME: Option<&str> = Some("հայերեն");
        /// `Some("hye")`
        pub const LANG_TERM: Option<&str> = Some("hye");
        /// `Some("%d%N%f%N%d%N%b%N%s %h 5e %r%N%C%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%d%N%f%N%d%N%b%N%s %h 5e %r%N%C%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_ADDRESS"],
        ]);
        /// `Some("Pablo Saratxaga")`
        pub const CONTACT: Option<&str> = Some("Pablo Saratxaga");
        /// `Some("2001-01-26")`
        pub const DATE: Option<&str> = Some("2001-01-26");
        /// `Some("pablo@mandrakesoft.com")`
        pub const EMAIL: Option<&str> = Some("pablo@mandrakesoft.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Armenian")`
        pub const LANGUAGE: Option<&str> = Some("Armenian");
        /// `Some("0.4")`
        pub const REVISION: Option<&str> = Some("0.4");
        /// `Some("http://www.freenet.am/armscii")`
        pub const SOURCE: Option<&str> = Some("http://www.freenet.am/armscii");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Armenia")`
        pub const TERRITORY: Option<&str> = Some("Armenia");
        /// `Some("Armenian language locale for Armenia")`
        pub const TITLE: Option<&str> = Some("Armenian language locale for Armenia");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNոՈ]"`
        pub const NOEXPR: &str = "^[-0nNոՈ]";
        /// `Some("ոչ")`
        pub const NOSTR: Option<&str> = Some("ոչ");
        /// `"^[+1yYsSաԱ]"`
        pub const YESEXPR: &str = "^[+1yYsSաԱ]";
        /// `Some("այո")`
        pub const YESSTR: Option<&str> = Some("այո");
    }
    pub mod LC_MONETARY {
        /// `"֏"`
        pub const CURRENCY_SYMBOL: &str = "֏";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"AMD "`
        pub const INT_CURR_SYMBOL: &str = "AMD ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("374")`
        pub const INT_PREFIX: Option<&str> = Some("374");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["Կրկ", "Երկ", "Երք", "Չրք", "Հնգ", "Ուր", "Շբթ"]`
        pub const ABDAY: &[&str] = &["Կրկ", "Երկ", "Երք", "Չրք", "Հնգ", "Ուր", "Շբթ"];
        /// `&["Հնվ", "Փտր", "Մար", "Ապր", "Մայ", "Հնս", "Հլս", "Օգս", "Սեպ", "Հոկ", "Նմբ", "Դեկ"]`
        pub const ABMON: &[&str] = &["Հնվ", "Փտր", "Մար", "Ապր", "Մայ", "Հնս", "Հլս", "Օգս", "Սեպ", "Հոկ", "Նմբ", "Դեկ"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `Some(&["հունվար", "փետրվար", "մարտ", "ապրիլ", "մայիս", "հունիս", "հուլիս", "օգոստոս", "սեպտեմբեր", "հոկտեմբեր", "նոյեմբեր", "դեկտեմբեր"])`
        pub const ALT_MON: Option<&[&str]> = Some(&["հունվար", "փետրվար", "մարտ", "ապրիլ", "մայիս", "հունիս", "հուլիս", "օգոստոս", "սեպտեմբեր", "հոկտեմբեր", "նոյեմբեր", "դեկտեմբեր"]);
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Կիրակի", "Երկուշաբթի", "Երեքշաբթի", "Չորեքշաբթի", "Հինգշաբթի", "Ուրբաթ", "Շաբաթ"]`
        pub const DAY: &[&str] = &["Կիրակի", "Երկուշաբթի", "Երեքշաբթի", "Չորեքշաբթի", "Հինգշաբթի", "Ուրբաթ", "Շաբաթ"];
        /// `"%m/%d/%y"`
        pub const D_FMT: &str = "%m/%d/%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["հունվարի", "փետրվարի", "մարտի", "ապրիլի", "մայիսի", "հունիսի", "հուլիսի", "օգոստոսի", "սեպտեմբերի", "հոկտեմբերի", "նոյեմբերի", "դեկտեմբերի"]`
        pub const MON: &[&str] = &["հունվարի", "փետրվարի", "մարտի", "ապրիլի", "մայիսի", "հունիսի", "հուլիսի", "օգոստոսի", "սեպտեմբերի", "հոկտեմբերի", "նոյեմբերի", "դեկտեմբերի"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ia_FR {
    pub mod LC_ADDRESS {
        /// `Some("FRA")`
        pub const COUNTRY_AB3: Option<&str> = Some("FRA");
        /// `Some("FR")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("FR");
        /// `Some("F")`
        pub const COUNTRY_CAR: Option<&str> = Some("F");
        /// `Some("979-10")`
        pub const COUNTRY_ISBN: Option<&str> = Some("979-10");
        /// `Some("Francia")`
        pub const COUNTRY_NAME: Option<&str> = Some("Francia");
        /// `Some(250)`
        pub const COUNTRY_NUM: Option<i64> = Some(250);
        /// `Some("F")`
        pub const COUNTRY_POST: Option<&str> = Some("F");
        /// `Some("ia")`
        pub const LANG_AB: Option<&str> = Some("ia");
        /// `Some("ina")`
        pub const LANG_LIB: Option<&str> = Some("ina");
        /// `Some("interlingua")`
        pub const LANG_NAME: Option<&str> = Some("interlingua");
        /// `Some("ina")`
        pub const LANG_TERM: Option<&str> = Some("ina");
        /// `Some("%f%N%a%N%d%N%b%N%s%t%h%t%e%t%r%N%z%t%T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s%t%h%t%e%t%r%N%z%t%T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
        ]);
        /// `Some("Nik Kalach")`
        pub const CONTACT: Option<&str> = Some("Nik Kalach");
        /// `Some("2013-03-01")`
        pub const DATE: Option<&str> = Some("2013-03-01");
        /// `Some("nikka@fedoraproject.org")`
        pub const EMAIL: Option<&str> = Some("nikka@fedoraproject.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Interlingua")`
        pub const LANGUAGE: Option<&str> = Some("Interlingua");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Fedora Project")`
        pub const SOURCE: Option<&str> = Some("Fedora Project");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("France")`
        pub const TERRITORY: Option<&str> = Some("France");
        /// `Some("Interlingua locale for France")`
        pub const TITLE: Option<&str> = Some("Interlingua locale for France");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("No")`
        pub const NOSTR: Option<&str> = Some("No");
        /// `"^[+1sSyY]"`
        pub const YESEXPR: &str = "^[+1sSyY]";
        /// `Some("Si")`
        pub const YESSTR: Option<&str> = Some("Si");
    }
    pub use super::fr_FR::LC_MONETARY;
    pub use super::fr_FR::LC_NUMERIC;
    pub use super::fr_FR::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["dom", "lun", "mar", "mer", "jov", "ven", "sab"]`
        pub const ABDAY: &[&str] = &["dom", "lun", "mar", "mer", "jov", "ven", "sab"];
        /// `&["jan", "feb", "mar", "apr", "mai", "jun", "jul", "aug", "sep", "oct", "nov", "dec"]`
        pub const ABMON: &[&str] = &["jan", "feb", "mar", "apr", "mai", "jun", "jul", "aug", "sep", "oct", "nov", "dec"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["AM", "PM"]`
        pub const AM_PM: &[&str] = &["AM", "PM"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["dominica", "lunedi", "martedi", "mercuridi", "jovedi", "venerdi", "sabbato"]`
        pub const DAY: &[&str] = &["dominica", "lunedi", "martedi", "mercuridi", "jovedi", "venerdi", "sabbato"];
        /// `"%d.%m.%Y"`
        pub const D_FMT: &str = "%d.%m.%Y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["januario", "februario", "martio", "april", "maio", "junio", "julio", "augusto", "septembre", "octobre", "novembre", "decembre"]`
        pub const MON: &[&str] = &["januario", "februario", "martio", "april", "maio", "junio", "julio", "augusto", "septembre", "octobre", "novembre", "decembre"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `"%l:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%l:%M:%S %p";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod id_ID {
    pub mod LC_ADDRESS {
        /// `Some("IDN")`
        pub const COUNTRY_AB3: Option<&str> = Some("IDN");
        /// `Some("ID")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("ID");
        /// `Some("RI")`
        pub const COUNTRY_CAR: Option<&str> = Some("RI");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Indonesia")`
        pub const COUNTRY_NAME: Option<&str> = Some("Indonesia");
        /// `Some(360)`
        pub const COUNTRY_NUM: Option<i64> = Some(360);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("id")`
        pub const LANG_AB: Option<&str> = Some("id");
        /// `Some("ind")`
        pub const LANG_LIB: Option<&str> = Some("ind");
        /// `Some("Indonesia")`
        pub const LANG_NAME: Option<&str> = Some("Indonesia");
        /// `Some("ind")`
        pub const LANG_TERM: Option<&str> = Some("ind");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Indonesian")`
        pub const LANGUAGE: Option<&str> = Some("Indonesian");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Indonesia")`
        pub const TERRITORY: Option<&str> = Some("Indonesia");
        /// `Some("Indonesian locale for Indonesia")`
        pub const TITLE: Option<&str> = Some("Indonesian locale for Indonesia");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0tTnN]"`
        pub const NOEXPR: &str = "^[-0tTnN]";
        /// `Some("tidak")`
        pub const NOSTR: Option<&str> = Some("tidak");
        /// `"^[+1yY]"`
        pub const YESEXPR: &str = "^[+1yY]";
        /// `Some("ya")`
        pub const YESSTR: Option<&str> = Some("ya");
    }
    pub mod LC_MONETARY {
        /// `"Rp"`
        pub const CURRENCY_SYMBOL: &str = "Rp";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"IDR "`
        pub const INT_CURR_SYMBOL: &str = "IDR ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const MON_THOUSANDS_SEP: &str = ".";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const THOUSANDS_SEP: &str = ".";
    }
    pub mod LC_TELEPHONE {
        /// `Some("62")`
        pub const INT_PREFIX: Option<&str> = Some("62");
        /// `Some("001,008,007,009")`
        pub const INT_SELECT: Option<&str> = Some("001,008,007,009");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["Min", "Sen", "Sel", "Rab", "Kam", "Jum", "Sab"]`
        pub const ABDAY: &[&str] = &["Min", "Sen", "Sel", "Rab", "Kam", "Jum", "Sab"];
        /// `&["Jan", "Feb", "Mar", "Apr", "Mei", "Jun", "Jul", "Agu", "Sep", "Okt", "Nov", "Des"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Mar", "Apr", "Mei", "Jun", "Jul", "Agu", "Sep", "Okt", "Nov", "Des"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Minggu", "Senin", "Selasa", "Rabu", "Kamis", "Jumat", "Sabtu"]`
        pub const DAY: &[&str] = &["Minggu", "Senin", "Selasa", "Rabu", "Kamis", "Jumat", "Sabtu"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Januari", "Februari", "Maret", "April", "Mei", "Juni", "Juli", "Agustus", "September", "Oktober", "November", "Desember"]`
        pub const MON: &[&str] = &["Januari", "Februari", "Maret", "April", "Mei", "Juni", "Juli", "Agustus", "September", "Oktober", "November", "Desember"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ig_NG {
    pub mod LC_ADDRESS {
        /// `Some("NGA")`
        pub const COUNTRY_AB3: Option<&str> = Some("NGA");
        /// `Some("NG")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("NG");
        /// `Some("WAN")`
        pub const COUNTRY_CAR: Option<&str> = Some("WAN");
        /// `Some("978")`
        pub const COUNTRY_ISBN: Option<&str> = Some("978");
        /// `Some("Naịjịrịa")`
        pub const COUNTRY_NAME: Option<&str> = Some("Naịjịrịa");
        /// `Some(566)`
        pub const COUNTRY_NUM: Option<i64> = Some(566);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ig")`
        pub const LANG_AB: Option<&str> = Some("ig");
        /// `Some("ibo")`
        pub const LANG_LIB: Option<&str> = Some("ibo");
        /// `Some("Asụsụ Igbo")`
        pub const LANG_NAME: Option<&str> = Some("Asụsụ Igbo");
        /// `Some("ibo")`
        pub const LANG_TERM: Option<&str> = Some("ibo");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2005-12-14")`
        pub const DATE: Option<&str> = Some("2005-12-14");
        /// `Some("pablo@mandriva.com")`
        pub const EMAIL: Option<&str> = Some("pablo@mandriva.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Igbo")`
        pub const LANGUAGE: Option<&str> = Some("Igbo");
        /// `Some("0.2")`
        pub const REVISION: Option<&str> = Some("0.2");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Nigeria")`
        pub const TERRITORY: Option<&str> = Some("Nigeria");
        /// `Some("Igbo locale for Nigeria")`
        pub const TITLE: Option<&str> = Some("Igbo locale for Nigeria");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0MmNn]"`
        pub const NOEXPR: &str = "^[-0MmNn]";
        /// `Some("Mba")`
        pub const NOSTR: Option<&str> = Some("Mba");
        /// `"^[+1EeIiYy]"`
        pub const YESEXPR: &str = "^[+1EeIiYy]";
        /// `Some("Eye")`
        pub const YESSTR: Option<&str> = Some("Eye");
    }
    pub use super::en_NG::LC_MONETARY;
    pub use super::en_NG::LC_NUMERIC;
    pub use super::en_NG::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["sọn", "mọn", "tuz", "wen", "tọs", "fra", "sat"]`
        pub const ABDAY: &[&str] = &["sọn", "mọn", "tuz", "wen", "tọs", "fra", "sat"];
        /// `&["Jen", "Feb", "Maa", "Epr", "Mee", "Juu", "Jul", "Ọgọ", "Sep", "Ọkt", "Nov", "Dis"]`
        pub const ABMON: &[&str] = &["Jen", "Feb", "Maa", "Epr", "Mee", "Juu", "Jul", "Ọgọ", "Sep", "Ọkt", "Nov", "Dis"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["AM", "PM"]`
        pub const AM_PM: &[&str] = &["AM", "PM"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["sọnde", "mọnde", "tuzde", "wenzde", "tọsde", "fraịde", "satọde"]`
        pub const DAY: &[&str] = &["sọnde", "mọnde", "tuzde", "wenzde", "tọsde", "fraịde", "satọde"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%A, %d %B %Y %T %Z"`
        pub const D_T_FMT: &str = "%A, %d %B %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Jenụwarị", "Febrụwarị", "Maachị", "Eprel", "Mee", "Juun", "Julaị", "Ọgọọst", "Septemba", "Ọktoba", "Novemba", "Disemba"]`
        pub const MON: &[&str] = &["Jenụwarị", "Febrụwarị", "Maachị", "Eprel", "Mee", "Juun", "Julaị", "Ọgọọst", "Septemba", "Ọktoba", "Novemba", "Disemba"];
        /// `"%I:%M:%S %p"`
        pub const T_FMT: &str = "%I:%M:%S %p";
        /// `"%I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ik_CA {
    pub mod LC_ADDRESS {
        /// `Some("CAN")`
        pub const COUNTRY_AB3: Option<&str> = Some("CAN");
        /// `Some("CA")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("CA");
        /// `Some("CDN")`
        pub const COUNTRY_CAR: Option<&str> = Some("CDN");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Kanada")`
        pub const COUNTRY_NAME: Option<&str> = Some("Kanada");
        /// `Some(124)`
        pub const COUNTRY_NUM: Option<i64> = Some(124);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ik")`
        pub const LANG_AB: Option<&str> = Some("ik");
        /// `Some("ipk")`
        pub const LANG_LIB: Option<&str> = Some("ipk");
        /// `Some("Iñupiatun")`
        pub const LANG_NAME: Option<&str> = Some("Iñupiatun");
        /// `Some("ipk")`
        pub const LANG_TERM: Option<&str> = Some("ipk");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2004-08-01")`
        pub const DATE: Option<&str> = Some("2004-08-01");
        /// `Some("pablo@mandriva.com")`
        pub const EMAIL: Option<&str> = Some("pablo@mandriva.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Inupiaq")`
        pub const LANGUAGE: Option<&str> = Some("Inupiaq");
        /// `Some("0.2")`
        pub const REVISION: Option<&str> = Some("0.2");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Canada")`
        pub const TERRITORY: Option<&str> = Some("Canada");
        /// `Some("Inupiaq locale for Canada")`
        pub const TITLE: Option<&str> = Some("Inupiaq locale for Canada");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNqQ]"`
        pub const NOEXPR: &str = "^[-0nNqQ]";
        /// `Some("qañaa")`
        pub const NOSTR: Option<&str> = Some("qañaa");
        /// `"^[+1yYiIaA]"`
        pub const YESEXPR: &str = "^[+1yYiIaA]";
        /// `Some("aa")`
        pub const YESSTR: Option<&str> = Some("aa");
    }
    pub use super::en_CA::LC_MONETARY;
    pub use super::en_CA::LC_NUMERIC;
    pub use super::en_CA::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Min", "Sav", "Ila", "Qit", "Sis", "Tal", "Maq"]`
        pub const ABDAY: &[&str] = &["Min", "Sav", "Ila", "Qit", "Sis", "Tal", "Maq"];
        /// `&["Sñt", "Sñs", "Pan", "Qil", "Sup", "Iġñ", "Itc", "Tiñ", "Ami", "Sik", "Nip", "Siq"]`
        pub const ABMON: &[&str] = &["Sñt", "Sñs", "Pan", "Qil", "Sup", "Iġñ", "Itc", "Tiñ", "Ami", "Sik", "Nip", "Siq"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["AM", "PM"]`
        pub const AM_PM: &[&str] = &["AM", "PM"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Minġuiqsioiq", "Savałłiq", "Ilaqtchiioiq", "Qitchiioiq", "Sisamiioiq", "Tallimmiioiq", "Maqinġuoiq"]`
        pub const DAY: &[&str] = &["Minġuiqsioiq", "Savałłiq", "Ilaqtchiioiq", "Qitchiioiq", "Sisamiioiq", "Tallimmiioiq", "Maqinġuoiq"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%a %d %b %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Siqiññaatchiaq", "Siqiññaasrugruk", "Paniqsiqsiivik", "Qilġich Tatqiat", "Suppivik", "Iġñivik", "Itchavik", "Tiññivik", "Amiġaiqsivik", "Sikkuvik", "Nippivik", "Siqiñġiḷaq"]`
        pub const MON: &[&str] = &["Siqiññaatchiaq", "Siqiññaasrugruk", "Paniqsiqsiivik", "Qilġich Tatqiat", "Suppivik", "Iġñivik", "Itchavik", "Tiññivik", "Amiġaiqsivik", "Sikkuvik", "Nippivik", "Siqiñġiḷaq"];
        /// `"%I:%M:%S %p"`
        pub const T_FMT: &str = "%I:%M:%S %p";
        /// `"%I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod is_IS {
    pub mod LC_ADDRESS {
        /// `Some("ISL")`
        pub const COUNTRY_AB3: Option<&str> = Some("ISL");
        /// `Some("IS")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IS");
        /// `Some("IS")`
        pub const COUNTRY_CAR: Option<&str> = Some("IS");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Ísland")`
        pub const COUNTRY_NAME: Option<&str> = Some("Ísland");
        /// `Some(352)`
        pub const COUNTRY_NUM: Option<i64> = Some(352);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("is")`
        pub const LANG_AB: Option<&str> = Some("is");
        /// `Some("ice")`
        pub const LANG_LIB: Option<&str> = Some("ice");
        /// `Some("íslenska")`
        pub const LANG_NAME: Option<&str> = Some("íslenska");
        /// `Some("isl")`
        pub const LANG_TERM: Option<&str> = Some("isl");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Keldnaholt-ITI', IS-112 Reykjavi'k, Iceland")`
        pub const ADDRESS: Option<&str> = Some("Keldnaholt-ITI', IS-112 Reykjavi'k, Iceland");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Icelandic")`
        pub const LANGUAGE: Option<&str> = Some("Icelandic");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Stadlarad I'slands")`
        pub const SOURCE: Option<&str> = Some("Stadlarad I'slands");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Iceland")`
        pub const TERRITORY: Option<&str> = Some("Iceland");
        /// `Some("Icelandic locale for Iceland")`
        pub const TITLE: Option<&str> = Some("Icelandic locale for Iceland");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("nei")`
        pub const NOSTR: Option<&str> = Some("nei");
        /// `"^[+1jJyY]"`
        pub const YESEXPR: &str = "^[+1jJyY]";
        /// `Some("já")`
        pub const YESSTR: Option<&str> = Some("já");
    }
    pub mod LC_MONETARY {
        /// `"kr"`
        pub const CURRENCY_SYMBOL: &str = "kr";
        /// `0`
        pub const FRAC_DIGITS: i64 = 0;
        /// `"ISK "`
        pub const INT_CURR_SYMBOL: &str = "ISK ";
        /// `0`
        pub const INT_FRAC_DIGITS: i64 = 0;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const MON_THOUSANDS_SEP: &str = ".";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const THOUSANDS_SEP: &str = ".";
    }
    pub mod LC_TELEPHONE {
        /// `Some("354")`
        pub const INT_PREFIX: Option<&str> = Some("354");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["sun", "mán", "þri", "mið", "fim", "fös", "lau"]`
        pub const ABDAY: &[&str] = &["sun", "mán", "þri", "mið", "fim", "fös", "lau"];
        /// `&["jan", "feb", "mar", "apr", "maí", "jún", "júl", "ágú", "sep", "okt", "nóv", "des"]`
        pub const ABMON: &[&str] = &["jan", "feb", "mar", "apr", "maí", "jún", "júl", "ágú", "sep", "okt", "nóv", "des"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["fh", "eh"]`
        pub const AM_PM: &[&str] = &["fh", "eh"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["sunnudagur", "mánudagur", "þriðjudagur", "miðvikudagur", "fimmtudagur", "föstudagur", "laugardagur"]`
        pub const DAY: &[&str] = &["sunnudagur", "mánudagur", "þriðjudagur", "miðvikudagur", "fimmtudagur", "föstudagur", "laugardagur"];
        /// `"%a %e.%b %Y"`
        pub const D_FMT: &str = "%a %e.%b %Y";
        /// `"%a %e.%b %Y, %T %Z"`
        pub const D_T_FMT: &str = "%a %e.%b %Y, %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["janúar", "febrúar", "mars", "apríl", "maí", "júní", "júlí", "ágúst", "september", "október", "nóvember", "desember"]`
        pub const MON: &[&str] = &["janúar", "febrúar", "mars", "apríl", "maí", "júní", "júlí", "ágúst", "september", "október", "nóvember", "desember"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `"%l:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%l:%M:%S %p";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod it_CH {
    pub mod LC_ADDRESS {
        /// `Some("CHE")`
        pub const COUNTRY_AB3: Option<&str> = Some("CHE");
        /// `Some("CH")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("CH");
        /// `Some("CH")`
        pub const COUNTRY_CAR: Option<&str> = Some("CH");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Svizzera")`
        pub const COUNTRY_NAME: Option<&str> = Some("Svizzera");
        /// `Some(756)`
        pub const COUNTRY_NUM: Option<i64> = Some(756);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("it")`
        pub const LANG_AB: Option<&str> = Some("it");
        /// `Some("ita")`
        pub const LANG_LIB: Option<&str> = Some("ita");
        /// `Some("italiano")`
        pub const LANG_NAME: Option<&str> = Some("italiano");
        /// `Some("ita")`
        pub const LANG_TERM: Option<&str> = Some("ita");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Italian")`
        pub const LANGUAGE: Option<&str> = Some("Italian");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Switzerland")`
        pub const TERRITORY: Option<&str> = Some("Switzerland");
        /// `Some("Italian locale for Switzerland")`
        pub const TITLE: Option<&str> = Some("Italian locale for Switzerland");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("no")`
        pub const NOSTR: Option<&str> = Some("no");
        /// `"^[+1sSjJoOyY]"`
        pub const YESEXPR: &str = "^[+1sSjJoOyY]";
        /// `Some("sì")`
        pub const YESSTR: Option<&str> = Some("sì");
    }
    pub use super::de_CH::LC_MONETARY;
    pub use super::de_CH::LC_NUMERIC;
    pub mod LC_TELEPHONE {
        /// `Some("41")`
        pub const INT_PREFIX: Option<&str> = Some("41");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["dom", "lun", "mar", "mer", "gio", "ven", "sab"]`
        pub const ABDAY: &[&str] = &["dom", "lun", "mar", "mer", "gio", "ven", "sab"];
        /// `&["gen", "feb", "mar", "apr", "mag", "giu", "lug", "ago", "set", "ott", "nov", "dic"]`
        pub const ABMON: &[&str] = &["gen", "feb", "mar", "apr", "mag", "giu", "lug", "ago", "set", "ott", "nov", "dic"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a %-d %b %Y, %T, %Z")`
        pub const DATE_FMT: Option<&str> = Some("%a %-d %b %Y, %T, %Z");
        /// `&["domenica", "lunedì", "martedì", "mercoledì", "giovedì", "venerdì", "sabato"]`
        pub const DAY: &[&str] = &["domenica", "lunedì", "martedì", "mercoledì", "giovedì", "venerdì", "sabato"];
        /// `"%d.%m.%Y"`
        pub const D_FMT: &str = "%d.%m.%Y";
        /// `"%a %-d %b %Y, %T"`
        pub const D_T_FMT: &str = "%a %-d %b %Y, %T";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["gennaio", "febbraio", "marzo", "aprile", "maggio", "giugno", "luglio", "agosto", "settembre", "ottobre", "novembre", "dicembre"]`
        pub const MON: &[&str] = &["gennaio", "febbraio", "marzo", "aprile", "maggio", "giugno", "luglio", "agosto", "settembre", "ottobre", "novembre", "dicembre"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod it_IT {
    pub mod LC_ADDRESS {
        /// `Some("ITA")`
        pub const COUNTRY_AB3: Option<&str> = Some("ITA");
        /// `Some("IT")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IT");
        /// `Some("I")`
        pub const COUNTRY_CAR: Option<&str> = Some("I");
        /// `Some("978-88,979-12")`
        pub const COUNTRY_ISBN: Option<&str> = Some("978-88,979-12");
        /// `Some("Italia")`
        pub const COUNTRY_NAME: Option<&str> = Some("Italia");
        /// `Some(380)`
        pub const COUNTRY_NUM: Option<i64> = Some(380);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("it")`
        pub const LANG_AB: Option<&str> = Some("it");
        /// `Some("ita")`
        pub const LANG_LIB: Option<&str> = Some("ita");
        /// `Some("italiano")`
        pub const LANG_NAME: Option<&str> = Some("italiano");
        /// `Some("ita")`
        pub const LANG_TERM: Option<&str> = Some("ita");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Italian")`
        pub const LANGUAGE: Option<&str> = Some("Italian");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Italy")`
        pub const TERRITORY: Option<&str> = Some("Italy");
        /// `Some("Italian locale for Italy")`
        pub const TITLE: Option<&str> = Some("Italian locale for Italy");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("no")`
        pub const NOSTR: Option<&str> = Some("no");
        /// `"^[+1sSyY]"`
        pub const YESEXPR: &str = "^[+1sSyY]";
        /// `Some("sì")`
        pub const YESSTR: Option<&str> = Some("sì");
    }
    pub mod LC_MONETARY {
        /// `"€"`
        pub const CURRENCY_SYMBOL: &str = "€";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"EUR "`
        pub const INT_CURR_SYMBOL: &str = "EUR ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const MON_THOUSANDS_SEP: &str = ".";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const THOUSANDS_SEP: &str = ".";
    }
    pub mod LC_TELEPHONE {
        /// `Some("39")`
        pub const INT_PREFIX: Option<&str> = Some("39");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["dom", "lun", "mar", "mer", "gio", "ven", "sab"]`
        pub const ABDAY: &[&str] = &["dom", "lun", "mar", "mer", "gio", "ven", "sab"];
        /// `&["gen", "feb", "mar", "apr", "mag", "giu", "lug", "ago", "set", "ott", "nov", "dic"]`
        pub const ABMON: &[&str] = &["gen", "feb", "mar", "apr", "mag", "giu", "lug", "ago", "set", "ott", "nov", "dic"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a %-d %b %Y, %T, %Z")`
        pub const DATE_FMT: Option<&str> = Some("%a %-d %b %Y, %T, %Z");
        /// `&["domenica", "lunedì", "martedì", "mercoledì", "giovedì", "venerdì", "sabato"]`
        pub const DAY: &[&str] = &["domenica", "lunedì", "martedì", "mercoledì", "giovedì", "venerdì", "sabato"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%a %-d %b %Y, %T"`
        pub const D_T_FMT: &str = "%a %-d %b %Y, %T";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["gennaio", "febbraio", "marzo", "aprile", "maggio", "giugno", "luglio", "agosto", "settembre", "ottobre", "novembre", "dicembre"]`
        pub const MON: &[&str] = &["gennaio", "febbraio", "marzo", "aprile", "maggio", "giugno", "luglio", "agosto", "settembre", "ottobre", "novembre", "dicembre"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod it_IT_euro {
    pub use super::it_IT::LC_ADDRESS;
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("https://www.gnu.org/software/libc/")`
        pub const ADDRESS: Option<&str> = Some("https://www.gnu.org/software/libc/");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-08-20")`
        pub const DATE: Option<&str> = Some("2000-08-20");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Italian")`
        pub const LANGUAGE: Option<&str> = Some("Italian");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Free Software Foundation, Inc.")`
        pub const SOURCE: Option<&str> = Some("Free Software Foundation, Inc.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Italy")`
        pub const TERRITORY: Option<&str> = Some("Italy");
        /// `Some("Italian locale for Italy with Euro")`
        pub const TITLE: Option<&str> = Some("Italian locale for Italy with Euro");
    }
    pub use super::it_IT::LC_MESSAGES;
    pub use super::it_IT::LC_MONETARY;
    pub use super::it_IT::LC_NUMERIC;
    pub use super::it_IT::LC_TELEPHONE;
    pub use super::it_IT::LC_TIME;
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod iu_CA {
    pub mod LC_ADDRESS {
        /// `Some("CAN")`
        pub const COUNTRY_AB3: Option<&str> = Some("CAN");
        /// `Some("CA")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("CA");
        /// `Some("CDN")`
        pub const COUNTRY_CAR: Option<&str> = Some("CDN");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("ᑲᓇᑕ")`
        pub const COUNTRY_NAME: Option<&str> = Some("ᑲᓇᑕ");
        /// `Some(124)`
        pub const COUNTRY_NUM: Option<i64> = Some(124);
        /// `Some("CA")`
        pub const COUNTRY_POST: Option<&str> = Some("CA");
        /// `Some("iu")`
        pub const LANG_AB: Option<&str> = Some("iu");
        /// `Some("iku")`
        pub const LANG_LIB: Option<&str> = Some("iku");
        /// `Some("ᐃᓄᒃᑎᑐᑦ")`
        pub const LANG_NAME: Option<&str> = Some("ᐃᓄᒃᑎᑐᑦ");
        /// `Some("iku")`
        pub const LANG_TERM: Option<&str> = Some("iku");
        /// `Some("%a%N%f%N%d%N%b%N%h %s %e %r%N%T %z%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%a%N%f%N%d%N%b%N%h %s %e %r%N%T %z%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_ADDRESS"],
        ]);
        /// `Some("Pablo Saratxaga")`
        pub const CONTACT: Option<&str> = Some("Pablo Saratxaga");
        /// `Some("2001-05-04")`
        pub const DATE: Option<&str> = Some("2001-05-04");
        /// `Some("pablo@mandriva.com")`
        pub const EMAIL: Option<&str> = Some("pablo@mandriva.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Inuktitut")`
        pub const LANGUAGE: Option<&str> = Some("Inuktitut");
        /// `Some("0.1")`
        pub const REVISION: Option<&str> = Some("0.1");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Canada")`
        pub const TERRITORY: Option<&str> = Some("Canada");
        /// `Some("Inuktitut language locale for Nunavut, Canada")`
        pub const TITLE: Option<&str> = Some("Inuktitut language locale for Nunavut, Canada");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `None`
        pub const NOSTR: Option<&str> = None;
        /// `"^[+1yYsS]"`
        pub const YESEXPR: &str = "^[+1yYsS]";
        /// `None`
        pub const YESSTR: Option<&str> = None;
    }
    pub use super::en_CA::LC_MONETARY;
    pub use super::en_CA::LC_NUMERIC;
    pub use super::en_CA::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["ᓈ", "ᓇ", "ᓕ", "ᐱ", "ᕿ", "ᐅ", "ᓯ"]`
        pub const ABDAY: &[&str] = &["ᓈ", "ᓇ", "ᓕ", "ᐱ", "ᕿ", "ᐅ", "ᓯ"];
        /// `&["ᔮᓄ", "ᕕᕗ", "ᒪᔅ", "ᐃᐳ", "ᒪᐃ", "ᔪᓂ", "ᔪᓚ", "ᐊᒋ", "ᓯᑎ", "ᐊᑦ", "ᓄᕕ", "ᑎᓯ"]`
        pub const ABMON: &[&str] = &["ᔮᓄ", "ᕕᕗ", "ᒪᔅ", "ᐃᐳ", "ᒪᐃ", "ᔪᓂ", "ᔪᓚ", "ᐊᒋ", "ᓯᑎ", "ᐊᑦ", "ᓄᕕ", "ᑎᓯ"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["AM", "PM"]`
        pub const AM_PM: &[&str] = &["AM", "PM"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["ᓈᑦᑎᖑᔭᕐᕕᒃ", "ᓇᒡᒐᔾᔭᐅ", "ᓇᒡᒐᔾᔭᐅᓕᖅᑭᑦ", "ᐱᖓᓲᓕᖅᓯᐅᑦ", "ᕿᑎᖅᑰᑦ", "ᐅᓪᓗᕈᓘᑐᐃᓇᖅ", "ᓯᕙᑖᕕᒃ"]`
        pub const DAY: &[&str] = &["ᓈᑦᑎᖑᔭᕐᕕᒃ", "ᓇᒡᒐᔾᔭᐅ", "ᓇᒡᒐᔾᔭᐅᓕᖅᑭᑦ", "ᐱᖓᓲᓕᖅᓯᐅᑦ", "ᕿᑎᖅᑰᑦ", "ᐅᓪᓗᕈᓘᑐᐃᓇᖅ", "ᓯᕙᑖᕕᒃ"];
        /// `"%m/%d/%y"`
        pub const D_FMT: &str = "%m/%d/%y";
        /// `"%a %d %b %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["ᔮᓄᐊᓕ", "ᕕᕗᐊᓕ", "ᒪᔅᓯ", "ᐃᐳᓗ", "ᒪᐃ", "ᔪᓂ", "ᔪᓚᐃ", "ᐊᒋᓯ", "ᓯᑎᕙ", "ᐊᑦᑐᕙ", "ᓄᕕᕙ", "ᑎᓯᕝᕙ"]`
        pub const MON: &[&str] = &["ᔮᓄᐊᓕ", "ᕕᕗᐊᓕ", "ᒪᔅᓯ", "ᐃᐳᓗ", "ᒪᐃ", "ᔪᓂ", "ᔪᓚᐃ", "ᐊᒋᓯ", "ᓯᑎᕙ", "ᐊᑦᑐᕙ", "ᓄᕕᕙ", "ᑎᓯᕝᕙ"];
        /// `"%I:%M:%S %p"`
        pub const T_FMT: &str = "%I:%M:%S %p";
        /// `"%I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ja_JP {
    pub mod LC_ADDRESS {
        /// `Some("JPN")`
        pub const COUNTRY_AB3: Option<&str> = Some("JPN");
        /// `Some("JP")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("JP");
        /// `Some("J")`
        pub const COUNTRY_CAR: Option<&str> = Some("J");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("日本")`
        pub const COUNTRY_NAME: Option<&str> = Some("日本");
        /// `Some(392)`
        pub const COUNTRY_NUM: Option<i64> = Some(392);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ja")`
        pub const LANG_AB: Option<&str> = Some("ja");
        /// `Some("jpn")`
        pub const LANG_LIB: Option<&str> = Some("jpn");
        /// `Some("日本語")`
        pub const LANG_NAME: Option<&str> = Some("日本語");
        /// `Some("jpn")`
        pub const LANG_TERM: Option<&str> = Some("jpn");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-07-20")`
        pub const DATE: Option<&str> = Some("2000-07-20");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Japanese")`
        pub const LANGUAGE: Option<&str> = Some("Japanese");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("HANATAKA, Shinya, hanataka@abyss.rim.or.jp")`
        pub const SOURCE: Option<&str> = Some("HANATAKA, Shinya, hanataka@abyss.rim.or.jp");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Japan")`
        pub const TERRITORY: Option<&str> = Some("Japan");
        /// `Some("Japanese language locale for Japan")`
        pub const TITLE: Option<&str> = Some("Japanese language locale for Japan");
    }
    pub mod LC_MESSAGES {
        /// `"^([-0nNｎＮ]|いいえ|イイエ)"`
        pub const NOEXPR: &str = "^([-0nNｎＮ]|いいえ|イイエ)";
        /// `Some("いいえ")`
        pub const NOSTR: Option<&str> = Some("いいえ");
        /// `"^([+1yYｙＹ]|はい|ハイ)"`
        pub const YESEXPR: &str = "^([+1yYｙＹ]|はい|ハイ)";
        /// `Some("はい")`
        pub const YESSTR: Option<&str> = Some("はい");
    }
    pub mod LC_MONETARY {
        /// `"￥"`
        pub const CURRENCY_SYMBOL: &str = "￥";
        /// `0`
        pub const FRAC_DIGITS: i64 = 0;
        /// `"JPY "`
        pub const INT_CURR_SYMBOL: &str = "JPY ";
        /// `0`
        pub const INT_FRAC_DIGITS: i64 = 0;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `Some(2)`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = Some(2);
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `Some(2)`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = Some(2);
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `4`
        pub const N_SIGN_POSN: i64 = 4;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `4`
        pub const P_SIGN_POSN: i64 = 4;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("81")`
        pub const INT_PREFIX: Option<&str> = Some("81");
        /// `Some("010")`
        pub const INT_SELECT: Option<&str> = Some("010");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c ;%a ;%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c ;%a ;%l");
    }
    pub mod LC_TIME {
        /// `&["日", "月", "火", "水", "木", "金", "土"]`
        pub const ABDAY: &[&str] = &["日", "月", "火", "水", "木", "金", "土"];
        /// `&[" 1月", " 2月", " 3月", " 4月", " 5月", " 6月", " 7月", " 8月", " 9月", "10月", "11月", "12月"]`
        pub const ABMON: &[&str] = &[" 1月", " 2月", " 3月", " 4月", " 5月", " 6月", " 7月", " 8月", " 9月", "10月", "11月", "12月"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `Some(&["〇", "一", "二", "三", "四", "五", "六", "七", "八", "九", "十", "十一", "十二", "十三", "十四", "十五", "十六", "十七", "十八", "十九", "二十", "二十一", "二十二", "二十三", "二十四", "二十五", "二十六", "二十七", "二十八", "二十九", "三十", "三十一", "三十二", "三十三", "三十四", "三十五", "三十六", "三十七", "三十八", "三十九", "四十", "四十一", "四十二", "四十三", "四十四", "四十五", "四十六", "四十七", "四十八", "四十九", "五十", "五十一", "五十二", "五十三", "五十四", "五十五", "五十六", "五十七", "五十八", "五十九", "六十", "六十一", "六十二", "六十三", "六十四", "六十五", "六十六", "六十七", "六十八", "六十九", "七十", "七十一", "七十二", "七十三", "七十四", "七十五", "七十六", "七十七", "七十八", "七十九", "八十", "八十一", "八十二", "八十三", "八十四", "八十五", "八十六", "八十七", "八十八", "八十九", "九十", "九十一", "九十二", "九十三", "九十四", "九十五", "九十六", "九十七", "九十八", "九十九"])`
        pub const ALT_DIGITS: Option<&[&str]> = Some(&["〇", "一", "二", "三", "四", "五", "六", "七", "八", "九", "十", "十一", "十二", "十三", "十四", "十五", "十六", "十七", "十八", "十九", "二十", "二十一", "二十二", "二十三", "二十四", "二十五", "二十六", "二十七", "二十八", "二十九", "三十", "三十一", "三十二", "三十三", "三十四", "三十五", "三十六", "三十七", "三十八", "三十九", "四十", "四十一", "四十二", "四十三", "四十四", "四十五", "四十六", "四十七", "四十八", "四十九", "五十", "五十一", "五十二", "五十三", "五十四", "五十五", "五十六", "五十七", "五十八", "五十九", "六十", "六十一", "六十二", "六十三", "六十四", "六十五", "六十六", "六十七", "六十八", "六十九", "七十", "七十一", "七十二", "七十三", "七十四", "七十五", "七十六", "七十七", "七十八", "七十九", "八十", "八十一", "八十二", "八十三", "八十四", "八十五", "八十六", "八十七", "八十八", "八十九", "九十", "九十一", "九十二", "九十三", "九十四", "九十五", "九十六", "九十七", "九十八", "九十九"]);
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["午前", "午後"]`
        pub const AM_PM: &[&str] = &["午前", "午後"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%Y年 %b %e日 %A %H:%M:%S %Z")`
        pub const DATE_FMT: Option<&str> = Some("%Y年 %b %e日 %A %H:%M:%S %Z");
        /// `&["日曜日", "月曜日", "火曜日", "水曜日", "木曜日", "金曜日", "土曜日"]`
        pub const DAY: &[&str] = &["日曜日", "月曜日", "火曜日", "水曜日", "木曜日", "金曜日", "土曜日"];
        /// `"%Y年%m月%d日"`
        pub const D_FMT: &str = "%Y年%m月%d日";
        /// `"%Y年%m月%d日 %H時%M分%S秒"`
        pub const D_T_FMT: &str = "%Y年%m月%d日 %H時%M分%S秒";
        /// `Some(&["+:2:2020/01/01:+*:令和:%EC%Ey年", "+:1:2019/05/01:2019/12/31:令和:%EC元年", "+:2:1990/01/01:2019/04/30:平成:%EC%Ey年", "+:1:1989/01/08:1989/12/31:平成:%EC元年", "+:2:1927/01/01:1989/01/07:昭和:%EC%Ey年", "+:1:1926/12/25:1926/12/31:昭和:%EC元年", "+:2:1913/01/01:1926/12/24:大正:%EC%Ey年", "+:1:1912/07/30:1912/12/31:大正:%EC元年", "+:6:1873/01/01:1912/07/29:明治:%EC%Ey年", "+:1:0001/01/01:1872/12/31:西暦:%EC%Ey年", "+:1:-0001/12/31:-*:紀元前:%EC%Ey年"])`
        pub const ERA: Option<&[&str]> = Some(&["+:2:2020/01/01:+*:令和:%EC%Ey年", "+:1:2019/05/01:2019/12/31:令和:%EC元年", "+:2:1990/01/01:2019/04/30:平成:%EC%Ey年", "+:1:1989/01/08:1989/12/31:平成:%EC元年", "+:2:1927/01/01:1989/01/07:昭和:%EC%Ey年", "+:1:1926/12/25:1926/12/31:昭和:%EC元年", "+:2:1913/01/01:1926/12/24:大正:%EC%Ey年", "+:1:1912/07/30:1912/12/31:大正:%EC元年", "+:6:1873/01/01:1912/07/29:明治:%EC%Ey年", "+:1:0001/01/01:1872/12/31:西暦:%EC%Ey年", "+:1:-0001/12/31:-*:紀元前:%EC%Ey年"]);
        /// `Some("%EY%m月%d日")`
        pub const ERA_D_FMT: Option<&str> = Some("%EY%m月%d日");
        /// `Some("%EY%m月%d日 %H時%M分%S秒")`
        pub const ERA_D_T_FMT: Option<&str> = Some("%EY%m月%d日 %H時%M分%S秒");
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["1月", "2月", "3月", "4月", "5月", "6月", "7月", "8月", "9月", "10月", "11月", "12月"]`
        pub const MON: &[&str] = &["1月", "2月", "3月", "4月", "5月", "6月", "7月", "8月", "9月", "10月", "11月", "12月"];
        /// `"%H時%M分%S秒"`
        pub const T_FMT: &str = "%H時%M分%S秒";
        /// `"%p%I時%M分%S秒"`
        pub const T_FMT_AMPM: &str = "%p%I時%M分%S秒";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ka_GE {
    pub mod LC_ADDRESS {
        /// `Some("GEO")`
        pub const COUNTRY_AB3: Option<&str> = Some("GEO");
        /// `Some("GE")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("GE");
        /// `Some("GE")`
        pub const COUNTRY_CAR: Option<&str> = Some("GE");
        /// `Some("99928")`
        pub const COUNTRY_ISBN: Option<&str> = Some("99928");
        /// `Some("საქართველო")`
        pub const COUNTRY_NAME: Option<&str> = Some("საქართველო");
        /// `Some(268)`
        pub const COUNTRY_NUM: Option<i64> = Some(268);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ka")`
        pub const LANG_AB: Option<&str> = Some("ka");
        /// `Some("geo")`
        pub const LANG_LIB: Option<&str> = Some("geo");
        /// `Some("ქართული")`
        pub const LANG_NAME: Option<&str> = Some("ქართული");
        /// `Some("kat")`
        pub const LANG_TERM: Option<&str> = Some("kat");
        /// `Some("%d%N%f%N%d%N%b%N%s %h 5e %r%N%C%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%d%N%f%N%d%N%b%N%s %h 5e %r%N%C%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_MONETARY"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_MONETARY"],
        ]);
        /// `Some("Pablo Saratxaga")`
        pub const CONTACT: Option<&str> = Some("Pablo Saratxaga");
        /// `Some("2001-01-26")`
        pub const DATE: Option<&str> = Some("2001-01-26");
        /// `Some("srtxg@chanae.alphanet.ch")`
        pub const EMAIL: Option<&str> = Some("srtxg@chanae.alphanet.ch");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Georgian")`
        pub const LANGUAGE: Option<&str> = Some("Georgian");
        /// `Some("0.6")`
        pub const REVISION: Option<&str> = Some("0.6");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Georgia")`
        pub const TERRITORY: Option<&str> = Some("Georgia");
        /// `Some("Georgian language locale for Georgia")`
        pub const TITLE: Option<&str> = Some("Georgian language locale for Georgia");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNaAა]"`
        pub const NOEXPR: &str = "^[-0nNaAა]";
        /// `Some("არა")`
        pub const NOSTR: Option<&str> = Some("არა");
        /// `"^[+1yYkKxXკ]"`
        pub const YESEXPR: &str = "^[+1yYkKxXკ]";
        /// `Some("კი")`
        pub const YESSTR: Option<&str> = Some("კი");
    }
    pub mod LC_MONETARY {
        /// `"₾"`
        pub const CURRENCY_SYMBOL: &str = "₾";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"GEL "`
        pub const INT_CURR_SYMBOL: &str = "GEL ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const MON_THOUSANDS_SEP: &str = ".";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const THOUSANDS_SEP: &str = ".";
    }
    pub mod LC_TELEPHONE {
        /// `Some("995")`
        pub const INT_PREFIX: Option<&str> = Some("995");
        /// `Some("8~10")`
        pub const INT_SELECT: Option<&str> = Some("8~10");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["კვი", "ორშ", "სამ", "ოთხ", "ხუთ", "პარ", "შაბ"]`
        pub const ABDAY: &[&str] = &["კვი", "ორშ", "სამ", "ოთხ", "ხუთ", "პარ", "შაბ"];
        /// `&["იან", "თებ", "მარ", "აპრ", "მაი", "ივნ", "ივლ", "აგვ", "სექ", "ოქტ", "ნოე", "დეკ"]`
        pub const ABMON: &[&str] = &["იან", "თებ", "მარ", "აპრ", "მაი", "ივნ", "ივლ", "აგვ", "სექ", "ოქტ", "ნოე", "დეკ"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["კვირა", "ორშაბათი", "სამშაბათი", "ოთხშაბათი", "ხუთშაბათი", "პარასკევი", "შაბათი"]`
        pub const DAY: &[&str] = &["კვირა", "ორშაბათი", "სამშაბათი", "ოთხშაბათი", "ხუთშაბათი", "პარასკევი", "შაბათი"];
        /// `"%m/%d/%Y"`
        pub const D_FMT: &str = "%m/%d/%Y";
        /// `"%Y წლის %d %B, %T %Z"`
        pub const D_T_FMT: &str = "%Y წლის %d %B, %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["იანვარი", "თებერვალი", "მარტი", "აპრილი", "მაისი", "ივნისი", "ივლისი", "აგვისტო", "სექტემბერი", "ოქტომბერი", "ნოემბერი", "დეკემბერი"]`
        pub const MON: &[&str] = &["იანვარი", "თებერვალი", "მარტი", "აპრილი", "მაისი", "ივნისი", "ივლისი", "აგვისტო", "სექტემბერი", "ოქტომბერი", "ნოემბერი", "დეკემბერი"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod kab_DZ {
    pub mod LC_ADDRESS {
        /// `Some("DZA")`
        pub const COUNTRY_AB3: Option<&str> = Some("DZA");
        /// `Some("DZ")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("DZ");
        /// `Some("DZ")`
        pub const COUNTRY_CAR: Option<&str> = Some("DZ");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Lezzayer")`
        pub const COUNTRY_NAME: Option<&str> = Some("Lezzayer");
        /// `Some(12)`
        pub const COUNTRY_NUM: Option<i64> = Some(12);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("kab")`
        pub const LANG_LIB: Option<&str> = Some("kab");
        /// `Some("Taqbaylit")`
        pub const LANG_NAME: Option<&str> = Some("Taqbaylit");
        /// `Some("kab")`
        pub const LANG_TERM: Option<&str> = Some("kab");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `None`
        pub const ADDRESS: Option<&str> = None;
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2016-07-06")`
        pub const DATE: Option<&str> = Some("2016-07-06");
        /// `Some("belkacem77@gmail.com")`
        pub const EMAIL: Option<&str> = Some("belkacem77@gmail.com");
        /// `None`
        pub const FAX: Option<&str> = None;
        /// `Some("Kabyle")`
        pub const LANGUAGE: Option<&str> = Some("Kabyle");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `None`
        pub const TEL: Option<&str> = None;
        /// `Some("Algeria")`
        pub const TERRITORY: Option<&str> = Some("Algeria");
        /// `Some("Kabyle language locale for Algeria")`
        pub const TITLE: Option<&str> = Some("Kabyle language locale for Algeria");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNuUaA]"`
        pub const NOEXPR: &str = "^[-0nNuUaA]";
        /// `Some("Uhu")`
        pub const NOSTR: Option<&str> = Some("Uhu");
        /// `"^[+1yYiI]"`
        pub const YESEXPR: &str = "^[+1yYiI]";
        /// `Some("Ih")`
        pub const YESSTR: Option<&str> = Some("Ih");
    }
    pub mod LC_MONETARY {
        /// `"DA"`
        pub const CURRENCY_SYMBOL: &str = "DA";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"DZD "`
        pub const INT_CURR_SYMBOL: &str = "DZD ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `""`
        pub const MON_THOUSANDS_SEP: &str = "";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `""`
        pub const THOUSANDS_SEP: &str = "";
    }
    pub use super::ar_DZ::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Ace", "Ari", "Ara", "Aha", "Amh", "Sem", "Sed"]`
        pub const ABDAY: &[&str] = &["Ace", "Ari", "Ara", "Aha", "Amh", "Sem", "Sed"];
        /// `&["Yen", "Fur", "Meɣ", "Yeb", "May", "Yun", "Yul", "ɣuc", "Cte", "Tub", "Wam", "Duj"]`
        pub const ABMON: &[&str] = &["Yen", "Fur", "Meɣ", "Yeb", "May", "Yun", "Yul", "ɣuc", "Cte", "Tub", "Wam", "Duj"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["FT", "MD"]`
        pub const AM_PM: &[&str] = &["FT", "MD"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Acer", "Arim", "Aram", "Ahad", "Amhad", "Sem", "Sed"]`
        pub const DAY: &[&str] = &["Acer", "Arim", "Aram", "Ahad", "Amhad", "Sem", "Sed"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(7)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(7);
        /// `Some(1)`
        pub const FIRST_WORKDAY: Option<i64> = Some(1);
        /// `&["Yennayer", "Fuṛar", "Meɣres", "Yebrir", "Mayyu", "Yunyu", "Yulyu", "ɣuct", "Ctembeṛ", "Tubeṛ", "Wambeṛ", "Dujembeṛ"]`
        pub const MON: &[&str] = &["Yennayer", "Fuṛar", "Meɣres", "Yebrir", "Mayyu", "Yunyu", "Yulyu", "ɣuct", "Ctembeṛ", "Tubeṛ", "Wambeṛ", "Dujembeṛ"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `"%I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod kk_KZ {
    pub mod LC_ADDRESS {
        /// `Some("KAZ")`
        pub const COUNTRY_AB3: Option<&str> = Some("KAZ");
        /// `Some("KZ")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("KZ");
        /// `Some("KZ")`
        pub const COUNTRY_CAR: Option<&str> = Some("KZ");
        /// `Some("978-601")`
        pub const COUNTRY_ISBN: Option<&str> = Some("978-601");
        /// `Some("Қазақстан")`
        pub const COUNTRY_NAME: Option<&str> = Some("Қазақстан");
        /// `Some(398)`
        pub const COUNTRY_NUM: Option<i64> = Some(398);
        /// `Some("KAZ")`
        pub const COUNTRY_POST: Option<&str> = Some("KAZ");
        /// `Some("kk")`
        pub const LANG_AB: Option<&str> = Some("kk");
        /// `Some("kaz")`
        pub const LANG_LIB: Option<&str> = Some("kaz");
        /// `Some("қазақ тілі")`
        pub const LANG_NAME: Option<&str> = Some("қазақ тілі");
        /// `Some("kaz")`
        pub const LANG_TERM: Option<&str> = Some("kaz");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Timur Birsh")`
        pub const CONTACT: Option<&str> = Some("Timur Birsh");
        /// `Some("2016-03-22")`
        pub const DATE: Option<&str> = Some("2016-03-22");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Kazakh")`
        pub const LANGUAGE: Option<&str> = Some("Kazakh");
        /// `Some("1.1")`
        pub const REVISION: Option<&str> = Some("1.1");
        /// `Some("linukz.org")`
        pub const SOURCE: Option<&str> = Some("linukz.org");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Kazakhstan")`
        pub const TERRITORY: Option<&str> = Some("Kazakhstan");
        /// `Some("Kazakh locale for Kazakhstan")`
        pub const TITLE: Option<&str> = Some("Kazakh locale for Kazakhstan");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNЖжN]"`
        pub const NOEXPR: &str = "^[-0nNЖжN]";
        /// `Some("жоқ")`
        pub const NOSTR: Option<&str> = Some("жоқ");
        /// `"^[+1yYИи]"`
        pub const YESEXPR: &str = "^[+1yYИи]";
        /// `Some("иә")`
        pub const YESSTR: Option<&str> = Some("иә");
    }
    pub mod LC_MONETARY {
        /// `"₸"`
        pub const CURRENCY_SYMBOL: &str = "₸";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"KZT "`
        pub const INT_CURR_SYMBOL: &str = "KZT ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `Some(1)`
        pub const INT_N_CS_PRECEDES: Option<i64> = Some(1);
        /// `Some(1)`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = Some(1);
        /// `Some(4)`
        pub const INT_N_SIGN_POSN: Option<i64> = Some(4);
        /// `Some(1)`
        pub const INT_P_CS_PRECEDES: Option<i64> = Some(1);
        /// `Some(2)`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = Some(2);
        /// `Some(4)`
        pub const INT_P_SIGN_POSN: Option<i64> = Some(4);
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const MON_THOUSANDS_SEP: &str = "\u{202f}";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `2`
        pub const P_SEP_BY_SPACE: i64 = 2;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const THOUSANDS_SEP: &str = "\u{202f}";
    }
    pub mod LC_TELEPHONE {
        /// `Some("7")`
        pub const INT_PREFIX: Option<&str> = Some("7");
        /// `Some("8~10")`
        pub const INT_SELECT: Option<&str> = Some("8~10");
        /// `Some("(%A) %l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("(%A) %l");
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["Жк", "Дс", "Сс", "Ср", "Бс", "Жм", "Сб"]`
        pub const ABDAY: &[&str] = &["Жк", "Дс", "Сс", "Ср", "Бс", "Жм", "Сб"];
        /// `&["Қаң", "Ақп", "Нау", "Сәу", "Мам", "Мау", "Шіл", "Там", "Қыр", "Қаз", "Қар", "Жел"]`
        pub const ABMON: &[&str] = &["Қаң", "Ақп", "Нау", "Сәу", "Мам", "Мау", "Шіл", "Там", "Қыр", "Қаз", "Қар", "Жел"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Жексенбі", "Дүйсенбі", "Сейсенбі", "Сәрсенбі", "Бейсенбі", "Жұма", "Сенбі"]`
        pub const DAY: &[&str] = &["Жексенбі", "Дүйсенбі", "Сейсенбі", "Сәрсенбі", "Бейсенбі", "Жұма", "Сенбі"];
        /// `"%d.%m.%Y"`
        pub const D_FMT: &str = "%d.%m.%Y";
        /// `"%a %d %b %Y %T"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Қаңтар", "Ақпан", "Наурыз", "Сәуір", "Мамыр", "Маусым", "Шілде", "Тамыз", "Қыркүйек", "Қазан", "Қараша", "Желтоқсан"]`
        pub const MON: &[&str] = &["Қаңтар", "Ақпан", "Наурыз", "Сәуір", "Мамыр", "Маусым", "Шілде", "Тамыз", "Қыркүйек", "Қазан", "Қараша", "Желтоқсан"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod kl_GL {
    pub mod LC_ADDRESS {
        /// `Some("GRL")`
        pub const COUNTRY_AB3: Option<&str> = Some("GRL");
        /// `Some("GL")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("GL");
        /// `Some("KN")`
        pub const COUNTRY_CAR: Option<&str> = Some("KN");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Kalaallit Nunaat")`
        pub const COUNTRY_NAME: Option<&str> = Some("Kalaallit Nunaat");
        /// `Some(304)`
        pub const COUNTRY_NUM: Option<i64> = Some(304);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("kl")`
        pub const LANG_AB: Option<&str> = Some("kl");
        /// `Some("kal")`
        pub const LANG_LIB: Option<&str> = Some("kal");
        /// `Some("kalaallisut")`
        pub const LANG_NAME: Option<&str> = Some("kalaallisut");
        /// `Some("kal")`
        pub const LANG_TERM: Option<&str> = Some("kal");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Kollegievej 6, DK-2920 Charlottenlund, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Kollegievej 6, DK-2920 Charlottenlund, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Kalaallisut")`
        pub const LANGUAGE: Option<&str> = Some("Kalaallisut");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Danish Standards Association")`
        pub const SOURCE: Option<&str> = Some("Danish Standards Association");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Greenland")`
        pub const TERRITORY: Option<&str> = Some("Greenland");
        /// `Some("Greenlandic locale for Greenland")`
        pub const TITLE: Option<&str> = Some("Greenlandic locale for Greenland");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0Nn]"`
        pub const NOEXPR: &str = "^[-0Nn]";
        /// `Some("naagga")`
        pub const NOSTR: Option<&str> = Some("naagga");
        /// `"^[+1JjYyAa]"`
        pub const YESEXPR: &str = "^[+1JjYyAa]";
        /// `Some("aap")`
        pub const YESSTR: Option<&str> = Some("aap");
    }
    pub use super::da_DK::LC_MONETARY;
    pub use super::da_DK::LC_NUMERIC;
    pub mod LC_TELEPHONE {
        /// `Some("299")`
        pub const INT_PREFIX: Option<&str> = Some("299");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["sap", "ata", "mar", "pin", "sis", "tal", "arf"]`
        pub const ABDAY: &[&str] = &["sap", "ata", "mar", "pin", "sis", "tal", "arf"];
        /// `&["jan", "feb", "mar", "apr", "maj", "jun", "jul", "aug", "sep", "okt", "nov", "dec"]`
        pub const ABMON: &[&str] = &["jan", "feb", "mar", "apr", "maj", "jun", "jul", "aug", "sep", "okt", "nov", "dec"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `Some(&["januaari", "februaari", "marsi", "apriili", "maaji", "juuni", "juuli", "aggusti", "septembari", "oktobari", "novembari", "decembari"])`
        pub const ALT_MON: Option<&[&str]> = Some(&["januaari", "februaari", "marsi", "apriili", "maaji", "juuni", "juuli", "aggusti", "septembari", "oktobari", "novembari", "decembari"]);
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["sapaat", "ataasinngorneq", "marlunngorneq", "pingasunngorneq", "sisamanngorneq", "tallimanngorneq", "arfininngorneq"]`
        pub const DAY: &[&str] = &["sapaat", "ataasinngorneq", "marlunngorneq", "pingasunngorneq", "sisamanngorneq", "tallimanngorneq", "arfininngorneq"];
        /// `"%b %d %Y"`
        pub const D_FMT: &str = "%b %d %Y";
        /// `"%a %b %d %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %b %d %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["januaarip", "februaarip", "marsip", "apriilip", "maajip", "juunip", "juulip", "aggustip", "septembarip", "oktobarip", "novembarip", "decembarip"]`
        pub const MON: &[&str] = &["januaarip", "februaarip", "marsip", "apriilip", "maajip", "juunip", "juulip", "aggustip", "septembarip", "oktobarip", "novembarip", "decembarip"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod km_KH {
    pub mod LC_ADDRESS {
        /// `Some("KHM")`
        pub const COUNTRY_AB3: Option<&str> = Some("KHM");
        /// `Some("KH")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("KH");
        /// `Some("KH")`
        pub const COUNTRY_CAR: Option<&str> = Some("KH");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("កម\u{17d2}ព\u{17bb}ជា")`
        pub const COUNTRY_NAME: Option<&str> = Some("កម\u{17d2}ព\u{17bb}ជា");
        /// `Some(116)`
        pub const COUNTRY_NUM: Option<i64> = Some(116);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("km")`
        pub const LANG_AB: Option<&str> = Some("km");
        /// `Some("khm")`
        pub const LANG_LIB: Option<&str> = Some("khm");
        /// `Some("ខ\u{17d2}មែរ")`
        pub const LANG_NAME: Option<&str> = Some("ខ\u{17d2}មែរ");
        /// `Some("khm")`
        pub const LANG_TERM: Option<&str> = Some("khm");
        /// `Some("%f%N%a%N%d%N%r%t%e%t%b%N%h%t%s%N%T%N%S%N%z%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%r%t%e%t%b%N%h%t%s%N%T%N%S%N%z%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Jens Herden at: jens@khmeros.info")`
        pub const CONTACT: Option<&str> = Some("Jens Herden at: jens@khmeros.info");
        /// `Some("2005-3-15")`
        pub const DATE: Option<&str> = Some("2005-3-15");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Khmer")`
        pub const LANGUAGE: Option<&str> = Some("Khmer");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Cambodia")`
        pub const TERRITORY: Option<&str> = Some("Cambodia");
        /// `Some("Khmer locale for Cambodia")`
        pub const TITLE: Option<&str> = Some("Khmer locale for Cambodia");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNទ]"`
        pub const NOEXPR: &str = "^[-0nNទ]";
        /// `Some("ទេ")`
        pub const NOSTR: Option<&str> = Some("ទេ");
        /// `"^[+1yYប]"`
        pub const YESEXPR: &str = "^[+1yYប]";
        /// `Some("បាទ/ចាស")`
        pub const YESSTR: Option<&str> = Some("បាទ/ចាស");
    }
    pub mod LC_MONETARY {
        /// `"៛"`
        pub const CURRENCY_SYMBOL: &str = "៛";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"KHR "`
        pub const INT_CURR_SYMBOL: &str = "KHR ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `Some(0)`
        pub const INT_N_CS_PRECEDES: Option<i64> = Some(0);
        /// `Some(0)`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = Some(0);
        /// `Some(1)`
        pub const INT_N_SIGN_POSN: Option<i64> = Some(1);
        /// `Some(0)`
        pub const INT_P_CS_PRECEDES: Option<i64> = Some(0);
        /// `Some(0)`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = Some(0);
        /// `Some(1)`
        pub const INT_P_SIGN_POSN: Option<i64> = Some(1);
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("855")`
        pub const INT_PREFIX: Option<&str> = Some("855");
        /// `Some("001")`
        pub const INT_SELECT: Option<&str> = Some("001");
        /// `Some("(%A)%l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("(%A)%l");
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["អា", "ច", "អ", "ព\u{17bb}", "ព\u{17d2}រ", "ស\u{17bb}", "ស"]`
        pub const ABDAY: &[&str] = &["អា", "ច", "អ", "ព\u{17bb}", "ព\u{17d2}រ", "ស\u{17bb}", "ស"];
        /// `&["១", "២", "៣", "៤", "៥", "៦", "៧", "៨", "៩", "១០", "១១", "១២"]`
        pub const ABMON: &[&str] = &["១", "២", "៣", "៤", "៥", "៦", "៧", "៨", "៩", "១០", "១១", "១២"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["ព\u{17d2}រ\u{17b9}ក", "ល\u{17d2}ងាច"]`
        pub const AM_PM: &[&str] = &["ព\u{17d2}រ\u{17b9}ក", "ល\u{17d2}ងាច"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["ថ\u{17d2}ងៃ\u{200b}អាទ\u{17b7}ត\u{17d2}យ", "ថ\u{17d2}ងៃ\u{200b}ច\u{17d0}ន\u{17d2}ទ", "ថ\u{17d2}ងៃ\u{200b}អង\u{17d2}គារ", "ថ\u{17d2}ងៃ\u{200b}ព\u{17bb}ធ", "ថ\u{17d2}ងៃ\u{200b}ព\u{17d2}រហស\u{17d2}បត\u{17b7}\u{17cd}", "ថ\u{17d2}ងៃ\u{200b}ស\u{17bb}ក\u{17d2}រ", "ថ\u{17d2}ងៃ\u{200b}សៅរ\u{17cd}"]`
        pub const DAY: &[&str] = &["ថ\u{17d2}ងៃ\u{200b}អាទ\u{17b7}ត\u{17d2}យ", "ថ\u{17d2}ងៃ\u{200b}ច\u{17d0}ន\u{17d2}ទ", "ថ\u{17d2}ងៃ\u{200b}អង\u{17d2}គារ", "ថ\u{17d2}ងៃ\u{200b}ព\u{17bb}ធ", "ថ\u{17d2}ងៃ\u{200b}ព\u{17d2}រហស\u{17d2}បត\u{17b7}\u{17cd}", "ថ\u{17d2}ងៃ\u{200b}ស\u{17bb}ក\u{17d2}រ", "ថ\u{17d2}ងៃ\u{200b}សៅរ\u{17cd}"];
        /// `"%e %B %Y"`
        pub const D_FMT: &str = "%e %B %Y";
        /// `"%A ថ\u{17d2}ងៃ %e ខែ %B ឆ\u{17d2}នា\u{17c6} %Y, %H ម\u{17c9}ោង m នាទ\u{17b8} %S វ\u{17b7}នាទ\u{17b8}\u{200b}"`
        pub const D_T_FMT: &str = "%A ថ\u{17d2}ងៃ %e ខែ %B ឆ\u{17d2}នា\u{17c6} %Y, %H ម\u{17c9}ោង m នាទ\u{17b8} %S វ\u{17b7}នាទ\u{17b8}\u{200b}";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(1)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(1);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["មករា", "ក\u{17bb}ម\u{17d2}ភៈ", "ម\u{17b8}នា", "មេសា", "ឧសភា", "ម\u{17b7}ថ\u{17bb}នា", "កក\u{17d2}កដា", "ស\u{17b8}ហា", "កញ\u{17d2}ញា", "ត\u{17bb}លា", "វ\u{17b7}ច\u{17d2}ឆ\u{17b7}កា", "ធ\u{17d2}ន\u{17bc}"]`
        pub const MON: &[&str] = &["មករា", "ក\u{17bb}ម\u{17d2}ភៈ", "ម\u{17b8}នា", "មេសា", "ឧសភា", "ម\u{17b7}ថ\u{17bb}នា", "កក\u{17d2}កដា", "ស\u{17b8}ហា", "កញ\u{17d2}ញា", "ត\u{17bb}លា", "វ\u{17b7}ច\u{17d2}ឆ\u{17b7}កា", "ធ\u{17d2}ន\u{17bc}"];
        /// `"%H:%M:%S"`
        pub const T_FMT: &str = "%H:%M:%S";
        /// `"%l:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%l:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod kn_IN {
    pub mod LC_ADDRESS {
        /// `Some("IND")`
        pub const COUNTRY_AB3: Option<&str> = Some("IND");
        /// `Some("IN")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IN");
        /// `Some("IND")`
        pub const COUNTRY_CAR: Option<&str> = Some("IND");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("ಭಾರತ")`
        pub const COUNTRY_NAME: Option<&str> = Some("ಭಾರತ");
        /// `Some(356)`
        pub const COUNTRY_NUM: Option<i64> = Some(356);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("kn")`
        pub const LANG_AB: Option<&str> = Some("kn");
        /// `Some("kan")`
        pub const LANG_LIB: Option<&str> = Some("kan");
        /// `Some("ಕನ\u{ccd}ನಡ")`
        pub const LANG_NAME: Option<&str> = Some("ಕನ\u{ccd}ನಡ");
        /// `Some("kan")`
        pub const LANG_TERM: Option<&str> = Some("kan");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2002-11-28")`
        pub const DATE: Option<&str> = Some("2002-11-28");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Kannada")`
        pub const LANGUAGE: Option<&str> = Some("Kannada");
        /// `Some("0.1")`
        pub const REVISION: Option<&str> = Some("0.1");
        /// `Some("IndLinux.org")`
        pub const SOURCE: Option<&str> = Some("IndLinux.org");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("India")`
        pub const TERRITORY: Option<&str> = Some("India");
        /// `Some("Kannada language locale for India")`
        pub const TITLE: Option<&str> = Some("Kannada language locale for India");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNಇ]"`
        pub const NOEXPR: &str = "^[-0nNಇ]";
        /// `Some("ಇಲ\u{ccd}ಲ")`
        pub const NOSTR: Option<&str> = Some("ಇಲ\u{ccd}ಲ");
        /// `"^[+1yYಹ]"`
        pub const YESEXPR: &str = "^[+1yYಹ]";
        /// `Some("ಹ\u{ccc}ದು")`
        pub const YESSTR: Option<&str> = Some("ಹ\u{ccc}ದು");
    }
    pub use super::hi_IN::LC_MONETARY;
    pub use super::hi_IN::LC_NUMERIC;
    pub use super::hi_IN::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["ರ", "ಸ\u{ccb}", "ಮಂ", "ಬು", "ಗು", "ಶು", "ಶ"]`
        pub const ABDAY: &[&str] = &["ರ", "ಸ\u{ccb}", "ಮಂ", "ಬು", "ಗು", "ಶು", "ಶ"];
        /// `&["ಜನ", "ಫ\u{cc6}ಬ\u{ccd}ರ", "ಮಾರ\u{ccd}ಚ\u{ccd}", "ಏಪ\u{ccd}ರ\u{cbf}", "ಮ\u{cc7}", "ಜ\u{cc2}ನ\u{ccd}", "ಜುಲ\u{cc8}", "ಆ", "ಸ\u{cc6}ಪ\u{ccd}ಟ\u{cc6}ಂ", "ಅಕ\u{ccd}ಟ\u{ccb}", "ನವ\u{cc6}ಂ", "ಡ\u{cbf}ಸ\u{cc6}ಂ"]`
        pub const ABMON: &[&str] = &["ಜನ", "ಫ\u{cc6}ಬ\u{ccd}ರ", "ಮಾರ\u{ccd}ಚ\u{ccd}", "ಏಪ\u{ccd}ರ\u{cbf}", "ಮ\u{cc7}", "ಜ\u{cc2}ನ\u{ccd}", "ಜುಲ\u{cc8}", "ಆ", "ಸ\u{cc6}ಪ\u{ccd}ಟ\u{cc6}ಂ", "ಅಕ\u{ccd}ಟ\u{ccb}", "ನವ\u{cc6}ಂ", "ಡ\u{cbf}ಸ\u{cc6}ಂ"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["ಪ\u{cc2}ರ\u{ccd}ವಾಹ\u{ccd}ನ", "ಅಪರಾಹ\u{ccd}ನ"]`
        pub const AM_PM: &[&str] = &["ಪ\u{cc2}ರ\u{ccd}ವಾಹ\u{ccd}ನ", "ಅಪರಾಹ\u{ccd}ನ"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["ರವ\u{cbf}ವಾರ", "ಸ\u{ccb}ಮವಾರ", "ಮಂಗಳವಾರ", "ಬುಧವಾರ", "ಗುರುವಾರ", "ಶುಕ\u{ccd}ರವಾರ", "ಶನ\u{cbf}ವಾರ"]`
        pub const DAY: &[&str] = &["ರವ\u{cbf}ವಾರ", "ಸ\u{ccb}ಮವಾರ", "ಮಂಗಳವಾರ", "ಬುಧವಾರ", "ಗುರುವಾರ", "ಶುಕ\u{ccd}ರವಾರ", "ಶನ\u{cbf}ವಾರ"];
        /// `"%-d/%-m/%y"`
        pub const D_FMT: &str = "%-d/%-m/%y";
        /// `"%A %d %b %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A %d %b %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["ಜನವರ\u{cbf}", "ಫ\u{cc6}ಬ\u{ccd}ರವರ\u{cbf}", "ಮಾರ\u{ccd}ಚ\u{ccd}", "ಏಪ\u{ccd}ರ\u{cbf}ಲ\u{ccd}", "ಮ\u{cc7}", "ಜ\u{cc2}ನ\u{ccd}", "ಜುಲ\u{cc8}", "ಆಗಸ\u{ccd}ಟ\u{ccd}", "ಸ\u{cc6}ಪ\u{ccd}ಟ\u{cc6}ಂಬರ\u{ccd}", "ಅಕ\u{ccd}ಟ\u{ccb}ಬರ\u{ccd}", "ನವ\u{cc6}ಂಬರ\u{ccd}", "ಡ\u{cbf}ಸ\u{cc6}ಂಬರ\u{ccd}"]`
        pub const MON: &[&str] = &["ಜನವರ\u{cbf}", "ಫ\u{cc6}ಬ\u{ccd}ರವರ\u{cbf}", "ಮಾರ\u{ccd}ಚ\u{ccd}", "ಏಪ\u{ccd}ರ\u{cbf}ಲ\u{ccd}", "ಮ\u{cc7}", "ಜ\u{cc2}ನ\u{ccd}", "ಜುಲ\u{cc8}", "ಆಗಸ\u{ccd}ಟ\u{ccd}", "ಸ\u{cc6}ಪ\u{ccd}ಟ\u{cc6}ಂಬರ\u{ccd}", "ಅಕ\u{ccd}ಟ\u{ccb}ಬರ\u{ccd}", "ನವ\u{cc6}ಂಬರ\u{ccd}", "ಡ\u{cbf}ಸ\u{cc6}ಂಬರ\u{ccd}"];
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT: &str = "%I:%M:%S %p %Z";
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p %Z";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ko_KR {
    pub mod LC_ADDRESS {
        /// `Some("KOR")`
        pub const COUNTRY_AB3: Option<&str> = Some("KOR");
        /// `Some("KR")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("KR");
        /// `Some("ROK")`
        pub const COUNTRY_CAR: Option<&str> = Some("ROK");
        /// `Some("979-11,978-89")`
        pub const COUNTRY_ISBN: Option<&str> = Some("979-11,978-89");
        /// `Some("대한민국")`
        pub const COUNTRY_NAME: Option<&str> = Some("대한민국");
        /// `Some(410)`
        pub const COUNTRY_NUM: Option<i64> = Some(410);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ko")`
        pub const LANG_AB: Option<&str> = Some("ko");
        /// `Some("kor")`
        pub const LANG_LIB: Option<&str> = Some("kor");
        /// `Some("한국어")`
        pub const LANG_NAME: Option<&str> = Some("한국어");
        /// `Some("kor")`
        pub const LANG_TERM: Option<&str> = Some("kor");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-11-09")`
        pub const DATE: Option<&str> = Some("2000-11-09");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Korean")`
        pub const LANGUAGE: Option<&str> = Some("Korean");
        /// `Some("1.1")`
        pub const REVISION: Option<&str> = Some("1.1");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("South Korea")`
        pub const TERRITORY: Option<&str> = Some("South Korea");
        /// `Some("Korean locale for Republic of Korea")`
        pub const TITLE: Option<&str> = Some("Korean locale for Republic of Korea");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNｎＮ아]"`
        pub const NOEXPR: &str = "^[-0nNｎＮ아]";
        /// `Some("아니요")`
        pub const NOSTR: Option<&str> = Some("아니요");
        /// `"^[+1yYｙＹ예]"`
        pub const YESEXPR: &str = "^[+1yYｙＹ예]";
        /// `Some("예")`
        pub const YESSTR: Option<&str> = Some("예");
    }
    pub mod LC_MONETARY {
        /// `"₩"`
        pub const CURRENCY_SYMBOL: &str = "₩";
        /// `0`
        pub const FRAC_DIGITS: i64 = 0;
        /// `"KRW "`
        pub const INT_CURR_SYMBOL: &str = "KRW ";
        /// `0`
        pub const INT_FRAC_DIGITS: i64 = 0;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `4`
        pub const N_SIGN_POSN: i64 = 4;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("82")`
        pub const INT_PREFIX: Option<&str> = Some("82");
        /// `Some("001")`
        pub const INT_SELECT: Option<&str> = Some("001");
        /// `Some("(%A) %l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("(%A) %l");
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["일", "월", "화", "수", "목", "금", "토"]`
        pub const ABDAY: &[&str] = &["일", "월", "화", "수", "목", "금", "토"];
        /// `&[" 1월", " 2월", " 3월", " 4월", " 5월", " 6월", " 7월", " 8월", " 9월", "10월", "11월", "12월"]`
        pub const ABMON: &[&str] = &[" 1월", " 2월", " 3월", " 4월", " 5월", " 6월", " 7월", " 8월", " 9월", "10월", "11월", "12월"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["오전", "오후"]`
        pub const AM_PM: &[&str] = &["오전", "오후"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%Y. %m. %d. (%a) %H:%M:%S %Z")`
        pub const DATE_FMT: Option<&str> = Some("%Y. %m. %d. (%a) %H:%M:%S %Z");
        /// `&["일요일", "월요일", "화요일", "수요일", "목요일", "금요일", "토요일"]`
        pub const DAY: &[&str] = &["일요일", "월요일", "화요일", "수요일", "목요일", "금요일", "토요일"];
        /// `"%Y년 %m월 %d일"`
        pub const D_FMT: &str = "%Y년 %m월 %d일";
        /// `"%Y년 %m월 %d일 (%a) %p %I시 %M분 %S초"`
        pub const D_T_FMT: &str = "%Y년 %m월 %d일 (%a) %p %I시 %M분 %S초";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["1월", "2월", "3월", "4월", "5월", "6월", "7월", "8월", "9월", "10월", "11월", "12월"]`
        pub const MON: &[&str] = &["1월", "2월", "3월", "4월", "5월", "6월", "7월", "8월", "9월", "10월", "11월", "12월"];
        /// `"%H시 %M분 %S초"`
        pub const T_FMT: &str = "%H시 %M분 %S초";
        /// `"%p %I시 %M분 %S초"`
        pub const T_FMT_AMPM: &str = "%p %I시 %M분 %S초";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod kok_IN {
    pub mod LC_ADDRESS {
        /// `Some("IND")`
        pub const COUNTRY_AB3: Option<&str> = Some("IND");
        /// `Some("IN")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IN");
        /// `Some("IND")`
        pub const COUNTRY_CAR: Option<&str> = Some("IND");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("भारत")`
        pub const COUNTRY_NAME: Option<&str> = Some("भारत");
        /// `Some(356)`
        pub const COUNTRY_NUM: Option<i64> = Some(356);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("kok")`
        pub const LANG_LIB: Option<&str> = Some("kok");
        /// `Some("को\u{902}कणी")`
        pub const LANG_NAME: Option<&str> = Some("को\u{902}कणी");
        /// `Some("kok")`
        pub const LANG_TERM: Option<&str> = Some("kok");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Level-1, Tower 10, Cyber City, Magarpatta City, Hadapsar, Pune, Maharashtra, India-411013")`
        pub const ADDRESS: Option<&str> = Some("Level-1, Tower 10, Cyber City, Magarpatta City, Hadapsar, Pune, Maharashtra, India-411013");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2009,December,08")`
        pub const DATE: Option<&str> = Some("2009,December,08");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Konkani")`
        pub const LANGUAGE: Option<&str> = Some("Konkani");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Red Hat, Pune")`
        pub const SOURCE: Option<&str> = Some("Red Hat, Pune");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("India")`
        pub const TERRITORY: Option<&str> = Some("India");
        /// `Some("Konkani language locale for India")`
        pub const TITLE: Option<&str> = Some("Konkani language locale for India");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNन]"`
        pub const NOEXPR: &str = "^[-0nNन]";
        /// `Some("न\u{94d}ही")`
        pub const NOSTR: Option<&str> = Some("न\u{94d}ही");
        /// `"^[+1yYह]"`
        pub const YESEXPR: &str = "^[+1yYह]";
        /// `Some("हय")`
        pub const YESSTR: Option<&str> = Some("हय");
    }
    pub use super::hi_IN::LC_MONETARY;
    pub use super::hi_IN::LC_NUMERIC;
    pub use super::hi_IN::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["आयतार", "सोमार", "म\u{902}गळवार", "ब\u{941}धवार", "ब\u{947}र\u{947}सतार", "श\u{941}करार", "श\u{947}नवार"]`
        pub const ABDAY: &[&str] = &["आयतार", "सोमार", "म\u{902}गळवार", "ब\u{941}धवार", "ब\u{947}र\u{947}सतार", "श\u{941}करार", "श\u{947}नवार"];
        /// `&["जान\u{947}वारी", "फ\u{947}ब\u{94d}र\u{941}वारी", "मार\u{94d}च", "एप\u{94d}रिल", "म\u{947}", "ज\u{942}न", "ज\u{941}ल\u{948}", "ओगस\u{94d}ट", "स\u{947}प\u{94d}ट\u{947}\u{902}बर", "ओक\u{94d}टोबर", "नोव\u{94d}ह\u{947}\u{902}बर", "डिस\u{947}\u{902}बर"]`
        pub const ABMON: &[&str] = &["जान\u{947}वारी", "फ\u{947}ब\u{94d}र\u{941}वारी", "मार\u{94d}च", "एप\u{94d}रिल", "म\u{947}", "ज\u{942}न", "ज\u{941}ल\u{948}", "ओगस\u{94d}ट", "स\u{947}प\u{94d}ट\u{947}\u{902}बर", "ओक\u{94d}टोबर", "नोव\u{94d}ह\u{947}\u{902}बर", "डिस\u{947}\u{902}बर"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["म.प\u{942}.", "म.न\u{902}."]`
        pub const AM_PM: &[&str] = &["म.प\u{942}.", "म.न\u{902}."];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["आयतार", "सोमार", "म\u{902}गळवार", "ब\u{941}धवार", "ब\u{947}र\u{947}सतार", "श\u{941}करार", "श\u{947}नवार"]`
        pub const DAY: &[&str] = &["आयतार", "सोमार", "म\u{902}गळवार", "ब\u{941}धवार", "ब\u{947}र\u{947}सतार", "श\u{941}करार", "श\u{947}नवार"];
        /// `"%-d-%-m-%y"`
        pub const D_FMT: &str = "%-d-%-m-%y";
        /// `"%A %d %b %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A %d %b %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["जान\u{947}वारी", "फ\u{947}ब\u{94d}र\u{941}वारी", "मार\u{94d}च", "एप\u{94d}रिल", "म\u{947}", "ज\u{942}न", "ज\u{941}ल\u{948}", "ओगस\u{94d}ट", "स\u{947}प\u{94d}ट\u{947}\u{902}बर", "ओक\u{94d}टोबर", "नोव\u{94d}ह\u{947}\u{902}बर", "डिस\u{947}\u{902}बर"]`
        pub const MON: &[&str] = &["जान\u{947}वारी", "फ\u{947}ब\u{94d}र\u{941}वारी", "मार\u{94d}च", "एप\u{94d}रिल", "म\u{947}", "ज\u{942}न", "ज\u{941}ल\u{948}", "ओगस\u{94d}ट", "स\u{947}प\u{94d}ट\u{947}\u{902}बर", "ओक\u{94d}टोबर", "नोव\u{94d}ह\u{947}\u{902}बर", "डिस\u{947}\u{902}बर"];
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT: &str = "%I:%M:%S %p %Z";
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p %Z";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ks_IN {
    pub mod LC_ADDRESS {
        /// `Some("IND")`
        pub const COUNTRY_AB3: Option<&str> = Some("IND");
        /// `Some("IN")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IN");
        /// `Some("IND")`
        pub const COUNTRY_CAR: Option<&str> = Some("IND");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("ہ\u{650}ندوستان")`
        pub const COUNTRY_NAME: Option<&str> = Some("ہ\u{650}ندوستان");
        /// `Some(356)`
        pub const COUNTRY_NUM: Option<i64> = Some(356);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ks")`
        pub const LANG_AB: Option<&str> = Some("ks");
        /// `Some("kas")`
        pub const LANG_LIB: Option<&str> = Some("kas");
        /// `Some("کٲش\u{64f}ر")`
        pub const LANG_NAME: Option<&str> = Some("کٲش\u{64f}ر");
        /// `Some("kas")`
        pub const LANG_TERM: Option<&str> = Some("kas");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Marisfot III, Marigold Premises, East-Wing, Kalyaninagar, Pune, India-411014")`
        pub const ADDRESS: Option<&str> = Some("Marisfot III, Marigold Premises, East-Wing, Kalyaninagar, Pune, India-411014");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2009,April,06")`
        pub const DATE: Option<&str> = Some("2009,April,06");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Kashmiri")`
        pub const LANGUAGE: Option<&str> = Some("Kashmiri");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Red Hat, Pune")`
        pub const SOURCE: Option<&str> = Some("Red Hat, Pune");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("India")`
        pub const TERRITORY: Option<&str> = Some("India");
        /// `Some("Kashmiri language locale for India")`
        pub const TITLE: Option<&str> = Some("Kashmiri language locale for India");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNن]"`
        pub const NOEXPR: &str = "^[-0nNن]";
        /// `Some("نع")`
        pub const NOSTR: Option<&str> = Some("نع");
        /// `"^[+1yYآ]"`
        pub const YESEXPR: &str = "^[+1yYآ]";
        /// `Some("آ")`
        pub const YESSTR: Option<&str> = Some("آ");
    }
    pub use super::hi_IN::LC_MONETARY;
    pub use super::hi_IN::LC_NUMERIC;
    pub use super::hi_IN::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["آتهوار", "ژءنتروار", "بوءںوار", "بودهوار", "برىسوار", "جمع", "بٹوار"]`
        pub const ABDAY: &[&str] = &["آتهوار", "ژءنتروار", "بوءںوار", "بودهوار", "برىسوار", "جمع", "بٹوار"];
        /// `&["جنؤری", "فرؤری", "مار\u{655}چ", "اپریل", "می\u{654}", "جو\u{657}ن", "جو\u{657}لایی", "اگست", "ستمبر", "اکتو\u{657}بر", "نومبر", "دسمبر"]`
        pub const ABMON: &[&str] = &["جنؤری", "فرؤری", "مار\u{655}چ", "اپریل", "می\u{654}", "جو\u{657}ن", "جو\u{657}لایی", "اگست", "ستمبر", "اکتو\u{657}بر", "نومبر", "دسمبر"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["دوپھربرونھ", "دوپھرپتھ"]`
        pub const AM_PM: &[&str] = &["دوپھربرونھ", "دوپھرپتھ"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["آتهوار", "ژءندروار", "بوءںوار", "بودهوار", "برىسوار", "جمع", "بٹوار"]`
        pub const DAY: &[&str] = &["آتهوار", "ژءندروار", "بوءںوار", "بودهوار", "برىسوار", "جمع", "بٹوار"];
        /// `"%-m/%-d/%y"`
        pub const D_FMT: &str = "%-m/%-d/%y";
        /// `"%A %d %b %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A %d %b %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["جنؤری", "فرؤری", "مار\u{655}چ", "اپریل", "می\u{654}", "جو\u{657}ن", "جو\u{657}لایی", "اگست", "ستمبر", "اکتو\u{657}بر", "نومبر", "دسمبر"]`
        pub const MON: &[&str] = &["جنؤری", "فرؤری", "مار\u{655}چ", "اپریل", "می\u{654}", "جو\u{657}ن", "جو\u{657}لایی", "اگست", "ستمبر", "اکتو\u{657}بر", "نومبر", "دسمبر"];
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT: &str = "%I:%M:%S %p %Z";
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p %Z";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ks_IN_devanagari {
    pub mod LC_ADDRESS {
        /// `Some("IND")`
        pub const COUNTRY_AB3: Option<&str> = Some("IND");
        /// `Some("IN")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IN");
        /// `Some("IND")`
        pub const COUNTRY_CAR: Option<&str> = Some("IND");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("भारत")`
        pub const COUNTRY_NAME: Option<&str> = Some("भारत");
        /// `Some(356)`
        pub const COUNTRY_NUM: Option<i64> = Some(356);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ks")`
        pub const LANG_AB: Option<&str> = Some("ks");
        /// `Some("kas")`
        pub const LANG_LIB: Option<&str> = Some("kas");
        /// `Some("कॉश\u{941}र")`
        pub const LANG_NAME: Option<&str> = Some("कॉश\u{941}र");
        /// `Some("kas")`
        pub const LANG_TERM: Option<&str> = Some("kas");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2008-08-26")`
        pub const DATE: Option<&str> = Some("2008-08-26");
        /// `Some("ks-gnome-trans-commits@lists.code.indlinux.net")`
        pub const EMAIL: Option<&str> = Some("ks-gnome-trans-commits@lists.code.indlinux.net");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Kashmiri")`
        pub const LANGUAGE: Option<&str> = Some("Kashmiri");
        /// `Some("0.1")`
        pub const REVISION: Option<&str> = Some("0.1");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("India")`
        pub const TERRITORY: Option<&str> = Some("India");
        /// `Some("Kashmiri(devanagari) language locale for India")`
        pub const TITLE: Option<&str> = Some("Kashmiri(devanagari) language locale for India");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNन]"`
        pub const NOEXPR: &str = "^[-0nNन]";
        /// `Some("न")`
        pub const NOSTR: Option<&str> = Some("न");
        /// `"^[+1yYइ]"`
        pub const YESEXPR: &str = "^[+1yYइ]";
        /// `Some("इ\u{902}न")`
        pub const YESSTR: Option<&str> = Some("इ\u{902}न");
    }
    pub use super::ks_IN::LC_MONETARY;
    pub use super::ks_IN::LC_NUMERIC;
    pub use super::hi_IN::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["आथ ", "च\u{945}\u{93c}द\u{941}र", "बोम", "ब\u{94d}वद", "ब\u{94d}रस", "शोक\u{941}र", "बट\u{941}"]`
        pub const ABDAY: &[&str] = &["आथ ", "च\u{945}\u{93c}द\u{941}र", "बोम", "ब\u{94d}वद", "ब\u{94d}रस", "शोक\u{941}र", "बट\u{941}"];
        /// `&["जनवरी", "फ\u{93c}रवरी", "मार\u{94d}च", "अप\u{94d}र\u{947}ल", "मई", "ज\u{942}न", "ज\u{941}लाई", "अगस\u{94d}त", "सितम\u{94d}बर", "अक\u{94d}ट\u{942}बर", "नवम\u{94d}बर", "दिसम\u{94d}बर"]`
        pub const ABMON: &[&str] = &["जनवरी", "फ\u{93c}रवरी", "मार\u{94d}च", "अप\u{94d}र\u{947}ल", "मई", "ज\u{942}न", "ज\u{941}लाई", "अगस\u{94d}त", "सितम\u{94d}बर", "अक\u{94d}ट\u{942}बर", "नवम\u{94d}बर", "दिसम\u{94d}बर"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["प\u{942}र\u{94d}वाह\u{94d}न", "अपराह\u{94d}न"]`
        pub const AM_PM: &[&str] = &["प\u{942}र\u{94d}वाह\u{94d}न", "अपराह\u{94d}न"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["आथवार", "च\u{945}\u{93c}द\u{941}रवार", "बोमवार", "ब\u{94d}वदवार", "ब\u{94d}रसवार", "शोक\u{941}रवार", "बट\u{941}वार"]`
        pub const DAY: &[&str] = &["आथवार", "च\u{945}\u{93c}द\u{941}रवार", "बोमवार", "ब\u{94d}वदवार", "ब\u{94d}रसवार", "शोक\u{941}रवार", "बट\u{941}वार"];
        /// `"%-m/%-d/%y"`
        pub const D_FMT: &str = "%-m/%-d/%y";
        /// `"%A %d %b %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A %d %b %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["जनवरी", "फ\u{93c}रवरी", "मार\u{94d}च", "अप\u{94d}र\u{947}ल", "मई", "ज\u{942}न", "ज\u{941}लाई", "अगस\u{94d}त", "सितम\u{94d}बर", "अक\u{94d}ट\u{942}बर", "नवम\u{94d}बर", "दिसम\u{94d}बर"]`
        pub const MON: &[&str] = &["जनवरी", "फ\u{93c}रवरी", "मार\u{94d}च", "अप\u{94d}र\u{947}ल", "मई", "ज\u{942}न", "ज\u{941}लाई", "अगस\u{94d}त", "सितम\u{94d}बर", "अक\u{94d}ट\u{942}बर", "नवम\u{94d}बर", "दिसम\u{94d}बर"];
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT: &str = "%I:%M:%S %p %Z";
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p %Z";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ku_TR {
    pub mod LC_ADDRESS {
        /// `Some("TUR")`
        pub const COUNTRY_AB3: Option<&str> = Some("TUR");
        /// `Some("TR")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("TR");
        /// `Some("TR")`
        pub const COUNTRY_CAR: Option<&str> = Some("TR");
        /// `Some("975")`
        pub const COUNTRY_ISBN: Option<&str> = Some("975");
        /// `Some("Tirkiye")`
        pub const COUNTRY_NAME: Option<&str> = Some("Tirkiye");
        /// `Some(792)`
        pub const COUNTRY_NUM: Option<i64> = Some(792);
        /// `Some("TR")`
        pub const COUNTRY_POST: Option<&str> = Some("TR");
        /// `Some("ku")`
        pub const LANG_AB: Option<&str> = Some("ku");
        /// `Some("kur")`
        pub const LANG_LIB: Option<&str> = Some("kur");
        /// `Some("kurdî")`
        pub const LANG_NAME: Option<&str> = Some("kurdî");
        /// `Some("kur")`
        pub const LANG_TERM: Option<&str> = Some("kur");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Kader DILSIZ, Pablo Saratxaga")`
        pub const CONTACT: Option<&str> = Some("Kader DILSIZ, Pablo Saratxaga");
        /// `Some("2003-07-27")`
        pub const DATE: Option<&str> = Some("2003-07-27");
        /// `Some("kader@ikader.com, pablo@mandrakesoft.com")`
        pub const EMAIL: Option<&str> = Some("kader@ikader.com, pablo@mandrakesoft.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Kurdish")`
        pub const LANGUAGE: Option<&str> = Some("Kurdish");
        /// `Some("0.1")`
        pub const REVISION: Option<&str> = Some("0.1");
        /// `Some("Kader DILSIZ")`
        pub const SOURCE: Option<&str> = Some("Kader DILSIZ");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Turkey")`
        pub const TERRITORY: Option<&str> = Some("Turkey");
        /// `Some("Kurdish (latin) locale for Turkey")`
        pub const TITLE: Option<&str> = Some("Kurdish (latin) locale for Turkey");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("nâ")`
        pub const NOSTR: Option<&str> = Some("nâ");
        /// `"^[+1eEdDyY]"`
        pub const YESEXPR: &str = "^[+1eEdDyY]";
        /// `Some("erê")`
        pub const YESSTR: Option<&str> = Some("erê");
    }
    pub use super::tr_TR::LC_MONETARY;
    pub use super::tr_TR::LC_NUMERIC;
    pub use super::tr_TR::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["yêk", "dus", "sês", "çar", "pên", "înî", "sep"]`
        pub const ABDAY: &[&str] = &["yêk", "dus", "sês", "çar", "pên", "înî", "sep"];
        /// `&["Çil", "Sib", "Ada", "Nîs", "Gul", "Hez", "Tîr", "Teb", "Îlo", "Cot", "Mij", "Kan"]`
        pub const ABMON: &[&str] = &["Çil", "Sib", "Ada", "Nîs", "Gul", "Hez", "Tîr", "Teb", "Îlo", "Cot", "Mij", "Kan"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%A %d %B %Y")`
        pub const DATE_FMT: Option<&str> = Some("%A %d %B %Y");
        /// `&["yêksêm", "dusêm", "sêsêm", "çarsêm", "pêncsêm", "înî", "sept"]`
        pub const DAY: &[&str] = &["yêksêm", "dusêm", "sêsêm", "çarsêm", "pêncsêm", "înî", "sept"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%A %d %B %Y %T %Z"`
        pub const D_T_FMT: &str = "%A %d %B %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Çile", "Sibat", "Adar", "Nîsan", "Gulan", "Hezîran", "Tîrmeh", "Tebax", "Îlon", "Cotmeh", "Mijdar", "Kanûn"]`
        pub const MON: &[&str] = &["Çile", "Sibat", "Adar", "Nîsan", "Gulan", "Hezîran", "Tîrmeh", "Tebax", "Îlon", "Cotmeh", "Mijdar", "Kanûn"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod kw_GB {
    pub mod LC_ADDRESS {
        /// `Some("GBR")`
        pub const COUNTRY_AB3: Option<&str> = Some("GBR");
        /// `Some("GB")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("GB");
        /// `Some("GB")`
        pub const COUNTRY_CAR: Option<&str> = Some("GB");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Rywvaneth Unys")`
        pub const COUNTRY_NAME: Option<&str> = Some("Rywvaneth Unys");
        /// `Some(826)`
        pub const COUNTRY_NUM: Option<i64> = Some(826);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("kw")`
        pub const LANG_AB: Option<&str> = Some("kw");
        /// `Some("cor")`
        pub const LANG_LIB: Option<&str> = Some("cor");
        /// `Some("kernewek")`
        pub const LANG_NAME: Option<&str> = Some("kernewek");
        /// `Some("cor")`
        pub const LANG_TERM: Option<&str> = Some("cor");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Croí Lár, Ballinahalla, Maigh Cuilinn,, Co. Gaillimh, Ireland")`
        pub const ADDRESS: Option<&str> = Some("Croí Lár, Ballinahalla, Maigh Cuilinn,, Co. Gaillimh, Ireland");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Cornish")`
        pub const LANGUAGE: Option<&str> = Some("Cornish");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Alastair McKinstry")`
        pub const SOURCE: Option<&str> = Some("Alastair McKinstry");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("United Kingdom")`
        pub const TERRITORY: Option<&str> = Some("United Kingdom");
        /// `Some("Cornish locale for Britain")`
        pub const TITLE: Option<&str> = Some("Cornish locale for Britain");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("na")`
        pub const NOSTR: Option<&str> = Some("na");
        /// `"^[+1eEyY]"`
        pub const YESEXPR: &str = "^[+1eEyY]";
        /// `Some("ea")`
        pub const YESSTR: Option<&str> = Some("ea");
    }
    pub mod LC_MONETARY {
        /// `"£"`
        pub const CURRENCY_SYMBOL: &str = "£";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"GBP "`
        pub const INT_CURR_SYMBOL: &str = "GBP ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("44")`
        pub const INT_PREFIX: Option<&str> = Some("44");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["Sul", "Lun", "Mth", "Mhr", "Yow", "Gwe", "Sad"]`
        pub const ABDAY: &[&str] = &["Sul", "Lun", "Mth", "Mhr", "Yow", "Gwe", "Sad"];
        /// `&["Gen", "Hwe", "Meu", "Ebr", "Me", "Met", "Gor", "Est", "Gwn", "Hed", "Du", "Kev"]`
        pub const ABMON: &[&str] = &["Gen", "Hwe", "Meu", "Ebr", "Me", "Met", "Gor", "Est", "Gwn", "Hed", "Du", "Kev"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["De Sul", "De Lun", "De Merth", "De Merher", "De Yow", "De Gwener", "De Sadorn"]`
        pub const DAY: &[&str] = &["De Sul", "De Lun", "De Merth", "De Merher", "De Yow", "De Gwener", "De Sadorn"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["mis Genver", "mis Hwevrer", "mis Meurth", "mis Ebrel", "mis Me", "mis Metheven", "mis Gortheren", "mis Est", "mis Gwynngala", "mis Hedra", "mis Du", "mis Kevardhu"]`
        pub const MON: &[&str] = &["mis Genver", "mis Hwevrer", "mis Meurth", "mis Ebrel", "mis Me", "mis Metheven", "mis Gortheren", "mis Est", "mis Gwynngala", "mis Hedra", "mis Du", "mis Kevardhu"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ky_KG {
    pub mod LC_ADDRESS {
        /// `Some("KGZ")`
        pub const COUNTRY_AB3: Option<&str> = Some("KGZ");
        /// `Some("KG")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("KG");
        /// `Some("KS")`
        pub const COUNTRY_CAR: Option<&str> = Some("KS");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Кыргызстан")`
        pub const COUNTRY_NAME: Option<&str> = Some("Кыргызстан");
        /// `Some(417)`
        pub const COUNTRY_NUM: Option<i64> = Some(417);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ky")`
        pub const LANG_AB: Option<&str> = Some("ky");
        /// `Some("kir")`
        pub const LANG_LIB: Option<&str> = Some("kir");
        /// `Some("кыргызча")`
        pub const LANG_NAME: Option<&str> = Some("кыргызча");
        /// `Some("kir")`
        pub const LANG_TERM: Option<&str> = Some("kir");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Pablo Saratxaga, Timur Jamakeev")`
        pub const CONTACT: Option<&str> = Some("Pablo Saratxaga, Timur Jamakeev");
        /// `Some("2004-10-14")`
        pub const DATE: Option<&str> = Some("2004-10-14");
        /// `Some("srtxg@mandrakesoft.com, ztimur@mail.ru")`
        pub const EMAIL: Option<&str> = Some("srtxg@mandrakesoft.com, ztimur@mail.ru");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Kyrgyz")`
        pub const LANGUAGE: Option<&str> = Some("Kyrgyz");
        /// `Some("0.2")`
        pub const REVISION: Option<&str> = Some("0.2");
        /// `Some("Timur Jamakeev")`
        pub const SOURCE: Option<&str> = Some("Timur Jamakeev");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Kyrgyzstan")`
        pub const TERRITORY: Option<&str> = Some("Kyrgyzstan");
        /// `Some("Kyrgyz Language Locale for Kyrgyzstan")`
        pub const TITLE: Option<&str> = Some("Kyrgyz Language Locale for Kyrgyzstan");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNЖж]"`
        pub const NOEXPR: &str = "^[-0nNЖж]";
        /// `Some("жок")`
        pub const NOSTR: Option<&str> = Some("жок");
        /// `"^[+1yYОо]"`
        pub const YESEXPR: &str = "^[+1yYОо]";
        /// `Some("ооба")`
        pub const YESSTR: Option<&str> = Some("ооба");
    }
    pub mod LC_MONETARY {
        /// `"сом"`
        pub const CURRENCY_SYMBOL: &str = "сом";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"KGS "`
        pub const INT_CURR_SYMBOL: &str = "KGS ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const MON_THOUSANDS_SEP: &str = "\u{202f}";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const THOUSANDS_SEP: &str = "\u{202f}";
    }
    pub mod LC_TELEPHONE {
        /// `Some("996")`
        pub const INT_PREFIX: Option<&str> = Some("996");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["жк", "дш", "ше", "ша", "бш", "жм", "иш"]`
        pub const ABDAY: &[&str] = &["жк", "дш", "ше", "ша", "бш", "жм", "иш"];
        /// `&["янв", "фев", "мар", "апр", "май", "июн", "июл", "авг", "сен", "окт", "ноя", "дек"]`
        pub const ABMON: &[&str] = &["янв", "фев", "мар", "апр", "май", "июн", "июл", "авг", "сен", "окт", "ноя", "дек"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a, %e-%b %Y ж, %H:%M:%S %Z")`
        pub const DATE_FMT: Option<&str> = Some("%a, %e-%b %Y ж, %H:%M:%S %Z");
        /// `&["жекшемби", "дүйшөмбү", "шейшемби", "шаршемби", "бейшемби", "жума", "ишемби"]`
        pub const DAY: &[&str] = &["жекшемби", "дүйшөмбү", "шейшемби", "шаршемби", "бейшемби", "жума", "ишемби"];
        /// `"%d.%m.%Y"`
        pub const D_FMT: &str = "%d.%m.%Y";
        /// `"%a %d %b %Y %T"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["январь", "февраль", "март", "апрель", "май", "июнь", "июль", "август", "сентябрь", "октябрь", "ноябрь", "декабрь"]`
        pub const MON: &[&str] = &["январь", "февраль", "март", "апрель", "май", "июнь", "июль", "август", "сентябрь", "октябрь", "ноябрь", "декабрь"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod lb_LU {
    pub mod LC_ADDRESS {
        /// `Some("LUX")`
        pub const COUNTRY_AB3: Option<&str> = Some("LUX");
        /// `Some("LU")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("LU");
        /// `Some("L")`
        pub const COUNTRY_CAR: Option<&str> = Some("L");
        /// `Some("2")`
        pub const COUNTRY_ISBN: Option<&str> = Some("2");
        /// `Some("Lëtzebuerg")`
        pub const COUNTRY_NAME: Option<&str> = Some("Lëtzebuerg");
        /// `Some(442)`
        pub const COUNTRY_NUM: Option<i64> = Some(442);
        /// `Some("L")`
        pub const COUNTRY_POST: Option<&str> = Some("L");
        /// `Some("lb")`
        pub const LANG_AB: Option<&str> = Some("lb");
        /// `Some("ltz")`
        pub const LANG_LIB: Option<&str> = Some("ltz");
        /// `Some("Lëtzebuergesch")`
        pub const LANG_NAME: Option<&str> = Some("Lëtzebuergesch");
        /// `Some("ltz")`
        pub const LANG_TERM: Option<&str> = Some("ltz");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Pit Wenkin")`
        pub const CONTACT: Option<&str> = Some("Pit Wenkin");
        /// `Some("2011-01-28")`
        pub const DATE: Option<&str> = Some("2011-01-28");
        /// `Some("pit@wenkin.lu")`
        pub const EMAIL: Option<&str> = Some("pit@wenkin.lu");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Luxembourgish")`
        pub const LANGUAGE: Option<&str> = Some("Luxembourgish");
        /// `Some("0.2")`
        pub const REVISION: Option<&str> = Some("0.2");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Luxembourg")`
        pub const TERRITORY: Option<&str> = Some("Luxembourg");
        /// `Some("Luxembourgish locale for Luxembourg")`
        pub const TITLE: Option<&str> = Some("Luxembourgish locale for Luxembourg");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("nee")`
        pub const NOSTR: Option<&str> = Some("nee");
        /// `"^[+1jJyY]"`
        pub const YESEXPR: &str = "^[+1jJyY]";
        /// `Some("jo")`
        pub const YESSTR: Option<&str> = Some("jo");
    }
    pub use super::de_DE::LC_MONETARY;
    pub use super::de_DE::LC_NUMERIC;
    pub mod LC_TELEPHONE {
        /// `Some("352")`
        pub const INT_PREFIX: Option<&str> = Some("352");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `Some("%A %l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("%A %l");
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["So", "Mé", "Dë", "Më", "Do", "Fr", "Sa"]`
        pub const ABDAY: &[&str] = &["So", "Mé", "Dë", "Më", "Do", "Fr", "Sa"];
        /// `&["Jan", "Feb", "Mäe", "Abr", "Mee", "Jun", "Jul", "Aug", "Sep", "Okt", "Nov", "Dez"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Mäe", "Abr", "Mee", "Jun", "Jul", "Aug", "Sep", "Okt", "Nov", "Dez"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a %-d. %b %H:%M:%S %Z %Y")`
        pub const DATE_FMT: Option<&str> = Some("%a %-d. %b %H:%M:%S %Z %Y");
        /// `&["Sonndeg", "Méindeg", "Dënschdeg", "Mëttwoch", "Donneschdeg", "Freideg", "Samschdeg"]`
        pub const DAY: &[&str] = &["Sonndeg", "Méindeg", "Dënschdeg", "Mëttwoch", "Donneschdeg", "Freideg", "Samschdeg"];
        /// `"%d.%m.%Y"`
        pub const D_FMT: &str = "%d.%m.%Y";
        /// `"%a %d. %b %Y %T"`
        pub const D_T_FMT: &str = "%a %d. %b %Y %T";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Januar", "Februar", "Mäerz", "Abrëll", "Mee", "Juni", "Juli", "August", "September", "Oktober", "November", "Dezember"]`
        pub const MON: &[&str] = &["Januar", "Februar", "Mäerz", "Abrëll", "Mee", "Juni", "Juli", "August", "September", "Oktober", "November", "Dezember"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod lg_UG {
    pub mod LC_ADDRESS {
        /// `Some("UGA")`
        pub const COUNTRY_AB3: Option<&str> = Some("UGA");
        /// `Some("UG")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("UG");
        /// `Some("EAU")`
        pub const COUNTRY_CAR: Option<&str> = Some("EAU");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Yuganda")`
        pub const COUNTRY_NAME: Option<&str> = Some("Yuganda");
        /// `Some(800)`
        pub const COUNTRY_NUM: Option<i64> = Some(800);
        /// `Some("UG")`
        pub const COUNTRY_POST: Option<&str> = Some("UG");
        /// `Some("lg")`
        pub const LANG_AB: Option<&str> = Some("lg");
        /// `Some("lug")`
        pub const LANG_LIB: Option<&str> = Some("lug");
        /// `Some("Luganda")`
        pub const LANG_NAME: Option<&str> = Some("Luganda");
        /// `Some("lug")`
        pub const LANG_TERM: Option<&str> = Some("lug");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("c/o P.O. Box 5190 Kampala, Uganda")`
        pub const ADDRESS: Option<&str> = Some("c/o P.O. Box 5190 Kampala, Uganda");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Kizito Birabwa")`
        pub const CONTACT: Option<&str> = Some("Kizito Birabwa");
        /// `Some("2001-11-04")`
        pub const DATE: Option<&str> = Some("2001-11-04");
        /// `Some("kompyuta@kizito.uklinux.net")`
        pub const EMAIL: Option<&str> = Some("kompyuta@kizito.uklinux.net");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Ganda")`
        pub const LANGUAGE: Option<&str> = Some("Ganda");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Akademe ya Luganda")`
        pub const SOURCE: Option<&str> = Some("Akademe ya Luganda");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Uganda")`
        pub const TERRITORY: Option<&str> = Some("Uganda");
        /// `Some("Luganda locale for Uganda")`
        pub const TITLE: Option<&str> = Some("Luganda locale for Uganda");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("Nedda")`
        pub const NOSTR: Option<&str> = Some("Nedda");
        /// `"^[+1yY]"`
        pub const YESEXPR: &str = "^[+1yY]";
        /// `Some("Ye")`
        pub const YESSTR: Option<&str> = Some("Ye");
    }
    pub mod LC_MONETARY {
        /// `"USh"`
        pub const CURRENCY_SYMBOL: &str = "USh";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"UGX "`
        pub const INT_CURR_SYMBOL: &str = "UGX ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("256")`
        pub const INT_PREFIX: Option<&str> = Some("256");
        /// `Some("000")`
        pub const INT_SELECT: Option<&str> = Some("000");
        /// `Some("%a %l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("%a %l");
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["Sab", "Bal", "Lw2", "Lw3", "Lw4", "Lw5", "Lw6"]`
        pub const ABDAY: &[&str] = &["Sab", "Bal", "Lw2", "Lw3", "Lw4", "Lw5", "Lw6"];
        /// `&["Jan", "Feb", "Mar", "Apu", "Maa", "Juu", "Jul", "Agu", "Seb", "Oki", "Nov", "Des"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Mar", "Apu", "Maa", "Juu", "Jul", "Agu", "Seb", "Oki", "Nov", "Des"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Sabiiti", "Balaza", "Lwakubiri", "Lwakusatu", "Lwakuna", "Lwakutaano", "Lwamukaaga"]`
        pub const DAY: &[&str] = &["Sabiiti", "Balaza", "Lwakubiri", "Lwakusatu", "Lwakuna", "Lwakutaano", "Lwamukaaga"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Janwaliyo", "Febwaliyo", "Marisi", "Apuli", "Maayi", "Juuni", "Julaayi", "Agusito", "Sebuttemba", "Okitobba", "Novemba", "Desemba"]`
        pub const MON: &[&str] = &["Janwaliyo", "Febwaliyo", "Marisi", "Apuli", "Maayi", "Juuni", "Julaayi", "Agusito", "Sebuttemba", "Okitobba", "Novemba", "Desemba"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod li_BE {
    pub mod LC_ADDRESS {
        /// `Some("BEL")`
        pub const COUNTRY_AB3: Option<&str> = Some("BEL");
        /// `Some("BE")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("BE");
        /// `Some("B")`
        pub const COUNTRY_CAR: Option<&str> = Some("B");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Bèlsj")`
        pub const COUNTRY_NAME: Option<&str> = Some("Bèlsj");
        /// `Some(56)`
        pub const COUNTRY_NUM: Option<i64> = Some(56);
        /// `Some("B")`
        pub const COUNTRY_POST: Option<&str> = Some("B");
        /// `Some("li")`
        pub const LANG_AB: Option<&str> = Some("li");
        /// `Some("lim")`
        pub const LANG_LIB: Option<&str> = Some("lim");
        /// `Some("Lèmbörgs")`
        pub const LANG_NAME: Option<&str> = Some("Lèmbörgs");
        /// `Some("lim")`
        pub const LANG_TERM: Option<&str> = Some("lim");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_ADDRESS"],
        ]);
        /// `Some("Kenneth Christiansen, Pablo Saratxaga")`
        pub const CONTACT: Option<&str> = Some("Kenneth Christiansen, Pablo Saratxaga");
        /// `Some("2003-11-30")`
        pub const DATE: Option<&str> = Some("2003-11-30");
        /// `Some("kenneth@gnu.org, pablo@mandriva.com")`
        pub const EMAIL: Option<&str> = Some("kenneth@gnu.org, pablo@mandriva.com");
        /// `None`
        pub const FAX: Option<&str> = None;
        /// `Some("Limburgish")`
        pub const LANGUAGE: Option<&str> = Some("Limburgish");
        /// `Some("0.1")`
        pub const REVISION: Option<&str> = Some("0.1");
        /// `Some("information from Kenneth Christiansen")`
        pub const SOURCE: Option<&str> = Some("information from Kenneth Christiansen");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Belgium")`
        pub const TERRITORY: Option<&str> = Some("Belgium");
        /// `Some("Limburgish Language Locale for Belgium")`
        pub const TITLE: Option<&str> = Some("Limburgish Language Locale for Belgium");
    }
    pub use super::nl_BE::LC_MESSAGES;
    pub use super::nl_BE::LC_MONETARY;
    pub use super::nl_BE::LC_NUMERIC;
    pub use super::nl_BE::LC_TELEPHONE;
    pub use super::li_NL::LC_TIME;
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod li_NL {
    pub mod LC_ADDRESS {
        /// `Some("NLD")`
        pub const COUNTRY_AB3: Option<&str> = Some("NLD");
        /// `Some("NL")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("NL");
        /// `Some("NL")`
        pub const COUNTRY_CAR: Option<&str> = Some("NL");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Nederland")`
        pub const COUNTRY_NAME: Option<&str> = Some("Nederland");
        /// `Some(528)`
        pub const COUNTRY_NUM: Option<i64> = Some(528);
        /// `Some("NL")`
        pub const COUNTRY_POST: Option<&str> = Some("NL");
        /// `Some("li")`
        pub const LANG_AB: Option<&str> = Some("li");
        /// `Some("lim")`
        pub const LANG_LIB: Option<&str> = Some("lim");
        /// `Some("Lèmbörgs")`
        pub const LANG_NAME: Option<&str> = Some("Lèmbörgs");
        /// `Some("lim")`
        pub const LANG_TERM: Option<&str> = Some("lim");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_ADDRESS"],
        ]);
        /// `Some("Kenneth Christiansen, Pablo Saratxaga")`
        pub const CONTACT: Option<&str> = Some("Kenneth Christiansen, Pablo Saratxaga");
        /// `Some("2003-11-30")`
        pub const DATE: Option<&str> = Some("2003-11-30");
        /// `Some("kenneth@gnu.org, pablo@mandriva.com")`
        pub const EMAIL: Option<&str> = Some("kenneth@gnu.org, pablo@mandriva.com");
        /// `None`
        pub const FAX: Option<&str> = None;
        /// `Some("Limburgish")`
        pub const LANGUAGE: Option<&str> = Some("Limburgish");
        /// `Some("0.1")`
        pub const REVISION: Option<&str> = Some("0.1");
        /// `Some("information from Kenneth Christiansen")`
        pub const SOURCE: Option<&str> = Some("information from Kenneth Christiansen");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Netherlands")`
        pub const TERRITORY: Option<&str> = Some("Netherlands");
        /// `Some("Limburgish Language Locale for the Netherlands")`
        pub const TITLE: Option<&str> = Some("Limburgish Language Locale for the Netherlands");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `None`
        pub const NOSTR: Option<&str> = None;
        /// `"^[+1jJyY]"`
        pub const YESEXPR: &str = "^[+1jJyY]";
        /// `None`
        pub const YESSTR: Option<&str> = None;
    }
    pub use super::nl_NL::LC_MONETARY;
    pub use super::nl_NL::LC_NUMERIC;
    pub use super::nl_NL::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["zón", "mao", "dae", "goo", "dón", "vri", "zao"]`
        pub const ABDAY: &[&str] = &["zón", "mao", "dae", "goo", "dón", "vri", "zao"];
        /// `&["jan", "fib", "mie", "epr", "mei", "jun", "jul", "aug", "sep", "okt", "nov", "des"]`
        pub const ABMON: &[&str] = &["jan", "fib", "mie", "epr", "mei", "jun", "jul", "aug", "sep", "okt", "nov", "des"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["zóndig", "maondig", "daensdig", "goonsdig", "dónderdig", "vriedig", "zaoterdig"]`
        pub const DAY: &[&str] = &["zóndig", "maondig", "daensdig", "goonsdig", "dónderdig", "vriedig", "zaoterdig"];
        /// `"%d.%m.%Y"`
        pub const D_FMT: &str = "%d.%m.%Y";
        /// `"%a %d. %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d. %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["jannewarie", "fibberwarie", "miert", "eprèl", "meij", "junie", "julie", "augustus", "september", "oktober", "november", "desember"]`
        pub const MON: &[&str] = &["jannewarie", "fibberwarie", "miert", "eprèl", "meij", "junie", "julie", "augustus", "september", "oktober", "november", "desember"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod lij_IT {
    pub mod LC_ADDRESS {
        /// `Some("ITA")`
        pub const COUNTRY_AB3: Option<&str> = Some("ITA");
        /// `Some("IT")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IT");
        /// `Some("I")`
        pub const COUNTRY_CAR: Option<&str> = Some("I");
        /// `Some("978-88,979-12")`
        pub const COUNTRY_ISBN: Option<&str> = Some("978-88,979-12");
        /// `Some("Italia")`
        pub const COUNTRY_NAME: Option<&str> = Some("Italia");
        /// `Some(380)`
        pub const COUNTRY_NUM: Option<i64> = Some(380);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("lij")`
        pub const LANG_LIB: Option<&str> = Some("lij");
        /// `Some("Lìgure")`
        pub const LANG_NAME: Option<&str> = Some("Lìgure");
        /// `Some("lij")`
        pub const LANG_TERM: Option<&str> = Some("lij");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2010-10-22")`
        pub const DATE: Option<&str> = Some("2010-10-22");
        /// `Some("alessio.gastaldi@libero.it")`
        pub const EMAIL: Option<&str> = Some("alessio.gastaldi@libero.it");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Ligurian")`
        pub const LANGUAGE: Option<&str> = Some("Ligurian");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Gastaldi")`
        pub const SOURCE: Option<&str> = Some("Gastaldi");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Italy")`
        pub const TERRITORY: Option<&str> = Some("Italy");
        /// `Some("Ligurian locale for Italy")`
        pub const TITLE: Option<&str> = Some("Ligurian locale for Italy");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `None`
        pub const NOSTR: Option<&str> = None;
        /// `"^[+1sSyY]"`
        pub const YESEXPR: &str = "^[+1sSyY]";
        /// `None`
        pub const YESSTR: Option<&str> = None;
    }
    pub use super::it_IT::LC_MONETARY;
    pub use super::it_IT::LC_NUMERIC;
    pub mod LC_TELEPHONE {
        /// `Some("39")`
        pub const INT_PREFIX: Option<&str> = Some("39");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["dom", "lûn", "mar", "mer", "zêu", "ven", "sab"]`
        pub const ABDAY: &[&str] = &["dom", "lûn", "mar", "mer", "zêu", "ven", "sab"];
        /// `&["zen", "fev", "mar", "arv", "maz", "zûg", "lûg", "ago", "set", "ött", "nov", "dix"]`
        pub const ABMON: &[&str] = &["zen", "fev", "mar", "arv", "maz", "zûg", "lûg", "ago", "set", "ött", "nov", "dix"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a %e %b %Y, %H.%M.%S, %Z")`
        pub const DATE_FMT: Option<&str> = Some("%a %e %b %Y, %H.%M.%S, %Z");
        /// `&["domenega", "lûnedì", "martedì", "mercUrdì", "zêggia", "venardì", "sabbo"]`
        pub const DAY: &[&str] = &["domenega", "lûnedì", "martedì", "mercUrdì", "zêggia", "venardì", "sabbo"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["zenâ", "fevrâ", "marzo", "avrî", "mazzo", "zûgno", "lûggio", "agosto", "settembre", "ottobre", "novembre", "dixembre"]`
        pub const MON: &[&str] = &["zenâ", "fevrâ", "marzo", "avrî", "mazzo", "zûgno", "lûggio", "agosto", "settembre", "ottobre", "novembre", "dixembre"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ln_CD {
    pub mod LC_ADDRESS {
        /// `Some("COD")`
        pub const COUNTRY_AB3: Option<&str> = Some("COD");
        /// `Some("CD")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("CD");
        /// `Some("CGO")`
        pub const COUNTRY_CAR: Option<&str> = Some("CGO");
        /// `Some("2")`
        pub const COUNTRY_ISBN: Option<&str> = Some("2");
        /// `Some("Repibiki demokratiki ya Kongó")`
        pub const COUNTRY_NAME: Option<&str> = Some("Repibiki demokratiki ya Kongó");
        /// `Some(180)`
        pub const COUNTRY_NUM: Option<i64> = Some(180);
        /// `Some("CD")`
        pub const COUNTRY_POST: Option<&str> = Some("CD");
        /// `Some("ln")`
        pub const LANG_AB: Option<&str> = Some("ln");
        /// `Some("lin")`
        pub const LANG_LIB: Option<&str> = Some("lin");
        /// `Some("lingála")`
        pub const LANG_NAME: Option<&str> = Some("lingála");
        /// `Some("lin")`
        pub const LANG_TERM: Option<&str> = Some("lin");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Ubuntu DR Congo Team")`
        pub const ADDRESS: Option<&str> = Some("Ubuntu DR Congo Team");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("René Manassé GALEKWA")`
        pub const CONTACT: Option<&str> = Some("René Manassé GALEKWA");
        /// `Some("2016-03-03")`
        pub const DATE: Option<&str> = Some("2016-03-03");
        /// `Some("renemanasse@gmail.com")`
        pub const EMAIL: Option<&str> = Some("renemanasse@gmail.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Lingala")`
        pub const LANGUAGE: Option<&str> = Some("Lingala");
        /// `Some("1.1")`
        pub const REVISION: Option<&str> = Some("1.1");
        /// `Some("Ubuntu")`
        pub const SOURCE: Option<&str> = Some("Ubuntu");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Democratic Republic of the Congo")`
        pub const TERRITORY: Option<&str> = Some("Democratic Republic of the Congo");
        /// `Some("Lingala locale for Democratic Republic of the Congo")`
        pub const TITLE: Option<&str> = Some("Lingala locale for Democratic Republic of the Congo");
    }
    pub mod LC_MESSAGES {
        /// `"^[nNtT]"`
        pub const NOEXPR: &str = "^[nNtT]";
        /// `Some("Tɛ\u{302}")`
        pub const NOSTR: Option<&str> = Some("Tɛ\u{302}");
        /// `"^[yYiIíÍ]"`
        pub const YESEXPR: &str = "^[yYiIíÍ]";
        /// `Some("Íyo")`
        pub const YESSTR: Option<&str> = Some("Íyo");
    }
    pub mod LC_MONETARY {
        /// `"FC"`
        pub const CURRENCY_SYMBOL: &str = "FC";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"CDF "`
        pub const INT_CURR_SYMBOL: &str = "CDF ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `"."`
        pub const MON_THOUSANDS_SEP: &str = ".";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `""`
        pub const THOUSANDS_SEP: &str = "";
    }
    pub mod LC_TELEPHONE {
        /// `Some("243")`
        pub const INT_PREFIX: Option<&str> = Some("243");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `Some("%a %l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("%a %l");
        /// `Some("+%c %a%t%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a%t%l");
    }
    pub mod LC_TIME {
        /// `&["m1.", "m2.", "m3.", "m4.", "m5.", "m6.", "m7."]`
        pub const ABDAY: &[&str] = &["m1.", "m2.", "m3.", "m4.", "m5.", "m6.", "m7."];
        /// `&["yan", "fbl", "msi", "apl", "mai", "yun", "yul", "agt", "stb", "ɔtb", "nvb", "dsb"]`
        pub const ABMON: &[&str] = &["yan", "fbl", "msi", "apl", "mai", "yun", "yul", "agt", "stb", "ɔtb", "nvb", "dsb"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%A %-e %B %Y, %H:%M:%S (UTC%z)")`
        pub const DATE_FMT: Option<&str> = Some("%A %-e %B %Y, %H:%M:%S (UTC%z)");
        /// `&["Lomíngo", "Mosálá mɔ\u{30c}kɔ\u{301}", "Misálá míbalé", "Misálá mísáto", "Misálá mínei", "Misálá mítáno", "Mpɔ\u{301}sɔ"]`
        pub const DAY: &[&str] = &["Lomíngo", "Mosálá mɔ\u{30c}kɔ\u{301}", "Misálá míbalé", "Misálá mísáto", "Misálá mínei", "Misálá mítáno", "Mpɔ\u{301}sɔ"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["sánzá ya yambo", "sánzá ya míbalé", "sánzá ya mísáto", "sánzá ya mínei", "sánzá ya mítáno", "sánzá ya motóbá", "sánzá ya nsambo", "sánzá ya mwambe", "sánzá ya libwa", "sánzá ya zómi", "sánzá ya zómi na mɔ\u{30c}kɔ\u{301}", "sánzá ya zómi na míbalé"]`
        pub const MON: &[&str] = &["sánzá ya yambo", "sánzá ya míbalé", "sánzá ya mísáto", "sánzá ya mínei", "sánzá ya mítáno", "sánzá ya motóbá", "sánzá ya nsambo", "sánzá ya mwambe", "sánzá ya libwa", "sánzá ya zómi", "sánzá ya zómi na mɔ\u{30c}kɔ\u{301}", "sánzá ya zómi na míbalé"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod lo_LA {
    pub mod LC_ADDRESS {
        /// `Some("LAO")`
        pub const COUNTRY_AB3: Option<&str> = Some("LAO");
        /// `Some("LA")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("LA");
        /// `Some("LAO")`
        pub const COUNTRY_CAR: Option<&str> = Some("LAO");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("ລາວ")`
        pub const COUNTRY_NAME: Option<&str> = Some("ລາວ");
        /// `Some(418)`
        pub const COUNTRY_NUM: Option<i64> = Some(418);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("lo")`
        pub const LANG_AB: Option<&str> = Some("lo");
        /// `Some("lao")`
        pub const LANG_LIB: Option<&str> = Some("lao");
        /// `Some("ລາວ")`
        pub const LANG_NAME: Option<&str> = Some("ລາວ");
        /// `Some("lao")`
        pub const LANG_TERM: Option<&str> = Some("lao");
        /// `Some("%f%N%a%N%d%N%r%t%e%t%b%N%h%t%s%N%T%N%S%N%z%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%r%t%e%t%b%N%h%t%s%N%T%N%S%N%z%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Anousak Souphavanh at: anousak@muanglao.com")`
        pub const CONTACT: Option<&str> = Some("Anousak Souphavanh at: anousak@muanglao.com");
        /// `Some("2003-4-1")`
        pub const DATE: Option<&str> = Some("2003-4-1");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Lao")`
        pub const LANGUAGE: Option<&str> = Some("Lao");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Laos")`
        pub const TERRITORY: Option<&str> = Some("Laos");
        /// `Some("Lao locale for Laos")`
        pub const TITLE: Option<&str> = Some("Lao locale for Laos");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNບ]"`
        pub const NOEXPR: &str = "^[-0nNບ]";
        /// `Some("ບ\u{ecd}\u{ec8}ແມ\u{ec8}ນ")`
        pub const NOSTR: Option<&str> = Some("ບ\u{ecd}\u{ec8}ແມ\u{ec8}ນ");
        /// `"^[+1yYມ]"`
        pub const YESEXPR: &str = "^[+1yYມ]";
        /// `Some("ແມ\u{ec8}ນ")`
        pub const YESSTR: Option<&str> = Some("ແມ\u{ec8}ນ");
    }
    pub mod LC_MONETARY {
        /// `"₭"`
        pub const CURRENCY_SYMBOL: &str = "₭";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"LAK "`
        pub const INT_CURR_SYMBOL: &str = "LAK ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `2`
        pub const N_SEP_BY_SPACE: i64 = 2;
        /// `4`
        pub const N_SIGN_POSN: i64 = 4;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `2`
        pub const P_SEP_BY_SPACE: i64 = 2;
        /// `4`
        pub const P_SIGN_POSN: i64 = 4;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("856")`
        pub const INT_PREFIX: Option<&str> = Some("856");
        /// `Some("001")`
        pub const INT_SELECT: Option<&str> = Some("001");
        /// `Some("(%A)%l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("(%A)%l");
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["ອາ.", "ຈ.", "ຄ.", "ພ.", "ພຫ.", "ສ.", "ສ."]`
        pub const ABDAY: &[&str] = &["ອາ.", "ຈ.", "ຄ.", "ພ.", "ພຫ.", "ສ.", "ສ."];
        /// `&["ມ.ກ.", "ກ.ພ.", "ມ.ນ.", "ມ.ສ.", "ພ.ພ.", "ມ\u{eb4}.ຖ.", "ກ.ລ.", "ສ.ຫ.", "ກ.ຍ.", "ຕ.ລ.", "ພ.ຈ.", "ທ.ວ."]`
        pub const ABMON: &[&str] = &["ມ.ກ.", "ກ.ພ.", "ມ.ນ.", "ມ.ສ.", "ພ.ພ.", "ມ\u{eb4}.ຖ.", "ກ.ລ.", "ສ.ຫ.", "ກ.ຍ.", "ຕ.ລ.", "ພ.ຈ.", "ທ.ວ."];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["AM", "PM"]`
        pub const AM_PM: &[&str] = &["AM", "PM"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a %e %b %Ey %H:%M:%S %Z")`
        pub const DATE_FMT: Option<&str> = Some("%a %e %b %Ey %H:%M:%S %Z");
        /// `&["ອາທ\u{eb4}ດ", "ຈ\u{eb1}ນ", "ອ\u{eb1}ງຄານ", "ພ\u{eb8}ດ", "ພະຫ\u{eb1}ດ", "ສ\u{eb8}ກ", "ເສ\u{ebb}າ"]`
        pub const DAY: &[&str] = &["ອາທ\u{eb4}ດ", "ຈ\u{eb1}ນ", "ອ\u{eb1}ງຄານ", "ພ\u{eb8}ດ", "ພະຫ\u{eb1}ດ", "ສ\u{eb8}ກ", "ເສ\u{ebb}າ"];
        /// `"%d/%m/%Ey"`
        pub const D_FMT: &str = "%d/%m/%Ey";
        /// `"%a %e %b %Ey, %H:%M:%S"`
        pub const D_T_FMT: &str = "%a %e %b %Ey, %H:%M:%S";
        /// `Some(&["+:1:-543/01/01:+*:ພ.ສ.:%EC %Ey"])`
        pub const ERA: Option<&[&str]> = Some(&["+:1:-543/01/01:+*:ພ.ສ.:%EC %Ey"]);
        /// `Some("%e %b %Ey")`
        pub const ERA_D_FMT: Option<&str> = Some("%e %b %Ey");
        /// `Some("ວ\u{eb1}ນ%Aທ\u{eb5}\u{ec8} %e %B %EC %Ey, %H.%M.%S ນ.")`
        pub const ERA_D_T_FMT: Option<&str> = Some("ວ\u{eb1}ນ%Aທ\u{eb5}\u{ec8} %e %B %EC %Ey, %H.%M.%S ນ.");
        /// `Some("%H.%M.%S ນ.")`
        pub const ERA_T_FMT: Option<&str> = Some("%H.%M.%S ນ.");
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["ມ\u{eb1}ງກອນ", "ກ\u{eb8}ມພາ", "ມ\u{eb5}ນາ", "ເມສາ", "ພ\u{eb6}ດສະພາ", "ມ\u{eb4}ຖ\u{eb8}ນາ", "ກ\u{ecd}ລະກ\u{ebb}ດ", "ສ\u{eb4}ງຫາ", "ກ\u{eb1}ນຍາ", "ຕ\u{eb8}ລາ", "ພະຈ\u{eb4}ກ", "ທ\u{eb1}ນວາ"]`
        pub const MON: &[&str] = &["ມ\u{eb1}ງກອນ", "ກ\u{eb8}ມພາ", "ມ\u{eb5}ນາ", "ເມສາ", "ພ\u{eb6}ດສະພາ", "ມ\u{eb4}ຖ\u{eb8}ນາ", "ກ\u{ecd}ລະກ\u{ebb}ດ", "ສ\u{eb4}ງຫາ", "ກ\u{eb1}ນຍາ", "ຕ\u{eb8}ລາ", "ພະຈ\u{eb4}ກ", "ທ\u{eb1}ນວາ"];
        /// `"%H:%M:%S"`
        pub const T_FMT: &str = "%H:%M:%S";
        /// `"%I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod lt_LT {
    pub mod LC_ADDRESS {
        /// `Some("LTU")`
        pub const COUNTRY_AB3: Option<&str> = Some("LTU");
        /// `Some("LT")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("LT");
        /// `Some("LT")`
        pub const COUNTRY_CAR: Option<&str> = Some("LT");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Lietuva")`
        pub const COUNTRY_NAME: Option<&str> = Some("Lietuva");
        /// `Some(440)`
        pub const COUNTRY_NUM: Option<i64> = Some(440);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("lt")`
        pub const LANG_AB: Option<&str> = Some("lt");
        /// `Some("lit")`
        pub const LANG_LIB: Option<&str> = Some("lit");
        /// `Some("lietuvių")`
        pub const LANG_NAME: Option<&str> = Some("lietuvių");
        /// `Some("lit")`
        pub const LANG_TERM: Option<&str> = Some("lit");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("P.O Box 1147, Donelaicio 60, 3000 Kaunas, Lithuania")`
        pub const ADDRESS: Option<&str> = Some("P.O Box 1147, Donelaicio 60, 3000 Kaunas, Lithuania");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Lithuanian")`
        pub const LANGUAGE: Option<&str> = Some("Lithuanian");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Lithuanian Computer Society and")`
        pub const SOURCE: Option<&str> = Some("Lithuanian Computer Society and");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Lithuania")`
        pub const TERRITORY: Option<&str> = Some("Lithuania");
        /// `Some("Lithuanian locale for Lithuania")`
        pub const TITLE: Option<&str> = Some("Lithuanian locale for Lithuania");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0Nn]"`
        pub const NOEXPR: &str = "^[-0Nn]";
        /// `Some("ne")`
        pub const NOSTR: Option<&str> = Some("ne");
        /// `"^[+1TtYy]"`
        pub const YESEXPR: &str = "^[+1TtYy]";
        /// `Some("taip")`
        pub const YESSTR: Option<&str> = Some("taip");
    }
    pub mod LC_MONETARY {
        /// `"€"`
        pub const CURRENCY_SYMBOL: &str = "€";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"EUR "`
        pub const INT_CURR_SYMBOL: &str = "EUR ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const MON_THOUSANDS_SEP: &str = ".";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const THOUSANDS_SEP: &str = ".";
    }
    pub mod LC_TELEPHONE {
        /// `Some("370")`
        pub const INT_PREFIX: Option<&str> = Some("370");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["Sk", "Pr", "An", "Tr", "Kt", "Pn", "Št"]`
        pub const ABDAY: &[&str] = &["Sk", "Pr", "An", "Tr", "Kt", "Pn", "Št"];
        /// `&["saus.", "vas.", "kov.", "bal.", "geg.", "birž.", "liep.", "rugp.", "rugs.", "spal.", "lapkr.", "gruod."]`
        pub const ABMON: &[&str] = &["saus.", "vas.", "kov.", "bal.", "geg.", "birž.", "liep.", "rugp.", "rugs.", "spal.", "lapkr.", "gruod."];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `Some(&["sausis", "vasaris", "kovas", "balandis", "gegužė", "birželis", "liepa", "rugpjūtis", "rugsėjis", "spalis", "lapkritis", "gruodis"])`
        pub const ALT_MON: Option<&[&str]> = Some(&["sausis", "vasaris", "kovas", "balandis", "gegužė", "birželis", "liepa", "rugpjūtis", "rugsėjis", "spalis", "lapkritis", "gruodis"]);
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Sekmadienis", "Pirmadienis", "Antradienis", "Trečiadienis", "Ketvirtadienis", "Penktadienis", "Šeštadienis"]`
        pub const DAY: &[&str] = &["Sekmadienis", "Pirmadienis", "Antradienis", "Trečiadienis", "Ketvirtadienis", "Penktadienis", "Šeštadienis"];
        /// `"%Y-%m-%d"`
        pub const D_FMT: &str = "%Y-%m-%d";
        /// `"%Y m. %B %d d. %T"`
        pub const D_T_FMT: &str = "%Y m. %B %d d. %T";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["sausio", "vasario", "kovo", "balandžio", "gegužės", "birželio", "liepos", "rugpjūčio", "rugsėjo", "spalio", "lapkričio", "gruodžio"]`
        pub const MON: &[&str] = &["sausio", "vasario", "kovo", "balandžio", "gegužės", "birželio", "liepos", "rugpjūčio", "rugsėjo", "spalio", "lapkričio", "gruodžio"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod lv_LV {
    pub mod LC_ADDRESS {
        /// `Some("LVA")`
        pub const COUNTRY_AB3: Option<&str> = Some("LVA");
        /// `Some("LV")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("LV");
        /// `Some("LV")`
        pub const COUNTRY_CAR: Option<&str> = Some("LV");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Latvija")`
        pub const COUNTRY_NAME: Option<&str> = Some("Latvija");
        /// `Some(428)`
        pub const COUNTRY_NUM: Option<i64> = Some(428);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("lv")`
        pub const LANG_AB: Option<&str> = Some("lv");
        /// `Some("lav")`
        pub const LANG_LIB: Option<&str> = Some("lav");
        /// `Some("latviešu")`
        pub const LANG_NAME: Option<&str> = Some("latviešu");
        /// `Some("lav")`
        pub const LANG_TERM: Option<&str> = Some("lav");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("LU MII, Rainis boul. 29, LV-1459 Riga, Latvia")`
        pub const ADDRESS: Option<&str> = Some("LU MII, Rainis boul. 29, LV-1459 Riga, Latvia");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Latvian")`
        pub const LANGUAGE: Option<&str> = Some("Latvian");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Latvian Standard LVS 24-93")`
        pub const SOURCE: Option<&str> = Some("Latvian Standard LVS 24-93");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Latvia")`
        pub const TERRITORY: Option<&str> = Some("Latvia");
        /// `Some("Latvian locale for Latvia")`
        pub const TITLE: Option<&str> = Some("Latvian locale for Latvia");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0Nn]"`
        pub const NOEXPR: &str = "^[-0Nn]";
        /// `Some("nē")`
        pub const NOSTR: Option<&str> = Some("nē");
        /// `"^[+1JjYy]"`
        pub const YESEXPR: &str = "^[+1JjYy]";
        /// `Some("jā")`
        pub const YESSTR: Option<&str> = Some("jā");
    }
    pub mod LC_MONETARY {
        /// `"€"`
        pub const CURRENCY_SYMBOL: &str = "€";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"EUR "`
        pub const INT_CURR_SYMBOL: &str = "EUR ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const MON_THOUSANDS_SEP: &str = "\u{202f}";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `3`
        pub const N_SIGN_POSN: i64 = 3;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `3`
        pub const P_SIGN_POSN: i64 = 3;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const THOUSANDS_SEP: &str = "\u{202f}";
    }
    pub mod LC_TELEPHONE {
        /// `Some("371")`
        pub const INT_PREFIX: Option<&str> = Some("371");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["Sv", "P\u{a0}", "O\u{a0}", "T\u{a0}", "C\u{a0}", "Pk", "S\u{a0}"]`
        pub const ABDAY: &[&str] = &["Sv", "P\u{a0}", "O\u{a0}", "T\u{a0}", "C\u{a0}", "Pk", "S\u{a0}"];
        /// `&["jan", "feb", "mar", "apr", "mai", "jūn", "jūl", "aug", "sep", "okt", "nov", "dec"]`
        pub const ABMON: &[&str] = &["jan", "feb", "mar", "apr", "mai", "jūn", "jūl", "aug", "sep", "okt", "nov", "dec"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["svētdiena", "pirmdiena", "otrdiena", "trešdiena", "ceturtdiena", "piektdiena", "sestdiena"]`
        pub const DAY: &[&str] = &["svētdiena", "pirmdiena", "otrdiena", "trešdiena", "ceturtdiena", "piektdiena", "sestdiena"];
        /// `"%Y.%m.%d."`
        pub const D_FMT: &str = "%Y.%m.%d.";
        /// `"%A, %Y. gada %e. %B, plkst. %H un %M"`
        pub const D_T_FMT: &str = "%A, %Y. gada %e. %B, plkst. %H un %M";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["janvāris", "februāris", "marts", "aprīlis", "maijs", "jūnijs", "jūlijs", "augusts", "septembris", "oktobris", "novembris", "decembris"]`
        pub const MON: &[&str] = &["janvāris", "februāris", "marts", "aprīlis", "maijs", "jūnijs", "jūlijs", "augusts", "septembris", "oktobris", "novembris", "decembris"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod lzh_TW {
    pub mod LC_ADDRESS {
        /// `Some("TWN")`
        pub const COUNTRY_AB3: Option<&str> = Some("TWN");
        /// `Some("TW")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("TW");
        /// `Some("RC")`
        pub const COUNTRY_CAR: Option<&str> = Some("RC");
        /// `Some("957")`
        pub const COUNTRY_ISBN: Option<&str> = Some("957");
        /// `Some("中華民國")`
        pub const COUNTRY_NAME: Option<&str> = Some("中華民國");
        /// `Some(158)`
        pub const COUNTRY_NUM: Option<i64> = Some(158);
        /// `Some("TW")`
        pub const COUNTRY_POST: Option<&str> = Some("TW");
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("lzh")`
        pub const LANG_LIB: Option<&str> = Some("lzh");
        /// `Some("漢語文言")`
        pub const LANG_NAME: Option<&str> = Some("漢語文言");
        /// `Some("lzh")`
        pub const LANG_TERM: Option<&str> = Some("lzh");
        /// `Some("%c%N%T%N%s %h %e %r%N%b%N%d%N%f%N%a%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%c%N%T%N%s %h %e %r%N%b%N%d%N%f%N%a%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2013-06-02")`
        pub const DATE: Option<&str> = Some("2013-06-02");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Literary Chinese")`
        pub const LANGUAGE: Option<&str> = Some("Literary Chinese");
        /// `Some("0.1")`
        pub const REVISION: Option<&str> = Some("0.1");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Taiwan")`
        pub const TERRITORY: Option<&str> = Some("Taiwan");
        /// `Some("Literary Chinese locale for the Republic of China")`
        pub const TITLE: Option<&str> = Some("Literary Chinese locale for the Republic of China");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNｎＮ非]"`
        pub const NOEXPR: &str = "^[-0nNｎＮ非]";
        /// `None`
        pub const NOSTR: Option<&str> = None;
        /// `"^[+1yYｙＹ是]"`
        pub const YESEXPR: &str = "^[+1yYｙＹ是]";
        /// `None`
        pub const YESSTR: Option<&str> = None;
    }
    pub mod LC_MONETARY {
        /// `"NT$"`
        pub const CURRENCY_SYMBOL: &str = "NT$";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"TWD "`
        pub const INT_CURR_SYMBOL: &str = "TWD ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `Some(1)`
        pub const INT_N_CS_PRECEDES: Option<i64> = Some(1);
        /// `Some(0)`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = Some(0);
        /// `Some(1)`
        pub const INT_N_SIGN_POSN: Option<i64> = Some(1);
        /// `Some(1)`
        pub const INT_P_CS_PRECEDES: Option<i64> = Some(1);
        /// `Some(0)`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = Some(0);
        /// `Some(1)`
        pub const INT_P_SIGN_POSN: Option<i64> = Some(1);
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[4]`
        pub const MON_GROUPING: &[i64] = &[4];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[4]`
        pub const GROUPING: &[i64] = &[4];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("886")`
        pub const INT_PREFIX: Option<&str> = Some("886");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `Some("%A-%l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("%A-%l");
        /// `Some("+%c-%a-%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c-%a-%l");
    }
    pub mod LC_TIME {
        /// `&["日", "一", "二", "三", "四", "五", "六"]`
        pub const ABDAY: &[&str] = &["日", "一", "二", "三", "四", "五", "六"];
        /// `&[" 一 ", " 二 ", " 三 ", " 四 ", " 五 ", " 六 ", " 七 ", " 八 ", " 九 ", " 十 ", "十一", "十二"]`
        pub const ABMON: &[&str] = &[" 一 ", " 二 ", " 三 ", " 四 ", " 五 ", " 六 ", " 七 ", " 八 ", " 九 ", " 十 ", "十一", "十二"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `Some(&["〇", "一", "二", "三", "四", "五", "六", "七", "八", "九", "十", "十一", "十二", "十三", "十四", "十五", "十六", "十七", "十八", "十九", "廿", "廿一", "廿二", "廿三", "廿四", "廿五", "廿六", "廿七", "廿八", "廿九", "卅", "卅一"])`
        pub const ALT_DIGITS: Option<&[&str]> = Some(&["〇", "一", "二", "三", "四", "五", "六", "七", "八", "九", "十", "十一", "十二", "十三", "十四", "十五", "十六", "十七", "十八", "十九", "廿", "廿一", "廿二", "廿三", "廿四", "廿五", "廿六", "廿七", "廿八", "廿九", "卅", "卅一"]);
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["朝", "暮"]`
        pub const AM_PM: &[&str] = &["朝", "暮"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("公曆 %C%Oy年 %B %Oe日 %A %OH時%OM分%OS秒")`
        pub const DATE_FMT: Option<&str> = Some("公曆 %C%Oy年 %B %Oe日 %A %OH時%OM分%OS秒");
        /// `&["週日", "週一", "週二", "週三", "週四", "週五", "週六"]`
        pub const DAY: &[&str] = &["週日", "週一", "週二", "週三", "週四", "週五", "週六"];
        /// `"%OC%Oy年%B%Od日"`
        pub const D_FMT: &str = "%OC%Oy年%B%Od日";
        /// `"%OC%Oy年%B%Od日 (%A) %OH時%OM分%OS秒"`
        pub const D_T_FMT: &str = "%OC%Oy年%B%Od日 (%A) %OH時%OM分%OS秒";
        /// `Some(&["+:2:1913/01/01:+*:民國:%EC%Ey年", "+:1:1912/01/01:1912/12/31:民國:%EC元年", "+:1:1911/12/31:-*:民前:%EC%Ey年"])`
        pub const ERA: Option<&[&str]> = Some(&["+:2:1913/01/01:+*:民國:%EC%Ey年", "+:1:1912/01/01:1912/12/31:民國:%EC元年", "+:1:1911/12/31:-*:民前:%EC%Ey年"]);
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["一月", "二月", "三月", "四月", "五月", "六月", "七月", "八月", "九月", "十月", "十一月", "十二月"]`
        pub const MON: &[&str] = &["一月", "二月", "三月", "四月", "五月", "六月", "七月", "八月", "九月", "十月", "十一月", "十二月"];
        /// `"%OH時%OM分%OS秒"`
        pub const T_FMT: &str = "%OH時%OM分%OS秒";
        /// `"%p %OI時%OM分%OS秒"`
        pub const T_FMT_AMPM: &str = "%p %OI時%OM分%OS秒";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod mag_IN {
    pub mod LC_ADDRESS {
        /// `Some("IND")`
        pub const COUNTRY_AB3: Option<&str> = Some("IND");
        /// `Some("IN")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IN");
        /// `Some("IND")`
        pub const COUNTRY_CAR: Option<&str> = Some("IND");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("भारत")`
        pub const COUNTRY_NAME: Option<&str> = Some("भारत");
        /// `Some(356)`
        pub const COUNTRY_NUM: Option<i64> = Some(356);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("mag")`
        pub const LANG_LIB: Option<&str> = Some("mag");
        /// `Some("मगही")`
        pub const LANG_NAME: Option<&str> = Some("मगही");
        /// `Some("mag")`
        pub const LANG_TERM: Option<&str> = Some("mag");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2012-04-11")`
        pub const DATE: Option<&str> = Some("2012-04-11");
        /// `Some("bhashaghar@googlegroups.com")`
        pub const EMAIL: Option<&str> = Some("bhashaghar@googlegroups.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Magahi")`
        pub const LANGUAGE: Option<&str> = Some("Magahi");
        /// `Some("0.1")`
        pub const REVISION: Option<&str> = Some("0.1");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("India")`
        pub const TERRITORY: Option<&str> = Some("India");
        /// `Some("Magahi language locale for India")`
        pub const TITLE: Option<&str> = Some("Magahi language locale for India");
    }
    pub use super::hi_IN::LC_MESSAGES;
    pub use super::hi_IN::LC_MONETARY;
    pub use super::hi_IN::LC_NUMERIC;
    pub use super::hi_IN::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["एतवार", "सोमार", "म\u{902}गर", "ब\u{941}ध", "बिफ\u{947}", "स\u{942}क", "सनिचर"]`
        pub const ABDAY: &[&str] = &["एतवार", "सोमार", "म\u{902}गर", "ब\u{941}ध", "बिफ\u{947}", "स\u{942}क", "सनिचर"];
        /// `&["जनवरी", "फ\u{93c}रवरी", "मार\u{94d}च", "अप\u{94d}र\u{947}ल", "मई", "ज\u{942}न", "ज\u{941}लाई", "अगस\u{94d}त", "सितम\u{94d}बर", "अक\u{94d}ट\u{942}बर", "नवम\u{94d}बर", "दिसम\u{94d}बर"]`
        pub const ABMON: &[&str] = &["जनवरी", "फ\u{93c}रवरी", "मार\u{94d}च", "अप\u{94d}र\u{947}ल", "मई", "ज\u{942}न", "ज\u{941}लाई", "अगस\u{94d}त", "सितम\u{94d}बर", "अक\u{94d}ट\u{942}बर", "नवम\u{94d}बर", "दिसम\u{94d}बर"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["प\u{942}र\u{94d}वाह\u{94d}न", "अपराह\u{94d}न"]`
        pub const AM_PM: &[&str] = &["प\u{942}र\u{94d}वाह\u{94d}न", "अपराह\u{94d}न"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["एतवार", "सोमार", "म\u{902}गर", "ब\u{941}ध", "बिफ\u{947}", "स\u{942}क", "सनिचर"]`
        pub const DAY: &[&str] = &["एतवार", "सोमार", "म\u{902}गर", "ब\u{941}ध", "बिफ\u{947}", "स\u{942}क", "सनिचर"];
        /// `"%-d/%-m/%y"`
        pub const D_FMT: &str = "%-d/%-m/%y";
        /// `"%A %d %b %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A %d %b %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["जनवरी", "फ\u{93c}रवरी", "मार\u{94d}च", "अप\u{94d}र\u{947}ल", "मई", "ज\u{942}न", "ज\u{941}लाई", "अगस\u{94d}त", "सितम\u{94d}बर", "अक\u{94d}ट\u{942}बर", "नवम\u{94d}बर", "दिसम\u{94d}बर"]`
        pub const MON: &[&str] = &["जनवरी", "फ\u{93c}रवरी", "मार\u{94d}च", "अप\u{94d}र\u{947}ल", "मई", "ज\u{942}न", "ज\u{941}लाई", "अगस\u{94d}त", "सितम\u{94d}बर", "अक\u{94d}ट\u{942}बर", "नवम\u{94d}बर", "दिसम\u{94d}बर"];
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT: &str = "%I:%M:%S %p %Z";
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p %Z";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod mai_IN {
    pub mod LC_ADDRESS {
        /// `Some("IND")`
        pub const COUNTRY_AB3: Option<&str> = Some("IND");
        /// `Some("IN")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IN");
        /// `Some("IND")`
        pub const COUNTRY_CAR: Option<&str> = Some("IND");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("भारत")`
        pub const COUNTRY_NAME: Option<&str> = Some("भारत");
        /// `Some(356)`
        pub const COUNTRY_NUM: Option<i64> = Some(356);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("mai")`
        pub const LANG_LIB: Option<&str> = Some("mai");
        /// `Some("म\u{948}थिली")`
        pub const LANG_NAME: Option<&str> = Some("म\u{948}थिली");
        /// `Some("mai")`
        pub const LANG_TERM: Option<&str> = Some("mai");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("B-3/302, Lunkad Daffodills, Viman Nagar, Pune, India")`
        pub const ADDRESS: Option<&str> = Some("B-3/302, Lunkad Daffodills, Viman Nagar, Pune, India");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2017-07-24")`
        pub const DATE: Option<&str> = Some("2017-07-24");
        /// `Some("rajeshkajha@yahoo.com,akhilesh.k@samusng.com")`
        pub const EMAIL: Option<&str> = Some("rajeshkajha@yahoo.com,akhilesh.k@samusng.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Maithili")`
        pub const LANGUAGE: Option<&str> = Some("Maithili");
        /// `Some("1.1")`
        pub const REVISION: Option<&str> = Some("1.1");
        /// `Some("Maithili Computing Research Center, Pune, India")`
        pub const SOURCE: Option<&str> = Some("Maithili Computing Research Center, Pune, India");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("India")`
        pub const TERRITORY: Option<&str> = Some("India");
        /// `Some("Maithili language locale for India")`
        pub const TITLE: Option<&str> = Some("Maithili language locale for India");
    }
    pub use super::hi_IN::LC_MESSAGES;
    pub use super::hi_IN::LC_MONETARY;
    pub use super::hi_IN::LC_NUMERIC;
    pub use super::hi_IN::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["रवि", "सोम", "म\u{902}गल", "ब\u{941}ध", "ब\u{943}हस\u{94d}पती", "श\u{941}क\u{94d}र", "शनी"]`
        pub const ABDAY: &[&str] = &["रवि", "सोम", "म\u{902}गल", "ब\u{941}ध", "ब\u{943}हस\u{94d}पती", "श\u{941}क\u{94d}र", "शनी"];
        /// `&["ब\u{948}साख", "ज\u{947}ठ", "अषाढ\u{93c}", "सावोन", "भादो", "आसिन", "कातिक", "अगहन", "प\u{942}स", "माघ", "फाग\u{941}न", "च\u{948}ति"]`
        pub const ABMON: &[&str] = &["ब\u{948}साख", "ज\u{947}ठ", "अषाढ\u{93c}", "सावोन", "भादो", "आसिन", "कातिक", "अगहन", "प\u{942}स", "माघ", "फाग\u{941}न", "च\u{948}ति"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["प\u{942}र\u{94d}वाह\u{94d}न", "अपराह\u{94d}न"]`
        pub const AM_PM: &[&str] = &["प\u{942}र\u{94d}वाह\u{94d}न", "अपराह\u{94d}न"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["रविदिन", "सोमदिन", "म\u{902}गलदिन", "ब\u{941}धदिन", "ब\u{943}हस\u{94d}पतीदिन", "श\u{941}क\u{94d}रदिन", "शनीदिन"]`
        pub const DAY: &[&str] = &["रविदिन", "सोमदिन", "म\u{902}गलदिन", "ब\u{941}धदिन", "ब\u{943}हस\u{94d}पतीदिन", "श\u{941}क\u{94d}रदिन", "शनीदिन"];
        /// `"%-d/%-m/%y"`
        pub const D_FMT: &str = "%-d/%-m/%y";
        /// `"%A %d %b %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A %d %b %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["ब\u{948}साख", "ज\u{947}ठ", "अषाढ\u{93c}", "सावोन", "भादो", "आसिन", "कातिक", "अगहन", "प\u{942}स", "माघ", "फाग\u{941}न", "च\u{948}ति"]`
        pub const MON: &[&str] = &["ब\u{948}साख", "ज\u{947}ठ", "अषाढ\u{93c}", "सावोन", "भादो", "आसिन", "कातिक", "अगहन", "प\u{942}स", "माघ", "फाग\u{941}न", "च\u{948}ति"];
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT: &str = "%I:%M:%S %p %Z";
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p %Z";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod mai_NP {
    pub mod LC_ADDRESS {
        /// `Some("NPL")`
        pub const COUNTRY_AB3: Option<&str> = Some("NPL");
        /// `Some("NP")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("NP");
        /// `Some("NEP")`
        pub const COUNTRY_CAR: Option<&str> = Some("NEP");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("न\u{947}पाल")`
        pub const COUNTRY_NAME: Option<&str> = Some("न\u{947}पाल");
        /// `Some(524)`
        pub const COUNTRY_NUM: Option<i64> = Some(524);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("mai")`
        pub const LANG_LIB: Option<&str> = Some("mai");
        /// `Some("म\u{948}थिली")`
        pub const LANG_NAME: Option<&str> = Some("म\u{948}थिली");
        /// `Some("mai")`
        pub const LANG_TERM: Option<&str> = Some("mai");
        /// `Some("%f%N%h%s%N%T")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%h%s%N%T");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2017-07-24")`
        pub const DATE: Option<&str> = Some("2017-07-24");
        /// `Some("akhilesh.k@samusng.com")`
        pub const EMAIL: Option<&str> = Some("akhilesh.k@samusng.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Maithili")`
        pub const LANGUAGE: Option<&str> = Some("Maithili");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Samsung Electronics Co., Ltd.")`
        pub const SOURCE: Option<&str> = Some("Samsung Electronics Co., Ltd.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Nepal")`
        pub const TERRITORY: Option<&str> = Some("Nepal");
        /// `Some("Maithili language locale for Nepal")`
        pub const TITLE: Option<&str> = Some("Maithili language locale for Nepal");
    }
    pub use super::hi_IN::LC_MESSAGES;
    pub use super::ne_NP::LC_MONETARY;
    pub use super::ne_NP::LC_NUMERIC;
    pub use super::ne_NP::LC_TELEPHONE;
    pub use super::mai_IN::LC_TIME;
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod mfe_MU {
    pub mod LC_ADDRESS {
        /// `Some("MUS")`
        pub const COUNTRY_AB3: Option<&str> = Some("MUS");
        /// `Some("MU")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("MU");
        /// `Some("MS")`
        pub const COUNTRY_CAR: Option<&str> = Some("MS");
        /// `Some("978-613,978-620,978-99903,978-99949")`
        pub const COUNTRY_ISBN: Option<&str> = Some("978-613,978-620,978-99903,978-99949");
        /// `Some("Moris")`
        pub const COUNTRY_NAME: Option<&str> = Some("Moris");
        /// `Some(480)`
        pub const COUNTRY_NUM: Option<i64> = Some(480);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("mfe")`
        pub const LANG_LIB: Option<&str> = Some("mfe");
        /// `Some("kreol morisien")`
        pub const LANG_NAME: Option<&str> = Some("kreol morisien");
        /// `Some("mfe")`
        pub const LANG_TERM: Option<&str> = Some("mfe");
        /// `Some("%f%N%h%s%N%T")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%h%s%N%T");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2017-08-18")`
        pub const DATE: Option<&str> = Some("2017-08-18");
        /// `Some("akhilesh.k@samsung.com")`
        pub const EMAIL: Option<&str> = Some("akhilesh.k@samsung.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Morisyen")`
        pub const LANGUAGE: Option<&str> = Some("Morisyen");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Samsung Electronics Co., Ltd.")`
        pub const SOURCE: Option<&str> = Some("Samsung Electronics Co., Ltd.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Mauritius")`
        pub const TERRITORY: Option<&str> = Some("Mauritius");
        /// `Some("Morisyen locale for Mauritius")`
        pub const TITLE: Option<&str> = Some("Morisyen locale for Mauritius");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("Non")`
        pub const NOSTR: Option<&str> = Some("Non");
        /// `"^[+1yYwW]"`
        pub const YESEXPR: &str = "^[+1yYwW]";
        /// `Some("Wi")`
        pub const YESSTR: Option<&str> = Some("Wi");
    }
    pub mod LC_MONETARY {
        /// `"₨"`
        pub const CURRENCY_SYMBOL: &str = "₨";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"MUR "`
        pub const INT_CURR_SYMBOL: &str = "MUR ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `Some(1)`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = Some(1);
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `Some(1)`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = Some(1);
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const MON_THOUSANDS_SEP: &str = "\u{202f}";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `"\u{202f}"`
        pub const THOUSANDS_SEP: &str = "\u{202f}";
    }
    pub mod LC_TELEPHONE {
        /// `Some("230")`
        pub const INT_PREFIX: Option<&str> = Some("230");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %l");
    }
    pub mod LC_TIME {
        /// `&["dim", "lin", "mar", "mer", "ze", "van", "sam"]`
        pub const ABDAY: &[&str] = &["dim", "lin", "mar", "mer", "ze", "van", "sam"];
        /// `&["zan", "fev", "mar", "avr", "me", "zin", "zil", "out", "sep", "okt", "nov", "des"]`
        pub const ABMON: &[&str] = &["zan", "fev", "mar", "avr", "me", "zin", "zil", "out", "sep", "okt", "nov", "des"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["AM", "PM"]`
        pub const AM_PM: &[&str] = &["AM", "PM"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a %e %b %H:%M:%S %Z %Y")`
        pub const DATE_FMT: Option<&str> = Some("%a %e %b %H:%M:%S %Z %Y");
        /// `&["dimans", "lindi", "mardi", "merkredi", "zedi", "vandredi", "samdi"]`
        pub const DAY: &[&str] = &["dimans", "lindi", "mardi", "merkredi", "zedi", "vandredi", "samdi"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["zanvie", "fevriye", "mars", "avril", "me", "zin", "zilye", "out", "septam", "oktob", "novam", "desam"]`
        pub const MON: &[&str] = &["zanvie", "fevriye", "mars", "avril", "me", "zin", "zilye", "out", "septam", "oktob", "novam", "desam"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `"%l:%M:%S %P %Z"`
        pub const T_FMT_AMPM: &str = "%l:%M:%S %P %Z";
        /// `None`
        pub const WEEK: Option<&[i64]> = None;
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod mg_MG {
    pub mod LC_ADDRESS {
        /// `Some("MDG")`
        pub const COUNTRY_AB3: Option<&str> = Some("MDG");
        /// `Some("MG")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("MG");
        /// `Some("RM")`
        pub const COUNTRY_CAR: Option<&str> = Some("RM");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Madagasikara")`
        pub const COUNTRY_NAME: Option<&str> = Some("Madagasikara");
        /// `Some(450)`
        pub const COUNTRY_NUM: Option<i64> = Some(450);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("mg")`
        pub const LANG_AB: Option<&str> = Some("mg");
        /// `Some("mlg")`
        pub const LANG_LIB: Option<&str> = Some("mlg");
        /// `Some("Malagasy")`
        pub const LANG_NAME: Option<&str> = Some("Malagasy");
        /// `Some("mlg")`
        pub const LANG_TERM: Option<&str> = Some("mlg");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Rado Ramarotafika,Do-Risika RAFIEFERANTSIARONJY")`
        pub const CONTACT: Option<&str> = Some("Rado Ramarotafika,Do-Risika RAFIEFERANTSIARONJY");
        /// `Some("2005-02-02")`
        pub const DATE: Option<&str> = Some("2005-02-02");
        /// `Some("rado@linuxmg.org,dourix@free.fr")`
        pub const EMAIL: Option<&str> = Some("rado@linuxmg.org,dourix@free.fr");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Malagasy")`
        pub const LANGUAGE: Option<&str> = Some("Malagasy");
        /// `Some("1.1")`
        pub const REVISION: Option<&str> = Some("1.1");
        /// `Some("The Debian Project modified by GNU/Linux Malagasy")`
        pub const SOURCE: Option<&str> = Some("The Debian Project modified by GNU/Linux Malagasy");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Madagascar")`
        pub const TERRITORY: Option<&str> = Some("Madagascar");
        /// `Some("Malagasy locale for Madagascar")`
        pub const TITLE: Option<&str> = Some("Malagasy locale for Madagascar");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0tTnN]"`
        pub const NOEXPR: &str = "^[-0tTnN]";
        /// `Some("Tsia")`
        pub const NOSTR: Option<&str> = Some("Tsia");
        /// `"^[+1eEyY]"`
        pub const YESEXPR: &str = "^[+1eEyY]";
        /// `Some("Eny")`
        pub const YESSTR: Option<&str> = Some("Eny");
    }
    pub mod LC_MONETARY {
        /// `"Ar"`
        pub const CURRENCY_SYMBOL: &str = "Ar";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"MGA "`
        pub const INT_CURR_SYMBOL: &str = "MGA ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const MON_THOUSANDS_SEP: &str = "\u{202f}";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[0, 0]`
        pub const GROUPING: &[i64] = &[0, 0];
        /// `""`
        pub const THOUSANDS_SEP: &str = "";
    }
    pub mod LC_TELEPHONE {
        /// `Some("261")`
        pub const INT_PREFIX: Option<&str> = Some("261");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `Some("%a %l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("%a %l");
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["lhd", "lts", "tlt", "lrb", "lkm", "zom", "sab"]`
        pub const ABDAY: &[&str] = &["lhd", "lts", "tlt", "lrb", "lkm", "zom", "sab"];
        /// `&["Jan", "Feb", "Mar", "Apr", "Mey", "Jon", "Jol", "Aog", "Sep", "Okt", "Nov", "Des"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Mar", "Apr", "Mey", "Jon", "Jol", "Aog", "Sep", "Okt", "Nov", "Des"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["alahady", "alatsinainy", "talata", "alarobia", "alakamisy", "zoma", "sabotsy"]`
        pub const DAY: &[&str] = &["alahady", "alatsinainy", "talata", "alarobia", "alakamisy", "zoma", "sabotsy"];
        /// `"%d.%m.%Y"`
        pub const D_FMT: &str = "%d.%m.%Y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Janoary", "Febroary", "Martsa", "Aprily", "Mey", "Jona", "Jolay", "Aogositra", "Septambra", "Oktobra", "Novambra", "Desambra"]`
        pub const MON: &[&str] = &["Janoary", "Febroary", "Martsa", "Aprily", "Mey", "Jona", "Jolay", "Aogositra", "Septambra", "Oktobra", "Novambra", "Desambra"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod mhr_RU {
    pub mod LC_ADDRESS {
        /// `Some("RUS")`
        pub const COUNTRY_AB3: Option<&str> = Some("RUS");
        /// `Some("RU")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("RU");
        /// `Some("RUS")`
        pub const COUNTRY_CAR: Option<&str> = Some("RUS");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `None`
        pub const COUNTRY_NAME: Option<&str> = None;
        /// `Some(643)`
        pub const COUNTRY_NUM: Option<i64> = Some(643);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("mhr")`
        pub const LANG_LIB: Option<&str> = Some("mhr");
        /// `Some("марий йылме")`
        pub const LANG_NAME: Option<&str> = Some("марий йылме");
        /// `Some("mhr")`
        pub const LANG_TERM: Option<&str> = Some("mhr");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("112, Komsomolskaya Street, Yoshkar-Ola, Mari El Republic, Russia")`
        pub const ADDRESS: Option<&str> = Some("112, Komsomolskaya Street, Yoshkar-Ola, Mari El Republic, Russia");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Vyacheslav Kileev")`
        pub const CONTACT: Option<&str> = Some("Vyacheslav Kileev");
        /// `Some("2011-04-18")`
        pub const DATE: Option<&str> = Some("2011-04-18");
        /// `Some("slavakileev@yandex.ru")`
        pub const EMAIL: Option<&str> = Some("slavakileev@yandex.ru");
        /// `None`
        pub const FAX: Option<&str> = None;
        /// `Some("Meadow Mari")`
        pub const LANGUAGE: Option<&str> = Some("Meadow Mari");
        /// `Some("0.2")`
        pub const REVISION: Option<&str> = Some("0.2");
        /// `Some("PeshSajSoft Ltd.")`
        pub const SOURCE: Option<&str> = Some("PeshSajSoft Ltd.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Russia")`
        pub const TERRITORY: Option<&str> = Some("Russia");
        /// `Some("Mari locale for Russia")`
        pub const TITLE: Option<&str> = Some("Mari locale for Russia");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNУу]"`
        pub const NOEXPR: &str = "^[-0nNУу]";
        /// `None`
        pub const NOSTR: Option<&str> = None;
        /// `"^[+1yYТт]"`
        pub const YESEXPR: &str = "^[+1yYТт]";
        /// `None`
        pub const YESSTR: Option<&str> = None;
    }
    pub mod LC_MONETARY {
        /// `"₽"`
        pub const CURRENCY_SYMBOL: &str = "₽";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"RUB "`
        pub const INT_CURR_SYMBOL: &str = "RUB ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const MON_THOUSANDS_SEP: &str = "\u{202f}";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub use super::ru_RU::LC_NUMERIC;
    pub use super::ru_RU::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Ршр", "Шчм", "Кжм", "Вгч", "Изр", "Кгр", "Шмт"]`
        pub const ABDAY: &[&str] = &["Ршр", "Шчм", "Кжм", "Вгч", "Изр", "Кгр", "Шмт"];
        /// `&["Шрк", "Пгж", "Ӱрн", "Вшр", "Ага", "Пдш", "Срм", "Срл", "Идм", "Шыж", "Клм", "Тел"]`
        pub const ABMON: &[&str] = &["Шрк", "Пгж", "Ӱрн", "Вшр", "Ага", "Пдш", "Срм", "Срл", "Идм", "Шыж", "Клм", "Тел"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%A %Y %B %d %H:%M:%S %Z")`
        pub const DATE_FMT: Option<&str> = Some("%A %Y %B %d %H:%M:%S %Z");
        /// `&["Рушарня", "Шочмо", "Кушкыжмо", "Вӱргече", "Изарня", "Кугарня", "Шуматкече"]`
        pub const DAY: &[&str] = &["Рушарня", "Шочмо", "Кушкыжмо", "Вӱргече", "Изарня", "Кугарня", "Шуматкече"];
        /// `"%Y.%m.%d"`
        pub const D_FMT: &str = "%Y.%m.%d";
        /// `"%A %Y %B %d %T"`
        pub const D_T_FMT: &str = "%A %Y %B %d %T";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Шорыкйол", "Пургыж", "Ӱярня", "Вӱдшор", "Ага", "Пеледыш", "Сӱрем", "Сорла", "Идым", "Шыжа", "Кылме", "Теле"]`
        pub const MON: &[&str] = &["Шорыкйол", "Пургыж", "Ӱярня", "Вӱдшор", "Ага", "Пеледыш", "Сӱрем", "Сорла", "Идым", "Шыжа", "Кылме", "Теле"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod mi_NZ {
    pub mod LC_ADDRESS {
        /// `Some("NZL")`
        pub const COUNTRY_AB3: Option<&str> = Some("NZL");
        /// `Some("NZ")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("NZ");
        /// `Some("NZ")`
        pub const COUNTRY_CAR: Option<&str> = Some("NZ");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Aotearoa")`
        pub const COUNTRY_NAME: Option<&str> = Some("Aotearoa");
        /// `Some(554)`
        pub const COUNTRY_NUM: Option<i64> = Some(554);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("mi")`
        pub const LANG_AB: Option<&str> = Some("mi");
        /// `Some("mao")`
        pub const LANG_LIB: Option<&str> = Some("mao");
        /// `Some("Māori")`
        pub const LANG_NAME: Option<&str> = Some("Māori");
        /// `Some("mri")`
        pub const LANG_TERM: Option<&str> = Some("mri");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_MONETARY"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_MONETARY"],
        ]);
        /// `Some("James Gasson, Pablo Saratxaga")`
        pub const CONTACT: Option<&str> = Some("James Gasson, Pablo Saratxaga");
        /// `Some("2001-01-28")`
        pub const DATE: Option<&str> = Some("2001-01-28");
        /// `Some("james.gasson@clear.net.nz, pablo@mandrakesoft.com")`
        pub const EMAIL: Option<&str> = Some("james.gasson@clear.net.nz, pablo@mandrakesoft.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Maori")`
        pub const LANGUAGE: Option<&str> = Some("Maori");
        /// `Some("0.3")`
        pub const REVISION: Option<&str> = Some("0.3");
        /// `Some("James Gasson")`
        pub const SOURCE: Option<&str> = Some("James Gasson");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("New Zealand")`
        pub const TERRITORY: Option<&str> = Some("New Zealand");
        /// `Some("Maori language locale for New Zealand")`
        pub const TITLE: Option<&str> = Some("Maori language locale for New Zealand");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNkK]"`
        pub const NOEXPR: &str = "^[-0nNkK]";
        /// `Some("kāore")`
        pub const NOSTR: Option<&str> = Some("kāore");
        /// `"^[+1yYāĀäÄaA]"`
        pub const YESEXPR: &str = "^[+1yYāĀäÄaA]";
        /// `Some("āe")`
        pub const YESSTR: Option<&str> = Some("āe");
    }
    pub use super::en_NZ::LC_MONETARY;
    pub use super::en_NZ::LC_NUMERIC;
    pub use super::en_NZ::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Ta", "Ma", "Tū", "We", "Tāi", "Pa", "Hā"]`
        pub const ABDAY: &[&str] = &["Ta", "Ma", "Tū", "We", "Tāi", "Pa", "Hā"];
        /// `&["Kohi", "Hui", "Pou", "Pae", "Hara", "Pipi", "Hōngoi", "Here", "Mahu", "Whi-nu", "Whi-ra", "Haki"]`
        pub const ABMON: &[&str] = &["Kohi", "Hui", "Pou", "Pae", "Hara", "Pipi", "Hōngoi", "Here", "Mahu", "Whi-nu", "Whi-ra", "Haki"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Rātapu", "Mane", "Tūrei", "Wenerei", "Tāite", "Paraire", "Hātarei"]`
        pub const DAY: &[&str] = &["Rātapu", "Mane", "Tūrei", "Wenerei", "Tāite", "Paraire", "Hātarei"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"Te %A, te %d o %B, %Y %T %Z"`
        pub const D_T_FMT: &str = "Te %A, te %d o %B, %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Kohi-tātea", "Hui-tanguru", "Poutū-te-rangi", "Paenga-whāwhā", "Haratua", "Pipiri", "Hōngoingoi", "Here-turi-kōkā", "Mahuru", "Whiringa-ā-nuku", "Whiringa-ā-rangi", "Hakihea"]`
        pub const MON: &[&str] = &["Kohi-tātea", "Hui-tanguru", "Poutū-te-rangi", "Paenga-whāwhā", "Haratua", "Pipiri", "Hōngoingoi", "Here-turi-kōkā", "Mahuru", "Whiringa-ā-nuku", "Whiringa-ā-rangi", "Hakihea"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod miq_NI {
    pub mod LC_ADDRESS {
        /// `Some("NIC")`
        pub const COUNTRY_AB3: Option<&str> = Some("NIC");
        /// `Some("NI")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("NI");
        /// `Some("NIC")`
        pub const COUNTRY_CAR: Option<&str> = Some("NIC");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Nicaragua")`
        pub const COUNTRY_NAME: Option<&str> = Some("Nicaragua");
        /// `Some(558)`
        pub const COUNTRY_NUM: Option<i64> = Some(558);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("miq")`
        pub const LANG_LIB: Option<&str> = Some("miq");
        /// `Some("Mískitu")`
        pub const LANG_NAME: Option<&str> = Some("Mískitu");
        /// `Some("miq")`
        pub const LANG_TERM: Option<&str> = Some("miq");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `None`
        pub const ADDRESS: Option<&str> = None;
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
        ]);
        /// `None`
        pub const CONTACT: Option<&str> = None;
        /// `Some("2016-08-20")`
        pub const DATE: Option<&str> = Some("2016-08-20");
        /// `None`
        pub const EMAIL: Option<&str> = None;
        /// `None`
        pub const FAX: Option<&str> = None;
        /// `Some("Miskito")`
        pub const LANGUAGE: Option<&str> = Some("Miskito");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `None`
        pub const SOURCE: Option<&str> = None;
        /// `None`
        pub const TEL: Option<&str> = None;
        /// `Some("Nicaragua")`
        pub const TERRITORY: Option<&str> = Some("Nicaragua");
        /// `Some("Miskito language locale for Nicaragua")`
        pub const TITLE: Option<&str> = Some("Miskito language locale for Nicaragua");
    }
    pub use super::es_NI::LC_MESSAGES;
    pub use super::es_NI::LC_MONETARY;
    pub use super::es_NI::LC_NUMERIC;
    pub use super::es_NI::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["san", "mun", "tius", "wens", "taus", "prai", "sat"]`
        pub const ABDAY: &[&str] = &["san", "mun", "tius", "wens", "taus", "prai", "sat"];
        /// `&["siakwa kati", "kuswa kati", "kakamuk kati", "lî wainhka kati", "lih mairin kati", "lî kati", "pastara kati", "sikla kati", "wîs kati", "waupasa kati", "yahbra kati", "trisu kati"]`
        pub const ABMON: &[&str] = &["siakwa kati", "kuswa kati", "kakamuk kati", "lî wainhka kati", "lih mairin kati", "lî kati", "pastara kati", "sikla kati", "wîs kati", "waupasa kati", "yahbra kati", "trisu kati"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["VM", "NM"]`
        pub const AM_PM: &[&str] = &["VM", "NM"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a %b %e %H:%M:%S %Z %Y")`
        pub const DATE_FMT: Option<&str> = Some("%a %b %e %H:%M:%S %Z %Y");
        /// `&["sandi", "mundi", "tiusdi", "wensde", "tausde", "praidi", "satadi"]`
        pub const DAY: &[&str] = &["sandi", "mundi", "tiusdi", "wensde", "tausde", "praidi", "satadi"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(1)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(1);
        /// `Some(2)`
        pub const FIRST_WORKDAY: Option<i64> = Some(2);
        /// `&["siakwa kati", "kuswa kati", "kakamuk kati", "lî wainhka kati", "lih mairin kati", "lî kati", "pastara kati", "sikla kati", "wîs kati", "waupasa kati", "yahbra kati", "trisu kati"]`
        pub const MON: &[&str] = &["siakwa kati", "kuswa kati", "kakamuk kati", "lî wainhka kati", "lih mairin kati", "lî kati", "pastara kati", "sikla kati", "wîs kati", "waupasa kati", "yahbra kati", "trisu kati"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `"%I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p";
        /// `Some(&[7, 19971130, 7])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 7]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod mjw_IN {
    pub mod LC_ADDRESS {
        /// `Some("IND")`
        pub const COUNTRY_AB3: Option<&str> = Some("IND");
        /// `Some("IN")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IN");
        /// `Some("IND")`
        pub const COUNTRY_CAR: Option<&str> = Some("IND");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Bhorot")`
        pub const COUNTRY_NAME: Option<&str> = Some("Bhorot");
        /// `Some(356)`
        pub const COUNTRY_NUM: Option<i64> = Some(356);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("mjw")`
        pub const LANG_LIB: Option<&str> = Some("mjw");
        /// `Some("Karbi")`
        pub const LANG_NAME: Option<&str> = Some("Karbi");
        /// `Some("mjw")`
        pub const LANG_TERM: Option<&str> = Some("mjw");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2013-08-08")`
        pub const DATE: Option<&str> = Some("2013-08-08");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Karbi")`
        pub const LANGUAGE: Option<&str> = Some("Karbi");
        /// `Some("1.1")`
        pub const REVISION: Option<&str> = Some("1.1");
        /// `Some("Jor Teron")`
        pub const SOURCE: Option<&str> = Some("Jor Teron");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("India")`
        pub const TERRITORY: Option<&str> = Some("India");
        /// `Some("Karbi language locale for India")`
        pub const TITLE: Option<&str> = Some("Karbi language locale for India");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("no")`
        pub const NOSTR: Option<&str> = Some("no");
        /// `"^[+1yY]"`
        pub const YESEXPR: &str = "^[+1yY]";
        /// `Some("yes")`
        pub const YESSTR: Option<&str> = Some("yes");
    }
    pub use super::hi_IN::LC_MONETARY;
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3, 2]`
        pub const GROUPING: &[i64] = &[3, 2];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub use super::hi_IN::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Bhom", "Ur", "Dur", "Tkel", "Tkem", "Bhta", "Bhti"]`
        pub const ABDAY: &[&str] = &["Bhom", "Ur", "Dur", "Tkel", "Tkem", "Bhta", "Bhti"];
        /// `&["Ark", "Thang", "The", "Jang", "Aru", "Vos", "Jak", "Pai", "Chi", "Phe", "Phai", "Mati"]`
        pub const ABMON: &[&str] = &["Ark", "Thang", "The", "Jang", "Aru", "Vos", "Jak", "Pai", "Chi", "Phe", "Phai", "Mati"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["AM", "PM"]`
        pub const AM_PM: &[&str] = &["AM", "PM"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Bhomkuru", "Urmi", "Durmi", "Thelang", "Theman", "Bhomta", "Bhomti"]`
        pub const DAY: &[&str] = &["Bhomkuru", "Urmi", "Durmi", "Thelang", "Theman", "Bhomta", "Bhomti"];
        /// `"%-d/%-m/%y"`
        pub const D_FMT: &str = "%-d/%-m/%y";
        /// `"%A %d %B %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A %d %B %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(1)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(1);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Arkoi", "Thangthang", "There", "Jangmi", "Aru", "Vosik", "Jakhong", "Paipai", "Chiti", "Phere", "Phaikuni", "Matijong"]`
        pub const MON: &[&str] = &["Arkoi", "Thangthang", "There", "Jangmi", "Aru", "Vosik", "Jakhong", "Paipai", "Chiti", "Phere", "Phaikuni", "Matijong"];
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT: &str = "%I:%M:%S %p %Z";
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p %Z";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod mk_MK {
    pub mod LC_ADDRESS {
        /// `Some("MKD")`
        pub const COUNTRY_AB3: Option<&str> = Some("MKD");
        /// `Some("MK")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("MK");
        /// `Some("MK")`
        pub const COUNTRY_CAR: Option<&str> = Some("MK");
        /// `Some("9989")`
        pub const COUNTRY_ISBN: Option<&str> = Some("9989");
        /// `Some("Македонија")`
        pub const COUNTRY_NAME: Option<&str> = Some("Македонија");
        /// `Some(807)`
        pub const COUNTRY_NUM: Option<i64> = Some(807);
        /// `Some("MK")`
        pub const COUNTRY_POST: Option<&str> = Some("MK");
        /// `Some("mk")`
        pub const LANG_AB: Option<&str> = Some("mk");
        /// `Some("mac")`
        pub const LANG_LIB: Option<&str> = Some("mac");
        /// `Some("македонски")`
        pub const LANG_NAME: Option<&str> = Some("македонски");
        /// `Some("mkd")`
        pub const LANG_TERM: Option<&str> = Some("mkd");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Damjan Georgievski")`
        pub const CONTACT: Option<&str> = Some("Damjan Georgievski");
        /// `Some("2006-09-12")`
        pub const DATE: Option<&str> = Some("2006-09-12");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Macedonian")`
        pub const LANGUAGE: Option<&str> = Some("Macedonian");
        /// `Some("2.2")`
        pub const REVISION: Option<&str> = Some("2.2");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Macedonia")`
        pub const TERRITORY: Option<&str> = Some("Macedonia");
        /// `Some("Macedonian locale for Macedonia")`
        pub const TITLE: Option<&str> = Some("Macedonian locale for Macedonia");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNНн]"`
        pub const NOEXPR: &str = "^[-0nNНн]";
        /// `Some("не")`
        pub const NOSTR: Option<&str> = Some("не");
        /// `"^[+1yYДдdD]"`
        pub const YESEXPR: &str = "^[+1yYДдdD]";
        /// `Some("да")`
        pub const YESSTR: Option<&str> = Some("да");
    }
    pub mod LC_MONETARY {
        /// `"ден"`
        pub const CURRENCY_SYMBOL: &str = "ден";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"MKD "`
        pub const INT_CURR_SYMBOL: &str = "MKD ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const MON_THOUSANDS_SEP: &str = "\u{202f}";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const THOUSANDS_SEP: &str = "\u{202f}";
    }
    pub mod LC_TELEPHONE {
        /// `Some("389")`
        pub const INT_PREFIX: Option<&str> = Some("389");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `Some("%A %l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("%A %l");
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["нед", "пон", "вто", "сре", "чет", "пет", "саб"]`
        pub const ABDAY: &[&str] = &["нед", "пон", "вто", "сре", "чет", "пет", "саб"];
        /// `&["јан", "фев", "мар", "апр", "мај", "јун", "јул", "авг", "сеп", "окт", "ное", "дек"]`
        pub const ABMON: &[&str] = &["јан", "фев", "мар", "апр", "мај", "јун", "јул", "авг", "сеп", "окт", "ное", "дек"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a, %d %b %H:%M:%S %Z %Y")`
        pub const DATE_FMT: Option<&str> = Some("%a, %d %b %H:%M:%S %Z %Y");
        /// `&["недела", "понеделник", "вторник", "среда", "четврток", "петок", "сабота"]`
        pub const DAY: &[&str] = &["недела", "понеделник", "вторник", "среда", "четврток", "петок", "сабота"];
        /// `"%d.%m.%Y"`
        pub const D_FMT: &str = "%d.%m.%Y";
        /// `"%a, %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a, %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["јануари", "февруари", "март", "април", "мај", "јуни", "јули", "август", "септември", "октомври", "ноември", "декември"]`
        pub const MON: &[&str] = &["јануари", "февруари", "март", "април", "мај", "јуни", "јули", "август", "септември", "октомври", "ноември", "декември"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ml_IN {
    pub mod LC_ADDRESS {
        /// `Some("IND")`
        pub const COUNTRY_AB3: Option<&str> = Some("IND");
        /// `Some("IN")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IN");
        /// `Some("IND")`
        pub const COUNTRY_CAR: Option<&str> = Some("IND");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("ഇന\u{d4d}ത\u{d4d}യ")`
        pub const COUNTRY_NAME: Option<&str> = Some("ഇന\u{d4d}ത\u{d4d}യ");
        /// `Some(356)`
        pub const COUNTRY_NUM: Option<i64> = Some(356);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ml")`
        pub const LANG_AB: Option<&str> = Some("ml");
        /// `Some("mal")`
        pub const LANG_LIB: Option<&str> = Some("mal");
        /// `Some("മലയ\u{d3e}ളം")`
        pub const LANG_NAME: Option<&str> = Some("മലയ\u{d3e}ളം");
        /// `Some("mal")`
        pub const LANG_TERM: Option<&str> = Some("mal");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2003-February-01")`
        pub const DATE: Option<&str> = Some("2003-February-01");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Malayalam")`
        pub const LANGUAGE: Option<&str> = Some("Malayalam");
        /// `Some("0.1")`
        pub const REVISION: Option<&str> = Some("0.1");
        /// `Some("Free Software Foundation of India, Trivandrum")`
        pub const SOURCE: Option<&str> = Some("Free Software Foundation of India, Trivandrum");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("India")`
        pub const TERRITORY: Option<&str> = Some("India");
        /// `Some("Malayalam language locale for India")`
        pub const TITLE: Option<&str> = Some("Malayalam language locale for India");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNഅ]"`
        pub const NOEXPR: &str = "^[-0nNഅ]";
        /// `Some("അല\u{d4d}ല")`
        pub const NOSTR: Option<&str> = Some("അല\u{d4d}ല");
        /// `"^[+1yYഉ]"`
        pub const YESEXPR: &str = "^[+1yYഉ]";
        /// `Some("ഉവ\u{d4d}വ\u{d4d}")`
        pub const YESSTR: Option<&str> = Some("ഉവ\u{d4d}വ\u{d4d}");
    }
    pub use super::hi_IN::LC_MONETARY;
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3, 2]`
        pub const GROUPING: &[i64] = &[3, 2];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub use super::hi_IN::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["ഞ\u{d3e}", "തി", "ചൊ", "ബ\u{d41}", "വ\u{d4d}യ\u{d3e}", "വെ", "ശ"]`
        pub const ABDAY: &[&str] = &["ഞ\u{d3e}", "തി", "ചൊ", "ബ\u{d41}", "വ\u{d4d}യ\u{d3e}", "വെ", "ശ"];
        /// `&["ജന\u{d41}", "ഫെബ\u{d4d}ര\u{d41}", "മ\u{d3e}ർ", "ഏപ\u{d4d}രി", "മേയ\u{d4d}", "ജ\u{d42}ൺ", "ജ\u{d42}ലൈ", "ഓഗ", "സെപ\u{d4d}റ\u{d4d}റം", "ഒക\u{d4d}ടോ", "നവം", "ഡിസം"]`
        pub const ABMON: &[&str] = &["ജന\u{d41}", "ഫെബ\u{d4d}ര\u{d41}", "മ\u{d3e}ർ", "ഏപ\u{d4d}രി", "മേയ\u{d4d}", "ജ\u{d42}ൺ", "ജ\u{d42}ലൈ", "ഓഗ", "സെപ\u{d4d}റ\u{d4d}റം", "ഒക\u{d4d}ടോ", "നവം", "ഡിസം"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["ര\u{d3e}വിലെ", "വൈക\u{d41}"]`
        pub const AM_PM: &[&str] = &["ര\u{d3e}വിലെ", "വൈക\u{d41}"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["ഞ\u{d3e}യര\u{d4d}\u{200d}", "തിങ\u{d4d}കള\u{d4d}\u{200d}", "ചൊവ\u{d4d}വ", "ബ\u{d41}ധന\u{d4d}\u{200d}", "വ\u{d4d}യ\u{d3e}ഴം", "വെള\u{d4d}ളി", "ശനി"]`
        pub const DAY: &[&str] = &["ഞ\u{d3e}യര\u{d4d}\u{200d}", "തിങ\u{d4d}കള\u{d4d}\u{200d}", "ചൊവ\u{d4d}വ", "ബ\u{d41}ധന\u{d4d}\u{200d}", "വ\u{d4d}യ\u{d3e}ഴം", "വെള\u{d4d}ളി", "ശനി"];
        /// `"%-d/%-m/%y"`
        pub const D_FMT: &str = "%-d/%-m/%y";
        /// `"%A %d %B %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A %d %B %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["ജന\u{d41}വരി", "ഫെബ\u{d4d}ര\u{d41}വരി", "മ\u{d3e}ർച\u{d4d}ച\u{d4d}", "ഏപ\u{d4d}രിൽ", "മേയ\u{d4d}", "ജ\u{d42}ൺ", "ജ\u{d42}ലൈ", "ഓഗസ\u{d4d}റ\u{d4d}റ\u{d4d}", "സെപ\u{d4d}റ\u{d4d}റംബർ", "ഒക\u{d4d}\u{200c}ടോബർ", "നവംബർ", "ഡിസംബർ"]`
        pub const MON: &[&str] = &["ജന\u{d41}വരി", "ഫെബ\u{d4d}ര\u{d41}വരി", "മ\u{d3e}ർച\u{d4d}ച\u{d4d}", "ഏപ\u{d4d}രിൽ", "മേയ\u{d4d}", "ജ\u{d42}ൺ", "ജ\u{d42}ലൈ", "ഓഗസ\u{d4d}റ\u{d4d}റ\u{d4d}", "സെപ\u{d4d}റ\u{d4d}റംബർ", "ഒക\u{d4d}\u{200c}ടോബർ", "നവംബർ", "ഡിസംബർ"];
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT: &str = "%I:%M:%S %p %Z";
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p %Z";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod mn_MN {
    pub mod LC_ADDRESS {
        /// `Some("MNG")`
        pub const COUNTRY_AB3: Option<&str> = Some("MNG");
        /// `Some("MN")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("MN");
        /// `Some("MGL")`
        pub const COUNTRY_CAR: Option<&str> = Some("MGL");
        /// `Some("99929")`
        pub const COUNTRY_ISBN: Option<&str> = Some("99929");
        /// `Some("Монгол")`
        pub const COUNTRY_NAME: Option<&str> = Some("Монгол");
        /// `Some(496)`
        pub const COUNTRY_NUM: Option<i64> = Some(496);
        /// `Some("MN")`
        pub const COUNTRY_POST: Option<&str> = Some("MN");
        /// `Some("mn")`
        pub const LANG_AB: Option<&str> = Some("mn");
        /// `Some("mon")`
        pub const LANG_LIB: Option<&str> = Some("mon");
        /// `Some("монгол")`
        pub const LANG_NAME: Option<&str> = Some("монгол");
        /// `Some("mon")`
        pub const LANG_TERM: Option<&str> = Some("mon");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2005-05-21")`
        pub const DATE: Option<&str> = Some("2005-05-21");
        /// `Some("badral@chinggis.com")`
        pub const EMAIL: Option<&str> = Some("badral@chinggis.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Mongolian")`
        pub const LANGUAGE: Option<&str> = Some("Mongolian");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Sanlig Badral")`
        pub const SOURCE: Option<&str> = Some("Sanlig Badral");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Mongolia")`
        pub const TERRITORY: Option<&str> = Some("Mongolia");
        /// `Some("Mongolian locale for Mongolia")`
        pub const TITLE: Option<&str> = Some("Mongolian locale for Mongolia");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0үҮnN]"`
        pub const NOEXPR: &str = "^[-0үҮnN]";
        /// `Some("үгүй")`
        pub const NOSTR: Option<&str> = Some("үгүй");
        /// `"^[+1тТyY]"`
        pub const YESEXPR: &str = "^[+1тТyY]";
        /// `Some("тийм")`
        pub const YESSTR: Option<&str> = Some("тийм");
    }
    pub mod LC_MONETARY {
        /// `"₮"`
        pub const CURRENCY_SYMBOL: &str = "₮";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"MNT "`
        pub const INT_CURR_SYMBOL: &str = "MNT ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const MON_THOUSANDS_SEP: &str = "\u{202f}";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const THOUSANDS_SEP: &str = ".";
    }
    pub mod LC_TELEPHONE {
        /// `Some("976")`
        pub const INT_PREFIX: Option<&str> = Some("976");
        /// `Some("001")`
        pub const INT_SELECT: Option<&str> = Some("001");
        /// `Some("(0%a) %l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("(0%a) %l");
        /// `Some("(+%c %a) %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("(+%c %a) %l");
    }
    pub mod LC_TIME {
        /// `&["Ня", "Да", "Мя", "Лх", "Пү", "Ба", "Бя"]`
        pub const ABDAY: &[&str] = &["Ня", "Да", "Мя", "Лх", "Пү", "Ба", "Бя"];
        /// `&["1-р сар", "2-р сар", "3-р сар", "4-р сар", "5-р сар", "6-р сар", "7-р сар", "8-р сар", "9-р сар", "10-р сар", "11-р сар", "12-р сар"]`
        pub const ABMON: &[&str] = &["1-р сар", "2-р сар", "3-р сар", "4-р сар", "5-р сар", "6-р сар", "7-р сар", "8-р сар", "9-р сар", "10-р сар", "11-р сар", "12-р сар"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%Z %Y оны %B %e, %a %H:%M:%S")`
        pub const DATE_FMT: Option<&str> = Some("%Z %Y оны %B %e, %a %H:%M:%S");
        /// `&["Ням", "Даваа", "Мягмар", "Лхагва", "Пүрэв", "Баасан", "Бямба"]`
        pub const DAY: &[&str] = &["Ням", "Даваа", "Мягмар", "Лхагва", "Пүрэв", "Баасан", "Бямба"];
        /// `"%Y.%m.%d"`
        pub const D_FMT: &str = "%Y.%m.%d";
        /// `"%Y %b %d, %a %T"`
        pub const D_T_FMT: &str = "%Y %b %d, %a %T";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Нэгдүгээр сар", "Хоёрдугаар сар", "Гуравдугаар сар", "Дөрөвдүгээр сар", "Тавдугаар сар", "Зургадугаар сар", "Долдугаар сар", "Наймдугаар сар", "Есдүгээр сар", "Аравдугаар сар", "Арван нэгдүгээр сар", "Арван хоёрдугаар сар"]`
        pub const MON: &[&str] = &["Нэгдүгээр сар", "Хоёрдугаар сар", "Гуравдугаар сар", "Дөрөвдүгээр сар", "Тавдугаар сар", "Зургадугаар сар", "Долдугаар сар", "Наймдугаар сар", "Есдүгээр сар", "Аравдугаар сар", "Арван нэгдүгээр сар", "Арван хоёрдугаар сар"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod mni_IN {
    pub mod LC_ADDRESS {
        /// `Some("IND")`
        pub const COUNTRY_AB3: Option<&str> = Some("IND");
        /// `Some("IN")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IN");
        /// `Some("IND")`
        pub const COUNTRY_CAR: Option<&str> = Some("IND");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("ভ\u{9be}রত")`
        pub const COUNTRY_NAME: Option<&str> = Some("ভ\u{9be}রত");
        /// `Some(356)`
        pub const COUNTRY_NUM: Option<i64> = Some(356);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("mni")`
        pub const LANG_LIB: Option<&str> = Some("mni");
        /// `Some("মৈতৈলোন\u{9cd}")`
        pub const LANG_NAME: Option<&str> = Some("মৈতৈলোন\u{9cd}");
        /// `Some("mni")`
        pub const LANG_TERM: Option<&str> = Some("mni");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Level 1, Tower X, Cybercity, Magarpatta City, Hadapsar, Pune-411013 ")`
        pub const ADDRESS: Option<&str> = Some("Level 1, Tower X, Cybercity, Magarpatta City, Hadapsar, Pune-411013 ");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2012-01-17")`
        pub const DATE: Option<&str> = Some("2012-01-17");
        /// `Some("libc-alpha@sourceware.org")`
        pub const EMAIL: Option<&str> = Some("libc-alpha@sourceware.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Manipuri")`
        pub const LANGUAGE: Option<&str> = Some("Manipuri");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Red Hat Pune")`
        pub const SOURCE: Option<&str> = Some("Red Hat Pune");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("India")`
        pub const TERRITORY: Option<&str> = Some("India");
        /// `Some("Manipuri language locale for India")`
        pub const TITLE: Option<&str> = Some("Manipuri language locale for India");
    }
    pub use super::en_IN::LC_MESSAGES;
    pub use super::bn_IN::LC_MONETARY;
    pub use super::bn_IN::LC_NUMERIC;
    pub use super::bn_IN::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["নোং", "নিং", "লৈব\u{9be}ক", "য়\u{9c1}ম", "শগোল", "ইর\u{9be}", "থ\u{9be}ং"]`
        pub const ABDAY: &[&str] = &["নোং", "নিং", "লৈব\u{9be}ক", "য়\u{9c1}ম", "শগোল", "ইর\u{9be}", "থ\u{9be}ং"];
        /// `&["জ\u{9be}ন", "ফেব", "ম\u{9be}র", "এপ\u{9cd}রি", "মে", "জ\u{9c1}ন", "জ\u{9c1}ল", "আগ", "সেপ", "ওক\u{9cd}ত", "নবে", "ডিস"]`
        pub const ABMON: &[&str] = &["জ\u{9be}ন", "ফেব", "ম\u{9be}র", "এপ\u{9cd}রি", "মে", "জ\u{9c1}ন", "জ\u{9c1}ল", "আগ", "সেপ", "ওক\u{9cd}ত", "নবে", "ডিস"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["এ.ম.", "প.ম."]`
        pub const AM_PM: &[&str] = &["এ.ম.", "প.ম."];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["নোংম\u{9be}ইজিং", "নিংথৌক\u{9be}ব\u{9be}", "লৈব\u{9be}কপোকপ\u{9be}", "য়\u{9c1}মশকৈশ\u{9be}", "শগোলশেন", "ইর\u{9be}ই", "থ\u{9be}ংজ"]`
        pub const DAY: &[&str] = &["নোংম\u{9be}ইজিং", "নিংথৌক\u{9be}ব\u{9be}", "লৈব\u{9be}কপোকপ\u{9be}", "য়\u{9c1}মশকৈশ\u{9be}", "শগোলশেন", "ইর\u{9be}ই", "থ\u{9be}ংজ"];
        /// `"%-d/%-m/%y"`
        pub const D_FMT: &str = "%-d/%-m/%y";
        /// `"%A %d %b %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A %d %b %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["জ\u{9be}ন\u{9c1}ৱ\u{9be}রি", "ফেব\u{9cd}র\u{9c1}ৱ\u{9be}রি", "ম\u{9be}র\u{9cd}চ", "এপ\u{9cd}রিল", "মে", "জ\u{9c1}ন", "জ\u{9c1}ল\u{9be}ই", "আগষ\u{9cd}ট", "সেপ\u{9cd}তেম\u{9cd}বর", "ওক\u{9cd}তোবর", "নবেম\u{9cd}বর", "ডিসেম\u{9cd}বর"]`
        pub const MON: &[&str] = &["জ\u{9be}ন\u{9c1}ৱ\u{9be}রি", "ফেব\u{9cd}র\u{9c1}ৱ\u{9be}রি", "ম\u{9be}র\u{9cd}চ", "এপ\u{9cd}রিল", "মে", "জ\u{9c1}ন", "জ\u{9c1}ল\u{9be}ই", "আগষ\u{9cd}ট", "সেপ\u{9cd}তেম\u{9cd}বর", "ওক\u{9cd}তোবর", "নবেম\u{9cd}বর", "ডিসেম\u{9cd}বর"];
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT: &str = "%I:%M:%S %p %Z";
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p %Z";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod mnw_MM {
    pub mod LC_ADDRESS {
        /// `Some("MMR")`
        pub const COUNTRY_AB3: Option<&str> = Some("MMR");
        /// `Some("MM")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("MM");
        /// `Some("MYA")`
        pub const COUNTRY_CAR: Option<&str> = Some("MYA");
        /// `Some("99971")`
        pub const COUNTRY_ISBN: Option<&str> = Some("99971");
        /// `Some("ဗ\u{105f}ာ")`
        pub const COUNTRY_NAME: Option<&str> = Some("ဗ\u{105f}ာ");
        /// `Some(104)`
        pub const COUNTRY_NUM: Option<i64> = Some(104);
        /// `Some("Myanmar")`
        pub const COUNTRY_POST: Option<&str> = Some("Myanmar");
        /// `Some("")`
        pub const LANG_AB: Option<&str> = Some("");
        /// `Some("mnw")`
        pub const LANG_LIB: Option<&str> = Some("mnw");
        /// `Some("မန\u{103a}")`
        pub const LANG_NAME: Option<&str> = Some("မန\u{103a}");
        /// `Some("mnw")`
        pub const LANG_TERM: Option<&str> = Some("mnw");
        /// `Some("%a%N%d%N%f%N%b%N%h%t%r%t%e%t%s%N%T%t%z%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%a%N%d%N%f%N%b%N%h%t%r%t%e%t%s%N%T%t%z%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Mawlamyine, Mon State, Myanmar")`
        pub const ADDRESS: Option<&str> = Some("Mawlamyine, Mon State, Myanmar");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
        ]);
        /// `Some("Talachan Mon")`
        pub const CONTACT: Option<&str> = Some("Talachan Mon");
        /// `Some("2019-10-27")`
        pub const DATE: Option<&str> = Some("2019-10-27");
        /// `Some("talachanmon@gmail.com")`
        pub const EMAIL: Option<&str> = Some("talachanmon@gmail.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Mon")`
        pub const LANGUAGE: Option<&str> = Some("Mon");
        /// `Some("1")`
        pub const REVISION: Option<&str> = Some("1");
        /// `Some("Mon Localization Team https://facebook.com/groups/monlocalization")`
        pub const SOURCE: Option<&str> = Some("Mon Localization Team https://facebook.com/groups/monlocalization");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Myanmar")`
        pub const TERRITORY: Option<&str> = Some("Myanmar");
        /// `Some("Mon language locale for Myanmar")`
        pub const TITLE: Option<&str> = Some("Mon language locale for Myanmar");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNဟ]"`
        pub const NOEXPR: &str = "^[-0nNဟ]";
        /// `Some("ဟ\u{103d}\u{1036}")`
        pub const NOSTR: Option<&str> = Some("ဟ\u{103d}\u{1036}");
        /// `"^[+1yYယ]"`
        pub const YESEXPR: &str = "^[+1yYယ]";
        /// `Some("ယ\u{103d}\u{1036}")`
        pub const YESSTR: Option<&str> = Some("ယ\u{103d}\u{1036}");
    }
    pub mod LC_MONETARY {
        /// `"K"`
        pub const CURRENCY_SYMBOL: &str = "K";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"MMK "`
        pub const INT_CURR_SYMBOL: &str = "MMK ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub use super::my_MM::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["ဒ\u{102d}\u{102f}တ\u{103a}", "စန\u{103a}", "ၚာ", "ဝါ", "တ\u{102d}", "သ\u{102d}\u{102f}က\u{103a}", "သဝ\u{103a}"]`
        pub const ABDAY: &[&str] = &["ဒ\u{102d}\u{102f}တ\u{103a}", "စန\u{103a}", "ၚာ", "ဝါ", "တ\u{102d}", "သ\u{102d}\u{102f}က\u{103a}", "သဝ\u{103a}"];
        /// `&["ဂျာန\u{103a}", "ဝ\u{103e}ေဝ\u{103a}", "မာတ\u{103a}", "အေ", "မေ", "ဂျ\u{102f}န\u{103a}", "ဂျ\u{1030}", "အဝ\u{103a}", "သ\u{102d}ပ\u{103a}", "အ\u{1036}က\u{103a}", "နဝ\u{103a}", "ဒ\u{1033}"]`
        pub const ABMON: &[&str] = &["ဂျာန\u{103a}", "ဝ\u{103e}ေဝ\u{103a}", "မာတ\u{103a}", "အေ", "မေ", "ဂျ\u{102f}န\u{103a}", "ဂျ\u{1030}", "အဝ\u{103a}", "သ\u{102d}ပ\u{103a}", "အ\u{1036}က\u{103a}", "နဝ\u{103a}", "ဒ\u{1033}"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `Some(&["၀၀", "၀၁", "၀၂", "၀၃", "၀၄", "၀၅", "၀၆", "၀၇", "၀၈", "၀၉", "၁၀", "၁၁", "၁၂", "၁၃", "၁၄", "၁၅", "၁၆", "၁၇", "၁၈", "၁၉", "၂၀", "၂၁", "၂၂", "၂၃", "၂၄", "၂၅", "၂၆", "၂၇", "၂၈", "၂၉", "၃၀", "၃၁", "၃၂", "၃၃", "၃၄", "၃၅", "၃၆", "၃၇", "၃၈", "၃၉", "၄၀", "၄၁", "၄၂", "၄၃", "၄၄", "၄၅", "၄၆", "၄၇", "၄၈", "၄၉", "၅၀", "၅၁", "၅၂", "၅၃", "၅၄", "၅၅", "၅၆", "၅၇", "၅၈", "၅၉", "၆၀", "၆၁", "၆၂", "၆၃", "၆၄", "၆၅", "၆၆", "၆၇", "၆၈", "၆၉", "၇၀", "၇၁", "၇၂", "၇၃", "၇၄", "၇၅", "၇၆", "၇၇", "၇၈", "၇၉", "၈၀", "၈၁", "၈၂", "၈၃", "၈၄", "၈၅", "၈၆", "၈၇", "၈၈", "၈၉", "၉၀", "၉၁", "၉၂", "၉၃", "၉၄", "၉၅", "၉၆", "၉၇", "၉၈", "၉၉"])`
        pub const ALT_DIGITS: Option<&[&str]> = Some(&["၀၀", "၀၁", "၀၂", "၀၃", "၀၄", "၀၅", "၀၆", "၀၇", "၀၈", "၀၉", "၁၀", "၁၁", "၁၂", "၁၃", "၁၄", "၁၅", "၁၆", "၁၇", "၁၈", "၁၉", "၂၀", "၂၁", "၂၂", "၂၃", "၂၄", "၂၅", "၂၆", "၂၇", "၂၈", "၂၉", "၃၀", "၃၁", "၃၂", "၃၃", "၃၄", "၃၅", "၃၆", "၃၇", "၃၈", "၃၉", "၄၀", "၄၁", "၄၂", "၄၃", "၄၄", "၄၅", "၄၆", "၄၇", "၄၈", "၄၉", "၅၀", "၅၁", "၅၂", "၅၃", "၅၄", "၅၅", "၅၆", "၅၇", "၅၈", "၅၉", "၆၀", "၆၁", "၆၂", "၆၃", "၆၄", "၆၅", "၆၆", "၆၇", "၆၈", "၆၉", "၇၀", "၇၁", "၇၂", "၇၃", "၇၄", "၇၅", "၇၆", "၇၇", "၇၈", "၇၉", "၈၀", "၈၁", "၈၂", "၈၃", "၈၄", "၈၅", "၈၆", "၈၇", "၈၈", "၈၉", "၉၀", "၉၁", "၉၂", "၉၃", "၉၄", "၉၅", "၉၆", "၉၇", "၉၈", "၉၉"]);
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["ဂယး", "တ\u{1039}ၚ\u{1032}စေၚ\u{103a}"]`
        pub const AM_PM: &[&str] = &["ဂယး", "တ\u{1039}ၚ\u{1032}စေၚ\u{103a}"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["အဒ\u{102d}\u{102f}တ\u{103a}", "စန\u{103a}", "အၚာ", "ဗ\u{102f}ဒ\u{1039}ဓဝါ", "ဗြ\u{1034}ဗ\u{1039}တ\u{102d}", "သ\u{102d}\u{102f}က\u{103a}", "သ\u{1039}ၚ\u{102d}သဝ\u{103a}"]`
        pub const DAY: &[&str] = &["အဒ\u{102d}\u{102f}တ\u{103a}", "စန\u{103a}", "အၚာ", "ဗ\u{102f}ဒ\u{1039}ဓဝါ", "ဗြ\u{1034}ဗ\u{1039}တ\u{102d}", "သ\u{102d}\u{102f}က\u{103a}", "သ\u{1039}ၚ\u{102d}သဝ\u{103a}"];
        /// `"%OC%Oy %b %Od %A"`
        pub const D_FMT: &str = "%OC%Oy %b %Od %A";
        /// `"%OC%Oy %b %Od %A %OI:%OM:%OS %Op %Z"`
        pub const D_T_FMT: &str = "%OC%Oy %b %Od %A %OI:%OM:%OS %Op %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["ဂျာန\u{103a}နျ\u{1030}အာရ\u{1033}", "ဝ\u{103e}ေဝ\u{103a}ဗြ\u{1030}အာရ\u{1033}", "မာတ\u{103a}ချ\u{103a}", "အေပြေယ\u{103a}လ\u{103a}", "မေ", "ဂျ\u{102f}န\u{103a}", "ဂျ\u{1030}လာၚ\u{103a}", "အဝ\u{103a}ဂါတ\u{103a}", "သ\u{102d}ပ\u{103a}တ\u{102e}ဗာ", "အ\u{1036}က\u{103a}တဝ\u{103a}ဗာ", "နဝ\u{103a}ဝါမ\u{103a}ဗာ", "ဒ\u{1033}သ\u{102e}ဗာ"]`
        pub const MON: &[&str] = &["ဂျာန\u{103a}နျ\u{1030}အာရ\u{1033}", "ဝ\u{103e}ေဝ\u{103a}ဗြ\u{1030}အာရ\u{1033}", "မာတ\u{103a}ချ\u{103a}", "အေပြေယ\u{103a}လ\u{103a}", "မေ", "ဂျ\u{102f}န\u{103a}", "ဂျ\u{1030}လာၚ\u{103a}", "အဝ\u{103a}ဂါတ\u{103a}", "သ\u{102d}ပ\u{103a}တ\u{102e}ဗာ", "အ\u{1036}က\u{103a}တဝ\u{103a}ဗာ", "နဝ\u{103a}ဝါမ\u{103a}ဗာ", "ဒ\u{1033}သ\u{102e}ဗာ"];
        /// `"%OI:%OM:%OS %p"`
        pub const T_FMT: &str = "%OI:%OM:%OS %p";
        /// `"%OI:%OM:%OS %p"`
        pub const T_FMT_AMPM: &str = "%OI:%OM:%OS %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod mr_IN {
    pub mod LC_ADDRESS {
        /// `Some("IND")`
        pub const COUNTRY_AB3: Option<&str> = Some("IND");
        /// `Some("IN")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IN");
        /// `Some("IND")`
        pub const COUNTRY_CAR: Option<&str> = Some("IND");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("भारत")`
        pub const COUNTRY_NAME: Option<&str> = Some("भारत");
        /// `Some(356)`
        pub const COUNTRY_NUM: Option<i64> = Some(356);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("mr")`
        pub const LANG_AB: Option<&str> = Some("mr");
        /// `Some("mar")`
        pub const LANG_LIB: Option<&str> = Some("mar");
        /// `Some("मराठी")`
        pub const LANG_NAME: Option<&str> = Some("मराठी");
        /// `Some("mar")`
        pub const LANG_TERM: Option<&str> = Some("mar");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan")`
        pub const ADDRESS: Option<&str> = Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-07-21")`
        pub const DATE: Option<&str> = Some("2000-07-21");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Marathi")`
        pub const LANGUAGE: Option<&str> = Some("Marathi");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("IBM Globalization Center of Competency, Yamato Software Laboratory")`
        pub const SOURCE: Option<&str> = Some("IBM Globalization Center of Competency, Yamato Software Laboratory");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("India")`
        pub const TERRITORY: Option<&str> = Some("India");
        /// `Some("Marathi language locale for India")`
        pub const TITLE: Option<&str> = Some("Marathi language locale for India");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNन]"`
        pub const NOEXPR: &str = "^[-0nNन]";
        /// `Some("नाही")`
        pub const NOSTR: Option<&str> = Some("नाही");
        /// `"^[+1yYह]"`
        pub const YESEXPR: &str = "^[+1yYह]";
        /// `Some("होय")`
        pub const YESSTR: Option<&str> = Some("होय");
    }
    pub use super::hi_IN::LC_MONETARY;
    pub use super::hi_IN::LC_NUMERIC;
    pub use super::hi_IN::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["रवि", "सोम", "म\u{902}गळ", "ब\u{941}ध", "ग\u{941}र\u{941}", "श\u{941}क\u{94d}र", "शनि"]`
        pub const ABDAY: &[&str] = &["रवि", "सोम", "म\u{902}गळ", "ब\u{941}ध", "ग\u{941}र\u{941}", "श\u{941}क\u{94d}र", "शनि"];
        /// `&["जान\u{947}", "फ\u{947}ब\u{94d}र\u{941}", "मार\u{94d}च", "एप\u{94d}रि", "म\u{947}", "ज\u{942}न", "ज\u{941}ल\u{948}", "ऑग", "सप\u{94d}ट\u{947}\u{902}", "ऑक\u{94d}टो", "नोव\u{94d}ह\u{947}\u{902}", "डिस\u{947}\u{902}"]`
        pub const ABMON: &[&str] = &["जान\u{947}", "फ\u{947}ब\u{94d}र\u{941}", "मार\u{94d}च", "एप\u{94d}रि", "म\u{947}", "ज\u{942}न", "ज\u{941}ल\u{948}", "ऑग", "सप\u{94d}ट\u{947}\u{902}", "ऑक\u{94d}टो", "नोव\u{94d}ह\u{947}\u{902}", "डिस\u{947}\u{902}"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["म.प\u{942}.", "म.न\u{902}."]`
        pub const AM_PM: &[&str] = &["म.प\u{942}.", "म.न\u{902}."];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["रविवार", "सोमवार", "म\u{902}गळवार", "ब\u{941}धवार", "ग\u{941}र\u{941}वार", "श\u{941}क\u{94d}रवार", "शनिवार"]`
        pub const DAY: &[&str] = &["रविवार", "सोमवार", "म\u{902}गळवार", "ब\u{941}धवार", "ग\u{941}र\u{941}वार", "श\u{941}क\u{94d}रवार", "शनिवार"];
        /// `"%-d/%-m/%y"`
        pub const D_FMT: &str = "%-d/%-m/%y";
        /// `"%A %d %b %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A %d %b %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["जान\u{947}वारी", "फ\u{947}ब\u{94d}र\u{941}वारी", "मार\u{94d}च", "एप\u{94d}रिल", "म\u{947}", "ज\u{942}न", "ज\u{941}ल\u{948}", "ऑगस\u{94d}ट", "सप\u{94d}ट\u{947}\u{902}बर", "ऑक\u{94d}टोबर", "नोव\u{94d}ह\u{947}\u{902}बर", "डिस\u{947}\u{902}बर"]`
        pub const MON: &[&str] = &["जान\u{947}वारी", "फ\u{947}ब\u{94d}र\u{941}वारी", "मार\u{94d}च", "एप\u{94d}रिल", "म\u{947}", "ज\u{942}न", "ज\u{941}ल\u{948}", "ऑगस\u{94d}ट", "सप\u{94d}ट\u{947}\u{902}बर", "ऑक\u{94d}टोबर", "नोव\u{94d}ह\u{947}\u{902}बर", "डिस\u{947}\u{902}बर"];
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT: &str = "%I:%M:%S %p %Z";
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p %Z";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ms_MY {
    pub mod LC_ADDRESS {
        /// `Some("MYS")`
        pub const COUNTRY_AB3: Option<&str> = Some("MYS");
        /// `Some("MY")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("MY");
        /// `Some("MAL")`
        pub const COUNTRY_CAR: Option<&str> = Some("MAL");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Malaysia")`
        pub const COUNTRY_NAME: Option<&str> = Some("Malaysia");
        /// `Some(458)`
        pub const COUNTRY_NUM: Option<i64> = Some(458);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ms")`
        pub const LANG_AB: Option<&str> = Some("ms");
        /// `Some("may")`
        pub const LANG_LIB: Option<&str> = Some("may");
        /// `Some("Melayu")`
        pub const LANG_NAME: Option<&str> = Some("Melayu");
        /// `Some("msa")`
        pub const LANG_TERM: Option<&str> = Some("msa");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan")`
        pub const ADDRESS: Option<&str> = Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2001, December, 10")`
        pub const DATE: Option<&str> = Some("2001, December, 10");
        /// `Some("bug-glibc-locales@gnu.org, sebol@ikhlas.com")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org, sebol@ikhlas.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Malay")`
        pub const LANGUAGE: Option<&str> = Some("Malay");
        /// `Some("0.92b")`
        pub const REVISION: Option<&str> = Some("0.92b");
        /// `Some("IBM Globalization Center of Competency, Yamato Software Laboratory")`
        pub const SOURCE: Option<&str> = Some("IBM Globalization Center of Competency, Yamato Software Laboratory");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Malaysia")`
        pub const TERRITORY: Option<&str> = Some("Malaysia");
        /// `Some("Malay language locale for Malaysia")`
        pub const TITLE: Option<&str> = Some("Malay language locale for Malaysia");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0tTnN]"`
        pub const NOEXPR: &str = "^[-0tTnN]";
        /// `Some("tidak")`
        pub const NOSTR: Option<&str> = Some("tidak");
        /// `"^[+1yY]"`
        pub const YESEXPR: &str = "^[+1yY]";
        /// `Some("ya")`
        pub const YESSTR: Option<&str> = Some("ya");
    }
    pub mod LC_MONETARY {
        /// `"RM"`
        pub const CURRENCY_SYMBOL: &str = "RM";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"MYR "`
        pub const INT_CURR_SYMBOL: &str = "MYR ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `0`
        pub const N_SIGN_POSN: i64 = 0;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("60")`
        pub const INT_PREFIX: Option<&str> = Some("60");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c ;%a ;%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c ;%a ;%l");
    }
    pub mod LC_TIME {
        /// `&["Ahd", "Isn", "Sel", "Rab", "Kha", "Jum", "Sab"]`
        pub const ABDAY: &[&str] = &["Ahd", "Isn", "Sel", "Rab", "Kha", "Jum", "Sab"];
        /// `&["Jan", "Feb", "Mac", "Apr", "Mei", "Jun", "Jul", "Ogos", "Sep", "Okt", "Nov", "Dis"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Mac", "Apr", "Mei", "Jun", "Jul", "Ogos", "Sep", "Okt", "Nov", "Dis"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Ahad", "Isnin", "Selasa", "Rabu", "Khamis", "Jumaat", "Sabtu"]`
        pub const DAY: &[&str] = &["Ahad", "Isnin", "Selasa", "Rabu", "Khamis", "Jumaat", "Sabtu"];
        /// `"%A %d %b %Y"`
        pub const D_FMT: &str = "%A %d %b %Y";
        /// `"%A %d %b %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A %d %b %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Januari", "Februari", "Mac", "April", "Mei", "Jun", "Julai", "Ogos", "September", "Oktober", "November", "Disember"]`
        pub const MON: &[&str] = &["Januari", "Februari", "Mac", "April", "Mei", "Jun", "Julai", "Ogos", "September", "Oktober", "November", "Disember"];
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT: &str = "%I:%M:%S %p %Z";
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p %Z";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod mt_MT {
    pub mod LC_ADDRESS {
        /// `Some("MLT")`
        pub const COUNTRY_AB3: Option<&str> = Some("MLT");
        /// `Some("MT")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("MT");
        /// `Some("M")`
        pub const COUNTRY_CAR: Option<&str> = Some("M");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Malta")`
        pub const COUNTRY_NAME: Option<&str> = Some("Malta");
        /// `Some(470)`
        pub const COUNTRY_NUM: Option<i64> = Some(470);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("mt")`
        pub const LANG_AB: Option<&str> = Some("mt");
        /// `Some("mlt")`
        pub const LANG_LIB: Option<&str> = Some("mlt");
        /// `Some("Malti")`
        pub const LANG_NAME: Option<&str> = Some("Malti");
        /// `Some("mlt")`
        pub const LANG_TERM: Option<&str> = Some("mlt");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan")`
        pub const ADDRESS: Option<&str> = Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-07-20")`
        pub const DATE: Option<&str> = Some("2000-07-20");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Maltese")`
        pub const LANGUAGE: Option<&str> = Some("Maltese");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("IBM Globalization Center of Competency, Yamato Software Laboratory")`
        pub const SOURCE: Option<&str> = Some("IBM Globalization Center of Competency, Yamato Software Laboratory");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("malta")`
        pub const TERRITORY: Option<&str> = Some("malta");
        /// `Some("Maltese language locale for Malta")`
        pub const TITLE: Option<&str> = Some("Maltese language locale for Malta");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNlL]"`
        pub const NOEXPR: &str = "^[-0nNlL]";
        /// `Some("le")`
        pub const NOSTR: Option<&str> = Some("le");
        /// `"^[+1yYiI]"`
        pub const YESEXPR: &str = "^[+1yYiI]";
        /// `Some("iva")`
        pub const YESSTR: Option<&str> = Some("iva");
    }
    pub mod LC_MONETARY {
        /// `"€"`
        pub const CURRENCY_SYMBOL: &str = "€";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"EUR "`
        pub const INT_CURR_SYMBOL: &str = "EUR ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("356")`
        pub const INT_PREFIX: Option<&str> = Some("356");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c ;%a ;%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c ;%a ;%l");
    }
    pub mod LC_TIME {
        /// `&["Ħad", "Tne", "Tli", "Erb", "Ħam", "Ġim", "Sib"]`
        pub const ABDAY: &[&str] = &["Ħad", "Tne", "Tli", "Erb", "Ħam", "Ġim", "Sib"];
        /// `&["Jan", "Fra", "Mar", "Apr", "Mej", "Ġun", "Lul", "Aww", "Set", "Ott", "Nov", "Diċ"]`
        pub const ABMON: &[&str] = &["Jan", "Fra", "Mar", "Apr", "Mej", "Ġun", "Lul", "Aww", "Set", "Ott", "Nov", "Diċ"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["il-Ħadd", "it-Tnejn", "it-Tlieta", "l-Erbgħa", "il-Ħamis", "il-Ġimgħa", "is-Sibt"]`
        pub const DAY: &[&str] = &["il-Ħadd", "it-Tnejn", "it-Tlieta", "l-Erbgħa", "il-Ħamis", "il-Ġimgħa", "is-Sibt"];
        /// `"%A, %d ta %b, %Y"`
        pub const D_FMT: &str = "%A, %d ta %b, %Y";
        /// `"%A, %d ta %b, %Y %H:%M:%S %Z"`
        pub const D_T_FMT: &str = "%A, %d ta %b, %Y %H:%M:%S %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Jannar", "Frar", "Marzu", "April", "Mejju", "Ġunju", "Lulju", "Awwissu", "Settembru", "Ottubru", "Novembru", "Diċembru"]`
        pub const MON: &[&str] = &["Jannar", "Frar", "Marzu", "April", "Mejju", "Ġunju", "Lulju", "Awwissu", "Settembru", "Ottubru", "Novembru", "Diċembru"];
        /// `"%H:%M:%S %Z"`
        pub const T_FMT: &str = "%H:%M:%S %Z";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod my_MM {
    pub mod LC_ADDRESS {
        /// `None`
        pub const COUNTRY_AB3: Option<&str> = None;
        /// `Some("MM")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("MM");
        /// `Some("MYA")`
        pub const COUNTRY_CAR: Option<&str> = Some("MYA");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("မြန\u{103a}မာ")`
        pub const COUNTRY_NAME: Option<&str> = Some("မြန\u{103a}မာ");
        /// `Some(104)`
        pub const COUNTRY_NUM: Option<i64> = Some(104);
        /// `Some("Myanmar")`
        pub const COUNTRY_POST: Option<&str> = Some("Myanmar");
        /// `Some("my")`
        pub const LANG_AB: Option<&str> = Some("my");
        /// `Some("bur")`
        pub const LANG_LIB: Option<&str> = Some("bur");
        /// `Some("မြန\u{103a}မာ")`
        pub const LANG_NAME: Option<&str> = Some("မြန\u{103a}မာ");
        /// `Some("mya")`
        pub const LANG_TERM: Option<&str> = Some("mya");
        /// `Some("%a%N%d%N%f%N%b%N%h%t%r%t%e%t%s%N%T%t%z%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%a%N%d%N%f%N%b%N%h%t%r%t%e%t%s%N%T%t%z%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Yangon, Myanmar")`
        pub const ADDRESS: Option<&str> = Some("Yangon, Myanmar");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_NAME"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_NAME"],
        ]);
        /// `Some("Keith Stribley")`
        pub const CONTACT: Option<&str> = Some("Keith Stribley");
        /// `Some("2009-10-02")`
        pub const DATE: Option<&str> = Some("2009-10-02");
        /// `Some("devel@thanlwinsoft.org")`
        pub const EMAIL: Option<&str> = Some("devel@thanlwinsoft.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Burmese")`
        pub const LANGUAGE: Option<&str> = Some("Burmese");
        /// `Some("1.3")`
        pub const REVISION: Option<&str> = Some("1.3");
        /// `Some("ThanLwinSoft http:/www.thanlwinsoft.org")`
        pub const SOURCE: Option<&str> = Some("ThanLwinSoft http:/www.thanlwinsoft.org");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Myanmar (Burma)")`
        pub const TERRITORY: Option<&str> = Some("Myanmar (Burma)");
        /// `Some("Burmese language locale for Myanmar")`
        pub const TITLE: Option<&str> = Some("Burmese language locale for Myanmar");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNမ]"`
        pub const NOEXPR: &str = "^[-0nNမ]";
        /// `Some("မဟ\u{102f}တ\u{103a}ဘ\u{1030}း")`
        pub const NOSTR: Option<&str> = Some("မဟ\u{102f}တ\u{103a}ဘ\u{1030}း");
        /// `"^[+1yYဟ]"`
        pub const YESEXPR: &str = "^[+1yYဟ]";
        /// `Some("ဟ\u{102f}တ\u{103a}တယ\u{103a}")`
        pub const YESSTR: Option<&str> = Some("ဟ\u{102f}တ\u{103a}တယ\u{103a}");
    }
    pub mod LC_MONETARY {
        /// `"K"`
        pub const CURRENCY_SYMBOL: &str = "K";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"MMK "`
        pub const INT_CURR_SYMBOL: &str = "MMK ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("95")`
        pub const INT_PREFIX: Option<&str> = Some("95");
        /// `Some("0")`
        pub const INT_SELECT: Option<&str> = Some("0");
        /// `Some("%a %l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("%a %l");
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["န\u{103d}ေ", "လာ", "ဂါ", "ဟ\u{1030}း", "တေး", "သော", "နေ"]`
        pub const ABDAY: &[&str] = &["န\u{103d}ေ", "လာ", "ဂါ", "ဟ\u{1030}း", "တေး", "သော", "နေ"];
        /// `&["ဇန\u{103a}", "ဖေ", "မတ\u{103a}", "ဧပြ\u{102e}", "မေ", "ဇ\u{103d}န\u{103a}", "ဇ\u{1030}", "ဩ", "စက\u{103a}", "အောက\u{103a}", "န\u{102d}\u{102f}", "ဒ\u{102e}"]`
        pub const ABMON: &[&str] = &["ဇန\u{103a}", "ဖေ", "မတ\u{103a}", "ဧပြ\u{102e}", "မေ", "ဇ\u{103d}န\u{103a}", "ဇ\u{1030}", "ဩ", "စက\u{103a}", "အောက\u{103a}", "န\u{102d}\u{102f}", "ဒ\u{102e}"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `Some(&["၀၀", "၀၁", "၀၂", "၀၃", "၀၄", "၀၅", "၀၆", "၀၇", "၀၈", "၀၉", "၁၀", "၁၁", "၁၂", "၁၃", "၁၄", "၁၅", "၁၆", "၁၇", "၁၈", "၁၉", "၂၀", "၂၁", "၂၂", "၂၃", "၂၄", "၂၅", "၂၆", "၂၇", "၂၈", "၂၉", "၃၀", "၃၁", "၃၂", "၃၃", "၃၄", "၃၅", "၃၆", "၃၇", "၃၈", "၃၉", "၄၀", "၄၁", "၄၂", "၄၃", "၄၄", "၄၅", "၄၆", "၄၇", "၄၈", "၄၉", "၅၀", "၅၁", "၅၂", "၅၃", "၅၄", "၅၅", "၅၆", "၅၇", "၅၈", "၅၉", "၆၀", "၆၁", "၆၂", "၆၃", "၆၄", "၆၅", "၆၆", "၆၇", "၆၈", "၆၉", "၇၀", "၇၁", "၇၂", "၇၃", "၇၄", "၇၅", "၇၆", "၇၇", "၇၈", "၇၉", "၈၀", "၈၁", "၈၂", "၈၃", "၈၄", "၈၅", "၈၆", "၈၇", "၈၈", "၈၉", "၉၀", "၉၁", "၉၂", "၉၃", "၉၄", "၉၅", "၉၆", "၉၇", "၉၈", "၉၉"])`
        pub const ALT_DIGITS: Option<&[&str]> = Some(&["၀၀", "၀၁", "၀၂", "၀၃", "၀၄", "၀၅", "၀၆", "၀၇", "၀၈", "၀၉", "၁၀", "၁၁", "၁၂", "၁၃", "၁၄", "၁၅", "၁၆", "၁၇", "၁၈", "၁၉", "၂၀", "၂၁", "၂၂", "၂၃", "၂၄", "၂၅", "၂၆", "၂၇", "၂၈", "၂၉", "၃၀", "၃၁", "၃၂", "၃၃", "၃၄", "၃၅", "၃၆", "၃၇", "၃၈", "၃၉", "၄၀", "၄၁", "၄၂", "၄၃", "၄၄", "၄၅", "၄၆", "၄၇", "၄၈", "၄၉", "၅၀", "၅၁", "၅၂", "၅၃", "၅၄", "၅၅", "၅၆", "၅၇", "၅၈", "၅၉", "၆၀", "၆၁", "၆၂", "၆၃", "၆၄", "၆၅", "၆၆", "၆၇", "၆၈", "၆၉", "၇၀", "၇၁", "၇၂", "၇၃", "၇၄", "၇၅", "၇၆", "၇၇", "၇၈", "၇၉", "၈၀", "၈၁", "၈၂", "၈၃", "၈၄", "၈၅", "၈၆", "၈၇", "၈၈", "၈၉", "၉၀", "၉၁", "၉၂", "၉၃", "၉၄", "၉၅", "၉၆", "၉၇", "၉၈", "၉၉"]);
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["န\u{1036}နက\u{103a}", "ညနေ"]`
        pub const AM_PM: &[&str] = &["န\u{1036}နက\u{103a}", "ညနေ"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["တနင\u{103a}\u{1039}ဂန\u{103d}ေ", "တနင\u{103a}\u{1039}လာ", "အင\u{103a}\u{1039}ဂါ", "ဗ\u{102f}ဒ\u{1039}ဓဟ\u{1030}း", "ကြာသပတေး", "သောကြာ", "စနေ"]`
        pub const DAY: &[&str] = &["တနင\u{103a}\u{1039}ဂန\u{103d}ေ", "တနင\u{103a}\u{1039}လာ", "အင\u{103a}\u{1039}ဂါ", "ဗ\u{102f}ဒ\u{1039}ဓဟ\u{1030}း", "ကြာသပတေး", "သောကြာ", "စနေ"];
        /// `"%OC%Oy %b %Od %A"`
        pub const D_FMT: &str = "%OC%Oy %b %Od %A";
        /// `"%OC%Oy %b %Od %A %OI:%OM:%OS %Op %Z"`
        pub const D_T_FMT: &str = "%OC%Oy %b %Od %A %OI:%OM:%OS %Op %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["ဇန\u{103a}နဝါရ\u{102e}", "ဖေဖော\u{103a}ဝါရ\u{102e}", "မတ\u{103a}", "ဧပြ\u{102e}", "မေ", "ဇ\u{103d}န\u{103a}", "ဇ\u{1030}လ\u{102d}\u{102f}င\u{103a}", "ဩဂ\u{102f}တ\u{103a}", "စက\u{103a}တင\u{103a}ဘာ", "အောက\u{103a}တ\u{102d}\u{102f}ဘာ", "န\u{102d}\u{102f}ဝင\u{103a}ဘာ", "ဒ\u{102e}ဇင\u{103a}ဘာ"]`
        pub const MON: &[&str] = &["ဇန\u{103a}နဝါရ\u{102e}", "ဖေဖော\u{103a}ဝါရ\u{102e}", "မတ\u{103a}", "ဧပြ\u{102e}", "မေ", "ဇ\u{103d}န\u{103a}", "ဇ\u{1030}လ\u{102d}\u{102f}င\u{103a}", "ဩဂ\u{102f}တ\u{103a}", "စက\u{103a}တင\u{103a}ဘာ", "အောက\u{103a}တ\u{102d}\u{102f}ဘာ", "န\u{102d}\u{102f}ဝင\u{103a}ဘာ", "ဒ\u{102e}ဇင\u{103a}ဘာ"];
        /// `"%OI:%OM:%OS %p"`
        pub const T_FMT: &str = "%OI:%OM:%OS %p";
        /// `"%OI:%OM:%OS %p"`
        pub const T_FMT_AMPM: &str = "%OI:%OM:%OS %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod nan_TW {
    pub mod LC_ADDRESS {
        /// `Some("TWN")`
        pub const COUNTRY_AB3: Option<&str> = Some("TWN");
        /// `Some("TW")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("TW");
        /// `Some("RC")`
        pub const COUNTRY_CAR: Option<&str> = Some("RC");
        /// `Some("957")`
        pub const COUNTRY_ISBN: Option<&str> = Some("957");
        /// `Some("中華民國")`
        pub const COUNTRY_NAME: Option<&str> = Some("中華民國");
        /// `Some(158)`
        pub const COUNTRY_NUM: Option<i64> = Some(158);
        /// `Some("TW")`
        pub const COUNTRY_POST: Option<&str> = Some("TW");
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("nan")`
        pub const LANG_LIB: Option<&str> = Some("nan");
        /// `Some("閩南語")`
        pub const LANG_NAME: Option<&str> = Some("閩南語");
        /// `Some("nan")`
        pub const LANG_TERM: Option<&str> = Some("nan");
        /// `Some("%c%N%T%N%s %h %e %r%N%b%N%d%N%f%N%a%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%c%N%T%N%s %h %e %r%N%b%N%d%N%f%N%a%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2013-06-02")`
        pub const DATE: Option<&str> = Some("2013-06-02");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Min Nan Chinese")`
        pub const LANGUAGE: Option<&str> = Some("Min Nan Chinese");
        /// `Some("0.1")`
        pub const REVISION: Option<&str> = Some("0.1");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Taiwan")`
        pub const TERRITORY: Option<&str> = Some("Taiwan");
        /// `Some("Min Nan Chinese locale for the Republic of China")`
        pub const TITLE: Option<&str> = Some("Min Nan Chinese locale for the Republic of China");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNｎＮ伓]"`
        pub const NOEXPR: &str = "^[-0nNｎＮ伓]";
        /// `None`
        pub const NOSTR: Option<&str> = None;
        /// `"^[+1yYｙＹ是]"`
        pub const YESEXPR: &str = "^[+1yYｙＹ是]";
        /// `None`
        pub const YESSTR: Option<&str> = None;
    }
    pub mod LC_MONETARY {
        /// `"NT$"`
        pub const CURRENCY_SYMBOL: &str = "NT$";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"TWD "`
        pub const INT_CURR_SYMBOL: &str = "TWD ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `Some(1)`
        pub const INT_N_CS_PRECEDES: Option<i64> = Some(1);
        /// `Some(0)`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = Some(0);
        /// `Some(1)`
        pub const INT_N_SIGN_POSN: Option<i64> = Some(1);
        /// `Some(1)`
        pub const INT_P_CS_PRECEDES: Option<i64> = Some(1);
        /// `Some(0)`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = Some(0);
        /// `Some(1)`
        pub const INT_P_SIGN_POSN: Option<i64> = Some(1);
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[4]`
        pub const MON_GROUPING: &[i64] = &[4];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[4]`
        pub const GROUPING: &[i64] = &[4];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("886")`
        pub const INT_PREFIX: Option<&str> = Some("886");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `Some("%A-%l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("%A-%l");
        /// `Some("+%c-%a-%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c-%a-%l");
    }
    pub mod LC_TIME {
        /// `&["日", "一", "二", "三", "四", "五", "六"]`
        pub const ABDAY: &[&str] = &["日", "一", "二", "三", "四", "五", "六"];
        /// `&[" 1月", " 2月", " 3月", " 4月", " 5月", " 6月", " 7月", " 8月", " 9月", "10月", "11月", "12月"]`
        pub const ABMON: &[&str] = &[" 1月", " 2月", " 3月", " 4月", " 5月", " 6月", " 7月", " 8月", " 9月", "10月", "11月", "12月"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["頂晡", "下晡"]`
        pub const AM_PM: &[&str] = &["頂晡", "下晡"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%Y年 %b %e日 %A %H:%M:%S %Z")`
        pub const DATE_FMT: Option<&str> = Some("%Y年 %b %e日 %A %H:%M:%S %Z");
        /// `&["禮拜日", "禮拜一", "禮拜二", "禮拜三", "禮拜四", "禮拜五", "禮拜六"]`
        pub const DAY: &[&str] = &["禮拜日", "禮拜一", "禮拜二", "禮拜三", "禮拜四", "禮拜五", "禮拜六"];
        /// `"%Y年%m月%d日"`
        pub const D_FMT: &str = "%Y年%m月%d日";
        /// `"%Y年%m月%d日 (%A) %H點%M分%S秒"`
        pub const D_T_FMT: &str = "%Y年%m月%d日 (%A) %H點%M分%S秒";
        /// `Some(&["+:2:1913/01/01:+*:民國:%EC%Ey年", "+:1:1912/01/01:1912/12/31:民國:%EC元年", "+:1:1911/12/31:-*:民前:%EC%Ey年"])`
        pub const ERA: Option<&[&str]> = Some(&["+:2:1913/01/01:+*:民國:%EC%Ey年", "+:1:1912/01/01:1912/12/31:民國:%EC元年", "+:1:1911/12/31:-*:民前:%EC%Ey年"]);
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["一月", "二月", "三月", "四月", "五月", "六月", "七月", "八月", "九月", "十月", "十一月", "十二月"]`
        pub const MON: &[&str] = &["一月", "二月", "三月", "四月", "五月", "六月", "七月", "八月", "九月", "十月", "十一月", "十二月"];
        /// `"%H點%M分%S秒"`
        pub const T_FMT: &str = "%H點%M分%S秒";
        /// `"%p %I點%M分%S秒"`
        pub const T_FMT_AMPM: &str = "%p %I點%M分%S秒";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod nan_TW_latin {
    pub mod LC_ADDRESS {
        /// `Some("TWN")`
        pub const COUNTRY_AB3: Option<&str> = Some("TWN");
        /// `Some("TW")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("TW");
        /// `Some("RC")`
        pub const COUNTRY_CAR: Option<&str> = Some("RC");
        /// `Some("957")`
        pub const COUNTRY_ISBN: Option<&str> = Some("957");
        /// `Some("Tâi-oân")`
        pub const COUNTRY_NAME: Option<&str> = Some("Tâi-oân");
        /// `Some(158)`
        pub const COUNTRY_NUM: Option<i64> = Some(158);
        /// `Some("TW")`
        pub const COUNTRY_POST: Option<&str> = Some("TW");
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("nan")`
        pub const LANG_LIB: Option<&str> = Some("nan");
        /// `Some("Bân-lâm-gú")`
        pub const LANG_NAME: Option<&str> = Some("Bân-lâm-gú");
        /// `Some("nan")`
        pub const LANG_TERM: Option<&str> = Some("nan");
        /// `Some("%f%N%a%N%d%N%b%N%r %e %h %s%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%r %e %h %s%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_MONETARY"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_MONETARY"],
        ]);
        /// `Some("Arne Goetje")`
        pub const CONTACT: Option<&str> = Some("Arne Goetje");
        /// `Some("2008-06-16")`
        pub const DATE: Option<&str> = Some("2008-06-16");
        /// `Some("arne@canonical.com")`
        pub const EMAIL: Option<&str> = Some("arne@canonical.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Min Nan Chinese")`
        pub const LANGUAGE: Option<&str> = Some("Min Nan Chinese");
        /// `Some("0.1")`
        pub const REVISION: Option<&str> = Some("0.1");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Taiwan")`
        pub const TERRITORY: Option<&str> = Some("Taiwan");
        /// `Some("Minnan language locale for Taiwan")`
        pub const TITLE: Option<&str> = Some("Minnan language locale for Taiwan");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNmM]"`
        pub const NOEXPR: &str = "^[-0nNmM]";
        /// `Some("M\u{304}-Sī")`
        pub const NOSTR: Option<&str> = Some("M\u{304}-Sī");
        /// `"^[+1yYsS]"`
        pub const YESEXPR: &str = "^[+1yYsS]";
        /// `Some("Sī")`
        pub const YESSTR: Option<&str> = Some("Sī");
    }
    pub use super::zh_TW::LC_MONETARY;
    pub use super::zh_TW::LC_NUMERIC;
    pub use super::zh_TW::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["lp", "p1", "p2", "p3", "p4", "p5", "p6"]`
        pub const ABDAY: &[&str] = &["lp", "p1", "p2", "p3", "p4", "p5", "p6"];
        /// `&["1g", "2g", "3g", "4g", "5g", "6g", "7g", "8g", "9g", "10g", "11g", "12g"]`
        pub const ABMON: &[&str] = &["1g", "2g", "3g", "4g", "5g", "6g", "7g", "8g", "9g", "10g", "11g", "12g"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["téng-po\u{358}", "ē-po\u{358}"]`
        pub const AM_PM: &[&str] = &["téng-po\u{358}", "ē-po\u{358}"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%Y %b %d (%a) %H:%M:%S %Z")`
        pub const DATE_FMT: Option<&str> = Some("%Y %b %d (%a) %H:%M:%S %Z");
        /// `&["lé-pài-ji\u{30d}t", "pài-it", "pài-jī", "pài-saⁿ", "pài-sì", "pài-gō\u{358}", "pài-la\u{30d}k"]`
        pub const DAY: &[&str] = &["lé-pài-ji\u{30d}t", "pài-it", "pài-jī", "pài-saⁿ", "pài-sì", "pài-gō\u{358}", "pài-la\u{30d}k"];
        /// `"%F"`
        pub const D_FMT: &str = "%F";
        /// `"%Y %b %d (%a) %H:%M:%S %Z"`
        pub const D_T_FMT: &str = "%Y %b %d (%a) %H:%M:%S %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["1goe\u{30d}h", "2goe\u{30d}h", "3goe\u{30d}h", "4goe\u{30d}h", "5goe\u{30d}h", "6goe\u{30d}h", "7goe\u{30d}h", "8goe\u{30d}h", "9goe\u{30d}h", "10goe\u{30d}h", "11goe\u{30d}h", "12goe\u{30d}h"]`
        pub const MON: &[&str] = &["1goe\u{30d}h", "2goe\u{30d}h", "3goe\u{30d}h", "4goe\u{30d}h", "5goe\u{30d}h", "6goe\u{30d}h", "7goe\u{30d}h", "8goe\u{30d}h", "9goe\u{30d}h", "10goe\u{30d}h", "11goe\u{30d}h", "12goe\u{30d}h"];
        /// `"%I:%M:%S %p"`
        pub const T_FMT: &str = "%I:%M:%S %p";
        /// `"%I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod nb_NO {
    pub mod LC_ADDRESS {
        /// `Some("NOR")`
        pub const COUNTRY_AB3: Option<&str> = Some("NOR");
        /// `Some("NO")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("NO");
        /// `Some("N")`
        pub const COUNTRY_CAR: Option<&str> = Some("N");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Norge")`
        pub const COUNTRY_NAME: Option<&str> = Some("Norge");
        /// `Some(578)`
        pub const COUNTRY_NUM: Option<i64> = Some(578);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("nb")`
        pub const LANG_AB: Option<&str> = Some("nb");
        /// `Some("nob")`
        pub const LANG_LIB: Option<&str> = Some("nob");
        /// `Some("norsk bokmål")`
        pub const LANG_NAME: Option<&str> = Some("norsk bokmål");
        /// `Some("nob")`
        pub const LANG_TERM: Option<&str> = Some("nob");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("University Library, Drammensveien 41, N-9242 Oslo, Norge")`
        pub const ADDRESS: Option<&str> = Some("University Library, Drammensveien 41, N-9242 Oslo, Norge");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Norwegian Bokmål")`
        pub const LANGUAGE: Option<&str> = Some("Norwegian Bokmål");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Norsk Standardiseringsforbund")`
        pub const SOURCE: Option<&str> = Some("Norsk Standardiseringsforbund");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Norway")`
        pub const TERRITORY: Option<&str> = Some("Norway");
        /// `Some("Norwegian (Bokmål) locale for Norway")`
        pub const TITLE: Option<&str> = Some("Norwegian (Bokmål) locale for Norway");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0Nn]"`
        pub const NOEXPR: &str = "^[-0Nn]";
        /// `Some("nei")`
        pub const NOSTR: Option<&str> = Some("nei");
        /// `"^[+1JjYy]"`
        pub const YESEXPR: &str = "^[+1JjYy]";
        /// `Some("ja")`
        pub const YESSTR: Option<&str> = Some("ja");
    }
    pub mod LC_MONETARY {
        /// `"kr"`
        pub const CURRENCY_SYMBOL: &str = "kr";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"NOK "`
        pub const INT_CURR_SYMBOL: &str = "NOK ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const MON_THOUSANDS_SEP: &str = "\u{202f}";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `4`
        pub const N_SIGN_POSN: i64 = 4;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `4`
        pub const P_SIGN_POSN: i64 = 4;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const THOUSANDS_SEP: &str = "\u{202f}";
    }
    pub mod LC_TELEPHONE {
        /// `Some("47")`
        pub const INT_PREFIX: Option<&str> = Some("47");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `Some("%l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("%l");
        /// `Some("+%c %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %l");
    }
    pub mod LC_TIME {
        /// `&["sø.", "ma.", "ti.", "on.", "to.", "fr.", "lø."]`
        pub const ABDAY: &[&str] = &["sø.", "ma.", "ti.", "on.", "to.", "fr.", "lø."];
        /// `&["jan.", "feb.", "mars", "april", "mai", "juni", "juli", "aug.", "sep.", "okt.", "nov.", "des."]`
        pub const ABMON: &[&str] = &["jan.", "feb.", "mars", "april", "mai", "juni", "juli", "aug.", "sep.", "okt.", "nov.", "des."];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a %d. %b %H:%M:%S %z %Y")`
        pub const DATE_FMT: Option<&str> = Some("%a %d. %b %H:%M:%S %z %Y");
        /// `&["søndag", "mandag", "tirsdag", "onsdag", "torsdag", "fredag", "lørdag"]`
        pub const DAY: &[&str] = &["søndag", "mandag", "tirsdag", "onsdag", "torsdag", "fredag", "lørdag"];
        /// `"%d. %b %Y"`
        pub const D_FMT: &str = "%d. %b %Y";
        /// `"%a %d. %b %Y kl. %H.%M %z"`
        pub const D_T_FMT: &str = "%a %d. %b %Y kl. %H.%M %z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["januar", "februar", "mars", "april", "mai", "juni", "juli", "august", "september", "oktober", "november", "desember"]`
        pub const MON: &[&str] = &["januar", "februar", "mars", "april", "mai", "juni", "juli", "august", "september", "oktober", "november", "desember"];
        /// `"kl. %H.%M %z"`
        pub const T_FMT: &str = "kl. %H.%M %z";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod nds_DE {
    pub mod LC_ADDRESS {
        /// `Some("DEU")`
        pub const COUNTRY_AB3: Option<&str> = Some("DEU");
        /// `Some("DE")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("DE");
        /// `Some("D")`
        pub const COUNTRY_CAR: Option<&str> = Some("D");
        /// `Some("3")`
        pub const COUNTRY_ISBN: Option<&str> = Some("3");
        /// `Some("Düütschland")`
        pub const COUNTRY_NAME: Option<&str> = Some("Düütschland");
        /// `Some(276)`
        pub const COUNTRY_NUM: Option<i64> = Some(276);
        /// `Some("D")`
        pub const COUNTRY_POST: Option<&str> = Some("D");
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("nds")`
        pub const LANG_LIB: Option<&str> = Some("nds");
        /// `Some("Neddersass’sch")`
        pub const LANG_NAME: Option<&str> = Some("Neddersass’sch");
        /// `Some("nds")`
        pub const LANG_TERM: Option<&str> = Some("nds");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_ADDRESS"],
        ]);
        /// `Some("Kenneth Christiansen, Pablo Saratxaga")`
        pub const CONTACT: Option<&str> = Some("Kenneth Christiansen, Pablo Saratxaga");
        /// `Some("2003-11-30")`
        pub const DATE: Option<&str> = Some("2003-11-30");
        /// `Some("kenneth@gnu.org, pablo@mandrakesoft.com")`
        pub const EMAIL: Option<&str> = Some("kenneth@gnu.org, pablo@mandrakesoft.com");
        /// `None`
        pub const FAX: Option<&str> = None;
        /// `Some("Low German")`
        pub const LANGUAGE: Option<&str> = Some("Low German");
        /// `Some("0.1")`
        pub const REVISION: Option<&str> = Some("0.1");
        /// `Some("information from Kenneth Christiansen")`
        pub const SOURCE: Option<&str> = Some("information from Kenneth Christiansen");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Germany")`
        pub const TERRITORY: Option<&str> = Some("Germany");
        /// `Some("Low(lands) Saxon Language Locale for Germany")`
        pub const TITLE: Option<&str> = Some("Low(lands) Saxon Language Locale for Germany");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("nee")`
        pub const NOSTR: Option<&str> = Some("nee");
        /// `"^[+1jJyY]"`
        pub const YESEXPR: &str = "^[+1jJyY]";
        /// `Some("jo")`
        pub const YESSTR: Option<&str> = Some("jo");
    }
    pub use super::de_DE::LC_MONETARY;
    pub use super::de_DE::LC_NUMERIC;
    pub use super::de_DE::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Sdag", "Maan", "Ding", "Migg", "Dunn", "Free", "Svd."]`
        pub const ABDAY: &[&str] = &["Sdag", "Maan", "Ding", "Migg", "Dunn", "Free", "Svd."];
        /// `&["Jan", "Feb", "Mär", "Apr", "Mai", "Jun", "Jul", "Aug", "Sep", "Okt", "Nov", "Dez"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Mär", "Apr", "Mai", "Jun", "Jul", "Aug", "Sep", "Okt", "Nov", "Dez"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Sünndag", "Maandag", "Dingsdag", "Middeweek", "Dunnersdag", "Freedag", "Sünnavend"]`
        pub const DAY: &[&str] = &["Sünndag", "Maandag", "Dingsdag", "Middeweek", "Dunnersdag", "Freedag", "Sünnavend"];
        /// `"%d.%m.%Y"`
        pub const D_FMT: &str = "%d.%m.%Y";
        /// `"%a %d. %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d. %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Jannuaar", "Feberwaar", "März", "April", "Mai", "Juni", "Juli", "August", "September", "Oktober", "November", "Dezember"]`
        pub const MON: &[&str] = &["Jannuaar", "Feberwaar", "März", "April", "Mai", "Juni", "Juli", "August", "September", "Oktober", "November", "Dezember"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod nds_NL {
    pub mod LC_ADDRESS {
        /// `Some("NLD")`
        pub const COUNTRY_AB3: Option<&str> = Some("NLD");
        /// `Some("NL")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("NL");
        /// `Some("NL")`
        pub const COUNTRY_CAR: Option<&str> = Some("NL");
        /// `Some("3")`
        pub const COUNTRY_ISBN: Option<&str> = Some("3");
        /// `Some("Nedderlannen")`
        pub const COUNTRY_NAME: Option<&str> = Some("Nedderlannen");
        /// `Some(528)`
        pub const COUNTRY_NUM: Option<i64> = Some(528);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("nds")`
        pub const LANG_LIB: Option<&str> = Some("nds");
        /// `Some("Neddersass’sch")`
        pub const LANG_NAME: Option<&str> = Some("Neddersass’sch");
        /// `Some("nds")`
        pub const LANG_TERM: Option<&str> = Some("nds");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_ADDRESS"],
        ]);
        /// `Some("Kenneth Christiansen, Pablo Saratxaga")`
        pub const CONTACT: Option<&str> = Some("Kenneth Christiansen, Pablo Saratxaga");
        /// `Some("2003-11-30")`
        pub const DATE: Option<&str> = Some("2003-11-30");
        /// `Some("kenneth@gnu.org, pablo@mandrakesoft.com")`
        pub const EMAIL: Option<&str> = Some("kenneth@gnu.org, pablo@mandrakesoft.com");
        /// `None`
        pub const FAX: Option<&str> = None;
        /// `Some("Low Saxon")`
        pub const LANGUAGE: Option<&str> = Some("Low Saxon");
        /// `Some("0.1")`
        pub const REVISION: Option<&str> = Some("0.1");
        /// `Some("information from Kenneth Christiansen")`
        pub const SOURCE: Option<&str> = Some("information from Kenneth Christiansen");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Netherlands")`
        pub const TERRITORY: Option<&str> = Some("Netherlands");
        /// `Some("Low(lands) Saxon Language Locale for the Netherlands")`
        pub const TITLE: Option<&str> = Some("Low(lands) Saxon Language Locale for the Netherlands");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("nee")`
        pub const NOSTR: Option<&str> = Some("nee");
        /// `"^[+1jJyY]"`
        pub const YESEXPR: &str = "^[+1jJyY]";
        /// `Some("jo")`
        pub const YESSTR: Option<&str> = Some("jo");
    }
    pub use super::nl_NL::LC_MONETARY;
    pub use super::nl_NL::LC_NUMERIC;
    pub use super::nl_NL::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Sdg", "Mdg", "Dsg", "Mwk", "Ddg", "Fdg", "Swd"]`
        pub const ABDAY: &[&str] = &["Sdg", "Mdg", "Dsg", "Mwk", "Ddg", "Fdg", "Swd"];
        /// `&["Jan", "Feb", "Moz", "Apr", "Mai", "Jun", "Jul", "Aug", "Sep", "Okt", "Now", "Dez"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Moz", "Apr", "Mai", "Jun", "Jul", "Aug", "Sep", "Okt", "Now", "Dez"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Sinndag", "Mondag", "Dingsdag", "Meddwäakj", "Donnadag", "Friedag", "Sinnowend"]`
        pub const DAY: &[&str] = &["Sinndag", "Mondag", "Dingsdag", "Meddwäakj", "Donnadag", "Friedag", "Sinnowend"];
        /// `"%d.%m.%Y"`
        pub const D_FMT: &str = "%d.%m.%Y";
        /// `"%a %d. %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d. %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Jaunuwoa", "Februwoa", "Moaz", "Aprell", "Mai", "Juni", "Juli", "August", "Septamba", "Oktoba", "Nowamba", "Dezamba"]`
        pub const MON: &[&str] = &["Jaunuwoa", "Februwoa", "Moaz", "Aprell", "Mai", "Juni", "Juli", "August", "Septamba", "Oktoba", "Nowamba", "Dezamba"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ne_NP {
    pub mod LC_ADDRESS {
        /// `Some("NPL")`
        pub const COUNTRY_AB3: Option<&str> = Some("NPL");
        /// `Some("NP")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("NP");
        /// `Some("NEP")`
        pub const COUNTRY_CAR: Option<&str> = Some("NEP");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("न\u{947}पाल")`
        pub const COUNTRY_NAME: Option<&str> = Some("न\u{947}पाल");
        /// `Some(524)`
        pub const COUNTRY_NUM: Option<i64> = Some(524);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ne")`
        pub const LANG_AB: Option<&str> = Some("ne");
        /// `Some("nep")`
        pub const LANG_LIB: Option<&str> = Some("nep");
        /// `Some("न\u{947}पाली")`
        pub const LANG_NAME: Option<&str> = Some("न\u{947}पाली");
        /// `Some("nep")`
        pub const LANG_TERM: Option<&str> = Some("nep");
        /// `Some("%f%N%h%s%N%T")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%h%s%N%T");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Rato Bangala, Patan Dhoka,Lalitpur,Nepal")`
        pub const ADDRESS: Option<&str> = Some("Rato Bangala, Patan Dhoka,Lalitpur,Nepal");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2003-05-12")`
        pub const DATE: Option<&str> = Some("2003-05-12");
        /// `Some("info@mpp.org.np")`
        pub const EMAIL: Option<&str> = Some("info@mpp.org.np");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Nepali")`
        pub const LANGUAGE: Option<&str> = Some("Nepali");
        /// `Some("1.1")`
        pub const REVISION: Option<&str> = Some("1.1");
        /// `Some("Madan Puraskar Pustakalaya")`
        pub const SOURCE: Option<&str> = Some("Madan Puraskar Pustakalaya");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Nepal")`
        pub const TERRITORY: Option<&str> = Some("Nepal");
        /// `Some("Nepali language locale for Nepal")`
        pub const TITLE: Option<&str> = Some("Nepali language locale for Nepal");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("होइन")`
        pub const NOSTR: Option<&str> = Some("होइन");
        /// `"^[+1yY]"`
        pub const YESEXPR: &str = "^[+1yY]";
        /// `Some("हो")`
        pub const YESSTR: Option<&str> = Some("हो");
    }
    pub mod LC_MONETARY {
        /// `"र\u{942}"`
        pub const CURRENCY_SYMBOL: &str = "र\u{942}";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"NPR "`
        pub const INT_CURR_SYMBOL: &str = "NPR ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("977")`
        pub const INT_PREFIX: Option<&str> = Some("977");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a%t%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a%t%l");
    }
    pub mod LC_TIME {
        /// `&["आइत", "सोम", "मङ\u{94d}गल", "ब\u{941}ध", "बिही", "श\u{941}क\u{94d}र", "शनि"]`
        pub const ABDAY: &[&str] = &["आइत", "सोम", "मङ\u{94d}गल", "ब\u{941}ध", "बिही", "श\u{941}क\u{94d}र", "शनि"];
        /// `&["जन", "फ\u{947}ब\u{94d}र\u{941}अरी", "मार\u{94d}च", "अप\u{94d}रिल", "म\u{947}", "ज\u{941}न", "ज\u{941}लाई", "अगस\u{94d}ट", "स\u{947}प\u{94d}ट\u{947}म\u{94d}बर", "अक\u{94d}टोबर", "नोभ\u{947}म\u{94d}बर", "डिस\u{947}म\u{94d}बर"]`
        pub const ABMON: &[&str] = &["जन", "फ\u{947}ब\u{94d}र\u{941}अरी", "मार\u{94d}च", "अप\u{94d}रिल", "म\u{947}", "ज\u{941}न", "ज\u{941}लाई", "अगस\u{94d}ट", "स\u{947}प\u{94d}ट\u{947}म\u{94d}बर", "अक\u{94d}टोबर", "नोभ\u{947}म\u{94d}बर", "डिस\u{947}म\u{94d}बर"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["प\u{942}र\u{94d}वाह\u{94d}न", "अपराह\u{94d}न"]`
        pub const AM_PM: &[&str] = &["प\u{942}र\u{94d}वाह\u{94d}न", "अपराह\u{94d}न"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["आइतबार", "सोमबार", "मङ\u{94d}गलबार", "ब\u{941}धबार", "बिहीबार", "श\u{941}क\u{94d}रबार", "शनिबार"]`
        pub const DAY: &[&str] = &["आइतबार", "सोमबार", "मङ\u{94d}गलबार", "ब\u{941}धबार", "बिहीबार", "श\u{941}क\u{94d}रबार", "शनिबार"];
        /// `"%y/%-m/%-d"`
        pub const D_FMT: &str = "%y/%-m/%-d";
        /// `"%Y %B %d %I:%M:%S %p"`
        pub const D_T_FMT: &str = "%Y %B %d %I:%M:%S %p";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["जनवरी", "फ\u{947}ब\u{94d}र\u{941}अरी", "मार\u{94d}च", "अप\u{94d}रिल", "म\u{947}", "ज\u{941}न", "ज\u{941}लाई", "अगस\u{94d}ट", "स\u{947}प\u{94d}ट\u{947}म\u{94d}बर", "अक\u{94d}टोबर", "नोभ\u{947}म\u{94d}बर", "डिस\u{947}म\u{94d}बर"]`
        pub const MON: &[&str] = &["जनवरी", "फ\u{947}ब\u{94d}र\u{941}अरी", "मार\u{94d}च", "अप\u{94d}रिल", "म\u{947}", "ज\u{941}न", "ज\u{941}लाई", "अगस\u{94d}ट", "स\u{947}प\u{94d}ट\u{947}म\u{94d}बर", "अक\u{94d}टोबर", "नोभ\u{947}म\u{94d}बर", "डिस\u{947}म\u{94d}बर"];
        /// `"%H:%M:%S"`
        pub const T_FMT: &str = "%H:%M:%S";
        /// `"%I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod nhn_MX {
    pub mod LC_ADDRESS {
        /// `Some("MEX")`
        pub const COUNTRY_AB3: Option<&str> = Some("MEX");
        /// `Some("MX")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("MX");
        /// `Some("MEX")`
        pub const COUNTRY_CAR: Option<&str> = Some("MEX");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Mexihco")`
        pub const COUNTRY_NAME: Option<&str> = Some("Mexihco");
        /// `Some(484)`
        pub const COUNTRY_NUM: Option<i64> = Some(484);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("nhn")`
        pub const LANG_LIB: Option<&str> = Some("nhn");
        /// `Some("Tlahco nāhuatlahtōlli")`
        pub const LANG_NAME: Option<&str> = Some("Tlahco nāhuatlahtōlli");
        /// `Some("nhn")`
        pub const LANG_TERM: Option<&str> = Some("nhn");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2012-08-19")`
        pub const DATE: Option<&str> = Some("2012-08-19");
        /// `Some("libc-alpha@sourceware.org")`
        pub const EMAIL: Option<&str> = Some("libc-alpha@sourceware.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Central Nahuatl")`
        pub const LANGUAGE: Option<&str> = Some("Central Nahuatl");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Mexico")`
        pub const TERRITORY: Option<&str> = Some("Mexico");
        /// `Some("Central Nahuatl for Mexico")`
        pub const TITLE: Option<&str> = Some("Central Nahuatl for Mexico");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNaA]"`
        pub const NOEXPR: &str = "^[-0nNaA]";
        /// `Some("ahmo")`
        pub const NOSTR: Option<&str> = Some("ahmo");
        /// `"^[+1sSqQyY]"`
        pub const YESEXPR: &str = "^[+1sSqQyY]";
        /// `Some("quema")`
        pub const YESSTR: Option<&str> = Some("quema");
    }
    pub use super::es_MX::LC_MONETARY;
    pub use super::es_MX::LC_NUMERIC;
    pub use super::es_MX::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["teo", "cei", "ome", "yei", "nau", "mac", "chi"]`
        pub const ABDAY: &[&str] = &["teo", "cei", "ome", "yei", "nau", "mac", "chi"];
        /// `&["ene", "feb", "mar", "abr", "may", "jun", "jul", "ago", "sep", "oct", "nov", "dic"]`
        pub const ABMON: &[&str] = &["ene", "feb", "mar", "abr", "may", "jun", "jul", "ago", "sep", "oct", "nov", "dic"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["teoilhuitl", "ceilhuitl", "omeilhuitl", "yeilhuitl", "nahuilhuitl", "macuililhuitl", "chicuaceilhuitl"]`
        pub const DAY: &[&str] = &["teoilhuitl", "ceilhuitl", "omeilhuitl", "yeilhuitl", "nahuilhuitl", "macuililhuitl", "chicuaceilhuitl"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"]`
        pub const MON: &[&str] = &["enero", "febrero", "marzo", "abril", "mayo", "junio", "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod niu_NU {
    pub mod LC_ADDRESS {
        /// `Some("NIU")`
        pub const COUNTRY_AB3: Option<&str> = Some("NIU");
        /// `Some("NU")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("NU");
        /// `None`
        pub const COUNTRY_CAR: Option<&str> = None;
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Niuē")`
        pub const COUNTRY_NAME: Option<&str> = Some("Niuē");
        /// `Some(570)`
        pub const COUNTRY_NUM: Option<i64> = Some(570);
        /// `Some("NU")`
        pub const COUNTRY_POST: Option<&str> = Some("NU");
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("niu")`
        pub const LANG_LIB: Option<&str> = Some("niu");
        /// `Some("ko e vagahau Niuē")`
        pub const LANG_NAME: Option<&str> = Some("ko e vagahau Niuē");
        /// `Some("niu")`
        pub const LANG_TERM: Option<&str> = Some("niu");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("PO Box 229, Fatalupe, Alofi, Niue")`
        pub const ADDRESS: Option<&str> = Some("PO Box 229, Fatalupe, Alofi, Niue");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
        ]);
        /// `Some("Emani Fakaotimanava-Lui")`
        pub const CONTACT: Option<&str> = Some("Emani Fakaotimanava-Lui");
        /// `Some("2012-07-18")`
        pub const DATE: Option<&str> = Some("2012-07-18");
        /// `Some("emani@niue.nu")`
        pub const EMAIL: Option<&str> = Some("emani@niue.nu");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Niuean")`
        pub const LANGUAGE: Option<&str> = Some("Niuean");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RockET Systems")`
        pub const SOURCE: Option<&str> = Some("RockET Systems");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Niue")`
        pub const TERRITORY: Option<&str> = Some("Niue");
        /// `Some("Niuean (Vagahau Niue) locale for Niue")`
        pub const TITLE: Option<&str> = Some("Niuean (Vagahau Niue) locale for Niue");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("Nakai")`
        pub const NOSTR: Option<&str> = Some("Nakai");
        /// `"^[+1yYēĒ]"`
        pub const YESEXPR: &str = "^[+1yYēĒ]";
        /// `Some("Talia")`
        pub const YESSTR: Option<&str> = Some("Talia");
    }
    pub use super::en_NZ::LC_MONETARY;
    pub use super::en_NZ::LC_NUMERIC;
    pub mod LC_TELEPHONE {
        /// `Some("683")`
        pub const INT_PREFIX: Option<&str> = Some("683");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["Tapu", "Gofua", "Ua", "Lotu", "Tuloto", "Falaile", "Faiumu"]`
        pub const ABDAY: &[&str] = &["Tapu", "Gofua", "Ua", "Lotu", "Tuloto", "Falaile", "Faiumu"];
        /// `&["Ian", "Fep", "Mas", "Ape", "Me", "Iun", "Iul", "Aok", "Sep", "Oke", "Nov", "Tes"]`
        pub const ABMON: &[&str] = &["Ian", "Fep", "Mas", "Ape", "Me", "Iun", "Iul", "Aok", "Sep", "Oke", "Nov", "Tes"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["AM", "PM"]`
        pub const AM_PM: &[&str] = &["AM", "PM"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Aho Tapu", "Aho Gofua", "Aho Ua", "Aho Lotu", "Aho Tuloto", "Aho Falaile", "Aho Faiumu"]`
        pub const DAY: &[&str] = &["Aho Tapu", "Aho Gofua", "Aho Ua", "Aho Lotu", "Aho Tuloto", "Aho Falaile", "Aho Faiumu"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Ianuali", "Fepuali", "Masi", "Apelila", "Me", "Iuni", "Iulai", "Aokuso", "Sepetema", "Oketopa", "Novema", "Tesemo"]`
        pub const MON: &[&str] = &["Ianuali", "Fepuali", "Masi", "Apelila", "Me", "Iuni", "Iulai", "Aokuso", "Sepetema", "Oketopa", "Novema", "Tesemo"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `"%I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod niu_NZ {
    pub mod LC_ADDRESS {
        /// `Some("NZL")`
        pub const COUNTRY_AB3: Option<&str> = Some("NZL");
        /// `Some("NZ")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("NZ");
        /// `Some("NZ")`
        pub const COUNTRY_CAR: Option<&str> = Some("NZ");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `None`
        pub const COUNTRY_NAME: Option<&str> = None;
        /// `Some(554)`
        pub const COUNTRY_NUM: Option<i64> = Some(554);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("niu")`
        pub const LANG_LIB: Option<&str> = Some("niu");
        /// `Some("ko e vagahau Niuē")`
        pub const LANG_NAME: Option<&str> = Some("ko e vagahau Niuē");
        /// `Some("niu")`
        pub const LANG_TERM: Option<&str> = Some("niu");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("PO Box 229, Fatalupe, Alofi, Niue")`
        pub const ADDRESS: Option<&str> = Some("PO Box 229, Fatalupe, Alofi, Niue");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
        ]);
        /// `Some("Emani Fakaotimanava-Lui")`
        pub const CONTACT: Option<&str> = Some("Emani Fakaotimanava-Lui");
        /// `Some("2012-07-18")`
        pub const DATE: Option<&str> = Some("2012-07-18");
        /// `Some("emani@niue.nu")`
        pub const EMAIL: Option<&str> = Some("emani@niue.nu");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Niuean")`
        pub const LANGUAGE: Option<&str> = Some("Niuean");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RockET Systems")`
        pub const SOURCE: Option<&str> = Some("RockET Systems");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("New Zealand")`
        pub const TERRITORY: Option<&str> = Some("New Zealand");
        /// `Some("Niuean (Vagahau Niue) locale for New Zealand")`
        pub const TITLE: Option<&str> = Some("Niuean (Vagahau Niue) locale for New Zealand");
    }
    pub use super::niu_NU::LC_MESSAGES;
    pub use super::en_NZ::LC_MONETARY;
    pub use super::en_NZ::LC_NUMERIC;
    pub use super::en_NZ::LC_TELEPHONE;
    pub use super::niu_NU::LC_TIME;
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod nl_AW {
    pub mod LC_ADDRESS {
        /// `Some("ABW")`
        pub const COUNTRY_AB3: Option<&str> = Some("ABW");
        /// `Some("AW")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("AW");
        /// `Some("AUA")`
        pub const COUNTRY_CAR: Option<&str> = Some("AUA");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Aruba")`
        pub const COUNTRY_NAME: Option<&str> = Some("Aruba");
        /// `Some(533)`
        pub const COUNTRY_NUM: Option<i64> = Some(533);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("nl")`
        pub const LANG_AB: Option<&str> = Some("nl");
        /// `Some("dut")`
        pub const LANG_LIB: Option<&str> = Some("dut");
        /// `Some("Nederlands")`
        pub const LANG_NAME: Option<&str> = Some("Nederlands");
        /// `Some("nld")`
        pub const LANG_TERM: Option<&str> = Some("nld");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("https://www.gnu.org/software/libc/")`
        pub const ADDRESS: Option<&str> = Some("https://www.gnu.org/software/libc/");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2008-09-16")`
        pub const DATE: Option<&str> = Some("2008-09-16");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Dutch")`
        pub const LANGUAGE: Option<&str> = Some("Dutch");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Free Software Foundation, Inc.")`
        pub const SOURCE: Option<&str> = Some("Free Software Foundation, Inc.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Aruba")`
        pub const TERRITORY: Option<&str> = Some("Aruba");
        /// `Some("Dutch language locale for Aruba")`
        pub const TITLE: Option<&str> = Some("Dutch language locale for Aruba");
    }
    pub use super::nl_NL::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"Afl."`
        pub const CURRENCY_SYMBOL: &str = "Afl.";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"AWG "`
        pub const INT_CURR_SYMBOL: &str = "AWG ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const MON_THOUSANDS_SEP: &str = ".";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `2`
        pub const N_SEP_BY_SPACE: i64 = 2;
        /// `4`
        pub const N_SIGN_POSN: i64 = 4;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub use super::nl_NL::LC_NUMERIC;
    pub mod LC_TELEPHONE {
        /// `Some("297")`
        pub const INT_PREFIX: Option<&str> = Some("297");
        /// `Some("011")`
        pub const INT_SELECT: Option<&str> = Some("011");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["zo", "ma", "di", "wo", "do", "vr", "za"]`
        pub const ABDAY: &[&str] = &["zo", "ma", "di", "wo", "do", "vr", "za"];
        /// `&["jan", "feb", "mrt", "apr", "mei", "jun", "jul", "aug", "sep", "okt", "nov", "dec"]`
        pub const ABMON: &[&str] = &["jan", "feb", "mrt", "apr", "mei", "jun", "jul", "aug", "sep", "okt", "nov", "dec"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a %e %b %Y %k:%M:%S %Z")`
        pub const DATE_FMT: Option<&str> = Some("%a %e %b %Y %k:%M:%S %Z");
        /// `&["zondag", "maandag", "dinsdag", "woensdag", "donderdag", "vrijdag", "zaterdag"]`
        pub const DAY: &[&str] = &["zondag", "maandag", "dinsdag", "woensdag", "donderdag", "vrijdag", "zaterdag"];
        /// `"%d-%m-%y"`
        pub const D_FMT: &str = "%d-%m-%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["januari", "februari", "maart", "april", "mei", "juni", "juli", "augustus", "september", "oktober", "november", "december"]`
        pub const MON: &[&str] = &["januari", "februari", "maart", "april", "mei", "juni", "juli", "augustus", "september", "oktober", "november", "december"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod nl_BE {
    pub mod LC_ADDRESS {
        /// `Some("BEL")`
        pub const COUNTRY_AB3: Option<&str> = Some("BEL");
        /// `Some("BE")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("BE");
        /// `Some("B")`
        pub const COUNTRY_CAR: Option<&str> = Some("B");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("België")`
        pub const COUNTRY_NAME: Option<&str> = Some("België");
        /// `Some(56)`
        pub const COUNTRY_NUM: Option<i64> = Some(56);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("nl")`
        pub const LANG_AB: Option<&str> = Some("nl");
        /// `Some("dut")`
        pub const LANG_LIB: Option<&str> = Some("dut");
        /// `Some("Nederlands")`
        pub const LANG_NAME: Option<&str> = Some("Nederlands");
        /// `Some("nld")`
        pub const LANG_TERM: Option<&str> = Some("nld");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Flemish")`
        pub const LANGUAGE: Option<&str> = Some("Flemish");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Belgium")`
        pub const TERRITORY: Option<&str> = Some("Belgium");
        /// `Some("Dutch locale for Belgium")`
        pub const TITLE: Option<&str> = Some("Dutch locale for Belgium");
    }
    pub use super::nl_NL::LC_MESSAGES;
    pub use super::nl_NL::LC_MONETARY;
    pub use super::fr_BE::LC_NUMERIC;
    pub use super::fr_BE::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["zo", "ma", "di", "wo", "do", "vr", "za"]`
        pub const ABDAY: &[&str] = &["zo", "ma", "di", "wo", "do", "vr", "za"];
        /// `&["jan", "feb", "mrt", "apr", "mei", "jun", "jul", "aug", "sep", "okt", "nov", "dec"]`
        pub const ABMON: &[&str] = &["jan", "feb", "mrt", "apr", "mei", "jun", "jul", "aug", "sep", "okt", "nov", "dec"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["zondag", "maandag", "dinsdag", "woensdag", "donderdag", "vrijdag", "zaterdag"]`
        pub const DAY: &[&str] = &["zondag", "maandag", "dinsdag", "woensdag", "donderdag", "vrijdag", "zaterdag"];
        /// `"%d-%m-%y"`
        pub const D_FMT: &str = "%d-%m-%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["januari", "februari", "maart", "april", "mei", "juni", "juli", "augustus", "september", "oktober", "november", "december"]`
        pub const MON: &[&str] = &["januari", "februari", "maart", "april", "mei", "juni", "juli", "augustus", "september", "oktober", "november", "december"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod nl_BE_euro {
    pub use super::nl_BE::LC_ADDRESS;
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("https://www.gnu.org/software/libc/")`
        pub const ADDRESS: Option<&str> = Some("https://www.gnu.org/software/libc/");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-08-21")`
        pub const DATE: Option<&str> = Some("2000-08-21");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Dutch")`
        pub const LANGUAGE: Option<&str> = Some("Dutch");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Free Software Foundation, Inc.")`
        pub const SOURCE: Option<&str> = Some("Free Software Foundation, Inc.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Belgium")`
        pub const TERRITORY: Option<&str> = Some("Belgium");
        /// `Some("Dutch locale for Belgium with Euro")`
        pub const TITLE: Option<&str> = Some("Dutch locale for Belgium with Euro");
    }
    pub use super::nl_BE::LC_MESSAGES;
    pub use super::fr_BE_euro::LC_MONETARY;
    pub use super::nl_BE::LC_NUMERIC;
    pub use super::nl_BE::LC_TELEPHONE;
    pub use super::nl_BE::LC_TIME;
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod nl_NL {
    pub mod LC_ADDRESS {
        /// `Some("NLD")`
        pub const COUNTRY_AB3: Option<&str> = Some("NLD");
        /// `Some("NL")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("NL");
        /// `Some("NL")`
        pub const COUNTRY_CAR: Option<&str> = Some("NL");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Nederland")`
        pub const COUNTRY_NAME: Option<&str> = Some("Nederland");
        /// `Some(528)`
        pub const COUNTRY_NUM: Option<i64> = Some(528);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("nl")`
        pub const LANG_AB: Option<&str> = Some("nl");
        /// `Some("dut")`
        pub const LANG_LIB: Option<&str> = Some("dut");
        /// `Some("Nederlands")`
        pub const LANG_NAME: Option<&str> = Some("Nederlands");
        /// `Some("nld")`
        pub const LANG_TERM: Option<&str> = Some("nld");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Dutch")`
        pub const LANGUAGE: Option<&str> = Some("Dutch");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Netherlands")`
        pub const TERRITORY: Option<&str> = Some("Netherlands");
        /// `Some("Dutch locale for the Netherlands")`
        pub const TITLE: Option<&str> = Some("Dutch locale for the Netherlands");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("nee")`
        pub const NOSTR: Option<&str> = Some("nee");
        /// `"^[+1jJyY]"`
        pub const YESEXPR: &str = "^[+1jJyY]";
        /// `Some("ja")`
        pub const YESSTR: Option<&str> = Some("ja");
    }
    pub mod LC_MONETARY {
        /// `"€"`
        pub const CURRENCY_SYMBOL: &str = "€";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"EUR "`
        pub const INT_CURR_SYMBOL: &str = "EUR ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const MON_THOUSANDS_SEP: &str = ".";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `2`
        pub const N_SEP_BY_SPACE: i64 = 2;
        /// `4`
        pub const N_SIGN_POSN: i64 = 4;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const THOUSANDS_SEP: &str = ".";
    }
    pub mod LC_TELEPHONE {
        /// `Some("31")`
        pub const INT_PREFIX: Option<&str> = Some("31");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["zo", "ma", "di", "wo", "do", "vr", "za"]`
        pub const ABDAY: &[&str] = &["zo", "ma", "di", "wo", "do", "vr", "za"];
        /// `&["jan", "feb", "mrt", "apr", "mei", "jun", "jul", "aug", "sep", "okt", "nov", "dec"]`
        pub const ABMON: &[&str] = &["jan", "feb", "mrt", "apr", "mei", "jun", "jul", "aug", "sep", "okt", "nov", "dec"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a %e %b %Y %k:%M:%S %Z")`
        pub const DATE_FMT: Option<&str> = Some("%a %e %b %Y %k:%M:%S %Z");
        /// `&["zondag", "maandag", "dinsdag", "woensdag", "donderdag", "vrijdag", "zaterdag"]`
        pub const DAY: &[&str] = &["zondag", "maandag", "dinsdag", "woensdag", "donderdag", "vrijdag", "zaterdag"];
        /// `"%d-%m-%y"`
        pub const D_FMT: &str = "%d-%m-%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["januari", "februari", "maart", "april", "mei", "juni", "juli", "augustus", "september", "oktober", "november", "december"]`
        pub const MON: &[&str] = &["januari", "februari", "maart", "april", "mei", "juni", "juli", "augustus", "september", "oktober", "november", "december"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod nl_NL_euro {
    pub use super::nl_NL::LC_ADDRESS;
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("https://www.gnu.org/software/libc/")`
        pub const ADDRESS: Option<&str> = Some("https://www.gnu.org/software/libc/");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-08-20")`
        pub const DATE: Option<&str> = Some("2000-08-20");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Dutch")`
        pub const LANGUAGE: Option<&str> = Some("Dutch");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Free Software Foundation, Inc.")`
        pub const SOURCE: Option<&str> = Some("Free Software Foundation, Inc.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Netherlands")`
        pub const TERRITORY: Option<&str> = Some("Netherlands");
        /// `Some("Dutch locale for the Netherlands with Euro")`
        pub const TITLE: Option<&str> = Some("Dutch locale for the Netherlands with Euro");
    }
    pub use super::nl_NL::LC_MESSAGES;
    pub use super::nl_NL::LC_MONETARY;
    pub use super::nl_NL::LC_NUMERIC;
    pub use super::nl_NL::LC_TELEPHONE;
    pub use super::nl_NL::LC_TIME;
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod nn_NO {
    pub mod LC_ADDRESS {
        /// `Some("NOR")`
        pub const COUNTRY_AB3: Option<&str> = Some("NOR");
        /// `Some("NO")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("NO");
        /// `Some("N")`
        pub const COUNTRY_CAR: Option<&str> = Some("N");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Noreg")`
        pub const COUNTRY_NAME: Option<&str> = Some("Noreg");
        /// `Some(578)`
        pub const COUNTRY_NUM: Option<i64> = Some(578);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("nn")`
        pub const LANG_AB: Option<&str> = Some("nn");
        /// `Some("nno")`
        pub const LANG_LIB: Option<&str> = Some("nno");
        /// `Some("nynorsk")`
        pub const LANG_NAME: Option<&str> = Some("nynorsk");
        /// `Some("nno")`
        pub const LANG_TERM: Option<&str> = Some("nno");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan")`
        pub const ADDRESS: Option<&str> = Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-08-31")`
        pub const DATE: Option<&str> = Some("2000-08-31");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Norwegian Nynorsk")`
        pub const LANGUAGE: Option<&str> = Some("Norwegian Nynorsk");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("IBM Globalization Center of Competency, Yamato Software Laboratory")`
        pub const SOURCE: Option<&str> = Some("IBM Globalization Center of Competency, Yamato Software Laboratory");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Norway")`
        pub const TERRITORY: Option<&str> = Some("Norway");
        /// `Some("Nynorsk language locale for Norway")`
        pub const TITLE: Option<&str> = Some("Nynorsk language locale for Norway");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0Nn]"`
        pub const NOEXPR: &str = "^[-0Nn]";
        /// `Some("nei")`
        pub const NOSTR: Option<&str> = Some("nei");
        /// `"^[+1JjYy]"`
        pub const YESEXPR: &str = "^[+1JjYy]";
        /// `Some("ja")`
        pub const YESSTR: Option<&str> = Some("ja");
    }
    pub mod LC_MONETARY {
        /// `"kr"`
        pub const CURRENCY_SYMBOL: &str = "kr";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"NOK "`
        pub const INT_CURR_SYMBOL: &str = "NOK ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `"\u{202f}"`
        pub const MON_THOUSANDS_SEP: &str = "\u{202f}";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `3`
        pub const N_SIGN_POSN: i64 = 3;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub use super::nb_NO::LC_NUMERIC;
    pub use super::nb_NO::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["su.", "må.", "ty.", "on.", "to.", "fr.", "la."]`
        pub const ABDAY: &[&str] = &["su.", "må.", "ty.", "on.", "to.", "fr.", "la."];
        /// `&["jan.", "feb.", "mars", "april", "mai", "juni", "juli", "aug.", "sep.", "okt.", "nov.", "des."]`
        pub const ABMON: &[&str] = &["jan.", "feb.", "mars", "april", "mai", "juni", "juli", "aug.", "sep.", "okt.", "nov.", "des."];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a %d. %b %H:%M:%S %z %Y")`
        pub const DATE_FMT: Option<&str> = Some("%a %d. %b %H:%M:%S %z %Y");
        /// `&["sundag ", "måndag ", "tysdag ", "onsdag ", "torsdag ", "fredag ", "laurdag "]`
        pub const DAY: &[&str] = &["sundag ", "måndag ", "tysdag ", "onsdag ", "torsdag ", "fredag ", "laurdag "];
        /// `"%d. %b %Y"`
        pub const D_FMT: &str = "%d. %b %Y";
        /// `"%a %d. %b %Y kl. %H.%M %z"`
        pub const D_T_FMT: &str = "%a %d. %b %Y kl. %H.%M %z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["januar", "februar", "mars", "april", "mai", "juni", "juli", "august", "september", "oktober", "november", "desember"]`
        pub const MON: &[&str] = &["januar", "februar", "mars", "april", "mai", "juni", "juli", "august", "september", "oktober", "november", "desember"];
        /// `"kl. %H.%M %z"`
        pub const T_FMT: &str = "kl. %H.%M %z";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod nr_ZA {
    pub mod LC_ADDRESS {
        /// `Some("ZAF")`
        pub const COUNTRY_AB3: Option<&str> = Some("ZAF");
        /// `Some("ZA")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("ZA");
        /// `Some("ZA")`
        pub const COUNTRY_CAR: Option<&str> = Some("ZA");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("iSewula Afrika")`
        pub const COUNTRY_NAME: Option<&str> = Some("iSewula Afrika");
        /// `Some(710)`
        pub const COUNTRY_NUM: Option<i64> = Some(710);
        /// `Some("ZA")`
        pub const COUNTRY_POST: Option<&str> = Some("ZA");
        /// `Some("nr")`
        pub const LANG_AB: Option<&str> = Some("nr");
        /// `Some("nbl")`
        pub const LANG_LIB: Option<&str> = Some("nbl");
        /// `Some("isiNdebele")`
        pub const LANG_NAME: Option<&str> = Some("isiNdebele");
        /// `Some("nbl")`
        pub const LANG_TERM: Option<&str> = Some("nbl");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("PO Box 28364, Sunnyside, 0132, South Africa")`
        pub const ADDRESS: Option<&str> = Some("PO Box 28364, Sunnyside, 0132, South Africa");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Dwayne Bailey")`
        pub const CONTACT: Option<&str> = Some("Dwayne Bailey");
        /// `Some("2005-10-13")`
        pub const DATE: Option<&str> = Some("2005-10-13");
        /// `Some("dwayne@translate.org.za")`
        pub const EMAIL: Option<&str> = Some("dwayne@translate.org.za");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("South Ndebele")`
        pub const LANGUAGE: Option<&str> = Some("South Ndebele");
        /// `Some("0.3")`
        pub const REVISION: Option<&str> = Some("0.3");
        /// `Some("Zuza Software Foundation (Translate.org.za)")`
        pub const SOURCE: Option<&str> = Some("Zuza Software Foundation (Translate.org.za)");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("South Africa")`
        pub const TERRITORY: Option<&str> = Some("South Africa");
        /// `Some("Southern Ndebele locale for South Africa")`
        pub const TITLE: Option<&str> = Some("Southern Ndebele locale for South Africa");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `None`
        pub const NOSTR: Option<&str> = None;
        /// `"^[+1yY]"`
        pub const YESEXPR: &str = "^[+1yY]";
        /// `None`
        pub const YESSTR: Option<&str> = None;
    }
    pub use super::en_ZA::LC_MONETARY;
    pub use super::en_ZA::LC_NUMERIC;
    pub use super::en_ZA::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Son", "Mvu", "Bil", "Tha", "Ne", "Hla", "Gqi"]`
        pub const ABDAY: &[&str] = &["Son", "Mvu", "Bil", "Tha", "Ne", "Hla", "Gqi"];
        /// `&["Jan", "Feb", "Mat", "Apr", "Mey", "Jun", "Jul", "Arh", "Sep", "Okt", "Usi", "Dis"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Mat", "Apr", "Mey", "Jun", "Jul", "Arh", "Sep", "Okt", "Usi", "Dis"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a %b %-e %H:%M:%S %Z %Y")`
        pub const DATE_FMT: Option<&str> = Some("%a %b %-e %H:%M:%S %Z %Y");
        /// `&["uSonto", "uMvulo", "uLesibili", "lesithathu", "uLesine", "ngoLesihlanu", "umGqibelo"]`
        pub const DAY: &[&str] = &["uSonto", "uMvulo", "uLesibili", "lesithathu", "uLesine", "ngoLesihlanu", "umGqibelo"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%a %-e %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %-e %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Janabari", "uFeberbari", "uMatjhi", "u-Apreli", "Meyi", "Juni", "Julayi", "Arhostosi", "Septemba", "Oktoba", "Usinyikhaba", "Disemba"]`
        pub const MON: &[&str] = &["Janabari", "uFeberbari", "uMatjhi", "u-Apreli", "Meyi", "Juni", "Julayi", "Arhostosi", "Septemba", "Oktoba", "Usinyikhaba", "Disemba"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod nso_ZA {
    pub mod LC_ADDRESS {
        /// `Some("ZAF")`
        pub const COUNTRY_AB3: Option<&str> = Some("ZAF");
        /// `Some("ZA")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("ZA");
        /// `Some("ZA")`
        pub const COUNTRY_CAR: Option<&str> = Some("ZA");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Afrika Borwa")`
        pub const COUNTRY_NAME: Option<&str> = Some("Afrika Borwa");
        /// `Some(710)`
        pub const COUNTRY_NUM: Option<i64> = Some(710);
        /// `Some("ZA")`
        pub const COUNTRY_POST: Option<&str> = Some("ZA");
        /// `Some("")`
        pub const LANG_AB: Option<&str> = Some("");
        /// `Some("nso")`
        pub const LANG_LIB: Option<&str> = Some("nso");
        /// `Some("Sesotho sa Leboa")`
        pub const LANG_NAME: Option<&str> = Some("Sesotho sa Leboa");
        /// `Some("nso")`
        pub const LANG_TERM: Option<&str> = Some("nso");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("PO Box 28364, Sunnyside, 0132, South Africa")`
        pub const ADDRESS: Option<&str> = Some("PO Box 28364, Sunnyside, 0132, South Africa");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Dwayne Bailey")`
        pub const CONTACT: Option<&str> = Some("Dwayne Bailey");
        /// `Some("2005-10-13")`
        pub const DATE: Option<&str> = Some("2005-10-13");
        /// `Some("dwayne@translate.org.za")`
        pub const EMAIL: Option<&str> = Some("dwayne@translate.org.za");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Northern Sotho")`
        pub const LANGUAGE: Option<&str> = Some("Northern Sotho");
        /// `Some("0.3")`
        pub const REVISION: Option<&str> = Some("0.3");
        /// `Some("Zuza Software Foundation (Translate.org.za)")`
        pub const SOURCE: Option<&str> = Some("Zuza Software Foundation (Translate.org.za)");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("South Africa")`
        pub const TERRITORY: Option<&str> = Some("South Africa");
        /// `Some("Northern Sotho locale for South Africa")`
        pub const TITLE: Option<&str> = Some("Northern Sotho locale for South Africa");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNaA]"`
        pub const NOEXPR: &str = "^[-0nNaA]";
        /// `None`
        pub const NOSTR: Option<&str> = None;
        /// `"^[+1yYeE]"`
        pub const YESEXPR: &str = "^[+1yYeE]";
        /// `None`
        pub const YESSTR: Option<&str> = None;
    }
    pub use super::en_ZA::LC_MONETARY;
    pub use super::en_ZA::LC_NUMERIC;
    pub use super::en_ZA::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Son", "Moš", "Bed", "Rar", "Ne", "Hla", "Mok"]`
        pub const ABDAY: &[&str] = &["Son", "Moš", "Bed", "Rar", "Ne", "Hla", "Mok"];
        /// `&["Jan", "Feb", "Mat", "Apr", "Mei", "Jun", "Jul", "Ago", "Set", "Okt", "Nof", "Dis"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Mat", "Apr", "Mei", "Jun", "Jul", "Ago", "Set", "Okt", "Nof", "Dis"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["LaMorena", "Mošupologo", "Labobedi", "Laboraro", "Labone", "Labohlano", "Mokibelo"]`
        pub const DAY: &[&str] = &["LaMorena", "Mošupologo", "Labobedi", "Laboraro", "Labone", "Labohlano", "Mokibelo"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Janaware", "Febereware", "Matšhe", "Aprele", "Mei", "June", "Julae", "Agostose", "Setemere", "Oktobere", "Nofemere", "Disemere"]`
        pub const MON: &[&str] = &["Janaware", "Febereware", "Matšhe", "Aprele", "Mei", "June", "Julae", "Agostose", "Setemere", "Oktobere", "Nofemere", "Disemere"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod oc_FR {
    pub mod LC_ADDRESS {
        /// `Some("FRA")`
        pub const COUNTRY_AB3: Option<&str> = Some("FRA");
        /// `Some("FR")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("FR");
        /// `Some("F")`
        pub const COUNTRY_CAR: Option<&str> = Some("F");
        /// `Some("979-10")`
        pub const COUNTRY_ISBN: Option<&str> = Some("979-10");
        /// `Some("França")`
        pub const COUNTRY_NAME: Option<&str> = Some("França");
        /// `Some(250)`
        pub const COUNTRY_NUM: Option<i64> = Some(250);
        /// `Some("F")`
        pub const COUNTRY_POST: Option<&str> = Some("F");
        /// `Some("oc")`
        pub const LANG_AB: Option<&str> = Some("oc");
        /// `Some("oci")`
        pub const LANG_LIB: Option<&str> = Some("oci");
        /// `Some("occitan")`
        pub const LANG_NAME: Option<&str> = Some("occitan");
        /// `Some("oci")`
        pub const LANG_TERM: Option<&str> = Some("oci");
        /// `Some("%d%N%f%N%d%N%b%N%s %h 5e %r%N%C%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%d%N%f%N%d%N%b%N%s %h 5e %r%N%C%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Los Genets, 12290 Pont de Salars")`
        pub const ADDRESS: Option<&str> = Some("Los Genets, 12290 Pont de Salars");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_TELEPHONE"],
        ]);
        /// `Some("Jean-Paul Fraysse")`
        pub const CONTACT: Option<&str> = Some("Jean-Paul Fraysse");
        /// `Some("2018-04-12")`
        pub const DATE: Option<&str> = Some("2018-04-12");
        /// `Some("Jean-Paul.Fraysse@wanadoo.fr")`
        pub const EMAIL: Option<&str> = Some("Jean-Paul.Fraysse@wanadoo.fr");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Occitan")`
        pub const LANGUAGE: Option<&str> = Some("Occitan");
        /// `Some("0.3")`
        pub const REVISION: Option<&str> = Some("0.3");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("France")`
        pub const TERRITORY: Option<&str> = Some("France");
        /// `Some("Occitan Language Locale for France")`
        pub const TITLE: Option<&str> = Some("Occitan Language Locale for France");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `None`
        pub const NOSTR: Option<&str> = None;
        /// `"^[+1oOsSyY]"`
        pub const YESEXPR: &str = "^[+1oOsSyY]";
        /// `None`
        pub const YESSTR: Option<&str> = None;
    }
    pub use super::fr_FR::LC_MONETARY;
    pub use super::fr_FR::LC_NUMERIC;
    pub use super::fr_FR::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["dg.", "dl.", "dm.", "dc.", "dj.", "dv.", "ds."]`
        pub const ABDAY: &[&str] = &["dg.", "dl.", "dm.", "dc.", "dj.", "dv.", "ds."];
        /// `&["gen.", "febr.", "març", "abr.", "mai", "junh", "jul.", "ago.", "set.", "oct.", "nov.", "dec."]`
        pub const ABMON: &[&str] = &["gen.", "febr.", "març", "abr.", "mai", "junh", "jul.", "ago.", "set.", "oct.", "nov.", "dec."];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `Some(&["genièr", "febrièr", "març", "abrial", "mai", "junh", "julhet", "agost", "setembre", "octòbre", "novembre", "decembre"])`
        pub const ALT_MON: Option<&[&str]> = Some(&["genièr", "febrièr", "març", "abrial", "mai", "junh", "julhet", "agost", "setembre", "octòbre", "novembre", "decembre"]);
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["dimenge", "diluns", "dimars", "dimècres", "dijóus", "divendres", "dissabte"]`
        pub const DAY: &[&str] = &["dimenge", "diluns", "dimars", "dimècres", "dijóus", "divendres", "dissabte"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["de genièr", "de febrièr", "de març", "d’abrial", "de mai", "de junh", "de julhet", "d’agost", "de setembre", "d’octòbre", "de novembre", "de decembre"]`
        pub const MON: &[&str] = &["de genièr", "de febrièr", "de març", "d’abrial", "de mai", "de junh", "de julhet", "d’agost", "de setembre", "d’octòbre", "de novembre", "de decembre"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod om_ET {
    pub mod LC_ADDRESS {
        /// `Some("ETH")`
        pub const COUNTRY_AB3: Option<&str> = Some("ETH");
        /// `Some("ET")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("ET");
        /// `Some("ETH")`
        pub const COUNTRY_CAR: Option<&str> = Some("ETH");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Itoophiyaa")`
        pub const COUNTRY_NAME: Option<&str> = Some("Itoophiyaa");
        /// `Some(231)`
        pub const COUNTRY_NUM: Option<i64> = Some(231);
        /// `Some("ETH")`
        pub const COUNTRY_POST: Option<&str> = Some("ETH");
        /// `Some("om")`
        pub const LANG_AB: Option<&str> = Some("om");
        /// `Some("orm")`
        pub const LANG_LIB: Option<&str> = Some("orm");
        /// `Some("Oromoo")`
        pub const LANG_NAME: Option<&str> = Some("Oromoo");
        /// `Some("orm")`
        pub const LANG_TERM: Option<&str> = Some("orm");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("7802 Solomon Seal Dr., Springfield, VA 22152, USA")`
        pub const ADDRESS: Option<&str> = Some("7802 Solomon Seal Dr., Springfield, VA 22152, USA");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TIME"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TIME"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2003-07-05")`
        pub const DATE: Option<&str> = Some("2003-07-05");
        /// `Some("locales@geez.org")`
        pub const EMAIL: Option<&str> = Some("locales@geez.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Oromo")`
        pub const LANGUAGE: Option<&str> = Some("Oromo");
        /// `Some("0.20")`
        pub const REVISION: Option<&str> = Some("0.20");
        /// `Some("Ge'ez Frontier Foundation & Sagalee Oromoo Publishing Co. Inc.")`
        pub const SOURCE: Option<&str> = Some("Ge'ez Frontier Foundation & Sagalee Oromoo Publishing Co. Inc.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Ethiopia")`
        pub const TERRITORY: Option<&str> = Some("Ethiopia");
        /// `Some("Oromo language locale for Ethiopia.")`
        pub const TITLE: Option<&str> = Some("Oromo language locale for Ethiopia.");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0mMnN]"`
        pub const NOEXPR: &str = "^[-0mMnN]";
        /// `Some("miti")`
        pub const NOSTR: Option<&str> = Some("miti");
        /// `"^[+1eEyY]"`
        pub const YESEXPR: &str = "^[+1eEyY]";
        /// `Some("eeyyee")`
        pub const YESSTR: Option<&str> = Some("eeyyee");
    }
    pub use super::ti_ET::LC_MONETARY;
    pub use super::ti_ET::LC_NUMERIC;
    pub use super::ti_ET::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Dil", "Wix", "Qib", "Rob", "Kam", "Jim", "San"]`
        pub const ABDAY: &[&str] = &["Dil", "Wix", "Qib", "Rob", "Kam", "Jim", "San"];
        /// `&["Ama", "Gur", "Bit", "Elb", "Cam", "Wax", "Ado", "Hag", "Ful", "Onk", "Sad", "Mud"]`
        pub const ABMON: &[&str] = &["Ama", "Gur", "Bit", "Elb", "Cam", "Wax", "Ado", "Hag", "Ful", "Onk", "Sad", "Mud"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["WD", "WB"]`
        pub const AM_PM: &[&str] = &["WD", "WB"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%A, %B %e, %r %Z %Y")`
        pub const DATE_FMT: Option<&str> = Some("%A, %B %e, %r %Z %Y");
        /// `&["Dilbata", "Wiixata", "Qibxata", "Roobii", "Kamiisa", "Jimaata", "Sanbata"]`
        pub const DAY: &[&str] = &["Dilbata", "Wiixata", "Qibxata", "Roobii", "Kamiisa", "Jimaata", "Sanbata"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%A, %B %e, %Y %l:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A, %B %e, %Y %l:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Amajjii", "Guraandhala", "Bitooteessa", "Elba", "Caamsa", "Waxabajjii", "Adooleessa", "Hagayya", "Fuulbana", "Onkololeessa", "Sadaasa", "Muddee"]`
        pub const MON: &[&str] = &["Amajjii", "Guraandhala", "Bitooteessa", "Elba", "Caamsa", "Waxabajjii", "Adooleessa", "Hagayya", "Fuulbana", "Onkololeessa", "Sadaasa", "Muddee"];
        /// `"%l:%M:%S %p"`
        pub const T_FMT: &str = "%l:%M:%S %p";
        /// `"%l:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%l:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod om_KE {
    pub mod LC_ADDRESS {
        /// `Some("KEN")`
        pub const COUNTRY_AB3: Option<&str> = Some("KEN");
        /// `Some("KE")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("KE");
        /// `Some("EAK")`
        pub const COUNTRY_CAR: Option<&str> = Some("EAK");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Keeniyaa")`
        pub const COUNTRY_NAME: Option<&str> = Some("Keeniyaa");
        /// `Some(404)`
        pub const COUNTRY_NUM: Option<i64> = Some(404);
        /// `Some("KEN")`
        pub const COUNTRY_POST: Option<&str> = Some("KEN");
        /// `Some("om")`
        pub const LANG_AB: Option<&str> = Some("om");
        /// `Some("orm")`
        pub const LANG_LIB: Option<&str> = Some("orm");
        /// `Some("Oromoo")`
        pub const LANG_NAME: Option<&str> = Some("Oromoo");
        /// `Some("orm")`
        pub const LANG_TERM: Option<&str> = Some("orm");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("7802 Solomon Seal Dr., Springfield, VA 22152, USA")`
        pub const ADDRESS: Option<&str> = Some("7802 Solomon Seal Dr., Springfield, VA 22152, USA");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TIME"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TIME"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2003-07-05")`
        pub const DATE: Option<&str> = Some("2003-07-05");
        /// `Some("locales@geez.org")`
        pub const EMAIL: Option<&str> = Some("locales@geez.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Oromo")`
        pub const LANGUAGE: Option<&str> = Some("Oromo");
        /// `Some("0.20")`
        pub const REVISION: Option<&str> = Some("0.20");
        /// `Some("Ge'ez Frontier Foundation & Sagalee Oromoo Publishing Co. Inc.")`
        pub const SOURCE: Option<&str> = Some("Ge'ez Frontier Foundation & Sagalee Oromoo Publishing Co. Inc.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Kenya")`
        pub const TERRITORY: Option<&str> = Some("Kenya");
        /// `Some("Oromo language locale for Kenya.")`
        pub const TITLE: Option<&str> = Some("Oromo language locale for Kenya.");
    }
    pub use super::om_ET::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"Ksh"`
        pub const CURRENCY_SYMBOL: &str = "Ksh";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"KES "`
        pub const INT_CURR_SYMBOL: &str = "KES ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("254")`
        pub const INT_PREFIX: Option<&str> = Some("254");
        /// `Some("000")`
        pub const INT_SELECT: Option<&str> = Some("000");
        /// `Some("%a-%l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("%a-%l");
        /// `Some("%c-%a-%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("%c-%a-%l");
    }
    pub use super::om_ET::LC_TIME;
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod or_IN {
    pub mod LC_ADDRESS {
        /// `Some("IND")`
        pub const COUNTRY_AB3: Option<&str> = Some("IND");
        /// `Some("IN")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IN");
        /// `Some("IND")`
        pub const COUNTRY_CAR: Option<&str> = Some("IND");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("ଭ\u{b3e}ରତ")`
        pub const COUNTRY_NAME: Option<&str> = Some("ଭ\u{b3e}ରତ");
        /// `Some(356)`
        pub const COUNTRY_NUM: Option<i64> = Some(356);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("or")`
        pub const LANG_AB: Option<&str> = Some("or");
        /// `Some("ori")`
        pub const LANG_LIB: Option<&str> = Some("ori");
        /// `Some("ଓଡ\u{b3c}\u{b3f}ଆ")`
        pub const LANG_NAME: Option<&str> = Some("ଓଡ\u{b3c}\u{b3f}ଆ");
        /// `Some("ori")`
        pub const LANG_TERM: Option<&str> = Some("ori");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan")`
        pub const ADDRESS: Option<&str> = Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2006-05-25")`
        pub const DATE: Option<&str> = Some("2006-05-25");
        /// `Some("bug-glibc@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Odia")`
        pub const LANGUAGE: Option<&str> = Some("Odia");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("IBM AP Linux Technology Center, Yamato Software Laboratory")`
        pub const SOURCE: Option<&str> = Some("IBM AP Linux Technology Center, Yamato Software Laboratory");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("India")`
        pub const TERRITORY: Option<&str> = Some("India");
        /// `Some("Odia language locale for India")`
        pub const TITLE: Option<&str> = Some("Odia language locale for India");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNନ]"`
        pub const NOEXPR: &str = "^[-0nNନ]";
        /// `Some("ନ\u{b3e}")`
        pub const NOSTR: Option<&str> = Some("ନ\u{b3e}");
        /// `"^[+1yYହ]"`
        pub const YESEXPR: &str = "^[+1yYହ]";
        /// `Some("ହ\u{b01}")`
        pub const YESSTR: Option<&str> = Some("ହ\u{b01}");
    }
    pub use super::hi_IN::LC_MONETARY;
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3, 2]`
        pub const GROUPING: &[i64] = &[3, 2];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("91")`
        pub const INT_PREFIX: Option<&str> = Some("91");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c ;%a ;%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c ;%a ;%l");
    }
    pub mod LC_TIME {
        /// `&["ରବ\u{b3f}", "ସୋମ", "ମଙ\u{b4d}ଗଳ", "ବ\u{b41}ଧ", "ଗ\u{b41}ର\u{b41}", "ଶ\u{b41}କ\u{b4d}ର", "ଶନ\u{b3f}"]`
        pub const ABDAY: &[&str] = &["ରବ\u{b3f}", "ସୋମ", "ମଙ\u{b4d}ଗଳ", "ବ\u{b41}ଧ", "ଗ\u{b41}ର\u{b41}", "ଶ\u{b41}କ\u{b4d}ର", "ଶନ\u{b3f}"];
        /// `&["ଜ\u{b3e}ନ\u{b41}ଆରୀ", "ଫେବ\u{b43}ଆରୀ", "ମ\u{b3e}ର\u{b4d}ଚ\u{b4d}ଚ", "ଅପ\u{b4d}ରେଲ", "ମଇ", "ଜ\u{b41}ନ", "ଜ\u{b41}ଲ\u{b3e}ଇ", "ଅଗଷ\u{b4d}ଟ", "ସେପ\u{b4d}ଟେମ\u{b4d}ବର", "ଅକ\u{b4d}ଟୋବର", "ନଭେମ\u{b4d}ବର", "ଡ\u{b3f}ସେମ\u{b4d}ବର"]`
        pub const ABMON: &[&str] = &["ଜ\u{b3e}ନ\u{b41}ଆରୀ", "ଫେବ\u{b43}ଆରୀ", "ମ\u{b3e}ର\u{b4d}ଚ\u{b4d}ଚ", "ଅପ\u{b4d}ରେଲ", "ମଇ", "ଜ\u{b41}ନ", "ଜ\u{b41}ଲ\u{b3e}ଇ", "ଅଗଷ\u{b4d}ଟ", "ସେପ\u{b4d}ଟେମ\u{b4d}ବର", "ଅକ\u{b4d}ଟୋବର", "ନଭେମ\u{b4d}ବର", "ଡ\u{b3f}ସେମ\u{b4d}ବର"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `Some(&["୦", "୧", "୨", "୩", "୪", "୫", "୬", "୭", "୮", "୯", "୧୦", "୧୧", "୧୨", "୧୩", "୧୪", "୧୫", "୧୬", "୧୭", "୧୮", "୧୯", "୨୦", "୨୧", "୨୨", "୨୩", "୨୪", "୨୫", "୨୬", "୨୭", "୨୮", "୨୯", "୩୦", "୩୧", "୩୨", "୩୩", "୩୪", "୩୫", "୩୬", "୩୭", "୩୮", "୩୯", "୪୦", "୪୧", "୪୨", "୪୩", "୪୪", "୪୫", "୪୬", "୪୭", "୪୮", "୪୯", "୫୦", "୫୧", "୫୨", "୫୩", "୫୪", "୫୫", "୫୬", "୫୭", "୫୮", "୫୯", "୬୦", "୬୧", "୬୨", "୬୩", "୬୪", "୬୫", "୬୬", "୬୭", "୬୮", "୬୯", "୭୦", "୭୧", "୭୨", "୭୩", "୭୪", "୭୫", "୭୬", "୭୭", "୭୮", "୭୯", "୮୦", "୮୧", "୮୨", "୮୩", "୮୪", "୮୫", "୮୬", "୮୭", "୮୮", "୮୯", "୯୦", "୯୧", "୯୨", "୯୩", "୯୪", "୯୫", "୯୬", "୯୭", "୯୮", "୯୯"])`
        pub const ALT_DIGITS: Option<&[&str]> = Some(&["୦", "୧", "୨", "୩", "୪", "୫", "୬", "୭", "୮", "୯", "୧୦", "୧୧", "୧୨", "୧୩", "୧୪", "୧୫", "୧୬", "୧୭", "୧୮", "୧୯", "୨୦", "୨୧", "୨୨", "୨୩", "୨୪", "୨୫", "୨୬", "୨୭", "୨୮", "୨୯", "୩୦", "୩୧", "୩୨", "୩୩", "୩୪", "୩୫", "୩୬", "୩୭", "୩୮", "୩୯", "୪୦", "୪୧", "୪୨", "୪୩", "୪୪", "୪୫", "୪୬", "୪୭", "୪୮", "୪୯", "୫୦", "୫୧", "୫୨", "୫୩", "୫୪", "୫୫", "୫୬", "୫୭", "୫୮", "୫୯", "୬୦", "୬୧", "୬୨", "୬୩", "୬୪", "୬୫", "୬୬", "୬୭", "୬୮", "୬୯", "୭୦", "୭୧", "୭୨", "୭୩", "୭୪", "୭୫", "୭୬", "୭୭", "୭୮", "୭୯", "୮୦", "୮୧", "୮୨", "୮୩", "୮୪", "୮୫", "୮୬", "୮୭", "୮୮", "୮୯", "୯୦", "୯୧", "୯୨", "୯୩", "୯୪", "୯୫", "୯୬", "୯୭", "୯୮", "୯୯"]);
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["AM", "PM"]`
        pub const AM_PM: &[&str] = &["AM", "PM"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["ରବ\u{b3f}ବ\u{b3e}ର", "ସୋମବ\u{b3e}ର", "ମଙ\u{b4d}ଗଳବ\u{b3e}ର", "ବ\u{b41}ଧବ\u{b3e}ର", "ଗ\u{b41}ର\u{b41}ବ\u{b3e}ର", "ଶ\u{b41}କ\u{b4d}ରବ\u{b3e}ର", "ଶନ\u{b3f}ବ\u{b3e}ର"]`
        pub const DAY: &[&str] = &["ରବ\u{b3f}ବ\u{b3e}ର", "ସୋମବ\u{b3e}ର", "ମଙ\u{b4d}ଗଳବ\u{b3e}ର", "ବ\u{b41}ଧବ\u{b3e}ର", "ଗ\u{b41}ର\u{b41}ବ\u{b3e}ର", "ଶ\u{b41}କ\u{b4d}ରବ\u{b3e}ର", "ଶନ\u{b3f}ବ\u{b3e}ର"];
        /// `"%Od-%Om-%Oy"`
        pub const D_FMT: &str = "%Od-%Om-%Oy";
        /// `"%Oe %B %Oy %OI:%OM:%OS %p %Z"`
        pub const D_T_FMT: &str = "%Oe %B %Oy %OI:%OM:%OS %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["ଜ\u{b3e}ନ\u{b41}ଆରୀ", "ଫେବ\u{b43}ଆରୀ", "ମ\u{b3e}ର\u{b4d}ଚ\u{b4d}ଚ", "ଅପ\u{b4d}ରେଲ", "ମଇ", "ଜ\u{b41}ନ", "ଜ\u{b41}ଲ\u{b3e}ଇ", "ଅଗଷ\u{b4d}ଟ", "ସେପ\u{b4d}ଟେମ\u{b4d}ବର", "ଅକ\u{b4d}ଟୋବର", "ନଭେମ\u{b4d}ବର", "ଡ\u{b3f}ସେମ\u{b4d}ବର"]`
        pub const MON: &[&str] = &["ଜ\u{b3e}ନ\u{b41}ଆରୀ", "ଫେବ\u{b43}ଆରୀ", "ମ\u{b3e}ର\u{b4d}ଚ\u{b4d}ଚ", "ଅପ\u{b4d}ରେଲ", "ମଇ", "ଜ\u{b41}ନ", "ଜ\u{b41}ଲ\u{b3e}ଇ", "ଅଗଷ\u{b4d}ଟ", "ସେପ\u{b4d}ଟେମ\u{b4d}ବର", "ଅକ\u{b4d}ଟୋବର", "ନଭେମ\u{b4d}ବର", "ଡ\u{b3f}ସେମ\u{b4d}ବର"];
        /// `"%OI:%OM:%OS %p"`
        pub const T_FMT: &str = "%OI:%OM:%OS %p";
        /// `"%OI:%OM:%OS %p"`
        pub const T_FMT_AMPM: &str = "%OI:%OM:%OS %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod os_RU {
    pub mod LC_ADDRESS {
        /// `Some("RUS")`
        pub const COUNTRY_AB3: Option<&str> = Some("RUS");
        /// `Some("RU")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("RU");
        /// `Some("RUS")`
        pub const COUNTRY_CAR: Option<&str> = Some("RUS");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Уӕрӕсе")`
        pub const COUNTRY_NAME: Option<&str> = Some("Уӕрӕсе");
        /// `Some(643)`
        pub const COUNTRY_NUM: Option<i64> = Some(643);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("os")`
        pub const LANG_AB: Option<&str> = Some("os");
        /// `Some("oss")`
        pub const LANG_LIB: Option<&str> = Some("oss");
        /// `Some("ирон")`
        pub const LANG_NAME: Option<&str> = Some("ирон");
        /// `Some("oss")`
        pub const LANG_TERM: Option<&str> = Some("oss");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Ossetic")`
        pub const LANGUAGE: Option<&str> = Some("Ossetic");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Russia")`
        pub const TERRITORY: Option<&str> = Some("Russia");
        /// `Some("Ossetian locale for Russia")`
        pub const TITLE: Option<&str> = Some("Ossetian locale for Russia");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNнН]"`
        pub const NOEXPR: &str = "^[-0nNнН]";
        /// `Some("нӕйы")`
        pub const NOSTR: Option<&str> = Some("нӕйы");
        /// `"^[+1yYуУ]"`
        pub const YESEXPR: &str = "^[+1yYуУ]";
        /// `Some("уойы")`
        pub const YESSTR: Option<&str> = Some("уойы");
    }
    pub use super::ru_RU::LC_MONETARY;
    pub use super::ru_RU::LC_NUMERIC;
    pub use super::ru_RU::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Хцб", "Крс", "Дцг", "Æрт", "Цпр", "Мрб", "Сбт"]`
        pub const ABDAY: &[&str] = &["Хцб", "Крс", "Дцг", "Æрт", "Цпр", "Мрб", "Сбт"];
        /// `&["Янв", "Фев", "Мар", "Апр", "Май", "Июн", "Июл", "Авг", "Сен", "Окт", "Ноя", "Дек"]`
        pub const ABMON: &[&str] = &["Янв", "Фев", "Мар", "Апр", "Май", "Июн", "Июл", "Авг", "Сен", "Окт", "Ноя", "Дек"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `Some(&["Январь", "Февраль", "Мартъи", "Апрель", "Май", "Июнь", "Июль", "Август", "Сентябрь", "Октябрь", "Ноябрь", "Декабрь"])`
        pub const ALT_MON: Option<&[&str]> = Some(&["Январь", "Февраль", "Мартъи", "Апрель", "Май", "Июнь", "Июль", "Август", "Сентябрь", "Октябрь", "Ноябрь", "Декабрь"]);
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Хуыцаубон", "Къуырисæр", "Дыццæг", "Æртыццæг", "Цыппæрæм", "Майрæмбон", "Сабат"]`
        pub const DAY: &[&str] = &["Хуыцаубон", "Къуырисæр", "Дыццæг", "Æртыццæг", "Цыппæрæм", "Майрæмбон", "Сабат"];
        /// `"%d.%m.%Y"`
        pub const D_FMT: &str = "%d.%m.%Y";
        /// `"%a %d %b %Y %T"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["январы", "февралы", "мартъийы", "апрелы", "майы", "июны", "июлы", "августы", "сентябры", "октябры", "ноябры", "декабры"]`
        pub const MON: &[&str] = &["январы", "февралы", "мартъийы", "апрелы", "майы", "июны", "июлы", "августы", "сентябры", "октябры", "ноябры", "декабры"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod pa_IN {
    pub mod LC_ADDRESS {
        /// `Some("IND")`
        pub const COUNTRY_AB3: Option<&str> = Some("IND");
        /// `Some("IN")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IN");
        /// `Some("IND")`
        pub const COUNTRY_CAR: Option<&str> = Some("IND");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("ਭਾਰਤ")`
        pub const COUNTRY_NAME: Option<&str> = Some("ਭਾਰਤ");
        /// `Some(356)`
        pub const COUNTRY_NUM: Option<i64> = Some(356);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("pa")`
        pub const LANG_AB: Option<&str> = Some("pa");
        /// `Some("pan")`
        pub const LANG_LIB: Option<&str> = Some("pan");
        /// `Some("ਪ\u{a70}ਜਾਬੀ")`
        pub const LANG_NAME: Option<&str> = Some("ਪ\u{a70}ਜਾਬੀ");
        /// `Some("pan")`
        pub const LANG_TERM: Option<&str> = Some("pan");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2004-09-30")`
        pub const DATE: Option<&str> = Some("2004-09-30");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Punjabi")`
        pub const LANGUAGE: Option<&str> = Some("Punjabi");
        /// `Some("0.2")`
        pub const REVISION: Option<&str> = Some("0.2");
        /// `Some("IndLinux.org")`
        pub const SOURCE: Option<&str> = Some("IndLinux.org");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("India")`
        pub const TERRITORY: Option<&str> = Some("India");
        /// `Some("Punjabi language locale for Indian Punjabi(Gurmukhi)")`
        pub const TITLE: Option<&str> = Some("Punjabi language locale for Indian Punjabi(Gurmukhi)");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNਨ]"`
        pub const NOEXPR: &str = "^[-0nNਨ]";
        /// `Some("ਨਹੀ\u{a02}")`
        pub const NOSTR: Option<&str> = Some("ਨਹੀ\u{a02}");
        /// `"^[+1yYਹ]"`
        pub const YESEXPR: &str = "^[+1yYਹ]";
        /// `Some("ਹਾ\u{a02}")`
        pub const YESSTR: Option<&str> = Some("ਹਾ\u{a02}");
    }
    pub use super::hi_IN::LC_MONETARY;
    pub use super::hi_IN::LC_NUMERIC;
    pub use super::hi_IN::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["ਐਤ", "ਸ\u{a4b}ਮ", "ਮ\u{a70}ਗਲ", "ਬ\u{a41}\u{a71}ਧ", "ਵੀਰ", "ਸ਼\u{a41}\u{a71}ਕਰ", "ਸ਼ਨਿ\u{a71}ਚਰ"]`
        pub const ABDAY: &[&str] = &["ਐਤ", "ਸ\u{a4b}ਮ", "ਮ\u{a70}ਗਲ", "ਬ\u{a41}\u{a71}ਧ", "ਵੀਰ", "ਸ਼\u{a41}\u{a71}ਕਰ", "ਸ਼ਨਿ\u{a71}ਚਰ"];
        /// `&["ਜਨ", "ਫ\u{a3c}ਰ", "ਮਾਰਚ", "ਅਪ\u{a4d}ਰ\u{a48}", "ਮਈ", "ਜ\u{a42}ਨ", "ਜ\u{a41}ਲਾ", "ਅਗ", "ਸਤ\u{a70}", "ਅਕਤ\u{a42}", "ਨਵ\u{a70}", "ਦਸ\u{a70}"]`
        pub const ABMON: &[&str] = &["ਜਨ", "ਫ\u{a3c}ਰ", "ਮਾਰਚ", "ਅਪ\u{a4d}ਰ\u{a48}", "ਮਈ", "ਜ\u{a42}ਨ", "ਜ\u{a41}ਲਾ", "ਅਗ", "ਸਤ\u{a70}", "ਅਕਤ\u{a42}", "ਨਵ\u{a70}", "ਦਸ\u{a70}"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["ਸਵ\u{a47}ਰ\u{a47}", "ਸ਼ਾਮ"]`
        pub const AM_PM: &[&str] = &["ਸਵ\u{a47}ਰ\u{a47}", "ਸ਼ਾਮ"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["ਐਤਵਾਰ", "ਸ\u{a4b}ਮਵਾਰ", "ਮ\u{a70}ਗਲਵਾਰ", "ਬ\u{a41}\u{a71}ਧਵਾਰ", "ਵੀਰਵਾਰ", "ਸ਼\u{a41}\u{a71}ਕਰਵਾਰ", "ਸ਼ਨਿ\u{a71}ਚਰਵਾਰ"]`
        pub const DAY: &[&str] = &["ਐਤਵਾਰ", "ਸ\u{a4b}ਮਵਾਰ", "ਮ\u{a70}ਗਲਵਾਰ", "ਬ\u{a41}\u{a71}ਧਵਾਰ", "ਵੀਰਵਾਰ", "ਸ਼\u{a41}\u{a71}ਕਰਵਾਰ", "ਸ਼ਨਿ\u{a71}ਚਰਵਾਰ"];
        /// `"%-d/%-m/%y"`
        pub const D_FMT: &str = "%-d/%-m/%y";
        /// `"%A %d %b %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A %d %b %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["ਜਨਵਰੀ", "ਫ\u{a3c}ਰਵਰੀ", "ਮਾਰਚ", "ਅਪ\u{a4d}ਰ\u{a48}ਲ", "ਮਈ", "ਜ\u{a42}ਨ", "ਜ\u{a41}ਲਾਈ", "ਅਗਸਤ", "ਸਤ\u{a70}ਬਰ", "ਅਕਤ\u{a42}ਬਰ", "ਨਵ\u{a70}ਬਰ", "ਦਸ\u{a70}ਬਰ"]`
        pub const MON: &[&str] = &["ਜਨਵਰੀ", "ਫ\u{a3c}ਰਵਰੀ", "ਮਾਰਚ", "ਅਪ\u{a4d}ਰ\u{a48}ਲ", "ਮਈ", "ਜ\u{a42}ਨ", "ਜ\u{a41}ਲਾਈ", "ਅਗਸਤ", "ਸਤ\u{a70}ਬਰ", "ਅਕਤ\u{a42}ਬਰ", "ਨਵ\u{a70}ਬਰ", "ਦਸ\u{a70}ਬਰ"];
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT: &str = "%I:%M:%S %p %Z";
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p %Z";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod pa_PK {
    pub mod LC_ADDRESS {
        /// `Some("PAK")`
        pub const COUNTRY_AB3: Option<&str> = Some("PAK");
        /// `Some("PK")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("PK");
        /// `Some("PK")`
        pub const COUNTRY_CAR: Option<&str> = Some("PK");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("پکستان")`
        pub const COUNTRY_NAME: Option<&str> = Some("پکستان");
        /// `Some(586)`
        pub const COUNTRY_NUM: Option<i64> = Some(586);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("pa")`
        pub const LANG_AB: Option<&str> = Some("pa");
        /// `Some("pan")`
        pub const LANG_LIB: Option<&str> = Some("pan");
        /// `Some("پنجابی")`
        pub const LANG_NAME: Option<&str> = Some("پنجابی");
        /// `Some("pan")`
        pub const LANG_TERM: Option<&str> = Some("pan");
        /// `Some("%a%N%f%N%d%N%b%N%h %s %e %r%N%T %z%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%a%N%f%N%d%N%b%N%h %s %e %r%N%T %z%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-07-11")`
        pub const DATE: Option<&str> = Some("2000-07-11");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Punjabi")`
        pub const LANGUAGE: Option<&str> = Some("Punjabi");
        /// `Some("0.3")`
        pub const REVISION: Option<&str> = Some("0.3");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Pakistan")`
        pub const TERRITORY: Option<&str> = Some("Pakistan");
        /// `Some("Punjabi (Shahmukhi) Language Locale for Pakistan")`
        pub const TITLE: Option<&str> = Some("Punjabi (Shahmukhi) Language Locale for Pakistan");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNنo]"`
        pub const NOEXPR: &str = "^[-0nNنo]";
        /// `Some("نهيں")`
        pub const NOSTR: Option<&str> = Some("نهيں");
        /// `"^[+1yYهبf]"`
        pub const YESEXPR: &str = "^[+1yYهبf]";
        /// `Some("بلكل")`
        pub const YESSTR: Option<&str> = Some("بلكل");
    }
    pub use super::ur_PK::LC_MONETARY;
    pub use super::ur_PK::LC_NUMERIC;
    pub use super::ur_PK::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["اتوار", "پير", "منگل", "بدھ", "جمعرات", "جمعه", "هفته"]`
        pub const ABDAY: &[&str] = &["اتوار", "پير", "منگل", "بدھ", "جمعرات", "جمعه", "هفته"];
        /// `&["جنوري", "فروري", "مارچ", "اپريل", "م\u{653}ی", "جون", "جولاي", "اگست", "ستمبر", "اكتوبر", "نومبر", "دسمبر"]`
        pub const ABMON: &[&str] = &["جنوري", "فروري", "مارچ", "اپريل", "م\u{653}ی", "جون", "جولاي", "اگست", "ستمبر", "اكتوبر", "نومبر", "دسمبر"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["ص", "ش"]`
        pub const AM_PM: &[&str] = &["ص", "ش"];
        /// `Some(3)`
        pub const CAL_DIRECTION: Option<i64> = Some(3);
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["اتوار", "پير", "منگل", "بدھ", "جمعرات", "جمعه", "هفته"]`
        pub const DAY: &[&str] = &["اتوار", "پير", "منگل", "بدھ", "جمعرات", "جمعه", "هفته"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"و %H:%M:%S %Z ت %d %B %Y"`
        pub const D_T_FMT: &str = "و %H:%M:%S %Z ت %d %B %Y";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["جنوري", "فروري", "مارچ", "اپريل", "م\u{653}ی", "جون", "جولاي", "اگست", "ستمبر", "اكتوبر", "نومبر", "دسمبر"]`
        pub const MON: &[&str] = &["جنوري", "فروري", "مارچ", "اپريل", "م\u{653}ی", "جون", "جولاي", "اگست", "ستمبر", "اكتوبر", "نومبر", "دسمبر"];
        /// `"%H:%M:%S"`
        pub const T_FMT: &str = "%H:%M:%S";
        /// `"%P %I:%M:%S"`
        pub const T_FMT_AMPM: &str = "%P %I:%M:%S";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod pap_AW {
    pub mod LC_ADDRESS {
        /// `Some("ABW")`
        pub const COUNTRY_AB3: Option<&str> = Some("ABW");
        /// `Some("AW")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("AW");
        /// `Some("AUA")`
        pub const COUNTRY_CAR: Option<&str> = Some("AUA");
        /// `Some("")`
        pub const COUNTRY_ISBN: Option<&str> = Some("");
        /// `Some("Aruba")`
        pub const COUNTRY_NAME: Option<&str> = Some("Aruba");
        /// `Some(533)`
        pub const COUNTRY_NUM: Option<i64> = Some(533);
        /// `Some("AW")`
        pub const COUNTRY_POST: Option<&str> = Some("AW");
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("pap")`
        pub const LANG_LIB: Option<&str> = Some("pap");
        /// `Some("Papiamento")`
        pub const LANG_NAME: Option<&str> = Some("Papiamento");
        /// `Some("pap")`
        pub const LANG_TERM: Option<&str> = Some("pap");
        /// `Some("%d%N%f%N%d%N%b%N%s %h 5e %r%N%C%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%d%N%f%N%d%N%b%N%s %h 5e %r%N%C%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_ADDRESS"],
        ]);
        /// `Some("Pablo Saratxaga")`
        pub const CONTACT: Option<&str> = Some("Pablo Saratxaga");
        /// `Some("2013-09-27")`
        pub const DATE: Option<&str> = Some("2013-09-27");
        /// `Some("pablo@mandrakesoft.com")`
        pub const EMAIL: Option<&str> = Some("pablo@mandrakesoft.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Papiamento")`
        pub const LANGUAGE: Option<&str> = Some("Papiamento");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("information from native speaker")`
        pub const SOURCE: Option<&str> = Some("information from native speaker");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Aruba")`
        pub const TERRITORY: Option<&str> = Some("Aruba");
        /// `Some("Papiamento Language for Aruba")`
        pub const TITLE: Option<&str> = Some("Papiamento Language for Aruba");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("nò")`
        pub const NOSTR: Option<&str> = Some("nò");
        /// `"^[+1sSyY]"`
        pub const YESEXPR: &str = "^[+1sSyY]";
        /// `Some("sí")`
        pub const YESSTR: Option<&str> = Some("sí");
    }
    pub mod LC_MONETARY {
        /// `"ƒ"`
        pub const CURRENCY_SYMBOL: &str = "ƒ";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"AWG "`
        pub const INT_CURR_SYMBOL: &str = "AWG ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const MON_THOUSANDS_SEP: &str = "\u{202f}";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `2`
        pub const N_SIGN_POSN: i64 = 2;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[0, 0]`
        pub const GROUPING: &[i64] = &[0, 0];
        /// `""`
        pub const THOUSANDS_SEP: &str = "";
    }
    pub mod LC_TELEPHONE {
        /// `Some("297")`
        pub const INT_PREFIX: Option<&str> = Some("297");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["do", "lu", "ma", "we", "ra", "bi", "sa"]`
        pub const ABDAY: &[&str] = &["do", "lu", "ma", "we", "ra", "bi", "sa"];
        /// `&["Yan", "Feb", "Mar", "Apr", "Mei", "Yün", "Yül", "Oug", "Sèp", "Okt", "Nov", "Des"]`
        pub const ABMON: &[&str] = &["Yan", "Feb", "Mar", "Apr", "Mei", "Yün", "Yül", "Oug", "Sèp", "Okt", "Nov", "Des"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Djadomingo", "Djaluna", "Djamars", "Djawebs", "Djarason", "Djabierne", "Djasabra"]`
        pub const DAY: &[&str] = &["Djadomingo", "Djaluna", "Djamars", "Djawebs", "Djarason", "Djabierne", "Djasabra"];
        /// `"%d-%m-%y"`
        pub const D_FMT: &str = "%d-%m-%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Yanüari", "Febrüari", "Mart", "Aprel", "Mei", "Yüni", "Yüli", "Ougùstùs", "Sèptèmber", "Oktober", "Novèmber", "Desèmber"]`
        pub const MON: &[&str] = &["Yanüari", "Febrüari", "Mart", "Aprel", "Mei", "Yüni", "Yüli", "Ougùstùs", "Sèptèmber", "Oktober", "Novèmber", "Desèmber"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod pap_CW {
    pub mod LC_ADDRESS {
        /// `Some("CUW")`
        pub const COUNTRY_AB3: Option<&str> = Some("CUW");
        /// `Some("CW")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("CW");
        /// `Some("")`
        pub const COUNTRY_CAR: Option<&str> = Some("");
        /// `Some("978-99904")`
        pub const COUNTRY_ISBN: Option<&str> = Some("978-99904");
        /// `Some("Kòrsou")`
        pub const COUNTRY_NAME: Option<&str> = Some("Kòrsou");
        /// `Some(531)`
        pub const COUNTRY_NUM: Option<i64> = Some(531);
        /// `Some("CW")`
        pub const COUNTRY_POST: Option<&str> = Some("CW");
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("pap")`
        pub const LANG_LIB: Option<&str> = Some("pap");
        /// `Some("Papiamento")`
        pub const LANG_NAME: Option<&str> = Some("Papiamento");
        /// `Some("pap")`
        pub const LANG_TERM: Option<&str> = Some("pap");
        /// `Some("%d%N%f%N%d%N%b%N%s %h 5e %r%N%C%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%d%N%f%N%d%N%b%N%s %h 5e %r%N%C%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_ADDRESS"],
        ]);
        /// `Some("Pablo Saratxaga")`
        pub const CONTACT: Option<&str> = Some("Pablo Saratxaga");
        /// `Some("2013-09-27")`
        pub const DATE: Option<&str> = Some("2013-09-27");
        /// `Some("pablo@mandrakesoft.com")`
        pub const EMAIL: Option<&str> = Some("pablo@mandrakesoft.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Papiamento")`
        pub const LANGUAGE: Option<&str> = Some("Papiamento");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("information from native speaker")`
        pub const SOURCE: Option<&str> = Some("information from native speaker");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Curaçao")`
        pub const TERRITORY: Option<&str> = Some("Curaçao");
        /// `Some("Papiamento language for Curaçao")`
        pub const TITLE: Option<&str> = Some("Papiamento language for Curaçao");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("nò")`
        pub const NOSTR: Option<&str> = Some("nò");
        /// `"^[+1sSyY]"`
        pub const YESEXPR: &str = "^[+1sSyY]";
        /// `Some("sí")`
        pub const YESSTR: Option<&str> = Some("sí");
    }
    pub mod LC_MONETARY {
        /// `"ƒ"`
        pub const CURRENCY_SYMBOL: &str = "ƒ";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"ANG "`
        pub const INT_CURR_SYMBOL: &str = "ANG ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const MON_THOUSANDS_SEP: &str = "\u{202f}";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `2`
        pub const N_SIGN_POSN: i64 = 2;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[0, 0]`
        pub const GROUPING: &[i64] = &[0, 0];
        /// `""`
        pub const THOUSANDS_SEP: &str = "";
    }
    pub mod LC_TELEPHONE {
        /// `Some("599")`
        pub const INT_PREFIX: Option<&str> = Some("599");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["do", "lu", "ma", "we", "ra", "bi", "sa"]`
        pub const ABDAY: &[&str] = &["do", "lu", "ma", "we", "ra", "bi", "sa"];
        /// `&["Yan", "Feb", "Mar", "Apr", "Mei", "Yün", "Yül", "Oug", "Sèp", "Okt", "Nov", "Des"]`
        pub const ABMON: &[&str] = &["Yan", "Feb", "Mar", "Apr", "Mei", "Yün", "Yül", "Oug", "Sèp", "Okt", "Nov", "Des"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Djadomingo", "Djaluna", "Djamars", "Djawebs", "Djarason", "Djabierne", "Djasabra"]`
        pub const DAY: &[&str] = &["Djadomingo", "Djaluna", "Djamars", "Djawebs", "Djarason", "Djabierne", "Djasabra"];
        /// `"%d-%m-%y"`
        pub const D_FMT: &str = "%d-%m-%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Yanüari", "Febrüari", "Mart", "Aprel", "Mei", "Yüni", "Yüli", "Ougùstùs", "Sèptèmber", "Oktober", "Novèmber", "Desèmber"]`
        pub const MON: &[&str] = &["Yanüari", "Febrüari", "Mart", "Aprel", "Mei", "Yüni", "Yüli", "Ougùstùs", "Sèptèmber", "Oktober", "Novèmber", "Desèmber"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod pl_PL {
    pub mod LC_ADDRESS {
        /// `Some("POL")`
        pub const COUNTRY_AB3: Option<&str> = Some("POL");
        /// `Some("PL")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("PL");
        /// `Some("PL")`
        pub const COUNTRY_CAR: Option<&str> = Some("PL");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Polska")`
        pub const COUNTRY_NAME: Option<&str> = Some("Polska");
        /// `Some(616)`
        pub const COUNTRY_NUM: Option<i64> = Some(616);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("pl")`
        pub const LANG_AB: Option<&str> = Some("pl");
        /// `Some("pol")`
        pub const LANG_LIB: Option<&str> = Some("pol");
        /// `Some("polski")`
        pub const LANG_NAME: Option<&str> = Some("polski");
        /// `Some("pol")`
        pub const LANG_TERM: Option<&str> = Some("pol");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jo/rgens Alle 8, DK-1615 Ko/benhavn V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jo/rgens Alle 8, DK-1615 Ko/benhavn V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Polish")`
        pub const LANGUAGE: Option<&str> = Some("Polish");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Poland")`
        pub const TERRITORY: Option<&str> = Some("Poland");
        /// `Some("Polish locale for Poland")`
        pub const TITLE: Option<&str> = Some("Polish locale for Poland");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("nie")`
        pub const NOSTR: Option<&str> = Some("nie");
        /// `"^[+1TtYy]"`
        pub const YESEXPR: &str = "^[+1TtYy]";
        /// `Some("tak")`
        pub const YESSTR: Option<&str> = Some("tak");
    }
    pub mod LC_MONETARY {
        /// `"zł"`
        pub const CURRENCY_SYMBOL: &str = "zł";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"PLN "`
        pub const INT_CURR_SYMBOL: &str = "PLN ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `"\u{202f}"`
        pub const MON_THOUSANDS_SEP: &str = "\u{202f}";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `"\u{202f}"`
        pub const THOUSANDS_SEP: &str = "\u{202f}";
    }
    pub mod LC_TELEPHONE {
        /// `Some("48")`
        pub const INT_PREFIX: Option<&str> = Some("48");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["nie", "pon", "wto", "śro", "czw", "pią", "sob"]`
        pub const ABDAY: &[&str] = &["nie", "pon", "wto", "śro", "czw", "pią", "sob"];
        /// `&["sty", "lut", "mar", "kwi", "maj", "cze", "lip", "sie", "wrz", "paź", "lis", "gru"]`
        pub const ABMON: &[&str] = &["sty", "lut", "mar", "kwi", "maj", "cze", "lip", "sie", "wrz", "paź", "lis", "gru"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `Some(&["styczeń", "luty", "marzec", "kwiecień", "maj", "czerwiec", "lipiec", "sierpień", "wrzesień", "październik", "listopad", "grudzień"])`
        pub const ALT_MON: Option<&[&str]> = Some(&["styczeń", "luty", "marzec", "kwiecień", "maj", "czerwiec", "lipiec", "sierpień", "wrzesień", "październik", "listopad", "grudzień"]);
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a, %-d %b %Y, %T %Z")`
        pub const DATE_FMT: Option<&str> = Some("%a, %-d %b %Y, %T %Z");
        /// `&["niedziela", "poniedziałek", "wtorek", "środa", "czwartek", "piątek", "sobota"]`
        pub const DAY: &[&str] = &["niedziela", "poniedziałek", "wtorek", "środa", "czwartek", "piątek", "sobota"];
        /// `"%d.%m.%Y"`
        pub const D_FMT: &str = "%d.%m.%Y";
        /// `"%a, %-d %b %Y, %T"`
        pub const D_T_FMT: &str = "%a, %-d %b %Y, %T";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["stycznia", "lutego", "marca", "kwietnia", "maja", "czerwca", "lipca", "sierpnia", "września", "października", "listopada", "grudnia"]`
        pub const MON: &[&str] = &["stycznia", "lutego", "marca", "kwietnia", "maja", "czerwca", "lipca", "sierpnia", "września", "października", "listopada", "grudnia"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ps_AF {
    pub mod LC_ADDRESS {
        /// `Some("AFG")`
        pub const COUNTRY_AB3: Option<&str> = Some("AFG");
        /// `Some("AF")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("AF");
        /// `Some("AFG")`
        pub const COUNTRY_CAR: Option<&str> = Some("AFG");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("افغانستان")`
        pub const COUNTRY_NAME: Option<&str> = Some("افغانستان");
        /// `Some(4)`
        pub const COUNTRY_NUM: Option<i64> = Some(4);
        /// `Some("AF")`
        pub const COUNTRY_POST: Option<&str> = Some("AF");
        /// `Some("ps")`
        pub const LANG_AB: Option<&str> = Some("ps");
        /// `Some("pus")`
        pub const LANG_LIB: Option<&str> = Some("pus");
        /// `Some("پښتو")`
        pub const LANG_NAME: Option<&str> = Some("پښتو");
        /// `Some("pus")`
        pub const LANG_TERM: Option<&str> = Some("pus");
        /// `Some("%a%N%f%N%d%N%b%N%s %h %e %r%N%C-%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%a%N%f%N%d%N%b%N%s %h %e %r%N%C-%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("see e-mail.")`
        pub const ADDRESS: Option<&str> = Some("see e-mail.");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
        ]);
        /// `Some("Nasir Gulzade")`
        pub const CONTACT: Option<&str> = Some("Nasir Gulzade");
        /// `Some("2009-01-16")`
        pub const DATE: Option<&str> = Some("2009-01-16");
        /// `Some("nasirgulzade@hotmail.com")`
        pub const EMAIL: Option<&str> = Some("nasirgulzade@hotmail.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Pashto")`
        pub const LANGUAGE: Option<&str> = Some("Pashto");
        /// `Some("0.2")`
        pub const REVISION: Option<&str> = Some("0.2");
        /// `Some("Nasir Gulzade")`
        pub const SOURCE: Option<&str> = Some("Nasir Gulzade");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Afghanistan")`
        pub const TERRITORY: Option<&str> = Some("Afghanistan");
        /// `Some("Pashto locale for Afghanistan")`
        pub const TITLE: Option<&str> = Some("Pashto locale for Afghanistan");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNخنo]"`
        pub const NOEXPR: &str = "^[-0nNخنo]";
        /// `Some("نه")`
        pub const NOSTR: Option<&str> = Some("نه");
        /// `"^[+1yYبf]"`
        pub const YESEXPR: &str = "^[+1yYبf]";
        /// `Some("هو")`
        pub const YESSTR: Option<&str> = Some("هو");
    }
    pub mod LC_MONETARY {
        /// `"؋"`
        pub const CURRENCY_SYMBOL: &str = "؋";
        /// `0`
        pub const FRAC_DIGITS: i64 = 0;
        /// `"AFN "`
        pub const INT_CURR_SYMBOL: &str = "AFN ";
        /// `0`
        pub const INT_FRAC_DIGITS: i64 = 0;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"٫"`
        pub const MON_DECIMAL_POINT: &str = "٫";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `"٬"`
        pub const MON_THOUSANDS_SEP: &str = "٬";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"٫"`
        pub const DECIMAL_POINT: &str = "٫";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `"٬"`
        pub const THOUSANDS_SEP: &str = "٬";
    }
    pub mod LC_TELEPHONE {
        /// `Some("93")`
        pub const INT_PREFIX: Option<&str> = Some("93");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `Some("%A‒%l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("%A‒%l");
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["ی.", "د.", "س.", "چ.", "پ.", "ج.", "ش."]`
        pub const ABDAY: &[&str] = &["ی.", "د.", "س.", "چ.", "پ.", "ج.", "ش."];
        /// `&["جنوري", "فبروري", "مارچ", "اپریل", "مۍ", "جون", "جولای", "اګست", "سپتمبر", "اکتوبر", "نومبر", "دسمبر"]`
        pub const ABMON: &[&str] = &["جنوري", "فبروري", "مارچ", "اپریل", "مۍ", "جون", "جولای", "اګست", "سپتمبر", "اکتوبر", "نومبر", "دسمبر"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["غ.م.", "غ.و."]`
        pub const AM_PM: &[&str] = &["غ.م.", "غ.و."];
        /// `Some(3)`
        pub const CAL_DIRECTION: Option<i64> = Some(3);
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["یکشنبه", "دوشنبه", "سه\u{200c}شنبه", "چارشنبه", "پنجشنبه", "جمعه", "شنبه"]`
        pub const DAY: &[&str] = &["یکشنبه", "دوشنبه", "سه\u{200c}شنبه", "چارشنبه", "پنجشنبه", "جمعه", "شنبه"];
        /// `"د %Y د %B %e"`
        pub const D_FMT: &str = "د %Y د %B %e";
        /// `"%A د %Y د %B %e، %Z %H:%M:%S"`
        pub const D_T_FMT: &str = "%A د %Y د %B %e، %Z %H:%M:%S";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(7)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(7);
        /// `Some(7)`
        pub const FIRST_WORKDAY: Option<i64> = Some(7);
        /// `&["جنوري", "فبروري", "مارچ", "اپریل", "مۍ", "جون", "جولای", "اګست", "سپتمبر", "اکتوبر", "نومبر", "دسمبر"]`
        pub const MON: &[&str] = &["جنوري", "فبروري", "مارچ", "اپریل", "مۍ", "جون", "جولای", "اګست", "سپتمبر", "اکتوبر", "نومبر", "دسمبر"];
        /// `"%H:%M:%S"`
        pub const T_FMT: &str = "%H:%M:%S";
        /// `"\u{202b}%I:%M:%S %p\u{202c}"`
        pub const T_FMT_AMPM: &str = "\u{202b}%I:%M:%S %p\u{202c}";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod pt_BR {
    pub mod LC_ADDRESS {
        /// `Some("BRA")`
        pub const COUNTRY_AB3: Option<&str> = Some("BRA");
        /// `Some("BR")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("BR");
        /// `Some("BR")`
        pub const COUNTRY_CAR: Option<&str> = Some("BR");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Brasil")`
        pub const COUNTRY_NAME: Option<&str> = Some("Brasil");
        /// `Some(76)`
        pub const COUNTRY_NUM: Option<i64> = Some(76);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("pt")`
        pub const LANG_AB: Option<&str> = Some("pt");
        /// `Some("por")`
        pub const LANG_LIB: Option<&str> = Some("por");
        /// `Some("português")`
        pub const LANG_NAME: Option<&str> = Some("português");
        /// `Some("por")`
        pub const LANG_TERM: Option<&str> = Some("por");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Brazilian Portuguese")`
        pub const LANGUAGE: Option<&str> = Some("Brazilian Portuguese");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Brazil")`
        pub const TERRITORY: Option<&str> = Some("Brazil");
        /// `Some("Portuguese locale for Brasil")`
        pub const TITLE: Option<&str> = Some("Portuguese locale for Brasil");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("não")`
        pub const NOSTR: Option<&str> = Some("não");
        /// `"^[+1SsyY]"`
        pub const YESEXPR: &str = "^[+1SsyY]";
        /// `Some("sim")`
        pub const YESSTR: Option<&str> = Some("sim");
    }
    pub mod LC_MONETARY {
        /// `"R$"`
        pub const CURRENCY_SYMBOL: &str = "R$";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"BRL "`
        pub const INT_CURR_SYMBOL: &str = "BRL ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const MON_THOUSANDS_SEP: &str = ".";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const THOUSANDS_SEP: &str = ".";
    }
    pub mod LC_TELEPHONE {
        /// `Some("55")`
        pub const INT_PREFIX: Option<&str> = Some("55");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["dom", "seg", "ter", "qua", "qui", "sex", "sáb"]`
        pub const ABDAY: &[&str] = &["dom", "seg", "ter", "qua", "qui", "sex", "sáb"];
        /// `&["jan", "fev", "mar", "abr", "mai", "jun", "jul", "ago", "set", "out", "nov", "dez"]`
        pub const ABMON: &[&str] = &["jan", "fev", "mar", "abr", "mai", "jun", "jul", "ago", "set", "out", "nov", "dez"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["domingo", "segunda", "terça", "quarta", "quinta", "sexta", "sábado"]`
        pub const DAY: &[&str] = &["domingo", "segunda", "terça", "quarta", "quinta", "sexta", "sábado"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["janeiro", "fevereiro", "março", "abril", "maio", "junho", "julho", "agosto", "setembro", "outubro", "novembro", "dezembro"]`
        pub const MON: &[&str] = &["janeiro", "fevereiro", "março", "abril", "maio", "junho", "julho", "agosto", "setembro", "outubro", "novembro", "dezembro"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod pt_PT {
    pub mod LC_ADDRESS {
        /// `Some("PRT")`
        pub const COUNTRY_AB3: Option<&str> = Some("PRT");
        /// `Some("PT")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("PT");
        /// `Some("P")`
        pub const COUNTRY_CAR: Option<&str> = Some("P");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Portugal")`
        pub const COUNTRY_NAME: Option<&str> = Some("Portugal");
        /// `Some(620)`
        pub const COUNTRY_NUM: Option<i64> = Some(620);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("pt")`
        pub const LANG_AB: Option<&str> = Some("pt");
        /// `Some("por")`
        pub const LANG_LIB: Option<&str> = Some("por");
        /// `Some("português")`
        pub const LANG_NAME: Option<&str> = Some("português");
        /// `Some("por")`
        pub const LANG_TERM: Option<&str> = Some("por");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("European Portuguese")`
        pub const LANGUAGE: Option<&str> = Some("European Portuguese");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Portugal")`
        pub const TERRITORY: Option<&str> = Some("Portugal");
        /// `Some("Portuguese locale for Portugal")`
        pub const TITLE: Option<&str> = Some("Portuguese locale for Portugal");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("não")`
        pub const NOSTR: Option<&str> = Some("não");
        /// `"^[+1SsyY]"`
        pub const YESEXPR: &str = "^[+1SsyY]";
        /// `Some("sim")`
        pub const YESSTR: Option<&str> = Some("sim");
    }
    pub mod LC_MONETARY {
        /// `"€"`
        pub const CURRENCY_SYMBOL: &str = "€";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"EUR "`
        pub const INT_CURR_SYMBOL: &str = "EUR ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const MON_THOUSANDS_SEP: &str = ".";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[0, 0]`
        pub const GROUPING: &[i64] = &[0, 0];
        /// `""`
        pub const THOUSANDS_SEP: &str = "";
    }
    pub mod LC_TELEPHONE {
        /// `Some("351")`
        pub const INT_PREFIX: Option<&str> = Some("351");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `Some("%l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("%l");
        /// `Some("+%c %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %l");
    }
    pub mod LC_TIME {
        /// `&["dom", "seg", "ter", "qua", "qui", "sex", "sáb"]`
        pub const ABDAY: &[&str] = &["dom", "seg", "ter", "qua", "qui", "sex", "sáb"];
        /// `&["jan", "fev", "mar", "abr", "mai", "jun", "jul", "ago", "set", "out", "nov", "dez"]`
        pub const ABMON: &[&str] = &["jan", "fev", "mar", "abr", "mai", "jun", "jul", "ago", "set", "out", "nov", "dez"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["domingo", "segunda", "terça", "quarta", "quinta", "sexta", "sábado"]`
        pub const DAY: &[&str] = &["domingo", "segunda", "terça", "quarta", "quinta", "sexta", "sábado"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["janeiro", "fevereiro", "março", "abril", "maio", "junho", "julho", "agosto", "setembro", "outubro", "novembro", "dezembro"]`
        pub const MON: &[&str] = &["janeiro", "fevereiro", "março", "abril", "maio", "junho", "julho", "agosto", "setembro", "outubro", "novembro", "dezembro"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod pt_PT_euro {
    pub use super::pt_PT::LC_ADDRESS;
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("https://www.gnu.org/software/libc/")`
        pub const ADDRESS: Option<&str> = Some("https://www.gnu.org/software/libc/");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-08-20")`
        pub const DATE: Option<&str> = Some("2000-08-20");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Portuguese")`
        pub const LANGUAGE: Option<&str> = Some("Portuguese");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Free Software Foundation, Inc.")`
        pub const SOURCE: Option<&str> = Some("Free Software Foundation, Inc.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Portugal")`
        pub const TERRITORY: Option<&str> = Some("Portugal");
        /// `Some("Portuguese locale for Portugal with Euro")`
        pub const TITLE: Option<&str> = Some("Portuguese locale for Portugal with Euro");
    }
    pub use super::pt_PT::LC_MESSAGES;
    pub use super::pt_PT::LC_MONETARY;
    pub use super::pt_PT::LC_NUMERIC;
    pub use super::pt_PT::LC_TELEPHONE;
    pub use super::pt_PT::LC_TIME;
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod quz_PE {
    pub mod LC_ADDRESS {
        /// `Some("PER")`
        pub const COUNTRY_AB3: Option<&str> = Some("PER");
        /// `Some("PE")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("PE");
        /// `Some("PE")`
        pub const COUNTRY_CAR: Option<&str> = Some("PE");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `None`
        pub const COUNTRY_NAME: Option<&str> = None;
        /// `Some(604)`
        pub const COUNTRY_NUM: Option<i64> = Some(604);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("quz")`
        pub const LANG_LIB: Option<&str> = Some("quz");
        /// `Some("Qusqu runasimi")`
        pub const LANG_NAME: Option<&str> = Some("Qusqu runasimi");
        /// `Some("quz")`
        pub const LANG_TERM: Option<&str> = Some("quz");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
        ]);
        /// `Some("sugarlabs.org")`
        pub const CONTACT: Option<&str> = Some("sugarlabs.org");
        /// `Some("2013-08-24")`
        pub const DATE: Option<&str> = Some("2013-08-24");
        /// `Some("libc-alpha@sourceware.org")`
        pub const EMAIL: Option<&str> = Some("libc-alpha@sourceware.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Cusco Quechua")`
        pub const LANGUAGE: Option<&str> = Some("Cusco Quechua");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Sugar Labs / OLPC")`
        pub const SOURCE: Option<&str> = Some("Sugar Labs / OLPC");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Peru")`
        pub const TERRITORY: Option<&str> = Some("Peru");
        /// `Some("Cusco Quechua locale for Peru")`
        pub const TITLE: Option<&str> = Some("Cusco Quechua locale for Peru");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0mMnN]"`
        pub const NOEXPR: &str = "^[-0mMnN]";
        /// `Some("Manan")`
        pub const NOSTR: Option<&str> = Some("Manan");
        /// `"^[+1aAsSyY]"`
        pub const YESEXPR: &str = "^[+1aAsSyY]";
        /// `Some("Arí")`
        pub const YESSTR: Option<&str> = Some("Arí");
    }
    pub use super::es_PE::LC_MONETARY;
    pub use super::es_PE::LC_NUMERIC;
    pub use super::es_PE::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["tum", "lun", "mar", "miy", "juy", "wiy", "saw"]`
        pub const ABDAY: &[&str] = &["tum", "lun", "mar", "miy", "juy", "wiy", "saw"];
        /// `&["ini", "phi", "mar", "awr", "may", "hun", "hul", "agu", "sip", "ukt", "nuw", "tis"]`
        pub const ABMON: &[&str] = &["ini", "phi", "mar", "awr", "may", "hun", "hul", "agu", "sip", "ukt", "nuw", "tis"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["AM", "PM"]`
        pub const AM_PM: &[&str] = &["AM", "PM"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["tuminku", "lunis", "martis", "miyirkulis", "juywis", "wiyirnis", "sawatu"]`
        pub const DAY: &[&str] = &["tuminku", "lunis", "martis", "miyirkulis", "juywis", "wiyirnis", "sawatu"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["iniru", "phiwriru", "marsu", "awril", "mayu", "huniyu", "huliyu", "agustu", "siptiyimri", "uktuwri", "nuwiyimri", "tisiyimri"]`
        pub const MON: &[&str] = &["iniru", "phiwriru", "marsu", "awril", "mayu", "huniyu", "huliyu", "agustu", "siptiyimri", "uktuwri", "nuwiyimri", "tisiyimri"];
        /// `"%I:%M:%S %p"`
        pub const T_FMT: &str = "%I:%M:%S %p";
        /// `"%I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod raj_IN {
    pub mod LC_ADDRESS {
        /// `Some("IND")`
        pub const COUNTRY_AB3: Option<&str> = Some("IND");
        /// `Some("IN")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IN");
        /// `Some("IND")`
        pub const COUNTRY_CAR: Option<&str> = Some("IND");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("भारत")`
        pub const COUNTRY_NAME: Option<&str> = Some("भारत");
        /// `Some(356)`
        pub const COUNTRY_NUM: Option<i64> = Some(356);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("raj")`
        pub const LANG_LIB: Option<&str> = Some("raj");
        /// `Some("राजस\u{94d}थानी")`
        pub const LANG_NAME: Option<&str> = Some("राजस\u{94d}थानी");
        /// `Some("raj")`
        pub const LANG_TERM: Option<&str> = Some("raj");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2014-06-13")`
        pub const DATE: Option<&str> = Some("2014-06-13");
        /// `Some("meghrajsuthar03@gmail.com")`
        pub const EMAIL: Option<&str> = Some("meghrajsuthar03@gmail.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Rajasthani")`
        pub const LANGUAGE: Option<&str> = Some("Rajasthani");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("India")`
        pub const TERRITORY: Option<&str> = Some("India");
        /// `Some("Rajasthani language locale for India")`
        pub const TITLE: Option<&str> = Some("Rajasthani language locale for India");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `None`
        pub const NOSTR: Option<&str> = None;
        /// `"^[+1yY]"`
        pub const YESEXPR: &str = "^[+1yY]";
        /// `None`
        pub const YESSTR: Option<&str> = None;
    }
    pub use super::hi_IN::LC_MONETARY;
    pub use super::hi_IN::LC_NUMERIC;
    pub use super::hi_IN::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["रवि", "सोम", "म\u{902}गल", "ब\u{941}ध", "ब\u{943}हस\u{94d}पति", "श\u{941}क\u{94d}र", "शनि"]`
        pub const ABDAY: &[&str] = &["रवि", "सोम", "म\u{902}गल", "ब\u{941}ध", "ब\u{943}हस\u{94d}पति", "श\u{941}क\u{94d}र", "शनि"];
        /// `&["जन", "फर", "मार\u{94d}च", "अप\u{94d}र\u{948}", "मई", "ज\u{942}न", "ज\u{941}ल", "अग", "सित", "अक\u{94d}ट\u{942}", "नव", "दिस"]`
        pub const ABMON: &[&str] = &["जन", "फर", "मार\u{94d}च", "अप\u{94d}र\u{948}", "मई", "ज\u{942}न", "ज\u{941}ल", "अग", "सित", "अक\u{94d}ट\u{942}", "नव", "दिस"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["प\u{942}र\u{94d}वाह\u{94d}न", "अपराह\u{94d}न"]`
        pub const AM_PM: &[&str] = &["प\u{942}र\u{94d}वाह\u{94d}न", "अपराह\u{94d}न"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["रविवार", "सोमवार", "म\u{902}गल\u{94d}लवार", "ब\u{941}धवार", "ब\u{943}हस\u{94d}पतिवार", "श\u{941}क\u{94d}रवार", "शनिवार"]`
        pub const DAY: &[&str] = &["रविवार", "सोमवार", "म\u{902}गल\u{94d}लवार", "ब\u{941}धवार", "ब\u{943}हस\u{94d}पतिवार", "श\u{941}क\u{94d}रवार", "शनिवार"];
        /// `"%-d/%-m/%y"`
        pub const D_FMT: &str = "%-d/%-m/%y";
        /// `"%A %d %b %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A %d %b %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["जनवरी", "फरवरी", "मार\u{94d}च", "अप\u{94d}र\u{948}ल", "मई", "ज\u{942}न", "ज\u{941}लाई", "अगस\u{94d}त", "सित\u{902}बर", "अक\u{94d}ट\u{942}बर", "नव\u{902}बर", "दिस\u{902}बर"]`
        pub const MON: &[&str] = &["जनवरी", "फरवरी", "मार\u{94d}च", "अप\u{94d}र\u{948}ल", "मई", "ज\u{942}न", "ज\u{941}लाई", "अगस\u{94d}त", "सित\u{902}बर", "अक\u{94d}ट\u{942}बर", "नव\u{902}बर", "दिस\u{902}बर"];
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT: &str = "%I:%M:%S %p %Z";
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p %Z";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ro_RO {
    pub mod LC_ADDRESS {
        /// `Some("ROU")`
        pub const COUNTRY_AB3: Option<&str> = Some("ROU");
        /// `Some("RO")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("RO");
        /// `Some("RO")`
        pub const COUNTRY_CAR: Option<&str> = Some("RO");
        /// `Some("973")`
        pub const COUNTRY_ISBN: Option<&str> = Some("973");
        /// `Some("România")`
        pub const COUNTRY_NAME: Option<&str> = Some("România");
        /// `Some(642)`
        pub const COUNTRY_NUM: Option<i64> = Some(642);
        /// `Some("RO")`
        pub const COUNTRY_POST: Option<&str> = Some("RO");
        /// `Some("ro")`
        pub const LANG_AB: Option<&str> = Some("ro");
        /// `Some("rum")`
        pub const LANG_LIB: Option<&str> = Some("rum");
        /// `Some("română")`
        pub const LANG_NAME: Option<&str> = Some("română");
        /// `Some("ron")`
        pub const LANG_TERM: Option<&str> = Some("ron");
        /// `Some("%f%N%a%d%N%s%t%h%N%b%t%e%t%r%N%z%t%T%N%S%t%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%d%N%s%t%h%N%b%t%e%t%r%N%z%t%T%N%S%t%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jorgens Alle 8, DK-1615 Kobenhavn V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jorgens Alle 8, DK-1615 Kobenhavn V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Romanian")`
        pub const LANGUAGE: Option<&str> = Some("Romanian");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Romania")`
        pub const TERRITORY: Option<&str> = Some("Romania");
        /// `Some("Romanian locale for Romania")`
        pub const TITLE: Option<&str> = Some("Romanian locale for Romania");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("nu")`
        pub const NOSTR: Option<&str> = Some("nu");
        /// `"^[+1DdYy]"`
        pub const YESEXPR: &str = "^[+1DdYy]";
        /// `Some("da")`
        pub const YESSTR: Option<&str> = Some("da");
    }
    pub mod LC_MONETARY {
        /// `"Lei"`
        pub const CURRENCY_SYMBOL: &str = "Lei";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"RON "`
        pub const INT_CURR_SYMBOL: &str = "RON ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const MON_THOUSANDS_SEP: &str = ".";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const THOUSANDS_SEP: &str = ".";
    }
    pub mod LC_TELEPHONE {
        /// `Some("40")`
        pub const INT_PREFIX: Option<&str> = Some("40");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["Du", "Lu", "Ma", "Mi", "Jo", "Vi", "Sb"]`
        pub const ABDAY: &[&str] = &["Du", "Lu", "Ma", "Mi", "Jo", "Vi", "Sb"];
        /// `&["ian", "feb", "mar", "apr", "mai", "iun", "iul", "aug", "sep", "oct", "nov", "dec"]`
        pub const ABMON: &[&str] = &["ian", "feb", "mar", "apr", "mai", "iun", "iul", "aug", "sep", "oct", "nov", "dec"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%A %-e %B %Y, %H:%M:%S %z")`
        pub const DATE_FMT: Option<&str> = Some("%A %-e %B %Y, %H:%M:%S %z");
        /// `&["duminică", "luni", "marți", "miercuri", "joi", "vineri", "sâmbătă"]`
        pub const DAY: &[&str] = &["duminică", "luni", "marți", "miercuri", "joi", "vineri", "sâmbătă"];
        /// `"%d.%m.%Y"`
        pub const D_FMT: &str = "%d.%m.%Y";
        /// `"%a %d %b %Y %T %z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["ianuarie", "februarie", "martie", "aprilie", "mai", "iunie", "iulie", "august", "septembrie", "octombrie", "noiembrie", "decembrie"]`
        pub const MON: &[&str] = &["ianuarie", "februarie", "martie", "aprilie", "mai", "iunie", "iulie", "august", "septembrie", "octombrie", "noiembrie", "decembrie"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ru_RU {
    pub mod LC_ADDRESS {
        /// `Some("RUS")`
        pub const COUNTRY_AB3: Option<&str> = Some("RUS");
        /// `Some("RU")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("RU");
        /// `Some("RUS")`
        pub const COUNTRY_CAR: Option<&str> = Some("RUS");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Россия")`
        pub const COUNTRY_NAME: Option<&str> = Some("Россия");
        /// `Some(643)`
        pub const COUNTRY_NUM: Option<i64> = Some(643);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ru")`
        pub const LANG_AB: Option<&str> = Some("ru");
        /// `Some("rus")`
        pub const LANG_LIB: Option<&str> = Some("rus");
        /// `Some("русский")`
        pub const LANG_NAME: Option<&str> = Some("русский");
        /// `Some("rus")`
        pub const LANG_TERM: Option<&str> = Some("rus");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jorgens Alle 8, DK-1615 Kobenhavn V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jorgens Alle 8, DK-1615 Kobenhavn V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Russian")`
        pub const LANGUAGE: Option<&str> = Some("Russian");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Russia")`
        pub const TERRITORY: Option<&str> = Some("Russia");
        /// `Some("Russian locale for Russia")`
        pub const TITLE: Option<&str> = Some("Russian locale for Russia");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNНн]"`
        pub const NOEXPR: &str = "^[-0nNНн]";
        /// `Some("нет")`
        pub const NOSTR: Option<&str> = Some("нет");
        /// `"^[+1yYДд]"`
        pub const YESEXPR: &str = "^[+1yYДд]";
        /// `Some("да")`
        pub const YESSTR: Option<&str> = Some("да");
    }
    pub mod LC_MONETARY {
        /// `"₽"`
        pub const CURRENCY_SYMBOL: &str = "₽";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"RUB "`
        pub const INT_CURR_SYMBOL: &str = "RUB ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const MON_THOUSANDS_SEP: &str = "\u{202f}";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const THOUSANDS_SEP: &str = "\u{202f}";
    }
    pub mod LC_TELEPHONE {
        /// `Some("7")`
        pub const INT_PREFIX: Option<&str> = Some("7");
        /// `Some("0~10")`
        pub const INT_SELECT: Option<&str> = Some("0~10");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["Вс", "Пн", "Вт", "Ср", "Чт", "Пт", "Сб"]`
        pub const ABDAY: &[&str] = &["Вс", "Пн", "Вт", "Ср", "Чт", "Пт", "Сб"];
        /// `&["янв", "фев", "мар", "апр", "мая", "июн", "июл", "авг", "сен", "окт", "ноя", "дек"]`
        pub const ABMON: &[&str] = &["янв", "фев", "мар", "апр", "мая", "июн", "июл", "авг", "сен", "окт", "ноя", "дек"];
        /// `Some(&["янв", "фев", "мар", "апр", "май", "июн", "июл", "авг", "сен", "окт", "ноя", "дек"])`
        pub const AB_ALT_MON: Option<&[&str]> = Some(&["янв", "фев", "мар", "апр", "май", "июн", "июл", "авг", "сен", "окт", "ноя", "дек"]);
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `Some(&["Январь", "Февраль", "Март", "Апрель", "Май", "Июнь", "Июль", "Август", "Сентябрь", "Октябрь", "Ноябрь", "Декабрь"])`
        pub const ALT_MON: Option<&[&str]> = Some(&["Январь", "Февраль", "Март", "Апрель", "Май", "Июнь", "Июль", "Август", "Сентябрь", "Октябрь", "Ноябрь", "Декабрь"]);
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Воскресенье", "Понедельник", "Вторник", "Среда", "Четверг", "Пятница", "Суббота"]`
        pub const DAY: &[&str] = &["Воскресенье", "Понедельник", "Вторник", "Среда", "Четверг", "Пятница", "Суббота"];
        /// `"%d.%m.%Y"`
        pub const D_FMT: &str = "%d.%m.%Y";
        /// `"%a %d %b %Y %T"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["января", "февраля", "марта", "апреля", "мая", "июня", "июля", "августа", "сентября", "октября", "ноября", "декабря"]`
        pub const MON: &[&str] = &["января", "февраля", "марта", "апреля", "мая", "июня", "июля", "августа", "сентября", "октября", "ноября", "декабря"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ru_UA {
    pub mod LC_ADDRESS {
        /// `Some("UKR")`
        pub const COUNTRY_AB3: Option<&str> = Some("UKR");
        /// `Some("UA")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("UA");
        /// `Some("UA")`
        pub const COUNTRY_CAR: Option<&str> = Some("UA");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Украина")`
        pub const COUNTRY_NAME: Option<&str> = Some("Украина");
        /// `Some(804)`
        pub const COUNTRY_NUM: Option<i64> = Some(804);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ru")`
        pub const LANG_AB: Option<&str> = Some("ru");
        /// `Some("rus")`
        pub const LANG_LIB: Option<&str> = Some("rus");
        /// `Some("русский")`
        pub const LANG_NAME: Option<&str> = Some("русский");
        /// `Some("rus")`
        pub const LANG_TERM: Option<&str> = Some("rus");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Russian")`
        pub const LANGUAGE: Option<&str> = Some("Russian");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RFC 2319")`
        pub const SOURCE: Option<&str> = Some("RFC 2319");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Ukraine")`
        pub const TERRITORY: Option<&str> = Some("Ukraine");
        /// `Some("Russian locale for Ukraine")`
        pub const TITLE: Option<&str> = Some("Russian locale for Ukraine");
    }
    pub use super::ru_RU::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"₴"`
        pub const CURRENCY_SYMBOL: &str = "₴";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"UAH "`
        pub const INT_CURR_SYMBOL: &str = "UAH ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const MON_THOUSANDS_SEP: &str = "\u{202f}";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const THOUSANDS_SEP: &str = ".";
    }
    pub mod LC_TELEPHONE {
        /// `Some("380")`
        pub const INT_PREFIX: Option<&str> = Some("380");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub use super::ru_RU::LC_TIME;
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod rw_RW {
    pub mod LC_ADDRESS {
        /// `Some("RWA")`
        pub const COUNTRY_AB3: Option<&str> = Some("RWA");
        /// `Some("RW")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("RW");
        /// `Some("RWA")`
        pub const COUNTRY_CAR: Option<&str> = Some("RWA");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("U Rwanda")`
        pub const COUNTRY_NAME: Option<&str> = Some("U Rwanda");
        /// `Some(646)`
        pub const COUNTRY_NUM: Option<i64> = Some(646);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("rw")`
        pub const LANG_AB: Option<&str> = Some("rw");
        /// `Some("kin")`
        pub const LANG_LIB: Option<&str> = Some("kin");
        /// `Some("Kinyarwanda")`
        pub const LANG_NAME: Option<&str> = Some("Kinyarwanda");
        /// `Some("kin")`
        pub const LANG_TERM: Option<&str> = Some("kin");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Rwanda, Africa")`
        pub const ADDRESS: Option<&str> = Some("Rwanda, Africa");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
        ]);
        /// `Some("Steve Murphy")`
        pub const CONTACT: Option<&str> = Some("Steve Murphy");
        /// `Some("2004-02-24")`
        pub const DATE: Option<&str> = Some("2004-02-24");
        /// `Some("murf@e-tools.com")`
        pub const EMAIL: Option<&str> = Some("murf@e-tools.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Kinyarwanda")`
        pub const LANGUAGE: Option<&str> = Some("Kinyarwanda");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Rwanda")`
        pub const SOURCE: Option<&str> = Some("Rwanda");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Rwanda")`
        pub const TERRITORY: Option<&str> = Some("Rwanda");
        /// `Some("Kinyarwanda language locale for Rwanda")`
        pub const TITLE: Option<&str> = Some("Kinyarwanda language locale for Rwanda");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNoO]"`
        pub const NOEXPR: &str = "^[-0nNoO]";
        /// `None`
        pub const NOSTR: Option<&str> = None;
        /// `"^[+1yY]"`
        pub const YESEXPR: &str = "^[+1yY]";
        /// `None`
        pub const YESSTR: Option<&str> = None;
    }
    pub mod LC_MONETARY {
        /// `"FRw"`
        pub const CURRENCY_SYMBOL: &str = "FRw";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"RWF "`
        pub const INT_CURR_SYMBOL: &str = "RWF ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const MON_THOUSANDS_SEP: &str = ".";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[-1]`
        pub const GROUPING: &[i64] = &[-1];
        /// `""`
        pub const THOUSANDS_SEP: &str = "";
    }
    pub mod LC_TELEPHONE {
        /// `Some("250")`
        pub const INT_PREFIX: Option<&str> = Some("250");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `Some("%a%l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("%a%l");
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["Mwe", "Mbe", "Kab", "Gtu", "Kan", "Gnu", "Gnd"]`
        pub const ABDAY: &[&str] = &["Mwe", "Mbe", "Kab", "Gtu", "Kan", "Gnu", "Gnd"];
        /// `&["Mut", "Gas", "Wer", "Mat", "Gic", "Kam", "Nya", "Kan", "Nze", "Ukw", "Ugu", "Uku"]`
        pub const ABMON: &[&str] = &["Mut", "Gas", "Wer", "Mat", "Gic", "Kam", "Nya", "Kan", "Nze", "Ukw", "Ugu", "Uku"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Ku cyumweru", "Kuwa mbere", "Kuwa kabiri", "Kuwa gatatu", "Kuwa kane", "Kuwa gatanu", "Kuwa gatandatu"]`
        pub const DAY: &[&str] = &["Ku cyumweru", "Kuwa mbere", "Kuwa kabiri", "Kuwa gatatu", "Kuwa kane", "Kuwa gatanu", "Kuwa gatandatu"];
        /// `"%d.%m.%Y"`
        pub const D_FMT: &str = "%d.%m.%Y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Mutarama", "Gashyantare", "Werurwe", "Mata", "Gicuransi", "Kamena", "Nyakanga", "Kanama", "Nzeli", "Ukwakira", "Ugushyingo", "Ukuboza"]`
        pub const MON: &[&str] = &["Mutarama", "Gashyantare", "Werurwe", "Mata", "Gicuransi", "Kamena", "Nyakanga", "Kanama", "Nzeli", "Ukwakira", "Ugushyingo", "Ukuboza"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod sa_IN {
    pub mod LC_ADDRESS {
        /// `Some("IND")`
        pub const COUNTRY_AB3: Option<&str> = Some("IND");
        /// `Some("IN")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IN");
        /// `Some("IND")`
        pub const COUNTRY_CAR: Option<&str> = Some("IND");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("भारतम\u{94d}")`
        pub const COUNTRY_NAME: Option<&str> = Some("भारतम\u{94d}");
        /// `Some(356)`
        pub const COUNTRY_NUM: Option<i64> = Some(356);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("sa")`
        pub const LANG_AB: Option<&str> = Some("sa");
        /// `Some("san")`
        pub const LANG_LIB: Option<&str> = Some("san");
        /// `Some("स\u{902}स\u{94d}क\u{943}तम\u{94d}")`
        pub const LANG_NAME: Option<&str> = Some("स\u{902}स\u{94d}क\u{943}तम\u{94d}");
        /// `Some("san")`
        pub const LANG_TERM: Option<&str> = Some("san");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Christian Perrier")`
        pub const CONTACT: Option<&str> = Some("Christian Perrier");
        /// `Some("2005-09-25")`
        pub const DATE: Option<&str> = Some("2005-09-25");
        /// `Some("bubulle@debian.org")`
        pub const EMAIL: Option<&str> = Some("bubulle@debian.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Sanskrit")`
        pub const LANGUAGE: Option<&str> = Some("Sanskrit");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("The Debian project")`
        pub const SOURCE: Option<&str> = Some("The Debian project");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("India")`
        pub const TERRITORY: Option<&str> = Some("India");
        /// `Some("Sanskrit language locale for India")`
        pub const TITLE: Option<&str> = Some("Sanskrit language locale for India");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNन]"`
        pub const NOEXPR: &str = "^[-0nNन]";
        /// `Some("न")`
        pub const NOSTR: Option<&str> = Some("न");
        /// `"^[+1aAyYआ]"`
        pub const YESEXPR: &str = "^[+1aAyYआ]";
        /// `Some("आम\u{94d}")`
        pub const YESSTR: Option<&str> = Some("आम\u{94d}");
    }
    pub use super::hi_IN::LC_MONETARY;
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("91")`
        pub const INT_PREFIX: Option<&str> = Some("91");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c ;%a ;%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c ;%a ;%l");
    }
    pub mod LC_TIME {
        /// `&["रविः", "सोम:", "म\u{902}गल:", "ब\u{941}ध:", "ब\u{943}हस\u{94d}पतिः", "श\u{941}क\u{94d}र", "शनि:"]`
        pub const ABDAY: &[&str] = &["रविः", "सोम:", "म\u{902}गल:", "ब\u{941}ध:", "ब\u{943}हस\u{94d}पतिः", "श\u{941}क\u{94d}र", "शनि:"];
        /// `&["जनवरी", "फ\u{93c}रवरी", "मार\u{94d}च", "अप\u{94d}र\u{947}ल", "मई", "ज\u{942}न", "ज\u{941}लाई", "अगस\u{94d}त", "सितम\u{94d}बर", "अक\u{94d}ट\u{942}बर", "नवम\u{94d}बर", "दिसम\u{94d}बर"]`
        pub const ABMON: &[&str] = &["जनवरी", "फ\u{93c}रवरी", "मार\u{94d}च", "अप\u{94d}र\u{947}ल", "मई", "ज\u{942}न", "ज\u{941}लाई", "अगस\u{94d}त", "सितम\u{94d}बर", "अक\u{94d}ट\u{942}बर", "नवम\u{94d}बर", "दिसम\u{94d}बर"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["प\u{942}र\u{94d}वाह\u{94d}न", "अपराह\u{94d}न"]`
        pub const AM_PM: &[&str] = &["प\u{942}र\u{94d}वाह\u{94d}न", "अपराह\u{94d}न"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["रविवासर:", "सोमवासर:", "म\u{902}गलवासर:", "ब\u{941}धवासर:", "ब\u{943}हस\u{94d}पतिवासरः", "श\u{941}क\u{94d}रवासर", "शनिवासर:"]`
        pub const DAY: &[&str] = &["रविवासर:", "सोमवासर:", "म\u{902}गलवासर:", "ब\u{941}धवासर:", "ब\u{943}हस\u{94d}पतिवासरः", "श\u{941}क\u{94d}रवासर", "शनिवासर:"];
        /// `"%-d-%m-%y"`
        pub const D_FMT: &str = "%-d-%m-%y";
        /// `"%A %d %b %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A %d %b %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["जनवरी", "फ\u{93c}रवरी", "मार\u{94d}च", "अप\u{94d}र\u{947}ल", "मई", "ज\u{942}न", "ज\u{941}लाई", "अगस\u{94d}त", "सितम\u{94d}बर", "अक\u{94d}ट\u{942}बर", "नवम\u{94d}बर", "दिसम\u{94d}बर"]`
        pub const MON: &[&str] = &["जनवरी", "फ\u{93c}रवरी", "मार\u{94d}च", "अप\u{94d}र\u{947}ल", "मई", "ज\u{942}न", "ज\u{941}लाई", "अगस\u{94d}त", "सितम\u{94d}बर", "अक\u{94d}ट\u{942}बर", "नवम\u{94d}बर", "दिसम\u{94d}बर"];
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT: &str = "%I:%M:%S %p %Z";
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p %Z";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod sah_RU {
    pub mod LC_ADDRESS {
        /// `Some("RUS")`
        pub const COUNTRY_AB3: Option<&str> = Some("RUS");
        /// `Some("RU")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("RU");
        /// `Some("RUS")`
        pub const COUNTRY_CAR: Option<&str> = Some("RUS");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Арассыыйа")`
        pub const COUNTRY_NAME: Option<&str> = Some("Арассыыйа");
        /// `Some(643)`
        pub const COUNTRY_NUM: Option<i64> = Some(643);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("")`
        pub const LANG_AB: Option<&str> = Some("");
        /// `Some("sah")`
        pub const LANG_LIB: Option<&str> = Some("sah");
        /// `Some("саха тыла")`
        pub const LANG_NAME: Option<&str> = Some("саха тыла");
        /// `Some("sah")`
        pub const LANG_TERM: Option<&str> = Some("sah");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
        ]);
        /// `Some("Valery Timiriliyev")`
        pub const CONTACT: Option<&str> = Some("Valery Timiriliyev");
        /// `Some("2018-07-06")`
        pub const DATE: Option<&str> = Some("2018-07-06");
        /// `Some("timiriliyev@gmail.com")`
        pub const EMAIL: Option<&str> = Some("timiriliyev@gmail.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Sakha")`
        pub const LANGUAGE: Option<&str> = Some("Sakha");
        /// `Some("1.1.0")`
        pub const REVISION: Option<&str> = Some("1.1.0");
        /// `Some("Valery Timiriliyev")`
        pub const SOURCE: Option<&str> = Some("Valery Timiriliyev");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Russian Federation")`
        pub const TERRITORY: Option<&str> = Some("Russian Federation");
        /// `Some("Sakha (Yakut) locale for Russian Federation")`
        pub const TITLE: Option<&str> = Some("Sakha (Yakut) locale for Russian Federation");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNнНсС]"`
        pub const NOEXPR: &str = "^[-0nNнНсС]";
        /// `Some("суох")`
        pub const NOSTR: Option<&str> = Some("суох");
        /// `"^[+1yYдДэЭ]"`
        pub const YESEXPR: &str = "^[+1yYдДэЭ]";
        /// `Some("ээх")`
        pub const YESSTR: Option<&str> = Some("ээх");
    }
    pub use super::ru_RU::LC_MONETARY;
    pub use super::ru_RU::LC_NUMERIC;
    pub use super::ru_RU::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["бс", "бн", "оп", "ср", "чп", "бт", "сб"]`
        pub const ABDAY: &[&str] = &["бс", "бн", "оп", "ср", "чп", "бт", "сб"];
        /// `&["тохс", "олун", "кул", "муус", "ыам", "бэс", "от", "атыр", "бал", "алт", "сэт", "ахс"]`
        pub const ABMON: &[&str] = &["тохс", "олун", "кул", "муус", "ыам", "бэс", "от", "атыр", "бал", "алт", "сэт", "ахс"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `Some(&["тохсунньу", "олунньу", "кулун тутар", "муус устар", "ыам ыйа", "бэс ыйа", "от ыйа", "атырдьах ыйа", "балаҕан ыйа", "алтынньы", "сэтинньи", "ахсынньы"])`
        pub const ALT_MON: Option<&[&str]> = Some(&["тохсунньу", "олунньу", "кулун тутар", "муус устар", "ыам ыйа", "бэс ыйа", "от ыйа", "атырдьах ыйа", "балаҕан ыйа", "алтынньы", "сэтинньи", "ахсынньы"]);
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a %Y с. %B %e к. %T (%Z)")`
        pub const DATE_FMT: Option<&str> = Some("%a %Y с. %B %e к. %T (%Z)");
        /// `&["баскыһыанньа", "бэнидиэнньик", "оптуорунньук", "сэрэдэ", "чэппиэр", "бээтинсэ", "субуота"]`
        pub const DAY: &[&str] = &["баскыһыанньа", "бэнидиэнньик", "оптуорунньук", "сэрэдэ", "чэппиэр", "бээтинсэ", "субуота"];
        /// `"%Y.%m.%d"`
        pub const D_FMT: &str = "%Y.%m.%d";
        /// `"%a %Y %b %d %T (%Z)"`
        pub const D_T_FMT: &str = "%a %Y %b %d %T (%Z)";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `Some(2)`
        pub const FIRST_WORKDAY: Option<i64> = Some(2);
        /// `&["тохсунньу", "олунньу", "кулун тутар", "муус устар", "ыам ыйын", "бэс ыйын", "от ыйын", "атырдьах ыйын", "балаҕан ыйын", "алтынньы", "сэтинньи", "ахсынньы"]`
        pub const MON: &[&str] = &["тохсунньу", "олунньу", "кулун тутар", "муус устар", "ыам ыйын", "бэс ыйын", "от ыйын", "атырдьах ыйын", "балаҕан ыйын", "алтынньы", "сэтинньи", "ахсынньы"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod sat_IN {
    pub mod LC_ADDRESS {
        /// `Some("IND")`
        pub const COUNTRY_AB3: Option<&str> = Some("IND");
        /// `Some("IN")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IN");
        /// `Some("IND")`
        pub const COUNTRY_CAR: Option<&str> = Some("IND");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("ᱥᱤᱧᱚᱛ")`
        pub const COUNTRY_NAME: Option<&str> = Some("ᱥᱤᱧᱚᱛ");
        /// `Some(356)`
        pub const COUNTRY_NUM: Option<i64> = Some(356);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("sat")`
        pub const LANG_LIB: Option<&str> = Some("sat");
        /// `Some("ᱥᱟᱱᱛᱟᱲᱤ")`
        pub const LANG_NAME: Option<&str> = Some("ᱥᱟᱱᱛᱟᱲᱤ");
        /// `Some("sat")`
        pub const LANG_TERM: Option<&str> = Some("sat");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Level 1, Tower X, Cybercity, Magarpatta City, Hadapsar, Pune-411013 ")`
        pub const ADDRESS: Option<&str> = Some("Level 1, Tower X, Cybercity, Magarpatta City, Hadapsar, Pune-411013 ");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2012-01-17")`
        pub const DATE: Option<&str> = Some("2012-01-17");
        /// `Some("libc-alpha@sourceware.org")`
        pub const EMAIL: Option<&str> = Some("libc-alpha@sourceware.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Santali")`
        pub const LANGUAGE: Option<&str> = Some("Santali");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Red Hat Pune")`
        pub const SOURCE: Option<&str> = Some("Red Hat Pune");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("India")`
        pub const TERRITORY: Option<&str> = Some("India");
        /// `Some("Santali language locale for India")`
        pub const TITLE: Option<&str> = Some("Santali language locale for India");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNब]"`
        pub const NOEXPR: &str = "^[-0nNब]";
        /// `Some("बाङ")`
        pub const NOSTR: Option<&str> = Some("बाङ");
        /// `"^[+1yYह]"`
        pub const YESEXPR: &str = "^[+1yYह]";
        /// `Some("होय")`
        pub const YESSTR: Option<&str> = Some("होय");
    }
    pub use super::hi_IN::LC_MONETARY;
    pub use super::hi_IN::LC_NUMERIC;
    pub use super::hi_IN::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["सि\u{902}ग\u{947}", "ओत\u{947}", "बाल\u{947}", "साग\u{941}न", "सारदी", "जार\u{941}म", "ञ\u{941}ह\u{941}म"]`
        pub const ABDAY: &[&str] = &["सि\u{902}ग\u{947}", "ओत\u{947}", "बाल\u{947}", "साग\u{941}न", "सारदी", "जार\u{941}म", "ञ\u{941}ह\u{941}म"];
        /// `&["जनवरी", "फरवरी", "मार\u{94d}च", "अप\u{94d}र\u{947}ल", "मई", "ज\u{941}न", "ज\u{941}लाई", "अगस\u{94d}त", "सितम\u{94d}बर", "अखथबर", "नवम\u{94d}बर", "दिसम\u{94d}बर"]`
        pub const ABMON: &[&str] = &["जनवरी", "फरवरी", "मार\u{94d}च", "अप\u{94d}र\u{947}ल", "मई", "ज\u{941}न", "ज\u{941}लाई", "अगस\u{94d}त", "सितम\u{94d}बर", "अखथबर", "नवम\u{94d}बर", "दिसम\u{94d}बर"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["AM", "PM"]`
        pub const AM_PM: &[&str] = &["AM", "PM"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["सि\u{902}ग\u{947}मा\u{901}हा\u{901}", "ओत\u{947}मा\u{901}हा\u{901}", "बाल\u{947}मा\u{901}हा\u{901}", "साग\u{941}नमा\u{901}हा\u{901}", "सारदीमा\u{901}हा\u{901}", "जार\u{941}ममा\u{901}हा\u{901}", "ञ\u{941}ह\u{941}ममा\u{901}हा\u{901}"]`
        pub const DAY: &[&str] = &["सि\u{902}ग\u{947}मा\u{901}हा\u{901}", "ओत\u{947}मा\u{901}हा\u{901}", "बाल\u{947}मा\u{901}हा\u{901}", "साग\u{941}नमा\u{901}हा\u{901}", "सारदीमा\u{901}हा\u{901}", "जार\u{941}ममा\u{901}हा\u{901}", "ञ\u{941}ह\u{941}ममा\u{901}हा\u{901}"];
        /// `"%-d/%-m/%y"`
        pub const D_FMT: &str = "%-d/%-m/%y";
        /// `"%A %d %b %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A %d %b %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["जनवरी", "फरवरी", "मार\u{94d}च", "अप\u{94d}र\u{947}ल", "मई", "ज\u{941}न", "ज\u{941}लाई", "अगस\u{94d}त", "सितम\u{94d}बर", "अखथबर", "नवम\u{94d}बर", "दिसम\u{94d}बर"]`
        pub const MON: &[&str] = &["जनवरी", "फरवरी", "मार\u{94d}च", "अप\u{94d}र\u{947}ल", "मई", "ज\u{941}न", "ज\u{941}लाई", "अगस\u{94d}त", "सितम\u{94d}बर", "अखथबर", "नवम\u{94d}बर", "दिसम\u{94d}बर"];
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT: &str = "%I:%M:%S %p %Z";
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p %Z";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod sc_IT {
    pub mod LC_ADDRESS {
        /// `Some("ITA")`
        pub const COUNTRY_AB3: Option<&str> = Some("ITA");
        /// `Some("IT")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IT");
        /// `Some("I")`
        pub const COUNTRY_CAR: Option<&str> = Some("I");
        /// `Some("978-88,979-12")`
        pub const COUNTRY_ISBN: Option<&str> = Some("978-88,979-12");
        /// `Some("Itàlia")`
        pub const COUNTRY_NAME: Option<&str> = Some("Itàlia");
        /// `Some(380)`
        pub const COUNTRY_NUM: Option<i64> = Some(380);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("sc")`
        pub const LANG_AB: Option<&str> = Some("sc");
        /// `Some("srd")`
        pub const LANG_LIB: Option<&str> = Some("srd");
        /// `Some("sardu")`
        pub const LANG_NAME: Option<&str> = Some("sardu");
        /// `Some("srd")`
        pub const LANG_TERM: Option<&str> = Some("srd");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Massimeddu Cireddu")`
        pub const CONTACT: Option<&str> = Some("Massimeddu Cireddu");
        /// `Some("2017-07-21")`
        pub const DATE: Option<&str> = Some("2017-07-21");
        /// `Some("massimeddu@gmail.com")`
        pub const EMAIL: Option<&str> = Some("massimeddu@gmail.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Sardinian")`
        pub const LANGUAGE: Option<&str> = Some("Sardinian");
        /// `Some("0.2")`
        pub const REVISION: Option<&str> = Some("0.2");
        /// `Some("Sardinian Translators Team")`
        pub const SOURCE: Option<&str> = Some("Sardinian Translators Team");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Italy")`
        pub const TERRITORY: Option<&str> = Some("Italy");
        /// `Some("Sardinian locale for Italy")`
        pub const TITLE: Option<&str> = Some("Sardinian locale for Italy");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("nono")`
        pub const NOSTR: Option<&str> = Some("nono");
        /// `"^[+1eEaAsSoOyY]"`
        pub const YESEXPR: &str = "^[+1eEaAsSoOyY]";
        /// `Some("eja")`
        pub const YESSTR: Option<&str> = Some("eja");
    }
    pub use super::it_IT::LC_MONETARY;
    pub use super::it_IT::LC_NUMERIC;
    pub use super::it_IT::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Dom", "Lun", "Mar", "Mèr", "Giò", "Che", "Sàb"]`
        pub const ABDAY: &[&str] = &["Dom", "Lun", "Mar", "Mèr", "Giò", "Che", "Sàb"];
        /// `&["Ghe", "Fre", "Mar", "Abr", "Maj", "Làm", "Arg", "Aus", "Cab", "Lad", "Onn", "Nad"]`
        pub const ABMON: &[&str] = &["Ghe", "Fre", "Mar", "Abr", "Maj", "Làm", "Arg", "Aus", "Cab", "Lad", "Onn", "Nad"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a %e %b %Y %H:%M:%S %Z")`
        pub const DATE_FMT: Option<&str> = Some("%a %e %b %Y %H:%M:%S %Z");
        /// `&["Domìnigu", "Lunis", "Martis", "Mèrcuris", "Giòbia", "Chenàbura", "Sàbadu"]`
        pub const DAY: &[&str] = &["Domìnigu", "Lunis", "Martis", "Mèrcuris", "Giòbia", "Chenàbura", "Sàbadu"];
        /// `"%d. %m. %y"`
        pub const D_FMT: &str = "%d. %m. %y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Ghennàrgiu", "Freàrgiu", "Martzu", "Abrile", "Maju", "Làmpadas", "Argiolas/Trìulas", "Austu", "Cabudanni", "Santugaine/Ladàmine", "Onniasantu/Santandria", "Nadale/Idas"]`
        pub const MON: &[&str] = &["Ghennàrgiu", "Freàrgiu", "Martzu", "Abrile", "Maju", "Làmpadas", "Argiolas/Trìulas", "Austu", "Cabudanni", "Santugaine/Ladàmine", "Onniasantu/Santandria", "Nadale/Idas"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod sd_IN {
    pub mod LC_ADDRESS {
        /// `Some("IND")`
        pub const COUNTRY_AB3: Option<&str> = Some("IND");
        /// `Some("IN")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IN");
        /// `Some("IND")`
        pub const COUNTRY_CAR: Option<&str> = Some("IND");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("انڊيا")`
        pub const COUNTRY_NAME: Option<&str> = Some("انڊيا");
        /// `Some(356)`
        pub const COUNTRY_NUM: Option<i64> = Some(356);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("sd")`
        pub const LANG_AB: Option<&str> = Some("sd");
        /// `Some("snd")`
        pub const LANG_LIB: Option<&str> = Some("snd");
        /// `Some("سنڌي")`
        pub const LANG_NAME: Option<&str> = Some("سنڌي");
        /// `Some("snd")`
        pub const LANG_TERM: Option<&str> = Some("snd");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Marisfot III, Marigold Premises, East-Wing, Kalyaninagar, Pune, India-411014")`
        pub const ADDRESS: Option<&str> = Some("Marisfot III, Marigold Premises, East-Wing, Kalyaninagar, Pune, India-411014");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2008,September,09")`
        pub const DATE: Option<&str> = Some("2008,September,09");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Sindhi")`
        pub const LANGUAGE: Option<&str> = Some("Sindhi");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Red Hat, Pune")`
        pub const SOURCE: Option<&str> = Some("Red Hat, Pune");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("India")`
        pub const TERRITORY: Option<&str> = Some("India");
        /// `Some("Sindhi language locale for India")`
        pub const TITLE: Option<&str> = Some("Sindhi language locale for India");
    }
    pub use super::ar_EG::LC_MESSAGES;
    pub use super::hi_IN::LC_MONETARY;
    pub use super::hi_IN::LC_NUMERIC;
    pub use super::hi_IN::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["آرتوار\u{64f}", "سومر\u{64f}", "منگل\u{64f}", "ٻ\u{64f}ڌر\u{64f}", "وسپت", "ج\u{64f}مو", "ڇنڇر"]`
        pub const ABDAY: &[&str] = &["آرتوار\u{64f}", "سومر\u{64f}", "منگل\u{64f}", "ٻ\u{64f}ڌر\u{64f}", "وسپت", "ج\u{64f}مو", "ڇنڇر"];
        /// `&["جنوري", "فبروري", "مارچ", "اپريل", "مي", "جون", "جولاء\u{650}", "آگسٽ", "سيپٽيمبر", "آڪٽوبر", "نومبر", "ڊسمبر"]`
        pub const ABMON: &[&str] = &["جنوري", "فبروري", "مارچ", "اپريل", "مي", "جون", "جولاء\u{650}", "آگسٽ", "سيپٽيمبر", "آڪٽوبر", "نومبر", "ڊسمبر"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["AM", "PM"]`
        pub const AM_PM: &[&str] = &["AM", "PM"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["آرتوار\u{64f}", "سومر\u{64f}", "منگل\u{64f}", "ٻ\u{64f}ڌر\u{64f}", "وسپت", "ج\u{64f}مو", "ڇنڇر"]`
        pub const DAY: &[&str] = &["آرتوار\u{64f}", "سومر\u{64f}", "منگل\u{64f}", "ٻ\u{64f}ڌر\u{64f}", "وسپت", "ج\u{64f}مو", "ڇنڇر"];
        /// `"%-d/%-m/%y"`
        pub const D_FMT: &str = "%-d/%-m/%y";
        /// `"%A %d %b %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A %d %b %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["جنوري", "فبروري", "مارچ", "اپريل", "مي", "جون", "جولاء\u{650}", "آگسٽ", "سيپٽيمبر", "آڪٽوبر", "نومبر", "ڊسمبر"]`
        pub const MON: &[&str] = &["جنوري", "فبروري", "مارچ", "اپريل", "مي", "جون", "جولاء\u{650}", "آگسٽ", "سيپٽيمبر", "آڪٽوبر", "نومبر", "ڊسمبر"];
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT: &str = "%I:%M:%S %p %Z";
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p %Z";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod sd_IN_devanagari {
    pub mod LC_ADDRESS {
        /// `Some("IND")`
        pub const COUNTRY_AB3: Option<&str> = Some("IND");
        /// `Some("IN")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IN");
        /// `Some("IND")`
        pub const COUNTRY_CAR: Option<&str> = Some("IND");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("भारत")`
        pub const COUNTRY_NAME: Option<&str> = Some("भारत");
        /// `Some(356)`
        pub const COUNTRY_NUM: Option<i64> = Some(356);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("sd")`
        pub const LANG_AB: Option<&str> = Some("sd");
        /// `Some("snd")`
        pub const LANG_LIB: Option<&str> = Some("snd");
        /// `Some("सिन\u{94d}धी")`
        pub const LANG_NAME: Option<&str> = Some("सिन\u{94d}धी");
        /// `Some("snd")`
        pub const LANG_TERM: Option<&str> = Some("snd");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Marisfot III, Marigold Premises, East-Wing, Kalyaninagar, Pune, India-411014")`
        pub const ADDRESS: Option<&str> = Some("Marisfot III, Marigold Premises, East-Wing, Kalyaninagar, Pune, India-411014");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2008-08-26")`
        pub const DATE: Option<&str> = Some("2008-08-26");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Sindhi")`
        pub const LANGUAGE: Option<&str> = Some("Sindhi");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Red Hat, Pune")`
        pub const SOURCE: Option<&str> = Some("Red Hat, Pune");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("India")`
        pub const TERRITORY: Option<&str> = Some("India");
        /// `Some("Sindhi language locale for India")`
        pub const TITLE: Option<&str> = Some("Sindhi language locale for India");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `None`
        pub const NOSTR: Option<&str> = None;
        /// `"^[+1yY]"`
        pub const YESEXPR: &str = "^[+1yY]";
        /// `None`
        pub const YESSTR: Option<&str> = None;
    }
    pub use super::sd_IN::LC_MONETARY;
    pub use super::hi_IN::LC_NUMERIC;
    pub use super::hi_IN::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["आर\u{94d}तवार\u{942}", "स\u{942}मर\u{942}", "म\u{902}गल\u{942}", "ॿ\u{941}धर\u{942}", "विस\u{94d}पति", "ज\u{941}मो", "छ\u{902}छस"]`
        pub const ABDAY: &[&str] = &["आर\u{94d}तवार\u{942}", "स\u{942}मर\u{942}", "म\u{902}गल\u{942}", "ॿ\u{941}धर\u{942}", "विस\u{94d}पति", "ज\u{941}मो", "छ\u{902}छस"];
        /// `&["जनवरी", "फबरवरी", "मार\u{94d}चि", "अप\u{94d}र\u{947}ल", "म\u{947}", "ज\u{942}नि", "ज\u{942}लाइ", "आगस\u{94d}ट\u{941}", "स\u{947}प\u{94d}ट\u{947}\u{902}बर\u{942}", "आक\u{94d}ट\u{942}बर\u{942}", "नव\u{902}बर\u{942}", "ॾिस\u{902}बर\u{942}"]`
        pub const ABMON: &[&str] = &["जनवरी", "फबरवरी", "मार\u{94d}चि", "अप\u{94d}र\u{947}ल", "म\u{947}", "ज\u{942}नि", "ज\u{942}लाइ", "आगस\u{94d}ट\u{941}", "स\u{947}प\u{94d}ट\u{947}\u{902}बर\u{942}", "आक\u{94d}ट\u{942}बर\u{942}", "नव\u{902}बर\u{942}", "ॾिस\u{902}बर\u{942}"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["म.प\u{942}.", "म.न\u{902}."]`
        pub const AM_PM: &[&str] = &["म.प\u{942}.", "म.न\u{902}."];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["आर\u{94d}तवार\u{942}", "स\u{942}मर\u{942}", "म\u{902}गल\u{942}", "ॿ\u{941}धर\u{942}", "विस\u{94d}पति", "ज\u{941}मो", "छ\u{902}छस"]`
        pub const DAY: &[&str] = &["आर\u{94d}तवार\u{942}", "स\u{942}मर\u{942}", "म\u{902}गल\u{942}", "ॿ\u{941}धर\u{942}", "विस\u{94d}पति", "ज\u{941}मो", "छ\u{902}छस"];
        /// `"%-d/%-m/%y"`
        pub const D_FMT: &str = "%-d/%-m/%y";
        /// `"%A %d %b %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A %d %b %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["जनवरी", "फबरवरी", "मार\u{94d}चि", "अप\u{94d}र\u{947}ल", "म\u{947}", "ज\u{942}नि", "ज\u{942}लाइ", "आगस\u{94d}ट\u{941}", "स\u{947}प\u{94d}ट\u{947}\u{902}बर\u{942}", "आक\u{94d}ट\u{942}बर\u{942}", "नव\u{902}बर\u{942}", "ॾिस\u{902}बर\u{942}"]`
        pub const MON: &[&str] = &["जनवरी", "फबरवरी", "मार\u{94d}चि", "अप\u{94d}र\u{947}ल", "म\u{947}", "ज\u{942}नि", "ज\u{942}लाइ", "आगस\u{94d}ट\u{941}", "स\u{947}प\u{94d}ट\u{947}\u{902}बर\u{942}", "आक\u{94d}ट\u{942}बर\u{942}", "नव\u{902}बर\u{942}", "ॾिस\u{902}बर\u{942}"];
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT: &str = "%I:%M:%S %p %Z";
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p %Z";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod se_NO {
    pub mod LC_ADDRESS {
        /// `Some("NOR")`
        pub const COUNTRY_AB3: Option<&str> = Some("NOR");
        /// `Some("NO")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("NO");
        /// `Some("N")`
        pub const COUNTRY_CAR: Option<&str> = Some("N");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Norga")`
        pub const COUNTRY_NAME: Option<&str> = Some("Norga");
        /// `Some(578)`
        pub const COUNTRY_NUM: Option<i64> = Some(578);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("se")`
        pub const LANG_AB: Option<&str> = Some("se");
        /// `Some("sme")`
        pub const LANG_LIB: Option<&str> = Some("sme");
        /// `Some("davvisámegiella")`
        pub const LANG_NAME: Option<&str> = Some("davvisámegiella");
        /// `Some("sme")`
        pub const LANG_TERM: Option<&str> = Some("sme");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
        ]);
        /// `Some("Børre Gaup")`
        pub const CONTACT: Option<&str> = Some("Børre Gaup");
        /// `Some("2001-11-09")`
        pub const DATE: Option<&str> = Some("2001-11-09");
        /// `Some("boerre.gaup@pc.nu")`
        pub const EMAIL: Option<&str> = Some("boerre.gaup@pc.nu");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Northern Sami")`
        pub const LANGUAGE: Option<&str> = Some("Northern Sami");
        /// `Some("0.1")`
        pub const REVISION: Option<&str> = Some("0.1");
        /// `Some("http://www.hum.uit.no/a/trond/loc.html")`
        pub const SOURCE: Option<&str> = Some("http://www.hum.uit.no/a/trond/loc.html");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Norway")`
        pub const TERRITORY: Option<&str> = Some("Norway");
        /// `Some("Northern Saami language locale for Norway")`
        pub const TITLE: Option<&str> = Some("Northern Saami language locale for Norway");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0IiNn]"`
        pub const NOEXPR: &str = "^[-0IiNn]";
        /// `Some("ii")`
        pub const NOSTR: Option<&str> = Some("ii");
        /// `"^[+1JjYy]"`
        pub const YESEXPR: &str = "^[+1JjYy]";
        /// `Some("jo")`
        pub const YESSTR: Option<&str> = Some("jo");
    }
    pub mod LC_MONETARY {
        /// `"kr"`
        pub const CURRENCY_SYMBOL: &str = "kr";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"NOK "`
        pub const INT_CURR_SYMBOL: &str = "NOK ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const MON_THOUSANDS_SEP: &str = ".";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `4`
        pub const N_SIGN_POSN: i64 = 4;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `4`
        pub const P_SIGN_POSN: i64 = 4;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const THOUSANDS_SEP: &str = ".";
    }
    pub mod LC_TELEPHONE {
        /// `Some("47")`
        pub const INT_PREFIX: Option<&str> = Some("47");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `Some("%l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("%l");
        /// `Some("+%c %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %l");
    }
    pub mod LC_TIME {
        /// `&["sotn", "vuos", "maŋ", "gask", "duor", "bear", "láv"]`
        pub const ABDAY: &[&str] = &["sotn", "vuos", "maŋ", "gask", "duor", "bear", "láv"];
        /// `&["ođđj", "guov", "njuk", "cuoŋ", "mies", "geas", "suoi", "borg", "čakč", "golg", "skáb", "juov"]`
        pub const ABMON: &[&str] = &["ođđj", "guov", "njuk", "cuoŋ", "mies", "geas", "suoi", "borg", "čakč", "golg", "skáb", "juov"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%A, %B %d. b. %Y %H:%M:%S %Z")`
        pub const DATE_FMT: Option<&str> = Some("%A, %B %d. b. %Y %H:%M:%S %Z");
        /// `&["sotnabeaivi", "vuossárga", "maŋŋebarga", "gaskavahkku", "duorasdat", "bearjadat", "lávvardat"]`
        pub const DAY: &[&str] = &["sotnabeaivi", "vuossárga", "maŋŋebarga", "gaskavahkku", "duorasdat", "bearjadat", "lávvardat"];
        /// `"%Y-%m-%d"`
        pub const D_FMT: &str = "%Y-%m-%d";
        /// `"%a, %b %e. b. %Y %T %Z"`
        pub const D_T_FMT: &str = "%a, %b %e. b. %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["ođđajagemánnu", "guovvamánnu", "njukčamánnu", "cuoŋománnu", "miessemánnu", "geassemánnu", "suoidnemánnu", "borgemánnu", "čakčamánnu", "golggotmánnu", "skábmamánnu", "juovlamánnu"]`
        pub const MON: &[&str] = &["ođđajagemánnu", "guovvamánnu", "njukčamánnu", "cuoŋománnu", "miessemánnu", "geassemánnu", "suoidnemánnu", "borgemánnu", "čakčamánnu", "golggotmánnu", "skábmamánnu", "juovlamánnu"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod sgs_LT {
    pub mod LC_ADDRESS {
        /// `Some("LTU")`
        pub const COUNTRY_AB3: Option<&str> = Some("LTU");
        /// `Some("LT")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("LT");
        /// `Some("LT")`
        pub const COUNTRY_CAR: Option<&str> = Some("LT");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Lietova")`
        pub const COUNTRY_NAME: Option<&str> = Some("Lietova");
        /// `Some(440)`
        pub const COUNTRY_NUM: Option<i64> = Some(440);
        /// `Some("LT")`
        pub const COUNTRY_POST: Option<&str> = Some("LT");
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("sgs")`
        pub const LANG_LIB: Option<&str> = Some("sgs");
        /// `Some("Žemaitėškā")`
        pub const LANG_NAME: Option<&str> = Some("Žemaitėškā");
        /// `Some("sgs")`
        pub const LANG_TERM: Option<&str> = Some("sgs");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Arnas Udovičius")`
        pub const CONTACT: Option<&str> = Some("Arnas Udovičius");
        /// `Some("2016-04-21")`
        pub const DATE: Option<&str> = Some("2016-04-21");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Samogitian")`
        pub const LANGUAGE: Option<&str> = Some("Samogitian");
        /// `Some("1.1")`
        pub const REVISION: Option<&str> = Some("1.1");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Lithuania")`
        pub const TERRITORY: Option<&str> = Some("Lithuania");
        /// `Some("Samogitian language locale for Lithuania")`
        pub const TITLE: Option<&str> = Some("Samogitian language locale for Lithuania");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0Nn]"`
        pub const NOEXPR: &str = "^[-0Nn]";
        /// `Some("nē")`
        pub const NOSTR: Option<&str> = Some("nē");
        /// `"^[+1TtYy]"`
        pub const YESEXPR: &str = "^[+1TtYy]";
        /// `Some("tǡp")`
        pub const YESSTR: Option<&str> = Some("tǡp");
    }
    pub use super::lt_LT::LC_MONETARY;
    pub use super::lt_LT::LC_NUMERIC;
    pub use super::lt_LT::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Nd", "Pn", "Ot", "Sr", "Čt", "Pt", "Sb"]`
        pub const ABDAY: &[&str] = &["Nd", "Pn", "Ot", "Sr", "Čt", "Pt", "Sb"];
        /// `&["Sau", "Vas", "Kuo", "Bal", "Geg", "Bėr", "Lëp", "Rgp", "Sie", "Spa", "Lap", "Grd"]`
        pub const ABMON: &[&str] = &["Sau", "Vas", "Kuo", "Bal", "Geg", "Bėr", "Lëp", "Rgp", "Sie", "Spa", "Lap", "Grd"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["nedielės dëna", "panedielis", "oterninks", "sereda", "četvergs", "petnīčė", "sobata"]`
        pub const DAY: &[&str] = &["nedielės dëna", "panedielis", "oterninks", "sereda", "četvergs", "petnīčė", "sobata"];
        /// `"%Y.%m.%d"`
        pub const D_FMT: &str = "%Y.%m.%d";
        /// `"%Y m. %B %d d. %T"`
        pub const D_T_FMT: &str = "%Y m. %B %d d. %T";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["sausė", "vasarė", "kuova", "balondė", "gegožės", "bėrželė", "lëpas", "rogpjūtė", "siejės", "spalė", "lapkrėstė", "grůdė"]`
        pub const MON: &[&str] = &["sausė", "vasarė", "kuova", "balondė", "gegožės", "bėrželė", "lëpas", "rogpjūtė", "siejės", "spalė", "lapkrėstė", "grůdė"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod shn_MM {
    pub mod LC_ADDRESS {
        /// `None`
        pub const COUNTRY_AB3: Option<&str> = None;
        /// `Some("MM")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("MM");
        /// `Some("MYA")`
        pub const COUNTRY_CAR: Option<&str> = Some("MYA");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("မ\u{102d}\u{1030}င\u{103a}းမၢၼ\u{103a}ႈ")`
        pub const COUNTRY_NAME: Option<&str> = Some("မ\u{102d}\u{1030}င\u{103a}းမၢၼ\u{103a}ႈ");
        /// `Some(104)`
        pub const COUNTRY_NUM: Option<i64> = Some(104);
        /// `Some("Myanmar")`
        pub const COUNTRY_POST: Option<&str> = Some("Myanmar");
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("shn")`
        pub const LANG_LIB: Option<&str> = Some("shn");
        /// `Some("လ\u{102d}ၵ\u{103a}ႈတ\u{1086}း")`
        pub const LANG_NAME: Option<&str> = Some("လ\u{102d}ၵ\u{103a}ႈတ\u{1086}း");
        /// `Some("shn")`
        pub const LANG_TERM: Option<&str> = Some("shn");
        /// `Some("%a%N%d%N%f%N%b%N%h%t%r%t%e%t%s%N%T%t%z%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%a%N%d%N%f%N%b%N%h%t%r%t%e%t%s%N%T%t%z%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Yangon, Myanmar")`
        pub const ADDRESS: Option<&str> = Some("Yangon, Myanmar");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_NAME"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_NAME"],
        ]);
        /// `Some("Bone Pyae Sone")`
        pub const CONTACT: Option<&str> = Some("Bone Pyae Sone");
        /// `Some("2012-07-01")`
        pub const DATE: Option<&str> = Some("2012-07-01");
        /// `Some("bone.burma@mail.com")`
        pub const EMAIL: Option<&str> = Some("bone.burma@mail.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Shan")`
        pub const LANGUAGE: Option<&str> = Some("Shan");
        /// `Some("1.3")`
        pub const REVISION: Option<&str> = Some("1.3");
        /// `Some("ubuntu Myanmar LoCo Team http:/www.ubuntu-mm.net")`
        pub const SOURCE: Option<&str> = Some("ubuntu Myanmar LoCo Team http:/www.ubuntu-mm.net");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Myanmar")`
        pub const TERRITORY: Option<&str> = Some("Myanmar");
        /// `Some("Shan language locale for Myanmar")`
        pub const TITLE: Option<&str> = Some("Shan language locale for Myanmar");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNမ]"`
        pub const NOEXPR: &str = "^[-0nNမ]";
        /// `Some("မဟ\u{102f}တ\u{103a}ဘ\u{1030}း")`
        pub const NOSTR: Option<&str> = Some("မဟ\u{102f}တ\u{103a}ဘ\u{1030}း");
        /// `"^[+1yYဟ]"`
        pub const YESEXPR: &str = "^[+1yYဟ]";
        /// `Some("ဟ\u{102f}တ\u{103a}တယ\u{103a}")`
        pub const YESSTR: Option<&str> = Some("ဟ\u{102f}တ\u{103a}တယ\u{103a}");
    }
    pub mod LC_MONETARY {
        /// `"Ks"`
        pub const CURRENCY_SYMBOL: &str = "Ks";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"MMK "`
        pub const INT_CURR_SYMBOL: &str = "MMK ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub use super::my_MM::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["တ\u{102d}တ\u{1037}\u{103a}", "ၸၼ\u{103a}", "ၵၢၼ\u{103a}း", "ပ\u{102f}တ\u{1037}\u{103a}", "ၽတ\u{103a}း", "သ\u{102f}ၵ\u{103a}း", "သဝ\u{103a}"]`
        pub const ABDAY: &[&str] = &["တ\u{102d}တ\u{1037}\u{103a}", "ၸၼ\u{103a}", "ၵၢၼ\u{103a}း", "ပ\u{102f}တ\u{1037}\u{103a}", "ၽတ\u{103a}း", "သ\u{102f}ၵ\u{103a}း", "သဝ\u{103a}"];
        /// `&["လ\u{102d}\u{1030}ၼ\u{103a}ၵမ\u{103a}", "လ\u{102d}\u{1030}ၼ\u{103a}သၢမ\u{103a}", "လ\u{102d}\u{1030}ၼ\u{103a}သ\u{102e}", "လ\u{102d}\u{1030}ၼ\u{103a}ႁႃႈ", "လ\u{102d}\u{1030}ၼ\u{103a}ႁ\u{1030}ၵ\u{103a}း", "လ\u{102d}\u{1030}ၼ\u{103a}ၸ\u{1035}တ\u{103a}း", "လ\u{102d}\u{1030}ၼ\u{103a}ပ\u{1085}တ\u{103a}ႇ", "လ\u{102d}\u{1030}ၼ\u{103a}ၵဝ\u{103a}ႈ", "လ\u{102d}\u{1030}ၼ\u{103a}သ\u{102d}ပ\u{103a}း", "လ\u{102d}\u{1030}ၼ\u{103a}သ\u{102d}ပ\u{103a}းဢ\u{102d}တ\u{103a}း", "လ\u{102d}\u{1030}ၼ\u{103a}သ\u{102d}ပ\u{103a}းဢ\u{102d}တ\u{103a}းသ\u{103d}င\u{103a}", "လ\u{102d}\u{1030}ၼ\u{103a}ၸ\u{1035}င\u{103a}"]`
        pub const ABMON: &[&str] = &["လ\u{102d}\u{1030}ၼ\u{103a}ၵမ\u{103a}", "လ\u{102d}\u{1030}ၼ\u{103a}သၢမ\u{103a}", "လ\u{102d}\u{1030}ၼ\u{103a}သ\u{102e}", "လ\u{102d}\u{1030}ၼ\u{103a}ႁႃႈ", "လ\u{102d}\u{1030}ၼ\u{103a}ႁ\u{1030}ၵ\u{103a}း", "လ\u{102d}\u{1030}ၼ\u{103a}ၸ\u{1035}တ\u{103a}း", "လ\u{102d}\u{1030}ၼ\u{103a}ပ\u{1085}တ\u{103a}ႇ", "လ\u{102d}\u{1030}ၼ\u{103a}ၵဝ\u{103a}ႈ", "လ\u{102d}\u{1030}ၼ\u{103a}သ\u{102d}ပ\u{103a}း", "လ\u{102d}\u{1030}ၼ\u{103a}သ\u{102d}ပ\u{103a}းဢ\u{102d}တ\u{103a}း", "လ\u{102d}\u{1030}ၼ\u{103a}သ\u{102d}ပ\u{103a}းဢ\u{102d}တ\u{103a}းသ\u{103d}င\u{103a}", "လ\u{102d}\u{1030}ၼ\u{103a}ၸ\u{1035}င\u{103a}"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `Some(&["႐႐", "႐႑", "႐႒", "႐႓", "႐႔", "႐႕", "႐႖", "႐႗", "႐႘", "႐႙", "႑႐", "႑႑", "႑႒", "႑႓", "႑႔", "႑႕", "႑႖", "႑႗", "႑႘", "႑႙", "႒႐", "႒႑", "႒႒", "႒႓", "႒႔", "႒႕", "႒႖", "႒႗", "႒႘", "႒႙", "႓႐", "႓႑", "႓႒", "႓႓", "႓႔", "႓႕", "႓႖", "႓႗", "႓႘", "႓႙", "႔႐", "႔႑", "႔႒", "႔႓", "႔႔", "႔႕", "႔႖", "႔႗", "႔႘", "႔႙", "႕႐", "႕႑", "႕႒", "႕႓", "႕႔", "႕႕", "႕႖", "႕႗", "႕႘", "႕႙", "႖႐", "႖႑", "႖႒", "႖႓", "႖႔", "႖႕", "႖႖", "႖႗", "႖႘", "႖႙", "႗႐", "႗႑", "႗႒", "႗႓", "႗႔", "႗႕", "႗႖", "႗႗", "႗႘", "႗႙", "႘႐", "႘႑", "႘႒", "႘႓", "႘႔", "႘႕", "႘႖", "႘႗", "႘႘", "႘႙", "႙႐", "႙႑", "႙႒", "႙႓", "႙႔", "႙႕", "႙႖", "႙႗", "႙႘", "႙႙"])`
        pub const ALT_DIGITS: Option<&[&str]> = Some(&["႐႐", "႐႑", "႐႒", "႐႓", "႐႔", "႐႕", "႐႖", "႐႗", "႐႘", "႐႙", "႑႐", "႑႑", "႑႒", "႑႓", "႑႔", "႑႕", "႑႖", "႑႗", "႑႘", "႑႙", "႒႐", "႒႑", "႒႒", "႒႓", "႒႔", "႒႕", "႒႖", "႒႗", "႒႘", "႒႙", "႓႐", "႓႑", "႓႒", "႓႓", "႓႔", "႓႕", "႓႖", "႓႗", "႓႘", "႓႙", "႔႐", "႔႑", "႔႒", "႔႓", "႔႔", "႔႕", "႔႖", "႔႗", "႔႘", "႔႙", "႕႐", "႕႑", "႕႒", "႕႓", "႕႔", "႕႕", "႕႖", "႕႗", "႕႘", "႕႙", "႖႐", "႖႑", "႖႒", "႖႓", "႖႔", "႖႕", "႖႖", "႖႗", "႖႘", "႖႙", "႗႐", "႗႑", "႗႒", "႗႓", "႗႔", "႗႕", "႗႖", "႗႗", "႗႘", "႗႙", "႘႐", "႘႑", "႘႒", "႘႓", "႘႔", "႘႕", "႘႖", "႘႗", "႘႘", "႘႙", "႙႐", "႙႑", "႙႒", "႙႓", "႙႔", "႙႕", "႙႖", "႙႗", "႙႘", "႙႙"]);
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["ၵၢင\u{103a}ၼ\u{1082}\u{103a}", "တၢမ\u{103a}းၶမ\u{103a}ႈ"]`
        pub const AM_PM: &[&str] = &["ၵၢင\u{103a}ၼ\u{1082}\u{103a}", "တၢမ\u{103a}းၶမ\u{103a}ႈ"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["ဝၼ\u{103a}းဢႃးတ\u{102d}တ\u{103a}ႉ", "ဝၼ\u{103a}းၸၼ\u{103a}", "ဝၼ\u{103a}း\u{200b}ဢၢင\u{103a}း\u{200b}ၵၢၼ\u{103a}း", "ဝၼ\u{103a}းပ\u{1030}တ\u{103a}ႉ", "ဝၼ\u{103a}းၽတ\u{103a}း", "ဝၼ\u{103a}းသ\u{102f}ၵ\u{103a}း", "ဝၼ\u{103a}းသဝ\u{103a}"]`
        pub const DAY: &[&str] = &["ဝၼ\u{103a}းဢႃးတ\u{102d}တ\u{103a}ႉ", "ဝၼ\u{103a}းၸၼ\u{103a}", "ဝၼ\u{103a}း\u{200b}ဢၢင\u{103a}း\u{200b}ၵၢၼ\u{103a}း", "ဝၼ\u{103a}းပ\u{1030}တ\u{103a}ႉ", "ဝၼ\u{103a}းၽတ\u{103a}း", "ဝၼ\u{103a}းသ\u{102f}ၵ\u{103a}း", "ဝၼ\u{103a}းသဝ\u{103a}"];
        /// `"%OC%Oy %b %Od %A"`
        pub const D_FMT: &str = "%OC%Oy %b %Od %A";
        /// `"%OC%Oy %b %Od %A %OI:%OM:%OS %Op %Z"`
        pub const D_T_FMT: &str = "%OC%Oy %b %Od %A %OI:%OM:%OS %Op %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["လ\u{102d}\u{1030}ၼ\u{103a}ၵမ\u{103a}", "လ\u{102d}\u{1030}ၼ\u{103a}သၢမ\u{103a}", "လ\u{102d}\u{1030}ၼ\u{103a}သ\u{102e}", "လ\u{102d}\u{1030}ၼ\u{103a}ႁႃႈ", "လ\u{102d}\u{1030}ၼ\u{103a}ႁ\u{1030}ၵ\u{103a}း", "လ\u{102d}\u{1030}ၼ\u{103a}ၸ\u{1035}တ\u{103a}း", "လ\u{102d}\u{1030}ၼ\u{103a}ပ\u{1085}တ\u{103a}ႇ", "လ\u{102d}\u{1030}ၼ\u{103a}ၵဝ\u{103a}ႈ", "လ\u{102d}\u{1030}ၼ\u{103a}သ\u{102d}ပ\u{103a}း", "လ\u{102d}\u{1030}ၼ\u{103a}သ\u{102d}ပ\u{103a}းဢ\u{102d}တ\u{103a}း", "လ\u{102d}\u{1030}ၼ\u{103a}သ\u{102d}ပ\u{103a}းဢ\u{102d}တ\u{103a}းသ\u{103d}င\u{103a}", "လ\u{102d}\u{1030}ၼ\u{103a}ၸ\u{1035}င\u{103a}"]`
        pub const MON: &[&str] = &["လ\u{102d}\u{1030}ၼ\u{103a}ၵမ\u{103a}", "လ\u{102d}\u{1030}ၼ\u{103a}သၢမ\u{103a}", "လ\u{102d}\u{1030}ၼ\u{103a}သ\u{102e}", "လ\u{102d}\u{1030}ၼ\u{103a}ႁႃႈ", "လ\u{102d}\u{1030}ၼ\u{103a}ႁ\u{1030}ၵ\u{103a}း", "လ\u{102d}\u{1030}ၼ\u{103a}ၸ\u{1035}တ\u{103a}း", "လ\u{102d}\u{1030}ၼ\u{103a}ပ\u{1085}တ\u{103a}ႇ", "လ\u{102d}\u{1030}ၼ\u{103a}ၵဝ\u{103a}ႈ", "လ\u{102d}\u{1030}ၼ\u{103a}သ\u{102d}ပ\u{103a}း", "လ\u{102d}\u{1030}ၼ\u{103a}သ\u{102d}ပ\u{103a}းဢ\u{102d}တ\u{103a}း", "လ\u{102d}\u{1030}ၼ\u{103a}သ\u{102d}ပ\u{103a}းဢ\u{102d}တ\u{103a}းသ\u{103d}င\u{103a}", "လ\u{102d}\u{1030}ၼ\u{103a}ၸ\u{1035}င\u{103a}"];
        /// `"%OH:%OM:%OS %p"`
        pub const T_FMT: &str = "%OH:%OM:%OS %p";
        /// `"%OI:%OM:%OS %p"`
        pub const T_FMT_AMPM: &str = "%OI:%OM:%OS %p";
        /// `None`
        pub const WEEK: Option<&[i64]> = None;
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod shs_CA {
    pub mod LC_ADDRESS {
        /// `Some("CAN")`
        pub const COUNTRY_AB3: Option<&str> = Some("CAN");
        /// `Some("CA")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("CA");
        /// `Some("CDN")`
        pub const COUNTRY_CAR: Option<&str> = Some("CDN");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `None`
        pub const COUNTRY_NAME: Option<&str> = None;
        /// `Some(124)`
        pub const COUNTRY_NUM: Option<i64> = Some(124);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("shs")`
        pub const LANG_LIB: Option<&str> = Some("shs");
        /// `Some("Secwepemctsín")`
        pub const LANG_NAME: Option<&str> = Some("Secwepemctsín");
        /// `Some("shs")`
        pub const LANG_TERM: Option<&str> = Some("shs");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("745 Ska-Hiish Dr, Chase BC V0E 1M3")`
        pub const ADDRESS: Option<&str> = Some("745 Ska-Hiish Dr, Chase BC V0E 1M3");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2008-01-15")`
        pub const DATE: Option<&str> = Some("2008-01-15");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Shuswap")`
        pub const LANGUAGE: Option<&str> = Some("Shuswap");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Neskie Manuel")`
        pub const SOURCE: Option<&str> = Some("Neskie Manuel");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Canada")`
        pub const TERRITORY: Option<&str> = Some("Canada");
        /// `Some("Shuswap locale for Canada")`
        pub const TITLE: Option<&str> = Some("Shuswap locale for Canada");
    }
    pub use super::en_CA::LC_MESSAGES;
    pub use super::en_CA::LC_MONETARY;
    pub use super::en_CA::LC_NUMERIC;
    pub use super::en_CA::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Sxe", "Spe", "Sel", "Ske", "Sme", "Sts", "Stq"]`
        pub const ABDAY: &[&str] = &["Sxe", "Spe", "Sel", "Ske", "Sme", "Sts", "Stq"];
        /// `&["Kwe", "Tsi", "Sqe", "Éwt", "Ell", "Tsp", "Tqw", "Ct\u{313}é", "Qel", "Wél", "U7l", "Tet"]`
        pub const ABMON: &[&str] = &["Kwe", "Tsi", "Sqe", "Éwt", "Ell", "Tsp", "Tqw", "Ct\u{313}é", "Qel", "Wél", "U7l", "Tet"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["AM", "PM"]`
        pub const AM_PM: &[&str] = &["AM", "PM"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Sxetspesq\u{313}t", "Spetkesq\u{313}t", "Selesq\u{313}t", "Skellesq\u{313}t", "Smesesq\u{313}t", "Stselkstesq\u{313}t", "Stqmekstesq\u{313}t"]`
        pub const DAY: &[&str] = &["Sxetspesq\u{313}t", "Spetkesq\u{313}t", "Selesq\u{313}t", "Skellesq\u{313}t", "Smesesq\u{313}t", "Stselkstesq\u{313}t", "Stqmekstesq\u{313}t"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%a %d %b %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Pellkwet\u{313}min", "Pelctsipwen\u{313}ten", "Pellsqépts", "Peslléwten", "Pell7ell7é7llqten", "Pelltspéntsk", "Pelltqwelq\u{313}wél\u{313}t", "Pellct\u{313}éxel\u{313}cten", "Pesqelqlélten", "Pesllwélsten", "Pellc7ell7é7llcwten\u{313}", "Pelltetétq\u{313}em"]`
        pub const MON: &[&str] = &["Pellkwet\u{313}min", "Pelctsipwen\u{313}ten", "Pellsqépts", "Peslléwten", "Pell7ell7é7llqten", "Pelltspéntsk", "Pelltqwelq\u{313}wél\u{313}t", "Pellct\u{313}éxel\u{313}cten", "Pesqelqlélten", "Pesllwélsten", "Pellc7ell7é7llcwten\u{313}", "Pelltetétq\u{313}em"];
        /// `"%I:%M:%S %p"`
        pub const T_FMT: &str = "%I:%M:%S %p";
        /// `"%I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod si_LK {
    pub mod LC_ADDRESS {
        /// `Some("LKA")`
        pub const COUNTRY_AB3: Option<&str> = Some("LKA");
        /// `Some("LK")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("LK");
        /// `Some("CL")`
        pub const COUNTRY_CAR: Option<&str> = Some("CL");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("ශ\u{dca}\u{200d}ර\u{dd3} ලංක\u{dcf}ව")`
        pub const COUNTRY_NAME: Option<&str> = Some("ශ\u{dca}\u{200d}ර\u{dd3} ලංක\u{dcf}ව");
        /// `Some(144)`
        pub const COUNTRY_NUM: Option<i64> = Some(144);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("si")`
        pub const LANG_AB: Option<&str> = Some("si");
        /// `Some("sin")`
        pub const LANG_LIB: Option<&str> = Some("sin");
        /// `Some("ස\u{dd2}ංහල")`
        pub const LANG_NAME: Option<&str> = Some("ස\u{dd2}ංහල");
        /// `Some("sin")`
        pub const LANG_TERM: Option<&str> = Some("sin");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2004.10.01")`
        pub const DATE: Option<&str> = Some("2004.10.01");
        /// `Some("libc-locales@sourceware.org")`
        pub const EMAIL: Option<&str> = Some("libc-locales@sourceware.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Sinhala")`
        pub const LANGUAGE: Option<&str> = Some("Sinhala");
        /// `Some("0.9")`
        pub const REVISION: Option<&str> = Some("0.9");
        /// `Some("Lanka Linux User Group (LKLUG) www.lug.lk, sinhala.linux.lk")`
        pub const SOURCE: Option<&str> = Some("Lanka Linux User Group (LKLUG) www.lug.lk, sinhala.linux.lk");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Sri Lanka")`
        pub const TERRITORY: Option<&str> = Some("Sri Lanka");
        /// `Some("Sinhala language locale for Sri Lanka")`
        pub const TITLE: Option<&str> = Some("Sinhala language locale for Sri Lanka");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNන]"`
        pub const NOEXPR: &str = "^[-0nNන]";
        /// `Some("නැත")`
        pub const NOSTR: Option<&str> = Some("නැත");
        /// `"^[+1yYඔ]"`
        pub const YESEXPR: &str = "^[+1yYඔ]";
        /// `Some("ඔව\u{dca}")`
        pub const YESSTR: Option<&str> = Some("ඔව\u{dca}");
    }
    pub mod LC_MONETARY {
        /// `"ර\u{dd4}"`
        pub const CURRENCY_SYMBOL: &str = "ර\u{dd4}";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"LKR "`
        pub const INT_CURR_SYMBOL: &str = "LKR ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("94")`
        pub const INT_PREFIX: Option<&str> = Some("94");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `Some("(%a) %l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("(%a) %l");
        /// `Some("+%c (%a) %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c (%a) %l");
    }
    pub mod LC_TIME {
        /// `&["ඉ", "ස", "අ", "බ", "බ\u{dca}\u{200d}ර", "ස\u{dd2}", "සෙ"]`
        pub const ABDAY: &[&str] = &["ඉ", "ස", "අ", "බ", "බ\u{dca}\u{200d}ර", "ස\u{dd2}", "සෙ"];
        /// `&["ජන", "පෙබ", "ම\u{dcf}ර\u{dca}", "අප\u{dca}\u{200d}ර\u{dd2}", "මැය\u{dd2}", "ජ\u{dd6}න\u{dd2}", "ජ\u{dd6}ල\u{dd2}", "අගෝ", "සැප\u{dca}", "ඔක\u{dca}", "නෙවැ", "දෙසැ"]`
        pub const ABMON: &[&str] = &["ජන", "පෙබ", "ම\u{dcf}ර\u{dca}", "අප\u{dca}\u{200d}ර\u{dd2}", "මැය\u{dd2}", "ජ\u{dd6}න\u{dd2}", "ජ\u{dd6}ල\u{dd2}", "අගෝ", "සැප\u{dca}", "ඔක\u{dca}", "නෙවැ", "දෙසැ"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["පෙ.ව.", "ප.ව."]`
        pub const AM_PM: &[&str] = &["පෙ.ව.", "ප.ව."];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%Y %B %e වැන\u{dd2} %A %H:%M:%S %z")`
        pub const DATE_FMT: Option<&str> = Some("%Y %B %e වැන\u{dd2} %A %H:%M:%S %z");
        /// `&["ඉර\u{dd2}ද\u{dcf}", "සඳ\u{dd4}ද\u{dcf}", "අඟහර\u{dd4}ව\u{dcf}ද\u{dcf}", "බද\u{dcf}ද\u{dcf}", "බ\u{dca}\u{200d}රහස\u{dca}පත\u{dd2}න\u{dca}ද\u{dcf}", "ස\u{dd2}ක\u{dd4}ර\u{dcf}ද\u{dcf}", "සෙනස\u{dd4}ර\u{dcf}ද\u{dcf}"]`
        pub const DAY: &[&str] = &["ඉර\u{dd2}ද\u{dcf}", "සඳ\u{dd4}ද\u{dcf}", "අඟහර\u{dd4}ව\u{dcf}ද\u{dcf}", "බද\u{dcf}ද\u{dcf}", "බ\u{dca}\u{200d}රහස\u{dca}පත\u{dd2}න\u{dca}ද\u{dcf}", "ස\u{dd2}ක\u{dd4}ර\u{dcf}ද\u{dcf}", "සෙනස\u{dd4}ර\u{dcf}ද\u{dcf}"];
        /// `"%Y-%m-%d"`
        pub const D_FMT: &str = "%Y-%m-%d";
        /// `"%Y-%m-%d %H:%M:%S %z"`
        pub const D_T_FMT: &str = "%Y-%m-%d %H:%M:%S %z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["ජනව\u{dcf}ර\u{dd2}", "පෙබරව\u{dcf}ර\u{dd2}", "ම\u{dcf}ර\u{dca}ත\u{dd4}", "අප\u{dca}\u{200d}රේල\u{dca}", "මැය\u{dd2}", "ජ\u{dd6}න\u{dd2}", "ජ\u{dd6}ල\u{dd2}", "අගෝස\u{dca}ත\u{dd4}", "සැප\u{dca}තැම\u{dca}බර\u{dca}", "ඔක\u{dca}තෝබර\u{dca}", "නොවැම\u{dca}බර\u{dca}", "දෙසැම\u{dca}බර\u{dca}"]`
        pub const MON: &[&str] = &["ජනව\u{dcf}ර\u{dd2}", "පෙබරව\u{dcf}ර\u{dd2}", "ම\u{dcf}ර\u{dca}ත\u{dd4}", "අප\u{dca}\u{200d}රේල\u{dca}", "මැය\u{dd2}", "ජ\u{dd6}න\u{dd2}", "ජ\u{dd6}ල\u{dd2}", "අගෝස\u{dca}ත\u{dd4}", "සැප\u{dca}තැම\u{dca}බර\u{dca}", "ඔක\u{dca}තෝබර\u{dca}", "නොවැම\u{dca}බර\u{dca}", "දෙසැම\u{dca}බර\u{dca}"];
        /// `"%H:%M:%S"`
        pub const T_FMT: &str = "%H:%M:%S";
        /// `"%p %I:%M:%S"`
        pub const T_FMT_AMPM: &str = "%p %I:%M:%S";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod sid_ET {
    pub mod LC_ADDRESS {
        /// `Some("ETH")`
        pub const COUNTRY_AB3: Option<&str> = Some("ETH");
        /// `Some("ET")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("ET");
        /// `Some("ETH")`
        pub const COUNTRY_CAR: Option<&str> = Some("ETH");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Itiyoophiya")`
        pub const COUNTRY_NAME: Option<&str> = Some("Itiyoophiya");
        /// `Some(231)`
        pub const COUNTRY_NUM: Option<i64> = Some(231);
        /// `Some("ETH")`
        pub const COUNTRY_POST: Option<&str> = Some("ETH");
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("sid")`
        pub const LANG_LIB: Option<&str> = Some("sid");
        /// `Some("Sidaamu Afo")`
        pub const LANG_NAME: Option<&str> = Some("Sidaamu Afo");
        /// `Some("sid")`
        pub const LANG_TERM: Option<&str> = Some("sid");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("7802 Solomon Seal Dr., Springfield, VA 22152, USA")`
        pub const ADDRESS: Option<&str> = Some("7802 Solomon Seal Dr., Springfield, VA 22152, USA");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_TIME"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_TIME"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2003-07-05")`
        pub const DATE: Option<&str> = Some("2003-07-05");
        /// `Some("locales@geez.org")`
        pub const EMAIL: Option<&str> = Some("locales@geez.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Sidamo")`
        pub const LANGUAGE: Option<&str> = Some("Sidamo");
        /// `Some("0.20")`
        pub const REVISION: Option<&str> = Some("0.20");
        /// `Some("Ge'ez Frontier Foundation")`
        pub const SOURCE: Option<&str> = Some("Ge'ez Frontier Foundation");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Ethiopia")`
        pub const TERRITORY: Option<&str> = Some("Ethiopia");
        /// `Some("Sidama language locale for Ethiopia.")`
        pub const TITLE: Option<&str> = Some("Sidama language locale for Ethiopia.");
    }
    pub use super::ti_ET::LC_MESSAGES;
    pub use super::ti_ET::LC_MONETARY;
    pub use super::ti_ET::LC_NUMERIC;
    pub use super::ti_ET::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Sam", "San", "Mak", "Row", "Ham", "Arb", "Qid"]`
        pub const ABDAY: &[&str] = &["Sam", "San", "Mak", "Row", "Ham", "Arb", "Qid"];
        /// `&["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["soodo", "hawwaro"]`
        pub const AM_PM: &[&str] = &["soodo", "hawwaro"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%A, %B %e, %r %Z %Y")`
        pub const DATE_FMT: Option<&str> = Some("%A, %B %e, %r %Z %Y");
        /// `&["Sambata", "Sanyo", "Maakisanyo", "Roowe", "Hamuse", "Arbe", "Qidaame"]`
        pub const DAY: &[&str] = &["Sambata", "Sanyo", "Maakisanyo", "Roowe", "Hamuse", "Arbe", "Qidaame"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%A, %B %e, %Y %l:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A, %B %e, %Y %l:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"]`
        pub const MON: &[&str] = &["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"];
        /// `"%l:%M:%S %p"`
        pub const T_FMT: &str = "%l:%M:%S %p";
        /// `"%l:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%l:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod sk_SK {
    pub mod LC_ADDRESS {
        /// `Some("SVK")`
        pub const COUNTRY_AB3: Option<&str> = Some("SVK");
        /// `Some("SK")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("SK");
        /// `Some("SK")`
        pub const COUNTRY_CAR: Option<&str> = Some("SK");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Slovensko")`
        pub const COUNTRY_NAME: Option<&str> = Some("Slovensko");
        /// `Some(703)`
        pub const COUNTRY_NUM: Option<i64> = Some(703);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("sk")`
        pub const LANG_AB: Option<&str> = Some("sk");
        /// `Some("slo")`
        pub const LANG_LIB: Option<&str> = Some("slo");
        /// `Some("slovenčina")`
        pub const LANG_NAME: Option<&str> = Some("slovenčina");
        /// `Some("slk")`
        pub const LANG_TERM: Option<&str> = Some("slk");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Narcisová 56, SK-821 01 Bratislava, Slovak Republic")`
        pub const ADDRESS: Option<&str> = Some("Narcisová 56, SK-821 01 Bratislava, Slovak Republic");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Slovak")`
        pub const LANGUAGE: Option<&str> = Some("Slovak");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Slovakia")`
        pub const TERRITORY: Option<&str> = Some("Slovakia");
        /// `Some("Slovak locale for Slovak")`
        pub const TITLE: Option<&str> = Some("Slovak locale for Slovak");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("nie")`
        pub const NOSTR: Option<&str> = Some("nie");
        /// `"^[+1aAáÁyY]"`
        pub const YESEXPR: &str = "^[+1aAáÁyY]";
        /// `Some("áno")`
        pub const YESSTR: Option<&str> = Some("áno");
    }
    pub mod LC_MONETARY {
        /// `"€"`
        pub const CURRENCY_SYMBOL: &str = "€";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"EUR "`
        pub const INT_CURR_SYMBOL: &str = "EUR ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const MON_THOUSANDS_SEP: &str = "\u{202f}";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const THOUSANDS_SEP: &str = "\u{202f}";
    }
    pub mod LC_TELEPHONE {
        /// `Some("421")`
        pub const INT_PREFIX: Option<&str> = Some("421");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["Ne", "Po", "Ut", "St", "Št", "Pi", "So"]`
        pub const ABDAY: &[&str] = &["Ne", "Po", "Ut", "St", "Št", "Pi", "So"];
        /// `&["jan", "feb", "mar", "apr", "máj", "jún", "júl", "aug", "sep", "okt", "nov", "dec"]`
        pub const ABMON: &[&str] = &["jan", "feb", "mar", "apr", "máj", "jún", "júl", "aug", "sep", "okt", "nov", "dec"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Nedeľa", "Pondelok", "Utorok", "Streda", "Štvrtok", "Piatok", "Sobota"]`
        pub const DAY: &[&str] = &["Nedeľa", "Pondelok", "Utorok", "Streda", "Štvrtok", "Piatok", "Sobota"];
        /// `"%d.%m.%Y"`
        pub const D_FMT: &str = "%d.%m.%Y";
        /// `"%a\u{a0}%e.\u{a0}%B\u{a0}%Y,\u{a0}%H:%M:%S\u{a0}%Z"`
        pub const D_T_FMT: &str = "%a\u{a0}%e.\u{a0}%B\u{a0}%Y,\u{a0}%H:%M:%S\u{a0}%Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["január", "február", "marec", "apríl", "máj", "jún", "júl", "august", "september", "október", "november", "december"]`
        pub const MON: &[&str] = &["január", "február", "marec", "apríl", "máj", "jún", "júl", "august", "september", "október", "november", "december"];
        /// `"%H:%M:%S"`
        pub const T_FMT: &str = "%H:%M:%S";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod sl_SI {
    pub mod LC_ADDRESS {
        /// `Some("SVN")`
        pub const COUNTRY_AB3: Option<&str> = Some("SVN");
        /// `Some("SI")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("SI");
        /// `Some("SLO")`
        pub const COUNTRY_CAR: Option<&str> = Some("SLO");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Slovenija")`
        pub const COUNTRY_NAME: Option<&str> = Some("Slovenija");
        /// `Some(705)`
        pub const COUNTRY_NUM: Option<i64> = Some(705);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("sl")`
        pub const LANG_AB: Option<&str> = Some("sl");
        /// `Some("slv")`
        pub const LANG_LIB: Option<&str> = Some("slv");
        /// `Some("slovenščina")`
        pub const LANG_NAME: Option<&str> = Some("slovenščina");
        /// `Some("slv")`
        pub const LANG_TERM: Option<&str> = Some("slv");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Kotnikova 6,, Ljubljana, Slovenia")`
        pub const ADDRESS: Option<&str> = Some("Kotnikova 6,, Ljubljana, Slovenia");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Slovenian")`
        pub const LANGUAGE: Option<&str> = Some("Slovenian");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("USM/MZT")`
        pub const SOURCE: Option<&str> = Some("USM/MZT");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Slovenia")`
        pub const TERRITORY: Option<&str> = Some("Slovenia");
        /// `Some("Slovenian locale for Slovenia")`
        pub const TITLE: Option<&str> = Some("Slovenian locale for Slovenia");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0Nn]"`
        pub const NOEXPR: &str = "^[-0Nn]";
        /// `Some("ne")`
        pub const NOSTR: Option<&str> = Some("ne");
        /// `"^[+1YyJj]"`
        pub const YESEXPR: &str = "^[+1YyJj]";
        /// `Some("da")`
        pub const YESSTR: Option<&str> = Some("da");
    }
    pub mod LC_MONETARY {
        /// `"€"`
        pub const CURRENCY_SYMBOL: &str = "€";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"EUR "`
        pub const INT_CURR_SYMBOL: &str = "EUR ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const MON_THOUSANDS_SEP: &str = "\u{202f}";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[0, 0]`
        pub const GROUPING: &[i64] = &[0, 0];
        /// `"\u{202f}"`
        pub const THOUSANDS_SEP: &str = "\u{202f}";
    }
    pub mod LC_TELEPHONE {
        /// `Some("386")`
        pub const INT_PREFIX: Option<&str> = Some("386");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["ned", "pon", "tor", "sre", "čet", "pet", "sob"]`
        pub const ABDAY: &[&str] = &["ned", "pon", "tor", "sre", "čet", "pet", "sob"];
        /// `&["jan", "feb", "mar", "apr", "maj", "jun", "jul", "avg", "sep", "okt", "nov", "dec"]`
        pub const ABMON: &[&str] = &["jan", "feb", "mar", "apr", "maj", "jun", "jul", "avg", "sep", "okt", "nov", "dec"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["nedelja", "ponedeljek", "torek", "sreda", "četrtek", "petek", "sobota"]`
        pub const DAY: &[&str] = &["nedelja", "ponedeljek", "torek", "sreda", "četrtek", "petek", "sobota"];
        /// `"%d. %m. %Y"`
        pub const D_FMT: &str = "%d. %m. %Y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["januar", "februar", "marec", "april", "maj", "junij", "julij", "avgust", "september", "oktober", "november", "december"]`
        pub const MON: &[&str] = &["januar", "februar", "marec", "april", "maj", "junij", "julij", "avgust", "september", "oktober", "november", "december"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod sm_WS {
    pub mod LC_ADDRESS {
        /// `Some("WSM")`
        pub const COUNTRY_AB3: Option<&str> = Some("WSM");
        /// `Some("WS")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("WS");
        /// `None`
        pub const COUNTRY_CAR: Option<&str> = None;
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Sāmoa")`
        pub const COUNTRY_NAME: Option<&str> = Some("Sāmoa");
        /// `Some(882)`
        pub const COUNTRY_NUM: Option<i64> = Some(882);
        /// `Some("%a %N   %f %N   %d %N   %b %N   %h %s %e %r %N   %T, %c  %N")`
        pub const COUNTRY_POST: Option<&str> = Some("%a %N   %f %N   %d %N   %b %N   %h %s %e %r %N   %T, %c  %N");
        /// `Some("sm")`
        pub const LANG_AB: Option<&str> = Some("sm");
        /// `Some("smo")`
        pub const LANG_LIB: Option<&str> = Some("smo");
        /// `Some("Gagana faʻa Sāmoa")`
        pub const LANG_NAME: Option<&str> = Some("Gagana faʻa Sāmoa");
        /// `Some("smo")`
        pub const LANG_TERM: Option<&str> = Some("smo");
        /// `Some("%a%N%f%N%d%N%b%N%h %s %e %r%N%T, %S %z%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%a%N%f%N%d%N%b%N%h %s %e %r%N%T, %S %z%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2017-03-21")`
        pub const DATE: Option<&str> = Some("2017-03-21");
        /// `Some("akhilesh.k@samsung.com")`
        pub const EMAIL: Option<&str> = Some("akhilesh.k@samsung.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Samoan")`
        pub const LANGUAGE: Option<&str> = Some("Samoan");
        /// `Some("0.1")`
        pub const REVISION: Option<&str> = Some("0.1");
        /// `Some("Samsung Electronics Co., Ltd.")`
        pub const SOURCE: Option<&str> = Some("Samsung Electronics Co., Ltd.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Samoa")`
        pub const TERRITORY: Option<&str> = Some("Samoa");
        /// `Some("Samoan language locale for Samoa")`
        pub const TITLE: Option<&str> = Some("Samoan language locale for Samoa");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0lLnN]"`
        pub const NOEXPR: &str = "^[-0lLnN]";
        /// `Some("leai")`
        pub const NOSTR: Option<&str> = Some("leai");
        /// `"^[+1iIyY]"`
        pub const YESEXPR: &str = "^[+1iIyY]";
        /// `Some("ioe")`
        pub const YESSTR: Option<&str> = Some("ioe");
    }
    pub mod LC_MONETARY {
        /// `"WS$"`
        pub const CURRENCY_SYMBOL: &str = "WS$";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"WST "`
        pub const INT_CURR_SYMBOL: &str = "WST ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `Some(1)`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = Some(1);
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `Some(1)`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = Some(1);
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("685")`
        pub const INT_PREFIX: Option<&str> = Some("685");
        /// `Some("0")`
        pub const INT_SELECT: Option<&str> = Some("0");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["Aso Sa", "Aso Gaf", "Aso Lua", "Aso Lul", "Aso Tof", "Aso Far", "Aso To'"]`
        pub const ABDAY: &[&str] = &["Aso Sa", "Aso Gaf", "Aso Lua", "Aso Lul", "Aso Tof", "Aso Far", "Aso To'"];
        /// `&["Ian", "Fep", "Mat", "Ape", "Me", "Iun", "Iul", "Aug", "Set", "Oke", "Nov", "Tes"]`
        pub const ABMON: &[&str] = &["Ian", "Fep", "Mat", "Ape", "Me", "Iun", "Iul", "Aug", "Set", "Oke", "Nov", "Tes"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Aso Sa", "Aso Gafua", "Aso Lua", "Aso Lulu", "Aso Tofi", "Aso Farail", "Aso To'ana'i"]`
        pub const DAY: &[&str] = &["Aso Sa", "Aso Gafua", "Aso Lua", "Aso Lulu", "Aso Tofi", "Aso Farail", "Aso To'ana'i"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Ianuari", "Fepuari", "Mati", "Aperila", "Me", "Iuni", "Iulai", "Auguso", "Setema", "Oketopa", "Novema", "Tesema"]`
        pub const MON: &[&str] = &["Ianuari", "Fepuari", "Mati", "Aperila", "Me", "Iuni", "Iulai", "Auguso", "Setema", "Oketopa", "Novema", "Tesema"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `None`
        pub const WEEK: Option<&[i64]> = None;
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod so_DJ {
    pub mod LC_ADDRESS {
        /// `Some("DJI")`
        pub const COUNTRY_AB3: Option<&str> = Some("DJI");
        /// `Some("DJ")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("DJ");
        /// `Some("DJI")`
        pub const COUNTRY_CAR: Option<&str> = Some("DJI");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Jabuuti")`
        pub const COUNTRY_NAME: Option<&str> = Some("Jabuuti");
        /// `Some(262)`
        pub const COUNTRY_NUM: Option<i64> = Some(262);
        /// `Some("DJI")`
        pub const COUNTRY_POST: Option<&str> = Some("DJI");
        /// `Some("so")`
        pub const LANG_AB: Option<&str> = Some("so");
        /// `Some("som")`
        pub const LANG_LIB: Option<&str> = Some("som");
        /// `Some("Soomaali")`
        pub const LANG_NAME: Option<&str> = Some("Soomaali");
        /// `Some("som")`
        pub const LANG_TERM: Option<&str> = Some("som");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("7802 Solomon Seal Dr., Springfield, VA 22152, USA")`
        pub const ADDRESS: Option<&str> = Some("7802 Solomon Seal Dr., Springfield, VA 22152, USA");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_TIME"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_TIME"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2003-07-05")`
        pub const DATE: Option<&str> = Some("2003-07-05");
        /// `Some("locales@geez.org")`
        pub const EMAIL: Option<&str> = Some("locales@geez.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Somali")`
        pub const LANGUAGE: Option<&str> = Some("Somali");
        /// `Some("0.20")`
        pub const REVISION: Option<&str> = Some("0.20");
        /// `Some("Ge'ez Frontier Foundation")`
        pub const SOURCE: Option<&str> = Some("Ge'ez Frontier Foundation");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Djibouti")`
        pub const TERRITORY: Option<&str> = Some("Djibouti");
        /// `Some("Somali language locale for Djibouti.")`
        pub const TITLE: Option<&str> = Some("Somali language locale for Djibouti.");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("maya")`
        pub const NOSTR: Option<&str> = Some("maya");
        /// `"^[+1oOyY]"`
        pub const YESEXPR: &str = "^[+1oOyY]";
        /// `Some("haa")`
        pub const YESSTR: Option<&str> = Some("haa");
    }
    pub use super::aa_DJ::LC_MONETARY;
    pub use super::aa_DJ::LC_NUMERIC;
    pub use super::aa_DJ::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Axd", "Isn", "Tal", "Arb", "Kha", "Jim", "Sab"]`
        pub const ABDAY: &[&str] = &["Axd", "Isn", "Tal", "Arb", "Kha", "Jim", "Sab"];
        /// `&["Kob", "Lab", "Sad", "Afr", "Sha", "Lix", "Tod", "Sid", "Sag", "Tob", "KIT", "LIT"]`
        pub const ABMON: &[&str] = &["Kob", "Lab", "Sad", "Afr", "Sha", "Lix", "Tod", "Sid", "Sag", "Tob", "KIT", "LIT"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["subaxnimo", "galabnimo"]`
        pub const AM_PM: &[&str] = &["subaxnimo", "galabnimo"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a %b %e %r %Z %Y")`
        pub const DATE_FMT: Option<&str> = Some("%a %b %e %r %Z %Y");
        /// `&["Axad", "Isniin", "Salaaso", "Arbaco", "Khamiis", "Jimco", "Sabti"]`
        pub const DAY: &[&str] = &["Axad", "Isniin", "Salaaso", "Arbaco", "Khamiis", "Jimco", "Sabti"];
        /// `"%d.%m.%Y"`
        pub const D_FMT: &str = "%d.%m.%Y";
        /// `"%a %d %b %Y %l:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %l:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(7)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(7);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Bisha Koobaad", "Bisha Labaad", "Bisha Saddexaad", "Bisha Afraad", "Bisha Shanaad", "Bisha Lixaad", "Bisha Todobaad", "Bisha Sideedaad", "Bisha Sagaalaad", "Bisha Tobnaad", "Bisha Kow iyo Tobnaad", "Bisha Laba iyo Tobnaad"]`
        pub const MON: &[&str] = &["Bisha Koobaad", "Bisha Labaad", "Bisha Saddexaad", "Bisha Afraad", "Bisha Shanaad", "Bisha Lixaad", "Bisha Todobaad", "Bisha Sideedaad", "Bisha Sagaalaad", "Bisha Tobnaad", "Bisha Kow iyo Tobnaad", "Bisha Laba iyo Tobnaad"];
        /// `"%l:%M:%S %p"`
        pub const T_FMT: &str = "%l:%M:%S %p";
        /// `"%l:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%l:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod so_ET {
    pub mod LC_ADDRESS {
        /// `Some("ETH")`
        pub const COUNTRY_AB3: Option<&str> = Some("ETH");
        /// `Some("ET")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("ET");
        /// `Some("ETH")`
        pub const COUNTRY_CAR: Option<&str> = Some("ETH");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Itoobiya")`
        pub const COUNTRY_NAME: Option<&str> = Some("Itoobiya");
        /// `Some(231)`
        pub const COUNTRY_NUM: Option<i64> = Some(231);
        /// `Some("ETH")`
        pub const COUNTRY_POST: Option<&str> = Some("ETH");
        /// `Some("so")`
        pub const LANG_AB: Option<&str> = Some("so");
        /// `Some("som")`
        pub const LANG_LIB: Option<&str> = Some("som");
        /// `Some("Soomaali")`
        pub const LANG_NAME: Option<&str> = Some("Soomaali");
        /// `Some("som")`
        pub const LANG_TERM: Option<&str> = Some("som");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("7802 Solomon Seal Dr., Springfield, VA 22152, USA")`
        pub const ADDRESS: Option<&str> = Some("7802 Solomon Seal Dr., Springfield, VA 22152, USA");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_TIME"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_TIME"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2003-07-05")`
        pub const DATE: Option<&str> = Some("2003-07-05");
        /// `Some("locales@geez.org")`
        pub const EMAIL: Option<&str> = Some("locales@geez.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Somali")`
        pub const LANGUAGE: Option<&str> = Some("Somali");
        /// `Some("0.20")`
        pub const REVISION: Option<&str> = Some("0.20");
        /// `Some("Ge'ez Frontier Foundation")`
        pub const SOURCE: Option<&str> = Some("Ge'ez Frontier Foundation");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Ethiopia")`
        pub const TERRITORY: Option<&str> = Some("Ethiopia");
        /// `Some("Somali language locale for Ethiopia")`
        pub const TITLE: Option<&str> = Some("Somali language locale for Ethiopia");
    }
    pub use super::ti_ET::LC_MESSAGES;
    pub use super::ti_ET::LC_MONETARY;
    pub use super::ti_ET::LC_NUMERIC;
    pub use super::ti_ET::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Axd", "Isn", "Sal", "Arb", "Kha", "Jim", "Sab"]`
        pub const ABDAY: &[&str] = &["Axd", "Isn", "Sal", "Arb", "Kha", "Jim", "Sab"];
        /// `&["Kob", "Lab", "Sad", "Afr", "Sha", "Lix", "Tod", "Sid", "Sag", "Tob", "KIT", "LIT"]`
        pub const ABMON: &[&str] = &["Kob", "Lab", "Sad", "Afr", "Sha", "Lix", "Tod", "Sid", "Sag", "Tob", "KIT", "LIT"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["subaxnimo", "galabnimo"]`
        pub const AM_PM: &[&str] = &["subaxnimo", "galabnimo"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%A, %B %e, %r %Z %Y")`
        pub const DATE_FMT: Option<&str> = Some("%A, %B %e, %r %Z %Y");
        /// `&["Axad", "Isniin", "Salaaso", "Arbaco", "Khamiis", "Jimco", "Sabti"]`
        pub const DAY: &[&str] = &["Axad", "Isniin", "Salaaso", "Arbaco", "Khamiis", "Jimco", "Sabti"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%A, %B %e, %Y %l:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A, %B %e, %Y %l:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Bisha Koobaad", "Bisha Labaad", "Bisha Saddexaad", "Bisha Afraad", "Bisha Shanaad", "Bisha Lixaad", "Bisha Todobaad", "Bisha Sideedaad", "Bisha Sagaalaad", "Bisha Tobnaad", "Bisha Kow iyo Tobnaad", "Bisha Laba iyo Tobnaad"]`
        pub const MON: &[&str] = &["Bisha Koobaad", "Bisha Labaad", "Bisha Saddexaad", "Bisha Afraad", "Bisha Shanaad", "Bisha Lixaad", "Bisha Todobaad", "Bisha Sideedaad", "Bisha Sagaalaad", "Bisha Tobnaad", "Bisha Kow iyo Tobnaad", "Bisha Laba iyo Tobnaad"];
        /// `"%l:%M:%S %p"`
        pub const T_FMT: &str = "%l:%M:%S %p";
        /// `"%l:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%l:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod so_KE {
    pub mod LC_ADDRESS {
        /// `Some("KEN")`
        pub const COUNTRY_AB3: Option<&str> = Some("KEN");
        /// `Some("KE")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("KE");
        /// `Some("EAK")`
        pub const COUNTRY_CAR: Option<&str> = Some("EAK");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Kiiniya")`
        pub const COUNTRY_NAME: Option<&str> = Some("Kiiniya");
        /// `Some(404)`
        pub const COUNTRY_NUM: Option<i64> = Some(404);
        /// `Some("KEN")`
        pub const COUNTRY_POST: Option<&str> = Some("KEN");
        /// `Some("so")`
        pub const LANG_AB: Option<&str> = Some("so");
        /// `Some("som")`
        pub const LANG_LIB: Option<&str> = Some("som");
        /// `Some("Soomaali")`
        pub const LANG_NAME: Option<&str> = Some("Soomaali");
        /// `Some("som")`
        pub const LANG_TERM: Option<&str> = Some("som");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("7802 Solomon Seal Dr., Springfield, VA 22152, USA")`
        pub const ADDRESS: Option<&str> = Some("7802 Solomon Seal Dr., Springfield, VA 22152, USA");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_TIME"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_TIME"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2003-07-05")`
        pub const DATE: Option<&str> = Some("2003-07-05");
        /// `Some("locales@geez.org")`
        pub const EMAIL: Option<&str> = Some("locales@geez.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Somali")`
        pub const LANGUAGE: Option<&str> = Some("Somali");
        /// `Some("0.20")`
        pub const REVISION: Option<&str> = Some("0.20");
        /// `Some("Ge'ez Frontier Foundation")`
        pub const SOURCE: Option<&str> = Some("Ge'ez Frontier Foundation");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Kenya")`
        pub const TERRITORY: Option<&str> = Some("Kenya");
        /// `Some("Somali language locale for Kenya")`
        pub const TITLE: Option<&str> = Some("Somali language locale for Kenya");
    }
    pub use super::so_SO::LC_MESSAGES;
    pub use super::om_KE::LC_MONETARY;
    pub use super::om_KE::LC_NUMERIC;
    pub use super::om_KE::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Axd", "Isn", "Sal", "Arb", "Kha", "Jim", "Sab"]`
        pub const ABDAY: &[&str] = &["Axd", "Isn", "Sal", "Arb", "Kha", "Jim", "Sab"];
        /// `&["Kob", "Lab", "Sad", "Afr", "Sha", "Lix", "Tod", "Sid", "Sag", "Tob", "KIT", "LIT"]`
        pub const ABMON: &[&str] = &["Kob", "Lab", "Sad", "Afr", "Sha", "Lix", "Tod", "Sid", "Sag", "Tob", "KIT", "LIT"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["subaxnimo", "galabnimo"]`
        pub const AM_PM: &[&str] = &["subaxnimo", "galabnimo"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%A, %B %e, %X %Z %Y")`
        pub const DATE_FMT: Option<&str> = Some("%A, %B %e, %X %Z %Y");
        /// `&["Axad", "Isniin", "Salaaso", "Arbaco", "Khamiis", "Jimco", "Sabti"]`
        pub const DAY: &[&str] = &["Axad", "Isniin", "Salaaso", "Arbaco", "Khamiis", "Jimco", "Sabti"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%A, %B %e, %Y %T %Z"`
        pub const D_T_FMT: &str = "%A, %B %e, %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Bisha Koobaad", "Bisha Labaad", "Bisha Saddexaad", "Bisha Afraad", "Bisha Shanaad", "Bisha Lixaad", "Bisha Todobaad", "Bisha Sideedaad", "Bisha Sagaalaad", "Bisha Tobnaad", "Bisha Kow iyo Tobnaad", "Bisha Laba iyo Tobnaad"]`
        pub const MON: &[&str] = &["Bisha Koobaad", "Bisha Labaad", "Bisha Saddexaad", "Bisha Afraad", "Bisha Shanaad", "Bisha Lixaad", "Bisha Todobaad", "Bisha Sideedaad", "Bisha Sagaalaad", "Bisha Tobnaad", "Bisha Kow iyo Tobnaad", "Bisha Laba iyo Tobnaad"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `"%l:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%l:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod so_SO {
    pub mod LC_ADDRESS {
        /// `Some("SOM")`
        pub const COUNTRY_AB3: Option<&str> = Some("SOM");
        /// `Some("SO")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("SO");
        /// `Some("SO")`
        pub const COUNTRY_CAR: Option<&str> = Some("SO");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Soomaaliya")`
        pub const COUNTRY_NAME: Option<&str> = Some("Soomaaliya");
        /// `Some(706)`
        pub const COUNTRY_NUM: Option<i64> = Some(706);
        /// `Some("SOM")`
        pub const COUNTRY_POST: Option<&str> = Some("SOM");
        /// `Some("so")`
        pub const LANG_AB: Option<&str> = Some("so");
        /// `Some("som")`
        pub const LANG_LIB: Option<&str> = Some("som");
        /// `Some("Soomaali")`
        pub const LANG_NAME: Option<&str> = Some("Soomaali");
        /// `Some("som")`
        pub const LANG_TERM: Option<&str> = Some("som");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("7802 Solomon Seal Dr., Springfield, VA 22152, USA")`
        pub const ADDRESS: Option<&str> = Some("7802 Solomon Seal Dr., Springfield, VA 22152, USA");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_TIME"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_TIME"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2003-07-05")`
        pub const DATE: Option<&str> = Some("2003-07-05");
        /// `Some("locales@geez.org")`
        pub const EMAIL: Option<&str> = Some("locales@geez.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Somali")`
        pub const LANGUAGE: Option<&str> = Some("Somali");
        /// `Some("0.20")`
        pub const REVISION: Option<&str> = Some("0.20");
        /// `Some("Ge'ez Frontier Foundation")`
        pub const SOURCE: Option<&str> = Some("Ge'ez Frontier Foundation");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Somalia")`
        pub const TERRITORY: Option<&str> = Some("Somalia");
        /// `Some("Somali language locale for Somalia")`
        pub const TITLE: Option<&str> = Some("Somali language locale for Somalia");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("maya")`
        pub const NOSTR: Option<&str> = Some("maya");
        /// `"^[+1yY]"`
        pub const YESEXPR: &str = "^[+1yY]";
        /// `Some("haa")`
        pub const YESSTR: Option<&str> = Some("haa");
    }
    pub mod LC_MONETARY {
        /// `"S"`
        pub const CURRENCY_SYMBOL: &str = "S";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"SOS "`
        pub const INT_CURR_SYMBOL: &str = "SOS ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("252")`
        pub const INT_PREFIX: Option<&str> = Some("252");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `Some("%a-%l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("%a-%l");
        /// `Some("%c-%a-%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("%c-%a-%l");
    }
    pub mod LC_TIME {
        /// `&["Axd", "Isn", "Sal", "Arb", "Kha", "Jim", "Sab"]`
        pub const ABDAY: &[&str] = &["Axd", "Isn", "Sal", "Arb", "Kha", "Jim", "Sab"];
        /// `&["Kob", "Lab", "Sad", "Afr", "Sha", "Lix", "Tod", "Sid", "Sag", "Tob", "KIT", "LIT"]`
        pub const ABMON: &[&str] = &["Kob", "Lab", "Sad", "Afr", "Sha", "Lix", "Tod", "Sid", "Sag", "Tob", "KIT", "LIT"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["subaxnimo", "galabnimo"]`
        pub const AM_PM: &[&str] = &["subaxnimo", "galabnimo"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%A, %B %e, %r %Z %Y")`
        pub const DATE_FMT: Option<&str> = Some("%A, %B %e, %r %Z %Y");
        /// `&["Axad", "Isniin", "Salaaso", "Arbaco", "Khamiis", "Jimco", "Sabti"]`
        pub const DAY: &[&str] = &["Axad", "Isniin", "Salaaso", "Arbaco", "Khamiis", "Jimco", "Sabti"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%A, %B %e, %Y %l:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A, %B %e, %Y %l:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Bisha Koobaad", "Bisha Labaad", "Bisha Saddexaad", "Bisha Afraad", "Bisha Shanaad", "Bisha Lixaad", "Bisha Todobaad", "Bisha Sideedaad", "Bisha Sagaalaad", "Bisha Tobnaad", "Bisha Kow iyo Tobnaad", "Bisha Laba iyo Tobnaad"]`
        pub const MON: &[&str] = &["Bisha Koobaad", "Bisha Labaad", "Bisha Saddexaad", "Bisha Afraad", "Bisha Shanaad", "Bisha Lixaad", "Bisha Todobaad", "Bisha Sideedaad", "Bisha Sagaalaad", "Bisha Tobnaad", "Bisha Kow iyo Tobnaad", "Bisha Laba iyo Tobnaad"];
        /// `"%l:%M:%S %p"`
        pub const T_FMT: &str = "%l:%M:%S %p";
        /// `"%l:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%l:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod sq_AL {
    pub mod LC_ADDRESS {
        /// `Some("ALB")`
        pub const COUNTRY_AB3: Option<&str> = Some("ALB");
        /// `Some("AL")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("AL");
        /// `Some("AL")`
        pub const COUNTRY_CAR: Option<&str> = Some("AL");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Shqipëri")`
        pub const COUNTRY_NAME: Option<&str> = Some("Shqipëri");
        /// `Some(8)`
        pub const COUNTRY_NUM: Option<i64> = Some(8);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("sq")`
        pub const LANG_AB: Option<&str> = Some("sq");
        /// `Some("alb")`
        pub const LANG_LIB: Option<&str> = Some("alb");
        /// `Some("shqip")`
        pub const LANG_NAME: Option<&str> = Some("shqip");
        /// `Some("sqi")`
        pub const LANG_TERM: Option<&str> = Some("sqi");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan")`
        pub const ADDRESS: Option<&str> = Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2004-07-01")`
        pub const DATE: Option<&str> = Some("2004-07-01");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Albanian")`
        pub const LANGUAGE: Option<&str> = Some("Albanian");
        /// `Some("1.1")`
        pub const REVISION: Option<&str> = Some("1.1");
        /// `Some("IBM Globalization Center of Competency, Yamato Software Laboratory")`
        pub const SOURCE: Option<&str> = Some("IBM Globalization Center of Competency, Yamato Software Laboratory");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Albania")`
        pub const TERRITORY: Option<&str> = Some("Albania");
        /// `Some("Albanian language locale for Albania")`
        pub const TITLE: Option<&str> = Some("Albanian language locale for Albania");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNjJ]"`
        pub const NOEXPR: &str = "^[-0nNjJ]";
        /// `Some("jo")`
        pub const NOSTR: Option<&str> = Some("jo");
        /// `"^[+1yYpP]"`
        pub const YESEXPR: &str = "^[+1yYpP]";
        /// `Some("po")`
        pub const YESSTR: Option<&str> = Some("po");
    }
    pub mod LC_MONETARY {
        /// `"L"`
        pub const CURRENCY_SYMBOL: &str = "L";
        /// `3`
        pub const FRAC_DIGITS: i64 = 3;
        /// `"ALL "`
        pub const INT_CURR_SYMBOL: &str = "ALL ";
        /// `3`
        pub const INT_FRAC_DIGITS: i64 = 3;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `"."`
        pub const MON_THOUSANDS_SEP: &str = ".";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `"."`
        pub const THOUSANDS_SEP: &str = ".";
    }
    pub mod LC_TELEPHONE {
        /// `Some("355")`
        pub const INT_PREFIX: Option<&str> = Some("355");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c ;%a ;%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c ;%a ;%l");
    }
    pub mod LC_TIME {
        /// `&["Die", "Hën", "Mar", "Mër", "Enj", "Pre", "Sht"]`
        pub const ABDAY: &[&str] = &["Die", "Hën", "Mar", "Mër", "Enj", "Pre", "Sht"];
        /// `&["Jan", "Shk", "Mar", "Pri", "Maj", "Qer", "Kor", "Gsh", "Sht", "Tet", "Nën", "Dhj"]`
        pub const ABMON: &[&str] = &["Jan", "Shk", "Mar", "Pri", "Maj", "Qer", "Kor", "Gsh", "Sht", "Tet", "Nën", "Dhj"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["PD", "MD"]`
        pub const AM_PM: &[&str] = &["PD", "MD"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a %-d %b %Y %I:%M:%S.%p %Z")`
        pub const DATE_FMT: Option<&str> = Some("%a %-d %b %Y %I:%M:%S.%p %Z");
        /// `&["e diel", "e hënë", "e martë", "e mërkurë", "e enjte", "e premte", "e shtunë"]`
        pub const DAY: &[&str] = &["e diel", "e hënë", "e martë", "e mërkurë", "e enjte", "e premte", "e shtunë"];
        /// `"%-d.%-m.%y"`
        pub const D_FMT: &str = "%-d.%-m.%y";
        /// `"%a %-d %b %Y %I:%M:%S.%p"`
        pub const D_T_FMT: &str = "%a %-d %b %Y %I:%M:%S.%p";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["janar", "shkurt", "mars", "prill", "maj", "qershor", "korrik", "gusht", "shtator", "tetor", "nëntor", "dhjetor"]`
        pub const MON: &[&str] = &["janar", "shkurt", "mars", "prill", "maj", "qershor", "korrik", "gusht", "shtator", "tetor", "nëntor", "dhjetor"];
        /// `"%I:%M:%S.%p %Z"`
        pub const T_FMT: &str = "%I:%M:%S.%p %Z";
        /// `"%I:%M:%S.%p %Z"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S.%p %Z";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod sq_MK {
    pub mod LC_ADDRESS {
        /// `Some("MKD")`
        pub const COUNTRY_AB3: Option<&str> = Some("MKD");
        /// `Some("MK")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("MK");
        /// `Some("MK")`
        pub const COUNTRY_CAR: Option<&str> = Some("MK");
        /// `Some("9989")`
        pub const COUNTRY_ISBN: Option<&str> = Some("9989");
        /// `Some("Maqedoni")`
        pub const COUNTRY_NAME: Option<&str> = Some("Maqedoni");
        /// `Some(807)`
        pub const COUNTRY_NUM: Option<i64> = Some(807);
        /// `Some("MK")`
        pub const COUNTRY_POST: Option<&str> = Some("MK");
        /// `Some("sq")`
        pub const LANG_AB: Option<&str> = Some("sq");
        /// `Some("alb")`
        pub const LANG_LIB: Option<&str> = Some("alb");
        /// `Some("shqip")`
        pub const LANG_NAME: Option<&str> = Some("shqip");
        /// `Some("sqi")`
        pub const LANG_TERM: Option<&str> = Some("sqi");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("465 4th St. 2nd Floor, Palisades Park, NJ 07650, U.S.A.")`
        pub const ADDRESS: Option<&str> = Some("465 4th St. 2nd Floor, Palisades Park, NJ 07650, U.S.A.");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2010-04-06")`
        pub const DATE: Option<&str> = Some("2010-04-06");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Albanian")`
        pub const LANGUAGE: Option<&str> = Some("Albanian");
        /// `Some("0.01")`
        pub const REVISION: Option<&str> = Some("0.01");
        /// `Some("Albanian Language Team")`
        pub const SOURCE: Option<&str> = Some("Albanian Language Team");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Macedonia")`
        pub const TERRITORY: Option<&str> = Some("Macedonia");
        /// `Some("Albanian language locale for Macedonia")`
        pub const TITLE: Option<&str> = Some("Albanian language locale for Macedonia");
    }
    pub use super::sq_AL::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"den"`
        pub const CURRENCY_SYMBOL: &str = "den";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"MKD "`
        pub const INT_CURR_SYMBOL: &str = "MKD ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const MON_THOUSANDS_SEP: &str = "\u{202f}";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub use super::sq_AL::LC_NUMERIC;
    pub use super::mk_MK::LC_TELEPHONE;
    pub use super::sq_AL::LC_TIME;
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod sr_ME {
    pub mod LC_ADDRESS {
        /// `Some("MNE")`
        pub const COUNTRY_AB3: Option<&str> = Some("MNE");
        /// `Some("ME")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("ME");
        /// `Some("MNE")`
        pub const COUNTRY_CAR: Option<&str> = Some("MNE");
        /// `Some("86")`
        pub const COUNTRY_ISBN: Option<&str> = Some("86");
        /// `Some("Црна Гора")`
        pub const COUNTRY_NAME: Option<&str> = Some("Црна Гора");
        /// `Some(499)`
        pub const COUNTRY_NUM: Option<i64> = Some(499);
        /// `Some("MNE")`
        pub const COUNTRY_POST: Option<&str> = Some("MNE");
        /// `Some("sr")`
        pub const LANG_AB: Option<&str> = Some("sr");
        /// `Some("srp")`
        pub const LANG_LIB: Option<&str> = Some("srp");
        /// `Some("српски")`
        pub const LANG_NAME: Option<&str> = Some("српски");
        /// `Some("srp")`
        pub const LANG_TERM: Option<&str> = Some("srp");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Danilo Segan")`
        pub const CONTACT: Option<&str> = Some("Danilo Segan");
        /// `Some("2006-10-11")`
        pub const DATE: Option<&str> = Some("2006-10-11");
        /// `Some("bug-glibc@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Serbian")`
        pub const LANGUAGE: Option<&str> = Some("Serbian");
        /// `Some("1.2")`
        pub const REVISION: Option<&str> = Some("1.2");
        /// `Some("sr_YU, sr_CS locale")`
        pub const SOURCE: Option<&str> = Some("sr_YU, sr_CS locale");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Montenegro")`
        pub const TERRITORY: Option<&str> = Some("Montenegro");
        /// `Some("Serbian locale for Montenegro")`
        pub const TITLE: Option<&str> = Some("Serbian locale for Montenegro");
    }
    pub use super::sr_RS::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"€"`
        pub const CURRENCY_SYMBOL: &str = "€";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"EUR "`
        pub const INT_CURR_SYMBOL: &str = "EUR ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const MON_THOUSANDS_SEP: &str = ".";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub use super::sr_RS::LC_NUMERIC;
    pub mod LC_TELEPHONE {
        /// `Some("382")`
        pub const INT_PREFIX: Option<&str> = Some("382");
        /// `Some("99")`
        pub const INT_SELECT: Option<&str> = Some("99");
        /// `Some("%A %l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("%A %l");
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["нед", "пон", "уто", "сри", "чет", "пет", "суб"]`
        pub const ABDAY: &[&str] = &["нед", "пон", "уто", "сри", "чет", "пет", "суб"];
        /// `&["јан", "феб", "мар", "апр", "мај", "јун", "јул", "авг", "сеп", "окт", "нов", "дец"]`
        pub const ABMON: &[&str] = &["јан", "феб", "мар", "апр", "мај", "јун", "јул", "авг", "сеп", "окт", "нов", "дец"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a, %e. %b %Y.  %H:%M:%S %Z")`
        pub const DATE_FMT: Option<&str> = Some("%a, %e. %b %Y.  %H:%M:%S %Z");
        /// `&["недјеља", "понедељак", "уторак", "сриједа", "четвртак", "петак", "субота"]`
        pub const DAY: &[&str] = &["недјеља", "понедељак", "уторак", "сриједа", "четвртак", "петак", "субота"];
        /// `"%d.%m.%Y."`
        pub const D_FMT: &str = "%d.%m.%Y.";
        /// `"%A, %d. %B %Y. %T %Z"`
        pub const D_T_FMT: &str = "%A, %d. %B %Y. %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["јануар", "фебруар", "март", "април", "мај", "јун", "јул", "август", "септембар", "октобар", "новембар", "децембар"]`
        pub const MON: &[&str] = &["јануар", "фебруар", "март", "април", "мај", "јун", "јул", "август", "септембар", "октобар", "новембар", "децембар"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `"%T"`
        pub const T_FMT_AMPM: &str = "%T";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod sr_RS {
    pub mod LC_ADDRESS {
        /// `Some("SRB")`
        pub const COUNTRY_AB3: Option<&str> = Some("SRB");
        /// `Some("RS")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("RS");
        /// `Some("SRB")`
        pub const COUNTRY_CAR: Option<&str> = Some("SRB");
        /// `Some("86")`
        pub const COUNTRY_ISBN: Option<&str> = Some("86");
        /// `Some("Србија")`
        pub const COUNTRY_NAME: Option<&str> = Some("Србија");
        /// `Some(688)`
        pub const COUNTRY_NUM: Option<i64> = Some(688);
        /// `Some("RS")`
        pub const COUNTRY_POST: Option<&str> = Some("RS");
        /// `Some("sr")`
        pub const LANG_AB: Option<&str> = Some("sr");
        /// `Some("srp")`
        pub const LANG_LIB: Option<&str> = Some("srp");
        /// `Some("српски")`
        pub const LANG_NAME: Option<&str> = Some("српски");
        /// `Some("srp")`
        pub const LANG_TERM: Option<&str> = Some("srp");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Danilo Segan")`
        pub const CONTACT: Option<&str> = Some("Danilo Segan");
        /// `Some("2006-10-09")`
        pub const DATE: Option<&str> = Some("2006-10-09");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Serbian")`
        pub const LANGUAGE: Option<&str> = Some("Serbian");
        /// `Some("1.3")`
        pub const REVISION: Option<&str> = Some("1.3");
        /// `Some("sr_YU, sr_CS locale")`
        pub const SOURCE: Option<&str> = Some("sr_YU, sr_CS locale");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Serbia")`
        pub const TERRITORY: Option<&str> = Some("Serbia");
        /// `Some("Serbian locale for Serbia")`
        pub const TITLE: Option<&str> = Some("Serbian locale for Serbia");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNНн]"`
        pub const NOEXPR: &str = "^[-0nNНн]";
        /// `Some("не")`
        pub const NOSTR: Option<&str> = Some("не");
        /// `"^[+1yYdDДд]"`
        pub const YESEXPR: &str = "^[+1yYdDДд]";
        /// `Some("да")`
        pub const YESSTR: Option<&str> = Some("да");
    }
    pub mod LC_MONETARY {
        /// `"дин"`
        pub const CURRENCY_SYMBOL: &str = "дин";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"RSD "`
        pub const INT_CURR_SYMBOL: &str = "RSD ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const MON_THOUSANDS_SEP: &str = ".";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[0, 0]`
        pub const GROUPING: &[i64] = &[0, 0];
        /// `""`
        pub const THOUSANDS_SEP: &str = "";
    }
    pub mod LC_TELEPHONE {
        /// `Some("381")`
        pub const INT_PREFIX: Option<&str> = Some("381");
        /// `Some("99")`
        pub const INT_SELECT: Option<&str> = Some("99");
        /// `Some("%A %l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("%A %l");
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["нед", "пон", "уто", "сре", "чет", "пет", "суб"]`
        pub const ABDAY: &[&str] = &["нед", "пон", "уто", "сре", "чет", "пет", "суб"];
        /// `&["јан", "феб", "мар", "апр", "мај", "јун", "јул", "авг", "сеп", "окт", "нов", "дец"]`
        pub const ABMON: &[&str] = &["јан", "феб", "мар", "апр", "мај", "јун", "јул", "авг", "сеп", "окт", "нов", "дец"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a, %e. %b %Y.  %H:%M:%S %Z")`
        pub const DATE_FMT: Option<&str> = Some("%a, %e. %b %Y.  %H:%M:%S %Z");
        /// `&["недеља", "понедељак", "уторак", "среда", "четвртак", "петак", "субота"]`
        pub const DAY: &[&str] = &["недеља", "понедељак", "уторак", "среда", "четвртак", "петак", "субота"];
        /// `"%d.%m.%Y."`
        pub const D_FMT: &str = "%d.%m.%Y.";
        /// `"%A, %d. %B %Y. %T %Z"`
        pub const D_T_FMT: &str = "%A, %d. %B %Y. %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["јануар", "фебруар", "март", "април", "мај", "јун", "јул", "август", "септембар", "октобар", "новембар", "децембар"]`
        pub const MON: &[&str] = &["јануар", "фебруар", "март", "април", "мај", "јун", "јул", "август", "септембар", "октобар", "новембар", "децембар"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `"%T"`
        pub const T_FMT_AMPM: &str = "%T";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod sr_RS_latin {
    pub mod LC_ADDRESS {
        /// `Some("SRB")`
        pub const COUNTRY_AB3: Option<&str> = Some("SRB");
        /// `Some("RS")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("RS");
        /// `Some("SRB")`
        pub const COUNTRY_CAR: Option<&str> = Some("SRB");
        /// `Some("86")`
        pub const COUNTRY_ISBN: Option<&str> = Some("86");
        /// `Some("Srbija")`
        pub const COUNTRY_NAME: Option<&str> = Some("Srbija");
        /// `Some(688)`
        pub const COUNTRY_NUM: Option<i64> = Some(688);
        /// `Some("RS")`
        pub const COUNTRY_POST: Option<&str> = Some("RS");
        /// `Some("sr")`
        pub const LANG_AB: Option<&str> = Some("sr");
        /// `Some("srp")`
        pub const LANG_LIB: Option<&str> = Some("srp");
        /// `Some("srpski")`
        pub const LANG_NAME: Option<&str> = Some("srpski");
        /// `Some("srp")`
        pub const LANG_TERM: Option<&str> = Some("srp");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Danilo Segan")`
        pub const CONTACT: Option<&str> = Some("Danilo Segan");
        /// `Some("2006-10-09")`
        pub const DATE: Option<&str> = Some("2006-10-09");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Serbian")`
        pub const LANGUAGE: Option<&str> = Some("Serbian");
        /// `Some("1.3")`
        pub const REVISION: Option<&str> = Some("1.3");
        /// `Some("sr_YU, sr_CS locale")`
        pub const SOURCE: Option<&str> = Some("sr_YU, sr_CS locale");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Serbia")`
        pub const TERRITORY: Option<&str> = Some("Serbia");
        /// `Some("Serbian Latin locale for Serbia")`
        pub const TITLE: Option<&str> = Some("Serbian Latin locale for Serbia");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("ne")`
        pub const NOSTR: Option<&str> = Some("ne");
        /// `"^[+1dDyY]"`
        pub const YESEXPR: &str = "^[+1dDyY]";
        /// `Some("da")`
        pub const YESSTR: Option<&str> = Some("da");
    }
    pub mod LC_MONETARY {
        /// `"din"`
        pub const CURRENCY_SYMBOL: &str = "din";
        /// `0`
        pub const FRAC_DIGITS: i64 = 0;
        /// `"RSD "`
        pub const INT_CURR_SYMBOL: &str = "RSD ";
        /// `0`
        pub const INT_FRAC_DIGITS: i64 = 0;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const MON_THOUSANDS_SEP: &str = ".";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub use super::sr_RS::LC_NUMERIC;
    pub use super::sr_RS::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["ned", "pon", "uto", "sre", "čet", "pet", "sub"]`
        pub const ABDAY: &[&str] = &["ned", "pon", "uto", "sre", "čet", "pet", "sub"];
        /// `&["jan", "feb", "mar", "apr", "maj", "jun", "jul", "avg", "sep", "okt", "nov", "dec"]`
        pub const ABMON: &[&str] = &["jan", "feb", "mar", "apr", "maj", "jun", "jul", "avg", "sep", "okt", "nov", "dec"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a, %e. %b %Y.  %H:%M:%S %Z")`
        pub const DATE_FMT: Option<&str> = Some("%a, %e. %b %Y.  %H:%M:%S %Z");
        /// `&["nedelja", "ponedeljak", "utorak", "sreda", "četvrtak", "petak", "subota"]`
        pub const DAY: &[&str] = &["nedelja", "ponedeljak", "utorak", "sreda", "četvrtak", "petak", "subota"];
        /// `"%d.%m.%Y."`
        pub const D_FMT: &str = "%d.%m.%Y.";
        /// `"%A, %d. %B %Y. %T %Z"`
        pub const D_T_FMT: &str = "%A, %d. %B %Y. %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["januar", "februar", "mart", "april", "maj", "jun", "jul", "avgust", "septembar", "oktobar", "novembar", "decembar"]`
        pub const MON: &[&str] = &["januar", "februar", "mart", "april", "maj", "jun", "jul", "avgust", "septembar", "oktobar", "novembar", "decembar"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `"%T"`
        pub const T_FMT_AMPM: &str = "%T";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ss_ZA {
    pub mod LC_ADDRESS {
        /// `Some("ZAF")`
        pub const COUNTRY_AB3: Option<&str> = Some("ZAF");
        /// `Some("ZA")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("ZA");
        /// `Some("ZA")`
        pub const COUNTRY_CAR: Option<&str> = Some("ZA");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("iNingizimu Afrika")`
        pub const COUNTRY_NAME: Option<&str> = Some("iNingizimu Afrika");
        /// `Some(710)`
        pub const COUNTRY_NUM: Option<i64> = Some(710);
        /// `Some("ZA")`
        pub const COUNTRY_POST: Option<&str> = Some("ZA");
        /// `Some("ss")`
        pub const LANG_AB: Option<&str> = Some("ss");
        /// `Some("ssw")`
        pub const LANG_LIB: Option<&str> = Some("ssw");
        /// `Some("siSwati")`
        pub const LANG_NAME: Option<&str> = Some("siSwati");
        /// `Some("ssw")`
        pub const LANG_TERM: Option<&str> = Some("ssw");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("PO Box 28364, Sunnyside, 0132, South Africa")`
        pub const ADDRESS: Option<&str> = Some("PO Box 28364, Sunnyside, 0132, South Africa");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Dwayne Bailey")`
        pub const CONTACT: Option<&str> = Some("Dwayne Bailey");
        /// `Some("2005-10-13")`
        pub const DATE: Option<&str> = Some("2005-10-13");
        /// `Some("dwayne@translate.org.za")`
        pub const EMAIL: Option<&str> = Some("dwayne@translate.org.za");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Swati")`
        pub const LANGUAGE: Option<&str> = Some("Swati");
        /// `Some("0.3")`
        pub const REVISION: Option<&str> = Some("0.3");
        /// `Some("Zuza Software Foundation (Translate.org.za)")`
        pub const SOURCE: Option<&str> = Some("Zuza Software Foundation (Translate.org.za)");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("South Africa")`
        pub const TERRITORY: Option<&str> = Some("South Africa");
        /// `Some("Swati locale for South Africa")`
        pub const TITLE: Option<&str> = Some("Swati locale for South Africa");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNaA]"`
        pub const NOEXPR: &str = "^[-0nNaA]";
        /// `None`
        pub const NOSTR: Option<&str> = None;
        /// `"^[+1yY]"`
        pub const YESEXPR: &str = "^[+1yY]";
        /// `None`
        pub const YESSTR: Option<&str> = None;
    }
    pub use super::en_ZA::LC_MONETARY;
    pub use super::en_ZA::LC_NUMERIC;
    pub use super::en_ZA::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Son", "Mso", "Bil", "Tsa", "Ne", "Hla", "Mgc"]`
        pub const ABDAY: &[&str] = &["Son", "Mso", "Bil", "Tsa", "Ne", "Hla", "Mgc"];
        /// `&["Bhi", "Van", "Vul", "Mab", "Khk", "Nhl", "Kho", "Ngc", "Nyo", "Imp", "Lwe", "Ngo"]`
        pub const ABMON: &[&str] = &["Bhi", "Van", "Vul", "Mab", "Khk", "Nhl", "Kho", "Ngc", "Nyo", "Imp", "Lwe", "Ngo"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a %b %-e %H:%M:%S %Z %Y")`
        pub const DATE_FMT: Option<&str> = Some("%a %b %-e %H:%M:%S %Z %Y");
        /// `&["Lisontfo", "uMsombuluko", "Lesibili", "Lesitsatfu", "Lesine", "Lesihlanu", "uMgcibelo"]`
        pub const DAY: &[&str] = &["Lisontfo", "uMsombuluko", "Lesibili", "Lesitsatfu", "Lesine", "Lesihlanu", "uMgcibelo"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%a %-e %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %-e %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Bhimbidvwane", "iNdlovane", "iNdlovulenkhulu", "Mabasa", "Inkhwenkhweti", "iNhlaba", "Kholwane", "iNgci", "iNyoni", "Impala", "Lweti", "iNgongoni"]`
        pub const MON: &[&str] = &["Bhimbidvwane", "iNdlovane", "iNdlovulenkhulu", "Mabasa", "Inkhwenkhweti", "iNhlaba", "Kholwane", "iNgci", "iNyoni", "Impala", "Lweti", "iNgongoni"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod st_ZA {
    pub mod LC_ADDRESS {
        /// `Some("ZAF")`
        pub const COUNTRY_AB3: Option<&str> = Some("ZAF");
        /// `Some("ZA")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("ZA");
        /// `Some("ZA")`
        pub const COUNTRY_CAR: Option<&str> = Some("ZA");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Afrika Borwa")`
        pub const COUNTRY_NAME: Option<&str> = Some("Afrika Borwa");
        /// `Some(710)`
        pub const COUNTRY_NUM: Option<i64> = Some(710);
        /// `Some("ZA")`
        pub const COUNTRY_POST: Option<&str> = Some("ZA");
        /// `Some("st")`
        pub const LANG_AB: Option<&str> = Some("st");
        /// `Some("sot")`
        pub const LANG_LIB: Option<&str> = Some("sot");
        /// `Some("Sesotho")`
        pub const LANG_NAME: Option<&str> = Some("Sesotho");
        /// `Some("sot")`
        pub const LANG_TERM: Option<&str> = Some("sot");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("PO Box 28364, Sunnyside, 0132, South Africa")`
        pub const ADDRESS: Option<&str> = Some("PO Box 28364, Sunnyside, 0132, South Africa");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Dwayne Bailey")`
        pub const CONTACT: Option<&str> = Some("Dwayne Bailey");
        /// `Some("2005-10-13")`
        pub const DATE: Option<&str> = Some("2005-10-13");
        /// `Some("dwayne@translate.org.za")`
        pub const EMAIL: Option<&str> = Some("dwayne@translate.org.za");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Southern Sotho")`
        pub const LANGUAGE: Option<&str> = Some("Southern Sotho");
        /// `Some("0.3")`
        pub const REVISION: Option<&str> = Some("0.3");
        /// `Some("Zuza Software Foundation (Translate.org.za)")`
        pub const SOURCE: Option<&str> = Some("Zuza Software Foundation (Translate.org.za)");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("South Africa")`
        pub const TERRITORY: Option<&str> = Some("South Africa");
        /// `Some("Sotho locale for South Africa")`
        pub const TITLE: Option<&str> = Some("Sotho locale for South Africa");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `None`
        pub const NOSTR: Option<&str> = None;
        /// `"^[+1yY]"`
        pub const YESEXPR: &str = "^[+1yY]";
        /// `None`
        pub const YESSTR: Option<&str> = None;
    }
    pub use super::en_ZA::LC_MONETARY;
    pub use super::en_ZA::LC_NUMERIC;
    pub use super::en_ZA::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Son", "Mma", "Bed", "Rar", "Ne", "Hla", "Moq"]`
        pub const ABDAY: &[&str] = &["Son", "Mma", "Bed", "Rar", "Ne", "Hla", "Moq"];
        /// `&["Phe", "Hla", "TlH", "Mme", "Mot", "Jan", "Upu", "Pha", "Leo", "Mph", "Pud", "Tsh"]`
        pub const ABMON: &[&str] = &["Phe", "Hla", "TlH", "Mme", "Mot", "Jan", "Upu", "Pha", "Leo", "Mph", "Pud", "Tsh"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a %b %-e %H:%M:%S %Z %Y")`
        pub const DATE_FMT: Option<&str> = Some("%a %b %-e %H:%M:%S %Z %Y");
        /// `&["Sontaha", "Mantaha", "Labobedi", "Laboraro", "Labone", "Labohlano", "Moqebelo"]`
        pub const DAY: &[&str] = &["Sontaha", "Mantaha", "Labobedi", "Laboraro", "Labone", "Labohlano", "Moqebelo"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%a %-e %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %-e %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Pherekgong", "Hlakola", "Tlhakubele", "Mmese", "Motsheanong", "Phupjane", "Phupu", "Phato", "Leotse", "Mphalane", "Pudungwana", "Tshitwe"]`
        pub const MON: &[&str] = &["Pherekgong", "Hlakola", "Tlhakubele", "Mmese", "Motsheanong", "Phupjane", "Phupu", "Phato", "Leotse", "Mphalane", "Pudungwana", "Tshitwe"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod sv_FI {
    pub mod LC_ADDRESS {
        /// `Some("FIN")`
        pub const COUNTRY_AB3: Option<&str> = Some("FIN");
        /// `Some("FI")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("FI");
        /// `Some("FIN")`
        pub const COUNTRY_CAR: Option<&str> = Some("FIN");
        /// `Some("952")`
        pub const COUNTRY_ISBN: Option<&str> = Some("952");
        /// `Some("Finland")`
        pub const COUNTRY_NAME: Option<&str> = Some("Finland");
        /// `Some(246)`
        pub const COUNTRY_NUM: Option<i64> = Some(246);
        /// `Some("FI")`
        pub const COUNTRY_POST: Option<&str> = Some("FI");
        /// `Some("sv")`
        pub const LANG_AB: Option<&str> = Some("sv");
        /// `Some("swe")`
        pub const LANG_LIB: Option<&str> = Some("swe");
        /// `Some("svenska")`
        pub const LANG_NAME: Option<&str> = Some("svenska");
        /// `Some("swe")`
        pub const LANG_TERM: Option<&str> = Some("swe");
        /// `Some("%f%N%d%N%b%N%a%N%s %h%t%e%t%r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%d%N%b%N%a%N%s %h%t%e%t%r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Swedish")`
        pub const LANGUAGE: Option<&str> = Some("Swedish");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Finland")`
        pub const TERRITORY: Option<&str> = Some("Finland");
        /// `Some("Swedish locale for Finland")`
        pub const TITLE: Option<&str> = Some("Swedish locale for Finland");
    }
    pub use super::sv_SE::LC_MESSAGES;
    pub use super::fi_FI::LC_MONETARY;
    pub use super::fi_FI::LC_NUMERIC;
    pub use super::fi_FI::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["sön", "mån", "tis", "ons", "tor", "fre", "lör"]`
        pub const ABDAY: &[&str] = &["sön", "mån", "tis", "ons", "tor", "fre", "lör"];
        /// `&["jan", "feb", "mar", "apr", "maj", "jun", "jul", "aug", "sep", "okt", "nov", "dec"]`
        pub const ABMON: &[&str] = &["jan", "feb", "mar", "apr", "maj", "jun", "jul", "aug", "sep", "okt", "nov", "dec"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a %-d.%-m.%Y %H.%M.%S %z")`
        pub const DATE_FMT: Option<&str> = Some("%a %-d.%-m.%Y %H.%M.%S %z");
        /// `&["söndag", "måndag", "tisdag", "onsdag", "torsdag", "fredag", "lördag"]`
        pub const DAY: &[&str] = &["söndag", "måndag", "tisdag", "onsdag", "torsdag", "fredag", "lördag"];
        /// `"%d.%m.%Y"`
        pub const D_FMT: &str = "%d.%m.%Y";
        /// `"%a %e. %B %Y %H.%M.%S"`
        pub const D_T_FMT: &str = "%a %e. %B %Y %H.%M.%S";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["januari", "februari", "mars", "april", "maj", "juni", "juli", "augusti", "september", "oktober", "november", "december"]`
        pub const MON: &[&str] = &["januari", "februari", "mars", "april", "maj", "juni", "juli", "augusti", "september", "oktober", "november", "december"];
        /// `"%H.%M.%S"`
        pub const T_FMT: &str = "%H.%M.%S";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod sv_FI_euro {
    pub use super::sv_FI::LC_ADDRESS;
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("https://www.gnu.org/software/libc/")`
        pub const ADDRESS: Option<&str> = Some("https://www.gnu.org/software/libc/");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-08-21")`
        pub const DATE: Option<&str> = Some("2000-08-21");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Swedish")`
        pub const LANGUAGE: Option<&str> = Some("Swedish");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Free Software Foundation, Inc.")`
        pub const SOURCE: Option<&str> = Some("Free Software Foundation, Inc.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Finland")`
        pub const TERRITORY: Option<&str> = Some("Finland");
        /// `Some("Swedish locale for Finland with Euro")`
        pub const TITLE: Option<&str> = Some("Swedish locale for Finland with Euro");
    }
    pub use super::sv_FI::LC_MESSAGES;
    pub use super::fi_FI_euro::LC_MONETARY;
    pub use super::fi_FI::LC_NUMERIC;
    pub use super::sv_FI::LC_TELEPHONE;
    pub use super::sv_FI::LC_TIME;
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod sv_SE {
    pub mod LC_ADDRESS {
        /// `Some("SWE")`
        pub const COUNTRY_AB3: Option<&str> = Some("SWE");
        /// `Some("SE")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("SE");
        /// `Some("S")`
        pub const COUNTRY_CAR: Option<&str> = Some("S");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Sverige")`
        pub const COUNTRY_NAME: Option<&str> = Some("Sverige");
        /// `Some(752)`
        pub const COUNTRY_NUM: Option<i64> = Some(752);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("sv")`
        pub const LANG_AB: Option<&str> = Some("sv");
        /// `Some("swe")`
        pub const LANG_LIB: Option<&str> = Some("swe");
        /// `Some("svenska")`
        pub const LANG_NAME: Option<&str> = Some("svenska");
        /// `Some("swe")`
        pub const LANG_TERM: Option<&str> = Some("swe");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Swedish")`
        pub const LANGUAGE: Option<&str> = Some("Swedish");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Sweden")`
        pub const TERRITORY: Option<&str> = Some("Sweden");
        /// `Some("Swedish locale for Sweden")`
        pub const TITLE: Option<&str> = Some("Swedish locale for Sweden");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("nej")`
        pub const NOSTR: Option<&str> = Some("nej");
        /// `"^[+1jJyY]"`
        pub const YESEXPR: &str = "^[+1jJyY]";
        /// `Some("ja")`
        pub const YESSTR: Option<&str> = Some("ja");
    }
    pub mod LC_MONETARY {
        /// `"kr"`
        pub const CURRENCY_SYMBOL: &str = "kr";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"SEK "`
        pub const INT_CURR_SYMBOL: &str = "SEK ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const MON_THOUSANDS_SEP: &str = "\u{202f}";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const THOUSANDS_SEP: &str = "\u{202f}";
    }
    pub mod LC_TELEPHONE {
        /// `Some("46")`
        pub const INT_PREFIX: Option<&str> = Some("46");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `Some("0%a-%l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("0%a-%l");
        /// `Some("+%c-%a-%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c-%a-%l");
    }
    pub mod LC_TIME {
        /// `&["sön", "mån", "tis", "ons", "tor", "fre", "lör"]`
        pub const ABDAY: &[&str] = &["sön", "mån", "tis", "ons", "tor", "fre", "lör"];
        /// `&["jan", "feb", "mar", "apr", "maj", "jun", "jul", "aug", "sep", "okt", "nov", "dec"]`
        pub const ABMON: &[&str] = &["jan", "feb", "mar", "apr", "maj", "jun", "jul", "aug", "sep", "okt", "nov", "dec"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["söndag", "måndag", "tisdag", "onsdag", "torsdag", "fredag", "lördag"]`
        pub const DAY: &[&str] = &["söndag", "måndag", "tisdag", "onsdag", "torsdag", "fredag", "lördag"];
        /// `"%Y-%m-%d"`
        pub const D_FMT: &str = "%Y-%m-%d";
        /// `"%a %e %b %Y %H:%M:%S"`
        pub const D_T_FMT: &str = "%a %e %b %Y %H:%M:%S";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["januari", "februari", "mars", "april", "maj", "juni", "juli", "augusti", "september", "oktober", "november", "december"]`
        pub const MON: &[&str] = &["januari", "februari", "mars", "april", "maj", "juni", "juli", "augusti", "september", "oktober", "november", "december"];
        /// `"%H:%M:%S"`
        pub const T_FMT: &str = "%H:%M:%S";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod sw_KE {
    pub mod LC_ADDRESS {
        /// `Some("KEN")`
        pub const COUNTRY_AB3: Option<&str> = Some("KEN");
        /// `Some("KE")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("KE");
        /// `Some("EAK")`
        pub const COUNTRY_CAR: Option<&str> = Some("EAK");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Kenya")`
        pub const COUNTRY_NAME: Option<&str> = Some("Kenya");
        /// `Some(404)`
        pub const COUNTRY_NUM: Option<i64> = Some(404);
        /// `Some("KE")`
        pub const COUNTRY_POST: Option<&str> = Some("KE");
        /// `Some("sw")`
        pub const LANG_AB: Option<&str> = Some("sw");
        /// `Some("swa")`
        pub const LANG_LIB: Option<&str> = Some("swa");
        /// `Some("Kiswahili")`
        pub const LANG_NAME: Option<&str> = Some("Kiswahili");
        /// `Some("swa")`
        pub const LANG_TERM: Option<&str> = Some("swa");
        /// `Some("%a%N%f%N%d%N%b%N%h %s %e %r%N%T, %S %z%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%a%N%f%N%d%N%b%N%h %s %e %r%N%T, %S %z%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `None`
        pub const ADDRESS: Option<&str> = None;
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
        ]);
        /// `Some("Martin Benjamin")`
        pub const CONTACT: Option<&str> = Some("Martin Benjamin");
        /// `Some("2011-03-07")`
        pub const DATE: Option<&str> = Some("2011-03-07");
        /// `Some("locales@kamusi.org")`
        pub const EMAIL: Option<&str> = Some("locales@kamusi.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Swahili")`
        pub const LANGUAGE: Option<&str> = Some("Swahili");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Kamusi Project")`
        pub const SOURCE: Option<&str> = Some("Kamusi Project");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Kenya")`
        pub const TERRITORY: Option<&str> = Some("Kenya");
        /// `Some("Swahili locale for Kenya")`
        pub const TITLE: Option<&str> = Some("Swahili locale for Kenya");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0hHlL]"`
        pub const NOEXPR: &str = "^[-0hHlL]";
        /// `Some("Hapana")`
        pub const NOSTR: Option<&str> = Some("Hapana");
        /// `"^[+1nNyY]"`
        pub const YESEXPR: &str = "^[+1nNyY]";
        /// `Some("Ndiyo")`
        pub const YESSTR: Option<&str> = Some("Ndiyo");
    }
    pub use super::om_KE::LC_MONETARY;
    pub use super::om_KE::LC_NUMERIC;
    pub mod LC_TELEPHONE {
        /// `Some("254")`
        pub const INT_PREFIX: Option<&str> = Some("254");
        /// `Some("000")`
        pub const INT_SELECT: Option<&str> = Some("000");
        /// `Some("%A %l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("%A %l");
        /// `Some("%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["J2", "J3", "J4", "J5", "Alh", "Ij", "J1"]`
        pub const ABDAY: &[&str] = &["J2", "J3", "J4", "J5", "Alh", "Ij", "J1"];
        /// `&["Jan", "Feb", "Mac", "Apr", "Mei", "Jun", "Jul", "Ago", "Sep", "Okt", "Nov", "Des"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Mac", "Apr", "Mei", "Jun", "Jul", "Ago", "Sep", "Okt", "Nov", "Des"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["asubuhi", "alasiri"]`
        pub const AM_PM: &[&str] = &["asubuhi", "alasiri"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Jumapili", "Jumatatu", "Jumanne", "Jumatano", "Alhamisi", "Ijumaa", "Jumamosi"]`
        pub const DAY: &[&str] = &["Jumapili", "Jumatatu", "Jumanne", "Jumatano", "Alhamisi", "Ijumaa", "Jumamosi"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%e %B %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%e %B %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Januari", "Februari", "Machi", "Aprili", "Mei", "Juni", "Julai", "Agosti", "Septemba", "Oktoba", "Novemba", "Desemba"]`
        pub const MON: &[&str] = &["Januari", "Februari", "Machi", "Aprili", "Mei", "Juni", "Julai", "Agosti", "Septemba", "Oktoba", "Novemba", "Desemba"];
        /// `"%I:%M:%S %p"`
        pub const T_FMT: &str = "%I:%M:%S %p";
        /// `"%I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod sw_TZ {
    pub mod LC_ADDRESS {
        /// `Some("TZA")`
        pub const COUNTRY_AB3: Option<&str> = Some("TZA");
        /// `Some("TZ")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("TZ");
        /// `Some("EAT")`
        pub const COUNTRY_CAR: Option<&str> = Some("EAT");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Tanzania")`
        pub const COUNTRY_NAME: Option<&str> = Some("Tanzania");
        /// `Some(834)`
        pub const COUNTRY_NUM: Option<i64> = Some(834);
        /// `Some("TZ")`
        pub const COUNTRY_POST: Option<&str> = Some("TZ");
        /// `Some("sw")`
        pub const LANG_AB: Option<&str> = Some("sw");
        /// `Some("swa")`
        pub const LANG_LIB: Option<&str> = Some("swa");
        /// `Some("Kiswahili")`
        pub const LANG_NAME: Option<&str> = Some("Kiswahili");
        /// `Some("swa")`
        pub const LANG_TERM: Option<&str> = Some("swa");
        /// `Some("%a%N%f%N%d%N%b%N%h %s %e %r%N%T, %S %z%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%a%N%f%N%d%N%b%N%h %s %e %r%N%T, %S %z%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `None`
        pub const ADDRESS: Option<&str> = None;
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
        ]);
        /// `Some("Martin Benjamin")`
        pub const CONTACT: Option<&str> = Some("Martin Benjamin");
        /// `Some("2011-03-07")`
        pub const DATE: Option<&str> = Some("2011-03-07");
        /// `Some("locales@kamusi.org")`
        pub const EMAIL: Option<&str> = Some("locales@kamusi.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Swahili")`
        pub const LANGUAGE: Option<&str> = Some("Swahili");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Kamusi Project")`
        pub const SOURCE: Option<&str> = Some("Kamusi Project");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Tanzania")`
        pub const TERRITORY: Option<&str> = Some("Tanzania");
        /// `Some("Swahili locale for Tanzania")`
        pub const TITLE: Option<&str> = Some("Swahili locale for Tanzania");
    }
    pub use super::sw_KE::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"TSh"`
        pub const CURRENCY_SYMBOL: &str = "TSh";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"TZS "`
        pub const INT_CURR_SYMBOL: &str = "TZS ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub use super::sw_KE::LC_NUMERIC;
    pub mod LC_TELEPHONE {
        /// `Some("255")`
        pub const INT_PREFIX: Option<&str> = Some("255");
        /// `Some("000")`
        pub const INT_SELECT: Option<&str> = Some("000");
        /// `Some("%A %l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("%A %l");
        /// `Some("%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["J2", "J3", "J4", "J5", "Alh", "Ij", "J1"]`
        pub const ABDAY: &[&str] = &["J2", "J3", "J4", "J5", "Alh", "Ij", "J1"];
        /// `&["Jan", "Feb", "Mac", "Apr", "Mei", "Jun", "Jul", "Ago", "Sep", "Okt", "Nov", "Des"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Mac", "Apr", "Mei", "Jun", "Jul", "Ago", "Sep", "Okt", "Nov", "Des"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["asubuhi", "alasiri"]`
        pub const AM_PM: &[&str] = &["asubuhi", "alasiri"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Jumapili", "Jumatatu", "Jumanne", "Jumatano", "Alhamisi", "Ijumaa", "Jumamosi"]`
        pub const DAY: &[&str] = &["Jumapili", "Jumatatu", "Jumanne", "Jumatano", "Alhamisi", "Ijumaa", "Jumamosi"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%e %B %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%e %B %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Januari", "Februari", "Machi", "Aprili", "Mei", "Juni", "Julai", "Agosti", "Septemba", "Oktoba", "Novemba", "Desemba"]`
        pub const MON: &[&str] = &["Januari", "Februari", "Machi", "Aprili", "Mei", "Juni", "Julai", "Agosti", "Septemba", "Oktoba", "Novemba", "Desemba"];
        /// `"%I:%M:%S %p"`
        pub const T_FMT: &str = "%I:%M:%S %p";
        /// `"%I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod szl_PL {
    pub mod LC_ADDRESS {
        /// `Some("POL")`
        pub const COUNTRY_AB3: Option<&str> = Some("POL");
        /// `Some("PL")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("PL");
        /// `Some("PL")`
        pub const COUNTRY_CAR: Option<&str> = Some("PL");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Polska")`
        pub const COUNTRY_NAME: Option<&str> = Some("Polska");
        /// `Some(616)`
        pub const COUNTRY_NUM: Option<i64> = Some(616);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("szl")`
        pub const LANG_LIB: Option<&str> = Some("szl");
        /// `Some("ślōnskŏ gŏdka")`
        pub const LANG_NAME: Option<&str> = Some("ślōnskŏ gŏdka");
        /// `Some("szl")`
        pub const LANG_TERM: Option<&str> = Some("szl");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
        ]);
        /// `Some("Przemyslaw Buczkowski")`
        pub const CONTACT: Option<&str> = Some("Przemyslaw Buczkowski");
        /// `Some("2012-07-17")`
        pub const DATE: Option<&str> = Some("2012-07-17");
        /// `Some("libc-alpha@sourceware.org")`
        pub const EMAIL: Option<&str> = Some("libc-alpha@sourceware.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Silesian")`
        pub const LANGUAGE: Option<&str> = Some("Silesian");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("szl_PL locale")`
        pub const SOURCE: Option<&str> = Some("szl_PL locale");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Poland")`
        pub const TERRITORY: Option<&str> = Some("Poland");
        /// `Some("Silesian locale for Poland")`
        pub const TITLE: Option<&str> = Some("Silesian locale for Poland");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `None`
        pub const NOSTR: Option<&str> = None;
        /// `"^[+1JjTtYy]"`
        pub const YESEXPR: &str = "^[+1JjTtYy]";
        /// `None`
        pub const YESSTR: Option<&str> = None;
    }
    pub use super::pl_PL::LC_MONETARY;
    pub use super::pl_PL::LC_NUMERIC;
    pub use super::pl_PL::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["niy", "pyń", "wto", "str", "szt", "piō", "sob"]`
        pub const ABDAY: &[&str] = &["niy", "pyń", "wto", "str", "szt", "piō", "sob"];
        /// `&["sty", "lut", "mar", "kwi", "moj", "czy", "lip", "sie", "wrz", "paź", "lis", "gru"]`
        pub const ABMON: &[&str] = &["sty", "lut", "mar", "kwi", "moj", "czy", "lip", "sie", "wrz", "paź", "lis", "gru"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `Some(&["styczyń", "luty", "marzec", "kwieciyń", "moj", "czyrwiec", "lipiec", "sierpiyń", "wrzesiyń", "październik", "listopad", "grudziyń"])`
        pub const ALT_MON: Option<&[&str]> = Some(&["styczyń", "luty", "marzec", "kwieciyń", "moj", "czyrwiec", "lipiec", "sierpiyń", "wrzesiyń", "październik", "listopad", "grudziyń"]);
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a, %-d %b %Y, %T %Z")`
        pub const DATE_FMT: Option<&str> = Some("%a, %-d %b %Y, %T %Z");
        /// `&["niydziela", "pyńdziałek", "wtorek", "strzoda", "sztwortek", "piōntek", "sobota"]`
        pub const DAY: &[&str] = &["niydziela", "pyńdziałek", "wtorek", "strzoda", "sztwortek", "piōntek", "sobota"];
        /// `"%d.%m.%Y"`
        pub const D_FMT: &str = "%d.%m.%Y";
        /// `"%a, %-d %b %Y, %T"`
        pub const D_T_FMT: &str = "%a, %-d %b %Y, %T";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["stycznia", "lutego", "marca", "kwietnia", "moja", "czyrwca", "lipca", "sierpnia", "września", "października", "listopada", "grudnia"]`
        pub const MON: &[&str] = &["stycznia", "lutego", "marca", "kwietnia", "moja", "czyrwca", "lipca", "sierpnia", "września", "października", "listopada", "grudnia"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ta_IN {
    pub mod LC_ADDRESS {
        /// `Some("IND")`
        pub const COUNTRY_AB3: Option<&str> = Some("IND");
        /// `Some("IN")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IN");
        /// `Some("IND")`
        pub const COUNTRY_CAR: Option<&str> = Some("IND");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("இந\u{bcd}திய\u{bbe}")`
        pub const COUNTRY_NAME: Option<&str> = Some("இந\u{bcd}திய\u{bbe}");
        /// `Some(356)`
        pub const COUNTRY_NUM: Option<i64> = Some(356);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ta")`
        pub const LANG_AB: Option<&str> = Some("ta");
        /// `Some("tam")`
        pub const LANG_LIB: Option<&str> = Some("tam");
        /// `Some("தமிழ\u{bcd}")`
        pub const LANG_NAME: Option<&str> = Some("தமிழ\u{bcd}");
        /// `Some("tam")`
        pub const LANG_TERM: Option<&str> = Some("tam");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan")`
        pub const ADDRESS: Option<&str> = Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000,October,27 (XML source:2000,July,20)")`
        pub const DATE: Option<&str> = Some("2000,October,27 (XML source:2000,July,20)");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Tamil")`
        pub const LANGUAGE: Option<&str> = Some("Tamil");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("IBM Globalization Center of Competency, Yamato Software Laboratory")`
        pub const SOURCE: Option<&str> = Some("IBM Globalization Center of Competency, Yamato Software Laboratory");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("India")`
        pub const TERRITORY: Option<&str> = Some("India");
        /// `Some("Tamil language locale for India")`
        pub const TITLE: Option<&str> = Some("Tamil language locale for India");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNஇ]"`
        pub const NOEXPR: &str = "^[-0nNஇ]";
        /// `Some("இல\u{bcd}லை")`
        pub const NOSTR: Option<&str> = Some("இல\u{bcd}லை");
        /// `"^[+1yYஆ]"`
        pub const YESEXPR: &str = "^[+1yYஆ]";
        /// `Some("ஆம\u{bcd}")`
        pub const YESSTR: Option<&str> = Some("ஆம\u{bcd}");
    }
    pub use super::hi_IN::LC_MONETARY;
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3, 2]`
        pub const GROUPING: &[i64] = &[3, 2];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub use super::hi_IN::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["ஞ\u{bbe}", "தி", "செ", "பு", "வி", "வெ", "ச"]`
        pub const ABDAY: &[&str] = &["ஞ\u{bbe}", "தி", "செ", "பு", "வி", "வெ", "ச"];
        /// `&["ஜன.", "பிப\u{bcd}.", "ம\u{bbe}ர\u{bcd}.", "ஏப\u{bcd}.", "மே", "ஜூன\u{bcd}", "ஜூலை", "ஆக.", "செப\u{bcd}.", "அக\u{bcd}.", "நவ.", "டிச."]`
        pub const ABMON: &[&str] = &["ஜன.", "பிப\u{bcd}.", "ம\u{bbe}ர\u{bcd}.", "ஏப\u{bcd}.", "மே", "ஜூன\u{bcd}", "ஜூலை", "ஆக.", "செப\u{bcd}.", "அக\u{bcd}.", "நவ.", "டிச."];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["க\u{bbe}லை", "ம\u{bbe}லை"]`
        pub const AM_PM: &[&str] = &["க\u{bbe}லை", "ம\u{bbe}லை"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["ஞ\u{bbe}யிறு", "திங\u{bcd}கள\u{bcd}", "செவ\u{bcd}வ\u{bbe}ய\u{bcd}", "புதன\u{bcd}", "விய\u{bbe}ழன\u{bcd}", "வெள\u{bcd}ளி", "சனி"]`
        pub const DAY: &[&str] = &["ஞ\u{bbe}யிறு", "திங\u{bcd}கள\u{bcd}", "செவ\u{bcd}வ\u{bbe}ய\u{bcd}", "புதன\u{bcd}", "விய\u{bbe}ழன\u{bcd}", "வெள\u{bcd}ளி", "சனி"];
        /// `"%-d/%-m/%y"`
        pub const D_FMT: &str = "%-d/%-m/%y";
        /// `"%A %d %B %Y %p %I:%M:%S %Z"`
        pub const D_T_FMT: &str = "%A %d %B %Y %p %I:%M:%S %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["ஜனவரி", "பிப\u{bcd}ரவரி", "ம\u{bbe}ர\u{bcd}ச\u{bcd}", "ஏப\u{bcd}ரல\u{bcd}", "மே", "ஜூன\u{bcd}", "ஜூலை", "ஆகஸ\u{bcd}ட\u{bcd}", "செப\u{bcd}டம\u{bcd}பர\u{bcd}", "அக\u{bcd}டோபர\u{bcd}", "நவம\u{bcd}பர\u{bcd}", "டிசம\u{bcd}பர\u{bcd}"]`
        pub const MON: &[&str] = &["ஜனவரி", "பிப\u{bcd}ரவரி", "ம\u{bbe}ர\u{bcd}ச\u{bcd}", "ஏப\u{bcd}ரல\u{bcd}", "மே", "ஜூன\u{bcd}", "ஜூலை", "ஆகஸ\u{bcd}ட\u{bcd}", "செப\u{bcd}டம\u{bcd}பர\u{bcd}", "அக\u{bcd}டோபர\u{bcd}", "நவம\u{bcd}பர\u{bcd}", "டிசம\u{bcd}பர\u{bcd}"];
        /// `"%p %I:%M:%S %Z"`
        pub const T_FMT: &str = "%p %I:%M:%S %Z";
        /// `"%p %I:%M:%S %Z"`
        pub const T_FMT_AMPM: &str = "%p %I:%M:%S %Z";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ta_LK {
    pub mod LC_ADDRESS {
        /// `Some("LKA")`
        pub const COUNTRY_AB3: Option<&str> = Some("LKA");
        /// `Some("LK")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("LK");
        /// `Some("CL")`
        pub const COUNTRY_CAR: Option<&str> = Some("CL");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("இலங\u{bcd}கை")`
        pub const COUNTRY_NAME: Option<&str> = Some("இலங\u{bcd}கை");
        /// `Some(144)`
        pub const COUNTRY_NUM: Option<i64> = Some(144);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ta")`
        pub const LANG_AB: Option<&str> = Some("ta");
        /// `Some("tam")`
        pub const LANG_LIB: Option<&str> = Some("tam");
        /// `Some("தமிழ\u{bcd}")`
        pub const LANG_NAME: Option<&str> = Some("தமிழ\u{bcd}");
        /// `Some("tam")`
        pub const LANG_TERM: Option<&str> = Some("tam");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("30/36Q -2/1,Charles Apartments, De Silva Cross Rd, Kalubowila, Dehiwela, SriLanka.")`
        pub const ADDRESS: Option<&str> = Some("30/36Q -2/1,Charles Apartments, De Silva Cross Rd, Kalubowila, Dehiwela, SriLanka.");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("94-777-315206")`
        pub const CONTACT: Option<&str> = Some("94-777-315206");
        /// `Some("2011,August,11")`
        pub const DATE: Option<&str> = Some("2011,August,11");
        /// `Some("yogaraj.ubuntu@gmail.com")`
        pub const EMAIL: Option<&str> = Some("yogaraj.ubuntu@gmail.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Tamil")`
        pub const LANGUAGE: Option<&str> = Some("Tamil");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("J.Yogaraj")`
        pub const SOURCE: Option<&str> = Some("J.Yogaraj");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Sri Lanka")`
        pub const TERRITORY: Option<&str> = Some("Sri Lanka");
        /// `Some("Tamil language locale for Sri Lanka")`
        pub const TITLE: Option<&str> = Some("Tamil language locale for Sri Lanka");
    }
    pub use super::ta_IN::LC_MESSAGES;
    pub use super::si_LK::LC_MONETARY;
    pub use super::ta_IN::LC_NUMERIC;
    pub use super::si_LK::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["ஞ\u{bbe}", "தி", "செ", "பு", "வி", "வெ", "ச"]`
        pub const ABDAY: &[&str] = &["ஞ\u{bbe}", "தி", "செ", "பு", "வி", "வெ", "ச"];
        /// `&["ஜன", "பிப\u{bcd}", "ம\u{bbe}ர\u{bcd}", "ஏப\u{bcd}", "மே", "ஜூன\u{bcd}", "ஜூலை", "ஆக", "செப\u{bcd}", "அக\u{bcd}", "நவ", "டிச"]`
        pub const ABMON: &[&str] = &["ஜன", "பிப\u{bcd}", "ம\u{bbe}ர\u{bcd}", "ஏப\u{bcd}", "மே", "ஜூன\u{bcd}", "ஜூலை", "ஆக", "செப\u{bcd}", "அக\u{bcd}", "நவ", "டிச"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["க\u{bbe}லை", "ம\u{bbe}லை"]`
        pub const AM_PM: &[&str] = &["க\u{bbe}லை", "ம\u{bbe}லை"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["ஞ\u{bbe}யிறு", "திங\u{bcd}கள\u{bcd}", "செவ\u{bcd}வ\u{bbe}ய\u{bcd}", "புதன\u{bcd}", "விய\u{bbe}ழன\u{bcd}", "வெள\u{bcd}ளி", "சனி"]`
        pub const DAY: &[&str] = &["ஞ\u{bbe}யிறு", "திங\u{bcd}கள\u{bcd}", "செவ\u{bcd}வ\u{bbe}ய\u{bcd}", "புதன\u{bcd}", "விய\u{bbe}ழன\u{bcd}", "வெள\u{bcd}ளி", "சனி"];
        /// `"%-d/%-m/%y"`
        pub const D_FMT: &str = "%-d/%-m/%y";
        /// `"%A %d %B %Y %H:%M:%S %Z"`
        pub const D_T_FMT: &str = "%A %d %B %Y %H:%M:%S %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["ஜனவரி", "பிப\u{bcd}ரவரி", "ம\u{bbe}ர\u{bcd}ச\u{bcd}", "ஏப\u{bcd}ரல\u{bcd}", "மே", "ஜூன\u{bcd}", "ஜூலை", "ஆகஸ\u{bcd}ட\u{bcd}", "செப\u{bcd}டம\u{bcd}பர\u{bcd}", "அக\u{bcd}டோபர\u{bcd}", "நவம\u{bcd}பர\u{bcd}", "டிசம\u{bcd}பர\u{bcd}"]`
        pub const MON: &[&str] = &["ஜனவரி", "பிப\u{bcd}ரவரி", "ம\u{bbe}ர\u{bcd}ச\u{bcd}", "ஏப\u{bcd}ரல\u{bcd}", "மே", "ஜூன\u{bcd}", "ஜூலை", "ஆகஸ\u{bcd}ட\u{bcd}", "செப\u{bcd}டம\u{bcd}பர\u{bcd}", "அக\u{bcd}டோபர\u{bcd}", "நவம\u{bcd}பர\u{bcd}", "டிசம\u{bcd}பர\u{bcd}"];
        /// `"%H:%M:%S %Z"`
        pub const T_FMT: &str = "%H:%M:%S %Z";
        /// `"%l:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%l:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod tcy_IN {
    pub mod LC_ADDRESS {
        /// `Some("IND")`
        pub const COUNTRY_AB3: Option<&str> = Some("IND");
        /// `Some("IN")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IN");
        /// `Some("IND")`
        pub const COUNTRY_CAR: Option<&str> = Some("IND");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("ಭಾರತ")`
        pub const COUNTRY_NAME: Option<&str> = Some("ಭಾರತ");
        /// `Some(356)`
        pub const COUNTRY_NUM: Option<i64> = Some(356);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("")`
        pub const LANG_AB: Option<&str> = Some("");
        /// `Some("tcy")`
        pub const LANG_LIB: Option<&str> = Some("tcy");
        /// `Some("ತುಳು")`
        pub const LANG_NAME: Option<&str> = Some("ತುಳು");
        /// `Some("tcy")`
        pub const LANG_TERM: Option<&str> = Some("tcy");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2014-09-30")`
        pub const DATE: Option<&str> = Some("2014-09-30");
        /// `Some("alexey.merzlyakov@samsung.com")`
        pub const EMAIL: Option<&str> = Some("alexey.merzlyakov@samsung.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Tulu")`
        pub const LANGUAGE: Option<&str> = Some("Tulu");
        /// `Some("0.1")`
        pub const REVISION: Option<&str> = Some("0.1");
        /// `Some("IndLinux.org, Samsung Electronics Co., Ltd.")`
        pub const SOURCE: Option<&str> = Some("IndLinux.org, Samsung Electronics Co., Ltd.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("India")`
        pub const TERRITORY: Option<&str> = Some("India");
        /// `Some("Tulu language locale for India")`
        pub const TITLE: Option<&str> = Some("Tulu language locale for India");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `None`
        pub const NOSTR: Option<&str> = None;
        /// `"^[+1yY]"`
        pub const YESEXPR: &str = "^[+1yY]";
        /// `None`
        pub const YESSTR: Option<&str> = None;
    }
    pub use super::hi_IN::LC_MONETARY;
    pub use super::hi_IN::LC_NUMERIC;
    pub use super::hi_IN::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["ಐ", "ಸ\u{ccb}", "ಅಂ", "ಬು", "ಗು", "ಶು", "ಶ"]`
        pub const ABDAY: &[&str] = &["ಐ", "ಸ\u{ccb}", "ಅಂ", "ಬು", "ಗು", "ಶು", "ಶ"];
        /// `&["ಜ", "ಫ\u{cc6}", "ಮಾ", "ಏ", "ಮ\u{cc7}", "ಜ\u{cc2}", "ಜು", "ಆ", "ಸ\u{cc6}", "ಅ", "ನ", "ಡ\u{cbf}"]`
        pub const ABMON: &[&str] = &["ಜ", "ಫ\u{cc6}", "ಮಾ", "ಏ", "ಮ\u{cc7}", "ಜ\u{cc2}", "ಜು", "ಆ", "ಸ\u{cc6}", "ಅ", "ನ", "ಡ\u{cbf}"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["ಕಾಂಡ\u{cc6}", "ಬಯ\u{ccd}ಯ"]`
        pub const AM_PM: &[&str] = &["ಕಾಂಡ\u{cc6}", "ಬಯ\u{ccd}ಯ"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["ಐಥಾರ", "ಸ\u{ccb}ಮಾರ", "ಅಂಗರ\u{cc6}", "ಬುಧಾರ", "ಗುರುವಾರ", "ಶುಕ\u{ccd}ರರ", "ಶನ\u{cbf}ವಾರ"]`
        pub const DAY: &[&str] = &["ಐಥಾರ", "ಸ\u{ccb}ಮಾರ", "ಅಂಗರ\u{cc6}", "ಬುಧಾರ", "ಗುರುವಾರ", "ಶುಕ\u{ccd}ರರ", "ಶನ\u{cbf}ವಾರ"];
        /// `"%-d/%-m/%y"`
        pub const D_FMT: &str = "%-d/%-m/%y";
        /// `"%A %d %b %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A %d %b %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["ಜನವರ\u{cbf}", "ಫ\u{cc6}ಬ\u{ccd}ರುವರ\u{cbf}", "ಮಾರ\u{ccd}ಚ\u{ccd}", "ಏಪ\u{ccd}ರ\u{cbf}ಲ\u{ccd}\u{200c}\u{200c}", "ಮ\u{cc7}", "ಜ\u{cc2}ನ\u{ccd}", "ಜುಲ\u{cc8}", "ಆಗಸ\u{ccd}ಟ\u{ccd}", "ಸ\u{cc6}ಪ\u{ccd}ಟ\u{cc6}ಂಬರ\u{ccd}\u{200c}", "ಅಕ\u{ccd}ಟ\u{ccb}ಬರ\u{ccd}", "ನವ\u{cc6}ಂಬರ\u{ccd}", "ಡ\u{cbf}ಸ\u{cc6}ಂಬರ\u{ccd}"]`
        pub const MON: &[&str] = &["ಜನವರ\u{cbf}", "ಫ\u{cc6}ಬ\u{ccd}ರುವರ\u{cbf}", "ಮಾರ\u{ccd}ಚ\u{ccd}", "ಏಪ\u{ccd}ರ\u{cbf}ಲ\u{ccd}\u{200c}\u{200c}", "ಮ\u{cc7}", "ಜ\u{cc2}ನ\u{ccd}", "ಜುಲ\u{cc8}", "ಆಗಸ\u{ccd}ಟ\u{ccd}", "ಸ\u{cc6}ಪ\u{ccd}ಟ\u{cc6}ಂಬರ\u{ccd}\u{200c}", "ಅಕ\u{ccd}ಟ\u{ccb}ಬರ\u{ccd}", "ನವ\u{cc6}ಂಬರ\u{ccd}", "ಡ\u{cbf}ಸ\u{cc6}ಂಬರ\u{ccd}"];
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT: &str = "%I:%M:%S %p %Z";
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p %Z";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod te_IN {
    pub mod LC_ADDRESS {
        /// `Some("IND")`
        pub const COUNTRY_AB3: Option<&str> = Some("IND");
        /// `Some("IN")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IN");
        /// `Some("IND")`
        pub const COUNTRY_CAR: Option<&str> = Some("IND");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("భ\u{c3e}రతద\u{c47}శం")`
        pub const COUNTRY_NAME: Option<&str> = Some("భ\u{c3e}రతద\u{c47}శం");
        /// `Some(356)`
        pub const COUNTRY_NUM: Option<i64> = Some(356);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("te")`
        pub const LANG_AB: Option<&str> = Some("te");
        /// `Some("tel")`
        pub const LANG_LIB: Option<&str> = Some("tel");
        /// `Some("త\u{c46}లుగు")`
        pub const LANG_NAME: Option<&str> = Some("త\u{c46}లుగు");
        /// `Some("tel")`
        pub const LANG_TERM: Option<&str> = Some("tel");
        /// `Some("%a%N%d%N%f%N%r%t%e%N%h%t%b%N%s%t%N%T%t%z%N%S%N%c")`
        pub const POSTAL_FMT: Option<&str> = Some("%a%N%d%N%f%N%r%t%e%N%h%t%b%N%s%t%N%T%t%z%N%S%N%c");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan")`
        pub const ADDRESS: Option<&str> = Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2004-10-05")`
        pub const DATE: Option<&str> = Some("2004-10-05");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Telugu")`
        pub const LANGUAGE: Option<&str> = Some("Telugu");
        /// `Some("0.95")`
        pub const REVISION: Option<&str> = Some("0.95");
        /// `Some("IBM Globalization Center of Competency, Yamato Software Laboratory")`
        pub const SOURCE: Option<&str> = Some("IBM Globalization Center of Competency, Yamato Software Laboratory");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("India")`
        pub const TERRITORY: Option<&str> = Some("India");
        /// `Some("Telugu language locale for India")`
        pub const TITLE: Option<&str> = Some("Telugu language locale for India");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNవ]"`
        pub const NOEXPR: &str = "^[-0nNవ]";
        /// `Some("వద\u{c4d}దు")`
        pub const NOSTR: Option<&str> = Some("వద\u{c4d}దు");
        /// `"^[+1yYఅ]"`
        pub const YESEXPR: &str = "^[+1yYఅ]";
        /// `Some("అవును")`
        pub const YESSTR: Option<&str> = Some("అవును");
    }
    pub use super::hi_IN::LC_MONETARY;
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3, 2]`
        pub const GROUPING: &[i64] = &[3, 2];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub use super::hi_IN::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["ఆద\u{c3f}", "స\u{c4b}మ", "మంగళ", "బుధ", "గురు", "శుక\u{c4d}ర", "శన\u{c3f}"]`
        pub const ABDAY: &[&str] = &["ఆద\u{c3f}", "స\u{c4b}మ", "మంగళ", "బుధ", "గురు", "శుక\u{c4d}ర", "శన\u{c3f}"];
        /// `&["జన", "ఫ\u{c3f}బ\u{c4d}ర", "మ\u{c3e}ర\u{c4d}చ\u{c3f}", "ఏప\u{c4d}ర\u{c3f}", "మ\u{c47}", "జూన\u{c4d}", "జుల\u{c48}", "ఆగ", "స\u{c46}ప\u{c4d}ట\u{c46}ం", "అక\u{c4d}ట\u{c4b}", "నవం", "డ\u{c3f}స\u{c46}ం"]`
        pub const ABMON: &[&str] = &["జన", "ఫ\u{c3f}బ\u{c4d}ర", "మ\u{c3e}ర\u{c4d}చ\u{c3f}", "ఏప\u{c4d}ర\u{c3f}", "మ\u{c47}", "జూన\u{c4d}", "జుల\u{c48}", "ఆగ", "స\u{c46}ప\u{c4d}ట\u{c46}ం", "అక\u{c4d}ట\u{c4b}", "నవం", "డ\u{c3f}స\u{c46}ం"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["ఉ.", "స\u{c3e}."]`
        pub const AM_PM: &[&str] = &["ఉ.", "స\u{c3e}."];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["ఆద\u{c3f}వ\u{c3e}రం", "స\u{c4b}మవ\u{c3e}రం", "మంగళవ\u{c3e}రం", "బుధవ\u{c3e}రం", "గురువ\u{c3e}రం", "శుక\u{c4d}రవ\u{c3e}రం", "శన\u{c3f}వ\u{c3e}రం"]`
        pub const DAY: &[&str] = &["ఆద\u{c3f}వ\u{c3e}రం", "స\u{c4b}మవ\u{c3e}రం", "మంగళవ\u{c3e}రం", "బుధవ\u{c3e}రం", "గురువ\u{c3e}రం", "శుక\u{c4d}రవ\u{c3e}రం", "శన\u{c3f}వ\u{c3e}రం"];
        /// `"%d-%m-%y"`
        pub const D_FMT: &str = "%d-%m-%y";
        /// `"%B %d %A %Y %p%I.%M.%S %Z"`
        pub const D_T_FMT: &str = "%B %d %A %Y %p%I.%M.%S %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["జనవర\u{c3f}", "ఫ\u{c3f}బ\u{c4d}రవర\u{c3f}", "మ\u{c3e}ర\u{c4d}చ\u{c3f}", "ఏప\u{c4d}ర\u{c3f}ల\u{c4d}", "మ\u{c47}", "జూన\u{c4d}", "జుల\u{c48}", "ఆగస\u{c4d}టు", "స\u{c46}ప\u{c4d}ట\u{c46}ంబర\u{c4d}", "అక\u{c4d}ట\u{c4b}బర\u{c4d}", "నవంబర\u{c4d}", "డ\u{c3f}స\u{c46}ంబర\u{c4d}"]`
        pub const MON: &[&str] = &["జనవర\u{c3f}", "ఫ\u{c3f}బ\u{c4d}రవర\u{c3f}", "మ\u{c3e}ర\u{c4d}చ\u{c3f}", "ఏప\u{c4d}ర\u{c3f}ల\u{c4d}", "మ\u{c47}", "జూన\u{c4d}", "జుల\u{c48}", "ఆగస\u{c4d}టు", "స\u{c46}ప\u{c4d}ట\u{c46}ంబర\u{c4d}", "అక\u{c4d}ట\u{c4b}బర\u{c4d}", "నవంబర\u{c4d}", "డ\u{c3f}స\u{c46}ంబర\u{c4d}"];
        /// `"%p%I.%M.%S %Z"`
        pub const T_FMT: &str = "%p%I.%M.%S %Z";
        /// `"%p%I.%M.%S %Z"`
        pub const T_FMT_AMPM: &str = "%p%I.%M.%S %Z";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod tg_TJ {
    pub mod LC_ADDRESS {
        /// `Some("TJK")`
        pub const COUNTRY_AB3: Option<&str> = Some("TJK");
        /// `Some("TJ")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("TJ");
        /// `Some("TJ")`
        pub const COUNTRY_CAR: Option<&str> = Some("TJ");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Тоҷикистон")`
        pub const COUNTRY_NAME: Option<&str> = Some("Тоҷикистон");
        /// `Some(762)`
        pub const COUNTRY_NUM: Option<i64> = Some(762);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("tg")`
        pub const LANG_AB: Option<&str> = Some("tg");
        /// `Some("tgk")`
        pub const LANG_LIB: Option<&str> = Some("tgk");
        /// `Some("тоҷикӣ")`
        pub const LANG_NAME: Option<&str> = Some("тоҷикӣ");
        /// `Some("tgk")`
        pub const LANG_TERM: Option<&str> = Some("tgk");
        /// `Some("%a%N%f%N%d%N%b%N%h %s %e %r%N%T %z%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%a%N%f%N%d%N%b%N%h %s %e %r%N%T %z%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
        ]);
        /// `Some("Pablo Saratxaga, Roger Kovacs")`
        pub const CONTACT: Option<&str> = Some("Pablo Saratxaga, Roger Kovacs");
        /// `Some("2004-01-09")`
        pub const DATE: Option<&str> = Some("2004-01-09");
        /// `Some("pablo@mandrakesoft.com, ROGERKO@micromotion.com")`
        pub const EMAIL: Option<&str> = Some("pablo@mandrakesoft.com, ROGERKO@micromotion.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Tajik")`
        pub const LANGUAGE: Option<&str> = Some("Tajik");
        /// `Some("0.4")`
        pub const REVISION: Option<&str> = Some("0.4");
        /// `Some("Roger Kovacs")`
        pub const SOURCE: Option<&str> = Some("Roger Kovacs");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Tajikistan")`
        pub const TERRITORY: Option<&str> = Some("Tajikistan");
        /// `Some("Tajik language locale for Tajikistan")`
        pub const TITLE: Option<&str> = Some("Tajik language locale for Tajikistan");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNНн]"`
        pub const NOEXPR: &str = "^[-0nNНн]";
        /// `Some("не")`
        pub const NOSTR: Option<&str> = Some("не");
        /// `"^[+1yYҲҳХхДд]"`
        pub const YESEXPR: &str = "^[+1yYҲҳХхДд]";
        /// `Some("ҳа")`
        pub const YESSTR: Option<&str> = Some("ҳа");
    }
    pub mod LC_MONETARY {
        /// `"руб"`
        pub const CURRENCY_SYMBOL: &str = "руб";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"TJS "`
        pub const INT_CURR_SYMBOL: &str = "TJS ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const MON_THOUSANDS_SEP: &str = "\u{202f}";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const THOUSANDS_SEP: &str = ".";
    }
    pub mod LC_TELEPHONE {
        /// `Some("992")`
        pub const INT_PREFIX: Option<&str> = Some("992");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a%t%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a%t%l");
    }
    pub mod LC_TIME {
        /// `&["Вск", "Пнд", "Втр", "Срд", "Чтв", "Птн", "Сбт"]`
        pub const ABDAY: &[&str] = &["Вск", "Пнд", "Втр", "Срд", "Чтв", "Птн", "Сбт"];
        /// `&["Янв", "Фев", "Мар", "Апр", "Май", "Июн", "Июл", "Авг", "Сен", "Окт", "Ноя", "Дек"]`
        pub const ABMON: &[&str] = &["Янв", "Фев", "Мар", "Апр", "Май", "Июн", "Июл", "Авг", "Сен", "Окт", "Ноя", "Дек"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Воскресенье", "Понедельник", "Вторник", "Среда", "Четверг", "Пятница", "Суббота"]`
        pub const DAY: &[&str] = &["Воскресенье", "Понедельник", "Вторник", "Среда", "Четверг", "Пятница", "Суббота"];
        /// `"%d.%m.%Y"`
        pub const D_FMT: &str = "%d.%m.%Y";
        /// `"%a %d %b %Y %T"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Января", "Февраля", "Марта", "Апреля", "Мая", "Июня", "Июля", "Августа", "Сентября", "Октября", "Ноября", "Декабря"]`
        pub const MON: &[&str] = &["Января", "Февраля", "Марта", "Апреля", "Мая", "Июня", "Июля", "Августа", "Сентября", "Октября", "Ноября", "Декабря"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod th_TH {
    pub mod LC_ADDRESS {
        /// `Some("THA")`
        pub const COUNTRY_AB3: Option<&str> = Some("THA");
        /// `Some("TH")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("TH");
        /// `Some("T")`
        pub const COUNTRY_CAR: Option<&str> = Some("T");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("ไทย")`
        pub const COUNTRY_NAME: Option<&str> = Some("ไทย");
        /// `Some(764)`
        pub const COUNTRY_NUM: Option<i64> = Some(764);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("th")`
        pub const LANG_AB: Option<&str> = Some("th");
        /// `Some("tha")`
        pub const LANG_LIB: Option<&str> = Some("tha");
        /// `Some("ไทย")`
        pub const LANG_NAME: Option<&str> = Some("ไทย");
        /// `Some("tha")`
        pub const LANG_TERM: Option<&str> = Some("tha");
        /// `Some("%f%N%a%N%d%N%r%t%e%t%b%N%h%t%s%N%T%N%S%N%z%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%r%t%e%t%b%N%h%t%s%N%T%N%S%N%z%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-06-29")`
        pub const DATE: Option<&str> = Some("2000-06-29");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Thai")`
        pub const LANGUAGE: Option<&str> = Some("Thai");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Thailand")`
        pub const TERRITORY: Option<&str> = Some("Thailand");
        /// `Some("Thai locale for Thailand")`
        pub const TITLE: Option<&str> = Some("Thai locale for Thailand");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNม]"`
        pub const NOEXPR: &str = "^[-0nNม]";
        /// `Some("ไม\u{e48}ใช\u{e48}")`
        pub const NOSTR: Option<&str> = Some("ไม\u{e48}ใช\u{e48}");
        /// `"^[+1yYช]"`
        pub const YESEXPR: &str = "^[+1yYช]";
        /// `Some("ใช\u{e48}")`
        pub const YESSTR: Option<&str> = Some("ใช\u{e48}");
    }
    pub mod LC_MONETARY {
        /// `"฿"`
        pub const CURRENCY_SYMBOL: &str = "฿";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"THB "`
        pub const INT_CURR_SYMBOL: &str = "THB ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `2`
        pub const N_SEP_BY_SPACE: i64 = 2;
        /// `4`
        pub const N_SIGN_POSN: i64 = 4;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `2`
        pub const P_SEP_BY_SPACE: i64 = 2;
        /// `4`
        pub const P_SIGN_POSN: i64 = 4;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("66")`
        pub const INT_PREFIX: Option<&str> = Some("66");
        /// `Some("001")`
        pub const INT_SELECT: Option<&str> = Some("001");
        /// `Some("0-%a%l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("0-%a%l");
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["อา.", "จ.", "อ.", "พ.", "พฤ.", "ศ.", "ส."]`
        pub const ABDAY: &[&str] = &["อา.", "จ.", "อ.", "พ.", "พฤ.", "ศ.", "ส."];
        /// `&["ม.ค.", "ก.พ.", "ม\u{e35}.ค.", "เม.ย.", "พ.ค.", "ม\u{e34}.ย.", "ก.ค.", "ส.ค.", "ก.ย.", "ต.ค.", "พ.ย.", "ธ.ค."]`
        pub const ABMON: &[&str] = &["ม.ค.", "ก.พ.", "ม\u{e35}.ค.", "เม.ย.", "พ.ค.", "ม\u{e34}.ย.", "ก.ค.", "ส.ค.", "ก.ย.", "ต.ค.", "พ.ย.", "ธ.ค."];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["AM", "PM"]`
        pub const AM_PM: &[&str] = &["AM", "PM"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a %e %b %Ey %H:%M:%S %Z")`
        pub const DATE_FMT: Option<&str> = Some("%a %e %b %Ey %H:%M:%S %Z");
        /// `&["อาท\u{e34}ตย\u{e4c}", "จ\u{e31}นทร\u{e4c}", "อ\u{e31}งคาร", "พ\u{e38}ธ", "พฤห\u{e31}สบด\u{e35}", "ศ\u{e38}กร\u{e4c}", "เสาร\u{e4c}"]`
        pub const DAY: &[&str] = &["อาท\u{e34}ตย\u{e4c}", "จ\u{e31}นทร\u{e4c}", "อ\u{e31}งคาร", "พ\u{e38}ธ", "พฤห\u{e31}สบด\u{e35}", "ศ\u{e38}กร\u{e4c}", "เสาร\u{e4c}"];
        /// `"%d/%m/%Ey"`
        pub const D_FMT: &str = "%d/%m/%Ey";
        /// `"%a %e %b %Ey, %H:%M:%S"`
        pub const D_T_FMT: &str = "%a %e %b %Ey, %H:%M:%S";
        /// `Some(&["+:1:-543/01/01:+*:พ.ศ.:%EC %Ey"])`
        pub const ERA: Option<&[&str]> = Some(&["+:1:-543/01/01:+*:พ.ศ.:%EC %Ey"]);
        /// `Some("%e %b %Ey")`
        pub const ERA_D_FMT: Option<&str> = Some("%e %b %Ey");
        /// `Some("ว\u{e31}น%Aท\u{e35}\u{e48} %e %B %EC %Ey, %H.%M.%S น.")`
        pub const ERA_D_T_FMT: Option<&str> = Some("ว\u{e31}น%Aท\u{e35}\u{e48} %e %B %EC %Ey, %H.%M.%S น.");
        /// `Some("%H.%M.%S น.")`
        pub const ERA_T_FMT: Option<&str> = Some("%H.%M.%S น.");
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["มกราคม", "ก\u{e38}มภาพ\u{e31}นธ\u{e4c}", "ม\u{e35}นาคม", "เมษายน", "พฤษภาคม", "ม\u{e34}ถ\u{e38}นายน", "กรกฎาคม", "ส\u{e34}งหาคม", "ก\u{e31}นยายน", "ต\u{e38}ลาคม", "พฤศจ\u{e34}กายน", "ธ\u{e31}นวาคม"]`
        pub const MON: &[&str] = &["มกราคม", "ก\u{e38}มภาพ\u{e31}นธ\u{e4c}", "ม\u{e35}นาคม", "เมษายน", "พฤษภาคม", "ม\u{e34}ถ\u{e38}นายน", "กรกฎาคม", "ส\u{e34}งหาคม", "ก\u{e31}นยายน", "ต\u{e38}ลาคม", "พฤศจ\u{e34}กายน", "ธ\u{e31}นวาคม"];
        /// `"%H:%M:%S"`
        pub const T_FMT: &str = "%H:%M:%S";
        /// `"%I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod the_NP {
    pub mod LC_ADDRESS {
        /// `Some("NPL")`
        pub const COUNTRY_AB3: Option<&str> = Some("NPL");
        /// `Some("NP")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("NP");
        /// `Some("NEP")`
        pub const COUNTRY_CAR: Option<&str> = Some("NEP");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("न\u{947}पाल")`
        pub const COUNTRY_NAME: Option<&str> = Some("न\u{947}पाल");
        /// `Some(524)`
        pub const COUNTRY_NUM: Option<i64> = Some(524);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("the")`
        pub const LANG_LIB: Option<&str> = Some("the");
        /// `Some("थार\u{941}")`
        pub const LANG_NAME: Option<&str> = Some("थार\u{941}");
        /// `Some("the")`
        pub const LANG_TERM: Option<&str> = Some("the");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sangam Market, narayanghat, Nepal")`
        pub const ADDRESS: Option<&str> = Some("Sangam Market, narayanghat, Nepal");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2013-11-15")`
        pub const DATE: Option<&str> = Some("2013-11-15");
        /// `Some("info@chitwanix.com")`
        pub const EMAIL: Option<&str> = Some("info@chitwanix.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Chitwania Tharu")`
        pub const LANGUAGE: Option<&str> = Some("Chitwania Tharu");
        /// `Some("1.1")`
        pub const REVISION: Option<&str> = Some("1.1");
        /// `Some("Chitwanix OS Development")`
        pub const SOURCE: Option<&str> = Some("Chitwanix OS Development");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Nepal")`
        pub const TERRITORY: Option<&str> = Some("Nepal");
        /// `Some("Tharu language locale for Nepal")`
        pub const TITLE: Option<&str> = Some("Tharu language locale for Nepal");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `None`
        pub const NOSTR: Option<&str> = None;
        /// `"^[+1yY]"`
        pub const YESEXPR: &str = "^[+1yY]";
        /// `None`
        pub const YESSTR: Option<&str> = None;
    }
    pub use super::ne_NP::LC_MONETARY;
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub use super::ne_NP::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["आइत", "सोम", "म\u{902}गल", "ब\u{941}ध", "बिहि", "श\u{941}क\u{94d}र", "शनि"]`
        pub const ABDAY: &[&str] = &["आइत", "सोम", "म\u{902}गल", "ब\u{941}ध", "बिहि", "श\u{941}क\u{94d}र", "शनि"];
        /// `&["जनवरी", "फ\u{93c}रवरी", "मार\u{94d}च", "अप\u{94d}र\u{947}ल", "मई", "ज\u{942}न", "ज\u{941}लाई", "अगस\u{94d}त", "सितम\u{94d}बर", "अक\u{94d}ट\u{942}बर", "नवम\u{94d}बर", "दिसम\u{94d}बर"]`
        pub const ABMON: &[&str] = &["जनवरी", "फ\u{93c}रवरी", "मार\u{94d}च", "अप\u{94d}र\u{947}ल", "मई", "ज\u{942}न", "ज\u{941}लाई", "अगस\u{94d}त", "सितम\u{94d}बर", "अक\u{94d}ट\u{942}बर", "नवम\u{94d}बर", "दिसम\u{94d}बर"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["प\u{942}र\u{94d}वाह\u{94d}न", "अपराह\u{94d}न"]`
        pub const AM_PM: &[&str] = &["प\u{942}र\u{94d}वाह\u{94d}न", "अपराह\u{94d}न"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["आइतबार", "सोमबार", "म\u{902}गलबार", "ब\u{941}धबार", "बिहिबार", "श\u{941}क\u{94d}रबार", "शनिबार"]`
        pub const DAY: &[&str] = &["आइतबार", "सोमबार", "म\u{902}गलबार", "ब\u{941}धबार", "बिहिबार", "श\u{941}क\u{94d}रबार", "शनिबार"];
        /// `"%A %d %b %Y"`
        pub const D_FMT: &str = "%A %d %b %Y";
        /// `"%A %d %b %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A %d %b %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["जनवरी", "फ\u{93c}रवरी", "मार\u{94d}च", "अप\u{94d}र\u{947}ल", "मई", "ज\u{942}न", "ज\u{941}लाई", "अगस\u{94d}त", "सितम\u{94d}बर", "अक\u{94d}ट\u{942}बर", "नवम\u{94d}बर", "दिसम\u{94d}बर"]`
        pub const MON: &[&str] = &["जनवरी", "फ\u{93c}रवरी", "मार\u{94d}च", "अप\u{94d}र\u{947}ल", "मई", "ज\u{942}न", "ज\u{941}लाई", "अगस\u{94d}त", "सितम\u{94d}बर", "अक\u{94d}ट\u{942}बर", "नवम\u{94d}बर", "दिसम\u{94d}बर"];
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT: &str = "%I:%M:%S %p %Z";
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p %Z";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ti_ER {
    pub mod LC_ADDRESS {
        /// `Some("ERI")`
        pub const COUNTRY_AB3: Option<&str> = Some("ERI");
        /// `Some("ER")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("ER");
        /// `Some("ER")`
        pub const COUNTRY_CAR: Option<&str> = Some("ER");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("ኤርትራ")`
        pub const COUNTRY_NAME: Option<&str> = Some("ኤርትራ");
        /// `Some(232)`
        pub const COUNTRY_NUM: Option<i64> = Some(232);
        /// `Some("ERI")`
        pub const COUNTRY_POST: Option<&str> = Some("ERI");
        /// `Some("ti")`
        pub const LANG_AB: Option<&str> = Some("ti");
        /// `Some("tir")`
        pub const LANG_LIB: Option<&str> = Some("tir");
        /// `Some("ትግርኛ")`
        pub const LANG_NAME: Option<&str> = Some("ትግርኛ");
        /// `Some("tir")`
        pub const LANG_TERM: Option<&str> = Some("tir");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("7802 Solomon Seal Dr., Springfield, VA 22152, USA")`
        pub const ADDRESS: Option<&str> = Some("7802 Solomon Seal Dr., Springfield, VA 22152, USA");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_TIME"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_TIME"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2003-07-05")`
        pub const DATE: Option<&str> = Some("2003-07-05");
        /// `Some("locales@geez.org")`
        pub const EMAIL: Option<&str> = Some("locales@geez.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Tigrinya")`
        pub const LANGUAGE: Option<&str> = Some("Tigrinya");
        /// `Some("0.20")`
        pub const REVISION: Option<&str> = Some("0.20");
        /// `Some("Ge'ez Frontier Foundation")`
        pub const SOURCE: Option<&str> = Some("Ge'ez Frontier Foundation");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Eritrea")`
        pub const TERRITORY: Option<&str> = Some("Eritrea");
        /// `Some("Tigrigna language locale for Eritrea.")`
        pub const TITLE: Option<&str> = Some("Tigrigna language locale for Eritrea.");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNነ]"`
        pub const NOEXPR: &str = "^[-0nNነ]";
        /// `Some("ነኖእ")`
        pub const NOSTR: Option<&str> = Some("ነኖእ");
        /// `"^[+1yYእ]"`
        pub const YESEXPR: &str = "^[+1yYእ]";
        /// `Some("እወ")`
        pub const YESSTR: Option<&str> = Some("እወ");
    }
    pub mod LC_MONETARY {
        /// `"Nfk"`
        pub const CURRENCY_SYMBOL: &str = "Nfk";
        /// `0`
        pub const FRAC_DIGITS: i64 = 0;
        /// `"ERN "`
        pub const INT_CURR_SYMBOL: &str = "ERN ";
        /// `0`
        pub const INT_FRAC_DIGITS: i64 = 0;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[0, 0]`
        pub const GROUPING: &[i64] = &[0, 0];
        /// `""`
        pub const THOUSANDS_SEP: &str = "";
    }
    pub mod LC_TELEPHONE {
        /// `Some("291")`
        pub const INT_PREFIX: Option<&str> = Some("291");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `Some("%a-%l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("%a-%l");
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["ሰንበ", "ሰኑይ", "ሰሉስ", "ረቡዕ", "ሓሙስ", "ዓርቢ", "ቀዳም"]`
        pub const ABDAY: &[&str] = &["ሰንበ", "ሰኑይ", "ሰሉስ", "ረቡዕ", "ሓሙስ", "ዓርቢ", "ቀዳም"];
        /// `&["ጥሪ ", "ለካቲ", "መጋቢ", "ሚያዝ", "ግንቦ", "ሰነ ", "ሓምለ", "ነሓሰ", "መስከ", "ጥቅም", "ሕዳር", "ታሕሳ"]`
        pub const ABMON: &[&str] = &["ጥሪ ", "ለካቲ", "መጋቢ", "ሚያዝ", "ግንቦ", "ሰነ ", "ሓምለ", "ነሓሰ", "መስከ", "ጥቅም", "ሕዳር", "ታሕሳ"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["ንጉሆ ሰዓተ", "ድሕር ሰዓት"]`
        pub const AM_PM: &[&str] = &["ንጉሆ ሰዓተ", "ድሕር ሰዓት"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%A፡ %B %e መዓልቲ %r %Z %Y ዓ/ም")`
        pub const DATE_FMT: Option<&str> = Some("%A፡ %B %e መዓልቲ %r %Z %Y ዓ/ም");
        /// `&["ሰንበት", "ሰኑይ", "ሰሉስ", "ረቡዕ", "ሓሙስ", "ዓርቢ", "ቀዳም"]`
        pub const DAY: &[&str] = &["ሰንበት", "ሰኑይ", "ሰሉስ", "ረቡዕ", "ሓሙስ", "ዓርቢ", "ቀዳም"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%A፡ %B %e መዓልቲ %Y %l:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A፡ %B %e መዓልቲ %Y %l:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["ጥሪ", "ለካቲት", "መጋቢት", "ሚያዝያ", "ግንቦት", "ሰነ", "ሓምለ", "ነሓሰ", "መስከረም", "ጥቅምቲ", "ሕዳር", "ታሕሳስ"]`
        pub const MON: &[&str] = &["ጥሪ", "ለካቲት", "መጋቢት", "ሚያዝያ", "ግንቦት", "ሰነ", "ሓምለ", "ነሓሰ", "መስከረም", "ጥቅምቲ", "ሕዳር", "ታሕሳስ"];
        /// `"%l:%M:%S %p"`
        pub const T_FMT: &str = "%l:%M:%S %p";
        /// `"%l:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%l:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ti_ET {
    pub mod LC_ADDRESS {
        /// `Some("ETH")`
        pub const COUNTRY_AB3: Option<&str> = Some("ETH");
        /// `Some("ET")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("ET");
        /// `Some("ETH")`
        pub const COUNTRY_CAR: Option<&str> = Some("ETH");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("ኢትዮጵያ")`
        pub const COUNTRY_NAME: Option<&str> = Some("ኢትዮጵያ");
        /// `Some(231)`
        pub const COUNTRY_NUM: Option<i64> = Some(231);
        /// `Some("ETH")`
        pub const COUNTRY_POST: Option<&str> = Some("ETH");
        /// `Some("ti")`
        pub const LANG_AB: Option<&str> = Some("ti");
        /// `Some("tir")`
        pub const LANG_LIB: Option<&str> = Some("tir");
        /// `Some("ትግርኛ")`
        pub const LANG_NAME: Option<&str> = Some("ትግርኛ");
        /// `Some("tir")`
        pub const LANG_TERM: Option<&str> = Some("tir");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("7802 Solomon Seal Dr., Springfield, VA 22152, USA")`
        pub const ADDRESS: Option<&str> = Some("7802 Solomon Seal Dr., Springfield, VA 22152, USA");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_TIME"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_TIME"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2003-07-05")`
        pub const DATE: Option<&str> = Some("2003-07-05");
        /// `Some("locales@geez.org")`
        pub const EMAIL: Option<&str> = Some("locales@geez.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Tigrinya")`
        pub const LANGUAGE: Option<&str> = Some("Tigrinya");
        /// `Some("0.20")`
        pub const REVISION: Option<&str> = Some("0.20");
        /// `Some("Ge'ez Frontier Foundation")`
        pub const SOURCE: Option<&str> = Some("Ge'ez Frontier Foundation");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Ethiopia")`
        pub const TERRITORY: Option<&str> = Some("Ethiopia");
        /// `Some("Tigrigna language locale for Ethiopia.")`
        pub const TITLE: Option<&str> = Some("Tigrigna language locale for Ethiopia.");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0ነnN]"`
        pub const NOEXPR: &str = "^[-0ነnN]";
        /// `Some("ነኖእ")`
        pub const NOSTR: Option<&str> = Some("ነኖእ");
        /// `"^[+1እyY]"`
        pub const YESEXPR: &str = "^[+1እyY]";
        /// `Some("እወ")`
        pub const YESSTR: Option<&str> = Some("እወ");
    }
    pub mod LC_MONETARY {
        /// `"Br"`
        pub const CURRENCY_SYMBOL: &str = "Br";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"ETB "`
        pub const INT_CURR_SYMBOL: &str = "ETB ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("251")`
        pub const INT_PREFIX: Option<&str> = Some("251");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `Some("%a-%l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("%a-%l");
        /// `Some("%c-%a-%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("%c-%a-%l");
    }
    pub mod LC_TIME {
        /// `&["ሰንበ", "ሰኑይ", "ሰሉስ", "ረቡዕ", "ሓሙስ", "ዓርቢ", "ቀዳም"]`
        pub const ABDAY: &[&str] = &["ሰንበ", "ሰኑይ", "ሰሉስ", "ረቡዕ", "ሓሙስ", "ዓርቢ", "ቀዳም"];
        /// `&["ጃንዩ", "ፌብሩ", "ማርች", "ኤፕረ", "ሜይ ", "ጁን ", "ጁላይ", "ኦገስ", "ሴፕቴ", "ኦክተ", "ኖቬም", "ዲሴም"]`
        pub const ABMON: &[&str] = &["ጃንዩ", "ፌብሩ", "ማርች", "ኤፕረ", "ሜይ ", "ጁን ", "ጁላይ", "ኦገስ", "ሴፕቴ", "ኦክተ", "ኖቬም", "ዲሴም"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["ንጉሆ ሰዓተ", "ድሕር ሰዓት"]`
        pub const AM_PM: &[&str] = &["ንጉሆ ሰዓተ", "ድሕር ሰዓት"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%A፣ %B %e መዓልቲ %r %Z %Y ዓ/ም")`
        pub const DATE_FMT: Option<&str> = Some("%A፣ %B %e መዓልቲ %r %Z %Y ዓ/ም");
        /// `&["ሰንበት", "ሰኑይ", "ሰሉስ", "ረቡዕ", "ሓሙስ", "ዓርቢ", "ቀዳም"]`
        pub const DAY: &[&str] = &["ሰንበት", "ሰኑይ", "ሰሉስ", "ረቡዕ", "ሓሙስ", "ዓርቢ", "ቀዳም"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%A፣ %B %e መዓልቲ %Y %l:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A፣ %B %e መዓልቲ %Y %l:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["ጃንዩወሪ", "ፌብሩወሪ", "ማርች", "ኤፕረል", "ሜይ", "ጁን", "ጁላይ", "ኦገስት", "ሴፕቴምበር", "ኦክተውበር", "ኖቬምበር", "ዲሴምበር"]`
        pub const MON: &[&str] = &["ጃንዩወሪ", "ፌብሩወሪ", "ማርች", "ኤፕረል", "ሜይ", "ጁን", "ጁላይ", "ኦገስት", "ሴፕቴምበር", "ኦክተውበር", "ኖቬምበር", "ዲሴምበር"];
        /// `"%l:%M:%S %p"`
        pub const T_FMT: &str = "%l:%M:%S %p";
        /// `"%l:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%l:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod tig_ER {
    pub mod LC_ADDRESS {
        /// `Some("ERI")`
        pub const COUNTRY_AB3: Option<&str> = Some("ERI");
        /// `Some("ER")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("ER");
        /// `Some("ER")`
        pub const COUNTRY_CAR: Option<&str> = Some("ER");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("ኤርትራ")`
        pub const COUNTRY_NAME: Option<&str> = Some("ኤርትራ");
        /// `Some(232)`
        pub const COUNTRY_NUM: Option<i64> = Some(232);
        /// `Some("ERI")`
        pub const COUNTRY_POST: Option<&str> = Some("ERI");
        /// `Some("")`
        pub const LANG_AB: Option<&str> = Some("");
        /// `Some("tig")`
        pub const LANG_LIB: Option<&str> = Some("tig");
        /// `Some("ትግረ")`
        pub const LANG_NAME: Option<&str> = Some("ትግረ");
        /// `Some("tig")`
        pub const LANG_TERM: Option<&str> = Some("tig");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("7802 Solomon Seal Dr., Springfield, VA 22152, USA")`
        pub const ADDRESS: Option<&str> = Some("7802 Solomon Seal Dr., Springfield, VA 22152, USA");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_TIME"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_TIME"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2003-07-05")`
        pub const DATE: Option<&str> = Some("2003-07-05");
        /// `Some("locales@geez.org")`
        pub const EMAIL: Option<&str> = Some("locales@geez.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Tigre")`
        pub const LANGUAGE: Option<&str> = Some("Tigre");
        /// `Some("0.20")`
        pub const REVISION: Option<&str> = Some("0.20");
        /// `Some("Ge'ez Frontier Foundation")`
        pub const SOURCE: Option<&str> = Some("Ge'ez Frontier Foundation");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Eritrea")`
        pub const TERRITORY: Option<&str> = Some("Eritrea");
        /// `Some("Tigre language locale for Eritrea")`
        pub const TITLE: Option<&str> = Some("Tigre language locale for Eritrea");
    }
    pub use super::ti_ER::LC_MESSAGES;
    pub use super::ti_ER::LC_MONETARY;
    pub use super::ti_ER::LC_NUMERIC;
    pub use super::ti_ER::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["ሰ/ዓ", "ሰኖ ", "ታላሸ", "ኣረር", "ከሚሽ", "ጅምዓ", "ሰ/ን"]`
        pub const ABDAY: &[&str] = &["ሰ/ዓ", "ሰኖ ", "ታላሸ", "ኣረር", "ከሚሽ", "ጅምዓ", "ሰ/ን"];
        /// `&["ጥሪ ", "ለካቲ", "መጋቢ", "ሚያዝ", "ግንቦ", "ሰነ ", "ሓምለ", "ነሓሰ", "መስከ", "ጥቅም", "ሕዳር", "ታሕሳ"]`
        pub const ABMON: &[&str] = &["ጥሪ ", "ለካቲ", "መጋቢ", "ሚያዝ", "ግንቦ", "ሰነ ", "ሓምለ", "ነሓሰ", "መስከ", "ጥቅም", "ሕዳር", "ታሕሳ"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["ቀደም ሰር ምዕል", "ሓቆ ሰር ምዕል"]`
        pub const AM_PM: &[&str] = &["ቀደም ሰር ምዕል", "ሓቆ ሰር ምዕል"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%A፡ %B %e ዮም %r %Z %Y ዓ/ም")`
        pub const DATE_FMT: Option<&str> = Some("%A፡ %B %e ዮም %r %Z %Y ዓ/ም");
        /// `&["ሰንበት ዓባይ", "ሰኖ", "ታላሸኖ", "ኣረርባዓ", "ከሚሽ", "ጅምዓት", "ሰንበት ንኢሽ"]`
        pub const DAY: &[&str] = &["ሰንበት ዓባይ", "ሰኖ", "ታላሸኖ", "ኣረርባዓ", "ከሚሽ", "ጅምዓት", "ሰንበት ንኢሽ"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%A፡ %B %e ዮም %Y %l:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A፡ %B %e ዮም %Y %l:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["ጥሪ", "ለካቲት", "መጋቢት", "ሚያዝያ", "ግንቦት", "ሰነ", "ሓምለ", "ነሓሰ", "መስከረም", "ጥቅምቲ", "ሕዳር", "ታሕሳስ"]`
        pub const MON: &[&str] = &["ጥሪ", "ለካቲት", "መጋቢት", "ሚያዝያ", "ግንቦት", "ሰነ", "ሓምለ", "ነሓሰ", "መስከረም", "ጥቅምቲ", "ሕዳር", "ታሕሳስ"];
        /// `"%l:%M:%S %p"`
        pub const T_FMT: &str = "%l:%M:%S %p";
        /// `"%l:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%l:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod tk_TM {
    pub mod LC_ADDRESS {
        /// `Some("TKM")`
        pub const COUNTRY_AB3: Option<&str> = Some("TKM");
        /// `Some("TM")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("TM");
        /// `Some("TM")`
        pub const COUNTRY_CAR: Option<&str> = Some("TM");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Türkmenistan")`
        pub const COUNTRY_NAME: Option<&str> = Some("Türkmenistan");
        /// `Some(795)`
        pub const COUNTRY_NUM: Option<i64> = Some(795);
        /// `Some("TM")`
        pub const COUNTRY_POST: Option<&str> = Some("TM");
        /// `Some("tk")`
        pub const LANG_AB: Option<&str> = Some("tk");
        /// `Some("tuk")`
        pub const LANG_LIB: Option<&str> = Some("tuk");
        /// `Some("türkmen dili")`
        pub const LANG_NAME: Option<&str> = Some("türkmen dili");
        /// `Some("tuk")`
        pub const LANG_TERM: Option<&str> = Some("tuk");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
        ]);
        /// `Some("Pablo Saratxaga & Ghorban M. Tavakoly")`
        pub const CONTACT: Option<&str> = Some("Pablo Saratxaga & Ghorban M. Tavakoly");
        /// `Some("2011-04-09")`
        pub const DATE: Option<&str> = Some("2011-04-09");
        /// `Some("pablo@walon.org & gmt314@yahoo.com")`
        pub const EMAIL: Option<&str> = Some("pablo@walon.org & gmt314@yahoo.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Turkmen")`
        pub const LANGUAGE: Option<&str> = Some("Turkmen");
        /// `Some("0.5")`
        pub const REVISION: Option<&str> = Some("0.5");
        /// `Some("Ghorban M. Tavakoly")`
        pub const SOURCE: Option<&str> = Some("Ghorban M. Tavakoly");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Turkmenistan")`
        pub const TERRITORY: Option<&str> = Some("Turkmenistan");
        /// `Some("Turkmen locale for Turkmenistan")`
        pub const TITLE: Option<&str> = Some("Turkmen locale for Turkmenistan");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNýÝ]"`
        pub const NOEXPR: &str = "^[-0nNýÝ]";
        /// `Some("ýok")`
        pub const NOSTR: Option<&str> = Some("ýok");
        /// `"^[+1hH]"`
        pub const YESEXPR: &str = "^[+1hH]";
        /// `Some("hawa")`
        pub const YESSTR: Option<&str> = Some("hawa");
    }
    pub mod LC_MONETARY {
        /// `"MANAT"`
        pub const CURRENCY_SYMBOL: &str = "MANAT";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"TMM "`
        pub const INT_CURR_SYMBOL: &str = "TMM ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("993")`
        pub const INT_PREFIX: Option<&str> = Some("993");
        /// `Some("8\u{a0}10")`
        pub const INT_SELECT: Option<&str> = Some("8\u{a0}10");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["Duş", "Siş", "Çar", "Pen", "Ann", "Şen", "Ýek"]`
        pub const ABDAY: &[&str] = &["Duş", "Siş", "Çar", "Pen", "Ann", "Şen", "Ýek"];
        /// `&["Ýan", "Few", "Mar", "Apr", "Maý", "Iýn", "Iýl", "Awg", "Sen", "Okt", "Noý", "Dek"]`
        pub const ABMON: &[&str] = &["Ýan", "Few", "Mar", "Apr", "Maý", "Iýn", "Iýl", "Awg", "Sen", "Okt", "Noý", "Dek"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Duşenbe", "Sişenbe", "Çarşenbe", "Penşenbe", "Anna", "Şenbe", "Ýekşenbe"]`
        pub const DAY: &[&str] = &["Duşenbe", "Sişenbe", "Çarşenbe", "Penşenbe", "Anna", "Şenbe", "Ýekşenbe"];
        /// `"%d.%m.%Y"`
        pub const D_FMT: &str = "%d.%m.%Y";
        /// `"%d.%m.%Y %T"`
        pub const D_T_FMT: &str = "%d.%m.%Y %T";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Ýanwar", "Fewral", "Mart", "Aprel", "Maý", "Iýun", "Iýul", "Awgust", "Sentýabr", "Oktýabr", "Noýabr", "Dekabr"]`
        pub const MON: &[&str] = &["Ýanwar", "Fewral", "Mart", "Aprel", "Maý", "Iýun", "Iýul", "Awgust", "Sentýabr", "Oktýabr", "Noýabr", "Dekabr"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod tl_PH {
    pub mod LC_ADDRESS {
        /// `Some("PHL")`
        pub const COUNTRY_AB3: Option<&str> = Some("PHL");
        /// `Some("PH")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("PH");
        /// `Some("RP")`
        pub const COUNTRY_CAR: Option<&str> = Some("RP");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Pilipinas")`
        pub const COUNTRY_NAME: Option<&str> = Some("Pilipinas");
        /// `Some(608)`
        pub const COUNTRY_NUM: Option<i64> = Some(608);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("tl")`
        pub const LANG_AB: Option<&str> = Some("tl");
        /// `Some("tgl")`
        pub const LANG_LIB: Option<&str> = Some("tgl");
        /// `Some("Tagalog")`
        pub const LANG_NAME: Option<&str> = Some("Tagalog");
        /// `Some("tgl")`
        pub const LANG_TERM: Option<&str> = Some("tgl");
        /// `Some("%a%N%f%N%d%N%b%N%h %s %e %r%N%T %z%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%a%N%f%N%d%N%b%N%h %s %e %r%N%T %z%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_MONETARY"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_MONETARY"],
        ]);
        /// `Some("Rene Torres, Pablo Saratxaga")`
        pub const CONTACT: Option<&str> = Some("Rene Torres, Pablo Saratxaga");
        /// `Some("2001-01-28")`
        pub const DATE: Option<&str> = Some("2001-01-28");
        /// `Some("rgtorre@rocketmail.com, pablo@mandrakesoft.com")`
        pub const EMAIL: Option<&str> = Some("rgtorre@rocketmail.com, pablo@mandrakesoft.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Tagalog")`
        pub const LANGUAGE: Option<&str> = Some("Tagalog");
        /// `Some("0.2")`
        pub const REVISION: Option<&str> = Some("0.2");
        /// `Some("Rene Torres")`
        pub const SOURCE: Option<&str> = Some("Rene Torres");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Philippines")`
        pub const TERRITORY: Option<&str> = Some("Philippines");
        /// `Some("Tagalog language locale for Philippines")`
        pub const TITLE: Option<&str> = Some("Tagalog language locale for Philippines");
    }
    pub use super::fil_PH::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"₱"`
        pub const CURRENCY_SYMBOL: &str = "₱";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"PHP "`
        pub const INT_CURR_SYMBOL: &str = "PHP ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub use super::en_US::LC_NUMERIC;
    pub use super::en_PH::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Lin", "Lun", "Mar", "Miy", "Huw", "Biy", "Sab"]`
        pub const ABDAY: &[&str] = &["Lin", "Lun", "Mar", "Miy", "Huw", "Biy", "Sab"];
        /// `&["Ene", "Peb", "Mar", "Abr", "May", "Hun", "Hul", "Ago", "Sep", "Okt", "Nob", "Dis"]`
        pub const ABMON: &[&str] = &["Ene", "Peb", "Mar", "Abr", "May", "Hun", "Hul", "Ago", "Sep", "Okt", "Nob", "Dis"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["AM", "PM"]`
        pub const AM_PM: &[&str] = &["AM", "PM"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Linggo", "Lunes", "Martes", "Miyerkoles", "Huwebes", "Biyernes", "Sabado"]`
        pub const DAY: &[&str] = &["Linggo", "Lunes", "Martes", "Miyerkoles", "Huwebes", "Biyernes", "Sabado"];
        /// `"%m/%d/%y"`
        pub const D_FMT: &str = "%m/%d/%y";
        /// `"%a %d %b %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Enero", "Pebrero", "Marso", "Abril", "Mayo", "Hunyo", "Hulyo", "Agosto", "Septiyembre", "Oktubre", "Nobiyembre", "Disyembre"]`
        pub const MON: &[&str] = &["Enero", "Pebrero", "Marso", "Abril", "Mayo", "Hunyo", "Hulyo", "Agosto", "Septiyembre", "Oktubre", "Nobiyembre", "Disyembre"];
        /// `"%I:%M:%S %p"`
        pub const T_FMT: &str = "%I:%M:%S %p";
        /// `"%I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod tn_ZA {
    pub mod LC_ADDRESS {
        /// `Some("ZAF")`
        pub const COUNTRY_AB3: Option<&str> = Some("ZAF");
        /// `Some("ZA")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("ZA");
        /// `Some("ZA")`
        pub const COUNTRY_CAR: Option<&str> = Some("ZA");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Aforika Borwa")`
        pub const COUNTRY_NAME: Option<&str> = Some("Aforika Borwa");
        /// `Some(710)`
        pub const COUNTRY_NUM: Option<i64> = Some(710);
        /// `Some("ZA")`
        pub const COUNTRY_POST: Option<&str> = Some("ZA");
        /// `Some("tn")`
        pub const LANG_AB: Option<&str> = Some("tn");
        /// `Some("tsn")`
        pub const LANG_LIB: Option<&str> = Some("tsn");
        /// `Some("Setswana")`
        pub const LANG_NAME: Option<&str> = Some("Setswana");
        /// `Some("tsn")`
        pub const LANG_TERM: Option<&str> = Some("tsn");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("PO Box 28364, Sunnyside, 0132, South Africa")`
        pub const ADDRESS: Option<&str> = Some("PO Box 28364, Sunnyside, 0132, South Africa");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Dwayne Bailey")`
        pub const CONTACT: Option<&str> = Some("Dwayne Bailey");
        /// `Some("2005-10-13")`
        pub const DATE: Option<&str> = Some("2005-10-13");
        /// `Some("dwayne@translate.org.za")`
        pub const EMAIL: Option<&str> = Some("dwayne@translate.org.za");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Tswana")`
        pub const LANGUAGE: Option<&str> = Some("Tswana");
        /// `Some("0.4")`
        pub const REVISION: Option<&str> = Some("0.4");
        /// `Some("Zuza Software Foundation (Translate.org.za)")`
        pub const SOURCE: Option<&str> = Some("Zuza Software Foundation (Translate.org.za)");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("South Africa")`
        pub const TERRITORY: Option<&str> = Some("South Africa");
        /// `Some("Tswana locale for South Africa")`
        pub const TITLE: Option<&str> = Some("Tswana locale for South Africa");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `None`
        pub const NOSTR: Option<&str> = None;
        /// `"^[+1yYeE]"`
        pub const YESEXPR: &str = "^[+1yYeE]";
        /// `None`
        pub const YESSTR: Option<&str> = None;
    }
    pub use super::en_ZA::LC_MONETARY;
    pub use super::en_ZA::LC_NUMERIC;
    pub use super::en_ZA::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Tsh", "Mos", "Bed", "Rar", "Ne", "Tlh", "Mat"]`
        pub const ABDAY: &[&str] = &["Tsh", "Mos", "Bed", "Rar", "Ne", "Tlh", "Mat"];
        /// `&["Fer", "Tlh", "Mop", "Mor", "Mot", "See", "Phu", "Pha", "Lwe", "Dip", "Ngw", "Sed"]`
        pub const ABMON: &[&str] = &["Fer", "Tlh", "Mop", "Mor", "Mot", "See", "Phu", "Pha", "Lwe", "Dip", "Ngw", "Sed"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a %b %-e %H:%M:%S %Z %Y")`
        pub const DATE_FMT: Option<&str> = Some("%a %b %-e %H:%M:%S %Z %Y");
        /// `&["laTshipi", "Mosupologo", "Labobedi", "Laboraro", "Labone", "Labotlhano", "Lamatlhatso"]`
        pub const DAY: &[&str] = &["laTshipi", "Mosupologo", "Labobedi", "Laboraro", "Labone", "Labotlhano", "Lamatlhatso"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%a %-e %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %-e %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Ferikgong", "Tlhakole", "Mopitlwe", "Moranang", "Motsheganong", "Seetebosigo", "Phukwi", "Phatwe", "Lwetse", "Diphalane", "Ngwanatsele", "Sedimonthole"]`
        pub const MON: &[&str] = &["Ferikgong", "Tlhakole", "Mopitlwe", "Moranang", "Motsheganong", "Seetebosigo", "Phukwi", "Phatwe", "Lwetse", "Diphalane", "Ngwanatsele", "Sedimonthole"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod to_TO {
    pub mod LC_ADDRESS {
        /// `Some("TON")`
        pub const COUNTRY_AB3: Option<&str> = Some("TON");
        /// `Some("TO")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("TO");
        /// `Some("TON")`
        pub const COUNTRY_CAR: Option<&str> = Some("TON");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Tonga")`
        pub const COUNTRY_NAME: Option<&str> = Some("Tonga");
        /// `Some(776)`
        pub const COUNTRY_NUM: Option<i64> = Some(776);
        /// `Some("")`
        pub const COUNTRY_POST: Option<&str> = Some("");
        /// `Some("to")`
        pub const LANG_AB: Option<&str> = Some("to");
        /// `Some("ton")`
        pub const LANG_LIB: Option<&str> = Some("ton");
        /// `Some("lea fakatonga")`
        pub const LANG_NAME: Option<&str> = Some("lea fakatonga");
        /// `Some("ton")`
        pub const LANG_TERM: Option<&str> = Some("ton");
        /// `Some("%a%b%s%c")`
        pub const POSTAL_FMT: Option<&str> = Some("%a%b%s%c");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Samsung R & D Institute India - Delhi 2-A Sectort 126 Noida 201303")`
        pub const ADDRESS: Option<&str> = Some("Samsung R & D Institute India - Delhi 2-A Sectort 126 Noida 201303");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2017-03-23")`
        pub const DATE: Option<&str> = Some("2017-03-23");
        /// `Some("akhilesh.k@samsung.com")`
        pub const EMAIL: Option<&str> = Some("akhilesh.k@samsung.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Tongan")`
        pub const LANGUAGE: Option<&str> = Some("Tongan");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("International Components for Unicode")`
        pub const SOURCE: Option<&str> = Some("International Components for Unicode");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Tonga")`
        pub const TERRITORY: Option<&str> = Some("Tonga");
        /// `Some("Tongan language locale for Tonga")`
        pub const TITLE: Option<&str> = Some("Tongan language locale for Tonga");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("ʻIkai")`
        pub const NOSTR: Option<&str> = Some("ʻIkai");
        /// `"^[+1yY]"`
        pub const YESEXPR: &str = "^[+1yY]";
        /// `Some("ʻIo")`
        pub const YESSTR: Option<&str> = Some("ʻIo");
    }
    pub mod LC_MONETARY {
        /// `"T$"`
        pub const CURRENCY_SYMBOL: &str = "T$";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"TOP "`
        pub const INT_CURR_SYMBOL: &str = "TOP ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("676")`
        pub const INT_PREFIX: Option<&str> = Some("676");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %l");
    }
    pub mod LC_TIME {
        /// `&["Sāp", "Mōn", "Tūs", "Pul", "Tuʻa", "Fal", "Tok"]`
        pub const ABDAY: &[&str] = &["Sāp", "Mōn", "Tūs", "Pul", "Tuʻa", "Fal", "Tok"];
        /// `&["Sān", "Fēp", "Maʻa", "ʻEpe", "Mē", "Sun", "Siu", "ʻAok", "Sep", "ʻOka", "Nōv", "Tīs"]`
        pub const ABMON: &[&str] = &["Sān", "Fēp", "Maʻa", "ʻEpe", "Mē", "Sun", "Siu", "ʻAok", "Sep", "ʻOka", "Nōv", "Tīs"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["hengihengi", "efiafi"]`
        pub const AM_PM: &[&str] = &["hengihengi", "efiafi"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Sāpate", "Mōnite", "Tūsite", "Pulelulu", "Tuʻapulelulu", "Falaite", "Tokonaki"]`
        pub const DAY: &[&str] = &["Sāpate", "Mōnite", "Tūsite", "Pulelulu", "Tuʻapulelulu", "Falaite", "Tokonaki"];
        /// `"%A %d %b %Y"`
        pub const D_FMT: &str = "%A %d %b %Y";
        /// `"%A %d %b %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A %d %b %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Sānuali", "Fēpueli", "Maʻasi", "ʻEpeleli", "Mē", "Sune", "Siulai", "ʻAokosi", "Sepitema", "ʻOkatopa", "Nōvema", "Tīsema"]`
        pub const MON: &[&str] = &["Sānuali", "Fēpueli", "Maʻasi", "ʻEpeleli", "Mē", "Sune", "Siulai", "ʻAokosi", "Sepitema", "ʻOkatopa", "Nōvema", "Tīsema"];
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT: &str = "%I:%M:%S %p %Z";
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p %Z";
        /// `None`
        pub const WEEK: Option<&[i64]> = None;
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod tpi_PG {
    pub mod LC_ADDRESS {
        /// `Some("PNG")`
        pub const COUNTRY_AB3: Option<&str> = Some("PNG");
        /// `Some("PG")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("PG");
        /// `Some("PNG")`
        pub const COUNTRY_CAR: Option<&str> = Some("PNG");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Papua Niugini")`
        pub const COUNTRY_NAME: Option<&str> = Some("Papua Niugini");
        /// `Some(598)`
        pub const COUNTRY_NUM: Option<i64> = Some(598);
        /// `Some("")`
        pub const COUNTRY_POST: Option<&str> = Some("");
        /// `Some("")`
        pub const LANG_AB: Option<&str> = Some("");
        /// `Some("tpi")`
        pub const LANG_LIB: Option<&str> = Some("tpi");
        /// `Some("Tok Pisin")`
        pub const LANG_NAME: Option<&str> = Some("Tok Pisin");
        /// `Some("tpi")`
        pub const LANG_TERM: Option<&str> = Some("tpi");
        /// `Some("%a%s%z%C")`
        pub const POSTAL_FMT: Option<&str> = Some("%a%s%z%C");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2017-03-21")`
        pub const DATE: Option<&str> = Some("2017-03-21");
        /// `Some("akhilesh.k@samsung.com")`
        pub const EMAIL: Option<&str> = Some("akhilesh.k@samsung.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Tok Pisin")`
        pub const LANGUAGE: Option<&str> = Some("Tok Pisin");
        /// `Some("0.1")`
        pub const REVISION: Option<&str> = Some("0.1");
        /// `Some("Samsung Electronics Co., Ltd.")`
        pub const SOURCE: Option<&str> = Some("Samsung Electronics Co., Ltd.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Papua New Guinea")`
        pub const TERRITORY: Option<&str> = Some("Papua New Guinea");
        /// `Some("Tok Pisin language locale for Papua New Guinea")`
        pub const TITLE: Option<&str> = Some("Tok Pisin language locale for Papua New Guinea");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("nogat")`
        pub const NOSTR: Option<&str> = Some("nogat");
        /// `"^[+1yY]"`
        pub const YESEXPR: &str = "^[+1yY]";
        /// `Some("yes")`
        pub const YESSTR: Option<&str> = Some("yes");
    }
    pub mod LC_MONETARY {
        /// `"K"`
        pub const CURRENCY_SYMBOL: &str = "K";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"PGK "`
        pub const INT_CURR_SYMBOL: &str = "PGK ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `Some(1)`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = Some(1);
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `Some(1)`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = Some(1);
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("675")`
        pub const INT_PREFIX: Option<&str> = Some("675");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %l");
    }
    pub mod LC_TIME {
        /// `&["San", "Man", "Tun", "Tri", "Fon", "Fra", "Sar"]`
        pub const ABDAY: &[&str] = &["San", "Man", "Tun", "Tri", "Fon", "Fra", "Sar"];
        /// `&["Jan", "Feb", "Mas", "Epr", "Me", "Jun", "Jul", "Oga", "Sep", "Okt", "Nov", "Des"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Mas", "Epr", "Me", "Jun", "Jul", "Oga", "Sep", "Okt", "Nov", "Des"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["biknait", "apinun"]`
        pub const AM_PM: &[&str] = &["biknait", "apinun"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Sande", "Mande", "Tunde", "Trinde", "Fonde", "Fraide", "Sarere"]`
        pub const DAY: &[&str] = &["Sande", "Mande", "Tunde", "Trinde", "Fonde", "Fraide", "Sarere"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Janueri", "Februeri", "Mas", "Epril", "Me", "Jun", "Julai", "Ogas", "Septemba", "Oktoba", "Novemba", "Desemba"]`
        pub const MON: &[&str] = &["Janueri", "Februeri", "Mas", "Epril", "Me", "Jun", "Julai", "Ogas", "Septemba", "Oktoba", "Novemba", "Desemba"];
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT: &str = "%I:%M:%S %p %Z";
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p %Z";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod tr_CY {
    pub mod LC_ADDRESS {
        /// `Some("CYP")`
        pub const COUNTRY_AB3: Option<&str> = Some("CYP");
        /// `Some("CY")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("CY");
        /// `Some("CY")`
        pub const COUNTRY_CAR: Option<&str> = Some("CY");
        /// `Some("975")`
        pub const COUNTRY_ISBN: Option<&str> = Some("975");
        /// `Some("Kıbrıs")`
        pub const COUNTRY_NAME: Option<&str> = Some("Kıbrıs");
        /// `Some(196)`
        pub const COUNTRY_NUM: Option<i64> = Some(196);
        /// `Some("TR")`
        pub const COUNTRY_POST: Option<&str> = Some("TR");
        /// `Some("tr")`
        pub const LANG_AB: Option<&str> = Some("tr");
        /// `Some("tur")`
        pub const LANG_LIB: Option<&str> = Some("tur");
        /// `Some("Türkçe")`
        pub const LANG_NAME: Option<&str> = Some("Türkçe");
        /// `Some("tur")`
        pub const LANG_TERM: Option<&str> = Some("tur");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("https://www.gnu.org/software/libc/")`
        pub const ADDRESS: Option<&str> = Some("https://www.gnu.org/software/libc/");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2004-10-23")`
        pub const DATE: Option<&str> = Some("2004-10-23");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Turkish")`
        pub const LANGUAGE: Option<&str> = Some("Turkish");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Free Software Foundation, Inc.")`
        pub const SOURCE: Option<&str> = Some("Free Software Foundation, Inc.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Cyprus")`
        pub const TERRITORY: Option<&str> = Some("Cyprus");
        /// `Some("Turkish language locale for Cyprus")`
        pub const TITLE: Option<&str> = Some("Turkish language locale for Cyprus");
    }
    pub use super::tr_TR::LC_MESSAGES;
    pub use super::tr_TR::LC_MONETARY;
    pub use super::tr_TR::LC_NUMERIC;
    pub use super::el_CY::LC_TELEPHONE;
    pub use super::tr_TR::LC_TIME;
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod tr_TR {
    pub mod LC_ADDRESS {
        /// `Some("TUR")`
        pub const COUNTRY_AB3: Option<&str> = Some("TUR");
        /// `Some("TR")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("TR");
        /// `Some("TR")`
        pub const COUNTRY_CAR: Option<&str> = Some("TR");
        /// `Some("975")`
        pub const COUNTRY_ISBN: Option<&str> = Some("975");
        /// `Some("Türkiye")`
        pub const COUNTRY_NAME: Option<&str> = Some("Türkiye");
        /// `Some(792)`
        pub const COUNTRY_NUM: Option<i64> = Some(792);
        /// `Some("TR")`
        pub const COUNTRY_POST: Option<&str> = Some("TR");
        /// `Some("tr")`
        pub const LANG_AB: Option<&str> = Some("tr");
        /// `Some("tur")`
        pub const LANG_LIB: Option<&str> = Some("tur");
        /// `Some("Türkçe")`
        pub const LANG_NAME: Option<&str> = Some("Türkçe");
        /// `Some("tur")`
        pub const LANG_TERM: Option<&str> = Some("tur");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark")`
        pub const ADDRESS: Option<&str> = Some("Sankt Jørgens Alle 8, DK-1615 København V, Danmark");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2019-05-08")`
        pub const DATE: Option<&str> = Some("2019-05-08");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Turkish")`
        pub const LANGUAGE: Option<&str> = Some("Turkish");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("RAP")`
        pub const SOURCE: Option<&str> = Some("RAP");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Turkey")`
        pub const TERRITORY: Option<&str> = Some("Turkey");
        /// `Some("Turkish locale for Turkey")`
        pub const TITLE: Option<&str> = Some("Turkish locale for Turkey");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNhH]"`
        pub const NOEXPR: &str = "^[-0nNhH]";
        /// `Some("hayır")`
        pub const NOSTR: Option<&str> = Some("hayır");
        /// `"^[+1yYeE]"`
        pub const YESEXPR: &str = "^[+1yYeE]";
        /// `Some("evet")`
        pub const YESSTR: Option<&str> = Some("evet");
    }
    pub mod LC_MONETARY {
        /// `"₺"`
        pub const CURRENCY_SYMBOL: &str = "₺";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"TRY "`
        pub const INT_CURR_SYMBOL: &str = "TRY ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const MON_THOUSANDS_SEP: &str = ".";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const THOUSANDS_SEP: &str = ".";
    }
    pub mod LC_TELEPHONE {
        /// `Some("90")`
        pub const INT_PREFIX: Option<&str> = Some("90");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["Paz", "Pzt", "Sal", "Çrş", "Prş", "Cum", "Cts"]`
        pub const ABDAY: &[&str] = &["Paz", "Pzt", "Sal", "Çrş", "Prş", "Cum", "Cts"];
        /// `&["Oca", "Şub", "Mar", "Nis", "May", "Haz", "Tem", "Ağu", "Eyl", "Eki", "Kas", "Ara"]`
        pub const ABMON: &[&str] = &["Oca", "Şub", "Mar", "Nis", "May", "Haz", "Tem", "Ağu", "Eyl", "Eki", "Kas", "Ara"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["ÖÖ", "ÖS"]`
        pub const AM_PM: &[&str] = &["ÖÖ", "ÖS"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Pazar", "Pazartesi", "Salı", "Çarşamba", "Perşembe", "Cuma", "Cumartesi"]`
        pub const DAY: &[&str] = &["Pazar", "Pazartesi", "Salı", "Çarşamba", "Perşembe", "Cuma", "Cumartesi"];
        /// `"%d-%m-%Y"`
        pub const D_FMT: &str = "%d-%m-%Y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Ocak", "Şubat", "Mart", "Nisan", "Mayıs", "Haziran", "Temmuz", "Ağustos", "Eylül", "Ekim", "Kasım", "Aralık"]`
        pub const MON: &[&str] = &["Ocak", "Şubat", "Mart", "Nisan", "Mayıs", "Haziran", "Temmuz", "Ağustos", "Eylül", "Ekim", "Kasım", "Aralık"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `"%I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ts_ZA {
    pub mod LC_ADDRESS {
        /// `Some("ZAF")`
        pub const COUNTRY_AB3: Option<&str> = Some("ZAF");
        /// `Some("ZA")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("ZA");
        /// `Some("ZA")`
        pub const COUNTRY_CAR: Option<&str> = Some("ZA");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Afrika Dzonga")`
        pub const COUNTRY_NAME: Option<&str> = Some("Afrika Dzonga");
        /// `Some(710)`
        pub const COUNTRY_NUM: Option<i64> = Some(710);
        /// `Some("ZA")`
        pub const COUNTRY_POST: Option<&str> = Some("ZA");
        /// `Some("ts")`
        pub const LANG_AB: Option<&str> = Some("ts");
        /// `Some("tso")`
        pub const LANG_LIB: Option<&str> = Some("tso");
        /// `Some("Xitsonga")`
        pub const LANG_NAME: Option<&str> = Some("Xitsonga");
        /// `Some("tso")`
        pub const LANG_TERM: Option<&str> = Some("tso");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("PO Box 28364, Sunnyside, 0132, South Africa")`
        pub const ADDRESS: Option<&str> = Some("PO Box 28364, Sunnyside, 0132, South Africa");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Dwayne Bailey")`
        pub const CONTACT: Option<&str> = Some("Dwayne Bailey");
        /// `Some("2005-10-12")`
        pub const DATE: Option<&str> = Some("2005-10-12");
        /// `Some("dwayne@translate.org.za")`
        pub const EMAIL: Option<&str> = Some("dwayne@translate.org.za");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Tsonga")`
        pub const LANGUAGE: Option<&str> = Some("Tsonga");
        /// `Some("0.3")`
        pub const REVISION: Option<&str> = Some("0.3");
        /// `Some("Zuza Software Foundation (Translate.org.za)")`
        pub const SOURCE: Option<&str> = Some("Zuza Software Foundation (Translate.org.za)");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("South Africa")`
        pub const TERRITORY: Option<&str> = Some("South Africa");
        /// `Some("Tsonga locale for South Africa")`
        pub const TITLE: Option<&str> = Some("Tsonga locale for South Africa");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0eEnN]"`
        pub const NOEXPR: &str = "^[-0eEnN]";
        /// `Some("Ee")`
        pub const NOSTR: Option<&str> = Some("Ee");
        /// `"^[+1iIyY]"`
        pub const YESEXPR: &str = "^[+1iIyY]";
        /// `Some("Ina")`
        pub const YESSTR: Option<&str> = Some("Ina");
    }
    pub use super::en_ZA::LC_MONETARY;
    pub use super::en_ZA::LC_NUMERIC;
    pub use super::en_ZA::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Son", "Mus", "Bir", "Har", "Ne", "Tlh", "Mug"]`
        pub const ABDAY: &[&str] = &["Son", "Mus", "Bir", "Har", "Ne", "Tlh", "Mug"];
        /// `&["Sun", "Yan", "Kul", "Dzi", "Mud", "Kho", "Maw", "Mha", "Ndz", "Nhl", "Huk", "N'w"]`
        pub const ABMON: &[&str] = &["Sun", "Yan", "Kul", "Dzi", "Mud", "Kho", "Maw", "Mha", "Ndz", "Nhl", "Huk", "N'w"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a %b %-e %H:%M:%S %Z %Y")`
        pub const DATE_FMT: Option<&str> = Some("%a %b %-e %H:%M:%S %Z %Y");
        /// `&["Sonto", "Musumbhunuku", "Ravumbirhi", "Ravunharhu", "Ravumune", "Ravuntlhanu", "Mugqivela"]`
        pub const DAY: &[&str] = &["Sonto", "Musumbhunuku", "Ravumbirhi", "Ravunharhu", "Ravumune", "Ravuntlhanu", "Mugqivela"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%a %-e %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %-e %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Sunguti", "Nyenyenyani", "Nyenyankulu", "Dzivamisoko", "Mudyaxihi", "Khotavuxika", "Mawuwani", "Mhawuri", "Ndzhati", "Nhlangula", "Hukuri", "N'wendzamhala"]`
        pub const MON: &[&str] = &["Sunguti", "Nyenyenyani", "Nyenyankulu", "Dzivamisoko", "Mudyaxihi", "Khotavuxika", "Mawuwani", "Mhawuri", "Ndzhati", "Nhlangula", "Hukuri", "N'wendzamhala"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod tt_RU {
    pub mod LC_ADDRESS {
        /// `Some("RUS")`
        pub const COUNTRY_AB3: Option<&str> = Some("RUS");
        /// `Some("RU")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("RU");
        /// `Some("RUS")`
        pub const COUNTRY_CAR: Option<&str> = Some("RUS");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Русия")`
        pub const COUNTRY_NAME: Option<&str> = Some("Русия");
        /// `Some(643)`
        pub const COUNTRY_NUM: Option<i64> = Some(643);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("tt")`
        pub const LANG_AB: Option<&str> = Some("tt");
        /// `Some("tat")`
        pub const LANG_LIB: Option<&str> = Some("tat");
        /// `Some("татар")`
        pub const LANG_NAME: Option<&str> = Some("татар");
        /// `Some("tat")`
        pub const LANG_TERM: Option<&str> = Some("tat");
        /// `Some("???")`
        pub const POSTAL_FMT: Option<&str> = Some("???");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
        ]);
        /// `Some("Pablo Saratxaga, Rinat Norkin")`
        pub const CONTACT: Option<&str> = Some("Pablo Saratxaga, Rinat Norkin");
        /// `Some("2001-01-28")`
        pub const DATE: Option<&str> = Some("2001-01-28");
        /// `Some("pablo@mandrakesoft.com, rinat@taif.ru")`
        pub const EMAIL: Option<&str> = Some("pablo@mandrakesoft.com, rinat@taif.ru");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Tatar")`
        pub const LANGUAGE: Option<&str> = Some("Tatar");
        /// `Some("0.4")`
        pub const REVISION: Option<&str> = Some("0.4");
        /// `Some("Rinat Norkin")`
        pub const SOURCE: Option<&str> = Some("Rinat Norkin");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Russia")`
        pub const TERRITORY: Option<&str> = Some("Russia");
        /// `Some("Tatar language locale for Russia")`
        pub const TITLE: Option<&str> = Some("Tatar language locale for Russia");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNюЮ]"`
        pub const NOEXPR: &str = "^[-0nNюЮ]";
        /// `Some("юк")`
        pub const NOSTR: Option<&str> = Some("юк");
        /// `"^[+1yYәӘ]"`
        pub const YESEXPR: &str = "^[+1yYәӘ]";
        /// `Some("әйе")`
        pub const YESSTR: Option<&str> = Some("әйе");
    }
    pub mod LC_MONETARY {
        /// `"₽"`
        pub const CURRENCY_SYMBOL: &str = "₽";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"RUB "`
        pub const INT_CURR_SYMBOL: &str = "RUB ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const MON_THOUSANDS_SEP: &str = "\u{202f}";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub use super::ru_RU::LC_NUMERIC;
    pub use super::ru_RU::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["якш", "дүш", "сиш", "чәр", "пәнҗ", "җом", "шим"]`
        pub const ABDAY: &[&str] = &["якш", "дүш", "сиш", "чәр", "пәнҗ", "җом", "шим"];
        /// `&["гыйн", "фев", "мар", "апр", "май", "июнь", "июль", "авг", "сент", "окт", "нояб", "дек"]`
        pub const ABMON: &[&str] = &["гыйн", "фев", "мар", "апр", "май", "июнь", "июль", "авг", "сент", "окт", "нояб", "дек"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["якшәмбе", "дүшәмбе", "сишәмбе", "чәршәмбе", "пәнҗешәмбе", "җомга", "шимбә"]`
        pub const DAY: &[&str] = &["якшәмбе", "дүшәмбе", "сишәмбе", "чәршәмбе", "пәнҗешәмбе", "җомга", "шимбә"];
        /// `"%d.%m.%Y"`
        pub const D_FMT: &str = "%d.%m.%Y";
        /// `"%a %d %b %Y %T"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["гыйнвар", "февраль", "март", "апрель", "май", "июнь", "июль", "август", "сентябрь", "октябрь", "ноябрь", "декабрь"]`
        pub const MON: &[&str] = &["гыйнвар", "февраль", "март", "апрель", "май", "июнь", "июль", "август", "сентябрь", "октябрь", "ноябрь", "декабрь"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod tt_RU_iqtelif {
    pub mod LC_ADDRESS {
        /// `Some("RUS")`
        pub const COUNTRY_AB3: Option<&str> = Some("RUS");
        /// `Some("RU")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("RU");
        /// `Some("RUS")`
        pub const COUNTRY_CAR: Option<&str> = Some("RUS");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Urıs Patşahlıq")`
        pub const COUNTRY_NAME: Option<&str> = Some("Urıs Patşahlıq");
        /// `Some(643)`
        pub const COUNTRY_NUM: Option<i64> = Some(643);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("tt")`
        pub const LANG_AB: Option<&str> = Some("tt");
        /// `Some("tat")`
        pub const LANG_LIB: Option<&str> = Some("tat");
        /// `Some("tatar tele")`
        pub const LANG_NAME: Option<&str> = Some("tatar tele");
        /// `Some("tat")`
        pub const LANG_TERM: Option<&str> = Some("tat");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_NAME"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_NAME"],
        ]);
        /// `Some("Reshat Sabiq")`
        pub const CONTACT: Option<&str> = Some("Reshat Sabiq");
        /// `Some("2006-10-12")`
        pub const DATE: Option<&str> = Some("2006-10-12");
        /// `Some("tatar.iqtelif.i18n@gmail.com")`
        pub const EMAIL: Option<&str> = Some("tatar.iqtelif.i18n@gmail.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Tatar")`
        pub const LANGUAGE: Option<&str> = Some("Tatar");
        /// `Some("0.1")`
        pub const REVISION: Option<&str> = Some("0.1");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Russia")`
        pub const TERRITORY: Option<&str> = Some("Russia");
        /// `Some("Tatar language locale using IQTElif alphabet; for Tatarstan, Russian Federation")`
        pub const TITLE: Option<&str> = Some("Tatar language locale using IQTElif alphabet; for Tatarstan, Russian Federation");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0YyNn]"`
        pub const NOEXPR: &str = "^[-0YyNn]";
        /// `Some("yuq")`
        pub const NOSTR: Option<&str> = Some("yuq");
        /// `"^[+1Ee]"`
        pub const YESEXPR: &str = "^[+1Ee]";
        /// `Some("eyí")`
        pub const YESSTR: Option<&str> = Some("eyí");
    }
    pub mod LC_MONETARY {
        /// `"₽"`
        pub const CURRENCY_SYMBOL: &str = "₽";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"RUB "`
        pub const INT_CURR_SYMBOL: &str = "RUB ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const MON_THOUSANDS_SEP: &str = "\u{202f}";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const THOUSANDS_SEP: &str = ".";
    }
    pub mod LC_TELEPHONE {
        /// `Some("7")`
        pub const INT_PREFIX: Option<&str> = Some("7");
        /// `Some("0~10")`
        pub const INT_SELECT: Option<&str> = Some("0~10");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["Yek", "Düş", "Siş", "Çer", "Pen", "Com", "Şim"]`
        pub const ABDAY: &[&str] = &["Yek", "Düş", "Siş", "Çer", "Pen", "Com", "Şim"];
        /// `&["Ğın", "Fiw", "Mar", "Apr", "May", "Yün", "Yül", "Awg", "Sin", "Ükt", "Noy", "Dik"]`
        pub const ABMON: &[&str] = &["Ğın", "Fiw", "Mar", "Apr", "May", "Yün", "Yül", "Awg", "Sin", "Ükt", "Noy", "Dik"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["ÖA", "ÖS"]`
        pub const AM_PM: &[&str] = &["ÖA", "ÖS"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Yekşembí", "Düşembí", "Sişembí", "Çerşembí", "Pencíşembí", "Comğa", "Şimbe"]`
        pub const DAY: &[&str] = &["Yekşembí", "Düşembí", "Sişembí", "Çerşembí", "Pencíşembí", "Comğa", "Şimbe"];
        /// `"%d.%m.%Y"`
        pub const D_FMT: &str = "%d.%m.%Y";
        /// `"%a %d %b %Y %T"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Ğınwar", "Fiwral'", "Mart", "April", "May", "Yün", "Yül", "Awgust", "Sintebír", "Üktebír", "Noyebír", "Dikebír"]`
        pub const MON: &[&str] = &["Ğınwar", "Fiwral'", "Mart", "April", "May", "Yün", "Yül", "Awgust", "Sintebír", "Üktebír", "Noyebír", "Dikebír"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `"%I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ug_CN {
    pub mod LC_ADDRESS {
        /// `Some("CHN")`
        pub const COUNTRY_AB3: Option<&str> = Some("CHN");
        /// `Some("CN")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("CN");
        /// `Some("CHN")`
        pub const COUNTRY_CAR: Option<&str> = Some("CHN");
        /// `Some("7")`
        pub const COUNTRY_ISBN: Option<&str> = Some("7");
        /// `Some("جۇڭگو")`
        pub const COUNTRY_NAME: Option<&str> = Some("جۇڭگو");
        /// `Some(156)`
        pub const COUNTRY_NUM: Option<i64> = Some(156);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ug")`
        pub const LANG_AB: Option<&str> = Some("ug");
        /// `Some("uig")`
        pub const LANG_LIB: Option<&str> = Some("uig");
        /// `Some("ئۇيغۇرچە")`
        pub const LANG_NAME: Option<&str> = Some("ئۇيغۇرچە");
        /// `Some("uig")`
        pub const LANG_TERM: Option<&str> = Some("uig");
        /// `Some("%c%N%T%N%s %h %e %r%N%b%N%d%N%f%N%a%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%c%N%T%N%s %h %e %r%N%b%N%d%N%f%N%a%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
        ]);
        /// `Some("Eagle Burkut")`
        pub const CONTACT: Option<&str> = Some("Eagle Burkut");
        /// `Some("2011-02-26")`
        pub const DATE: Option<&str> = Some("2011-02-26");
        /// `Some("eagle.burkut@gmail.com")`
        pub const EMAIL: Option<&str> = Some("eagle.burkut@gmail.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Uyghur")`
        pub const LANGUAGE: Option<&str> = Some("Uyghur");
        /// `Some("2.00")`
        pub const REVISION: Option<&str> = Some("2.00");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("China")`
        pub const TERRITORY: Option<&str> = Some("China");
        /// `Some("Uyghur locale for China")`
        pub const TITLE: Option<&str> = Some("Uyghur locale for China");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNي]"`
        pub const NOEXPR: &str = "^[-0nNي]";
        /// `Some("ياق")`
        pub const NOSTR: Option<&str> = Some("ياق");
        /// `"^[+1yYھ]"`
        pub const YESEXPR: &str = "^[+1yYھ]";
        /// `Some("ھەئە")`
        pub const YESSTR: Option<&str> = Some("ھەئە");
    }
    pub use super::zh_CN::LC_MONETARY;
    pub use super::zh_CN::LC_NUMERIC;
    pub use super::zh_CN::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["ي", "د", "س", "چ", "پ", "ج", "ش"]`
        pub const ABDAY: &[&str] = &["ي", "د", "س", "چ", "پ", "ج", "ش"];
        /// `&["يانۋار", "فېۋرال", "مارت", "ئاپرېل", "ماي", "ئىيۇن", "ئىيۇل", "ئاۋغۇست", "سېنتەبىر", "ئۆكتەبىر", "نويابىر", "دېكابىر"]`
        pub const ABMON: &[&str] = &["يانۋار", "فېۋرال", "مارت", "ئاپرېل", "ماي", "ئىيۇن", "ئىيۇل", "ئاۋغۇست", "سېنتەبىر", "ئۆكتەبىر", "نويابىر", "دېكابىر"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a، %d-%m-%Y، %T")`
        pub const DATE_FMT: Option<&str> = Some("%a، %d-%m-%Y، %T");
        /// `&["يەكشەنبە", "دۈشەنبە", "سەيشەنبە", "چارشەنبە", "پەيشەنبە", "جۈمە", "شەنبە"]`
        pub const DAY: &[&str] = &["يەكشەنبە", "دۈشەنبە", "سەيشەنبە", "چارشەنبە", "پەيشەنبە", "جۈمە", "شەنبە"];
        /// `"%a، %d-%m-%Y"`
        pub const D_FMT: &str = "%a، %d-%m-%Y";
        /// `"%a، %d-%m-%Y، %T"`
        pub const D_T_FMT: &str = "%a، %d-%m-%Y، %T";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["يانۋار", "فېۋرال", "مارت", "ئاپرېل", "ماي", "ئىيۇن", "ئىيۇل", "ئاۋغۇست", "سېنتەبىر", "ئۆكتەبىر", "نويابىر", "دېكابىر"]`
        pub const MON: &[&str] = &["يانۋار", "فېۋرال", "مارت", "ئاپرېل", "ماي", "ئىيۇن", "ئىيۇل", "ئاۋغۇست", "سېنتەبىر", "ئۆكتەبىر", "نويابىر", "دېكابىر"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod uk_UA {
    pub mod LC_ADDRESS {
        /// `Some("UKR")`
        pub const COUNTRY_AB3: Option<&str> = Some("UKR");
        /// `Some("UA")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("UA");
        /// `Some("UA")`
        pub const COUNTRY_CAR: Option<&str> = Some("UA");
        /// `Some("966")`
        pub const COUNTRY_ISBN: Option<&str> = Some("966");
        /// `Some("Україна")`
        pub const COUNTRY_NAME: Option<&str> = Some("Україна");
        /// `Some(804)`
        pub const COUNTRY_NUM: Option<i64> = Some(804);
        /// `Some("UA")`
        pub const COUNTRY_POST: Option<&str> = Some("UA");
        /// `Some("uk")`
        pub const LANG_AB: Option<&str> = Some("uk");
        /// `Some("ukr")`
        pub const LANG_LIB: Option<&str> = Some("ukr");
        /// `Some("українська")`
        pub const LANG_NAME: Option<&str> = Some("українська");
        /// `Some("ukr")`
        pub const LANG_TERM: Option<&str> = Some("ukr");
        /// `Some("%a%t%f%t%d%N%s%t%h%t, %b%t%e%t%r%N%T%t%S%N%z%N%c")`
        pub const POSTAL_FMT: Option<&str> = Some("%a%t%f%t%d%N%s%t%h%t, %b%t%e%t%r%N%T%t%S%N%z%N%c");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("GNU libc maintainers")`
        pub const CONTACT: Option<&str> = Some("GNU libc maintainers");
        /// `Some("2006-05-20")`
        pub const DATE: Option<&str> = Some("2006-05-20");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Ukrainian")`
        pub const LANGUAGE: Option<&str> = Some("Ukrainian");
        /// `Some("2.1.12")`
        pub const REVISION: Option<&str> = Some("2.1.12");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Ukraine")`
        pub const TERRITORY: Option<&str> = Some("Ukraine");
        /// `Some("Ukrainian Language Locale for Ukraine")`
        pub const TITLE: Option<&str> = Some("Ukrainian Language Locale for Ukraine");
    }
    pub mod LC_MESSAGES {
        /// `"^([-0Nn]|[Нн][Іі])$"`
        pub const NOEXPR: &str = "^([-0Nn]|[Нн][Іі])$";
        /// `Some("ні")`
        pub const NOSTR: Option<&str> = Some("ні");
        /// `"^([+1Yy]|[Тт][Аа][Кк]?)$"`
        pub const YESEXPR: &str = "^([+1Yy]|[Тт][Аа][Кк]?)$";
        /// `Some("так")`
        pub const YESSTR: Option<&str> = Some("так");
    }
    pub mod LC_MONETARY {
        /// `"грн."`
        pub const CURRENCY_SYMBOL: &str = "грн.";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"UAH "`
        pub const INT_CURR_SYMBOL: &str = "UAH ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `Some(1)`
        pub const INT_N_CS_PRECEDES: Option<i64> = Some(1);
        /// `Some(1)`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = Some(1);
        /// `Some(4)`
        pub const INT_N_SIGN_POSN: Option<i64> = Some(4);
        /// `Some(1)`
        pub const INT_P_CS_PRECEDES: Option<i64> = Some(1);
        /// `Some(2)`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = Some(2);
        /// `Some(4)`
        pub const INT_P_SIGN_POSN: Option<i64> = Some(4);
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const MON_THOUSANDS_SEP: &str = "\u{202f}";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `2`
        pub const P_SEP_BY_SPACE: i64 = 2;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const THOUSANDS_SEP: &str = "\u{202f}";
    }
    pub mod LC_TELEPHONE {
        /// `Some("380")`
        pub const INT_PREFIX: Option<&str> = Some("380");
        /// `Some("8~10")`
        pub const INT_SELECT: Option<&str> = Some("8~10");
        /// `Some("(%A) %l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("(%A) %l");
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["нд", "пн", "вт", "ср", "чт", "пт", "сб"]`
        pub const ABDAY: &[&str] = &["нд", "пн", "вт", "ср", "чт", "пт", "сб"];
        /// `&["січ", "лют", "бер", "кві", "тра", "чер", "лип", "сер", "вер", "жов", "лис", "гру"]`
        pub const ABMON: &[&str] = &["січ", "лют", "бер", "кві", "тра", "чер", "лип", "сер", "вер", "жов", "лис", "гру"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `Some(&["січень", "лютий", "березень", "квітень", "травень", "червень", "липень", "серпень", "вересень", "жовтень", "листопад", "грудень"])`
        pub const ALT_MON: Option<&[&str]> = Some(&["січень", "лютий", "березень", "квітень", "травень", "червень", "липень", "серпень", "вересень", "жовтень", "листопад", "грудень"]);
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `Some(2)`
        pub const CAL_DIRECTION: Option<i64> = Some(2);
        /// `Some("%A, %-d %B %Y %X %z")`
        pub const DATE_FMT: Option<&str> = Some("%A, %-d %B %Y %X %z");
        /// `&["неділя", "понеділок", "вівторок", "середа", "четвер", "п'ятниця", "субота"]`
        pub const DAY: &[&str] = &["неділя", "понеділок", "вівторок", "середа", "четвер", "п'ятниця", "субота"];
        /// `"%d.%m.%y"`
        pub const D_FMT: &str = "%d.%m.%y";
        /// `"%a, %d-%b-%Y %T %z"`
        pub const D_T_FMT: &str = "%a, %d-%b-%Y %T %z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["січня", "лютого", "березня", "квітня", "травня", "червня", "липня", "серпня", "вересня", "жовтня", "листопада", "грудня"]`
        pub const MON: &[&str] = &["січня", "лютого", "березня", "квітня", "травня", "червня", "липня", "серпня", "вересня", "жовтня", "листопада", "грудня"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod unm_US {
    pub mod LC_ADDRESS {
        /// `Some("USA")`
        pub const COUNTRY_AB3: Option<&str> = Some("USA");
        /// `Some("US")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("US");
        /// `Some("USA")`
        pub const COUNTRY_CAR: Option<&str> = Some("USA");
        /// `Some("0")`
        pub const COUNTRY_ISBN: Option<&str> = Some("0");
        /// `Some("USA")`
        pub const COUNTRY_NAME: Option<&str> = Some("USA");
        /// `Some(840)`
        pub const COUNTRY_NUM: Option<i64> = Some(840);
        /// `Some("USA")`
        pub const COUNTRY_POST: Option<&str> = Some("USA");
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("unm")`
        pub const LANG_LIB: Option<&str> = Some("unm");
        /// `Some("Lenape")`
        pub const LANG_NAME: Option<&str> = Some("Lenape");
        /// `Some("unm")`
        pub const LANG_TERM: Option<&str> = Some("unm");
        /// `Some("%a%N%f%N%d%N%b%N%h %s %e %r%N%T, %S %z%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%a%N%f%N%d%N%b%N%h %s %e %r%N%T, %S %z%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2011-10-27")`
        pub const DATE: Option<&str> = Some("2011-10-27");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Unami Delaware")`
        pub const LANGUAGE: Option<&str> = Some("Unami Delaware");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("United States")`
        pub const TERRITORY: Option<&str> = Some("United States");
        /// `Some("Unami Delaware locale for the USA")`
        pub const TITLE: Option<&str> = Some("Unami Delaware locale for the USA");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0kKmM]"`
        pub const NOEXPR: &str = "^[-0kKmM]";
        /// `None`
        pub const NOSTR: Option<&str> = None;
        /// `"^[+1yY]"`
        pub const YESEXPR: &str = "^[+1yY]";
        /// `None`
        pub const YESSTR: Option<&str> = None;
    }
    pub mod LC_MONETARY {
        /// `"$"`
        pub const CURRENCY_SYMBOL: &str = "$";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"USD "`
        pub const INT_CURR_SYMBOL: &str = "USD ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `"\u{202f}"`
        pub const MON_THOUSANDS_SEP: &str = "\u{202f}";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[2, 2, 2, 3]`
        pub const GROUPING: &[i64] = &[2, 2, 2, 3];
        /// `"\u{202f}"`
        pub const THOUSANDS_SEP: &str = "\u{202f}";
    }
    pub use super::en_US::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["ken", "man", "tus", "lel", "tas", "pel", "sat"]`
        pub const ABDAY: &[&str] = &["ken", "man", "tus", "lel", "tas", "pel", "sat"];
        /// `&["eni", "chk", "xam", "kwe", "tai", "nip", "lai", "win", "tah", "puk", "kun", "mux"]`
        pub const ABMON: &[&str] = &["eni", "chk", "xam", "kwe", "tai", "nip", "lai", "win", "tah", "puk", "kun", "mux"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["kentuwei", "manteke", "tusteke", "lelai", "tasteke", "pelaiteke", "sateteke"]`
        pub const DAY: &[&str] = &["kentuwei", "manteke", "tusteke", "lelai", "tasteke", "pelaiteke", "sateteke"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["enikwsi", "chkwali", "xamokhwite", "kwetayoxe", "tainipen", "kichinipen", "lainipen", "winaminke", "kichitahkok", "puksit", "wini", "muxkotae"]`
        pub const MON: &[&str] = &["enikwsi", "chkwali", "xamokhwite", "kwetayoxe", "tainipen", "kichinipen", "lainipen", "winaminke", "kichitahkok", "puksit", "wini", "muxkotae"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ur_IN {
    pub mod LC_ADDRESS {
        /// `Some("IND")`
        pub const COUNTRY_AB3: Option<&str> = Some("IND");
        /// `Some("IN")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("IN");
        /// `Some("IND")`
        pub const COUNTRY_CAR: Option<&str> = Some("IND");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("بھارت")`
        pub const COUNTRY_NAME: Option<&str> = Some("بھارت");
        /// `Some(356)`
        pub const COUNTRY_NUM: Option<i64> = Some(356);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ur")`
        pub const LANG_AB: Option<&str> = Some("ur");
        /// `Some("urd")`
        pub const LANG_LIB: Option<&str> = Some("urd");
        /// `Some("اردو")`
        pub const LANG_NAME: Option<&str> = Some("اردو");
        /// `Some("urd")`
        pub const LANG_TERM: Option<&str> = Some("urd");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Marisfot III, Marigold Premises, East-Wing, Kalyaninagar, Pune, India-411014")`
        pub const ADDRESS: Option<&str> = Some("Marisfot III, Marigold Premises, East-Wing, Kalyaninagar, Pune, India-411014");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2009,June,09")`
        pub const DATE: Option<&str> = Some("2009,June,09");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Urdu")`
        pub const LANGUAGE: Option<&str> = Some("Urdu");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Red Hat, Pune")`
        pub const SOURCE: Option<&str> = Some("Red Hat, Pune");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("India")`
        pub const TERRITORY: Option<&str> = Some("India");
        /// `Some("Urdu language locale for India")`
        pub const TITLE: Option<&str> = Some("Urdu language locale for India");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNن]"`
        pub const NOEXPR: &str = "^[-0nNن]";
        /// `Some("نہیں")`
        pub const NOSTR: Option<&str> = Some("نہیں");
        /// `"^[+1yYہ]"`
        pub const YESEXPR: &str = "^[+1yYہ]";
        /// `Some("ہاں")`
        pub const YESSTR: Option<&str> = Some("ہاں");
    }
    pub use super::hi_IN::LC_MONETARY;
    pub use super::hi_IN::LC_NUMERIC;
    pub use super::hi_IN::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["اتوار", "پیر", "منگل", "بدھ", "جمعرات", "جمعہ", "سنیچر"]`
        pub const ABDAY: &[&str] = &["اتوار", "پیر", "منگل", "بدھ", "جمعرات", "جمعہ", "سنیچر"];
        /// `&["جنوری", "فروری", "مارچ", "اپریل", "مئی", "جون", "جولائی", "اگست", "ستمبر", "اکتوبر", "نومبر", "دسمبر"]`
        pub const ABMON: &[&str] = &["جنوری", "فروری", "مارچ", "اپریل", "مئی", "جون", "جولائی", "اگست", "ستمبر", "اکتوبر", "نومبر", "دسمبر"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["AM", "PM"]`
        pub const AM_PM: &[&str] = &["AM", "PM"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["اتوار", "پیر", "منگل", "بدھ", "جمعرات", "جمعہ", "سنیچر"]`
        pub const DAY: &[&str] = &["اتوار", "پیر", "منگل", "بدھ", "جمعرات", "جمعہ", "سنیچر"];
        /// `"%-d/%-m/%y"`
        pub const D_FMT: &str = "%-d/%-m/%y";
        /// `"%A %d %b %Y %I:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A %d %b %Y %I:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["جنوری", "فروری", "مارچ", "اپریل", "مئی", "جون", "جولائی", "اگست", "ستمبر", "اکتوبر", "نومبر", "دسمبر"]`
        pub const MON: &[&str] = &["جنوری", "فروری", "مارچ", "اپریل", "مئی", "جون", "جولائی", "اگست", "ستمبر", "اکتوبر", "نومبر", "دسمبر"];
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT: &str = "%I:%M:%S %p %Z";
        /// `"%I:%M:%S %p %Z"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p %Z";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ur_PK {
    pub mod LC_ADDRESS {
        /// `Some("PAK")`
        pub const COUNTRY_AB3: Option<&str> = Some("PAK");
        /// `Some("PK")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("PK");
        /// `Some("PK")`
        pub const COUNTRY_CAR: Option<&str> = Some("PK");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("پاکستان")`
        pub const COUNTRY_NAME: Option<&str> = Some("پاکستان");
        /// `Some(586)`
        pub const COUNTRY_NUM: Option<i64> = Some(586);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("ur")`
        pub const LANG_AB: Option<&str> = Some("ur");
        /// `Some("urd")`
        pub const LANG_LIB: Option<&str> = Some("urd");
        /// `Some("اردو")`
        pub const LANG_NAME: Option<&str> = Some("اردو");
        /// `Some("urd")`
        pub const LANG_TERM: Option<&str> = Some("urd");
        /// `Some("%a%N%f%N%d%N%b%N%h %s %e %r%N%T %z%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%a%N%f%N%d%N%b%N%h %s %e %r%N%T %z%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-07-11")`
        pub const DATE: Option<&str> = Some("2000-07-11");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Urdu")`
        pub const LANGUAGE: Option<&str> = Some("Urdu");
        /// `Some("0.3")`
        pub const REVISION: Option<&str> = Some("0.3");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Pakistan")`
        pub const TERRITORY: Option<&str> = Some("Pakistan");
        /// `Some("Urdu Language Locale for Pakistan")`
        pub const TITLE: Option<&str> = Some("Urdu Language Locale for Pakistan");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNنo]"`
        pub const NOEXPR: &str = "^[-0nNنo]";
        /// `Some("نہیں")`
        pub const NOSTR: Option<&str> = Some("نہیں");
        /// `"^[+1yYهبf]"`
        pub const YESEXPR: &str = "^[+1yYهبf]";
        /// `Some("ہاں")`
        pub const YESSTR: Option<&str> = Some("ہاں");
    }
    pub mod LC_MONETARY {
        /// `"Rs"`
        pub const CURRENCY_SYMBOL: &str = "Rs";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"PKR "`
        pub const INT_CURR_SYMBOL: &str = "PKR ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `2`
        pub const P_SIGN_POSN: i64 = 2;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("92")`
        pub const INT_PREFIX: Option<&str> = Some("92");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c ;%a ;%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c ;%a ;%l");
    }
    pub mod LC_TIME {
        /// `&["اتوار", "پير", "منگل", "بدھ", "جمعرات", "جمعه", "هفته"]`
        pub const ABDAY: &[&str] = &["اتوار", "پير", "منگل", "بدھ", "جمعرات", "جمعه", "هفته"];
        /// `&["جنوری", "فروری", "مارچ", "اپریل", "مئی", "جون", "جولائی", "اگست", "ستمبر", "اکتوبر", "نومبر", "دسمبر"]`
        pub const ABMON: &[&str] = &["جنوری", "فروری", "مارچ", "اپریل", "مئی", "جون", "جولائی", "اگست", "ستمبر", "اکتوبر", "نومبر", "دسمبر"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["ص", "ش"]`
        pub const AM_PM: &[&str] = &["ص", "ش"];
        /// `Some(3)`
        pub const CAL_DIRECTION: Option<i64> = Some(3);
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["اتوار", "پير", "منگل", "بدھ", "جمعرات", "جمعه", "هفته"]`
        pub const DAY: &[&str] = &["اتوار", "پير", "منگل", "بدھ", "جمعرات", "جمعه", "هفته"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"و %H:%M:%S %Z ت %d %B %Y"`
        pub const D_T_FMT: &str = "و %H:%M:%S %Z ت %d %B %Y";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["جنوری", "فروری", "مارچ", "اپریل", "مئی", "جون", "جولائی", "اگست", "ستمبر", "اکتوبر", "نومبر", "دسمبر"]`
        pub const MON: &[&str] = &["جنوری", "فروری", "مارچ", "اپریل", "مئی", "جون", "جولائی", "اگست", "ستمبر", "اکتوبر", "نومبر", "دسمبر"];
        /// `"%H:%M:%S"`
        pub const T_FMT: &str = "%H:%M:%S";
        /// `"%P %I:%M:%S"`
        pub const T_FMT_AMPM: &str = "%P %I:%M:%S";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod uz_UZ {
    pub mod LC_ADDRESS {
        /// `Some("UZB")`
        pub const COUNTRY_AB3: Option<&str> = Some("UZB");
        /// `Some("UZ")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("UZ");
        /// `Some("UZ")`
        pub const COUNTRY_CAR: Option<&str> = Some("UZ");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Oʻzbekiston")`
        pub const COUNTRY_NAME: Option<&str> = Some("Oʻzbekiston");
        /// `Some(860)`
        pub const COUNTRY_NUM: Option<i64> = Some(860);
        /// `Some("UZ")`
        pub const COUNTRY_POST: Option<&str> = Some("UZ");
        /// `Some("uz")`
        pub const LANG_AB: Option<&str> = Some("uz");
        /// `Some("uzb")`
        pub const LANG_LIB: Option<&str> = Some("uzb");
        /// `Some("o‘zbek")`
        pub const LANG_NAME: Option<&str> = Some("o‘zbek");
        /// `Some("uzb")`
        pub const LANG_TERM: Option<&str> = Some("uzb");
        /// `Some("%C%N%T%N%s %h%N%f%N%b%N%d%N%e %r%N%a%N%z")`
        pub const POSTAL_FMT: Option<&str> = Some("%C%N%T%N%s %h%N%f%N%b%N%d%N%e %r%N%a%N%z");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Bobir Ismailov, Pablo Saratxaga, Mashrab Kuvatov")`
        pub const CONTACT: Option<&str> = Some("Bobir Ismailov, Pablo Saratxaga, Mashrab Kuvatov");
        /// `Some("2003-06-27")`
        pub const DATE: Option<&str> = Some("2003-06-27");
        /// `Some("bobir_is@yahoo.com, pablo@mandrakesoft.com, kmashrab@uni-bremen.de")`
        pub const EMAIL: Option<&str> = Some("bobir_is@yahoo.com, pablo@mandrakesoft.com, kmashrab@uni-bremen.de");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Uzbek")`
        pub const LANGUAGE: Option<&str> = Some("Uzbek");
        /// `Some("0.5")`
        pub const REVISION: Option<&str> = Some("0.5");
        /// `Some("Bobir Ismailov")`
        pub const SOURCE: Option<&str> = Some("Bobir Ismailov");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Uzbekistan")`
        pub const TERRITORY: Option<&str> = Some("Uzbekistan");
        /// `Some("Uzbek (latin) locale for Uzbekistan")`
        pub const TITLE: Option<&str> = Some("Uzbek (latin) locale for Uzbekistan");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNyYjJЙй]"`
        pub const NOEXPR: &str = "^[-0nNyYjJЙй]";
        /// `Some("yo‘q")`
        pub const NOSTR: Option<&str> = Some("yo‘q");
        /// `"^[+1hHҲҳ]"`
        pub const YESEXPR: &str = "^[+1hHҲҳ]";
        /// `Some("ha")`
        pub const YESSTR: Option<&str> = Some("ha");
    }
    pub mod LC_MONETARY {
        /// `"soʻm"`
        pub const CURRENCY_SYMBOL: &str = "soʻm";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"UZS "`
        pub const INT_CURR_SYMBOL: &str = "UZS ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("998")`
        pub const INT_PREFIX: Option<&str> = Some("998");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["Yak", "Du", "Se", "Cho", "Pay", "Ju", "Sha"]`
        pub const ABDAY: &[&str] = &["Yak", "Du", "Se", "Cho", "Pay", "Ju", "Sha"];
        /// `&["Yan", "Fev", "Mar", "Apr", "May", "Iyn", "Iyl", "Avg", "Sen", "Okt", "Noy", "Dek"]`
        pub const ABMON: &[&str] = &["Yan", "Fev", "Mar", "Apr", "May", "Iyn", "Iyl", "Avg", "Sen", "Okt", "Noy", "Dek"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%d %B, %Y yil, %A")`
        pub const DATE_FMT: Option<&str> = Some("%d %B, %Y yil, %A");
        /// `&["Yakshanba", "Dushanba", "Seshanba", "Chorshanba", "Payshanba", "Juma", "Shanba"]`
        pub const DAY: &[&str] = &["Yakshanba", "Dushanba", "Seshanba", "Chorshanba", "Payshanba", "Juma", "Shanba"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%T, %d %B, %Y yil, %A"`
        pub const D_T_FMT: &str = "%T, %d %B, %Y yil, %A";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Yanvar", "Fevral", "Mart", "Aprel", "May", "Iyun", "Iyul", "Avgust", "Sentabr", "Oktabr", "Noyabr", "Dekabr"]`
        pub const MON: &[&str] = &["Yanvar", "Fevral", "Mart", "Aprel", "May", "Iyun", "Iyul", "Avgust", "Sentabr", "Oktabr", "Noyabr", "Dekabr"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod uz_UZ_cyrillic {
    pub mod LC_ADDRESS {
        /// `Some("UZB")`
        pub const COUNTRY_AB3: Option<&str> = Some("UZB");
        /// `Some("UZ")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("UZ");
        /// `Some("UZ")`
        pub const COUNTRY_CAR: Option<&str> = Some("UZ");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Ўзбекистон")`
        pub const COUNTRY_NAME: Option<&str> = Some("Ўзбекистон");
        /// `Some(860)`
        pub const COUNTRY_NUM: Option<i64> = Some(860);
        /// `Some("UZ")`
        pub const COUNTRY_POST: Option<&str> = Some("UZ");
        /// `Some("uz")`
        pub const LANG_AB: Option<&str> = Some("uz");
        /// `Some("uzb")`
        pub const LANG_LIB: Option<&str> = Some("uzb");
        /// `Some("Ўзбекча")`
        pub const LANG_NAME: Option<&str> = Some("Ўзбекча");
        /// `Some("uzb")`
        pub const LANG_TERM: Option<&str> = Some("uzb");
        /// `Some("%C%N%T%N%s %h%N%f%N%b%N%d%N%e %r%N%a%N%z")`
        pub const POSTAL_FMT: Option<&str> = Some("%C%N%T%N%s %h%N%f%N%b%N%d%N%e %r%N%a%N%z");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Mashrab Kuvatov, Pablo Saratxaga")`
        pub const CONTACT: Option<&str> = Some("Mashrab Kuvatov, Pablo Saratxaga");
        /// `Some("2003-05-30")`
        pub const DATE: Option<&str> = Some("2003-05-30");
        /// `Some("kmashrab@uni-bremen.de, pablo@mandrakesoft.com")`
        pub const EMAIL: Option<&str> = Some("kmashrab@uni-bremen.de, pablo@mandrakesoft.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Uzbek")`
        pub const LANGUAGE: Option<&str> = Some("Uzbek");
        /// `Some("0.1")`
        pub const REVISION: Option<&str> = Some("0.1");
        /// `Some("Mashrab Kuvatov")`
        pub const SOURCE: Option<&str> = Some("Mashrab Kuvatov");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Uzbekistan")`
        pub const TERRITORY: Option<&str> = Some("Uzbekistan");
        /// `Some("Uzbek (cyrillic) locale for Uzbekistan")`
        pub const TITLE: Option<&str> = Some("Uzbek (cyrillic) locale for Uzbekistan");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNyYjJЙй]"`
        pub const NOEXPR: &str = "^[-0nNyYjJЙй]";
        /// `Some("йўқ")`
        pub const NOSTR: Option<&str> = Some("йўқ");
        /// `"^[+1hHҲҳ]"`
        pub const YESEXPR: &str = "^[+1hHҲҳ]";
        /// `Some("ҳа")`
        pub const YESSTR: Option<&str> = Some("ҳа");
    }
    pub mod LC_MONETARY {
        /// `"сўм"`
        pub const CURRENCY_SYMBOL: &str = "сўм";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"UZS "`
        pub const INT_CURR_SYMBOL: &str = "UZS ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub use super::uz_UZ::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Якш", "Душ", "Сеш", "Чор", "Пай", "Жум", "Шан"]`
        pub const ABDAY: &[&str] = &["Якш", "Душ", "Сеш", "Чор", "Пай", "Жум", "Шан"];
        /// `&["Янв", "Фев", "Мар", "Апр", "Май", "Июн", "Июл", "Авг", "Сен", "Окт", "Ноя", "Дек"]`
        pub const ABMON: &[&str] = &["Янв", "Фев", "Мар", "Апр", "Май", "Июн", "Июл", "Авг", "Сен", "Окт", "Ноя", "Дек"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%d %B, %Y йил, %A")`
        pub const DATE_FMT: Option<&str> = Some("%d %B, %Y йил, %A");
        /// `&["Якшанба", "Душанба", "Сешанба", "Чоршанба", "Пайшанба", "Жума", "Шанба"]`
        pub const DAY: &[&str] = &["Якшанба", "Душанба", "Сешанба", "Чоршанба", "Пайшанба", "Жума", "Шанба"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%T, %d %B, %Y йил, %A"`
        pub const D_T_FMT: &str = "%T, %d %B, %Y йил, %A";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Январ", "Феврал", "Март", "Апрел", "Май", "Июн", "Июл", "Август", "Сентябр", "Октябр", "Ноябр", "Декабр"]`
        pub const MON: &[&str] = &["Январ", "Феврал", "Март", "Апрел", "Май", "Июн", "Июл", "Август", "Сентябр", "Октябр", "Ноябр", "Декабр"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod ve_ZA {
    pub mod LC_ADDRESS {
        /// `Some("ZAF")`
        pub const COUNTRY_AB3: Option<&str> = Some("ZAF");
        /// `Some("ZA")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("ZA");
        /// `Some("ZA")`
        pub const COUNTRY_CAR: Option<&str> = Some("ZA");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Afurika Tshipembe")`
        pub const COUNTRY_NAME: Option<&str> = Some("Afurika Tshipembe");
        /// `Some(710)`
        pub const COUNTRY_NUM: Option<i64> = Some(710);
        /// `Some("ZA")`
        pub const COUNTRY_POST: Option<&str> = Some("ZA");
        /// `Some("ve")`
        pub const LANG_AB: Option<&str> = Some("ve");
        /// `Some("ven")`
        pub const LANG_LIB: Option<&str> = Some("ven");
        /// `Some("Tshivenḓa")`
        pub const LANG_NAME: Option<&str> = Some("Tshivenḓa");
        /// `Some("ven")`
        pub const LANG_TERM: Option<&str> = Some("ven");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("PO Box 28364, Sunnyside, 0132, South Africa")`
        pub const ADDRESS: Option<&str> = Some("PO Box 28364, Sunnyside, 0132, South Africa");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Dwayne Bailey")`
        pub const CONTACT: Option<&str> = Some("Dwayne Bailey");
        /// `Some("2005-10-13")`
        pub const DATE: Option<&str> = Some("2005-10-13");
        /// `Some("dwayne@translate.org.za")`
        pub const EMAIL: Option<&str> = Some("dwayne@translate.org.za");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Venda")`
        pub const LANGUAGE: Option<&str> = Some("Venda");
        /// `Some("0.3")`
        pub const REVISION: Option<&str> = Some("0.3");
        /// `Some("Zuza Software Foundation (Translate.org.za)")`
        pub const SOURCE: Option<&str> = Some("Zuza Software Foundation (Translate.org.za)");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("South Africa")`
        pub const TERRITORY: Option<&str> = Some("South Africa");
        /// `Some("Venda locale for South Africa")`
        pub const TITLE: Option<&str> = Some("Venda locale for South Africa");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNhH]"`
        pub const NOEXPR: &str = "^[-0nNhH]";
        /// `None`
        pub const NOSTR: Option<&str> = None;
        /// `"^[+1yYeE]"`
        pub const YESEXPR: &str = "^[+1yYeE]";
        /// `None`
        pub const YESSTR: Option<&str> = None;
    }
    pub use super::en_ZA::LC_MONETARY;
    pub use super::en_ZA::LC_NUMERIC;
    pub use super::en_ZA::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Swo", "Mus", "Vhi", "Rar", "ṋa", "Ṱan", "Mug"]`
        pub const ABDAY: &[&str] = &["Swo", "Mus", "Vhi", "Rar", "ṋa", "Ṱan", "Mug"];
        /// `&["Pha", "Luh", "Fam", "Lam", "Shu", "Lwi", "Lwa", "Ngu", "Khu", "Tsh", "Ḽar", "Nye"]`
        pub const ABMON: &[&str] = &["Pha", "Luh", "Fam", "Lam", "Shu", "Lwi", "Lwa", "Ngu", "Khu", "Tsh", "Ḽar", "Nye"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Swondaha", "Musumbuluwo", "Ḽavhuvhili", "Ḽavhuraru", "Ḽavhuṋa", "Ḽavhuṱanu", "Mugivhela"]`
        pub const DAY: &[&str] = &["Swondaha", "Musumbuluwo", "Ḽavhuvhili", "Ḽavhuraru", "Ḽavhuṋa", "Ḽavhuṱanu", "Mugivhela"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Phando", "Luhuhi", "Ṱhafamuhwe", "Lambamai", "Shundunthule", "Fulwi", "Fulwana", "Ṱhangule", "Khubvumedzi", "Tshimedzi", "Ḽara", "Nyendavhusiku"]`
        pub const MON: &[&str] = &["Phando", "Luhuhi", "Ṱhafamuhwe", "Lambamai", "Shundunthule", "Fulwi", "Fulwana", "Ṱhangule", "Khubvumedzi", "Tshimedzi", "Ḽara", "Nyendavhusiku"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod vi_VN {
    pub mod LC_ADDRESS {
        /// `Some("VNM")`
        pub const COUNTRY_AB3: Option<&str> = Some("VNM");
        /// `Some("VN")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("VN");
        /// `Some("VN")`
        pub const COUNTRY_CAR: Option<&str> = Some("VN");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Việt Nam")`
        pub const COUNTRY_NAME: Option<&str> = Some("Việt Nam");
        /// `Some(704)`
        pub const COUNTRY_NUM: Option<i64> = Some(704);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("vi")`
        pub const LANG_AB: Option<&str> = Some("vi");
        /// `Some("vie")`
        pub const LANG_LIB: Option<&str> = Some("vie");
        /// `Some("Tiếng Việt")`
        pub const LANG_NAME: Option<&str> = Some("Tiếng Việt");
        /// `Some("vie")`
        pub const LANG_TERM: Option<&str> = Some("vie");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Pablo Saratxaga")`
        pub const CONTACT: Option<&str> = Some("Pablo Saratxaga");
        /// `Some("2004-01-09")`
        pub const DATE: Option<&str> = Some("2004-01-09");
        /// `Some("pablo@mandrakesoft.com")`
        pub const EMAIL: Option<&str> = Some("pablo@mandrakesoft.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Vietnamese")`
        pub const LANGUAGE: Option<&str> = Some("Vietnamese");
        /// `Some("1.1")`
        pub const REVISION: Option<&str> = Some("1.1");
        /// `Some("glibc locale and info from vietnamese native speakers")`
        pub const SOURCE: Option<&str> = Some("glibc locale and info from vietnamese native speakers");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Vietnam")`
        pub const TERRITORY: Option<&str> = Some("Vietnam");
        /// `Some("Vietnamese language locale for Vietnam")`
        pub const TITLE: Option<&str> = Some("Vietnamese language locale for Vietnam");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNkK]"`
        pub const NOEXPR: &str = "^[-0nNkK]";
        /// `Some("Không")`
        pub const NOSTR: Option<&str> = Some("Không");
        /// `"^[+1yYcC]"`
        pub const YESEXPR: &str = "^[+1yYcC]";
        /// `Some("Có")`
        pub const YESSTR: Option<&str> = Some("Có");
    }
    pub mod LC_MONETARY {
        /// `"₫"`
        pub const CURRENCY_SYMBOL: &str = "₫";
        /// `0`
        pub const FRAC_DIGITS: i64 = 0;
        /// `"VND "`
        pub const INT_CURR_SYMBOL: &str = "VND ";
        /// `0`
        pub const INT_FRAC_DIGITS: i64 = 0;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const MON_THOUSANDS_SEP: &str = ".";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `"."`
        pub const THOUSANDS_SEP: &str = ".";
    }
    pub mod LC_TELEPHONE {
        /// `Some("84")`
        pub const INT_PREFIX: Option<&str> = Some("84");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c ;%a ;%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c ;%a ;%l");
    }
    pub mod LC_TIME {
        /// `&["CN", "T2", "T3", "T4", "T5", "T6", "T7"]`
        pub const ABDAY: &[&str] = &["CN", "T2", "T3", "T4", "T5", "T6", "T7"];
        /// `&["Thg 1", "Thg 2", "Thg 3", "Thg 4", "Thg 5", "Thg 6", "Thg 7", "Thg 8", "Thg 9", "Thg 10", "Thg 11", "Thg 12"]`
        pub const ABMON: &[&str] = &["Thg 1", "Thg 2", "Thg 3", "Thg 4", "Thg 5", "Thg 6", "Thg 7", "Thg 8", "Thg 9", "Thg 10", "Thg 11", "Thg 12"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["AM", "PM"]`
        pub const AM_PM: &[&str] = &["AM", "PM"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%A, %d %B năm %Y %T %Z")`
        pub const DATE_FMT: Option<&str> = Some("%A, %d %B năm %Y %T %Z");
        /// `&["Chủ nhật", "Thứ hai", "Thứ ba", "Thứ tư", "Thứ năm", "Thứ sáu", "Thứ bảy"]`
        pub const DAY: &[&str] = &["Chủ nhật", "Thứ hai", "Thứ ba", "Thứ tư", "Thứ năm", "Thứ sáu", "Thứ bảy"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%A, %d %B Năm %Y %T %Z"`
        pub const D_T_FMT: &str = "%A, %d %B Năm %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Tháng 1", "Tháng 2", "Tháng 3", "Tháng 4", "Tháng 5", "Tháng 6", "Tháng 7", "Tháng 8", "Tháng 9", "Tháng 10", "Tháng 11", "Tháng 12"]`
        pub const MON: &[&str] = &["Tháng 1", "Tháng 2", "Tháng 3", "Tháng 4", "Tháng 5", "Tháng 6", "Tháng 7", "Tháng 8", "Tháng 9", "Tháng 10", "Tháng 11", "Tháng 12"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `"%I:%M %p"`
        pub const T_FMT_AMPM: &str = "%I:%M %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod wa_BE {
    pub mod LC_ADDRESS {
        /// `Some("BEL")`
        pub const COUNTRY_AB3: Option<&str> = Some("BEL");
        /// `Some("BE")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("BE");
        /// `Some("B")`
        pub const COUNTRY_CAR: Option<&str> = Some("B");
        /// `Some("2")`
        pub const COUNTRY_ISBN: Option<&str> = Some("2");
        /// `Some("Beldjike")`
        pub const COUNTRY_NAME: Option<&str> = Some("Beldjike");
        /// `Some(56)`
        pub const COUNTRY_NUM: Option<i64> = Some(56);
        /// `Some("B")`
        pub const COUNTRY_POST: Option<&str> = Some("B");
        /// `Some("wa")`
        pub const LANG_AB: Option<&str> = Some("wa");
        /// `Some("wln")`
        pub const LANG_LIB: Option<&str> = Some("wln");
        /// `Some("walon")`
        pub const LANG_NAME: Option<&str> = Some("walon");
        /// `Some("wln")`
        pub const LANG_TERM: Option<&str> = Some("wln");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_NAME"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_NAME"],
        ]);
        /// `Some("Pablo Saratxaga")`
        pub const CONTACT: Option<&str> = Some("Pablo Saratxaga");
        /// `Some("2003-08-25")`
        pub const DATE: Option<&str> = Some("2003-08-25");
        /// `Some("pablo@mandrakesoft.com")`
        pub const EMAIL: Option<&str> = Some("pablo@mandrakesoft.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Walloon")`
        pub const LANGUAGE: Option<&str> = Some("Walloon");
        /// `Some("0.9")`
        pub const REVISION: Option<&str> = Some("0.9");
        /// `Some("Djan SACRE")`
        pub const SOURCE: Option<&str> = Some("Djan SACRE");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Belgium")`
        pub const TERRITORY: Option<&str> = Some("Belgium");
        /// `Some("Walloon Language Locale for Belgium")`
        pub const TITLE: Option<&str> = Some("Walloon Language Locale for Belgium");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("Neni")`
        pub const NOSTR: Option<&str> = Some("Neni");
        /// `"^[+1oOyYaAwW]"`
        pub const YESEXPR: &str = "^[+1oOyYaAwW]";
        /// `Some("Oyi")`
        pub const YESSTR: Option<&str> = Some("Oyi");
    }
    pub use super::fr_BE::LC_MONETARY;
    pub use super::fr_BE::LC_NUMERIC;
    pub use super::fr_BE::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["dim", "lon", "mår", "mie", "dju", "vén", "sem"]`
        pub const ABDAY: &[&str] = &["dim", "lon", "mår", "mie", "dju", "vén", "sem"];
        /// `&["dja", "fev", "mås", "avr", "may", "djn", "djl", "awo", "set", "oct", "nôv", "dec"]`
        pub const ABMON: &[&str] = &["dja", "fev", "mås", "avr", "may", "djn", "djl", "awo", "set", "oct", "nôv", "dec"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `Some(&["djanvî", "fevrî", "måss", "avri", "may", "djun", "djulete", "awousse", "setimbe", "octôbe", "nôvimbe", "decimbe"])`
        pub const ALT_MON: Option<&[&str]> = Some(&["djanvî", "fevrî", "måss", "avri", "may", "djun", "djulete", "awousse", "setimbe", "octôbe", "nôvimbe", "decimbe"]);
        /// `&["AM", "PM"]`
        pub const AM_PM: &[&str] = &["AM", "PM"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["dimegne", "londi", "mårdi", "mierkidi", "djudi", "vénrdi", "semdi"]`
        pub const DAY: &[&str] = &["dimegne", "londi", "mårdi", "mierkidi", "djudi", "vénrdi", "semdi"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"Li %A %d %B %Y %T %Z"`
        pub const D_T_FMT: &str = "Li %A %d %B %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["di djanvî", "di fevrî", "di måss", "d’\u{a0}avri", "di may", "di djun", "di djulete", "d’\u{a0}awousse", "di setimbe", "d’\u{a0}octôbe", "di nôvimbe", "di decimbe"]`
        pub const MON: &[&str] = &["di djanvî", "di fevrî", "di måss", "d’\u{a0}avri", "di may", "di djun", "di djulete", "d’\u{a0}awousse", "di setimbe", "d’\u{a0}octôbe", "di nôvimbe", "di decimbe"];
        /// `"%H:%M:%S"`
        pub const T_FMT: &str = "%H:%M:%S";
        /// `"%I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod wa_BE_euro {
    pub use super::wa_BE::LC_ADDRESS;
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("https://www.gnu.org/software/libc/")`
        pub const ADDRESS: Option<&str> = Some("https://www.gnu.org/software/libc/");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2002-02-23")`
        pub const DATE: Option<&str> = Some("2002-02-23");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Walloon")`
        pub const LANGUAGE: Option<&str> = Some("Walloon");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Free Software Foundation, Inc.")`
        pub const SOURCE: Option<&str> = Some("Free Software Foundation, Inc.");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Belgium")`
        pub const TERRITORY: Option<&str> = Some("Belgium");
        /// `Some("Walloon locale for Belgium with Euro")`
        pub const TITLE: Option<&str> = Some("Walloon locale for Belgium with Euro");
    }
    pub use super::wa_BE::LC_MESSAGES;
    pub use super::fr_BE_euro::LC_MONETARY;
    pub use super::wa_BE::LC_NUMERIC;
    pub use super::wa_BE::LC_TELEPHONE;
    pub use super::wa_BE::LC_TIME;
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod wae_CH {
    pub mod LC_ADDRESS {
        /// `Some("CHE")`
        pub const COUNTRY_AB3: Option<&str> = Some("CHE");
        /// `Some("CH")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("CH");
        /// `Some("CH")`
        pub const COUNTRY_CAR: Option<&str> = Some("CH");
        /// `Some("3")`
        pub const COUNTRY_ISBN: Option<&str> = Some("3");
        /// `Some("Schwiz")`
        pub const COUNTRY_NAME: Option<&str> = Some("Schwiz");
        /// `Some(756)`
        pub const COUNTRY_NUM: Option<i64> = Some(756);
        /// `Some("DH")`
        pub const COUNTRY_POST: Option<&str> = Some("DH");
        /// `Some("")`
        pub const LANG_AB: Option<&str> = Some("");
        /// `Some("wae")`
        pub const LANG_LIB: Option<&str> = Some("wae");
        /// `Some("Walser")`
        pub const LANG_NAME: Option<&str> = Some("Walser");
        /// `Some("wae")`
        pub const LANG_TERM: Option<&str> = Some("wae");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
        ]);
        /// `Some("Walser Translation Team")`
        pub const CONTACT: Option<&str> = Some("Walser Translation Team");
        /// `Some("2010-10-25")`
        pub const DATE: Option<&str> = Some("2010-10-25");
        /// `Some("ml@translate-wae.ch")`
        pub const EMAIL: Option<&str> = Some("ml@translate-wae.ch");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Walser")`
        pub const LANGUAGE: Option<&str> = Some("Walser");
        /// `Some("0.1")`
        pub const REVISION: Option<&str> = Some("0.1");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Switzerland")`
        pub const TERRITORY: Option<&str> = Some("Switzerland");
        /// `Some("Walser locale for Switzerland")`
        pub const TITLE: Option<&str> = Some("Walser locale for Switzerland");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("nei")`
        pub const NOSTR: Option<&str> = Some("nei");
        /// `"^[+1jJyY]"`
        pub const YESEXPR: &str = "^[+1jJyY]";
        /// `Some("ja")`
        pub const YESSTR: Option<&str> = Some("ja");
    }
    pub use super::de_CH::LC_MONETARY;
    pub use super::de_CH::LC_NUMERIC;
    pub use super::de_CH::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Sun", "Män", "Zis", "Mit", "Fro", "Fri", "Sam"]`
        pub const ABDAY: &[&str] = &["Sun", "Män", "Zis", "Mit", "Fro", "Fri", "Sam"];
        /// `&["Jen", "Hor", "Mär", "Abr", "Mei", "Brá", "Hei", "Öig", "Her", "Wím", "Win", "Chr"]`
        pub const ABMON: &[&str] = &["Jen", "Hor", "Mär", "Abr", "Mei", "Brá", "Hei", "Öig", "Her", "Wím", "Win", "Chr"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a %-d %b %H:%M:%S %Z %Y")`
        pub const DATE_FMT: Option<&str> = Some("%a %-d %b %H:%M:%S %Z %Y");
        /// `&["Suntag", "Mäntag", "Zischtag", "Mittwuch", "Frontag", "Fritag", "Samschtag"]`
        pub const DAY: &[&str] = &["Suntag", "Mäntag", "Zischtag", "Mittwuch", "Frontag", "Fritag", "Samschtag"];
        /// `"%Y-%m-%d"`
        pub const D_FMT: &str = "%Y-%m-%d";
        /// `"%a %d. %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d. %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Jenner", "Hornig", "Märze", "Abrille", "Meije", "Bráčet", "Heiwet", "Öigšte", "Herbštmánet", "Wímánet", "Wintermánet", "Chrištmánet"]`
        pub const MON: &[&str] = &["Jenner", "Hornig", "Märze", "Abrille", "Meije", "Bráčet", "Heiwet", "Öigšte", "Herbštmánet", "Wímánet", "Wintermánet", "Chrištmánet"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 4])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 4]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod wal_ET {
    pub mod LC_ADDRESS {
        /// `Some("ETH")`
        pub const COUNTRY_AB3: Option<&str> = Some("ETH");
        /// `Some("ET")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("ET");
        /// `Some("ETH")`
        pub const COUNTRY_CAR: Option<&str> = Some("ETH");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("ኢትዮጵያ")`
        pub const COUNTRY_NAME: Option<&str> = Some("ኢትዮጵያ");
        /// `Some(231)`
        pub const COUNTRY_NUM: Option<i64> = Some(231);
        /// `Some("ETH")`
        pub const COUNTRY_POST: Option<&str> = Some("ETH");
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("wal")`
        pub const LANG_LIB: Option<&str> = Some("wal");
        /// `Some("ወላይታቱ")`
        pub const LANG_NAME: Option<&str> = Some("ወላይታቱ");
        /// `Some("wal")`
        pub const LANG_TERM: Option<&str> = Some("wal");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("7802 Solomon Seal Dr., Springfield, VA 22152, USA")`
        pub const ADDRESS: Option<&str> = Some("7802 Solomon Seal Dr., Springfield, VA 22152, USA");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_TIME"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_TIME"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2004-04-25")`
        pub const DATE: Option<&str> = Some("2004-04-25");
        /// `Some("locales@geez.org")`
        pub const EMAIL: Option<&str> = Some("locales@geez.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Wolaytta")`
        pub const LANGUAGE: Option<&str> = Some("Wolaytta");
        /// `Some("0.10")`
        pub const REVISION: Option<&str> = Some("0.10");
        /// `Some("Ge'ez Frontier Foundation")`
        pub const SOURCE: Option<&str> = Some("Ge'ez Frontier Foundation");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Ethiopia")`
        pub const TERRITORY: Option<&str> = Some("Ethiopia");
        /// `Some("Walaita language locale for Ethiopia.")`
        pub const TITLE: Option<&str> = Some("Walaita language locale for Ethiopia.");
    }
    pub use super::ti_ET::LC_MESSAGES;
    pub use super::ti_ET::LC_MONETARY;
    pub use super::ti_ET::LC_NUMERIC;
    pub use super::ti_ET::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["ወጋ ", "ሳይኖ", "ማቆሳ", "አሩዋ", "ሃሙሳ", "አርባ", "ቄራ "]`
        pub const ABDAY: &[&str] = &["ወጋ ", "ሳይኖ", "ማቆሳ", "አሩዋ", "ሃሙሳ", "አርባ", "ቄራ "];
        /// `&["ጃንዩ", "ፌብሩ", "ማርች", "ኤፕረ", "ሜይ ", "ጁን ", "ጁላይ", "ኦገስ", "ሴፕቴ", "ኦክተ", "ኖቬም", "ዲሴም"]`
        pub const ABMON: &[&str] = &["ጃንዩ", "ፌብሩ", "ማርች", "ኤፕረ", "ሜይ ", "ጁን ", "ጁላይ", "ኦገስ", "ሴፕቴ", "ኦክተ", "ኖቬም", "ዲሴም"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["ማለዶ", "ቃማ"]`
        pub const AM_PM: &[&str] = &["ማለዶ", "ቃማ"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%A፣ %B %e ጋላሳ %r %Z %Y ግ/ላ")`
        pub const DATE_FMT: Option<&str> = Some("%A፣ %B %e ጋላሳ %r %Z %Y ግ/ላ");
        /// `&["ወጋ", "ሳይኖ", "ማቆሳኛ", "አሩዋ", "ሃሙሳ", "አርባ", "ቄራ"]`
        pub const DAY: &[&str] = &["ወጋ", "ሳይኖ", "ማቆሳኛ", "አሩዋ", "ሃሙሳ", "አርባ", "ቄራ"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%A፣ %B %e ጋላሳ %Y %l:%M:%S %p %Z"`
        pub const D_T_FMT: &str = "%A፣ %B %e ጋላሳ %Y %l:%M:%S %p %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["ጃንዩወሪ", "ፌብሩወሪ", "ማርች", "ኤፕረል", "ሜይ", "ጁን", "ጁላይ", "ኦገስት", "ሴፕቴምበር", "ኦክተውበር", "ኖቬምበር", "ዲሴምበር"]`
        pub const MON: &[&str] = &["ጃንዩወሪ", "ፌብሩወሪ", "ማርች", "ኤፕረል", "ሜይ", "ጁን", "ጁላይ", "ኦገስት", "ሴፕቴምበር", "ኦክተውበር", "ኖቬምበር", "ዲሴምበር"];
        /// `"%l:%M:%S %p"`
        pub const T_FMT: &str = "%l:%M:%S %p";
        /// `"%l:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%l:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod wo_SN {
    pub mod LC_ADDRESS {
        /// `Some("SEN")`
        pub const COUNTRY_AB3: Option<&str> = Some("SEN");
        /// `Some("SN")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("SN");
        /// `Some("SN")`
        pub const COUNTRY_CAR: Option<&str> = Some("SN");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Senegaal")`
        pub const COUNTRY_NAME: Option<&str> = Some("Senegaal");
        /// `Some(686)`
        pub const COUNTRY_NUM: Option<i64> = Some(686);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("wo")`
        pub const LANG_AB: Option<&str> = Some("wo");
        /// `Some("wol")`
        pub const LANG_LIB: Option<&str> = Some("wol");
        /// `Some("Wolof")`
        pub const LANG_NAME: Option<&str> = Some("Wolof");
        /// `Some("wol")`
        pub const LANG_TERM: Option<&str> = Some("wol");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Christian Perrier")`
        pub const CONTACT: Option<&str> = Some("Christian Perrier");
        /// `Some("2004-09-08")`
        pub const DATE: Option<&str> = Some("2004-09-08");
        /// `Some("bubulle@debian.org")`
        pub const EMAIL: Option<&str> = Some("bubulle@debian.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Wolof")`
        pub const LANGUAGE: Option<&str> = Some("Wolof");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("The Debian Project")`
        pub const SOURCE: Option<&str> = Some("The Debian Project");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Senegal")`
        pub const TERRITORY: Option<&str> = Some("Senegal");
        /// `Some("Wolof locale for Senegal")`
        pub const TITLE: Option<&str> = Some("Wolof locale for Senegal");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0dDnN]"`
        pub const NOEXPR: &str = "^[-0dDnN]";
        /// `Some("déedet")`
        pub const NOSTR: Option<&str> = Some("déedet");
        /// `"^[+1wWyY]"`
        pub const YESEXPR: &str = "^[+1wWyY]";
        /// `Some("waaw")`
        pub const YESSTR: Option<&str> = Some("waaw");
    }
    pub mod LC_MONETARY {
        /// `"CFA"`
        pub const CURRENCY_SYMBOL: &str = "CFA";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"XOF "`
        pub const INT_CURR_SYMBOL: &str = "XOF ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `","`
        pub const MON_DECIMAL_POINT: &str = ",";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `"\u{202f}"`
        pub const MON_THOUSANDS_SEP: &str = "\u{202f}";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `0`
        pub const N_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `0`
        pub const P_CS_PRECEDES: i64 = 0;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `","`
        pub const DECIMAL_POINT: &str = ",";
        /// `&[0, 0]`
        pub const GROUPING: &[i64] = &[0, 0];
        /// `""`
        pub const THOUSANDS_SEP: &str = "";
    }
    pub mod LC_TELEPHONE {
        /// `Some("221")`
        pub const INT_PREFIX: Option<&str> = Some("221");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `Some("%a %l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("%a %l");
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["dib", "alt", "tal", "all", "alx", "ajj", "gaa"]`
        pub const ABDAY: &[&str] = &["dib", "alt", "tal", "all", "alx", "ajj", "gaa"];
        /// `&["san", "fee", "mar", "awr", "me ", "suw", "sul", "uut", "sep", "okt", "now", "des"]`
        pub const ABMON: &[&str] = &["san", "fee", "mar", "awr", "me ", "suw", "sul", "uut", "sep", "okt", "now", "des"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["dib'eer", "altine", "talaata", "allarba", "alxames", "ajjuma", "gaawu"]`
        pub const DAY: &[&str] = &["dib'eer", "altine", "talaata", "allarba", "alxames", "ajjuma", "gaawu"];
        /// `"%d.%m.%Y"`
        pub const D_FMT: &str = "%d.%m.%Y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["sanwiy'e", "feebriy'e", "mars", "awril", "me", "suwen", "sulet", "uut", "septaambar", "oktoobar", "nowaambar", "desaambar"]`
        pub const MON: &[&str] = &["sanwiy'e", "feebriy'e", "mars", "awril", "me", "suwen", "sulet", "uut", "septaambar", "oktoobar", "nowaambar", "desaambar"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod xh_ZA {
    pub mod LC_ADDRESS {
        /// `Some("ZAF")`
        pub const COUNTRY_AB3: Option<&str> = Some("ZAF");
        /// `Some("ZA")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("ZA");
        /// `Some("ZA")`
        pub const COUNTRY_CAR: Option<&str> = Some("ZA");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("uMzantsi Afrika")`
        pub const COUNTRY_NAME: Option<&str> = Some("uMzantsi Afrika");
        /// `Some(710)`
        pub const COUNTRY_NUM: Option<i64> = Some(710);
        /// `Some("ZA")`
        pub const COUNTRY_POST: Option<&str> = Some("ZA");
        /// `Some("xh")`
        pub const LANG_AB: Option<&str> = Some("xh");
        /// `Some("xho")`
        pub const LANG_LIB: Option<&str> = Some("xho");
        /// `Some("isiXhosa")`
        pub const LANG_NAME: Option<&str> = Some("isiXhosa");
        /// `Some("xho")`
        pub const LANG_TERM: Option<&str> = Some("xho");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("PO Box 28364, Sunnyside, 0132, South Africa")`
        pub const ADDRESS: Option<&str> = Some("PO Box 28364, Sunnyside, 0132, South Africa");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Dwayne Bailey")`
        pub const CONTACT: Option<&str> = Some("Dwayne Bailey");
        /// `Some("2005-10-13")`
        pub const DATE: Option<&str> = Some("2005-10-13");
        /// `Some("dwayne@translate.org.za")`
        pub const EMAIL: Option<&str> = Some("dwayne@translate.org.za");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Xhosa")`
        pub const LANGUAGE: Option<&str> = Some("Xhosa");
        /// `Some("0.3")`
        pub const REVISION: Option<&str> = Some("0.3");
        /// `Some("Zuza Software Foundation (Translate.org.za)")`
        pub const SOURCE: Option<&str> = Some("Zuza Software Foundation (Translate.org.za)");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("South Africa")`
        pub const TERRITORY: Option<&str> = Some("South Africa");
        /// `Some("Xhosa locale for South Africa")`
        pub const TITLE: Option<&str> = Some("Xhosa locale for South Africa");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNhH]"`
        pub const NOEXPR: &str = "^[-0nNhH]";
        /// `Some("hayi")`
        pub const NOSTR: Option<&str> = Some("hayi");
        /// `"^[+1yYeE]"`
        pub const YESEXPR: &str = "^[+1yYeE]";
        /// `Some("ewe")`
        pub const YESSTR: Option<&str> = Some("ewe");
    }
    pub use super::en_ZA::LC_MONETARY;
    pub use super::en_ZA::LC_NUMERIC;
    pub use super::en_ZA::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Caw", "Mvu", "Bin", "Tha", "Sin", "Hla", "Mgq"]`
        pub const ABDAY: &[&str] = &["Caw", "Mvu", "Bin", "Tha", "Sin", "Hla", "Mgq"];
        /// `&["Mqu", "Mdu", "Kwi", "Tsh", "Can", "Sil", "Kha", "Thu", "Msi", "Dwa", "Nka", "Mng"]`
        pub const ABMON: &[&str] = &["Mqu", "Mdu", "Kwi", "Tsh", "Can", "Sil", "Kha", "Thu", "Msi", "Dwa", "Nka", "Mng"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%a %b %-e %H:%M:%S %Z %Y")`
        pub const DATE_FMT: Option<&str> = Some("%a %b %-e %H:%M:%S %Z %Y");
        /// `&["iCawa", "uMvulo", "lwesiBini", "lwesiThathu", "ulweSine", "lwesiHlanu", "uMgqibelo"]`
        pub const DAY: &[&str] = &["iCawa", "uMvulo", "lwesiBini", "lwesiThathu", "ulweSine", "lwesiHlanu", "uMgqibelo"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%a %-e %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %-e %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["eyoMqungu", "eyoMdumba", "eyoKwindla", "uTshazimpuzi", "uCanzibe", "eyeSilimela", "eyeKhala", "eyeThupa", "eyoMsintsi", "eyeDwarha", "eyeNkanga", "eyoMnga"]`
        pub const MON: &[&str] = &["eyoMqungu", "eyoMdumba", "eyoKwindla", "uTshazimpuzi", "uCanzibe", "eyeSilimela", "eyeKhala", "eyeThupa", "eyoMsintsi", "eyeDwarha", "eyeNkanga", "eyoMnga"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod yi_US {
    pub mod LC_ADDRESS {
        /// `Some("USA")`
        pub const COUNTRY_AB3: Option<&str> = Some("USA");
        /// `Some("US")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("US");
        /// `Some("USA")`
        pub const COUNTRY_CAR: Option<&str> = Some("USA");
        /// `Some("0")`
        pub const COUNTRY_ISBN: Option<&str> = Some("0");
        /// `Some("פ\u{5bf}א\u{5b7}ראייניגטע שטא\u{5b7}טן")`
        pub const COUNTRY_NAME: Option<&str> = Some("פ\u{5bf}א\u{5b7}ראייניגטע שטא\u{5b7}טן");
        /// `Some(840)`
        pub const COUNTRY_NUM: Option<i64> = Some(840);
        /// `Some("USA")`
        pub const COUNTRY_POST: Option<&str> = Some("USA");
        /// `Some("yi")`
        pub const LANG_AB: Option<&str> = Some("yi");
        /// `Some("yid")`
        pub const LANG_LIB: Option<&str> = Some("yid");
        /// `Some("יי\u{5b4}דיש")`
        pub const LANG_NAME: Option<&str> = Some("יי\u{5b4}דיש");
        /// `Some("yid")`
        pub const LANG_TERM: Option<&str> = Some("yid");
        /// `Some("%d%N%f%N%d%N%b%N%s %h 5e %r%N%C%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%d%N%f%N%d%N%b%N%s %h 5e %r%N%C%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_MEASUREMENT"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_ADDRESS"],
        ]);
        /// `Some("Pablo Saratxaga")`
        pub const CONTACT: Option<&str> = Some("Pablo Saratxaga");
        /// `Some("2003-08-16")`
        pub const DATE: Option<&str> = Some("2003-08-16");
        /// `Some("pablo@mandrakesoft.com")`
        pub const EMAIL: Option<&str> = Some("pablo@mandrakesoft.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Yiddish")`
        pub const LANGUAGE: Option<&str> = Some("Yiddish");
        /// `Some("0.4")`
        pub const REVISION: Option<&str> = Some("0.4");
        /// `Some("http://www.uyip.org/")`
        pub const SOURCE: Option<&str> = Some("http://www.uyip.org/");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("United States")`
        pub const TERRITORY: Option<&str> = Some("United States");
        /// `Some("Yiddish Language locale for the USA")`
        pub const TITLE: Option<&str> = Some("Yiddish Language locale for the USA");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNנק]"`
        pub const NOEXPR: &str = "^[-0nNנק]";
        /// `Some("קײן")`
        pub const NOSTR: Option<&str> = Some("קײן");
        /// `"^[+1yYי]"`
        pub const YESEXPR: &str = "^[+1yYי]";
        /// `Some("יא\u{5b8}")`
        pub const YESSTR: Option<&str> = Some("יא\u{5b8}");
    }
    pub mod LC_MONETARY {
        /// `"$"`
        pub const CURRENCY_SYMBOL: &str = "$";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"USD "`
        pub const INT_CURR_SYMBOL: &str = "USD ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const N_SEP_BY_SPACE: i64 = 1;
        /// `2`
        pub const N_SIGN_POSN: i64 = 2;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `1`
        pub const P_SEP_BY_SPACE: i64 = 1;
        /// `2`
        pub const P_SIGN_POSN: i64 = 2;
    }
    pub use super::en_US::LC_NUMERIC;
    pub use super::en_US::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["זונ'", "מא\u{5b8}נ'", "דינ'", "מיט'", "דא\u{5b8}נ'", "פ\u{5bf}רײ\u{5b7}'", "שבת"]`
        pub const ABDAY: &[&str] = &["זונ'", "מא\u{5b8}נ'", "דינ'", "מיט'", "דא\u{5b8}נ'", "פ\u{5bf}רײ\u{5b7}'", "שבת"];
        /// `&["יא\u{5b7}נ", "פ\u{5bf}עב", "מא\u{5b7}ר", "א\u{5b7}פ\u{5bc}ר", "מײ\u{5b7} ", "יונ", "יול", "אױג", "סעפ\u{5bc}", "א\u{5b8}קט", "נא\u{5b8}װ", "דעצ"]`
        pub const ABMON: &[&str] = &["יא\u{5b7}נ", "פ\u{5bf}עב", "מא\u{5b7}ר", "א\u{5b7}פ\u{5bc}ר", "מײ\u{5b7} ", "יונ", "יול", "אױג", "סעפ\u{5bc}", "א\u{5b8}קט", "נא\u{5b8}װ", "דעצ"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["AM", "PM"]`
        pub const AM_PM: &[&str] = &["AM", "PM"];
        /// `Some(3)`
        pub const CAL_DIRECTION: Option<i64> = Some(3);
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["זונטיק", "מא\u{5b8}נטיק", "דינסטיק", "מיטװא\u{5b8}ך", "דא\u{5b8}נערשטיק", "פ\u{5bf}רײ\u{5b7}טיק", "שבת"]`
        pub const DAY: &[&str] = &["זונטיק", "מא\u{5b8}נטיק", "דינסטיק", "מיטװא\u{5b8}ך", "דא\u{5b8}נערשטיק", "פ\u{5bf}רײ\u{5b7}טיק", "שבת"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%Z %H:%M:%S %Y %b %d %a"`
        pub const D_T_FMT: &str = "%Z %H:%M:%S %Y %b %d %a";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["יא\u{5b7}נוא\u{5b7}ר", "פ\u{5bf}עברוא\u{5b7}ר", "מערץ", "א\u{5b7}פ\u{5bc}ריל", "מיי", "יוני", "יולי", "אויגוסט", "סעפ\u{5bc}טעמבער", "אקטאבער", "נאוועמבער", "דעצעמבער"]`
        pub const MON: &[&str] = &["יא\u{5b7}נוא\u{5b7}ר", "פ\u{5bf}עברוא\u{5b7}ר", "מערץ", "א\u{5b7}פ\u{5bc}ריל", "מיי", "יוני", "יולי", "אויגוסט", "סעפ\u{5bc}טעמבער", "אקטאבער", "נאוועמבער", "דעצעמבער"];
        /// `"%H:%M:%S"`
        pub const T_FMT: &str = "%H:%M:%S";
        /// `"%I:%M:%S %P"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %P";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod yo_NG {
    pub mod LC_ADDRESS {
        /// `Some("NGA")`
        pub const COUNTRY_AB3: Option<&str> = Some("NGA");
        /// `Some("NG")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("NG");
        /// `Some("WAN")`
        pub const COUNTRY_CAR: Option<&str> = Some("WAN");
        /// `Some("978")`
        pub const COUNTRY_ISBN: Option<&str> = Some("978");
        /// `Some("Orilẹ\u{300}-èdè Nàìjíríà")`
        pub const COUNTRY_NAME: Option<&str> = Some("Orilẹ\u{300}-èdè Nàìjíríà");
        /// `Some(566)`
        pub const COUNTRY_NUM: Option<i64> = Some(566);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("yo")`
        pub const LANG_AB: Option<&str> = Some("yo");
        /// `Some("yor")`
        pub const LANG_LIB: Option<&str> = Some("yor");
        /// `Some("Èdè Yorùbá")`
        pub const LANG_NAME: Option<&str> = Some("Èdè Yorùbá");
        /// `Some("yor")`
        pub const LANG_TERM: Option<&str> = Some("yor");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2005-11-20")`
        pub const DATE: Option<&str> = Some("2005-11-20");
        /// `Some("pablo@mandriva.com")`
        pub const EMAIL: Option<&str> = Some("pablo@mandriva.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Yoruba")`
        pub const LANGUAGE: Option<&str> = Some("Yoruba");
        /// `Some("0.2")`
        pub const REVISION: Option<&str> = Some("0.2");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Nigeria")`
        pub const TERRITORY: Option<&str> = Some("Nigeria");
        /// `Some("Yoruba locale for Nigeria")`
        pub const TITLE: Option<&str> = Some("Yoruba locale for Nigeria");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0rROoKk]"`
        pub const NOEXPR: &str = "^[-0rROoKk]";
        /// `Some("Bẹ\u{301}ẹ\u{300}kọ\u{301}")`
        pub const NOSTR: Option<&str> = Some("Bẹ\u{301}ẹ\u{300}kọ\u{301}");
        /// `"^[+1EeyYNn]"`
        pub const YESEXPR: &str = "^[+1EeyYNn]";
        /// `Some("Bẹ\u{301}ẹ\u{300}ni")`
        pub const YESSTR: Option<&str> = Some("Bẹ\u{301}ẹ\u{300}ni");
    }
    pub use super::en_NG::LC_MONETARY;
    pub use super::en_NG::LC_NUMERIC;
    pub use super::en_NG::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Àìkú", "Ajé", "Ìsẹ\u{301}gun", "Ọjọ\u{301}rú", "Ọjọ\u{301}bọ", "Ẹtì", "Àbámẹ\u{301}ta"]`
        pub const ABDAY: &[&str] = &["Àìkú", "Ajé", "Ìsẹ\u{301}gun", "Ọjọ\u{301}rú", "Ọjọ\u{301}bọ", "Ẹtì", "Àbámẹ\u{301}ta"];
        /// `&["Ṣẹ\u{301}rẹ\u{301}", "Èrèlè", "Ẹrẹ\u{300}nà", "Ìgbé", "Ẹ\u{300}bibi", "Òkúdu", "Agẹmọ", "Ògún", "Owewe", "Ọ\u{300}wàrà", "Bélú", "Ọ\u{300}pẹ\u{300}"]`
        pub const ABMON: &[&str] = &["Ṣẹ\u{301}rẹ\u{301}", "Èrèlè", "Ẹrẹ\u{300}nà", "Ìgbé", "Ẹ\u{300}bibi", "Òkúdu", "Agẹmọ", "Ògún", "Owewe", "Ọ\u{300}wàrà", "Bélú", "Ọ\u{300}pẹ\u{300}"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["AM", "PM"]`
        pub const AM_PM: &[&str] = &["AM", "PM"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["Ọjọ\u{301} Àìkú", "Ọjọ\u{301} Ajé", "Ọjọ\u{301} Ìsẹ\u{301}gun", "Ọjọ\u{301}rú", "Ọjọ\u{301}bọ", "Ọjọ\u{301} Ẹtì", "Ọjọ\u{301} Àbámẹ\u{301}ta"]`
        pub const DAY: &[&str] = &["Ọjọ\u{301} Àìkú", "Ọjọ\u{301} Ajé", "Ọjọ\u{301} Ìsẹ\u{301}gun", "Ọjọ\u{301}rú", "Ọjọ\u{301}bọ", "Ọjọ\u{301} Ẹtì", "Ọjọ\u{301} Àbámẹ\u{301}ta"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"ọjọ\u{301} %a, %d oṣù %b ọdún %Y %T %Z"`
        pub const D_T_FMT: &str = "ọjọ\u{301} %a, %d oṣù %b ọdún %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Oṣù Ṣẹ\u{301}rẹ\u{301}", "Oṣù Èrèlè", "Oṣù Ẹrẹ\u{300}nà", "Oṣù Ìgbé", "Oṣù Ẹ\u{300}bibi", "Oṣù Òkúdu", "Oṣù Agẹmọ", "Oṣù Ògún", "Oṣù Owewe", "Oṣù Ọ\u{300}wàrà", "Oṣù Bélú", "Oṣù Ọ\u{300}pẹ\u{300}"]`
        pub const MON: &[&str] = &["Oṣù Ṣẹ\u{301}rẹ\u{301}", "Oṣù Èrèlè", "Oṣù Ẹrẹ\u{300}nà", "Oṣù Ìgbé", "Oṣù Ẹ\u{300}bibi", "Oṣù Òkúdu", "Oṣù Agẹmọ", "Oṣù Ògún", "Oṣù Owewe", "Oṣù Ọ\u{300}wàrà", "Oṣù Bélú", "Oṣù Ọ\u{300}pẹ\u{300}"];
        /// `"%I:%M:%S %p"`
        pub const T_FMT: &str = "%I:%M:%S %p";
        /// `"%I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod yue_HK {
    pub mod LC_ADDRESS {
        /// `Some("HKG")`
        pub const COUNTRY_AB3: Option<&str> = Some("HKG");
        /// `Some("HK")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("HK");
        /// `Some("HK")`
        pub const COUNTRY_CAR: Option<&str> = Some("HK");
        /// `Some("962")`
        pub const COUNTRY_ISBN: Option<&str> = Some("962");
        /// `Some("中華人民共和國香港特別行政區")`
        pub const COUNTRY_NAME: Option<&str> = Some("中華人民共和國香港特別行政區");
        /// `Some(344)`
        pub const COUNTRY_NUM: Option<i64> = Some(344);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `None`
        pub const LANG_AB: Option<&str> = None;
        /// `Some("yue")`
        pub const LANG_LIB: Option<&str> = Some("yue");
        /// `Some("粵語")`
        pub const LANG_NAME: Option<&str> = Some("粵語");
        /// `Some("yue")`
        pub const LANG_TERM: Option<&str> = Some("yue");
        /// `Some("%c%T%s%h%N%b%e%r%N%f%N%d%N%a")`
        pub const POSTAL_FMT: Option<&str> = Some("%c%T%s%h%N%b%e%r%N%f%N%d%N%a");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2010-11-07")`
        pub const DATE: Option<&str> = Some("2010-11-07");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Cantonese")`
        pub const LANGUAGE: Option<&str> = Some("Cantonese");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Hong Kong SAR China")`
        pub const TERRITORY: Option<&str> = Some("Hong Kong SAR China");
        /// `Some("Yue Chinese (Cantonese) language locale for Hong Kong")`
        pub const TITLE: Option<&str> = Some("Yue Chinese (Cantonese) language locale for Hong Kong");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nN]"`
        pub const NOEXPR: &str = "^[-0nN]";
        /// `Some("唔係")`
        pub const NOSTR: Option<&str> = Some("唔係");
        /// `"^[+1yY]"`
        pub const YESEXPR: &str = "^[+1yY]";
        /// `Some("係")`
        pub const YESSTR: Option<&str> = Some("係");
    }
    pub use super::zh_HK::LC_MONETARY;
    pub use super::zh_HK::LC_NUMERIC;
    pub use super::zh_HK::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["日", "一", "二", "三", "四", "五", "六"]`
        pub const ABDAY: &[&str] = &["日", "一", "二", "三", "四", "五", "六"];
        /// `&["1月", "2月", "3月", "4月", "5月", "6月", "7月", "8月", "9月", "10月", "11月", "12月"]`
        pub const ABMON: &[&str] = &["1月", "2月", "3月", "4月", "5月", "6月", "7月", "8月", "9月", "10月", "11月", "12月"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["上午", "下午"]`
        pub const AM_PM: &[&str] = &["上午", "下午"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%Y年 %b %e日 %A %H:%M:%S %Z")`
        pub const DATE_FMT: Option<&str> = Some("%Y年 %b %e日 %A %H:%M:%S %Z");
        /// `&["星期日", "星期一", "星期二", "星期三", "星期四", "星期五", "星期六"]`
        pub const DAY: &[&str] = &["星期日", "星期一", "星期二", "星期三", "星期四", "星期五", "星期六"];
        /// `"%Y年%m月%d日 %A"`
        pub const D_FMT: &str = "%Y年%m月%d日 %A";
        /// `"%Y年%m月%d日 %A %H點%M分%S秒"`
        pub const D_T_FMT: &str = "%Y年%m月%d日 %A %H點%M分%S秒";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["1月", "2月", "3月", "4月", "5月", "6月", "7月", "8月", "9月", "10月", "11月", "12月"]`
        pub const MON: &[&str] = &["1月", "2月", "3月", "4月", "5月", "6月", "7月", "8月", "9月", "10月", "11月", "12月"];
        /// `"%H點%M分%S秒"`
        pub const T_FMT: &str = "%H點%M分%S秒";
        /// `"%p%I點%M分%S秒"`
        pub const T_FMT_AMPM: &str = "%p%I點%M分%S秒";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod yuw_PG {
    pub mod LC_ADDRESS {
        /// `Some("PNG")`
        pub const COUNTRY_AB3: Option<&str> = Some("PNG");
        /// `Some("PG")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("PG");
        /// `Some("PNG")`
        pub const COUNTRY_CAR: Option<&str> = Some("PNG");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("Papua New Guinea")`
        pub const COUNTRY_NAME: Option<&str> = Some("Papua New Guinea");
        /// `Some(598)`
        pub const COUNTRY_NUM: Option<i64> = Some(598);
        /// `Some("")`
        pub const COUNTRY_POST: Option<&str> = Some("");
        /// `Some("")`
        pub const LANG_AB: Option<&str> = Some("");
        /// `Some("yuw")`
        pub const LANG_LIB: Option<&str> = Some("yuw");
        /// `Some("Uruwa")`
        pub const LANG_NAME: Option<&str> = Some("Uruwa");
        /// `Some("yuw")`
        pub const LANG_TERM: Option<&str> = Some("yuw");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Hannah Sarvasy")`
        pub const CONTACT: Option<&str> = Some("Hannah Sarvasy");
        /// `Some("2016-12-07")`
        pub const DATE: Option<&str> = Some("2016-12-07");
        /// `Some("nungon.localization@gmail.com")`
        pub const EMAIL: Option<&str> = Some("nungon.localization@gmail.com");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Yau")`
        pub const LANGUAGE: Option<&str> = Some("Yau");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("Information from native speakers")`
        pub const SOURCE: Option<&str> = Some("Information from native speakers");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Papua New Guinea")`
        pub const TERRITORY: Option<&str> = Some("Papua New Guinea");
        /// `Some("YauNungon locale for Papua New Guinea")`
        pub const TITLE: Option<&str> = Some("YauNungon locale for Papua New Guinea");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNmM]"`
        pub const NOEXPR: &str = "^[-0nNmM]";
        /// `Some("muuno")`
        pub const NOSTR: Option<&str> = Some("muuno");
        /// `"^[+1yYöÖ]"`
        pub const YESEXPR: &str = "^[+1yYöÖ]";
        /// `Some("öö")`
        pub const YESSTR: Option<&str> = Some("öö");
    }
    pub mod LC_MONETARY {
        /// `"K"`
        pub const CURRENCY_SYMBOL: &str = "K";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"PGK "`
        pub const INT_CURR_SYMBOL: &str = "PGK ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const MON_GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3, 3]`
        pub const GROUPING: &[i64] = &[3, 3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("675")`
        pub const INT_PREFIX: Option<&str> = Some("675");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["sön", "mön", "sin", "mit", "soi", "nen", "sab"]`
        pub const ABDAY: &[&str] = &["sön", "mön", "sin", "mit", "soi", "nen", "sab"];
        /// `&["jen", "feb", "mas", "epr", "mei", "jun", "jul", "ögu", "sep", "ökt", "nöw", "dis"]`
        pub const ABMON: &[&str] = &["jen", "feb", "mas", "epr", "mei", "jun", "jul", "ögu", "sep", "ökt", "nöw", "dis"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["AM", "PM"]`
        pub const AM_PM: &[&str] = &["AM", "PM"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["sönda", "mönda", "sinda", "mitiwö", "sogipbono", "nenggo", "söndanggie"]`
        pub const DAY: &[&str] = &["sönda", "mönda", "sinda", "mitiwö", "sogipbono", "nenggo", "söndanggie"];
        /// `"%d/%m/%y"`
        pub const D_FMT: &str = "%d/%m/%y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["jenuari", "febuari", "mas", "epril", "mei", "jun", "julai", "ögus", "septemba", "öktoba", "nöwemba", "diksemba"]`
        pub const MON: &[&str] = &["jenuari", "febuari", "mas", "epril", "mei", "jun", "julai", "ögus", "septemba", "öktoba", "nöwemba", "diksemba"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `"%I:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%I:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod zh_CN {
    pub mod LC_ADDRESS {
        /// `Some("CHN")`
        pub const COUNTRY_AB3: Option<&str> = Some("CHN");
        /// `Some("CN")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("CN");
        /// `Some("CHN")`
        pub const COUNTRY_CAR: Option<&str> = Some("CHN");
        /// `Some("7")`
        pub const COUNTRY_ISBN: Option<&str> = Some("7");
        /// `Some("中华人民共和国")`
        pub const COUNTRY_NAME: Option<&str> = Some("中华人民共和国");
        /// `Some(156)`
        pub const COUNTRY_NUM: Option<i64> = Some(156);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("zh")`
        pub const LANG_AB: Option<&str> = Some("zh");
        /// `Some("chi")`
        pub const LANG_LIB: Option<&str> = Some("chi");
        /// `Some("中文")`
        pub const LANG_NAME: Option<&str> = Some("中文");
        /// `Some("zho")`
        pub const LANG_TERM: Option<&str> = Some("zho");
        /// `Some("%c%N%T%N%s %h %e %r%N%b%N%d%N%f%N%a%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%c%N%T%N%s %h %e %r%N%b%N%d%N%f%N%a%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-07-25")`
        pub const DATE: Option<&str> = Some("2000-07-25");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Chinese")`
        pub const LANGUAGE: Option<&str> = Some("Chinese");
        /// `Some("0.1")`
        pub const REVISION: Option<&str> = Some("0.1");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("China")`
        pub const TERRITORY: Option<&str> = Some("China");
        /// `Some("Chinese locale for Peoples Republic of China")`
        pub const TITLE: Option<&str> = Some("Chinese locale for Peoples Republic of China");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNｎＮ不否]"`
        pub const NOEXPR: &str = "^[-0nNｎＮ不否]";
        /// `Some("不是")`
        pub const NOSTR: Option<&str> = Some("不是");
        /// `"^[+1yYｙＹ是]"`
        pub const YESEXPR: &str = "^[+1yYｙＹ是]";
        /// `Some("是")`
        pub const YESSTR: Option<&str> = Some("是");
    }
    pub mod LC_MONETARY {
        /// `"￥"`
        pub const CURRENCY_SYMBOL: &str = "￥";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"CNY "`
        pub const INT_CURR_SYMBOL: &str = "CNY ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `Some(1)`
        pub const INT_N_CS_PRECEDES: Option<i64> = Some(1);
        /// `Some(0)`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = Some(0);
        /// `Some(1)`
        pub const INT_N_SIGN_POSN: Option<i64> = Some(1);
        /// `Some(1)`
        pub const INT_P_CS_PRECEDES: Option<i64> = Some(1);
        /// `Some(0)`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = Some(0);
        /// `Some(1)`
        pub const INT_P_SIGN_POSN: Option<i64> = Some(1);
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `4`
        pub const N_SIGN_POSN: i64 = 4;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `4`
        pub const P_SIGN_POSN: i64 = 4;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("86")`
        pub const INT_PREFIX: Option<&str> = Some("86");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `Some("0%a %l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("0%a %l");
        /// `Some("+%c %a %l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c %a %l");
    }
    pub mod LC_TIME {
        /// `&["日", "一", "二", "三", "四", "五", "六"]`
        pub const ABDAY: &[&str] = &["日", "一", "二", "三", "四", "五", "六"];
        /// `&["1月", "2月", "3月", "4月", "5月", "6月", "7月", "8月", "9月", "10月", "11月", "12月"]`
        pub const ABMON: &[&str] = &["1月", "2月", "3月", "4月", "5月", "6月", "7月", "8月", "9月", "10月", "11月", "12月"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["上午", "下午"]`
        pub const AM_PM: &[&str] = &["上午", "下午"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `Some("%Y年 %m月 %d日 %A %H:%M:%S %Z")`
        pub const DATE_FMT: Option<&str> = Some("%Y年 %m月 %d日 %A %H:%M:%S %Z");
        /// `&["星期日", "星期一", "星期二", "星期三", "星期四", "星期五", "星期六"]`
        pub const DAY: &[&str] = &["星期日", "星期一", "星期二", "星期三", "星期四", "星期五", "星期六"];
        /// `"%Y年%m月%d日"`
        pub const D_FMT: &str = "%Y年%m月%d日";
        /// `"%Y年%m月%d日 %A %H时%M分%S秒"`
        pub const D_T_FMT: &str = "%Y年%m月%d日 %A %H时%M分%S秒";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `Some(2)`
        pub const FIRST_WEEKDAY: Option<i64> = Some(2);
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["一月", "二月", "三月", "四月", "五月", "六月", "七月", "八月", "九月", "十月", "十一月", "十二月"]`
        pub const MON: &[&str] = &["一月", "二月", "三月", "四月", "五月", "六月", "七月", "八月", "九月", "十月", "十一月", "十二月"];
        /// `"%H时%M分%S秒"`
        pub const T_FMT: &str = "%H时%M分%S秒";
        /// `"%p %I时%M分%S秒"`
        pub const T_FMT_AMPM: &str = "%p %I时%M分%S秒";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod zh_HK {
    pub mod LC_ADDRESS {
        /// `Some("HKG")`
        pub const COUNTRY_AB3: Option<&str> = Some("HKG");
        /// `Some("HK")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("HK");
        /// `Some("HK")`
        pub const COUNTRY_CAR: Option<&str> = Some("HK");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `None`
        pub const COUNTRY_NAME: Option<&str> = None;
        /// `Some(344)`
        pub const COUNTRY_NUM: Option<i64> = Some(344);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("zh")`
        pub const LANG_AB: Option<&str> = Some("zh");
        /// `Some("chi")`
        pub const LANG_LIB: Option<&str> = Some("chi");
        /// `Some("繁體中文")`
        pub const LANG_NAME: Option<&str> = Some("繁體中文");
        /// `Some("zho")`
        pub const LANG_TERM: Option<&str> = Some("zho");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan")`
        pub const ADDRESS: Option<&str> = Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-07-20")`
        pub const DATE: Option<&str> = Some("2000-07-20");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Chinese")`
        pub const LANGUAGE: Option<&str> = Some("Chinese");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("IBM Globalization Center of Competency, Yamato Software Laboratory")`
        pub const SOURCE: Option<&str> = Some("IBM Globalization Center of Competency, Yamato Software Laboratory");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Hong Kong SAR China")`
        pub const TERRITORY: Option<&str> = Some("Hong Kong SAR China");
        /// `Some("Chinese language locale for Hong Kong")`
        pub const TITLE: Option<&str> = Some("Chinese language locale for Hong Kong");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNｎＮ不否]"`
        pub const NOEXPR: &str = "^[-0nNｎＮ不否]";
        /// `Some("否")`
        pub const NOSTR: Option<&str> = Some("否");
        /// `"^[+1yYｙＹ是]"`
        pub const YESEXPR: &str = "^[+1yYｙＹ是]";
        /// `Some("是")`
        pub const YESSTR: Option<&str> = Some("是");
    }
    pub mod LC_MONETARY {
        /// `"HK$"`
        pub const CURRENCY_SYMBOL: &str = "HK$";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"HKD "`
        pub const INT_CURR_SYMBOL: &str = "HKD ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `0`
        pub const N_SIGN_POSN: i64 = 0;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("852")`
        pub const INT_PREFIX: Option<&str> = Some("852");
        /// `Some("001")`
        pub const INT_SELECT: Option<&str> = Some("001");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c ;%a ;%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c ;%a ;%l");
    }
    pub mod LC_TIME {
        /// `&["日", "一", "二", "三", "四", "五", "六"]`
        pub const ABDAY: &[&str] = &["日", "一", "二", "三", "四", "五", "六"];
        /// `&["1月", "2月", "3月", "4月", "5月", "6月", "7月", "8月", "9月", "10月", "11月", "12月"]`
        pub const ABMON: &[&str] = &["1月", "2月", "3月", "4月", "5月", "6月", "7月", "8月", "9月", "10月", "11月", "12月"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["上午", "下午"]`
        pub const AM_PM: &[&str] = &["上午", "下午"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["星期日", "星期一", "星期二", "星期三", "星期四", "星期五", "星期六"]`
        pub const DAY: &[&str] = &["星期日", "星期一", "星期二", "星期三", "星期四", "星期五", "星期六"];
        /// `"%Y年%m月%d日 %A"`
        pub const D_FMT: &str = "%Y年%m月%d日 %A";
        /// `"%Y年%m月%d日 %A %H:%M:%S"`
        pub const D_T_FMT: &str = "%Y年%m月%d日 %A %H:%M:%S";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["一月", "二月", "三月", "四月", "五月", "六月", "七月", "八月", "九月", "十月", "十一月", "十二月"]`
        pub const MON: &[&str] = &["一月", "二月", "三月", "四月", "五月", "六月", "七月", "八月", "九月", "十月", "十一月", "十二月"];
        /// `"%p %I時%M分%S秒 %Z"`
        pub const T_FMT: &str = "%p %I時%M分%S秒 %Z";
        /// `"%p %I:%M:%S"`
        pub const T_FMT_AMPM: &str = "%p %I:%M:%S";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod zh_SG {
    pub mod LC_ADDRESS {
        /// `Some("SGP")`
        pub const COUNTRY_AB3: Option<&str> = Some("SGP");
        /// `Some("SG")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("SG");
        /// `Some("SGP")`
        pub const COUNTRY_CAR: Option<&str> = Some("SGP");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("新加坡")`
        pub const COUNTRY_NAME: Option<&str> = Some("新加坡");
        /// `Some(702)`
        pub const COUNTRY_NUM: Option<i64> = Some(702);
        /// `None`
        pub const COUNTRY_POST: Option<&str> = None;
        /// `Some("zh")`
        pub const LANG_AB: Option<&str> = Some("zh");
        /// `Some("chi")`
        pub const LANG_LIB: Option<&str> = Some("chi");
        /// `Some("简体中文")`
        pub const LANG_NAME: Option<&str> = Some("简体中文");
        /// `Some("zho")`
        pub const LANG_TERM: Option<&str> = Some("zho");
        /// `Some("%z%c%T%s%b%e%r")`
        pub const POSTAL_FMT: Option<&str> = Some("%z%c%T%s%b%e%r");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan")`
        pub const ADDRESS: Option<&str> = Some("1623-14, Shimotsuruma, Yamato-shi, Kanagawa-ken, 242-8502, Japan");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000,October,27 (XML source:2000,July,20)")`
        pub const DATE: Option<&str> = Some("2000,October,27 (XML source:2000,July,20)");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Chinese")`
        pub const LANGUAGE: Option<&str> = Some("Chinese");
        /// `Some("1.0")`
        pub const REVISION: Option<&str> = Some("1.0");
        /// `Some("IBM Globalization Center of Competency, Yamato Software Laboratory")`
        pub const SOURCE: Option<&str> = Some("IBM Globalization Center of Competency, Yamato Software Laboratory");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Singapore")`
        pub const TERRITORY: Option<&str> = Some("Singapore");
        /// `Some("Chinese language locale for Singapore")`
        pub const TITLE: Option<&str> = Some("Chinese language locale for Singapore");
    }
    pub use super::zh_CN::LC_MESSAGES;
    pub mod LC_MONETARY {
        /// `"$"`
        pub const CURRENCY_SYMBOL: &str = "$";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"SGD "`
        pub const INT_CURR_SYMBOL: &str = "SGD ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `None`
        pub const INT_N_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_N_SIGN_POSN: Option<i64> = None;
        /// `None`
        pub const INT_P_CS_PRECEDES: Option<i64> = None;
        /// `None`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = None;
        /// `None`
        pub const INT_P_SIGN_POSN: Option<i64> = None;
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `0`
        pub const N_SIGN_POSN: i64 = 0;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("65")`
        pub const INT_PREFIX: Option<&str> = Some("65");
        /// `Some("001")`
        pub const INT_SELECT: Option<&str> = Some("001");
        /// `None`
        pub const TEL_DOM_FMT: Option<&str> = None;
        /// `Some("+%c ;%a ;%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c ;%a ;%l");
    }
    pub mod LC_TIME {
        /// `&["日", "一", "二", "三", "四", "五", "六"]`
        pub const ABDAY: &[&str] = &["日", "一", "二", "三", "四", "五", "六"];
        /// `&["一月", "二月", "三月", "四月", "五月", "六月", "七月", "八月", "九月", "十月", "十一月", "十二月"]`
        pub const ABMON: &[&str] = &["一月", "二月", "三月", "四月", "五月", "六月", "七月", "八月", "九月", "十月", "十一月", "十二月"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["上午", "下午"]`
        pub const AM_PM: &[&str] = &["上午", "下午"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["星期日", "星期一", "星期二", "星期三", "星期四", "星期五", "星期六"]`
        pub const DAY: &[&str] = &["星期日", "星期一", "星期二", "星期三", "星期四", "星期五", "星期六"];
        /// `"%Y年%m月%d日"`
        pub const D_FMT: &str = "%Y年%m月%d日";
        /// `"%Y年%m月%d日 %H时%M分%S秒 %Z"`
        pub const D_T_FMT: &str = "%Y年%m月%d日 %H时%M分%S秒 %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["一月", "二月", "三月", "四月", "五月", "六月", "七月", "八月", "九月", "十月", "十一月", "十二月"]`
        pub const MON: &[&str] = &["一月", "二月", "三月", "四月", "五月", "六月", "七月", "八月", "九月", "十月", "十一月", "十二月"];
        /// `"%H时%M分%S秒 %Z"`
        pub const T_FMT: &str = "%H时%M分%S秒 %Z";
        /// `"%l:%M:%S %p"`
        pub const T_FMT_AMPM: &str = "%l:%M:%S %p";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod zh_TW {
    pub mod LC_ADDRESS {
        /// `Some("TWN")`
        pub const COUNTRY_AB3: Option<&str> = Some("TWN");
        /// `Some("TW")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("TW");
        /// `Some("RC")`
        pub const COUNTRY_CAR: Option<&str> = Some("RC");
        /// `Some("957")`
        pub const COUNTRY_ISBN: Option<&str> = Some("957");
        /// `Some("中華民國")`
        pub const COUNTRY_NAME: Option<&str> = Some("中華民國");
        /// `Some(158)`
        pub const COUNTRY_NUM: Option<i64> = Some(158);
        /// `Some("TW")`
        pub const COUNTRY_POST: Option<&str> = Some("TW");
        /// `Some("zh")`
        pub const LANG_AB: Option<&str> = Some("zh");
        /// `Some("chi")`
        pub const LANG_LIB: Option<&str> = Some("chi");
        /// `Some("中文")`
        pub const LANG_NAME: Option<&str> = Some("中文");
        /// `Some("zho")`
        pub const LANG_TERM: Option<&str> = Some("zho");
        /// `Some("%c%N%T%N%s %h %e %r%N%b%N%d%N%f%N%a%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%c%N%T%N%s %h %e %r%N%b%N%d%N%f%N%a%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("")`
        pub const ADDRESS: Option<&str> = Some("");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("")`
        pub const CONTACT: Option<&str> = Some("");
        /// `Some("2000-08-02")`
        pub const DATE: Option<&str> = Some("2000-08-02");
        /// `Some("bug-glibc-locales@gnu.org")`
        pub const EMAIL: Option<&str> = Some("bug-glibc-locales@gnu.org");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Chinese")`
        pub const LANGUAGE: Option<&str> = Some("Chinese");
        /// `Some("0.2")`
        pub const REVISION: Option<&str> = Some("0.2");
        /// `Some("")`
        pub const SOURCE: Option<&str> = Some("");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("Taiwan")`
        pub const TERRITORY: Option<&str> = Some("Taiwan");
        /// `Some("Chinese locale for Taiwan R.O.C.")`
        pub const TITLE: Option<&str> = Some("Chinese locale for Taiwan R.O.C.");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNｎＮ不否]"`
        pub const NOEXPR: &str = "^[-0nNｎＮ不否]";
        /// `Some("不是")`
        pub const NOSTR: Option<&str> = Some("不是");
        /// `"^[+1yYｙＹ是]"`
        pub const YESEXPR: &str = "^[+1yYｙＹ是]";
        /// `Some("是")`
        pub const YESSTR: Option<&str> = Some("是");
    }
    pub mod LC_MONETARY {
        /// `"NT$"`
        pub const CURRENCY_SYMBOL: &str = "NT$";
        /// `2`
        pub const FRAC_DIGITS: i64 = 2;
        /// `"TWD "`
        pub const INT_CURR_SYMBOL: &str = "TWD ";
        /// `2`
        pub const INT_FRAC_DIGITS: i64 = 2;
        /// `Some(1)`
        pub const INT_N_CS_PRECEDES: Option<i64> = Some(1);
        /// `Some(0)`
        pub const INT_N_SEP_BY_SPACE: Option<i64> = Some(0);
        /// `Some(1)`
        pub const INT_N_SIGN_POSN: Option<i64> = Some(1);
        /// `Some(1)`
        pub const INT_P_CS_PRECEDES: Option<i64> = Some(1);
        /// `Some(0)`
        pub const INT_P_SEP_BY_SPACE: Option<i64> = Some(0);
        /// `Some(1)`
        pub const INT_P_SIGN_POSN: Option<i64> = Some(1);
        /// `"."`
        pub const MON_DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const MON_GROUPING: &[i64] = &[3];
        /// `","`
        pub const MON_THOUSANDS_SEP: &str = ",";
        /// `"-"`
        pub const NEGATIVE_SIGN: &str = "-";
        /// `1`
        pub const N_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const N_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const N_SIGN_POSN: i64 = 1;
        /// `""`
        pub const POSITIVE_SIGN: &str = "";
        /// `1`
        pub const P_CS_PRECEDES: i64 = 1;
        /// `0`
        pub const P_SEP_BY_SPACE: i64 = 0;
        /// `1`
        pub const P_SIGN_POSN: i64 = 1;
    }
    pub mod LC_NUMERIC {
        /// `"."`
        pub const DECIMAL_POINT: &str = ".";
        /// `&[3]`
        pub const GROUPING: &[i64] = &[3];
        /// `","`
        pub const THOUSANDS_SEP: &str = ",";
    }
    pub mod LC_TELEPHONE {
        /// `Some("886")`
        pub const INT_PREFIX: Option<&str> = Some("886");
        /// `Some("00")`
        pub const INT_SELECT: Option<&str> = Some("00");
        /// `Some("%A-%l")`
        pub const TEL_DOM_FMT: Option<&str> = Some("%A-%l");
        /// `Some("+%c-%a-%l")`
        pub const TEL_INT_FMT: Option<&str> = Some("+%c-%a-%l");
    }
    pub mod LC_TIME {
        /// `&["日", "一", "二", "三", "四", "五", "六"]`
        pub const ABDAY: &[&str] = &["日", "一", "二", "三", "四", "五", "六"];
        /// `&[" 1月", " 2月", " 3月", " 4月", " 5月", " 6月", " 7月", " 8月", " 9月", "10月", "11月", "12月"]`
        pub const ABMON: &[&str] = &[" 1月", " 2月", " 3月", " 4月", " 5月", " 6月", " 7月", " 8月", " 9月", "10月", "11月", "12月"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["上午", "下午"]`
        pub const AM_PM: &[&str] = &["上午", "下午"];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["週日", "週一", "週二", "週三", "週四", "週五", "週六"]`
        pub const DAY: &[&str] = &["週日", "週一", "週二", "週三", "週四", "週五", "週六"];
        /// `"西元%Y年%m月%d日"`
        pub const D_FMT: &str = "西元%Y年%m月%d日";
        /// `"西元%Y年%m月%d日 (%A) %H時%M分%S秒"`
        pub const D_T_FMT: &str = "西元%Y年%m月%d日 (%A) %H時%M分%S秒";
        /// `Some(&["+:2:1913/01/01:+*:民國:%EC%Ey年", "+:1:1912/01/01:1912/12/31:民國:%EC元年", "+:1:1911/12/31:-*:民前:%EC%Ey年"])`
        pub const ERA: Option<&[&str]> = Some(&["+:2:1913/01/01:+*:民國:%EC%Ey年", "+:1:1912/01/01:1912/12/31:民國:%EC元年", "+:1:1911/12/31:-*:民前:%EC%Ey年"]);
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["一月", "二月", "三月", "四月", "五月", "六月", "七月", "八月", "九月", "十月", "十一月", "十二月"]`
        pub const MON: &[&str] = &["一月", "二月", "三月", "四月", "五月", "六月", "七月", "八月", "九月", "十月", "十一月", "十二月"];
        /// `"%H時%M分%S秒"`
        pub const T_FMT: &str = "%H時%M分%S秒";
        /// `"%p %I時%M分%S秒"`
        pub const T_FMT_AMPM: &str = "%p %I時%M分%S秒";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

#[allow(non_snake_case,non_camel_case_types,dead_code,unused_imports)]
pub mod zu_ZA {
    pub mod LC_ADDRESS {
        /// `Some("ZAF")`
        pub const COUNTRY_AB3: Option<&str> = Some("ZAF");
        /// `Some("ZA")`
        pub const COUNTRY_ABTWO: Option<&str> = Some("ZA");
        /// `Some("ZA")`
        pub const COUNTRY_CAR: Option<&str> = Some("ZA");
        /// `None`
        pub const COUNTRY_ISBN: Option<&str> = None;
        /// `Some("iNingizimu Afrika")`
        pub const COUNTRY_NAME: Option<&str> = Some("iNingizimu Afrika");
        /// `Some(710)`
        pub const COUNTRY_NUM: Option<i64> = Some(710);
        /// `Some("ZA")`
        pub const COUNTRY_POST: Option<&str> = Some("ZA");
        /// `Some("zu")`
        pub const LANG_AB: Option<&str> = Some("zu");
        /// `Some("zul")`
        pub const LANG_LIB: Option<&str> = Some("zul");
        /// `Some("isiZulu")`
        pub const LANG_NAME: Option<&str> = Some("isiZulu");
        /// `Some("zul")`
        pub const LANG_TERM: Option<&str> = Some("zul");
        /// `Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N")`
        pub const POSTAL_FMT: Option<&str> = Some("%f%N%a%N%d%N%b%N%s %h %e %r%N%z %T%N%c%N");
    }
    pub mod LC_IDENTIFICATION {
        /// `None`
        pub const ABBREVIATION: Option<&str> = None;
        /// `Some("Box 28364, Sunnyside, 0132, South Africa")`
        pub const ADDRESS: Option<&str> = Some("Box 28364, Sunnyside, 0132, South Africa");
        /// ```ignore
        /// &[
        ///     &["i18n:2012", "LC_IDENTIFICATION"],
        ///     &["i18n:2012", "LC_CTYPE"],
        ///     &["i18n:2012", "LC_COLLATE"],
        ///     &["i18n:2012", "LC_TIME"],
        ///     &["i18n:2012", "LC_NUMERIC"],
        ///     &["i18n:2012", "LC_MONETARY"],
        ///     &["i18n:2012", "LC_MESSAGES"],
        ///     &["i18n:2012", "LC_PAPER"],
        ///     &["i18n:2012", "LC_NAME"],
        ///     &["i18n:2012", "LC_ADDRESS"],
        ///     &["i18n:2012", "LC_TELEPHONE"],
        ///     &["i18n:2012", "LC_MEASUREMENT"],
        /// ]
        /// ```
        pub const CATEGORY: Option<&[&[&str]]> = Some(&[
            &["i18n:2012", "LC_IDENTIFICATION"],
            &["i18n:2012", "LC_CTYPE"],
            &["i18n:2012", "LC_COLLATE"],
            &["i18n:2012", "LC_TIME"],
            &["i18n:2012", "LC_NUMERIC"],
            &["i18n:2012", "LC_MONETARY"],
            &["i18n:2012", "LC_MESSAGES"],
            &["i18n:2012", "LC_PAPER"],
            &["i18n:2012", "LC_NAME"],
            &["i18n:2012", "LC_ADDRESS"],
            &["i18n:2012", "LC_TELEPHONE"],
            &["i18n:2012", "LC_MEASUREMENT"],
        ]);
        /// `Some("Dwayne Bailey")`
        pub const CONTACT: Option<&str> = Some("Dwayne Bailey");
        /// `Some("2005-10-13")`
        pub const DATE: Option<&str> = Some("2005-10-13");
        /// `Some("dwayne@translate.org.za")`
        pub const EMAIL: Option<&str> = Some("dwayne@translate.org.za");
        /// `Some("")`
        pub const FAX: Option<&str> = Some("");
        /// `Some("Zulu")`
        pub const LANGUAGE: Option<&str> = Some("Zulu");
        /// `Some("0.3")`
        pub const REVISION: Option<&str> = Some("0.3");
        /// `Some("Zuza Software Foundation (Translate.org.za)")`
        pub const SOURCE: Option<&str> = Some("Zuza Software Foundation (Translate.org.za)");
        /// `Some("")`
        pub const TEL: Option<&str> = Some("");
        /// `Some("South Africa")`
        pub const TERRITORY: Option<&str> = Some("South Africa");
        /// `Some("Zulu locale for South Africa")`
        pub const TITLE: Option<&str> = Some("Zulu locale for South Africa");
    }
    pub mod LC_MESSAGES {
        /// `"^[-0nNcC]"`
        pub const NOEXPR: &str = "^[-0nNcC]";
        /// `Some("cha")`
        pub const NOSTR: Option<&str> = Some("cha");
        /// `"^[+1yY]"`
        pub const YESEXPR: &str = "^[+1yY]";
        /// `Some("yebo")`
        pub const YESSTR: Option<&str> = Some("yebo");
    }
    pub use super::en_ZA::LC_MONETARY;
    pub use super::en_ZA::LC_NUMERIC;
    pub use super::en_ZA::LC_TELEPHONE;
    pub mod LC_TIME {
        /// `&["Son", "Mso", "Bil", "Tha", "Sin", "Hla", "Mgq"]`
        pub const ABDAY: &[&str] = &["Son", "Mso", "Bil", "Tha", "Sin", "Hla", "Mgq"];
        /// `&["Jan", "Feb", "Mas", "Eph", "Mey", "Jun", "Jul", "Aga", "Sep", "Okt", "Nov", "Dis"]`
        pub const ABMON: &[&str] = &["Jan", "Feb", "Mas", "Eph", "Mey", "Jun", "Jul", "Aga", "Sep", "Okt", "Nov", "Dis"];
        /// `None`
        pub const AB_ALT_MON: Option<&[&str]> = None;
        /// `None`
        pub const ALT_DIGITS: Option<&[&str]> = None;
        /// `None`
        pub const ALT_MON: Option<&[&str]> = None;
        /// `&["", ""]`
        pub const AM_PM: &[&str] = &["", ""];
        /// `None`
        pub const CAL_DIRECTION: Option<i64> = None;
        /// `None`
        pub const DATE_FMT: Option<&str> = None;
        /// `&["iSonto", "uMsombuluko", "uLwesibili", "uLwesithathu", "uLwesine", "uLwesihlanu", "uMgqibelo"]`
        pub const DAY: &[&str] = &["iSonto", "uMsombuluko", "uLwesibili", "uLwesithathu", "uLwesine", "uLwesihlanu", "uMgqibelo"];
        /// `"%d/%m/%Y"`
        pub const D_FMT: &str = "%d/%m/%Y";
        /// `"%a %d %b %Y %T %Z"`
        pub const D_T_FMT: &str = "%a %d %b %Y %T %Z";
        /// `None`
        pub const ERA: Option<&[&str]> = None;
        /// `None`
        pub const ERA_D_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_D_T_FMT: Option<&str> = None;
        /// `None`
        pub const ERA_T_FMT: Option<&str> = None;
        /// `None`
        pub const FIRST_WEEKDAY: Option<i64> = None;
        /// `None`
        pub const FIRST_WORKDAY: Option<i64> = None;
        /// `&["Januwari", "Februwari", "Mashi", "Ephreli", "Meyi", "Juni", "Julayi", "Agasti", "Septhemba", "Okthoba", "Novemba", "Disemba"]`
        pub const MON: &[&str] = &["Januwari", "Februwari", "Mashi", "Ephreli", "Meyi", "Juni", "Julayi", "Agasti", "Septhemba", "Okthoba", "Novemba", "Disemba"];
        /// `"%T"`
        pub const T_FMT: &str = "%T";
        /// `""`
        pub const T_FMT_AMPM: &str = "";
        /// `Some(&[7, 19971130, 1])`
        pub const WEEK: Option<&[i64]> = Some(&[7, 19971130, 1]);
    }
}

/// Locales matching the locales in `glibc`.
///
/// Most locale names follow the syntax `language[_territory][@modifier]`.
/// The `@` is replaced with `_` in the `enum` variant names.
///
/// The default locale is `POSIX`.
///
/// License note: The Free Software Foundation does not claim any copyright interest in the locale
/// data of the GNU C Library; they believe it is not copyrightable.
#[allow(non_camel_case_types,dead_code)]
#[derive(Copy, Clone, PartialEq, Eq, Hash)]
pub enum Locale {
    /// `POSIX`: POSIX Standard Locale.
    POSIX,
    /// `aa_DJ`: Afar language locale for Djibouti (Cadu/Laaqo Dialects).
    aa_DJ,
    /// `aa_ER`: Afar language locale for Eritrea (Cadu/Laaqo Dialects).
    aa_ER,
    /// `aa_ER@saaho`: Afar language locale for Eritrea (Saaho Dialect).
    aa_ER_saaho,
    /// `aa_ET`: Afar language locale for Ethiopia (Cadu/Carra Dialects).
    aa_ET,
    /// `af_ZA`: Afrikaans locale for South Africa.
    af_ZA,
    /// `agr_PE`: Awajún (agr) locale for Peru.
    agr_PE,
    /// `ak_GH`: Akan locale for Ghana.
    ak_GH,
    /// `am_ET`: Amharic language locale for Ethiopia.
    am_ET,
    /// `an_ES`: Aragonese locale for Spain.
    an_ES,
    /// `anp_IN`: Angika language locale for India.
    anp_IN,
    /// `ar_AE`: Arabic language locale for United Arab Emirates.
    ar_AE,
    /// `ar_BH`: Arabic language locale for Bahrain.
    ar_BH,
    /// `ar_DZ`: Arabic language locale for Algeria.
    ar_DZ,
    /// `ar_EG`: Arabic language locale for Egypt.
    ar_EG,
    /// `ar_IN`: Arabic language locale for India.
    ar_IN,
    /// `ar_IQ`: Arabic language locale for Iraq.
    ar_IQ,
    /// `ar_JO`: Arabic language locale for Jordan.
    ar_JO,
    /// `ar_KW`: Arabic language locale for Kuwait.
    ar_KW,
    /// `ar_LB`: Arabic language locale for Lebanon.
    ar_LB,
    /// `ar_LY`: Arabic language locale for Libyan Arab Jamahiriya.
    ar_LY,
    /// `ar_MA`: Arabic language locale for Morocco.
    ar_MA,
    /// `ar_OM`: Arabic language locale for Oman.
    ar_OM,
    /// `ar_QA`: Arabic language locale for Qatar.
    ar_QA,
    /// `ar_SA`: Arabic locale for Saudi Arabia.
    ar_SA,
    /// `ar_SD`: Arabic language locale for Sudan.
    ar_SD,
    /// `ar_SS`: Arabic language locale for South Sudan.
    ar_SS,
    /// `ar_SY`: Arabic language locale for Syrian Arab Republic.
    ar_SY,
    /// `ar_TN`: Arabic language locale for Tunisia.
    ar_TN,
    /// `ar_YE`: Arabic language locale for Yemen.
    ar_YE,
    /// `as_IN`: Assamese language locale for India.
    as_IN,
    /// `ast_ES`: Asturian locale for Spain.
    ast_ES,
    /// `ayc_PE`: Aymara (ayc) locale for Peru.
    ayc_PE,
    /// `az_AZ`: Azeri language locale for Azerbaijan (latin).
    az_AZ,
    /// `az_IR`: South Azerbaijani language locale for Iran.
    az_IR,
    /// `be_BY`: Belarusian locale for Belarus.
    be_BY,
    /// `be_BY@latin`: Belarusian Latin-Script locale for Belarus.
    be_BY_latin,
    /// `bem_ZM`: Bemba locale for Zambia.
    bem_ZM,
    /// `ber_DZ`: Berber language locale for Algeria (latin).
    ber_DZ,
    /// `ber_MA`: Berber language locale for Morocco (tifinagh).
    ber_MA,
    /// `bg_BG`: Bulgarian locale for Bulgaria.
    bg_BG,
    /// `bhb_IN`: Bhili(devanagari) language locale for India.
    bhb_IN,
    /// `bho_IN`: Bhojpuri language locale for India.
    bho_IN,
    /// `bho_NP`: Bhojpuri language locale for Nepal.
    bho_NP,
    /// `bi_VU`: Bislama language locale for Vanuatu.
    bi_VU,
    /// `bn_BD`: Bangla language locale for Bangladesh.
    bn_BD,
    /// `bn_IN`: Bangla language locale for India.
    bn_IN,
    /// `bo_CN`: Tibetan language locale for P.R. of China.
    bo_CN,
    /// `bo_IN`: Tibetan language locale for India.
    bo_IN,
    /// `br_FR`: Breton language locale for France.
    br_FR,
    /// `br_FR@euro`: Breton locale for France with Euro.
    br_FR_euro,
    /// `brx_IN`: Bodo language locale for India.
    brx_IN,
    /// `bs_BA`: Bosnian language locale for Bosnia and Herzegowina.
    bs_BA,
    /// `byn_ER`: Blin language locale for Eritrea.
    byn_ER,
    /// `ca_AD`: Catalan locale for Andorra .
    ca_AD,
    /// `ca_ES`: Catalan locale for Spain.
    ca_ES,
    /// `ca_ES@euro`: Catalan locale for Catalonia with Euro.
    ca_ES_euro,
    /// `ca_ES@valencia`: Valencian (southern Catalan) locale for Spain with Euro.
    ca_ES_valencia,
    /// `ca_FR`: Catalan locale for France .
    ca_FR,
    /// `ca_IT`: Catalan locale for Italy (L'Alguer) .
    ca_IT,
    /// `ce_RU`: Chechen locale for RUSSIAN FEDERATION.
    ce_RU,
    /// `chr_US`: Cherokee language locale for United States.
    chr_US,
    /// `cmn_TW`: Mandarin Chinese locale for the Republic of China.
    cmn_TW,
    /// `crh_UA`: Crimean Tatar (Crimean Turkish) language locale for Ukraine.
    crh_UA,
    /// `cs_CZ`: Czech locale for the Czech Republic.
    cs_CZ,
    /// `csb_PL`: Kashubian locale for Poland.
    csb_PL,
    /// `cv_RU`: Chuvash locale for Russia.
    cv_RU,
    /// `cy_GB`: Welsh language locale for Great Britain.
    cy_GB,
    /// `da_DK`: Danish locale for Denmark.
    da_DK,
    /// `de_AT`: German locale for Austria.
    de_AT,
    /// `de_AT@euro`: German locale for Austria with Euro.
    de_AT_euro,
    /// `de_BE`: German locale for Belgium.
    de_BE,
    /// `de_BE@euro`: German locale for Belgium with Euro.
    de_BE_euro,
    /// `de_CH`: German locale for Switzerland.
    de_CH,
    /// `de_DE`: German locale for Germany.
    de_DE,
    /// `de_DE@euro`: German locale for Germany with Euro.
    de_DE_euro,
    /// `de_IT`: German language locale for Italy.
    de_IT,
    /// `de_LI`: German locale for Liechtenstein.
    de_LI,
    /// `de_LU`: German locale for Luxemburg.
    de_LU,
    /// `de_LU@euro`: German locale for Luxemburg with Euro.
    de_LU_euro,
    /// `doi_IN`: Dogri language locale for India.
    doi_IN,
    /// `dsb_DE`: Lower Sorbian locale for Germany.
    dsb_DE,
    /// `dv_MV`: Dhivehi Language Locale for Maldives.
    dv_MV,
    /// `dz_BT`: Dzongkha language locale for Bhutan.
    dz_BT,
    /// `el_CY`: Greek locale for Cyprus.
    el_CY,
    /// `el_GR`: Greek locale for Greece.
    el_GR,
    /// `el_GR@euro`: Greek locale for Greece with Euro.
    el_GR_euro,
    /// `en_AG`: English language locale for Antigua and Barbuda.
    en_AG,
    /// `en_AU`: English locale for Australia.
    en_AU,
    /// `en_BW`: English locale for Botswana.
    en_BW,
    /// `en_CA`: English locale for Canada.
    en_CA,
    /// `en_DK`: English locale for Denmark.
    en_DK,
    /// `en_GB`: English locale for Britain.
    en_GB,
    /// `en_HK`: English locale for Hong Kong.
    en_HK,
    /// `en_IE`: English locale for Ireland.
    en_IE,
    /// `en_IE@euro`: English locale for Ireland with Euro.
    en_IE_euro,
    /// `en_IL`: English locale for Israel.
    en_IL,
    /// `en_IN`: English language locale for India.
    en_IN,
    /// `en_NG`: English locale for Nigeria.
    en_NG,
    /// `en_NZ`: English locale for New Zealand.
    en_NZ,
    /// `en_PH`: English language locale for Philippines.
    en_PH,
    /// `en_SC`: English locale for the Seychelles.
    en_SC,
    /// `en_SG`: English language locale for Singapore.
    en_SG,
    /// `en_US`: English locale for the USA.
    en_US,
    /// `en_ZA`: English locale for South Africa.
    en_ZA,
    /// `en_ZM`: English locale for Zambia.
    en_ZM,
    /// `en_ZW`: English locale for Zimbabwe.
    en_ZW,
    /// `eo`: Esperanto language locale.
    eo,
    /// `es_AR`: Spanish locale for Argentina.
    es_AR,
    /// `es_BO`: Spanish locale for Bolivia.
    es_BO,
    /// `es_CL`: Spanish locale for Chile.
    es_CL,
    /// `es_CO`: Spanish locale for Colombia.
    es_CO,
    /// `es_CR`: Spanish locale for Costa Rica.
    es_CR,
    /// `es_CU`: Spanish locale for Cuba.
    es_CU,
    /// `es_DO`: Spanish locale for Dominican Republic.
    es_DO,
    /// `es_EC`: Spanish locale for Ecuador.
    es_EC,
    /// `es_ES`: Spanish locale for Spain.
    es_ES,
    /// `es_ES@euro`: Spanish locale for Spain with Euro.
    es_ES_euro,
    /// `es_GT`: Spanish locale for Guatemala.
    es_GT,
    /// `es_HN`: Spanish locale for Honduras.
    es_HN,
    /// `es_MX`: Spanish locale for Mexico.
    es_MX,
    /// `es_NI`: Spanish locale for Nicaragua.
    es_NI,
    /// `es_PA`: Spanish locale for Panama.
    es_PA,
    /// `es_PE`: Spanish locale for Peru.
    es_PE,
    /// `es_PR`: Spanish locale for Puerto Rico.
    es_PR,
    /// `es_PY`: Spanish locale for Paraguay.
    es_PY,
    /// `es_SV`: Spanish locale for El Salvador.
    es_SV,
    /// `es_US`: Spanish locale for the USA.
    es_US,
    /// `es_UY`: Spanish locale for Uruguay.
    es_UY,
    /// `es_VE`: Spanish locale for Venezuela.
    es_VE,
    /// `et_EE`: Estonian locale for Estonia.
    et_EE,
    /// `eu_ES`: Basque locale for Spain.
    eu_ES,
    /// `eu_ES@euro`: Basque language locale for Spain with Euro.
    eu_ES_euro,
    /// `fa_IR`: Persian locale for Iran.
    fa_IR,
    /// `ff_SN`: Fulah locale for Senegal.
    ff_SN,
    /// `fi_FI`: Finnish locale for Finland.
    fi_FI,
    /// `fi_FI@euro`: Finnish locale for Finland with Euro.
    fi_FI_euro,
    /// `fil_PH`: Filipino language locale for Philippines.
    fil_PH,
    /// `fo_FO`: Faroese locale for Faroe Islands.
    fo_FO,
    /// `fr_BE`: French locale for Belgium.
    fr_BE,
    /// `fr_BE@euro`: French locale for Belgium with Euro.
    fr_BE_euro,
    /// `fr_CA`: French locale for Canada.
    fr_CA,
    /// `fr_CH`: French locale for Switzerland.
    fr_CH,
    /// `fr_FR`: French locale for France.
    fr_FR,
    /// `fr_FR@euro`: French locale for France with Euro.
    fr_FR_euro,
    /// `fr_LU`: French locale for Luxemburg.
    fr_LU,
    /// `fr_LU@euro`: French locale for Luxemburg with Euro.
    fr_LU_euro,
    /// `fur_IT`: Furlan locale for Italy.
    fur_IT,
    /// `fy_DE`: Sater Frisian and North Frisian Locale for Germany.
    fy_DE,
    /// `fy_NL`: Frisian locale for the Netherlands.
    fy_NL,
    /// `ga_IE`: Irish locale for Ireland.
    ga_IE,
    /// `ga_IE@euro`: Irish locale for Ireland with Euro.
    ga_IE_euro,
    /// `gd_GB`: Scots Gaelic language locale for Great Britain.
    gd_GB,
    /// `gez_ER`: Ge'ez language locale for Eritrea.
    gez_ER,
    /// `gez_ER@abegede`: Ge'ez language locale for Eritrea With Abegede Collation.
    gez_ER_abegede,
    /// `gez_ET`: Ge'ez language locale for Ethiopia.
    gez_ET,
    /// `gez_ET@abegede`: Ge'ez language locale for Ethiopia With Abegede Collation.
    gez_ET_abegede,
    /// `gl_ES`: Galician locale for Spain.
    gl_ES,
    /// `gl_ES@euro`: Galician locale for Spain with Euro.
    gl_ES_euro,
    /// `gu_IN`: Gujarati Language Locale For India.
    gu_IN,
    /// `gv_GB`: Manx Gaelic locale for Britain.
    gv_GB,
    /// `ha_NG`: Hausa locale for Nigeria.
    ha_NG,
    /// `hak_TW`: Hakka Chinese locale for the Republic of China.
    hak_TW,
    /// `he_IL`: Hebrew locale for Israel.
    he_IL,
    /// `hi_IN`: Hindi language locale for India.
    hi_IN,
    /// `hif_FJ`: Fiji Hindi (Latin) language locale for Fiji.
    hif_FJ,
    /// `hne_IN`: Chhattisgarhi language locale for India.
    hne_IN,
    /// `hr_HR`: Croatian locale for Croatia.
    hr_HR,
    /// `hsb_DE`: Upper Sorbian locale for Germany.
    hsb_DE,
    /// `ht_HT`: Kreyol locale for Haiti.
    ht_HT,
    /// `hu_HU`: Hungarian locale for Hungary.
    hu_HU,
    /// `hy_AM`: Armenian language locale for Armenia.
    hy_AM,
    /// `ia_FR`: Interlingua locale for France.
    ia_FR,
    /// `id_ID`: Indonesian locale for Indonesia.
    id_ID,
    /// `ig_NG`: Igbo locale for Nigeria.
    ig_NG,
    /// `ik_CA`: Inupiaq locale for Canada.
    ik_CA,
    /// `is_IS`: Icelandic locale for Iceland.
    is_IS,
    /// `it_CH`: Italian locale for Switzerland.
    it_CH,
    /// `it_IT`: Italian locale for Italy.
    it_IT,
    /// `it_IT@euro`: Italian locale for Italy with Euro.
    it_IT_euro,
    /// `iu_CA`: Inuktitut language locale for Nunavut, Canada.
    iu_CA,
    /// `ja_JP`: Japanese language locale for Japan.
    ja_JP,
    /// `ka_GE`: Georgian language locale for Georgia.
    ka_GE,
    /// `kab_DZ`: Kabyle language locale for Algeria.
    kab_DZ,
    /// `kk_KZ`: Kazakh locale for Kazakhstan.
    kk_KZ,
    /// `kl_GL`: Greenlandic locale for Greenland.
    kl_GL,
    /// `km_KH`: Khmer locale for Cambodia.
    km_KH,
    /// `kn_IN`: Kannada language locale for India.
    kn_IN,
    /// `ko_KR`: Korean locale for Republic of Korea.
    ko_KR,
    /// `kok_IN`: Konkani language locale for India.
    kok_IN,
    /// `ks_IN`: Kashmiri language locale for India.
    ks_IN,
    /// `ks_IN@devanagari`: Kashmiri(devanagari) language locale for India.
    ks_IN_devanagari,
    /// `ku_TR`: Kurdish (latin) locale for Turkey.
    ku_TR,
    /// `kw_GB`: Cornish locale for Britain.
    kw_GB,
    /// `ky_KG`: Kyrgyz Language Locale for Kyrgyzstan.
    ky_KG,
    /// `lb_LU`: Luxembourgish locale for Luxembourg.
    lb_LU,
    /// `lg_UG`: Luganda locale for Uganda.
    lg_UG,
    /// `li_BE`: Limburgish Language Locale for Belgium.
    li_BE,
    /// `li_NL`: Limburgish Language Locale for the Netherlands.
    li_NL,
    /// `lij_IT`: Ligurian locale for Italy.
    lij_IT,
    /// `ln_CD`: Lingala locale for Democratic Republic of the Congo.
    ln_CD,
    /// `lo_LA`: Lao locale for Laos.
    lo_LA,
    /// `lt_LT`: Lithuanian locale for Lithuania.
    lt_LT,
    /// `lv_LV`: Latvian locale for Latvia.
    lv_LV,
    /// `lzh_TW`: Literary Chinese locale for the Republic of China.
    lzh_TW,
    /// `mag_IN`: Magahi language locale for India.
    mag_IN,
    /// `mai_IN`: Maithili language locale for India.
    mai_IN,
    /// `mai_NP`: Maithili language locale for Nepal.
    mai_NP,
    /// `mfe_MU`: Morisyen locale for Mauritius.
    mfe_MU,
    /// `mg_MG`: Malagasy locale for Madagascar.
    mg_MG,
    /// `mhr_RU`: Mari locale for Russia.
    mhr_RU,
    /// `mi_NZ`: Maori language locale for New Zealand.
    mi_NZ,
    /// `miq_NI`: Miskito language locale for Nicaragua.
    miq_NI,
    /// `mjw_IN`: Karbi language locale for India.
    mjw_IN,
    /// `mk_MK`: Macedonian locale for Macedonia.
    mk_MK,
    /// `ml_IN`: Malayalam language locale for India.
    ml_IN,
    /// `mn_MN`: Mongolian locale for Mongolia.
    mn_MN,
    /// `mni_IN`: Manipuri language locale for India.
    mni_IN,
    /// `mnw_MM`: Mon language locale for Myanmar.
    mnw_MM,
    /// `mr_IN`: Marathi language locale for India.
    mr_IN,
    /// `ms_MY`: Malay language locale for Malaysia.
    ms_MY,
    /// `mt_MT`: Maltese language locale for Malta.
    mt_MT,
    /// `my_MM`: Burmese language locale for Myanmar.
    my_MM,
    /// `nan_TW`: Min Nan Chinese locale for the Republic of China.
    nan_TW,
    /// `nan_TW@latin`: Minnan language locale for Taiwan.
    nan_TW_latin,
    /// `nb_NO`: Norwegian (Bokmål) locale for Norway.
    nb_NO,
    /// `nds_DE`: Low(lands) Saxon Language Locale for Germany.
    nds_DE,
    /// `nds_NL`: Low(lands) Saxon Language Locale for the Netherlands.
    nds_NL,
    /// `ne_NP`: Nepali language locale for Nepal.
    ne_NP,
    /// `nhn_MX`: Central Nahuatl for Mexico.
    nhn_MX,
    /// `niu_NU`: Niuean (Vagahau Niue) locale for Niue.
    niu_NU,
    /// `niu_NZ`: Niuean (Vagahau Niue) locale for New Zealand.
    niu_NZ,
    /// `nl_AW`: Dutch language locale for Aruba.
    nl_AW,
    /// `nl_BE`: Dutch locale for Belgium.
    nl_BE,
    /// `nl_BE@euro`: Dutch locale for Belgium with Euro.
    nl_BE_euro,
    /// `nl_NL`: Dutch locale for the Netherlands.
    nl_NL,
    /// `nl_NL@euro`: Dutch locale for the Netherlands with Euro.
    nl_NL_euro,
    /// `nn_NO`: Nynorsk language locale for Norway.
    nn_NO,
    /// `nr_ZA`: Southern Ndebele locale for South Africa.
    nr_ZA,
    /// `nso_ZA`: Northern Sotho locale for South Africa.
    nso_ZA,
    /// `oc_FR`: Occitan Language Locale for France.
    oc_FR,
    /// `om_ET`: Oromo language locale for Ethiopia.
    om_ET,
    /// `om_KE`: Oromo language locale for Kenya.
    om_KE,
    /// `or_IN`: Odia language locale for India.
    or_IN,
    /// `os_RU`: Ossetian locale for Russia.
    os_RU,
    /// `pa_IN`: Punjabi language locale for Indian Punjabi(Gurmukhi).
    pa_IN,
    /// `pa_PK`: Punjabi (Shahmukhi) Language Locale for Pakistan.
    pa_PK,
    /// `pap_AW`: Papiamento Language for Aruba.
    pap_AW,
    /// `pap_CW`: Papiamento language for Curaçao.
    pap_CW,
    /// `pl_PL`: Polish locale for Poland.
    pl_PL,
    /// `ps_AF`: Pashto locale for Afghanistan.
    ps_AF,
    /// `pt_BR`: Portuguese locale for Brasil.
    pt_BR,
    /// `pt_PT`: Portuguese locale for Portugal.
    pt_PT,
    /// `pt_PT@euro`: Portuguese locale for Portugal with Euro.
    pt_PT_euro,
    /// `quz_PE`: Cusco Quechua locale for Peru.
    quz_PE,
    /// `raj_IN`: Rajasthani language locale for India.
    raj_IN,
    /// `ro_RO`: Romanian locale for Romania.
    ro_RO,
    /// `ru_RU`: Russian locale for Russia.
    ru_RU,
    /// `ru_UA`: Russian locale for Ukraine.
    ru_UA,
    /// `rw_RW`: Kinyarwanda language locale for Rwanda.
    rw_RW,
    /// `sa_IN`: Sanskrit language locale for India.
    sa_IN,
    /// `sah_RU`: Sakha (Yakut) locale for Russian Federation.
    sah_RU,
    /// `sat_IN`: Santali language locale for India.
    sat_IN,
    /// `sc_IT`: Sardinian locale for Italy.
    sc_IT,
    /// `sd_IN`: Sindhi language locale for India.
    sd_IN,
    /// `sd_IN@devanagari`: Sindhi language locale for India.
    sd_IN_devanagari,
    /// `se_NO`: Northern Saami language locale for Norway.
    se_NO,
    /// `sgs_LT`: Samogitian language locale for Lithuania.
    sgs_LT,
    /// `shn_MM`: Shan language locale for Myanmar.
    shn_MM,
    /// `shs_CA`: Shuswap locale for Canada.
    shs_CA,
    /// `si_LK`: Sinhala language locale for Sri Lanka.
    si_LK,
    /// `sid_ET`: Sidama language locale for Ethiopia.
    sid_ET,
    /// `sk_SK`: Slovak locale for Slovak.
    sk_SK,
    /// `sl_SI`: Slovenian locale for Slovenia.
    sl_SI,
    /// `sm_WS`: Samoan language locale for Samoa.
    sm_WS,
    /// `so_DJ`: Somali language locale for Djibouti.
    so_DJ,
    /// `so_ET`: Somali language locale for Ethiopia.
    so_ET,
    /// `so_KE`: Somali language locale for Kenya.
    so_KE,
    /// `so_SO`: Somali language locale for Somalia.
    so_SO,
    /// `sq_AL`: Albanian language locale for Albania.
    sq_AL,
    /// `sq_MK`: Albanian language locale for Macedonia.
    sq_MK,
    /// `sr_ME`: Serbian locale for Montenegro.
    sr_ME,
    /// `sr_RS`: Serbian locale for Serbia.
    sr_RS,
    /// `sr_RS@latin`: Serbian Latin locale for Serbia.
    sr_RS_latin,
    /// `ss_ZA`: Swati locale for South Africa.
    ss_ZA,
    /// `st_ZA`: Sotho locale for South Africa.
    st_ZA,
    /// `sv_FI`: Swedish locale for Finland.
    sv_FI,
    /// `sv_FI@euro`: Swedish locale for Finland with Euro.
    sv_FI_euro,
    /// `sv_SE`: Swedish locale for Sweden.
    sv_SE,
    /// `sw_KE`: Swahili locale for Kenya.
    sw_KE,
    /// `sw_TZ`: Swahili locale for Tanzania.
    sw_TZ,
    /// `szl_PL`: Silesian locale for Poland.
    szl_PL,
    /// `ta_IN`: Tamil language locale for India.
    ta_IN,
    /// `ta_LK`: Tamil language locale for Sri Lanka.
    ta_LK,
    /// `tcy_IN`: Tulu language locale for India.
    tcy_IN,
    /// `te_IN`: Telugu language locale for India.
    te_IN,
    /// `tg_TJ`: Tajik language locale for Tajikistan.
    tg_TJ,
    /// `th_TH`: Thai locale for Thailand.
    th_TH,
    /// `the_NP`: Tharu language locale for Nepal.
    the_NP,
    /// `ti_ER`: Tigrigna language locale for Eritrea.
    ti_ER,
    /// `ti_ET`: Tigrigna language locale for Ethiopia.
    ti_ET,
    /// `tig_ER`: Tigre language locale for Eritrea.
    tig_ER,
    /// `tk_TM`: Turkmen locale for Turkmenistan.
    tk_TM,
    /// `tl_PH`: Tagalog language locale for Philippines.
    tl_PH,
    /// `tn_ZA`: Tswana locale for South Africa.
    tn_ZA,
    /// `to_TO`: Tongan language locale for Tonga.
    to_TO,
    /// `tpi_PG`: Tok Pisin language locale for Papua New Guinea.
    tpi_PG,
    /// `tr_CY`: Turkish language locale for Cyprus.
    tr_CY,
    /// `tr_TR`: Turkish locale for Turkey.
    tr_TR,
    /// `ts_ZA`: Tsonga locale for South Africa.
    ts_ZA,
    /// `tt_RU`: Tatar language locale for Russia.
    tt_RU,
    /// `tt_RU@iqtelif`: Tatar language locale using IQTElif alphabet; for Tatarstan, Russian Federation.
    tt_RU_iqtelif,
    /// `ug_CN`: Uyghur locale for China.
    ug_CN,
    /// `uk_UA`: Ukrainian Language Locale for Ukraine.
    uk_UA,
    /// `unm_US`: Unami Delaware locale for the USA.
    unm_US,
    /// `ur_IN`: Urdu language locale for India.
    ur_IN,
    /// `ur_PK`: Urdu Language Locale for Pakistan.
    ur_PK,
    /// `uz_UZ`: Uzbek (latin) locale for Uzbekistan.
    uz_UZ,
    /// `uz_UZ@cyrillic`: Uzbek (cyrillic) locale for Uzbekistan.
    uz_UZ_cyrillic,
    /// `ve_ZA`: Venda locale for South Africa.
    ve_ZA,
    /// `vi_VN`: Vietnamese language locale for Vietnam.
    vi_VN,
    /// `wa_BE`: Walloon Language Locale for Belgium.
    wa_BE,
    /// `wa_BE@euro`: Walloon locale for Belgium with Euro.
    wa_BE_euro,
    /// `wae_CH`: Walser locale for Switzerland.
    wae_CH,
    /// `wal_ET`: Walaita language locale for Ethiopia.
    wal_ET,
    /// `wo_SN`: Wolof locale for Senegal.
    wo_SN,
    /// `xh_ZA`: Xhosa locale for South Africa.
    xh_ZA,
    /// `yi_US`: Yiddish Language locale for the USA.
    yi_US,
    /// `yo_NG`: Yoruba locale for Nigeria.
    yo_NG,
    /// `yue_HK`: Yue Chinese (Cantonese) language locale for Hong Kong.
    yue_HK,
    /// `yuw_PG`: YauNungon locale for Papua New Guinea.
    yuw_PG,
    /// `zh_CN`: Chinese locale for Peoples Republic of China.
    zh_CN,
    /// `zh_HK`: Chinese language locale for Hong Kong.
    zh_HK,
    /// `zh_SG`: Chinese language locale for Singapore.
    zh_SG,
    /// `zh_TW`: Chinese locale for Taiwan R.O.C.
    zh_TW,
    /// `zu_ZA`: Zulu locale for South Africa.
    zu_ZA,
}

impl core::fmt::Display for Locale {
    fn fmt(&self, f: &mut core::fmt::Formatter) -> core::fmt::Result {
        f.write_str(match self {
            Locale::POSIX => "POSIX",
            Locale::aa_DJ => "aa_DJ",
            Locale::aa_ER => "aa_ER",
            Locale::aa_ER_saaho => "aa_ER@saaho",
            Locale::aa_ET => "aa_ET",
            Locale::af_ZA => "af_ZA",
            Locale::agr_PE => "agr_PE",
            Locale::ak_GH => "ak_GH",
            Locale::am_ET => "am_ET",
            Locale::an_ES => "an_ES",
            Locale::anp_IN => "anp_IN",
            Locale::ar_AE => "ar_AE",
            Locale::ar_BH => "ar_BH",
            Locale::ar_DZ => "ar_DZ",
            Locale::ar_EG => "ar_EG",
            Locale::ar_IN => "ar_IN",
            Locale::ar_IQ => "ar_IQ",
            Locale::ar_JO => "ar_JO",
            Locale::ar_KW => "ar_KW",
            Locale::ar_LB => "ar_LB",
            Locale::ar_LY => "ar_LY",
            Locale::ar_MA => "ar_MA",
            Locale::ar_OM => "ar_OM",
            Locale::ar_QA => "ar_QA",
            Locale::ar_SA => "ar_SA",
            Locale::ar_SD => "ar_SD",
            Locale::ar_SS => "ar_SS",
            Locale::ar_SY => "ar_SY",
            Locale::ar_TN => "ar_TN",
            Locale::ar_YE => "ar_YE",
            Locale::as_IN => "as_IN",
            Locale::ast_ES => "ast_ES",
            Locale::ayc_PE => "ayc_PE",
            Locale::az_AZ => "az_AZ",
            Locale::az_IR => "az_IR",
            Locale::be_BY => "be_BY",
            Locale::be_BY_latin => "be_BY@latin",
            Locale::bem_ZM => "bem_ZM",
            Locale::ber_DZ => "ber_DZ",
            Locale::ber_MA => "ber_MA",
            Locale::bg_BG => "bg_BG",
            Locale::bhb_IN => "bhb_IN",
            Locale::bho_IN => "bho_IN",
            Locale::bho_NP => "bho_NP",
            Locale::bi_VU => "bi_VU",
            Locale::bn_BD => "bn_BD",
            Locale::bn_IN => "bn_IN",
            Locale::bo_CN => "bo_CN",
            Locale::bo_IN => "bo_IN",
            Locale::br_FR => "br_FR",
            Locale::br_FR_euro => "br_FR@euro",
            Locale::brx_IN => "brx_IN",
            Locale::bs_BA => "bs_BA",
            Locale::byn_ER => "byn_ER",
            Locale::ca_AD => "ca_AD",
            Locale::ca_ES => "ca_ES",
            Locale::ca_ES_euro => "ca_ES@euro",
            Locale::ca_ES_valencia => "ca_ES@valencia",
            Locale::ca_FR => "ca_FR",
            Locale::ca_IT => "ca_IT",
            Locale::ce_RU => "ce_RU",
            Locale::chr_US => "chr_US",
            Locale::cmn_TW => "cmn_TW",
            Locale::crh_UA => "crh_UA",
            Locale::cs_CZ => "cs_CZ",
            Locale::csb_PL => "csb_PL",
            Locale::cv_RU => "cv_RU",
            Locale::cy_GB => "cy_GB",
            Locale::da_DK => "da_DK",
            Locale::de_AT => "de_AT",
            Locale::de_AT_euro => "de_AT@euro",
            Locale::de_BE => "de_BE",
            Locale::de_BE_euro => "de_BE@euro",
            Locale::de_CH => "de_CH",
            Locale::de_DE => "de_DE",
            Locale::de_DE_euro => "de_DE@euro",
            Locale::de_IT => "de_IT",
            Locale::de_LI => "de_LI",
            Locale::de_LU => "de_LU",
            Locale::de_LU_euro => "de_LU@euro",
            Locale::doi_IN => "doi_IN",
            Locale::dsb_DE => "dsb_DE",
            Locale::dv_MV => "dv_MV",
            Locale::dz_BT => "dz_BT",
            Locale::el_CY => "el_CY",
            Locale::el_GR => "el_GR",
            Locale::el_GR_euro => "el_GR@euro",
            Locale::en_AG => "en_AG",
            Locale::en_AU => "en_AU",
            Locale::en_BW => "en_BW",
            Locale::en_CA => "en_CA",
            Locale::en_DK => "en_DK",
            Locale::en_GB => "en_GB",
            Locale::en_HK => "en_HK",
            Locale::en_IE => "en_IE",
            Locale::en_IE_euro => "en_IE@euro",
            Locale::en_IL => "en_IL",
            Locale::en_IN => "en_IN",
            Locale::en_NG => "en_NG",
            Locale::en_NZ => "en_NZ",
            Locale::en_PH => "en_PH",
            Locale::en_SC => "en_SC",
            Locale::en_SG => "en_SG",
            Locale::en_US => "en_US",
            Locale::en_ZA => "en_ZA",
            Locale::en_ZM => "en_ZM",
            Locale::en_ZW => "en_ZW",
            Locale::eo => "eo",
            Locale::es_AR => "es_AR",
            Locale::es_BO => "es_BO",
            Locale::es_CL => "es_CL",
            Locale::es_CO => "es_CO",
            Locale::es_CR => "es_CR",
            Locale::es_CU => "es_CU",
            Locale::es_DO => "es_DO",
            Locale::es_EC => "es_EC",
            Locale::es_ES => "es_ES",
            Locale::es_ES_euro => "es_ES@euro",
            Locale::es_GT => "es_GT",
            Locale::es_HN => "es_HN",
            Locale::es_MX => "es_MX",
            Locale::es_NI => "es_NI",
            Locale::es_PA => "es_PA",
            Locale::es_PE => "es_PE",
            Locale::es_PR => "es_PR",
            Locale::es_PY => "es_PY",
            Locale::es_SV => "es_SV",
            Locale::es_US => "es_US",
            Locale::es_UY => "es_UY",
            Locale::es_VE => "es_VE",
            Locale::et_EE => "et_EE",
            Locale::eu_ES => "eu_ES",
            Locale::eu_ES_euro => "eu_ES@euro",
            Locale::fa_IR => "fa_IR",
            Locale::ff_SN => "ff_SN",
            Locale::fi_FI => "fi_FI",
            Locale::fi_FI_euro => "fi_FI@euro",
            Locale::fil_PH => "fil_PH",
            Locale::fo_FO => "fo_FO",
            Locale::fr_BE => "fr_BE",
            Locale::fr_BE_euro => "fr_BE@euro",
            Locale::fr_CA => "fr_CA",
            Locale::fr_CH => "fr_CH",
            Locale::fr_FR => "fr_FR",
            Locale::fr_FR_euro => "fr_FR@euro",
            Locale::fr_LU => "fr_LU",
            Locale::fr_LU_euro => "fr_LU@euro",
            Locale::fur_IT => "fur_IT",
            Locale::fy_DE => "fy_DE",
            Locale::fy_NL => "fy_NL",
            Locale::ga_IE => "ga_IE",
            Locale::ga_IE_euro => "ga_IE@euro",
            Locale::gd_GB => "gd_GB",
            Locale::gez_ER => "gez_ER",
            Locale::gez_ER_abegede => "gez_ER@abegede",
            Locale::gez_ET => "gez_ET",
            Locale::gez_ET_abegede => "gez_ET@abegede",
            Locale::gl_ES => "gl_ES",
            Locale::gl_ES_euro => "gl_ES@euro",
            Locale::gu_IN => "gu_IN",
            Locale::gv_GB => "gv_GB",
            Locale::ha_NG => "ha_NG",
            Locale::hak_TW => "hak_TW",
            Locale::he_IL => "he_IL",
            Locale::hi_IN => "hi_IN",
            Locale::hif_FJ => "hif_FJ",
            Locale::hne_IN => "hne_IN",
            Locale::hr_HR => "hr_HR",
            Locale::hsb_DE => "hsb_DE",
            Locale::ht_HT => "ht_HT",
            Locale::hu_HU => "hu_HU",
            Locale::hy_AM => "hy_AM",
            Locale::ia_FR => "ia_FR",
            Locale::id_ID => "id_ID",
            Locale::ig_NG => "ig_NG",
            Locale::ik_CA => "ik_CA",
            Locale::is_IS => "is_IS",
            Locale::it_CH => "it_CH",
            Locale::it_IT => "it_IT",
            Locale::it_IT_euro => "it_IT@euro",
            Locale::iu_CA => "iu_CA",
            Locale::ja_JP => "ja_JP",
            Locale::ka_GE => "ka_GE",
            Locale::kab_DZ => "kab_DZ",
            Locale::kk_KZ => "kk_KZ",
            Locale::kl_GL => "kl_GL",
            Locale::km_KH => "km_KH",
            Locale::kn_IN => "kn_IN",
            Locale::ko_KR => "ko_KR",
            Locale::kok_IN => "kok_IN",
            Locale::ks_IN => "ks_IN",
            Locale::ks_IN_devanagari => "ks_IN@devanagari",
            Locale::ku_TR => "ku_TR",
            Locale::kw_GB => "kw_GB",
            Locale::ky_KG => "ky_KG",
            Locale::lb_LU => "lb_LU",
            Locale::lg_UG => "lg_UG",
            Locale::li_BE => "li_BE",
            Locale::li_NL => "li_NL",
            Locale::lij_IT => "lij_IT",
            Locale::ln_CD => "ln_CD",
            Locale::lo_LA => "lo_LA",
            Locale::lt_LT => "lt_LT",
            Locale::lv_LV => "lv_LV",
            Locale::lzh_TW => "lzh_TW",
            Locale::mag_IN => "mag_IN",
            Locale::mai_IN => "mai_IN",
            Locale::mai_NP => "mai_NP",
            Locale::mfe_MU => "mfe_MU",
            Locale::mg_MG => "mg_MG",
            Locale::mhr_RU => "mhr_RU",
            Locale::mi_NZ => "mi_NZ",
            Locale::miq_NI => "miq_NI",
            Locale::mjw_IN => "mjw_IN",
            Locale::mk_MK => "mk_MK",
            Locale::ml_IN => "ml_IN",
            Locale::mn_MN => "mn_MN",
            Locale::mni_IN => "mni_IN",
            Locale::mnw_MM => "mnw_MM",
            Locale::mr_IN => "mr_IN",
            Locale::ms_MY => "ms_MY",
            Locale::mt_MT => "mt_MT",
            Locale::my_MM => "my_MM",
            Locale::nan_TW => "nan_TW",
            Locale::nan_TW_latin => "nan_TW@latin",
            Locale::nb_NO => "nb_NO",
            Locale::nds_DE => "nds_DE",
            Locale::nds_NL => "nds_NL",
            Locale::ne_NP => "ne_NP",
            Locale::nhn_MX => "nhn_MX",
            Locale::niu_NU => "niu_NU",
            Locale::niu_NZ => "niu_NZ",
            Locale::nl_AW => "nl_AW",
            Locale::nl_BE => "nl_BE",
            Locale::nl_BE_euro => "nl_BE@euro",
            Locale::nl_NL => "nl_NL",
            Locale::nl_NL_euro => "nl_NL@euro",
            Locale::nn_NO => "nn_NO",
            Locale::nr_ZA => "nr_ZA",
            Locale::nso_ZA => "nso_ZA",
            Locale::oc_FR => "oc_FR",
            Locale::om_ET => "om_ET",
            Locale::om_KE => "om_KE",
            Locale::or_IN => "or_IN",
            Locale::os_RU => "os_RU",
            Locale::pa_IN => "pa_IN",
            Locale::pa_PK => "pa_PK",
            Locale::pap_AW => "pap_AW",
            Locale::pap_CW => "pap_CW",
            Locale::pl_PL => "pl_PL",
            Locale::ps_AF => "ps_AF",
            Locale::pt_BR => "pt_BR",
            Locale::pt_PT => "pt_PT",
            Locale::pt_PT_euro => "pt_PT@euro",
            Locale::quz_PE => "quz_PE",
            Locale::raj_IN => "raj_IN",
            Locale::ro_RO => "ro_RO",
            Locale::ru_RU => "ru_RU",
            Locale::ru_UA => "ru_UA",
            Locale::rw_RW => "rw_RW",
            Locale::sa_IN => "sa_IN",
            Locale::sah_RU => "sah_RU",
            Locale::sat_IN => "sat_IN",
            Locale::sc_IT => "sc_IT",
            Locale::sd_IN => "sd_IN",
            Locale::sd_IN_devanagari => "sd_IN@devanagari",
            Locale::se_NO => "se_NO",
            Locale::sgs_LT => "sgs_LT",
            Locale::shn_MM => "shn_MM",
            Locale::shs_CA => "shs_CA",
            Locale::si_LK => "si_LK",
            Locale::sid_ET => "sid_ET",
            Locale::sk_SK => "sk_SK",
            Locale::sl_SI => "sl_SI",
            Locale::sm_WS => "sm_WS",
            Locale::so_DJ => "so_DJ",
            Locale::so_ET => "so_ET",
            Locale::so_KE => "so_KE",
            Locale::so_SO => "so_SO",
            Locale::sq_AL => "sq_AL",
            Locale::sq_MK => "sq_MK",
            Locale::sr_ME => "sr_ME",
            Locale::sr_RS => "sr_RS",
            Locale::sr_RS_latin => "sr_RS@latin",
            Locale::ss_ZA => "ss_ZA",
            Locale::st_ZA => "st_ZA",
            Locale::sv_FI => "sv_FI",
            Locale::sv_FI_euro => "sv_FI@euro",
            Locale::sv_SE => "sv_SE",
            Locale::sw_KE => "sw_KE",
            Locale::sw_TZ => "sw_TZ",
            Locale::szl_PL => "szl_PL",
            Locale::ta_IN => "ta_IN",
            Locale::ta_LK => "ta_LK",
            Locale::tcy_IN => "tcy_IN",
            Locale::te_IN => "te_IN",
            Locale::tg_TJ => "tg_TJ",
            Locale::th_TH => "th_TH",
            Locale::the_NP => "the_NP",
            Locale::ti_ER => "ti_ER",
            Locale::ti_ET => "ti_ET",
            Locale::tig_ER => "tig_ER",
            Locale::tk_TM => "tk_TM",
            Locale::tl_PH => "tl_PH",
            Locale::tn_ZA => "tn_ZA",
            Locale::to_TO => "to_TO",
            Locale::tpi_PG => "tpi_PG",
            Locale::tr_CY => "tr_CY",
            Locale::tr_TR => "tr_TR",
            Locale::ts_ZA => "ts_ZA",
            Locale::tt_RU => "tt_RU",
            Locale::tt_RU_iqtelif => "tt_RU@iqtelif",
            Locale::ug_CN => "ug_CN",
            Locale::uk_UA => "uk_UA",
            Locale::unm_US => "unm_US",
            Locale::ur_IN => "ur_IN",
            Locale::ur_PK => "ur_PK",
            Locale::uz_UZ => "uz_UZ",
            Locale::uz_UZ_cyrillic => "uz_UZ@cyrillic",
            Locale::ve_ZA => "ve_ZA",
            Locale::vi_VN => "vi_VN",
            Locale::wa_BE => "wa_BE",
            Locale::wa_BE_euro => "wa_BE@euro",
            Locale::wae_CH => "wae_CH",
            Locale::wal_ET => "wal_ET",
            Locale::wo_SN => "wo_SN",
            Locale::xh_ZA => "xh_ZA",
            Locale::yi_US => "yi_US",
            Locale::yo_NG => "yo_NG",
            Locale::yue_HK => "yue_HK",
            Locale::yuw_PG => "yuw_PG",
            Locale::zh_CN => "zh_CN",
            Locale::zh_HK => "zh_HK",
            Locale::zh_SG => "zh_SG",
            Locale::zh_TW => "zh_TW",
            Locale::zu_ZA => "zu_ZA",
        })
    }
}

impl core::fmt::Debug for Locale {
    fn fmt(&self, f: &mut core::fmt::Formatter) -> core::fmt::Result {
        core::fmt::Display::fmt(self, f)
    }
}

#[cfg(feature = "defmt")]
impl defmt::Format for Locale {
    fn format(&self, f: defmt::Formatter) {
        defmt::write!(f, "{:?}", self);
    }
}

impl core::str::FromStr for Locale {
    type Err = UnknownLocale;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        core::convert::TryFrom::<&str>::try_from(s)
    }
}

impl core::convert::TryFrom<&str> for Locale {
    type Error = UnknownLocale;

    fn try_from(i: &str) -> Result<Self, Self::Error> {
        match i {
            "POSIX" => Ok(Locale::POSIX),
            "aa_DJ" => Ok(Locale::aa_DJ),
            "aa_ER" => Ok(Locale::aa_ER),
            "aa_ER@saaho" => Ok(Locale::aa_ER_saaho),
            "aa_ET" => Ok(Locale::aa_ET),
            "af_ZA" => Ok(Locale::af_ZA),
            "agr_PE" => Ok(Locale::agr_PE),
            "ak_GH" => Ok(Locale::ak_GH),
            "am_ET" => Ok(Locale::am_ET),
            "an_ES" => Ok(Locale::an_ES),
            "anp_IN" => Ok(Locale::anp_IN),
            "ar_AE" => Ok(Locale::ar_AE),
            "ar_BH" => Ok(Locale::ar_BH),
            "ar_DZ" => Ok(Locale::ar_DZ),
            "ar_EG" => Ok(Locale::ar_EG),
            "ar_IN" => Ok(Locale::ar_IN),
            "ar_IQ" => Ok(Locale::ar_IQ),
            "ar_JO" => Ok(Locale::ar_JO),
            "ar_KW" => Ok(Locale::ar_KW),
            "ar_LB" => Ok(Locale::ar_LB),
            "ar_LY" => Ok(Locale::ar_LY),
            "ar_MA" => Ok(Locale::ar_MA),
            "ar_OM" => Ok(Locale::ar_OM),
            "ar_QA" => Ok(Locale::ar_QA),
            "ar_SA" => Ok(Locale::ar_SA),
            "ar_SD" => Ok(Locale::ar_SD),
            "ar_SS" => Ok(Locale::ar_SS),
            "ar_SY" => Ok(Locale::ar_SY),
            "ar_TN" => Ok(Locale::ar_TN),
            "ar_YE" => Ok(Locale::ar_YE),
            "as_IN" => Ok(Locale::as_IN),
            "ast_ES" => Ok(Locale::ast_ES),
            "ayc_PE" => Ok(Locale::ayc_PE),
            "az_AZ" => Ok(Locale::az_AZ),
            "az_IR" => Ok(Locale::az_IR),
            "be_BY" => Ok(Locale::be_BY),
            "be_BY@latin" => Ok(Locale::be_BY_latin),
            "bem_ZM" => Ok(Locale::bem_ZM),
            "ber_DZ" => Ok(Locale::ber_DZ),
            "ber_MA" => Ok(Locale::ber_MA),
            "bg_BG" => Ok(Locale::bg_BG),
            "bhb_IN" => Ok(Locale::bhb_IN),
            "bho_IN" => Ok(Locale::bho_IN),
            "bho_NP" => Ok(Locale::bho_NP),
            "bi_VU" => Ok(Locale::bi_VU),
            "bn_BD" => Ok(Locale::bn_BD),
            "bn_IN" => Ok(Locale::bn_IN),
            "bo_CN" => Ok(Locale::bo_CN),
            "bo_IN" => Ok(Locale::bo_IN),
            "br_FR" => Ok(Locale::br_FR),
            "br_FR@euro" => Ok(Locale::br_FR_euro),
            "brx_IN" => Ok(Locale::brx_IN),
            "bs_BA" => Ok(Locale::bs_BA),
            "byn_ER" => Ok(Locale::byn_ER),
            "ca_AD" => Ok(Locale::ca_AD),
            "ca_ES" => Ok(Locale::ca_ES),
            "ca_ES@euro" => Ok(Locale::ca_ES_euro),
            "ca_ES@valencia" => Ok(Locale::ca_ES_valencia),
            "ca_FR" => Ok(Locale::ca_FR),
            "ca_IT" => Ok(Locale::ca_IT),
            "ce_RU" => Ok(Locale::ce_RU),
            "chr_US" => Ok(Locale::chr_US),
            "cmn_TW" => Ok(Locale::cmn_TW),
            "crh_UA" => Ok(Locale::crh_UA),
            "cs_CZ" => Ok(Locale::cs_CZ),
            "csb_PL" => Ok(Locale::csb_PL),
            "cv_RU" => Ok(Locale::cv_RU),
            "cy_GB" => Ok(Locale::cy_GB),
            "da_DK" => Ok(Locale::da_DK),
            "de_AT" => Ok(Locale::de_AT),
            "de_AT@euro" => Ok(Locale::de_AT_euro),
            "de_BE" => Ok(Locale::de_BE),
            "de_BE@euro" => Ok(Locale::de_BE_euro),
            "de_CH" => Ok(Locale::de_CH),
            "de_DE" => Ok(Locale::de_DE),
            "de_DE@euro" => Ok(Locale::de_DE_euro),
            "de_IT" => Ok(Locale::de_IT),
            "de_LI" => Ok(Locale::de_LI),
            "de_LU" => Ok(Locale::de_LU),
            "de_LU@euro" => Ok(Locale::de_LU_euro),
            "doi_IN" => Ok(Locale::doi_IN),
            "dsb_DE" => Ok(Locale::dsb_DE),
            "dv_MV" => Ok(Locale::dv_MV),
            "dz_BT" => Ok(Locale::dz_BT),
            "el_CY" => Ok(Locale::el_CY),
            "el_GR" => Ok(Locale::el_GR),
            "el_GR@euro" => Ok(Locale::el_GR_euro),
            "en_AG" => Ok(Locale::en_AG),
            "en_AU" => Ok(Locale::en_AU),
            "en_BW" => Ok(Locale::en_BW),
            "en_CA" => Ok(Locale::en_CA),
            "en_DK" => Ok(Locale::en_DK),
            "en_GB" => Ok(Locale::en_GB),
            "en_HK" => Ok(Locale::en_HK),
            "en_IE" => Ok(Locale::en_IE),
            "en_IE@euro" => Ok(Locale::en_IE_euro),
            "en_IL" => Ok(Locale::en_IL),
            "en_IN" => Ok(Locale::en_IN),
            "en_NG" => Ok(Locale::en_NG),
            "en_NZ" => Ok(Locale::en_NZ),
            "en_PH" => Ok(Locale::en_PH),
            "en_SC" => Ok(Locale::en_SC),
            "en_SG" => Ok(Locale::en_SG),
            "en_US" => Ok(Locale::en_US),
            "en_ZA" => Ok(Locale::en_ZA),
            "en_ZM" => Ok(Locale::en_ZM),
            "en_ZW" => Ok(Locale::en_ZW),
            "eo" => Ok(Locale::eo),
            "es_AR" => Ok(Locale::es_AR),
            "es_BO" => Ok(Locale::es_BO),
            "es_CL" => Ok(Locale::es_CL),
            "es_CO" => Ok(Locale::es_CO),
            "es_CR" => Ok(Locale::es_CR),
            "es_CU" => Ok(Locale::es_CU),
            "es_DO" => Ok(Locale::es_DO),
            "es_EC" => Ok(Locale::es_EC),
            "es_ES" => Ok(Locale::es_ES),
            "es_ES@euro" => Ok(Locale::es_ES_euro),
            "es_GT" => Ok(Locale::es_GT),
            "es_HN" => Ok(Locale::es_HN),
            "es_MX" => Ok(Locale::es_MX),
            "es_NI" => Ok(Locale::es_NI),
            "es_PA" => Ok(Locale::es_PA),
            "es_PE" => Ok(Locale::es_PE),
            "es_PR" => Ok(Locale::es_PR),
            "es_PY" => Ok(Locale::es_PY),
            "es_SV" => Ok(Locale::es_SV),
            "es_US" => Ok(Locale::es_US),
            "es_UY" => Ok(Locale::es_UY),
            "es_VE" => Ok(Locale::es_VE),
            "et_EE" => Ok(Locale::et_EE),
            "eu_ES" => Ok(Locale::eu_ES),
            "eu_ES@euro" => Ok(Locale::eu_ES_euro),
            "fa_IR" => Ok(Locale::fa_IR),
            "ff_SN" => Ok(Locale::ff_SN),
            "fi_FI" => Ok(Locale::fi_FI),
            "fi_FI@euro" => Ok(Locale::fi_FI_euro),
            "fil_PH" => Ok(Locale::fil_PH),
            "fo_FO" => Ok(Locale::fo_FO),
            "fr_BE" => Ok(Locale::fr_BE),
            "fr_BE@euro" => Ok(Locale::fr_BE_euro),
            "fr_CA" => Ok(Locale::fr_CA),
            "fr_CH" => Ok(Locale::fr_CH),
            "fr_FR" => Ok(Locale::fr_FR),
            "fr_FR@euro" => Ok(Locale::fr_FR_euro),
            "fr_LU" => Ok(Locale::fr_LU),
            "fr_LU@euro" => Ok(Locale::fr_LU_euro),
            "fur_IT" => Ok(Locale::fur_IT),
            "fy_DE" => Ok(Locale::fy_DE),
            "fy_NL" => Ok(Locale::fy_NL),
            "ga_IE" => Ok(Locale::ga_IE),
            "ga_IE@euro" => Ok(Locale::ga_IE_euro),
            "gd_GB" => Ok(Locale::gd_GB),
            "gez_ER" => Ok(Locale::gez_ER),
            "gez_ER@abegede" => Ok(Locale::gez_ER_abegede),
            "gez_ET" => Ok(Locale::gez_ET),
            "gez_ET@abegede" => Ok(Locale::gez_ET_abegede),
            "gl_ES" => Ok(Locale::gl_ES),
            "gl_ES@euro" => Ok(Locale::gl_ES_euro),
            "gu_IN" => Ok(Locale::gu_IN),
            "gv_GB" => Ok(Locale::gv_GB),
            "ha_NG" => Ok(Locale::ha_NG),
            "hak_TW" => Ok(Locale::hak_TW),
            "he_IL" => Ok(Locale::he_IL),
            "hi_IN" => Ok(Locale::hi_IN),
            "hif_FJ" => Ok(Locale::hif_FJ),
            "hne_IN" => Ok(Locale::hne_IN),
            "hr_HR" => Ok(Locale::hr_HR),
            "hsb_DE" => Ok(Locale::hsb_DE),
            "ht_HT" => Ok(Locale::ht_HT),
            "hu_HU" => Ok(Locale::hu_HU),
            "hy_AM" => Ok(Locale::hy_AM),
            "ia_FR" => Ok(Locale::ia_FR),
            "id_ID" => Ok(Locale::id_ID),
            "ig_NG" => Ok(Locale::ig_NG),
            "ik_CA" => Ok(Locale::ik_CA),
            "is_IS" => Ok(Locale::is_IS),
            "it_CH" => Ok(Locale::it_CH),
            "it_IT" => Ok(Locale::it_IT),
            "it_IT@euro" => Ok(Locale::it_IT_euro),
            "iu_CA" => Ok(Locale::iu_CA),
            "ja_JP" => Ok(Locale::ja_JP),
            "ka_GE" => Ok(Locale::ka_GE),
            "kab_DZ" => Ok(Locale::kab_DZ),
            "kk_KZ" => Ok(Locale::kk_KZ),
            "kl_GL" => Ok(Locale::kl_GL),
            "km_KH" => Ok(Locale::km_KH),
            "kn_IN" => Ok(Locale::kn_IN),
            "ko_KR" => Ok(Locale::ko_KR),
            "kok_IN" => Ok(Locale::kok_IN),
            "ks_IN" => Ok(Locale::ks_IN),
            "ks_IN@devanagari" => Ok(Locale::ks_IN_devanagari),
            "ku_TR" => Ok(Locale::ku_TR),
            "kw_GB" => Ok(Locale::kw_GB),
            "ky_KG" => Ok(Locale::ky_KG),
            "lb_LU" => Ok(Locale::lb_LU),
            "lg_UG" => Ok(Locale::lg_UG),
            "li_BE" => Ok(Locale::li_BE),
            "li_NL" => Ok(Locale::li_NL),
            "lij_IT" => Ok(Locale::lij_IT),
            "ln_CD" => Ok(Locale::ln_CD),
            "lo_LA" => Ok(Locale::lo_LA),
            "lt_LT" => Ok(Locale::lt_LT),
            "lv_LV" => Ok(Locale::lv_LV),
            "lzh_TW" => Ok(Locale::lzh_TW),
            "mag_IN" => Ok(Locale::mag_IN),
            "mai_IN" => Ok(Locale::mai_IN),
            "mai_NP" => Ok(Locale::mai_NP),
            "mfe_MU" => Ok(Locale::mfe_MU),
            "mg_MG" => Ok(Locale::mg_MG),
            "mhr_RU" => Ok(Locale::mhr_RU),
            "mi_NZ" => Ok(Locale::mi_NZ),
            "miq_NI" => Ok(Locale::miq_NI),
            "mjw_IN" => Ok(Locale::mjw_IN),
            "mk_MK" => Ok(Locale::mk_MK),
            "ml_IN" => Ok(Locale::ml_IN),
            "mn_MN" => Ok(Locale::mn_MN),
            "mni_IN" => Ok(Locale::mni_IN),
            "mnw_MM" => Ok(Locale::mnw_MM),
            "mr_IN" => Ok(Locale::mr_IN),
            "ms_MY" => Ok(Locale::ms_MY),
            "mt_MT" => Ok(Locale::mt_MT),
            "my_MM" => Ok(Locale::my_MM),
            "nan_TW" => Ok(Locale::nan_TW),
            "nan_TW@latin" => Ok(Locale::nan_TW_latin),
            "nb_NO" => Ok(Locale::nb_NO),
            "nds_DE" => Ok(Locale::nds_DE),
            "nds_NL" => Ok(Locale::nds_NL),
            "ne_NP" => Ok(Locale::ne_NP),
            "nhn_MX" => Ok(Locale::nhn_MX),
            "niu_NU" => Ok(Locale::niu_NU),
            "niu_NZ" => Ok(Locale::niu_NZ),
            "nl_AW" => Ok(Locale::nl_AW),
            "nl_BE" => Ok(Locale::nl_BE),
            "nl_BE@euro" => Ok(Locale::nl_BE_euro),
            "nl_NL" => Ok(Locale::nl_NL),
            "nl_NL@euro" => Ok(Locale::nl_NL_euro),
            "nn_NO" => Ok(Locale::nn_NO),
            "nr_ZA" => Ok(Locale::nr_ZA),
            "nso_ZA" => Ok(Locale::nso_ZA),
            "oc_FR" => Ok(Locale::oc_FR),
            "om_ET" => Ok(Locale::om_ET),
            "om_KE" => Ok(Locale::om_KE),
            "or_IN" => Ok(Locale::or_IN),
            "os_RU" => Ok(Locale::os_RU),
            "pa_IN" => Ok(Locale::pa_IN),
            "pa_PK" => Ok(Locale::pa_PK),
            "pap_AW" => Ok(Locale::pap_AW),
            "pap_CW" => Ok(Locale::pap_CW),
            "pl_PL" => Ok(Locale::pl_PL),
            "ps_AF" => Ok(Locale::ps_AF),
            "pt_BR" => Ok(Locale::pt_BR),
            "pt_PT" => Ok(Locale::pt_PT),
            "pt_PT@euro" => Ok(Locale::pt_PT_euro),
            "quz_PE" => Ok(Locale::quz_PE),
            "raj_IN" => Ok(Locale::raj_IN),
            "ro_RO" => Ok(Locale::ro_RO),
            "ru_RU" => Ok(Locale::ru_RU),
            "ru_UA" => Ok(Locale::ru_UA),
            "rw_RW" => Ok(Locale::rw_RW),
            "sa_IN" => Ok(Locale::sa_IN),
            "sah_RU" => Ok(Locale::sah_RU),
            "sat_IN" => Ok(Locale::sat_IN),
            "sc_IT" => Ok(Locale::sc_IT),
            "sd_IN" => Ok(Locale::sd_IN),
            "sd_IN@devanagari" => Ok(Locale::sd_IN_devanagari),
            "se_NO" => Ok(Locale::se_NO),
            "sgs_LT" => Ok(Locale::sgs_LT),
            "shn_MM" => Ok(Locale::shn_MM),
            "shs_CA" => Ok(Locale::shs_CA),
            "si_LK" => Ok(Locale::si_LK),
            "sid_ET" => Ok(Locale::sid_ET),
            "sk_SK" => Ok(Locale::sk_SK),
            "sl_SI" => Ok(Locale::sl_SI),
            "sm_WS" => Ok(Locale::sm_WS),
            "so_DJ" => Ok(Locale::so_DJ),
            "so_ET" => Ok(Locale::so_ET),
            "so_KE" => Ok(Locale::so_KE),
            "so_SO" => Ok(Locale::so_SO),
            "sq_AL" => Ok(Locale::sq_AL),
            "sq_MK" => Ok(Locale::sq_MK),
            "sr_ME" => Ok(Locale::sr_ME),
            "sr_RS" => Ok(Locale::sr_RS),
            "sr_RS@latin" => Ok(Locale::sr_RS_latin),
            "ss_ZA" => Ok(Locale::ss_ZA),
            "st_ZA" => Ok(Locale::st_ZA),
            "sv_FI" => Ok(Locale::sv_FI),
            "sv_FI@euro" => Ok(Locale::sv_FI_euro),
            "sv_SE" => Ok(Locale::sv_SE),
            "sw_KE" => Ok(Locale::sw_KE),
            "sw_TZ" => Ok(Locale::sw_TZ),
            "szl_PL" => Ok(Locale::szl_PL),
            "ta_IN" => Ok(Locale::ta_IN),
            "ta_LK" => Ok(Locale::ta_LK),
            "tcy_IN" => Ok(Locale::tcy_IN),
            "te_IN" => Ok(Locale::te_IN),
            "tg_TJ" => Ok(Locale::tg_TJ),
            "th_TH" => Ok(Locale::th_TH),
            "the_NP" => Ok(Locale::the_NP),
            "ti_ER" => Ok(Locale::ti_ER),
            "ti_ET" => Ok(Locale::ti_ET),
            "tig_ER" => Ok(Locale::tig_ER),
            "tk_TM" => Ok(Locale::tk_TM),
            "tl_PH" => Ok(Locale::tl_PH),
            "tn_ZA" => Ok(Locale::tn_ZA),
            "to_TO" => Ok(Locale::to_TO),
            "tpi_PG" => Ok(Locale::tpi_PG),
            "tr_CY" => Ok(Locale::tr_CY),
            "tr_TR" => Ok(Locale::tr_TR),
            "ts_ZA" => Ok(Locale::ts_ZA),
            "tt_RU" => Ok(Locale::tt_RU),
            "tt_RU@iqtelif" => Ok(Locale::tt_RU_iqtelif),
            "ug_CN" => Ok(Locale::ug_CN),
            "uk_UA" => Ok(Locale::uk_UA),
            "unm_US" => Ok(Locale::unm_US),
            "ur_IN" => Ok(Locale::ur_IN),
            "ur_PK" => Ok(Locale::ur_PK),
            "uz_UZ" => Ok(Locale::uz_UZ),
            "uz_UZ@cyrillic" => Ok(Locale::uz_UZ_cyrillic),
            "ve_ZA" => Ok(Locale::ve_ZA),
            "vi_VN" => Ok(Locale::vi_VN),
            "wa_BE" => Ok(Locale::wa_BE),
            "wa_BE@euro" => Ok(Locale::wa_BE_euro),
            "wae_CH" => Ok(Locale::wae_CH),
            "wal_ET" => Ok(Locale::wal_ET),
            "wo_SN" => Ok(Locale::wo_SN),
            "xh_ZA" => Ok(Locale::xh_ZA),
            "yi_US" => Ok(Locale::yi_US),
            "yo_NG" => Ok(Locale::yo_NG),
            "yue_HK" => Ok(Locale::yue_HK),
            "yuw_PG" => Ok(Locale::yuw_PG),
            "zh_CN" => Ok(Locale::zh_CN),
            "zh_HK" => Ok(Locale::zh_HK),
            "zh_SG" => Ok(Locale::zh_SG),
            "zh_TW" => Ok(Locale::zh_TW),
            "zu_ZA" => Ok(Locale::zu_ZA),
            _ => Err(UnknownLocale),
        }
    }
}

#[macro_export]
macro_rules! locale_match {
    ($locale:expr => $($item:ident)::+) => {{
        match $locale {
            $crate::Locale::POSIX => $crate::POSIX::$($item)::+,
            $crate::Locale::aa_DJ => $crate::aa_DJ::$($item)::+,
            $crate::Locale::aa_ER => $crate::aa_ER::$($item)::+,
            $crate::Locale::aa_ER_saaho => $crate::aa_ER_saaho::$($item)::+,
            $crate::Locale::aa_ET => $crate::aa_ET::$($item)::+,
            $crate::Locale::af_ZA => $crate::af_ZA::$($item)::+,
            $crate::Locale::agr_PE => $crate::agr_PE::$($item)::+,
            $crate::Locale::ak_GH => $crate::ak_GH::$($item)::+,
            $crate::Locale::am_ET => $crate::am_ET::$($item)::+,
            $crate::Locale::an_ES => $crate::an_ES::$($item)::+,
            $crate::Locale::anp_IN => $crate::anp_IN::$($item)::+,
            $crate::Locale::ar_AE => $crate::ar_AE::$($item)::+,
            $crate::Locale::ar_BH => $crate::ar_BH::$($item)::+,
            $crate::Locale::ar_DZ => $crate::ar_DZ::$($item)::+,
            $crate::Locale::ar_EG => $crate::ar_EG::$($item)::+,
            $crate::Locale::ar_IN => $crate::ar_IN::$($item)::+,
            $crate::Locale::ar_IQ => $crate::ar_IQ::$($item)::+,
            $crate::Locale::ar_JO => $crate::ar_JO::$($item)::+,
            $crate::Locale::ar_KW => $crate::ar_KW::$($item)::+,
            $crate::Locale::ar_LB => $crate::ar_LB::$($item)::+,
            $crate::Locale::ar_LY => $crate::ar_LY::$($item)::+,
            $crate::Locale::ar_MA => $crate::ar_MA::$($item)::+,
            $crate::Locale::ar_OM => $crate::ar_OM::$($item)::+,
            $crate::Locale::ar_QA => $crate::ar_QA::$($item)::+,
            $crate::Locale::ar_SA => $crate::ar_SA::$($item)::+,
            $crate::Locale::ar_SD => $crate::ar_SD::$($item)::+,
            $crate::Locale::ar_SS => $crate::ar_SS::$($item)::+,
            $crate::Locale::ar_SY => $crate::ar_SY::$($item)::+,
            $crate::Locale::ar_TN => $crate::ar_TN::$($item)::+,
            $crate::Locale::ar_YE => $crate::ar_YE::$($item)::+,
            $crate::Locale::as_IN => $crate::as_IN::$($item)::+,
            $crate::Locale::ast_ES => $crate::ast_ES::$($item)::+,
            $crate::Locale::ayc_PE => $crate::ayc_PE::$($item)::+,
            $crate::Locale::az_AZ => $crate::az_AZ::$($item)::+,
            $crate::Locale::az_IR => $crate::az_IR::$($item)::+,
            $crate::Locale::be_BY => $crate::be_BY::$($item)::+,
            $crate::Locale::be_BY_latin => $crate::be_BY_latin::$($item)::+,
            $crate::Locale::bem_ZM => $crate::bem_ZM::$($item)::+,
            $crate::Locale::ber_DZ => $crate::ber_DZ::$($item)::+,
            $crate::Locale::ber_MA => $crate::ber_MA::$($item)::+,
            $crate::Locale::bg_BG => $crate::bg_BG::$($item)::+,
            $crate::Locale::bhb_IN => $crate::bhb_IN::$($item)::+,
            $crate::Locale::bho_IN => $crate::bho_IN::$($item)::+,
            $crate::Locale::bho_NP => $crate::bho_NP::$($item)::+,
            $crate::Locale::bi_VU => $crate::bi_VU::$($item)::+,
            $crate::Locale::bn_BD => $crate::bn_BD::$($item)::+,
            $crate::Locale::bn_IN => $crate::bn_IN::$($item)::+,
            $crate::Locale::bo_CN => $crate::bo_CN::$($item)::+,
            $crate::Locale::bo_IN => $crate::bo_IN::$($item)::+,
            $crate::Locale::br_FR => $crate::br_FR::$($item)::+,
            $crate::Locale::br_FR_euro => $crate::br_FR_euro::$($item)::+,
            $crate::Locale::brx_IN => $crate::brx_IN::$($item)::+,
            $crate::Locale::bs_BA => $crate::bs_BA::$($item)::+,
            $crate::Locale::byn_ER => $crate::byn_ER::$($item)::+,
            $crate::Locale::ca_AD => $crate::ca_AD::$($item)::+,
            $crate::Locale::ca_ES => $crate::ca_ES::$($item)::+,
            $crate::Locale::ca_ES_euro => $crate::ca_ES_euro::$($item)::+,
            $crate::Locale::ca_ES_valencia => $crate::ca_ES_valencia::$($item)::+,
            $crate::Locale::ca_FR => $crate::ca_FR::$($item)::+,
            $crate::Locale::ca_IT => $crate::ca_IT::$($item)::+,
            $crate::Locale::ce_RU => $crate::ce_RU::$($item)::+,
            $crate::Locale::chr_US => $crate::chr_US::$($item)::+,
            $crate::Locale::cmn_TW => $crate::cmn_TW::$($item)::+,
            $crate::Locale::crh_UA => $crate::crh_UA::$($item)::+,
            $crate::Locale::cs_CZ => $crate::cs_CZ::$($item)::+,
            $crate::Locale::csb_PL => $crate::csb_PL::$($item)::+,
            $crate::Locale::cv_RU => $crate::cv_RU::$($item)::+,
            $crate::Locale::cy_GB => $crate::cy_GB::$($item)::+,
            $crate::Locale::da_DK => $crate::da_DK::$($item)::+,
            $crate::Locale::de_AT => $crate::de_AT::$($item)::+,
            $crate::Locale::de_AT_euro => $crate::de_AT_euro::$($item)::+,
            $crate::Locale::de_BE => $crate::de_BE::$($item)::+,
            $crate::Locale::de_BE_euro => $crate::de_BE_euro::$($item)::+,
            $crate::Locale::de_CH => $crate::de_CH::$($item)::+,
            $crate::Locale::de_DE => $crate::de_DE::$($item)::+,
            $crate::Locale::de_DE_euro => $crate::de_DE_euro::$($item)::+,
            $crate::Locale::de_IT => $crate::de_IT::$($item)::+,
            $crate::Locale::de_LI => $crate::de_LI::$($item)::+,
            $crate::Locale::de_LU => $crate::de_LU::$($item)::+,
            $crate::Locale::de_LU_euro => $crate::de_LU_euro::$($item)::+,
            $crate::Locale::doi_IN => $crate::doi_IN::$($item)::+,
            $crate::Locale::dsb_DE => $crate::dsb_DE::$($item)::+,
            $crate::Locale::dv_MV => $crate::dv_MV::$($item)::+,
            $crate::Locale::dz_BT => $crate::dz_BT::$($item)::+,
            $crate::Locale::el_CY => $crate::el_CY::$($item)::+,
            $crate::Locale::el_GR => $crate::el_GR::$($item)::+,
            $crate::Locale::el_GR_euro => $crate::el_GR_euro::$($item)::+,
            $crate::Locale::en_AG => $crate::en_AG::$($item)::+,
            $crate::Locale::en_AU => $crate::en_AU::$($item)::+,
            $crate::Locale::en_BW => $crate::en_BW::$($item)::+,
            $crate::Locale::en_CA => $crate::en_CA::$($item)::+,
            $crate::Locale::en_DK => $crate::en_DK::$($item)::+,
            $crate::Locale::en_GB => $crate::en_GB::$($item)::+,
            $crate::Locale::en_HK => $crate::en_HK::$($item)::+,
            $crate::Locale::en_IE => $crate::en_IE::$($item)::+,
            $crate::Locale::en_IE_euro => $crate::en_IE_euro::$($item)::+,
            $crate::Locale::en_IL => $crate::en_IL::$($item)::+,
            $crate::Locale::en_IN => $crate::en_IN::$($item)::+,
            $crate::Locale::en_NG => $crate::en_NG::$($item)::+,
            $crate::Locale::en_NZ => $crate::en_NZ::$($item)::+,
            $crate::Locale::en_PH => $crate::en_PH::$($item)::+,
            $crate::Locale::en_SC => $crate::en_SC::$($item)::+,
            $crate::Locale::en_SG => $crate::en_SG::$($item)::+,
            $crate::Locale::en_US => $crate::en_US::$($item)::+,
            $crate::Locale::en_ZA => $crate::en_ZA::$($item)::+,
            $crate::Locale::en_ZM => $crate::en_ZM::$($item)::+,
            $crate::Locale::en_ZW => $crate::en_ZW::$($item)::+,
            $crate::Locale::eo => $crate::eo::$($item)::+,
            $crate::Locale::es_AR => $crate::es_AR::$($item)::+,
            $crate::Locale::es_BO => $crate::es_BO::$($item)::+,
            $crate::Locale::es_CL => $crate::es_CL::$($item)::+,
            $crate::Locale::es_CO => $crate::es_CO::$($item)::+,
            $crate::Locale::es_CR => $crate::es_CR::$($item)::+,
            $crate::Locale::es_CU => $crate::es_CU::$($item)::+,
            $crate::Locale::es_DO => $crate::es_DO::$($item)::+,
            $crate::Locale::es_EC => $crate::es_EC::$($item)::+,
            $crate::Locale::es_ES => $crate::es_ES::$($item)::+,
            $crate::Locale::es_ES_euro => $crate::es_ES_euro::$($item)::+,
            $crate::Locale::es_GT => $crate::es_GT::$($item)::+,
            $crate::Locale::es_HN => $crate::es_HN::$($item)::+,
            $crate::Locale::es_MX => $crate::es_MX::$($item)::+,
            $crate::Locale::es_NI => $crate::es_NI::$($item)::+,
            $crate::Locale::es_PA => $crate::es_PA::$($item)::+,
            $crate::Locale::es_PE => $crate::es_PE::$($item)::+,
            $crate::Locale::es_PR => $crate::es_PR::$($item)::+,
            $crate::Locale::es_PY => $crate::es_PY::$($item)::+,
            $crate::Locale::es_SV => $crate::es_SV::$($item)::+,
            $crate::Locale::es_US => $crate::es_US::$($item)::+,
            $crate::Locale::es_UY => $crate::es_UY::$($item)::+,
            $crate::Locale::es_VE => $crate::es_VE::$($item)::+,
            $crate::Locale::et_EE => $crate::et_EE::$($item)::+,
            $crate::Locale::eu_ES => $crate::eu_ES::$($item)::+,
            $crate::Locale::eu_ES_euro => $crate::eu_ES_euro::$($item)::+,
            $crate::Locale::fa_IR => $crate::fa_IR::$($item)::+,
            $crate::Locale::ff_SN => $crate::ff_SN::$($item)::+,
            $crate::Locale::fi_FI => $crate::fi_FI::$($item)::+,
            $crate::Locale::fi_FI_euro => $crate::fi_FI_euro::$($item)::+,
            $crate::Locale::fil_PH => $crate::fil_PH::$($item)::+,
            $crate::Locale::fo_FO => $crate::fo_FO::$($item)::+,
            $crate::Locale::fr_BE => $crate::fr_BE::$($item)::+,
            $crate::Locale::fr_BE_euro => $crate::fr_BE_euro::$($item)::+,
            $crate::Locale::fr_CA => $crate::fr_CA::$($item)::+,
            $crate::Locale::fr_CH => $crate::fr_CH::$($item)::+,
            $crate::Locale::fr_FR => $crate::fr_FR::$($item)::+,
            $crate::Locale::fr_FR_euro => $crate::fr_FR_euro::$($item)::+,
            $crate::Locale::fr_LU => $crate::fr_LU::$($item)::+,
            $crate::Locale::fr_LU_euro => $crate::fr_LU_euro::$($item)::+,
            $crate::Locale::fur_IT => $crate::fur_IT::$($item)::+,
            $crate::Locale::fy_DE => $crate::fy_DE::$($item)::+,
            $crate::Locale::fy_NL => $crate::fy_NL::$($item)::+,
            $crate::Locale::ga_IE => $crate::ga_IE::$($item)::+,
            $crate::Locale::ga_IE_euro => $crate::ga_IE_euro::$($item)::+,
            $crate::Locale::gd_GB => $crate::gd_GB::$($item)::+,
            $crate::Locale::gez_ER => $crate::gez_ER::$($item)::+,
            $crate::Locale::gez_ER_abegede => $crate::gez_ER_abegede::$($item)::+,
            $crate::Locale::gez_ET => $crate::gez_ET::$($item)::+,
            $crate::Locale::gez_ET_abegede => $crate::gez_ET_abegede::$($item)::+,
            $crate::Locale::gl_ES => $crate::gl_ES::$($item)::+,
            $crate::Locale::gl_ES_euro => $crate::gl_ES_euro::$($item)::+,
            $crate::Locale::gu_IN => $crate::gu_IN::$($item)::+,
            $crate::Locale::gv_GB => $crate::gv_GB::$($item)::+,
            $crate::Locale::ha_NG => $crate::ha_NG::$($item)::+,
            $crate::Locale::hak_TW => $crate::hak_TW::$($item)::+,
            $crate::Locale::he_IL => $crate::he_IL::$($item)::+,
            $crate::Locale::hi_IN => $crate::hi_IN::$($item)::+,
            $crate::Locale::hif_FJ => $crate::hif_FJ::$($item)::+,
            $crate::Locale::hne_IN => $crate::hne_IN::$($item)::+,
            $crate::Locale::hr_HR => $crate::hr_HR::$($item)::+,
            $crate::Locale::hsb_DE => $crate::hsb_DE::$($item)::+,
            $crate::Locale::ht_HT => $crate::ht_HT::$($item)::+,
            $crate::Locale::hu_HU => $crate::hu_HU::$($item)::+,
            $crate::Locale::hy_AM => $crate::hy_AM::$($item)::+,
            $crate::Locale::ia_FR => $crate::ia_FR::$($item)::+,
            $crate::Locale::id_ID => $crate::id_ID::$($item)::+,
            $crate::Locale::ig_NG => $crate::ig_NG::$($item)::+,
            $crate::Locale::ik_CA => $crate::ik_CA::$($item)::+,
            $crate::Locale::is_IS => $crate::is_IS::$($item)::+,
            $crate::Locale::it_CH => $crate::it_CH::$($item)::+,
            $crate::Locale::it_IT => $crate::it_IT::$($item)::+,
            $crate::Locale::it_IT_euro => $crate::it_IT_euro::$($item)::+,
            $crate::Locale::iu_CA => $crate::iu_CA::$($item)::+,
            $crate::Locale::ja_JP => $crate::ja_JP::$($item)::+,
            $crate::Locale::ka_GE => $crate::ka_GE::$($item)::+,
            $crate::Locale::kab_DZ => $crate::kab_DZ::$($item)::+,
            $crate::Locale::kk_KZ => $crate::kk_KZ::$($item)::+,
            $crate::Locale::kl_GL => $crate::kl_GL::$($item)::+,
            $crate::Locale::km_KH => $crate::km_KH::$($item)::+,
            $crate::Locale::kn_IN => $crate::kn_IN::$($item)::+,
            $crate::Locale::ko_KR => $crate::ko_KR::$($item)::+,
            $crate::Locale::kok_IN => $crate::kok_IN::$($item)::+,
            $crate::Locale::ks_IN => $crate::ks_IN::$($item)::+,
            $crate::Locale::ks_IN_devanagari => $crate::ks_IN_devanagari::$($item)::+,
            $crate::Locale::ku_TR => $crate::ku_TR::$($item)::+,
            $crate::Locale::kw_GB => $crate::kw_GB::$($item)::+,
            $crate::Locale::ky_KG => $crate::ky_KG::$($item)::+,
            $crate::Locale::lb_LU => $crate::lb_LU::$($item)::+,
            $crate::Locale::lg_UG => $crate::lg_UG::$($item)::+,
            $crate::Locale::li_BE => $crate::li_BE::$($item)::+,
            $crate::Locale::li_NL => $crate::li_NL::$($item)::+,
            $crate::Locale::lij_IT => $crate::lij_IT::$($item)::+,
            $crate::Locale::ln_CD => $crate::ln_CD::$($item)::+,
            $crate::Locale::lo_LA => $crate::lo_LA::$($item)::+,
            $crate::Locale::lt_LT => $crate::lt_LT::$($item)::+,
            $crate::Locale::lv_LV => $crate::lv_LV::$($item)::+,
            $crate::Locale::lzh_TW => $crate::lzh_TW::$($item)::+,
            $crate::Locale::mag_IN => $crate::mag_IN::$($item)::+,
            $crate::Locale::mai_IN => $crate::mai_IN::$($item)::+,
            $crate::Locale::mai_NP => $crate::mai_NP::$($item)::+,
            $crate::Locale::mfe_MU => $crate::mfe_MU::$($item)::+,
            $crate::Locale::mg_MG => $crate::mg_MG::$($item)::+,
            $crate::Locale::mhr_RU => $crate::mhr_RU::$($item)::+,
            $crate::Locale::mi_NZ => $crate::mi_NZ::$($item)::+,
            $crate::Locale::miq_NI => $crate::miq_NI::$($item)::+,
            $crate::Locale::mjw_IN => $crate::mjw_IN::$($item)::+,
            $crate::Locale::mk_MK => $crate::mk_MK::$($item)::+,
            $crate::Locale::ml_IN => $crate::ml_IN::$($item)::+,
            $crate::Locale::mn_MN => $crate::mn_MN::$($item)::+,
            $crate::Locale::mni_IN => $crate::mni_IN::$($item)::+,
            $crate::Locale::mnw_MM => $crate::mnw_MM::$($item)::+,
            $crate::Locale::mr_IN => $crate::mr_IN::$($item)::+,
            $crate::Locale::ms_MY => $crate::ms_MY::$($item)::+,
            $crate::Locale::mt_MT => $crate::mt_MT::$($item)::+,
            $crate::Locale::my_MM => $crate::my_MM::$($item)::+,
            $crate::Locale::nan_TW => $crate::nan_TW::$($item)::+,
            $crate::Locale::nan_TW_latin => $crate::nan_TW_latin::$($item)::+,
            $crate::Locale::nb_NO => $crate::nb_NO::$($item)::+,
            $crate::Locale::nds_DE => $crate::nds_DE::$($item)::+,
            $crate::Locale::nds_NL => $crate::nds_NL::$($item)::+,
            $crate::Locale::ne_NP => $crate::ne_NP::$($item)::+,
            $crate::Locale::nhn_MX => $crate::nhn_MX::$($item)::+,
            $crate::Locale::niu_NU => $crate::niu_NU::$($item)::+,
            $crate::Locale::niu_NZ => $crate::niu_NZ::$($item)::+,
            $crate::Locale::nl_AW => $crate::nl_AW::$($item)::+,
            $crate::Locale::nl_BE => $crate::nl_BE::$($item)::+,
            $crate::Locale::nl_BE_euro => $crate::nl_BE_euro::$($item)::+,
            $crate::Locale::nl_NL => $crate::nl_NL::$($item)::+,
            $crate::Locale::nl_NL_euro => $crate::nl_NL_euro::$($item)::+,
            $crate::Locale::nn_NO => $crate::nn_NO::$($item)::+,
            $crate::Locale::nr_ZA => $crate::nr_ZA::$($item)::+,
            $crate::Locale::nso_ZA => $crate::nso_ZA::$($item)::+,
            $crate::Locale::oc_FR => $crate::oc_FR::$($item)::+,
            $crate::Locale::om_ET => $crate::om_ET::$($item)::+,
            $crate::Locale::om_KE => $crate::om_KE::$($item)::+,
            $crate::Locale::or_IN => $crate::or_IN::$($item)::+,
            $crate::Locale::os_RU => $crate::os_RU::$($item)::+,
            $crate::Locale::pa_IN => $crate::pa_IN::$($item)::+,
            $crate::Locale::pa_PK => $crate::pa_PK::$($item)::+,
            $crate::Locale::pap_AW => $crate::pap_AW::$($item)::+,
            $crate::Locale::pap_CW => $crate::pap_CW::$($item)::+,
            $crate::Locale::pl_PL => $crate::pl_PL::$($item)::+,
            $crate::Locale::ps_AF => $crate::ps_AF::$($item)::+,
            $crate::Locale::pt_BR => $crate::pt_BR::$($item)::+,
            $crate::Locale::pt_PT => $crate::pt_PT::$($item)::+,
            $crate::Locale::pt_PT_euro => $crate::pt_PT_euro::$($item)::+,
            $crate::Locale::quz_PE => $crate::quz_PE::$($item)::+,
            $crate::Locale::raj_IN => $crate::raj_IN::$($item)::+,
            $crate::Locale::ro_RO => $crate::ro_RO::$($item)::+,
            $crate::Locale::ru_RU => $crate::ru_RU::$($item)::+,
            $crate::Locale::ru_UA => $crate::ru_UA::$($item)::+,
            $crate::Locale::rw_RW => $crate::rw_RW::$($item)::+,
            $crate::Locale::sa_IN => $crate::sa_IN::$($item)::+,
            $crate::Locale::sah_RU => $crate::sah_RU::$($item)::+,
            $crate::Locale::sat_IN => $crate::sat_IN::$($item)::+,
            $crate::Locale::sc_IT => $crate::sc_IT::$($item)::+,
            $crate::Locale::sd_IN => $crate::sd_IN::$($item)::+,
            $crate::Locale::sd_IN_devanagari => $crate::sd_IN_devanagari::$($item)::+,
            $crate::Locale::se_NO => $crate::se_NO::$($item)::+,
            $crate::Locale::sgs_LT => $crate::sgs_LT::$($item)::+,
            $crate::Locale::shn_MM => $crate::shn_MM::$($item)::+,
            $crate::Locale::shs_CA => $crate::shs_CA::$($item)::+,
            $crate::Locale::si_LK => $crate::si_LK::$($item)::+,
            $crate::Locale::sid_ET => $crate::sid_ET::$($item)::+,
            $crate::Locale::sk_SK => $crate::sk_SK::$($item)::+,
            $crate::Locale::sl_SI => $crate::sl_SI::$($item)::+,
            $crate::Locale::sm_WS => $crate::sm_WS::$($item)::+,
            $crate::Locale::so_DJ => $crate::so_DJ::$($item)::+,
            $crate::Locale::so_ET => $crate::so_ET::$($item)::+,
            $crate::Locale::so_KE => $crate::so_KE::$($item)::+,
            $crate::Locale::so_SO => $crate::so_SO::$($item)::+,
            $crate::Locale::sq_AL => $crate::sq_AL::$($item)::+,
            $crate::Locale::sq_MK => $crate::sq_MK::$($item)::+,
            $crate::Locale::sr_ME => $crate::sr_ME::$($item)::+,
            $crate::Locale::sr_RS => $crate::sr_RS::$($item)::+,
            $crate::Locale::sr_RS_latin => $crate::sr_RS_latin::$($item)::+,
            $crate::Locale::ss_ZA => $crate::ss_ZA::$($item)::+,
            $crate::Locale::st_ZA => $crate::st_ZA::$($item)::+,
            $crate::Locale::sv_FI => $crate::sv_FI::$($item)::+,
            $crate::Locale::sv_FI_euro => $crate::sv_FI_euro::$($item)::+,
            $crate::Locale::sv_SE => $crate::sv_SE::$($item)::+,
            $crate::Locale::sw_KE => $crate::sw_KE::$($item)::+,
            $crate::Locale::sw_TZ => $crate::sw_TZ::$($item)::+,
            $crate::Locale::szl_PL => $crate::szl_PL::$($item)::+,
            $crate::Locale::ta_IN => $crate::ta_IN::$($item)::+,
            $crate::Locale::ta_LK => $crate::ta_LK::$($item)::+,
            $crate::Locale::tcy_IN => $crate::tcy_IN::$($item)::+,
            $crate::Locale::te_IN => $crate::te_IN::$($item)::+,
            $crate::Locale::tg_TJ => $crate::tg_TJ::$($item)::+,
            $crate::Locale::th_TH => $crate::th_TH::$($item)::+,
            $crate::Locale::the_NP => $crate::the_NP::$($item)::+,
            $crate::Locale::ti_ER => $crate::ti_ER::$($item)::+,
            $crate::Locale::ti_ET => $crate::ti_ET::$($item)::+,
            $crate::Locale::tig_ER => $crate::tig_ER::$($item)::+,
            $crate::Locale::tk_TM => $crate::tk_TM::$($item)::+,
            $crate::Locale::tl_PH => $crate::tl_PH::$($item)::+,
            $crate::Locale::tn_ZA => $crate::tn_ZA::$($item)::+,
            $crate::Locale::to_TO => $crate::to_TO::$($item)::+,
            $crate::Locale::tpi_PG => $crate::tpi_PG::$($item)::+,
            $crate::Locale::tr_CY => $crate::tr_CY::$($item)::+,
            $crate::Locale::tr_TR => $crate::tr_TR::$($item)::+,
            $crate::Locale::ts_ZA => $crate::ts_ZA::$($item)::+,
            $crate::Locale::tt_RU => $crate::tt_RU::$($item)::+,
            $crate::Locale::tt_RU_iqtelif => $crate::tt_RU_iqtelif::$($item)::+,
            $crate::Locale::ug_CN => $crate::ug_CN::$($item)::+,
            $crate::Locale::uk_UA => $crate::uk_UA::$($item)::+,
            $crate::Locale::unm_US => $crate::unm_US::$($item)::+,
            $crate::Locale::ur_IN => $crate::ur_IN::$($item)::+,
            $crate::Locale::ur_PK => $crate::ur_PK::$($item)::+,
            $crate::Locale::uz_UZ => $crate::uz_UZ::$($item)::+,
            $crate::Locale::uz_UZ_cyrillic => $crate::uz_UZ_cyrillic::$($item)::+,
            $crate::Locale::ve_ZA => $crate::ve_ZA::$($item)::+,
            $crate::Locale::vi_VN => $crate::vi_VN::$($item)::+,
            $crate::Locale::wa_BE => $crate::wa_BE::$($item)::+,
            $crate::Locale::wa_BE_euro => $crate::wa_BE_euro::$($item)::+,
            $crate::Locale::wae_CH => $crate::wae_CH::$($item)::+,
            $crate::Locale::wal_ET => $crate::wal_ET::$($item)::+,
            $crate::Locale::wo_SN => $crate::wo_SN::$($item)::+,
            $crate::Locale::xh_ZA => $crate::xh_ZA::$($item)::+,
            $crate::Locale::yi_US => $crate::yi_US::$($item)::+,
            $crate::Locale::yo_NG => $crate::yo_NG::$($item)::+,
            $crate::Locale::yue_HK => $crate::yue_HK::$($item)::+,
            $crate::Locale::yuw_PG => $crate::yuw_PG::$($item)::+,
            $crate::Locale::zh_CN => $crate::zh_CN::$($item)::+,
            $crate::Locale::zh_HK => $crate::zh_HK::$($item)::+,
            $crate::Locale::zh_SG => $crate::zh_SG::$($item)::+,
            $crate::Locale::zh_TW => $crate::zh_TW::$($item)::+,
            $crate::Locale::zu_ZA => $crate::zu_ZA::$($item)::+,
        }
    }}
}

impl Default for Locale {
    fn default() -> Self {
        Locale::POSIX
    }
}
